/*
 * Copyright (c) 2012, Texas Instruments Incorporated - http://www.ti.com/
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions are met:
 *       * Redistributions of source code must retain the above copyright
 *         notice, this list of conditions and the following disclaimer.
 *       * Redistributions in binary form must reproduce the above copyright
 *         notice, this list of conditions and the following disclaimer in the
 *         documentation and/or other materials provided with the distribution.
 *       * Neither the name of Texas Instruments Incorporated nor the
 *         names of its contributors may be used to endorse or promote products
 *         derived from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *   DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
 *   DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


/**
 *   @file  event_machine_private_functions.h
 *
 *   @brief
 *      Internal private functions used by the Open EM implementation for DSP.
 *      These functions are not available for use by the application.
 *
 */

#ifndef EVENT_MACHINE_PRIVATE_FUNCTIONS_DSP_H
#define EVENT_MACHINE_PRIVATE_FUNCTIONS_DSP_H

#include <stddef.h>
#include <inttypes.h>
#include <ti/runtime/openem/event_machine_macros.h>
#include <ti/runtime/openem/event_machine_group.h>
#include <ti/runtime/openem/event_machine_hw_ti_macros.h>
#include <ti/runtime/openem/event_machine_hw_ti_types.h>
#include <ti/runtime/openem/event_machine_hw_ti_functions.h>
#include <ti/runtime/openem/src/event_machine_hwpform.h>
#include <ti/runtime/openem/src/event_machine_private_macros.h>
#include <ti/runtime/openem/src/event_machine_private_types.h>
#include <ti/runtime/openem/src/event_machine_private_variables.h>

#if (EM_CHECK_LEVEL == 3)
#include <stdio.h>
#endif
/**
 * @internal
 *
 * The API is used to ensure data have been written to memory
 */
static inline void
Emti_mfence (
    void
)
{
    asm ("\tmfence\n\tnop\n\tmark 0\n\tmfence\n\tnop\n\tmark 0");
}

/**
 * @internal
 *
 * The API is used to ensure data have been written to memory after a cache coherency operation
 */
static inline void
Emti_mfenceWithCache (
    void
)
{
    asm ("\tmfence\n\tnop\n\tmark 0\n\tmfence\n\tnop\n\tmark 0\n\tnop 8\n\tnop 8");
}

/**
 * @internal
 *
 *  The function locks the semaphore to ensure critical sections
 *  across multiple cores.
 *
 * @return
 *  Not applicable.
 */
static inline void
Emti_lockSemaphore (
    void
)
{
    /*
     * Spin around till we get the semaphore. 
     */
    while ((CSL_semAcquireDirect (Emti_gvTcb.hwSemIdx)) == 0);
}

/**
 * @internal
 *
 *  The function unlocks the semaphore which had been locked
 *
 * @return
 *  Not applicable.
 */
static inline void
Emti_unlockSemaphore (
    void
)
{
    Emti_mfence ();
    /*
     * Release the semaphore. 
     */
    CSL_semReleaseSemaphore (Emti_gvTcb.hwSemIdx);
}

/**
 * @internal
 *
 *  The function is used for logging debug information messages.
 *
 *  The user can re-implement this function. But he would need
 *  to recompile the library and the objects that are using
 *  inlined Open EM functions.
 *
 * @return
 *  Not applicable.
 */
static inline void
Emti_logDbg (
    const char *fmt,
    ...
)
{
#if (EM_CHECK_LEVEL == 3)
    va_list vl;
    va_start (vl, fmt);
    vprintf (fmt, vl);
    va_end (vl);
#endif
    return;
}

/**
 * @internal
 *
 *  The function is used for logging debug error messages.
 *
 *  The user can re-implement this function. But he would need
 *  to recompile the library and the objects that are using
 *  inlined Open EM functions.
 *
 * @return
 *  Not applicable.
 */
static inline void
Emti_logError (
    const char *fmt,
    ...
)
{
#if (EM_CHECK_LEVEL == 3)
    va_list vl;
    va_start (vl, fmt);
    vprintf (fmt, vl);
    va_end (vl);
#endif
    return;
}

/**
 * @internal
 *
 * The API is used to writeback invalidate the L1D cache for the specified address and size.
 */
static inline void
Emti_cacheWbInvL1d (
    void *blockPtr,
    uint32_t byteCnt
)
{
    uint32_t lvInt = _disable_interrupts ();
    CACHE_wbInvL1d (blockPtr, byteCnt, CACHE_NOWAIT);
    Emti_mfenceWithCache ();
    _restore_interrupts (lvInt);
    return;
}

/**
 * @internal
 *
 * The API is used to writeback invalidate the L2 cache for the specified address and size.
 */
static inline void
Emti_cacheWbInvL2 (
    void *blockPtr,
    uint32_t byteCnt
)
{
    uint32_t lvInt = _disable_interrupts ();
    CACHE_wbInvL2 (blockPtr, byteCnt, CACHE_NOWAIT);
    Emti_mfenceWithCache ();
    _restore_interrupts (lvInt);
    return;
}

/**
 * @internal
 *
 * The API is used to invalidate the L1D cache for the specified address and size.
 */
static inline void
Emti_cacheInvL1d (
    void *blockPtr,
    uint32_t byteCnt
)
{
    uint32_t lvInt = _disable_interrupts ();
    CACHE_invL1d (blockPtr, byteCnt, CACHE_NOWAIT);
    Emti_mfenceWithCache ();
    _restore_interrupts (lvInt);
    return;
}

/**
 * @internal
 *
 * The API is used to invalidate the L2 cache for the specified address and size.
 */
static inline void
Emti_cacheInvL2 (
    void *blockPtr,
    uint32_t byteCnt
)
{
    uint32_t lvInt = _disable_interrupts ();
    CACHE_invL2 (blockPtr, byteCnt, CACHE_NOWAIT);
    Emti_mfenceWithCache ();
    _restore_interrupts (lvInt);
    return;
}

/**
 * @internal
 *
 * The API is used to invalidate the whole L2 cache.
 */
static inline void
Emti_cacheInvAllL2 (
    void
)
{
    CACHE_wbInvAllL2 (CACHE_WAIT);
    return;
}

/**
 * @internal
 *
 * The API is used to invalidate the XMC prefetch buffer
 */
static inline void
Emti_prefetchBufferInv (
    void
)
{
    CSL_XMC_invalidatePrefetchBuffer ();
    return;
}

/**
 * @internal
 *
 * The function invalidates the data buffer of the descriptor if required.
 *
 * @return 1 when invalidate operation has been applied, 0 otherwise.
 */
static inline uint32_t
Emti_inv (
    em_event_t event
)
{
    Cppi_HostDesc *lvHostDescPtr = Emti_getCppiHostDescPtr (event);

    ti_em_coh_mode_t lvCohMode = Emti_getCohMode (lvHostDescPtr);
    uint32_t lvDirtyFlag = Emti_getDirtyFlag (lvHostDescPtr);

    /*
     * check first buffer 
     */
    if ((lvCohMode == TI_EM_COH_MODE_ON) && (lvDirtyFlag == 1))
    {
        void *lvContentPtr;
        uint32_t lvContentsize;

        /*
         * clear buffer dirty flag 
         */
        Emti_setDirtyFlag (lvHostDescPtr, 0x0);

        /*
         * read buffer pointer from event packet descriptor (word 4) 
         */
        lvContentPtr = (void *) (lvHostDescPtr->buffPtr);

        /*
         * read buffer size from event packet descriptor (word 3, bit 0..21 
         */
        lvContentsize = lvHostDescPtr->buffLen;

        /*
         * invalidate cache for buffer 
         */
        if ((((uint32_t) (lvContentPtr)) & Emti_EXT_MEM_LOCATION))
            Emti_cacheInvL2 (lvContentPtr, lvContentsize);
        else
            Emti_cacheInvL1d (lvContentPtr, lvContentsize);

        return (1);
    }
    return (0);
}

/**
 * @internal
 *
 * The function writes back the data buffer of the descriptor if required.
 *
 * @return 1 when write back operation has been applied, 0 otherwise.
 */
static inline uint32_t
Emti_wbinv (
    em_event_t event
)
{
    Cppi_HostDesc *lvHostDescPtr = Emti_getCppiHostDescPtr (event);

    ti_em_coh_mode_t lvCohMode = Emti_getCohMode (lvHostDescPtr);
    uint32_t lvDirtyFlag = Emti_getDirtyFlag (lvHostDescPtr);

    /*
     * check cache coherency on first buffer 
     */
    if ((lvCohMode == TI_EM_COH_MODE_ON) && (lvDirtyFlag == 1))
    {
        void *lvContentPtr;
        uint32_t lvContentsize;

        /*
         * clear buffer dirty flag 
         */
        Emti_setDirtyFlag (lvHostDescPtr, 0x0);

        /*
         * read buffer pointer from event packet descriptor (word 4) 
         */
        lvContentPtr = (void *) (lvHostDescPtr->buffPtr);

        /*
         * read buffer size from event packet descriptor (word 3, bit 0..21 
         */
        lvContentsize = lvHostDescPtr->buffLen;

        /*
         * write back invalidate cache for buffer 
         */
        if ((((uint32_t) (lvContentPtr)) & Emti_EXT_MEM_LOCATION))
            Emti_cacheWbInvL2 (lvContentPtr, lvContentsize);
        else
            Emti_cacheWbInvL1d (lvContentPtr, lvContentsize);

        return (1);
    }

    return (0);
}

/**
 * @internal
 *
 * Provide physical to/from virtual descriptor address translations
 * 
 */
static inline void *
Emti_publicDesc_ptov (
    uint32_t paddr
)
{
#ifdef K2_PG1_MSMC_STALL_WA
    if ((paddr & 0xff000000) == 0x0c000000)
        return (void *)(paddr | 0x10000000);
    else
#endif /* K2_PG1_MSMC_STALL_WA */
        return (void *) paddr;
}

static inline uint32_t
Emti_publicDesc_vtop (
    void *vaddr
)
{
#ifdef K2_PG1_MSMC_STALL_WA
    if (((uint32_t)(vaddr) & 0xff000000) == 0x1c000000)
        return (uint32_t)(vaddr) & 0x0fffffff;
    else
#endif /* K2_PG1_MSMC_STALL_WA */
        return (uint32_t) vaddr;
}

static inline void *
Emti_privateDesc_ptov (
    uint32_t paddr
)
{
    return (void *) paddr;
}

static inline uint32_t
Emti_privateDesc_vtop (
    void *vaddr
)
{
    return (uint32_t) vaddr;
}

/**
 * @internal
 *
 * Provide physical to/from virtual buffer pointer address translations
 * 
 */
static inline void *
Emti_bufferPtr_ptov (
    Cppi_HostDesc * descAddr
)
{
    return (void *) descAddr->buffPtr;
}

static inline void *
Emti_origBufferPtr_ptov (
    Cppi_HostDesc * descAddr
)
{
    return (void *) descAddr->origBuffPtr;
}

static inline uint32_t
Emti_bufferPtr_vtop (
    void *vaddr,
    uint32_t poolId
)
{
    return (uint32_t) vaddr;
}

static inline uint32_t
Emti_getUncachedAlias (
    uint32_t naddr
)
{
#ifdef K2_PG1_MSMC_STALL_WA
    if ((naddr & 0xff000000) == 0x0c000000)
        return (uint32_t)(naddr | 0x10000000);
    else
#endif /* K2_PG1_MSMC_STALL_WA */
        return (uint32_t) naddr;
}

static inline uint32_t
Emti_getUncachedNative (
    uint32_t aaddr
)
{
#ifdef K2_PG1_MSMC_STALL_WA
    if (((uint32_t)(aaddr) & 0xff000000) == 0x1c000000)
        return (uint32_t)(aaddr) & 0x0fffffff;
    else
#endif /* K2_PG1_MSMC_STALL_WA */
        return (uint32_t) aaddr;
}

/**
 * @internal
 *
 * Provide physical to/from virtual QMMS register address translations
 * 
 */
static inline void *
Emti_QmssRegs_ptov (
    uint32_t paddr
)
{
    return (void *) paddr;
}

static inline uint32_t
Emti_QmssRegs_vtop (
    void *vaddr
)
{
    return (uint32_t) vaddr;
}

static inline void
Emti_queueEmpty (
    uint16_t queueIdx
)
{
    CSL_Qm_queue_managementQueue_mgmt_groupRegs *Emti_gvQmAddr;

    Emti_gvQmAddr = (void *) CSL_QMSS_CFG_QM_1_QUEUE_MANAGEMENT_REGS;

    Emti_gvQmAddr[queueIdx].QUEUE_REG_D = 0;
    return;
}

static inline void
Emti_queueDivert (
    uint16_t srcQueueIdx,
    uint16_t dstQueueIdx,
    uint8_t location
)
{
    uint32_t temp = 0;
    CSL_Qm_configRegs *Emti_gvQmAddr;
    if ((srcQueueIdx > 8191) && (dstQueueIdx > 8191))
        Emti_gvQmAddr = (CSL_Qm_configRegs *) CSL_QMSS_CFG_QM_2_CFG_REGS;
    else
        Emti_gvQmAddr = (CSL_Qm_configRegs *) CSL_QMSS_CFG_QM_1_CFG_REGS;
    CSL_FINS (temp, QM_CONFIG_QUEUE_DIVERSION_REG_SOURCE_QNUM, (srcQueueIdx & 0x1FFF));
    CSL_FINS (temp, QM_CONFIG_QUEUE_DIVERSION_REG_DEST_QNUM, (dstQueueIdx & 0x1FFF));
    CSL_FINS (temp, QM_CONFIG_QUEUE_DIVERSION_REG_HEAD_TAIL, location);

    Emti_gvQmAddr->QUEUE_DIVERSION_REG = temp;

    return;
}

static inline void
Emti_pushEvent (
    uint16_t queueIdx,
    void *eventPtr,
    short mapping
)
{
    CSL_Qm_queue_managementQueue_mgmt_groupRegs *Emti_gvQmAddr;

    Emti_gvQmAddr = (void *) CSL_QMSS_CFG_QM_1_QUEUE_MANAGEMENT_REGS;

    Emti_gvQmAddr[queueIdx].QUEUE_REG_D = (uint32_t)Emti_publicDesc_vtop((void *)eventPtr);
    return;
}

static inline void
Emti_pushEventToHead (
    uint16_t queueIdx,
    void *eventPtr,
    short mapping
)
{
    CSL_Qm_queue_managementQueue_mgmt_groupRegs *Emti_gvQmAddr;
    uint64_t lvDWord;

    Emti_gvQmAddr = (void *) CSL_QMSS_CFG_QM_1_QUEUE_PROXY_REGS;

    /*
     * build double word 
     */
#ifdef _BIG_ENDIAN
    {
        lvDWord = _itoll ((uint32_t) (Qmss_Location_HEAD << 31), (uint32_t)Emti_publicDesc_vtop((void *)eventPtr));
    }
#endif
#ifdef _LITTLE_ENDIAN
    {
        lvDWord = _itoll ((uint32_t)Emti_publicDesc_vtop((void *)eventPtr), (uint32_t) (Qmss_Location_HEAD << 31));
    }
#endif

    /*
     * push double word to QMSS 
     */
    *(uint64_t *) (&Emti_gvQmAddr[queueIdx].QUEUE_REG_C) = lvDWord;

    return;
}


static inline void *
Emti_popEvent (
    uint16_t queueIdx,
    short mapping
)
{
    CSL_Qm_queue_managementQueue_mgmt_groupRegs *Emti_gvQmAddr;

    Emti_gvQmAddr = (void *) CSL_QMSS_CFG_QM_1_QUEUE_MANAGEMENT_REGS;

    return (void *)Emti_publicDesc_ptov((uint32_t)Emti_gvQmAddr[queueIdx].QUEUE_REG_D);
}

static inline void
Emti_setQueueThreshold (
    uint16_t queueIdx,
    uint16_t hilo,
    uint8_t threshold
)
{
    CSL_Qm_queue_status_configRegs *Emti_gvQmAddr;

    Emti_gvQmAddr = (void *) CSL_QMSS_CFG_QM_1_QUEUE_PEEK_REGS;

    Emti_gvQmAddr->QUEUE_STATUS_CONFIG_GROUP[queueIdx].QUEUE_STATUS_CONFIG_REG_D =
        (threshold | (hilo << CSL_QM_QUEUE_STATUS_CONFIG_QUEUE_STATUS_CONFIG_REG_D_THRESHOLD_HILO_SHIFT));
    return;
}

static inline uint32_t
Emti_getQueueEntryCount (
    uint16_t queueIdx
)
{
    CSL_Qm_queue_status_configRegs *Emti_gvQmAddr;

    Emti_gvQmAddr = (void *) CSL_QMSS_CFG_QM_1_QUEUE_PEEK_REGS;

    return (uint32_t) (Emti_gvQmAddr->QUEUE_STATUS_CONFIG_GROUP[queueIdx].QUEUE_STATUS_CONFIG_REG_A);
}

static inline uint32_t
Emti_getPdspDataRamBaseAddress (
    uint8_t pdspIdx
)
{
    return (uint32_t) (Emti_CD_QUEUE_IDX_REG_ADD_PDSP1 +
                       pdspIdx * (Emti_CD_QUEUE_IDX_REG_ADD_PDSP2 - Emti_CD_QUEUE_IDX_REG_ADD_PDSP1));
}

/*
 * Put CPU memory to PDSP (nothing to do on DSP)
 */
static inline uint32_t
Emti_pdspPut (
    uint32_t slot,
    void *ptr,
    uint32_t size
)
{
#ifdef K2_PG1_MSMC_STALL_WA
    if (((uint32_t)(ptr) & 0xff000000) == 0x1c000000)
        return (uint32_t)(ptr) & 0x0fffffff;
    else
#endif /* K2_PG1_MSMC_STALL_WA */
        return (uint32_t) ptr;
}

/**
 * @internal
 *
 * Get PDSP memory to CPU  (nothing to do on DSP)
 *
 */
static inline void
Emti_pdspGet (
    uint32_t slot,
    void *ptr,
    uint32_t size
)
{
}

/**
 * @internal
 *
 * Map CPU memory to PDSP communication memory
 *
 */
static inline uint32_t
Emti_pdspMap (
    uint32_t slot,
    void **ptr
)
{
    return ((uint32_t) * ptr | 0x10000000 | (DNUM << 24));
}

/**
 * @internal
 *
 * Initialize internal shared memory
 *
 */
static inline int
Emti_sharedMemInit (
    unsigned int size,
    int init
)
{
    return 0;
}

/**
 * @internal
 *
 * Release internal shared memory
 *
 */
static inline void
Emti_sharedMemRelease (
    void
)
{
}

/**
 * @internal
 *
 * Allocate shared memory
 *
 */
static inline void *
Emti_sharedMemAlloc (
    unsigned int size
)
{
    return NULL;
}

/**
 * @internal
 *
 * Free shared memory
 *
 */
static inline void
Emti_sharedMemFree (
    void *ptr
)
{
}

#endif /* EVENT_MACHINE_PRIVATE_FUNCTIONS_DSP_H */
