/*
 * Copyright (c) 2012, Texas Instruments Incorporated - http://www.ti.com/
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions are met:
 *       * Redistributions of source code must retain the above copyright
 *         notice, this list of conditions and the following disclaimer.
 *       * Redistributions in binary form must reproduce the above copyright
 *         notice, this list of conditions and the following disclaimer in the
 *         documentation and/or other materials provided with the distribution.
 *       * Neither the name of Texas Instruments Incorporated nor the
 *         names of its contributors may be used to endorse or promote products
 *         derived from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *   DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
 *   DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


/**
 *   @file  em_pdk_hal.c
 *
 *   @brief   
 *      This file contains the routines to abstract some of CPPI/QMSS drivers functions
 *      for the event machine.
 *
 *  \par
 *  ============================================================================
 *  @n   (C) Copyright 2012, Texas Instruments, Inc.
 *  ============================================================================ 
 *  \par
 */
#include <string.h>

/* importing some of the Open EM defined symbols */
#include "my_event_machine.h"

#include <ti/drv/cppi/cppi_drv.h>
#include <ti/drv/cppi/cppi_desc.h>
#include <ti/drv/cppi/include/cppi_pvt.h>
#include <ti/drv/qmss/qmss_drv.h>
/*
 * Arrays to store the CPPI PDK handles
 */
Cppi_FlowHnd flowHandleArray[2][64];
Cppi_ChHnd rxChannelHandleArray[2][32];
Cppi_ChHnd txChannelHandleArray[2][32];

/*
 **********************************************************************************************************************
 **********************************************************************************************************************
 */

em_status_t
ti_em_hw_queue_open (
    int queue_idx
)
{
    Qmss_QueueType lvQueueType;
    Qmss_QueueHnd lvQueueHandle;
    uint16_t lvRelativeQueueIdx = queue_idx & 0x1FFF;
    uint8_t lvIsAllocated;

    if ((lvRelativeQueueIdx >= (QMSS_INFRASTRUCTURE_QUEUE_BASE)) &&
        (lvRelativeQueueIdx < (QMSS_INFRASTRUCTURE_QUEUE_BASE + QMSS_MAX_INFRASTRUCTURE_QUEUE)))
        lvQueueType = Qmss_QueueType_INFRASTRUCTURE_QUEUE;
    else
        lvQueueType = Qmss_QueueType_GENERAL_PURPOSE_QUEUE;
    lvQueueHandle = Qmss_queueOpen (lvQueueType, queue_idx, &lvIsAllocated);
    // The queue shall not be already used by another application.
    if ((lvQueueHandle < 0) || (lvIsAllocated != 1))
        return (em_status_t) EM_ERR;
    else
        return (em_status_t) EM_OK;
}

/*
 **********************************************************************************************************************
 **********************************************************************************************************************
 */

em_status_t
ti_em_hw_queue_close (
    int queue_idx
)
{
    Qmss_Result lvRes;
    // Empty the queue
    Qmss_queueEmpty ((Qmss_QueueHnd) queue_idx);
    // Close the queue
    lvRes = Qmss_queueClose ((Qmss_QueueHnd) queue_idx);
    if (lvRes != 0)
        return (em_status_t) EM_ERR;
    else
        return (em_status_t) EM_OK;
}

static inline int8_t
ti_em_retrieve_array_index (
    Cppi_CpDma dmaNum
)
{
    int8_t lvArrayIdx;

    switch (dmaNum)
    {
        case Cppi_CpDma_QMSS_CPDMA:
            lvArrayIdx = 0;
            break;
#ifdef TI_EM_C6634
        case Cppi_CpDma_QMSS_QM2_CPDMA:
            lvArrayIdx = 2;
            break;
#endif
        default:
            /*
             * Invalid CPPI
             */
            lvArrayIdx = -1;
            break;
    }
    return lvArrayIdx;
}

/*
 **********************************************************************************************************************
 **********************************************************************************************************************
 */

em_status_t
ti_em_rx_flow_open (
    int dma_idx,
    int flow_idx,
    int dst_queue_idx,
    int free_queue_idx,
    int error_handling
)
{
    Cppi_RxFlowCfg lvRxFlowCfg;
    Cppi_DMAObj *dmaObjPtr = (Cppi_DMAObj *) dma_idx;
    Cppi_FlowHnd lvRxFlowHnd;
    uint8_t lvIsAllocated;
    int8_t lvArrayIdx;

    lvArrayIdx = ti_em_retrieve_array_index (dmaObjPtr->dmaNum);
    if (lvArrayIdx == -1)
        return (em_status_t) EM_ERR;

    lvIsAllocated = 0;
    memset ((void *) &lvRxFlowCfg, 0, sizeof (Cppi_RxFlowCfg));

    lvRxFlowCfg.flowIdNum = flow_idx;
    lvRxFlowCfg.rx_dest_qnum = (dst_queue_idx & 0xFFF);
    lvRxFlowCfg.rx_dest_qmgr = (dst_queue_idx >> 12);
    lvRxFlowCfg.rx_desc_type = Cppi_DescType_HOST;
    /*
     * Mandatory parameters to configure the RX FLOW 
     */
    lvRxFlowCfg.rx_einfo_present = 1;
    lvRxFlowCfg.rx_dest_tag_lo_sel = 4;
    lvRxFlowCfg.rx_dest_tag_hi_sel = 5;
    lvRxFlowCfg.rx_src_tag_hi_sel = 4;
    lvRxFlowCfg.rx_psinfo_present = 1;
    lvRxFlowCfg.rx_ps_location = 0;
    lvRxFlowCfg.rx_error_handling = error_handling;
    lvRxFlowCfg.rx_fdq0_sz0_qnum = (free_queue_idx & 0xFFF);
    lvRxFlowCfg.rx_fdq0_sz0_qmgr = (free_queue_idx >> 12);
    lvRxFlowCfg.rx_fdq1_qnum = (free_queue_idx & 0xFFF);
    lvRxFlowCfg.rx_fdq1_qmgr = (free_queue_idx >> 12);
    lvRxFlowCfg.rx_fdq2_qnum = (free_queue_idx & 0xFFF);
    lvRxFlowCfg.rx_fdq2_qmgr = (free_queue_idx >> 12);
    lvRxFlowCfg.rx_fdq3_qnum = (free_queue_idx & 0xFFF);
    lvRxFlowCfg.rx_fdq3_qmgr = (free_queue_idx >> 12);
    /*
     * Other fields are left to be modified by user 
     */

    /*
     * Configure Rx flow 
     */
    lvRxFlowHnd = Cppi_configureRxFlow ((Cppi_Handle) dma_idx, &lvRxFlowCfg, &lvIsAllocated);

    /*
     * Store the handle for future usage
     */
    flowHandleArray[lvArrayIdx][flow_idx] = lvRxFlowHnd;

    // The RX Flow shall not be already used by another application.
    if ((lvRxFlowHnd == NULL) || (lvIsAllocated != 1))
        return (em_status_t) EM_ERR;
    else
        return (em_status_t) EM_OK;
}

/*
 **********************************************************************************************************************
 **********************************************************************************************************************
 */

em_status_t
ti_em_rx_flow_close (
    int dma_idx,
    int flow_idx
)
{
    Cppi_DMAObj *dmaObjPtr = (Cppi_DMAObj *) dma_idx;
    Cppi_FlowHnd lvRxFlowHnd;
    Cppi_Result lvRes;
    int8_t lvArrayIdx;

    lvArrayIdx = ti_em_retrieve_array_index (dmaObjPtr->dmaNum);
    if (lvArrayIdx == -1)
        return (em_status_t) EM_ERR;

    lvRxFlowHnd = flowHandleArray[lvArrayIdx][flow_idx];

    lvRes = Cppi_closeRxFlow (lvRxFlowHnd);

    if (lvRes != 0)
        return (em_status_t) EM_ERR;
    else
        return (em_status_t) EM_OK;
}

/*
 **********************************************************************************************************************
 **********************************************************************************************************************
 */

em_status_t
ti_em_xge_rx_flow_open (
    int dma_idx,
    int flow_idx,
    int dst_queue_idx,
    int free_queue_idx0,
    int free_queue_idx1,
    int error_handling
)
{
    return (em_status_t) EM_OK;
}

em_status_t
ti_em_xge_rx_flow_close (
    int dma_idx,
    int flow_idx
)
{
    return (em_status_t) EM_OK;
}

/*
 **********************************************************************************************************************
 **********************************************************************************************************************
 */

em_status_t
ti_em_chain_rx_flow_open (
    int dma_idx,
    int flow_idx,
    int dst_queue_idx,
    int free_queue_idxSizeA,
    int sizeA,
    int free_queue_idxSizeB,
    int sizeB,
    int free_queue_idxSizeC,
    int sizeC,
    int free_queue_idxSizeD,
    int sizeD,
    int free_queue_idxOverflow,
    int error_handling
)
{
    Cppi_RxFlowCfg lvRxFlowCfg;
    Cppi_DMAObj *dmaObjPtr = (Cppi_DMAObj *) dma_idx;
    Cppi_FlowHnd lvRxFlowHnd;
    uint8_t lvIsAllocated;
    int8_t lvArrayIdx;

    lvArrayIdx = ti_em_retrieve_array_index (dmaObjPtr->dmaNum);
    if (lvArrayIdx == -1)
        return (em_status_t) EM_ERR;

    lvIsAllocated = 0;
    memset ((void *) &lvRxFlowCfg, 0, sizeof (Cppi_RxFlowCfg));

    lvRxFlowCfg.flowIdNum = flow_idx;
    lvRxFlowCfg.rx_dest_qnum = (dst_queue_idx & 0xFFF);
    lvRxFlowCfg.rx_dest_qmgr = (dst_queue_idx >> 12);
    lvRxFlowCfg.rx_desc_type = Cppi_DescType_HOST;
    /*
     * Mandatory parameters to configure the RX FLOW 
     */
    lvRxFlowCfg.rx_einfo_present = 1;
    lvRxFlowCfg.rx_dest_tag_lo_sel = 4;
    lvRxFlowCfg.rx_dest_tag_hi_sel = 5;
    lvRxFlowCfg.rx_src_tag_hi_sel = 4;
    lvRxFlowCfg.rx_psinfo_present = 1;
    lvRxFlowCfg.rx_ps_location = 0;
    lvRxFlowCfg.rx_error_handling = error_handling;
    lvRxFlowCfg.rx_fdq0_sz0_qnum = (free_queue_idxSizeA & 0xFFF);
    lvRxFlowCfg.rx_fdq0_sz0_qmgr = (free_queue_idxSizeA >> 12);
    lvRxFlowCfg.rx_fdq0_sz1_qnum = (free_queue_idxSizeB & 0xFFF);
    lvRxFlowCfg.rx_fdq0_sz1_qmgr = (free_queue_idxSizeB >> 12);
    lvRxFlowCfg.rx_fdq0_sz2_qnum = (free_queue_idxSizeC & 0xFFF);
    lvRxFlowCfg.rx_fdq0_sz2_qmgr = (free_queue_idxSizeC >> 12);
    lvRxFlowCfg.rx_fdq0_sz3_qnum = (free_queue_idxSizeD & 0xFFF);
    lvRxFlowCfg.rx_fdq0_sz3_qmgr = (free_queue_idxSizeD >> 12);
    lvRxFlowCfg.rx_fdq1_qnum = (free_queue_idxOverflow & 0xFFF);
    lvRxFlowCfg.rx_fdq1_qmgr = (free_queue_idxOverflow >> 12);
    lvRxFlowCfg.rx_fdq2_qnum = (free_queue_idxOverflow & 0xFFF);
    lvRxFlowCfg.rx_fdq2_qmgr = (free_queue_idxOverflow >> 12);
    lvRxFlowCfg.rx_fdq3_qnum = (free_queue_idxOverflow & 0xFFF);
    lvRxFlowCfg.rx_fdq3_qmgr = (free_queue_idxOverflow >> 12);

    /* Check if thresholds need to be used and set them accordingly to queue sizes */
    if (free_queue_idxSizeB == free_queue_idxSizeA)
	goto skip_threshold;

    lvRxFlowCfg.rx_size_thresh0_en = 1;
    lvRxFlowCfg.rx_size_thresh0    = sizeA;
    
    if (free_queue_idxSizeC == free_queue_idxSizeB)
	goto skip_threshold;
    
    lvRxFlowCfg.rx_size_thresh1_en = 1;
    lvRxFlowCfg.rx_size_thresh1    = sizeB;
    
    if (free_queue_idxSizeD == free_queue_idxSizeC)
	goto skip_threshold;
    
    lvRxFlowCfg.rx_size_thresh2_en = 1;
    lvRxFlowCfg.rx_size_thresh2    = sizeC;

skip_threshold:

    /*
     * Configure Rx flow 
     */
    lvRxFlowHnd = Cppi_configureRxFlow ((Cppi_Handle) dma_idx, &lvRxFlowCfg, &lvIsAllocated);

    /*
     * Store the handle for future usage
     */
    flowHandleArray[lvArrayIdx][flow_idx] = lvRxFlowHnd;

    // The RX Flow shall not be already used by another application.
    if ((lvRxFlowHnd == NULL) || (lvIsAllocated != 1))
        return (em_status_t) EM_ERR;
    else
        return (em_status_t) EM_OK;
}

/*
 **********************************************************************************************************************
 **********************************************************************************************************************
 */

em_status_t
ti_em_rio_rx_flow_open (
    int dma_idx,
    int flow_idx,
    int dst_queue_idx,
    int free_queue_idxSizeA,
    int sizeA,
    int free_queue_idxSizeB,
    int sizeB,
    int free_queue_idxSizeC,
    int sizeC,
    int free_queue_idxSizeD,
    int sizeD,
    int free_queue_idxOverflow,
    int deviceIdx,
    int processIdx,
    int error_handling
)
{
    return (em_status_t) EM_OK;
}

/*
 **********************************************************************************************************************
 **********************************************************************************************************************
 */

em_status_t
ti_em_rio_tx_channel_open (
    int dma_idx
)
{

    return (em_status_t) EM_OK;
}

/*
 **********************************************************************************************************************
 **********************************************************************************************************************
 */

ti_em_queue_id_t
ti_em_rio_tx_queue_open (
    int rio_tx_queue_idx
)
{
    return (ti_em_queue_id_t) 0;
}

/*
 **********************************************************************************************************************
 **********************************************************************************************************************
 */

em_status_t
ti_em_tx_channel_open (
    int dma_idx,
    int channel_idx
)
{
    Cppi_TxChInitCfg lvTxChCfg;
    Cppi_DMAObj *dmaObjPtr = (Cppi_DMAObj *) dma_idx;
    Cppi_ChHnd lvTxChHnd;
    uint8_t lvIsAllocated;
    int8_t lvArrayIdx;

    lvArrayIdx = ti_em_retrieve_array_index (dmaObjPtr->dmaNum);
    if (lvArrayIdx == -1)
        return (em_status_t) EM_ERR;

    memset ((void *) &lvTxChCfg, 0, sizeof (Cppi_TxChInitCfg));

    lvTxChCfg.channelNum = channel_idx;
    lvTxChCfg.priority = 2;
    lvTxChCfg.txEnable = Cppi_ChState_CHANNEL_ENABLE;

    /*
     * Open Tx Channel 
     */
    lvTxChHnd = (Cppi_ChHnd) Cppi_txChannelOpen ((Cppi_Handle) dma_idx, &lvTxChCfg, &lvIsAllocated);

    txChannelHandleArray[lvArrayIdx][channel_idx] = lvTxChHnd;

    // The TX Channel shall not be already used by another application.
    if ((lvTxChHnd == NULL) || (lvIsAllocated != 1))
        return (em_status_t) EM_ERR;
    else
        return (em_status_t) EM_OK;
}

/*
 **********************************************************************************************************************
 **********************************************************************************************************************
 */

em_status_t
ti_em_tx_channel_close (
    int dma_idx,
    int channel_idx
)
{
    Cppi_DMAObj *dmaObjPtr = (Cppi_DMAObj *) dma_idx;
    Cppi_ChHnd lvTxChHnd;
    Cppi_Result lvRes;
    int8_t lvArrayIdx;

    lvArrayIdx = ti_em_retrieve_array_index (dmaObjPtr->dmaNum);
    if (lvArrayIdx == -1)
        return (em_status_t) EM_ERR;

    lvTxChHnd = txChannelHandleArray[lvArrayIdx][channel_idx];

    lvRes = Cppi_channelClose (lvTxChHnd);

    if (lvRes != 0)
        return (em_status_t) EM_ERR;
    else
        return (em_status_t) EM_OK;
}

/*
 **********************************************************************************************************************
 **********************************************************************************************************************
 */

em_status_t
ti_em_xge_tx_channel_open (
    int dma_idx,
    int channel_idx
)
{
    return (em_status_t) EM_OK;
}

em_status_t
ti_em_xge_tx_channel_close (
    int dma_idx,
    int channel_idx
)
{
    return (em_status_t) EM_OK;
}

/*
 **********************************************************************************************************************
 **********************************************************************************************************************
 */

em_status_t
ti_em_rx_channel_open (
    int dma_idx,
    int channel_idx
)
{
    Cppi_RxChInitCfg lvRxChCfg;
    Cppi_ChHnd lvRxChHnd;
    Cppi_DMAObj *dmaObjPtr = (Cppi_DMAObj *) dma_idx;
    uint8_t lvIsAllocated;
    int8_t lvArrayIdx;

    lvArrayIdx = ti_em_retrieve_array_index (dmaObjPtr->dmaNum);
    if (lvArrayIdx == -1)
        return (em_status_t) EM_ERR;

    memset ((void *) &lvRxChCfg, 0, sizeof (Cppi_RxChInitCfg));

    lvRxChCfg.channelNum = channel_idx;
    lvRxChCfg.rxEnable = Cppi_ChState_CHANNEL_ENABLE;

    /*
     * Open Rx Channel 
     */
    lvRxChHnd = Cppi_rxChannelOpen ((Cppi_Handle) dma_idx, &lvRxChCfg, &lvIsAllocated);

    rxChannelHandleArray[lvArrayIdx][channel_idx] = lvRxChHnd;

    // The RX Channel shall not be already used by another application.
    if ((lvRxChHnd == NULL) || (lvIsAllocated != 1))
        return (em_status_t) EM_ERR;
    else
        return (em_status_t) EM_OK;
}

/*
 **********************************************************************************************************************
 **********************************************************************************************************************
 */

em_status_t
ti_em_rx_channel_close (
    int dma_idx,
    int channel_idx
)
{
    Cppi_DMAObj *dmaObjPtr = (Cppi_DMAObj *) dma_idx;
    Cppi_ChHnd lvRxChHnd;
    Cppi_Result lvRes;
    int8_t lvArrayIdx;

    lvArrayIdx = ti_em_retrieve_array_index (dmaObjPtr->dmaNum);
    if (lvArrayIdx == -1)
        return (em_status_t) EM_ERR;

    lvRxChHnd = rxChannelHandleArray[lvArrayIdx][channel_idx];

    lvRes = Cppi_channelClose (lvRxChHnd);

    if (lvRes != 0)
        return (em_status_t) EM_ERR;
    else
        return (em_status_t) EM_OK;
}

/*
 **********************************************************************************************************************
 **********************************************************************************************************************
 */

em_status_t
ti_em_xge_rx_channel_open (
    int dma_idx,
    int channel_idx
)
{
    return (em_status_t) EM_OK;
}

em_status_t
ti_em_xge_rx_channel_close (
    int dma_idx,
    int channel_idx
)
{
    return (em_status_t) EM_OK;
}

/*
 **********************************************************************************************************************
 **********************************************************************************************************************
 */

ti_em_queue_id_t
ti_em_xge_tx_queue_base_idx_get (
)
{
    return (ti_em_queue_id_t) 0;
}

ti_em_queue_id_t
ti_em_xge_tx_queue_open (
    ti_em_queue_id_t queue_base_idx,
    int vlan_priority
)
{
    return (ti_em_queue_id_t) 0;
}

/*
 **********************************************************************************************************************
 **********************************************************************************************************************
 */

ti_em_queue_id_t
ti_em_get_absolute_queue_id (
    ti_em_dma_id_t dma_idx,
    ti_em_queue_id_t queue_idx
)
{
    Cppi_DMAObj *dmaObjPtr = (Cppi_DMAObj *) dma_idx;

    if (dmaObjPtr->dmaNum == Cppi_CpDma_QMSS_CPDMA)
        return (ti_em_queue_id_t) (queue_idx + QMSS_INFRASTRUCTURE_QUEUE_BASE);
    else
        return (ti_em_queue_id_t) (queue_idx + QMSS_INFRASTRUCTURE_QUEUE_BASE + 8192);
}
