/*
 * Copyright (c) 2012, Texas Instruments Incorporated - http://www.ti.com/
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions are met:
 *       * Redistributions of source code must retain the above copyright
 *         notice, this list of conditions and the following disclaimer.
 *       * Redistributions in binary form must reproduce the above copyright
 *         notice, this list of conditions and the following disclaimer in the
 *         documentation and/or other materials provided with the distribution.
 *       * Neither the name of Texas Instruments Incorporated nor the
 *         names of its contributors may be used to endorse or promote products
 *         derived from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *   DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
 *   DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


// This file implements additional functions, types, variables and constants to be used for enabling the event machine
// The implementations are likely application specific.

#include <stdio.h>
#include "my_receive.h"
#include "my_util.h"
#include "my_init.h"
#include "my_em_init.h"

#include <ti/csl/csl_cacheAux.h>

extern cregister volatile uint32_t DNUM;

#define MY_FLOW_NUM (4)
#define MY_JOB_NUM_PER_FLOW (8)
#define MY_JOB_NUM (MY_FLOW_NUM*MY_JOB_NUM_PER_FLOW)

#define MY_FFT_SIZE_MIN (1024)
#define MY_FFT_SIZE_MAX (8*1024)

#define MY_JOB_HDR_SIZE (CACHE_L2_LINESIZE)
#define MY_JOB_PLD_SIZE (2*MY_FFT_SIZE_MAX*sizeof(int16_t))
#define MY_JOB_SIZE (MY_JOB_HDR_SIZE+MY_JOB_PLD_SIZE)

typedef struct my_JobDsc_
{
    uint32_t jobIdx;
    uint32_t flowIdx;
    uint32_t fftSize;
    uint32_t dataSize;
    uint32_t successFlag;
    uint32_t coreIdx;
    uint32_t startTime;
    uint32_t stopTime;
} my_JobDsc;

#pragma DATA_ALIGN   (my_svRunningJobNum, CACHE_L2_LINESIZE)
#pragma DATA_SECTION (my_svRunningJobNum, ".my_svL1Mem");
uint32_t my_svRunningJobNum;
#pragma DATA_ALIGN   (my_svJobDscTbl, CACHE_L2_LINESIZE)
#pragma DATA_SECTION (my_svJobDscTbl, ".my_svL1Mem");
my_JobDsc my_svJobDscTbl[MY_JOB_NUM];

int
my_initJobs (
    void
)
{
    uint32_t i, j;

    if (MY_JOB_NUM > 2 * MY_EM_EVENT_NUM)
        return -1;
    if (MY_FLOW_NUM > MY_EM_PROC_QUEUE_NUM)
        return -1;
    if (MY_JOB_SIZE > MY_EM_EVENT_BUF_SIZE)
        return -1;
    if (sizeof (my_JobDsc) > MY_JOB_HDR_SIZE)
        return -1;

    for (i = 0; i < MY_FLOW_NUM; i++)
    {
        for (j = 0; j < MY_JOB_NUM_PER_FLOW; j++)
        {
            uint32_t lvJobIdx;
            uint32_t lvFftSize;

            lvFftSize = MY_FFT_SIZE_MIN;

            lvJobIdx = i * MY_JOB_NUM_PER_FLOW + j;
            my_svJobDscTbl[lvJobIdx].jobIdx = lvJobIdx;
            my_svJobDscTbl[lvJobIdx].flowIdx = i;
            my_svJobDscTbl[lvJobIdx].fftSize = lvFftSize;
            my_svJobDscTbl[lvJobIdx].dataSize = MY_JOB_HDR_SIZE + 2 * lvFftSize * sizeof (int16_t);
            my_svJobDscTbl[lvJobIdx].successFlag = 0;
            my_svJobDscTbl[lvJobIdx].coreIdx = MY_EM_CORE_NUM;
            my_svJobDscTbl[lvJobIdx].startTime = 0;
            my_svJobDscTbl[lvJobIdx].stopTime = 0;
        }
    }

    return 0;
}

void
my_fill_buffer (
    uint8_t * buffer_ptr,
    uint32_t seed,
    uint32_t size
)
{
    uint32_t *lvPtr = (uint32_t *) buffer_ptr;
    uint32_t i;
    for (i = 0; i < (size >> 2); i++)
    {
        uint32_t word = (seed << 20) + i;
        *lvPtr++ = word;
    }
}

uint32_t
my_check_buffer (
    uint8_t * buffer_ptr,
    uint32_t seed,
    uint32_t size
)
{
    uint32_t *lvPtr = (uint32_t *) buffer_ptr;
    uint32_t i;
    uint32_t nb_mismatches = 0;
    for (i = 0; i < (size >> 2); i++)
    {
        uint32_t word = (seed << 20) + i;
        if (*lvPtr++ != word)
            nb_mismatches++;
    }
    return nb_mismatches;
}

int
my_sourceJobs (
    void
)
{
    uint32_t i;
    uint32_t nb_starvation = 0;

    /*
     * initialize job counter 
     */
    my_svRunningJobNum = MY_JOB_NUM;

    /*
     * loop over all jobs to create 
     */
    for (i = 0; i < MY_JOB_NUM; i++)
    {
        volatile em_event_t lvEventHdl;
        uint8_t *lvJobBufPtr;
        my_JobDsc *lvJobDscPtr;
        uint32_t lvFlowIdx;
        uint32_t lvFftSize;
        uint32_t lvDataSize;
        uint32_t lvSuccessFlag;

        lvFlowIdx = my_svJobDscTbl[i].flowIdx;
        lvFftSize = my_svJobDscTbl[i].fftSize;
        lvDataSize = my_svJobDscTbl[i].dataSize;
        lvSuccessFlag = 1;


        /*
         * allocate event 
         */
        lvEventHdl = EM_EVENT_UNDEF;
        do
            lvEventHdl = em_alloc (lvDataSize, MY_EM_PROC_EVENT_TYPE, MY_EM_POOL_IDX);
        while (lvEventHdl == EM_EVENT_UNDEF);

        if (lvEventHdl == NULL)
            return -1;

        /*
         * acess event buffer 
         */
        lvJobBufPtr = (uint8_t *) em_event_pointer (lvEventHdl);

        /*
         * set ctrl information 
         */
        lvJobDscPtr = (my_JobDsc *) lvJobBufPtr;
        lvJobDscPtr->jobIdx = i;
        lvJobDscPtr->flowIdx = lvFlowIdx;
        lvJobDscPtr->fftSize = lvFftSize;
        lvJobDscPtr->dataSize = lvDataSize;
        lvJobDscPtr->successFlag = lvSuccessFlag;

        /*
         * set FFT input data 
         */
        lvJobBufPtr = lvJobBufPtr + MY_JOB_HDR_SIZE;
        my_fill_buffer (lvJobBufPtr, i, lvDataSize - MY_JOB_HDR_SIZE);

        /*
         * send event to EQs 
         */
        {
            em_queue_t lvD1QueueHandle;
            em_status_t send_state;
            lvD1QueueHandle = my_em_svProcQueueHdlTbl[lvFlowIdx] + 0x1000000;
            send_state = em_send (lvEventHdl, lvD1QueueHandle);
            while (send_state != EM_OK)
            {
                nb_starvation++;
                send_state = em_send (lvEventHdl, lvD1QueueHandle);
            }
        }
    }
    printf ("nb_starvation = %u\n", nb_starvation);

    return 0;
}

void
my_processJob (
    void *eoCtxtPtr,
    em_event_t eventHdl,
    em_event_type_t eventType,
    em_queue_t queueHdl,
    void *queueCtxtPtr
)
{
    uint8_t *lvJobBufPtr;
    my_JobDsc *lvJobDscPtr;
    volatile em_event_t lvOutputEventHdl;
    uint32_t lvJobIdx;
    uint32_t lvFftSize;
    uint32_t lvDataSize;
    uint32_t lvSuccessFlag;
    uint32_t lvStartTime;
    uint32_t lvStopTime;

    /*
     * prefetch request 
     */
    ti_em_preschedule ();

    /*
     * access event buffer of received event 
     */
    lvJobBufPtr = (uint8_t *) em_event_pointer (eventHdl);
    lvJobDscPtr = (my_JobDsc *) lvJobBufPtr;
    lvJobBufPtr = lvJobBufPtr + MY_JOB_HDR_SIZE;

    /*
     * retrieve ctrl information from received
     * event buffer 
     */
    lvJobIdx = lvJobDscPtr->jobIdx;
    lvFftSize = lvJobDscPtr->fftSize;
    lvDataSize = lvJobDscPtr->dataSize;
    lvSuccessFlag = lvJobDscPtr->successFlag;


    /*
     * allocate a new event for storing results
     */
    lvOutputEventHdl = EM_EVENT_UNDEF;
    do
        lvOutputEventHdl = em_alloc (lvDataSize, TI_EM_EVENT_TYPE_PRELOAD_OFF, MY_EM_POOL_IDX);
    while (lvOutputEventHdl == EM_EVENT_UNDEF);

    /*
     * access event buffer of new event 
     */
    lvJobBufPtr = (uint8_t *) em_event_pointer (lvOutputEventHdl);
    lvJobDscPtr = (my_JobDsc *) lvJobBufPtr;
    lvJobBufPtr = lvJobBufPtr + MY_JOB_HDR_SIZE;

    /*
     * fill new event buffer with ctrl information 
     */
    lvJobDscPtr->jobIdx = lvJobIdx;
    lvJobDscPtr->fftSize = lvFftSize;
    lvJobDscPtr->dataSize = lvDataSize;

    /*
     * get start time for statistics 
     */
    lvStartTime = my_readClockGlobal ();

    /*
     * The buffer should be not scattered anymore! 
     */
    {
        uint8_t *buff_ptr;
        size_t buff_size = ti_em_buffer_size (eventHdl);
        size_t left_size = ti_em_event_size (eventHdl);
        uint32_t i;

        if (buff_size != left_size)
            lvSuccessFlag = 0;

        buff_ptr = (uint8_t *) em_event_pointer (eventHdl);
        buff_ptr += MY_JOB_HDR_SIZE;

        for (i = MY_JOB_HDR_SIZE; i < buff_size; i++)
            *lvJobBufPtr++ = *buff_ptr++;

    }
    lvJobDscPtr->successFlag = lvSuccessFlag;


    /*
     * get stop time for statistics 
     */
    lvStopTime = my_readClockGlobal ();

    /*
     * store information for statistics 
     */
    my_svJobDscTbl[lvJobIdx].coreIdx = DNUM;
    my_svJobDscTbl[lvJobIdx].startTime = lvStartTime;
    my_svJobDscTbl[lvJobIdx].stopTime = lvStopTime;

    /*
     * free received event 
     */
    em_free (eventHdl);

    /*
     * send new event 
     */
    em_send (lvOutputEventHdl, my_em_svSinkQueueHdl);
}

void
my_sinkJob (
    void *eoCtxtPtr,
    em_event_t eventHdl,
    em_event_type_t eventType,
    em_queue_t queueHdl,
    void *queueCtxtPtr
)
{
    uint8_t *lvJobBufPtr;
    my_JobDsc *lvJobDscPtr;
    uint32_t lvJobIdx;
    uint32_t lvDataSize;

    /*
     * prefetch request 
     */
    ti_em_preschedule ();

    /*
     * access event buffer of received event 
     */
    lvJobBufPtr = (uint8_t *) em_event_pointer (eventHdl);
    lvJobDscPtr = (my_JobDsc *) lvJobBufPtr;
    lvJobBufPtr = lvJobBufPtr + MY_JOB_HDR_SIZE;

    /*
     * retrieve ctrl information from received
     * event buffer 
     */
    lvJobIdx = lvJobDscPtr->jobIdx;
    lvDataSize = lvJobDscPtr->dataSize;

    if (my_check_buffer (lvJobBufPtr, lvJobIdx, lvDataSize - MY_JOB_HDR_SIZE) != 0)
        my_svJobDscTbl[lvJobIdx].successFlag = 0;
    else
        my_svJobDscTbl[lvJobIdx].successFlag = 1;

    /*
     * free received event 
     */
    em_free (eventHdl);

    /*
     * decrement shared job counter 
     */
    my_lock ();
    my_svRunningJobNum--;
    my_unlock ();
}

int
my_statsJobs (
    void
)
{
    uint32_t lvJobIdx;
    uint32_t lvFlowIdx;
    uint32_t lvCoreIdx;
    uint32_t lvJobNum = 0;
    uint32_t lvMinStartTime = 0xffffffff;
    uint32_t lvMaxStopTime = 0;
    uint32_t lvRunTime = 0;
    uint32_t lvSuccessFlag = 1;
    uint32_t lvProcTime = 0;
    uint32_t lvStartTime;
    uint32_t lvStopTime;
    uint32_t lvMinProcTime = 0xffffffff;
    uint32_t lvMaxProcTime = 0;
    uint32_t lvMeanProcTime = 0;
    uint32_t lvMinProcTimeTbl[MY_EM_CORE_NUM];
    uint32_t lvMaxProcTimeTbl[MY_EM_CORE_NUM];
    uint32_t lvMeanProcTimeTbl[MY_EM_CORE_NUM];
    uint32_t lvJobNumTbl[MY_EM_CORE_NUM][MY_FLOW_NUM];
    uint32_t lvStopTimeTbl[MY_EM_CORE_NUM];
    uint32_t lvOvhTime;
    uint32_t lvMinOvhTime = 0xffffffff;
    uint32_t lvMaxOvhTime = 0;
    uint32_t lvMeanOvhTime = 0;
    uint32_t lvMeanOvhCount = 0;
    uint32_t lvMinOvhTimeTbl[MY_EM_CORE_NUM];
    uint32_t lvMaxOvhTimeTbl[MY_EM_CORE_NUM];
    uint32_t lvMeanOvhTimeTbl[MY_EM_CORE_NUM];
    uint32_t lvScenarioPassFailFlag = 1;

    /*
     * initialize local variables 
     */
    for (lvCoreIdx = 0; lvCoreIdx < MY_EM_CORE_NUM; lvCoreIdx++)
    {
        lvMinProcTimeTbl[lvCoreIdx] = 0xffffffff;
        lvMaxProcTimeTbl[lvCoreIdx] = 0;
        lvMeanProcTimeTbl[lvCoreIdx] = 0;
        for (lvFlowIdx = 0; lvFlowIdx < MY_FLOW_NUM; lvFlowIdx++)
            lvJobNumTbl[lvCoreIdx][lvFlowIdx] = 0;
        lvStopTimeTbl[lvCoreIdx] = 0;
        lvMinOvhTimeTbl[lvCoreIdx] = 0xffffffff;
        lvMaxOvhTimeTbl[lvCoreIdx] = 0;
        lvMeanOvhTimeTbl[lvCoreIdx] = 0;
    }

    /*
     * compute statistics 
     */
    for (lvJobIdx = 0; lvJobIdx < MY_JOB_NUM; lvJobIdx++)
    {
        lvStartTime = my_svJobDscTbl[lvJobIdx].startTime;
        lvStopTime = my_svJobDscTbl[lvJobIdx].stopTime;
        lvCoreIdx = my_svJobDscTbl[lvJobIdx].coreIdx;
        lvFlowIdx = my_svJobDscTbl[lvJobIdx].flowIdx;

        if (my_svJobDscTbl[lvJobIdx].successFlag == 0)
            lvSuccessFlag = 0;

        /*
         * detect start time for first event 
         */
        if (lvMinStartTime > lvStartTime)
            lvMinStartTime = lvStartTime;

        /*
         * detect stop time for last event 
         */
        if (lvMaxStopTime < lvStopTime)
            lvMaxStopTime = lvStopTime;

        /*
         * compute total duration 
         */
        lvRunTime = lvMaxStopTime - lvMinStartTime;

        /*
         * increment total number of jobs 
         */
        lvJobNum++;

        /*
         * compute job duration time 
         */
        lvProcTime = lvStopTime - lvStartTime;

        /*
         * check with min duration time 
         */
        if (lvProcTime < lvMinProcTime)
            lvMinProcTime = lvProcTime;
        if (lvProcTime < lvMinProcTimeTbl[lvCoreIdx])
            lvMinProcTimeTbl[lvCoreIdx] = lvProcTime;

        /*
         * check with max duration time 
         */
        if (lvProcTime > lvMaxProcTime)
            lvMaxProcTime = lvProcTime;
        if (lvProcTime > lvMaxProcTimeTbl[lvCoreIdx])
            lvMaxProcTimeTbl[lvCoreIdx] = lvProcTime;

        lvJobNumTbl[lvCoreIdx][lvFlowIdx]++;
        lvMeanProcTime += lvProcTime;
        lvMeanProcTimeTbl[lvCoreIdx] += lvProcTime;

        if (lvStopTimeTbl[lvCoreIdx] != 0)
        {
            /*
             * compute duration time 
             */
            lvOvhTime = lvStartTime - lvStopTimeTbl[lvCoreIdx];

            /*
             * check with min duration time 
             */
            if (lvOvhTime < lvMinOvhTime)
                lvMinOvhTime = lvOvhTime;
            if (lvOvhTime < lvMinOvhTimeTbl[lvCoreIdx])
                lvMinOvhTimeTbl[lvCoreIdx] = lvOvhTime;

            /*
             * check with max duration time 
             */
            if (lvOvhTime > lvMaxOvhTime)
                lvMaxOvhTime = lvOvhTime;
            if (lvOvhTime > lvMaxOvhTimeTbl[lvCoreIdx])
                lvMaxOvhTimeTbl[lvCoreIdx] = lvOvhTime;

            lvMeanOvhCount++;
            lvMeanOvhTime += lvOvhTime;
            lvMeanOvhTimeTbl[lvCoreIdx] += lvOvhTime;
        }
        lvStopTimeTbl[lvCoreIdx] = lvStopTime;
    }

    printf ("\n==========\n");

    if (lvSuccessFlag == 1)
        printf ("data check OK!\n");
    else
        printf ("data check failed!\n");

    /*
     * per core statistics 
     */
    for (lvCoreIdx = 0; lvCoreIdx < MY_EM_CORE_NUM; lvCoreIdx++)
    {
        uint32_t lvJobNum2 = 0;

        for (lvFlowIdx = 0; lvFlowIdx < MY_FLOW_NUM; lvFlowIdx++)
            lvJobNum2 += lvJobNumTbl[lvCoreIdx][lvFlowIdx];

        printf ("\nStatistics for core %d:\n", lvCoreIdx);
        printf ("Number of jobs : %d\n", lvJobNum2);
        printf ("Number of jobs per flow: ");
        for (lvFlowIdx = 0; lvFlowIdx < MY_FLOW_NUM; lvFlowIdx++)
            printf ("%d ", lvJobNumTbl[lvCoreIdx][lvFlowIdx]);
        printf ("\n");

        if (lvJobNum2 > 0)
            printf ("Min/mean/max processing cycles : %d/%d/%d\n", lvMinProcTimeTbl[lvCoreIdx],
                    lvMeanProcTimeTbl[lvCoreIdx] / lvJobNum2, lvMaxProcTimeTbl[lvCoreIdx]);

        if (lvJobNum2 > 1)
            printf ("Min/mean/max overhead cycles : %d/%d/%d\n\n", lvMinOvhTimeTbl[lvCoreIdx],
                    lvMeanOvhTimeTbl[lvCoreIdx] / (lvJobNum2 - 1), lvMaxOvhTimeTbl[lvCoreIdx]);

        if (lvJobNum2 == 0)
            lvScenarioPassFailFlag = 0;
    }

    /*
     * print statistics 
     */
    printf ("\nStatistics for all cores\n");
    printf ("Total number of jobs: %d\n", lvJobNum);
    printf ("Total number of jobs per flow: ");
    for (lvFlowIdx = 0; lvFlowIdx < MY_FLOW_NUM; lvFlowIdx++)
    {
        uint32_t lvJobNum2 = 0;

        for (lvCoreIdx = 0; lvCoreIdx < MY_EM_CORE_NUM; lvCoreIdx++)
            lvJobNum2 += lvJobNumTbl[lvCoreIdx][lvFlowIdx];
        printf ("%d ", lvJobNum2);
    }
    printf ("\n");

    if (lvJobNum > 0)
        printf ("Min/mean/max processing cycles : %d/%d/%d\n", lvMinProcTime, lvMeanProcTime / lvJobNum, lvMaxProcTime);

    if (lvMeanOvhCount > 0)
        printf ("Min/mean/max overhead cycles : %d/%d/%d\n\n", lvMinOvhTime, lvMeanOvhTime / (lvMeanOvhCount),
                lvMaxOvhTime);

    printf ("Start/stop/run cycles: %d/%d/%d\n", lvMinStartTime, lvMaxStopTime, lvRunTime);


    if ((lvSuccessFlag == 1) & (lvScenarioPassFailFlag == 1))
        return 0;
    else
        return -1;
}
