/*
 * Copyright (c) 2012, Texas Instruments Incorporated - http://www.ti.com/
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions are met:
 *       * Redistributions of source code must retain the above copyright
 *         notice, this list of conditions and the following disclaimer.
 *       * Redistributions in binary form must reproduce the above copyright
 *         notice, this list of conditions and the following disclaimer in the
 *         documentation and/or other materials provided with the distribution.
 *       * Neither the name of Texas Instruments Incorporated nor the
 *         names of its contributors may be used to endorse or promote products
 *         derived from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *   DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
 *   DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


// This file implements utility functions, types, variables and constants
// The implementations are likely application specific and may change from one device to another

#include "my_util.h"
#include "my_init.h"
#include "my_em_init.h"

#include <ti/csl/csl_xmc.h>
#include <ti/csl/csl_xmcAux.h>
#include <ti/csl/csl_msmc.h>
#include <ti/csl/csl_msmcAux.h>
#include <ti/csl/csl_semAux.h>
#include <ti/csl/cslr_tmr.h>
#include <ti/csl/csl_emif4fAux.h>
#include <ti/csl/csl_cacheAux.h>

#define MY_HW_TIM_IDX (0)
#define MY_HW_SEM_IDX (0)

typedef struct my_ClockDsc_
{
    uint32_t globalStartTime;
    uint32_t localStartTime;
} my_ClockDsc;

#pragma DATA_ALIGN   (my_svClockDscTbl, CACHE_L2_LINESIZE)
#pragma DATA_SECTION (my_svClockDscTbl, ".my_svL1Mem");
my_ClockDsc my_svClockDscTbl[MY_EM_CORE_NUM];

typedef struct _my_CoreBarrier
{
    uint8_t tbl0[MY_EM_CORE_NUM];
    uint8_t tbl1[MY_EM_CORE_NUM];
} my_CoreBarrier;

#pragma DATA_ALIGN   (my_svCoreBarrier, CACHE_L2_LINESIZE)
#pragma DATA_SECTION (my_svCoreBarrier, ".my_svL1Mem");
my_CoreBarrier my_svCoreBarrier;

extern cregister volatile uint32_t TSCL;
extern cregister volatile uint32_t DNUM;

uint8_t my_gvCoreBarrierSense = 0;

uint32_t my_gvEmifPerfCount1;
uint32_t my_gvEmifPerfCount2;

volatile uint32_t my_gvStartTime;
volatile uint32_t my_gvStopTime;

void
my_startEmifPerfCounters (
    void
)
{
    CSL_EMIF4F_GetPerfCounters (&my_gvEmifPerfCount1, &my_gvEmifPerfCount2);
}

void
my_stopEmifPerfCounters (
    uint32_t * count1Ptr,
    uint32_t * count2Ptr
)
{
    uint32_t lvCount1;
    uint32_t lvCount2;

    CSL_EMIF4F_GetPerfCounters (&lvCount1, &lvCount2);
    *count1Ptr = lvCount1 - my_gvEmifPerfCount1;
    *count2Ptr = lvCount2 - my_gvEmifPerfCount2;
}

int
my_initMemLocal (
    void
)
{
    CSL_XMC_XMPAXH lvMpaxh;
    CSL_XMC_XMPAXL lvMpaxl;
    int i;

    // Configure caches
    CACHE_setL1DSize (CACHE_L1_32KCACHE);
    CACHE_setL1PSize (CACHE_L1_32KCACHE);
    CACHE_setL2Size (CACHE_128KCACHE);

    // Enable prefetching for MAR12 (0x0C00_0000:0x0CFF_FFFF)
    CACHE_setMemRegionInfo (12, 0, 1);
    
    // Disable caching for MAR28 (0x1C00_0000:0x1CFF_FFFF)
    CACHE_disableCaching (28);
    // Disable prefetching for MAR28
    CACHE_setMemRegionInfo (28, 0, 0);

    // Set MAR attributes for MAR128:MAR135 (0x8000_0000:0x87FF_FFFF)
    for (i = 0; i < 8; i++)
    {
        // Enable caching
        CACHE_enableCaching (128 + i);
        // Enable prefetching
        CACHE_setMemRegionInfo (128 + i, 0, 1);
    }

    // Set MAR attributes for MAR136:MAR143 (0x8800_0000:0x8FFF_FFFF)
    for (i = 0; i < 8; i++)
    {
        // Disable caching
        CACHE_disableCaching (136 + i);
        // Disable prefetching
        CACHE_setMemRegionInfo (136 + i, 0, 0);
    }

    /*
     * QMSS memories are not cacheable 
     * - Keystone I  devices: MAR52 : 0x34xx_xxxx
     * - Keystone II devices: MAR35 : 0x23xx_xxxx 
     */
#ifdef TI_EM_C6634
    CACHE_disableCaching (35);
#else
    CACHE_disableCaching (52);
#endif

    // Map 4 KB from 0x2100_0000 to 0x1_0000_0000 (XMC)
    // Use segment 2
    lvMpaxh.segSize = 0xB;      // 4 KB
    lvMpaxh.bAddr = 0x21000;    // 32-bit address >> 12
    CSL_XMC_setXMPAXH (2, &lvMpaxh);
    lvMpaxl.ux = 1;
    lvMpaxl.uw = 1;
    lvMpaxl.ur = 1;
    lvMpaxl.sx = 1;
    lvMpaxl.sw = 1;
    lvMpaxl.sr = 1;
    lvMpaxl.rAddr = 0x100000;   // 36-bit address >> 12
    CSL_XMC_setXMPAXL (2, &lvMpaxl);

    // Map 1 MB from 0x1c00_0000 to 0x0_0C00_0000 (XMC)
    // Use segment 4
    lvMpaxh.segSize = 0x13;     // 1 MB
#ifndef K2_PG1_MSMC_STALL_WA
    lvMpaxh.bAddr = 0x88300;    // 32-bit address >> 12
#else
    lvMpaxh.bAddr = 0x1c000;    // 32-bit address >> 12
#endif /* K2_PG1_MSMC_STALL_WA */
    CSL_XMC_setXMPAXH (4, &lvMpaxh);
    lvMpaxl.ux = 1;
    lvMpaxl.uw = 1;
    lvMpaxl.ur = 1;
    lvMpaxl.sx = 1;
    lvMpaxl.sw = 1;
    lvMpaxl.sr = 1;
    lvMpaxl.rAddr = 0x00C000;   // 36-bit address >> 12
    CSL_XMC_setXMPAXL (4, &lvMpaxl);

    return 0;
}

int
my_initMemGlobal (
    void
)
{
    uint8_t lvCosCount1;
    uint8_t lvCosCount2;
    uint8_t lvOldCount;
    uint8_t lvWriteThreshold;
    uint8_t lvReadThreshold;
    CSL_MSMC_SESMPAXH lvSesMpaxh;
    CSL_MSMC_SESMPAXL lvSesMpaxl;
    EMIF4F_PERF_CONFIG lvPerfConfig;

    lvCosCount1 = 0x0;
    lvCosCount2 = 0x0;
    lvOldCount = 0xFF;
    CSL_EMIF4F_SetVBUSConfig (lvCosCount1, lvCosCount2, lvOldCount);

    lvWriteThreshold = 0x3;
    lvReadThreshold = 0x5;
    CSL_EMIF4F_SetReadWriteThreshold (lvWriteThreshold, lvReadThreshold);

    lvPerfConfig.mstIDEn = 0;
    lvPerfConfig.regEn = 0;
    lvPerfConfig.cntCfg = 2;
    CSL_EMIF4F_SetPerfCounterConfig (1, &lvPerfConfig);

    lvPerfConfig.mstIDEn = 0;
    lvPerfConfig.regEn = 0;
    lvPerfConfig.cntCfg = 3;
    CSL_EMIF4F_SetPerfCounterConfig (2, &lvPerfConfig);

#ifndef K2_PG1_MSMC_STALL_WA
    // Map 1 MB from 0x8830_0000 to 0x0_0C10_0000 (MSMC)
    // Use segment 2 for PRIVID 8 and 10 (QMSS)
    lvSesMpaxh.segSz = 0x13;
    lvSesMpaxh.baddr = 0x88300; // 32-bit address >> 12
    CSL_MSMC_setSESMPAXH (8, 2, &lvSesMpaxh);
    CSL_MSMC_setSESMPAXH (9, 2, &lvSesMpaxh);   /* Privilege Id for RIO (K1) or QMSS2 (K2)  */
    CSL_MSMC_setSESMPAXH (10, 2, &lvSesMpaxh);
    lvSesMpaxl.ux = 1;
    lvSesMpaxl.uw = 1;
    lvSesMpaxl.ur = 1;
    lvSesMpaxl.sx = 1;
    lvSesMpaxl.sw = 1;
    lvSesMpaxl.sr = 1;
    lvSesMpaxl.raddr = 0x00C000;    // 36-bit address >> 12
    CSL_MSMC_setSESMPAXL (8, 2, &lvSesMpaxl);
    CSL_MSMC_setSESMPAXL (9, 2, &lvSesMpaxl);   /* Privilege Id for RIO (K1) or QMSS2 (K2)  */
    CSL_MSMC_setSESMPAXL (10, 2, &lvSesMpaxl);
#endif /* K2_PG1_MSMC_STALL_WA */
    return 0;
}

int
my_initClockLocal (
    void
)
{
    volatile CSL_TmrRegs *lvTimerRegPtr;

    lvTimerRegPtr = (CSL_TmrRegs *) (CSL_TIMER_0_REGS + (CSL_TIMER_1_REGS - CSL_TIMER_0_REGS) * MY_HW_TIM_IDX);
    TSCL = 0;                   // start local timer

    my_svClockDscTbl[DNUM].localStartTime = TSCL;
    my_svClockDscTbl[DNUM].globalStartTime = lvTimerRegPtr->CNTLO;

    return 0;
}

int
my_initClockGlobal (
    void
)
{
    volatile CSL_TmrRegs *lvTimerRegPtr;

    lvTimerRegPtr = (CSL_TmrRegs *) (CSL_TIMER_0_REGS + (CSL_TIMER_1_REGS - CSL_TIMER_0_REGS) * MY_HW_TIM_IDX);

    lvTimerRegPtr->TGCR = 0x00000000;   // TIMMODE = 0 ; TIMHIRS = 0 ; TIMLORS = 0
    lvTimerRegPtr->PRDLO = (Uint32) 0xffffffff;
    lvTimerRegPtr->PRDHI = (Uint32) 0xffffffff;
    lvTimerRegPtr->CNTLO = 0;
    lvTimerRegPtr->CNTHI = 0;
    lvTimerRegPtr->TCR = 0x00000080;    // ENAMODE_LO = 2 ; CLKSRC_LO = 0 ; TI_EN_LO = 0
    lvTimerRegPtr->TGCR = 0x00000003;   // TIMMODE = 0 ; TIMHIRS = 1 ; TIMLORS = 1

    return 0;
}

uint32_t
my_readClockLocal (
    void
)
{
    return TSCL;
}

uint32_t
my_readClockGlobal (
    void
)
{
    return (TSCL - my_svClockDscTbl[DNUM].localStartTime + 6 * my_svClockDscTbl[DNUM].globalStartTime);
}

void
my_waitAtCoreBarrier (
    void
)
{
    volatile uint8_t *lvCoreBarrierPtr;
    volatile uint8_t lvGo = 0;
    int i;

    if (my_gvCoreBarrierSense == 0)
    {
        lvCoreBarrierPtr = my_svCoreBarrier.tbl0;
        my_gvCoreBarrierSense = 1;
    }
    else
    {
        lvCoreBarrierPtr = my_svCoreBarrier.tbl1;
        my_gvCoreBarrierSense = 0;
    }

    for (i = 0; i < MY_EM_CORE_NUM; i++)
    {
        *(lvCoreBarrierPtr + i) = 0;
    }

    while (!lvGo)
    {
        lvGo = 1;
        *(lvCoreBarrierPtr + DNUM) = 1;
        for (i = 0; i < MY_EM_CORE_NUM; i++)
        {
            lvGo = lvGo & *(lvCoreBarrierPtr + i);
        }
        *(lvCoreBarrierPtr + DNUM) = 1;
    }
}

void
my_lock (
    void
)
{
    while (!CSL_semAcquireDirect (MY_HW_SEM_IDX));
}

void
my_unlock (
    void
)
{
    my_mfence ();
    CSL_semReleaseSemaphore (MY_HW_SEM_IDX);
}

// Wait until data written to memory
void
my_mfence (
    void
)
{
    _mfence ();
    asm (" NOP 9");
    asm (" NOP 7");
}

// Write back and invalidate all cache
void
my_cacheWbInvL2 (
    uint8_t * bufferPtr,
    uint32_t bufferSize
)
{
    uint32_t lvInt = _disable_interrupts ();
    CACHE_wbInvL2 (bufferPtr, bufferSize, CACHE_NOWAIT);
    CSL_XMC_invalidatePrefetchBuffer ();
    my_mfence ();
    _restore_interrupts (lvInt);
    return;
}

void *
my_ntoa (
    uint32_t naddr
)
{
#ifdef K2_PG1_MSMC_STALL_WA
    if ((naddr & 0xff000000) == 0x0c000000)
        return (void *) (naddr | 0x10000000);
    else
#endif /* K2_PG1_MSMC_STALL_WA */
        return (void *) naddr;
}

uint32_t
my_aton (
    void *aaddr
)
{
#ifdef K2_PG1_MSMC_STALL_WA
    if (((uint32_t)(aaddr) & 0xff000000) == 0x1c000000)
        return (uint32_t)(aaddr) & 0x0fffffff;
    else
#endif /* K2_PG1_MSMC_STALL_WA */
        return (uint32_t) aaddr;
}
