/*
 * Copyright (c) 2012, Texas Instruments Incorporated - http://www.ti.com/
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions are met:
 *       * Redistributions of source code must retain the above copyright
 *         notice, this list of conditions and the following disclaimer.
 *       * Redistributions in binary form must reproduce the above copyright
 *         notice, this list of conditions and the following disclaimer in the
 *         documentation and/or other materials provided with the distribution.
 *       * Neither the name of Texas Instruments Incorporated nor the
 *         names of its contributors may be used to endorse or promote products
 *         derived from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *   DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
 *   DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


// This file implements additional functions, types, variables and constants to be used for initializing the EM
// The implementations are TI specific and are likely the same for all applications

#include <stdio.h>
#include "my_em_init.h"
#include "ti_em_init.h"
#include "my_util.h"
#include <ti/runtime/openem/firmware/event_machine_firmware.h>

#include <ti/drv/srio/srio_drv.h>

#include <ti/drv/cppi/cppi_desc.h>
#include <ti/csl/csl_cppi.h>
#include <ti/csl/csl_pscAux.h>
#include <ti/csl/csl_srio.h>
#include <ti/csl/csl_srioAux.h>
#include <ti/csl/csl_srioAuxPhyLayer.h>
#include <ti/csl/csl_bootcfgAux.h>

extern cregister volatile uint32_t DNUM;

#define Emti_REGION_NUM (3 + MY_EM_CORE_NUM)
#define Emti_PUBLIC_REGION_IDX (0)
#define Emti_PRIVATE0_REGION_IDX (1)
#define Emti_PRIVATE1_REGION_IDX (2)
#define Emti_LOCAL_REGION_BASE_IDX (3)

#define Emti_PDSP_NUM (1)
#define Emti_PDSP_IDX (0)

#define Emti_LOCAL_EVENT_NUM (2)



/**********************************************************************
 ************************* LOCAL Definitions **************************
 **********************************************************************/

/* These are the GARBAGE queues which are used by the TXU to dump the 
 * descriptor if there is an error instead of recycling the descriptor
 * to the free queue. */
#define GARBAGE_LEN_QUEUE            955
#define GARBAGE_TOUT_QUEUE            956
#define GARBAGE_RETRY_QUEUE            957
#define GARBAGE_TRANS_ERR_QUEUE        958
#define GARBAGE_PROG_QUEUE            959
#define GARBAGE_SSIZE_QUEUE            960

/* SRIO Device Information
 * - 16 bit Device Identifier.
 * - 8 bit Device Identifier.
 * - Vendor Identifier. 
 * - Device Revision. */
#define DEVICE_VENDOR_ID            0x30
#define DEVICE_REVISION             0x0

/* SRIO Assembly Information
 * - Assembly Identifier
 * - Assembly Vendor Identifier. 
 * - Assembly Device Revision. 
 * - Assembly Extension Features */
#define DEVICE_ASSEMBLY_ID          0x0
#define DEVICE_ASSEMBLY_VENDOR_ID   0x30
#define DEVICE_ASSEMBLY_REVISION    0x0
#define DEVICE_ASSEMBLY_INFO        0x0100

/**********************************************************************
 ************************* Extern Definitions *************************
 **********************************************************************/
const uint32_t DEVICE_ID1_16BIT = 0x0001;
const uint32_t DEVICE_ID1_8BIT = 0x01;
const uint32_t DEVICE_ID2_16BIT = 0x1000;
const uint32_t DEVICE_ID2_8BIT = 0x02;
const uint32_t DEVICE_ID3_16BIT = 0x0003;
const uint32_t DEVICE_ID3_8BIT_ID = 0x03;
const uint32_t DEVICE_ID4_16BIT = 0x0004;
const uint32_t DEVICE_ID4_8BIT_ID = 0x04;


#ifndef TI_EM_C6634
extern Qmss_GlobalConfigParams qmssGblCfgParams;
extern Cppi_GlobalConfigParams cppiGblCfgParams[];
#else
#include <ti/drv/cppi/device/k2k/src/cppi_device.c>
#include <ti/drv/qmss/device/k2k/src/qmss_device.c>
#endif

#pragma DATA_ALIGN   (Emti_svPublicEventDscMem, CACHE_L2_LINESIZE);
#pragma DATA_SECTION (Emti_svPublicEventDscMem, ".tiEmSvPublicEventDscMem");
uint8_t Emti_svPublicEventDscMem[(TI_EM_EVENT_NUM2) * MY_EM_PUBLIC_EVENT_DSC_SIZE];

#ifdef TI_EM_C6634
#pragma DATA_ALIGN   (Emti_svPrivateEventDscMem, CACHE_L2_LINESIZE);
#pragma DATA_SECTION (Emti_svPrivateEventDscMem, ".tiEmSvPrivateEventDscMem");
uint8_t Emti_svPrivateEventDscMem[TI_EM_PDSP_GLOBAL_DATA_SIZE];
#endif

#pragma DATA_ALIGN   (Emti_svPrivateEventDscMem1, CACHE_L2_LINESIZE);
#pragma DATA_SECTION (Emti_svPrivateEventDscMem1, ".tiEmSvPrivateEventDscMem1");
uint8_t Emti_svPrivateEventDscMem1[TI_EM_PDSP_GLOBAL_DATA_SIZE];

#pragma DATA_ALIGN   (Emti_svPrivateEventDscMem2, CACHE_L2_LINESIZE);
#pragma DATA_SECTION (Emti_svPrivateEventDscMem2, ".tiEmSvPrivateEventDscMem2");
uint8_t Emti_svPrivateEventDscMem2[TI_EM_PDSP_GLOBAL_DATA_SIZE];

#pragma DATA_ALIGN   (Emti_gvLocalEventDscMem, CACHE_L2_LINESIZE);
#pragma DATA_SECTION (Emti_gvLocalEventDscMem, ".tiEmGvLocalEventDscMem");
uint8_t Emti_gvLocalEventDscMem[(Emti_LOCAL_EVENT_NUM + 1) * MY_EM_PUBLIC_EVENT_DSC_SIZE];

typedef enum
{
    srio_lane_rate_1p250Gbps = 0,               /**< Rate of 1.250Gbps for SRIO lanes */
    srio_lane_rate_2p500Gbps,                   /**< Rate of 2.500Gbps for SRIO lanes */
    srio_lane_rate_3p125Gbps,                   /**< Rate of 3.125Gbps for SRIO lanes */
    srio_lane_rate_5p000Gbps                    /**< Rate of 5.000Gbps for SRIO lanes */
} srioLaneRateGbps_e;

/** These are the possible values for SRIO PLL input reference clock in MHZ */
typedef enum
{
    srio_ref_clock_125p00Mhz = 0,  /**< Reference clock of 125.00Mhz for SRIO PLL */
    srio_ref_clock_156p25Mhz,      /**< Reference clock of 156.25Mhz for SRIO PLL */
    srio_ref_clock_250p00Mhz,      /**< Reference clock of 250.00Mhz for SRIO PLL */
    srio_ref_clock_312p50Mhz       /**< Reference clock of 312.50Mhz for SRIO PLL */
} srioRefClockMhz_e;

/** These are the possible values for SRIO lane mode */
typedef enum
{
    srio_lanes_form_four_1x_ports = 0,           /**< SRIO lanes form four 1x ports */
    srio_lanes_form_one_2x_port_and_two_1x_ports,/**< SRIO lanes form one 2x port and two 1x ports */
    srio_lanes_form_two_1x_ports_and_one_2x_port,/**< SRIO lanes form two 1x ports and one 2x port */
    srio_lanes_form_two_2x_ports,                /**< SRIO lanes form two 2x ports */
    srio_lanes_form_one_4x_port                  /**< SRIO lanes form one 4x port */
} srioLanesMode_e;


uint32_t
Emti_roundDscNum (
    uint32_t eventNum
)
{
    uint32_t lvEventNum;
    uint32_t lvEventNumTmp;
    uint32_t lvEventNumMin = 0x00000020;    // 32
    uint32_t lvEventNumMax = 0x00004000;    // 16K

    lvEventNum = eventNum;
    if (lvEventNum < lvEventNumMin)
    {
        lvEventNum = lvEventNumMin;
    }
    else
    {
        lvEventNumTmp = lvEventNumMax >> 1;
        while (lvEventNum <= lvEventNumTmp)
        {
            lvEventNumTmp = lvEventNumTmp >> 1;
        }
        lvEventNum = lvEventNumTmp << 1;
    }

    return lvEventNum;
}

uint32_t
Emti_makeAddressGlobal (
    uint32_t coreIdx,
    uint32_t address
)
{
    if (address < 0x01000000)
    {
        address = (0x10 + coreIdx) << 24 | address;
    }

    return address;
}


static Int32
enable_srio (
    void
)
{
    /*
     * SRIO power domain is turned OFF by default. It needs to be turned on before doing any 
     * * SRIO device register access. This not required for the simulator. 
     */

    /*
     * Set SRIO Power domain to ON 
     */
    CSL_PSC_enablePowerDomain (CSL_PSC_PD_SRIO);

    /*
     * Enable the clocks too for SRIO 
     */
    CSL_PSC_setModuleNextState (CSL_PSC_LPSC_SRIO, PSC_MODSTATE_ENABLE);

    /*
     * Start the state transition 
     */
    CSL_PSC_startStateTransition (CSL_PSC_PD_SRIO);

    /*
     * Wait until the state transition process is completed. 
     */
    while (!CSL_PSC_isStateTransitionDone (CSL_PSC_PD_SRIO));

    /*
     * Return SRIO PSC status 
     */
    if ((CSL_PSC_getPowerDomainState (CSL_PSC_PD_SRIO) == PSC_PDSTATE_ON) &&
        (CSL_PSC_getModuleState (CSL_PSC_LPSC_SRIO) == PSC_MODSTATE_ENABLE))
    {
        /*
         * SRIO ON. Ready for use 
         */
        return 0;
    }
    else
    {
        /*
         * SRIO Power on failed. Return error 
         */
        return -1;
    }
}

#if (defined TI_EM_C6634  || defined TI_EM_C6638)
void
reg_poll (
    volatile unsigned int addr,
    volatile unsigned int poll_val,
    volatile unsigned int masked_bits
)
{
    volatile unsigned int read_data;

    read_data = (*(unsigned int *) (addr));
    read_data = read_data & masked_bits;
    while (read_data != (poll_val & masked_bits))
    {
        read_data = (*(unsigned int *) (addr));
        read_data = read_data & masked_bits;
    }
}

void
wait_for_pll_lock (
)
{
    reg_poll (CSL_SRIO_SERDES_CFG_REGS + 0x1ff4, 0xF0F, 0xF0F);
}

void
reg_wr_8 (
    Uint32 addr,
    unsigned char write_data
)
{
    *(unsigned char *) addr = write_data;
}

void
reg_wr_32 (
    Uint32 addr,
    Uint32 write_data
)
{
    (*(Uint32 *) (addr)) = write_data;
}

void
serdes_5G (
)
{
    /*
     * Uses Full Rate configuration by default 
     */
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x003, 0x00);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x014, 0x82);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x015, 0x82);

    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x060, 0x38);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x061, 0x24);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x062, 0x14);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x065, 0xc7);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x066, 0xc3);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x079, 0xc0);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x204, 0x80);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x207, 0x78);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x208, 0x26);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x20f, 0x02);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x214, 0x38);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x215, 0x6f);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x218, 0xe4);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x21a, 0x80);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x21b, 0x75);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x22d, 0x08);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x22e, 0x10);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x280, 0x86);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x282, 0x86);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x284, 0x85);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x286, 0x0f);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x287, 0x1d);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x28d, 0x2c);

    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x404, 0x80);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x407, 0x78);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x408, 0x26);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x40f, 0x02);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x414, 0x38);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x415, 0x6f);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x418, 0xe4);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x41a, 0x80);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x41b, 0x75);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x42d, 0x08);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x42e, 0x10);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x480, 0x86);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x482, 0x86);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x484, 0x85);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x486, 0x0f);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x487, 0x1d);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x48d, 0x2c);


    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x604, 0x80);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x607, 0x78);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x608, 0x26);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x60f, 0x02);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x614, 0x38);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x615, 0x6f);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x618, 0xe4);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x61a, 0x80);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x61b, 0x75);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x62d, 0x08);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x62e, 0x10);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x680, 0x86);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x682, 0x86);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x684, 0x85);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x686, 0x0f);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x687, 0x1d);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x68d, 0x2c);


    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x804, 0x80);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x807, 0x78);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x808, 0x26);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x80f, 0x02);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x814, 0x38);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x815, 0x6f);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x818, 0xe4);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x81a, 0x80);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x81b, 0x75);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x82d, 0x08);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x82e, 0x10);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x880, 0x86);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x882, 0x86);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x884, 0x85);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x886, 0x0f);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x887, 0x1d);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x88d, 0x2c);

    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0xa01, 0x80);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0xa0a, 0xd2);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0xa0b, 0x38);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0xa31, 0x8d);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0xa32, 0x8d);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0xa85, 0x06);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0xa97, 0x10);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0xaa3, 0x81);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0xabf, 0xff);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0xac0, 0x8b);

    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x000, 0x03);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0xa00, 0x5f);
}


int32_t
srioDefSerdesSetup (
    CSL_SrioHandle hSrio,
    srioRefClockMhz_e refClockMhz,
    srioLaneRateGbps_e linkRateGbps,
    int isLoopbackMode
)
{

    // Write 0 to Boot complete bit
    CSL_SRIO_SetBootComplete (hSrio, 0);

    // Disable pll before configuring the SerDes registers
    reg_wr_32 (CSL_SRIO_SERDES_CFG_REGS + 0x1ff4, 0x00000000);

    serdes_5G ();
    reg_wr_32 (CSL_SRIO_SERDES_CFG_REGS + 0x1fe0, 0xF0C0F0F0);
    reg_wr_32 (CSL_SRIO_SERDES_CFG_REGS + 0x1fe4, 0xF0C0F0F0);
    reg_wr_32 (CSL_SRIO_SERDES_CFG_REGS + 0x1fe8, 0xF0C0F0F0);
    reg_wr_32 (CSL_SRIO_SERDES_CFG_REGS + 0x1fec, 0xF0C0F0F0);

    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x203, 0x40);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x403, 0x40);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x603, 0x40);
    reg_wr_8 (CSL_SRIO_SERDES_CFG_REGS + 0x803, 0x40);


    // Enable pll vi athe pll_ctrl 0x0014
    reg_wr_32 (CSL_SRIO_SERDES_CFG_REGS + 0x1ff4, 0xe0000000);

    // Wait the SerDes PLL lock
    wait_for_pll_lock ();
    /*
     * TBD SRIO SERDES configuration for different modes 
     */

    return 0;
}

#elif (defined TI_EM_C6614 || defined TI_EM_C6670 || defined TI_EM_C6678)
void
srioDefSerdesSetup (
)
{
    /*
     * Refclock is at 250 MHz                               
     */
    /*
     * MPY fact i 5 in full rate mode                       
     */
    /*
     * => Line rate is 5 GHz                                
     */
    CSL_BootCfgSetSRIOSERDESConfigPLL (0x229);

    /*
     * Configure the SRIO SERDES Receive Configuration. 
     */
    CSL_BootCfgSetSRIOSERDESRxConfig (0, 0x00440485);
    CSL_BootCfgSetSRIOSERDESRxConfig (1, 0x00440485);
    CSL_BootCfgSetSRIOSERDESRxConfig (2, 0x00440485);
    CSL_BootCfgSetSRIOSERDESRxConfig (3, 0x00440485);

    /*
     * Configure the SRIO SERDES Transmit Configuration. 
     */
    CSL_BootCfgSetSRIOSERDESTxConfig (0, 0x00180785);
    CSL_BootCfgSetSRIOSERDESTxConfig (1, 0x00180785);
    CSL_BootCfgSetSRIOSERDESTxConfig (2, 0x00180785);
    CSL_BootCfgSetSRIOSERDESTxConfig (3, 0x00180785);

    /*
     * Loop around till the SERDES PLL is not locked. 
     */
    while (1)
    {
        uint32_t status;

        /*
         * Get the SRIO SERDES Status 
         */
        CSL_BootCfgGetSRIOSERDESStatus (&status);
        if (status & 0x1)
            break;
    }
}
#endif

#pragma CODE_SECTION(SrioDevice_init, ".text:SrioDevice_init");
int32_t
SrioDevice_init (
    void
)
{
    CSL_SrioHandle hSrio;
    int32_t i;
    SRIO_PE_FEATURES peFeatures;
    SRIO_OP_CAR opCar;
    Qmss_QueueHnd queueHnd;
    uint8_t isAllocated;
    uint32_t gargbageQueue[] = { GARBAGE_LEN_QUEUE, GARBAGE_TOUT_QUEUE,
        GARBAGE_RETRY_QUEUE, GARBAGE_TRANS_ERR_QUEUE,
        GARBAGE_PROG_QUEUE, GARBAGE_SSIZE_QUEUE
    };

    /*
     * Get the CSL SRIO Handle. 
     */
    hSrio = CSL_SRIO_Open (0);
    if (hSrio == NULL)
        return -1;

    /*
     * Code to disable SRIO reset isolation 
     */
    if (CSL_PSC_isModuleResetIsolationEnabled (CSL_PSC_LPSC_SRIO))
        CSL_PSC_disableModuleResetIsolation (CSL_PSC_LPSC_SRIO);

    /*
     * Disable the SRIO Global block 
     */
    CSL_SRIO_GlobalDisable (hSrio);

    /*
     * Disable each of the individual SRIO blocks. 
     */
    for (i = 0; i <= 9; i++)
        CSL_SRIO_DisableBlock (hSrio, i);

    /*
     * Set boot complete to be 0; we are not done with the initialization. 
     */
    CSL_SRIO_SetBootComplete (hSrio, 0);

    /*
     * Now enable the SRIO block and all the individual blocks also. 
     */
    CSL_SRIO_GlobalEnable (hSrio);
    for (i = 0; i <= 9; i++)
        CSL_SRIO_EnableBlock (hSrio, i);
    /*
     * CAUTION: HERE THE SRIO is configured so the data was pushed back to
     * the digital domain before the SERDES macros.
     */
    /*
     * Configure SRIO ports to operate in loopback mode. 
     */
    CSL_SRIO_SetLoopbackMode (hSrio, 0);
    CSL_SRIO_SetLoopbackMode (hSrio, 1);
    CSL_SRIO_SetLoopbackMode (hSrio, 2);
    CSL_SRIO_SetLoopbackMode (hSrio, 3);

    /*
     * Enable Automatic Priority Promotion of response packets. 
     */
    CSL_SRIO_EnableAutomaticPriorityPromotion (hSrio);

    /*
     * Set the SRIO Prescalar select to operate in the range of 44.7 to 89.5 
     */
    CSL_SRIO_SetPrescalarSelect (hSrio, 0);

    /*
     * Unlock the Boot Configuration Kicker 
     */
    CSL_BootCfgUnlockKicker ();

#if (defined TI_EM_C6634 || defined TI_EM_C6638)
    srioDefSerdesSetup (hSrio, srio_ref_clock_156p25Mhz, srio_lane_rate_5p000Gbps, 1);
#elif (defined TI_EM_C6614 || defined TI_EM_C6670 || defined TI_EM_C6678)
    srioDefSerdesSetup ();
#endif

    /*
     * Clear the LSU pending interrupts. 
     */
    CSL_SRIO_ClearLSUPendingInterrupt (hSrio, 0xFFFFFFFF, 0xFFFFFFFF);

    /*
     * Set the Device Information 
     */
    CSL_SRIO_SetDeviceInfo (hSrio, DEVICE_ID1_16BIT, DEVICE_VENDOR_ID, DEVICE_REVISION);

    /*
     * Set the Assembly Information 
     */
    CSL_SRIO_SetAssemblyInfo (hSrio, DEVICE_ASSEMBLY_ID, DEVICE_ASSEMBLY_VENDOR_ID, DEVICE_ASSEMBLY_REVISION,
                              DEVICE_ASSEMBLY_INFO);

    /*
     * TODO: Configure the processing element features
     * *  The SRIO RL file is missing the Re-transmit Suppression Support (Bit6) field definition 
     */
    peFeatures.isBridge = 0;
    peFeatures.isEndpoint = 0;
    peFeatures.isProcessor = 1;
    peFeatures.isSwitch = 0;
    peFeatures.isMultiport = 0;
    peFeatures.isFlowArbiterationSupported = 0;
    peFeatures.isMulticastSupported = 0;
    peFeatures.isExtendedRouteConfigSupported = 0;
    peFeatures.isStandardRouteConfigSupported = 1;
    peFeatures.isFlowControlSupported = 1;
    peFeatures.isCRFSupported = 0;
    peFeatures.isCTLSSupported = 1;
    peFeatures.isExtendedFeaturePtrValid = 1;
    peFeatures.numAddressBitSupported = 1;
    CSL_SRIO_SetProcessingElementFeatures (hSrio, &peFeatures);

    /*
     * Configure the source operation CAR 
     */
    memset ((void *) &opCar, 0, sizeof (opCar));
    opCar.portWriteOperationSupport = 1;
    opCar.atomicClearSupport = 1;
    opCar.atomicSetSupport = 1;
    opCar.atomicDecSupport = 1;
    opCar.atomicIncSupport = 1;
    opCar.atomicTestSwapSupport = 1;
    opCar.doorbellSupport = 1;
    opCar.dataMessageSupport = 1;
    opCar.writeResponseSupport = 1;
    opCar.streamWriteSupport = 1;
    opCar.writeSupport = 1;
    opCar.readSupport = 1;
    opCar.dataStreamingSupport = 1;
    CSL_SRIO_SetSourceOperationCAR (hSrio, &opCar);

    /*
     * Configure the destination operation CAR 
     */
    memset ((void *) &opCar, 0, sizeof (opCar));
    opCar.portWriteOperationSupport = 1;
    opCar.doorbellSupport = 1;
    opCar.dataMessageSupport = 1;
    opCar.writeResponseSupport = 1;
    opCar.streamWriteSupport = 1;
    opCar.writeSupport = 1;
    opCar.readSupport = 1;
    CSL_SRIO_SetDestOperationCAR (hSrio, &opCar);

    /*
     * Set the 16 bit and 8 bit identifier for the SRIO Device. 
     */
    CSL_SRIO_SetDeviceIDCSR (hSrio, DEVICE_ID1_8BIT, DEVICE_ID1_16BIT);

    /*
     * Enable TLM Base Routing Information for Maintainance Requests & ensure that
     * * the BRR's can be used by all the ports. 
     */
    CSL_SRIO_SetTLMPortBaseRoutingInfo (hSrio, 0, 1, 1, 1, 0);
    CSL_SRIO_SetTLMPortBaseRoutingInfo (hSrio, 0, 2, 1, 1, 0);
    CSL_SRIO_SetTLMPortBaseRoutingInfo (hSrio, 0, 3, 1, 1, 0);
    CSL_SRIO_SetTLMPortBaseRoutingInfo (hSrio, 1, 0, 1, 1, 0);

    /*
     * Configure the Base Routing Register to ensure that all packets matching the 
     * * Device Identifier & the Secondary Device Id are admitted. 
     */
    CSL_SRIO_SetTLMPortBaseRoutingPatternMatch (hSrio, 0, 1, DEVICE_ID2_16BIT, 0xFFFF);
    CSL_SRIO_SetTLMPortBaseRoutingPatternMatch (hSrio, 0, 2, DEVICE_ID3_16BIT, 0xFFFF);
    CSL_SRIO_SetTLMPortBaseRoutingPatternMatch (hSrio, 0, 3, DEVICE_ID4_16BIT, 0xFFFF);
    CSL_SRIO_SetTLMPortBaseRoutingPatternMatch (hSrio, 1, 0, DEVICE_ID2_8BIT, 0xFF);

    /*
     * We need to open the Garbage collection queues in the QMSS. This is done to ensure that 
     * * these queues are not opened by another system entity. 
     */
    for (i = 0; i < 6; i++)
    {
        /*
         * Open the Garabage queues 
         */
        queueHnd = Qmss_queueOpen (Qmss_QueueType_GENERAL_PURPOSE_QUEUE, gargbageQueue[i], &isAllocated);
        if (queueHnd < 0)
            return -1;

        /*
         * Make sure the queue has not been opened already; we dont the queues to be shared by some other
         * * entity in the system. 
         */
        if (isAllocated > 1)
            return -1;
    }

    /*
     * Set the Transmit Garbage Collection Information. 
     */
    CSL_SRIO_SetTxGarbageCollectionInfo (hSrio, GARBAGE_LEN_QUEUE, GARBAGE_TOUT_QUEUE, GARBAGE_RETRY_QUEUE,
                                         GARBAGE_TRANS_ERR_QUEUE, GARBAGE_PROG_QUEUE, GARBAGE_SSIZE_QUEUE);

    /*
     * Set the Host Device Identifier. 
     */
    CSL_SRIO_SetHostDeviceID (hSrio, DEVICE_ID1_16BIT);

    /*
     * Configure the component tag CSR 
     */
    CSL_SRIO_SetCompTagCSR (hSrio, 0x00000000);

    /*
     * Configure the PLM for all the ports. 
     */
    for (i = 0; i < 4; i++)
    {
        /*
         * Set the PLM Port Silence Timer. 
         */
        CSL_SRIO_SetPLMPortSilenceTimer (hSrio, i, 0x2);

        /*
         * TODO: We need to ensure that the Port 0 is configured to support both
         * * the 2x and 4x modes. The Port Width field is read only. So here we simply
         * * ensure that the Input and Output ports are enabled. 
         */
        CSL_SRIO_EnableInputPort (hSrio, i);
        CSL_SRIO_EnableOutputPort (hSrio, i);

        /*
         * Set the PLM Port Discovery Timer. 
         */
        CSL_SRIO_SetPLMPortDiscoveryTimer (hSrio, i, 0x2);

        /*
         * Reset the Port Write Reception capture. 
         */
        CSL_SRIO_SetPortWriteReceptionCapture (hSrio, i, 0x0);
    }

    /*
     * Set the Port link timeout CSR 
     */
    CSL_SRIO_SetPortLinkTimeoutCSR (hSrio, 0x000FFF);

    /*
     * Set the Port General CSR: Only executing as Master Enable 
     */
    CSL_SRIO_SetPortGeneralCSR (hSrio, 0, 1, 0);

    /*
     * Clear the sticky register bits. 
     */
    CSL_SRIO_SetLLMResetControl (hSrio, 1);

    /*
     * Set the device id to be 0 for the Maintenance Port-Write operation 
     * * to report errors to a system host. 
     */
    CSL_SRIO_SetPortWriteDeviceId (hSrio, 0x0, 0x0, 0x0);

    /*
     * Set the Data Streaming MTU 
     */
    CSL_SRIO_SetDataStreamingMTU (hSrio, 64);

    /*
     * Configure the path mode for the ports. 
     */
    for (i = 0; i < 4; i++)
        CSL_SRIO_SetPLMPortPathControlMode (hSrio, i, 0);

    /*
     * Set the LLM Port IP Prescalar. 
     */
    CSL_SRIO_SetLLMPortIPPrescalar (hSrio, 0x21);

    /*
     * Enable the peripheral. 
     */
    CSL_SRIO_EnablePeripheral (hSrio);

    /*
     * Configuration has been completed. 
     */
    CSL_SRIO_SetBootComplete (hSrio, 1);

    /*
     * This code checks if the ports are operational or not. The functionality is not supported 
     * * on the simulator. 
     */
    for (i = 0; i < 4; i++)
        while (CSL_SRIO_IsPortOk (hSrio, i) != TRUE);

    /*
     * Set all the queues 0 to operate at the same priority level and to send packets onto Port 0 
     */
    for (i = 0; i < 16; i++)
        CSL_SRIO_SetTxQueueSchedInfo (hSrio, i, 0, 0);

    /*
     * Set the Doorbell route to determine which routing table is to be used 
     * * This configuration implies that the Interrupt Routing Table is configured as 
     * * follows:-
     * *  Interrupt Destination 0 - INTDST 16 
     * *  Interrupt Destination 1 - INTDST 17 
     * *  Interrupt Destination 2 - INTDST 18
     * *  Interrupt Destination 3 - INTDST 19 
     */
    CSL_SRIO_SetDoorbellRoute (hSrio, 0);

    /*
     * Route the Doorbell interrupts. 
     * *  Doorbell Register 0 - All 16 Doorbits are routed to Interrupt Destination 0. 
     * *  Doorbell Register 1 - All 16 Doorbits are routed to Interrupt Destination 1. 
     * *  Doorbell Register 2 - All 16 Doorbits are routed to Interrupt Destination 2. 
     * *  Doorbell Register 3 - All 16 Doorbits are routed to Interrupt Destination 3. 
     */
    for (i = 0; i < 16; i++)
    {
        CSL_SRIO_RouteDoorbellInterrupts (hSrio, 0, i, 0);
        CSL_SRIO_RouteDoorbellInterrupts (hSrio, 1, i, 1);
        CSL_SRIO_RouteDoorbellInterrupts (hSrio, 2, i, 2);
        CSL_SRIO_RouteDoorbellInterrupts (hSrio, 3, i, 3);
    }

    /*
     * Initialization has been completed. 
     */
    return 0;
}

int
Emti_initPdspConfig (
    uint32_t pdspNum1,
    Qmss_PdspCfg * pdspConfigTbl1,
    Qmss_PdspCfg * pdspConfigTbl2,
    Qmss_PdspId * emPdspIdxPtr
)
{
    uint32_t lvPdspNum;

    lvPdspNum = pdspNum1 + Emti_PDSP_NUM;
    if (lvPdspNum > QMSS_MAX_PDSP)
        return -1;

    {
#if defined(TI_EM_C6670) || defined(TI_EM_C6678)
        *emPdspIdxPtr = Qmss_PdspId_PDSP1;
        pdspConfigTbl2[Emti_PDSP_IDX].pdspId = Qmss_PdspId_PDSP1;
        pdspConfigTbl2[Emti_PDSP_IDX].firmware = &em_scheduler_router_fw_c667x_pdsp1;
        pdspConfigTbl2[Emti_PDSP_IDX].size = sizeof (em_scheduler_router_fw_c667x_pdsp1);
#endif
#ifdef TI_EM_C6634
        *emPdspIdxPtr = Qmss_PdspId_PDSP1;
        pdspConfigTbl2[Emti_PDSP_IDX].pdspId = Qmss_PdspId_PDSP1;
        pdspConfigTbl2[Emti_PDSP_IDX].firmware = &em_scheduler_router_fw_c663x;
        pdspConfigTbl2[Emti_PDSP_IDX].size = sizeof (em_scheduler_router_fw_c663x);
#endif
    }

    return lvPdspNum;
}

// At least 1 region for public global events
// Exactly 1 region for private events
// At least 1 region per core for public local events
int
Emti_initRegionConfig (
    Qmss_PdspCfg * pdspConfigTbl,
    uint32_t regionNum1,
    Qmss_MemRegInfo * regionConfigTbl1,
    Qmss_MemRegInfo * regionConfigTbl2
)
{
    uint32_t lvRegionNum;
    uint32_t lvDscNum = 0;
    uint32_t lvLinkIdx = 0;
    Qmss_MemRegInfo *lvRegionConfigPtrTbl[QMSS_MAX_MEM_REGIONS];
    int i, j;

    lvRegionNum = regionNum1 + Emti_REGION_NUM;
    if (lvRegionNum > QMSS_MAX_MEM_REGIONS)
        return -1;

    // initialize descriptor regions for OpenEM
    // public event descriptors
    regionConfigTbl2[Emti_PUBLIC_REGION_IDX].descSize = (uint32_t) MY_EM_PUBLIC_EVENT_DSC_SIZE;
    regionConfigTbl2[Emti_PUBLIC_REGION_IDX].descNum = Emti_roundDscNum (1 + TI_EM_EVENT_NUM2);
    regionConfigTbl2[Emti_PUBLIC_REGION_IDX].descBase =
        (uint32_t *) Emti_makeAddressGlobal (0, (uint32_t)my_aton((void*)(Emti_svPublicEventDscMem)));
    regionConfigTbl2[Emti_PUBLIC_REGION_IDX].manageDescFlag = Qmss_ManageDesc_UNMANAGED_DESCRIPTOR;

    // private event descriptors

    regionConfigTbl2[Emti_PRIVATE0_REGION_IDX].descSize = (uint32_t) TI_EM_PRIVATE_EVENT_DSC_SIZE;
    regionConfigTbl2[Emti_PRIVATE0_REGION_IDX].descNum = Emti_roundDscNum (TI_EM_CD_PRIVATE_EVENT_NUM);
#if defined(TI_EM_C6670) || defined(TI_EM_C6678)
    if (pdspConfigTbl[Emti_PDSP_IDX].pdspId == Qmss_PdspId_PDSP1)
        regionConfigTbl2[Emti_PRIVATE0_REGION_IDX].descBase =
            (uint32_t *) Emti_makeAddressGlobal (0, (uint32_t) Emti_svPrivateEventDscMem1);
    else
        regionConfigTbl2[Emti_PRIVATE0_REGION_IDX].descBase =
            (uint32_t *) Emti_makeAddressGlobal (0, (uint32_t) Emti_svPrivateEventDscMem2);
#else
    regionConfigTbl2[Emti_PRIVATE0_REGION_IDX].descBase =
        (uint32_t *) Emti_makeAddressGlobal (0, (uint32_t) Emti_svPrivateEventDscMem);
#endif
    regionConfigTbl2[Emti_PRIVATE0_REGION_IDX].manageDescFlag = Qmss_ManageDesc_UNMANAGED_DESCRIPTOR;
    // private event descriptors
    regionConfigTbl2[Emti_PRIVATE1_REGION_IDX].descSize = (uint32_t) TI_EM_PRIVATE_EVENT_DSC_SIZE;
    regionConfigTbl2[Emti_PRIVATE1_REGION_IDX].descNum = Emti_roundDscNum (TI_EM_AP_PRIVATE_EVENT_NUM);
    regionConfigTbl2[Emti_PRIVATE1_REGION_IDX].descBase =
        (uint32_t *) ((uint32_t) regionConfigTbl2[Emti_PRIVATE0_REGION_IDX].descBase +
                      regionConfigTbl2[Emti_PRIVATE0_REGION_IDX].descNum *
                      regionConfigTbl2[Emti_PRIVATE0_REGION_IDX].descSize);
    regionConfigTbl2[Emti_PRIVATE1_REGION_IDX].manageDescFlag = Qmss_ManageDesc_UNMANAGED_DESCRIPTOR;
    memset ((void *) &Emti_svPrivateEventDscMem1, 0, sizeof (Emti_svPrivateEventDscMem1));
    memset ((void *) &Emti_svPrivateEventDscMem2, 0, sizeof (Emti_svPrivateEventDscMem2));
#ifdef TI_EM_C6634
    memset ((void *) &Emti_svPrivateEventDscMem, 0, sizeof (Emti_svPrivateEventDscMem));
#endif

    // local event descriptors
    for (i = 0; i < MY_EM_CORE_NUM; i++)
    {
        regionConfigTbl2[Emti_LOCAL_REGION_BASE_IDX + i].descSize = (uint32_t) MY_EM_PUBLIC_EVENT_DSC_SIZE;
        regionConfigTbl2[Emti_LOCAL_REGION_BASE_IDX + i].descNum = Emti_roundDscNum (1 + Emti_LOCAL_EVENT_NUM);
        regionConfigTbl2[Emti_LOCAL_REGION_BASE_IDX + i].descBase =
            (uint32_t *) Emti_makeAddressGlobal (i, (uint32_t) Emti_gvLocalEventDscMem);
        regionConfigTbl2[Emti_LOCAL_REGION_BASE_IDX + i].manageDescFlag = Qmss_ManageDesc_UNMANAGED_DESCRIPTOR;
    }

    // copy user descriptor regions
    for (i = 0; i < regionNum1; i++)
    {
        regionConfigTbl2[Emti_REGION_NUM + i].descSize = regionConfigTbl1[i].descSize;
        regionConfigTbl2[Emti_REGION_NUM + i].descNum = regionConfigTbl1[i].descNum;
        regionConfigTbl2[Emti_REGION_NUM + i].descBase = regionConfigTbl1[i].descBase;
        regionConfigTbl2[Emti_REGION_NUM + i].manageDescFlag = regionConfigTbl1[i].manageDescFlag;
    }

    // Sort descriptor regions according to memory base address.
    // Count total number of descriptors.
    for (i = 0; i < lvRegionNum; i++)
    {
        uint32_t lvRank = 0;
        for (j = 0; j < lvRegionNum; j++)
        {
            if (regionConfigTbl2[j].descBase < regionConfigTbl2[i].descBase)
                lvRank++;
        }
        lvRegionConfigPtrTbl[lvRank] = (&regionConfigTbl2[i]);
        lvDscNum += regionConfigTbl2[i].descNum;
    }

    // assign descriptor region index and linking RAM start index
    for (i = 0; i < lvRegionNum; i++)
    {
        lvRegionConfigPtrTbl[i]->memRegion = (Qmss_MemRegion) i;
        lvRegionConfigPtrTbl[i]->startIndex = lvLinkIdx;
        lvLinkIdx += lvRegionConfigPtrTbl[i]->descNum;
    }

    // update input descriptor regions
    for (i = 0; i < regionNum1; i++)
    {
        regionConfigTbl1[i].memRegion = regionConfigTbl2[Emti_REGION_NUM + i].memRegion;
        regionConfigTbl1[i].startIndex = regionConfigTbl2[Emti_REGION_NUM + i].startIndex;
    }

    return lvRegionNum;
}

em_status_t
Emti_initQmss (
    uint32_t regionNum,
    Qmss_MemRegInfo * regionConfigTbl,
    uint32_t pdspNum,
    Qmss_PdspCfg * pdspConfigTbl,
    uint8_t * extLinkTbl,
    Cppi_Handle * cppiHdlPtr
)
{
    uint32_t lvDscNum = 0;
    Qmss_InitCfg lvQmssCfg;
    Cppi_CpDmaInitCfg lvCppiCfg;
    int i, j;

    for (i = 0; i < regionNum; i++)
    {
        lvDscNum += regionConfigTbl[i].descNum;
    }

    /*
     * Set up the linking RAM configuration. Use the internal Linking RAM. 
     */
    lvQmssCfg.linkingRAM0Base = 0;  // Use linking RAM of Navigator SS
    lvQmssCfg.linkingRAM0Size = 0;  // Use all linking RAM of Navigator SS
    lvQmssCfg.linkingRAM1Base = (uint32_t) extLinkTbl;
    lvQmssCfg.maxDescNum = lvDscNum;
#ifdef TI_EM_C6634
    lvQmssCfg.mode = Qmss_Mode_JOINT_LOADBALANCED;
#endif

    // Set up the PDSP configuration
    for (i = 0; i < pdspNum; i++)
    {
        lvQmssCfg.pdspFirmware[i].pdspId = pdspConfigTbl[i].pdspId;
        lvQmssCfg.pdspFirmware[i].firmware = pdspConfigTbl[i].firmware;
        lvQmssCfg.pdspFirmware[i].size = pdspConfigTbl[i].size;
    }
    for (i = pdspNum; i < QMSS_MAX_PDSP; i++)
    {
        lvQmssCfg.pdspFirmware[i].pdspId = (Qmss_PdspId) i;
        lvQmssCfg.pdspFirmware[i].firmware = NULL;
        lvQmssCfg.pdspFirmware[i].size = 0;
    }

    /*
     * Initialize the QMSS 
     */
    if (Qmss_init (&lvQmssCfg, (Qmss_GlobalConfigParams *) & qmssGblCfgParams) != QMSS_SOK)
        return EM_ERR;

    /*
     * Start the QMSS. 
     */
    if (Qmss_start () != QMSS_SOK)
        return EM_ERR;

    /*
     * Flush all the queues 
     */
    for (i = 0; i < qmssGblCfgParams.maxQue; i++)
        Qmss_queueEmpty (i);

    // Insert descriptor regions.
    for (i = 0; i < regionNum; i++)
    {
        for (j = 0; j < regionNum; j++)
        {
            if (regionConfigTbl[j].memRegion == (Qmss_MemRegion) i)
            {
                if (Qmss_insertMemoryRegion (&regionConfigTbl[j]) < 0)
                    return EM_ERR;
            }
        }
    }

    /*
     * Initialize CPPI CPDMA 
     */
#ifdef TI_EM_C6634
    if (Cppi_init (&cppiGblCfgParams) != CPPI_SOK)
        return EM_ERR;
#else
    if (Cppi_init (cppiGblCfgParams) != CPPI_SOK)
        return EM_ERR;
#endif

    /*
     * Open cppi 
     */
    lvCppiCfg.dmaNum = Cppi_CpDma_QMSS_CPDMA;
    lvCppiCfg.writeFifoDepth = 3;   // 
    lvCppiCfg.timeoutCount = 0; //

#ifdef TI_EM_C6634
    lvCppiCfg.qm0BaseAddress = CSL_QMSS_DATA_QM1_QUEUE_MANAGEMENT_REGS + 0 * 4096;
    lvCppiCfg.qm1BaseAddress = CSL_QMSS_DATA_QM1_QUEUE_MANAGEMENT_REGS + 16 * 4096;
    lvCppiCfg.qm2BaseAddress = CSL_QMSS_DATA_QM1_QUEUE_MANAGEMENT_REGS + 32 * 4096;
    lvCppiCfg.qm3BaseAddress = CSL_QMSS_DATA_QM1_QUEUE_MANAGEMENT_REGS + 48 * 4096;
#else
    lvCppiCfg.qm0BaseAddress = CSL_QM_SS_DATA_QM_QUEUE_DEQUEUE_REGS + 0 * 4096;
    lvCppiCfg.qm1BaseAddress = CSL_QM_SS_DATA_QM_QUEUE_DEQUEUE_REGS + 16 * 4096;
    lvCppiCfg.qm2BaseAddress = CSL_QM_SS_DATA_QM_QUEUE_DEQUEUE_REGS + 32 * 4096;
    lvCppiCfg.qm3BaseAddress = CSL_QM_SS_DATA_QM_QUEUE_DEQUEUE_REGS + 48 * 4096;
#endif /* TI_EM_C6634 */

    /*
     * Open QMSS CPDMA 
     */
    *cppiHdlPtr = Cppi_open (&lvCppiCfg);

    /*
     * CPPI and QMSS Infrastructure is operational. 
     */
    return EM_OK;
}


em_status_t
Emti_initSrio (
    Cppi_Handle * srioCppiHdlPtr
)
{
    Cppi_CpDmaInitCfg srioCPDMACfg;

    if (enable_srio () == -1)
    {
        //printf("RIO power on failed\n");
        return EM_ERR;
    }
    SrioDevice_init ();

    /*
     * Initialize the SRIO Driver 
     */
    if (Srio_init () < 0)
    {
        //printf ("Error: SRIO Driver Initialization Failed\n");
        return EM_ERR;
    }

    /*
     * Initialize the SRIO CPDMA config structure. 
     */
    memset ((void *) &srioCPDMACfg, 0, sizeof (Cppi_CpDmaInitCfg));
    srioCPDMACfg.dmaNum = Cppi_CpDma_SRIO_CPDMA;
    /*
     * Open the SRIO CPDMA. 
     */
    *srioCppiHdlPtr = Cppi_open (&srioCPDMACfg);

    return EM_OK;
}

em_status_t
ti_em_init_private_events (
    Qmss_MemRegInfo * regionConfigTbl2
)
{
    Qmss_QueueHnd lvPrivateFreeQueueHandle;
    uint8_t lvIsAllocated;
    uint32_t lvEventDscAdd;
    uint32_t lvIndex;

    /*
     * Open the queues 
     */
    lvPrivateFreeQueueHandle =
        Qmss_queueOpen (Qmss_QueueType_GENERAL_PURPOSE_QUEUE, MY_EM_CD_PRIVATE_FREE_QUEUE_IDX, &lvIsAllocated);
    if ((lvPrivateFreeQueueHandle < 0) || (lvIsAllocated != 1))
        return EM_ERR_ALLOC_FAILED;

    /*
     * assume private region is at base address of the pdsp data ram, 
     * first descriptor is reserved for cd queue identification 
     */
    lvEventDscAdd = (uint32_t) (regionConfigTbl2[Emti_PRIVATE0_REGION_IDX].descBase) + TI_EM_PRIVATE_EVENT_DSC_SIZE;
    /*
     * push private descriptors in private free queue 
     */
    for (lvIndex = 0; lvIndex < (TI_EM_CD_PRIVATE_EVENT_NUM - 1); lvIndex++)
    {
        Qmss_queuePushDesc (lvPrivateFreeQueueHandle, (void *) lvEventDscAdd);
        lvEventDscAdd += TI_EM_PRIVATE_EVENT_DSC_SIZE;
    }

    lvPrivateFreeQueueHandle =
        Qmss_queueOpen (Qmss_QueueType_GENERAL_PURPOSE_QUEUE, MY_EM_AP_PRIVATE_FREE_QUEUE_IDX, &lvIsAllocated);
    if ((lvPrivateFreeQueueHandle < 0) || (lvIsAllocated != 1))
        return EM_ERR_ALLOC_FAILED;

    lvEventDscAdd = (uint32_t) (regionConfigTbl2[Emti_PRIVATE1_REGION_IDX].descBase);
    /*
     * push private descriptors in private free queue 
     */
    for (lvIndex = 0; lvIndex < TI_EM_AP_PRIVATE_EVENT_NUM; lvIndex++)
    {
        Qmss_queuePushDesc (lvPrivateFreeQueueHandle, (void *) lvEventDscAdd);
        lvEventDscAdd += TI_EM_PRIVATE_EVENT_DSC_SIZE;
    }
    return EM_OK;
}

em_status_t
Emti_initEvents (
    Qmss_MemRegInfo * regionConfigTbl,
    uint32_t poolNum,
    ti_em_pool_config2_t * poolConfig2Tbl
)
{
    uint8_t *lvDscPtr;
    uint32_t lvDscSize;
    uint32_t lvEventNumTotal = 1;
	uint32_t lvEventNumMax = regionConfigTbl[Emti_PUBLIC_REGION_IDX].descNum;
    int i;

    lvDscPtr = (uint8_t *)my_ntoa((uint32_t)(regionConfigTbl[Emti_PUBLIC_REGION_IDX].descBase));
    lvDscSize = regionConfigTbl[Emti_PUBLIC_REGION_IDX].descSize;
    lvDscPtr += lvDscSize;

    for (i = 0; i < poolNum; i++)
    {
        uint8_t *lvBufPtr;
        uint32_t lvBufSize;
        uint32_t lvEventNum;
        uint32_t lvQueueIdx;
        int j;

        lvBufPtr = poolConfig2Tbl[i].buffer_ptr;
        lvBufSize = poolConfig2Tbl[i].buffer_size;
        lvEventNum = poolConfig2Tbl[i].event_num;
        lvQueueIdx = poolConfig2Tbl[i].pool_config.free_queue_idx;
        lvEventNumTotal += lvEventNum;

		if (lvEventNumTotal > lvEventNumMax)
			return EM_ERR;

        if (lvEventNumTotal > (TI_EM_EVENT_NUM2 + 1))
            return EM_ERR;
        if (ti_em_hw_queue_open (lvQueueIdx) != EM_OK)
            return EM_ERR;

        for (j = 0; j < lvEventNum; j++)
        {
            Cppi_DescTag lvTag;

            ti_em_packet_set_default ((ti_em_packet_t *) lvDscPtr);

            ti_em_packet_set_pool_info ((ti_em_packet_t *) (lvDscPtr), (poolConfig2Tbl[i].pool_config));

            if (poolConfig2Tbl[i].buffer_ptr != NULL)
            {
                ti_em_buffer_config_t lvBufferConfig;

                lvBufferConfig.buffer_ptr = (void *) (lvBufPtr);
                lvBufferConfig.buffer_size = (uint32_t) (lvBufSize);
                lvBufferConfig.orig_buffer_ptr = (void *) (lvBufPtr);
                lvBufferConfig.orig_buffer_size = (uint32_t) (lvBufSize);

                if (i == MY_EM_EXIT_POOL_IDX)
                    lvBufferConfig.coh_mode = (ti_em_coh_mode_t) (TI_EM_COH_MODE_OFF);
                if (i == MY_EM_POOL_IDX)
                    lvBufferConfig.coh_mode = (ti_em_coh_mode_t) (MY_EM_COH_MODE);
                if (i == MY_EM_PROCESS_CHAIN0_POOL_IDX)
                    lvBufferConfig.coh_mode = (ti_em_coh_mode_t) (MY_EM_PROCESS_CHAIN0_COH_MODE);

                lvBufferConfig.free_func = (ti_em_free_func_t) NULL;
                lvBufferConfig.free_push_policy = (ti_em_push_policy_t) TI_EM_PUSH_POLICY_TAIL;

                ti_em_packet_set_buffer_info ((ti_em_packet_t *) lvDscPtr, lvBufferConfig);
            }
            Cppi_setPSLen (Cppi_DescType_HOST, (Cppi_Desc *) (lvDscPtr), poolConfig2Tbl[i].ps_word_num << 2);

            lvTag.srcTagHi = 0x00;
            lvTag.srcTagLo = 0xFF;
            lvTag.destTagHi = 0x00;
            lvTag.destTagLo = 0x00;
            Cppi_setTag (Cppi_DescType_HOST, (Cppi_Desc *) (lvDscPtr), &lvTag);

            my_mfence ();

            Qmss_queuePushDesc (lvQueueIdx, (void *)my_aton((void*)(lvDscPtr)));

            lvBufPtr = lvBufPtr + lvBufSize;
            lvDscPtr = lvDscPtr + lvDscSize;
        }
    }

    return EM_OK;
}

em_status_t
Emti_initPlEvents (
    Qmss_MemRegInfo * regionConfigTbl,
    uint32_t coreNum,
    ti_em_pl_pool_config2_t * poolConfig2Tbl
)
{
    ti_em_pool_config_t lvPoolConfig;
    int i;

    lvPoolConfig.buf_mode = TI_EM_BUF_MODE_TIGHT;

    for (i = 0; i < coreNum; i++)
    {
        uint32_t lvCoreIdx;
        uint8_t *lvDscPtr;
        uint32_t lvDscSize;
        uint8_t *lvBufPtr;
        uint32_t lvBufSize;
        uint32_t lvPsWordNum;
        uint32_t lvEventNum;
    	uint32_t lvEventNumMax;
        uint32_t lvQueueIdx;
        int j;

        lvCoreIdx = poolConfig2Tbl[i].core_idx;
        lvBufPtr = poolConfig2Tbl[i].buffer_ptr;
        lvBufSize = poolConfig2Tbl[i].buffer_size;
        lvEventNum = poolConfig2Tbl[i].event_num;
        lvPsWordNum = poolConfig2Tbl[i].ps_word_num;
        lvQueueIdx = poolConfig2Tbl[i].free_queue_idx;
    	
        lvEventNumMax = regionConfigTbl[Emti_LOCAL_REGION_BASE_IDX + lvCoreIdx].descNum;

		if (lvEventNum > lvEventNumMax)
			return EM_ERR;

        if (lvEventNum > Emti_LOCAL_EVENT_NUM+1)
            return EM_ERR;
            
        if (ti_em_hw_queue_open (lvQueueIdx) != EM_OK)
            return EM_ERR;

        lvPoolConfig.free_queue_idx = lvQueueIdx;

        lvDscPtr = (uint8_t *) regionConfigTbl[Emti_LOCAL_REGION_BASE_IDX + lvCoreIdx].descBase;
        lvDscSize = regionConfigTbl[Emti_LOCAL_REGION_BASE_IDX + lvCoreIdx].descSize;
        lvDscPtr += lvDscSize;

        for (j = 0; j < lvEventNum; j++)
        {
            Cppi_DescTag lvTag;

            ti_em_packet_set_default ((ti_em_packet_t *) lvDscPtr);
            ti_em_packet_set_pool_info ((ti_em_packet_t *) (lvDscPtr), lvPoolConfig);

            if (lvBufPtr != NULL)
            {
                ti_em_buffer_config_t lvBufferConfig;

                lvBufferConfig.buffer_ptr = (void *) (lvBufPtr);
                lvBufferConfig.buffer_size = (uint32_t) (lvBufSize);
                lvBufferConfig.orig_buffer_ptr = (void *) (lvBufPtr);
                lvBufferConfig.orig_buffer_size = (uint32_t) (lvBufSize);
                lvBufferConfig.coh_mode = (ti_em_coh_mode_t) (TI_EM_COH_MODE_OFF);
                lvBufferConfig.free_func = (ti_em_free_func_t) NULL;
                lvBufferConfig.free_push_policy = (ti_em_push_policy_t) TI_EM_PUSH_POLICY_TAIL;

                ti_em_packet_set_buffer_info ((ti_em_packet_t *) lvDscPtr, lvBufferConfig);
            }

            Cppi_setPSLen (Cppi_DescType_HOST, (Cppi_Desc *) (lvDscPtr), lvPsWordNum << 2);

            lvTag.srcTagHi = 0x00;
            lvTag.srcTagLo = 0xFF;
            lvTag.destTagHi = 0x00;
            lvTag.destTagLo = 0x00;
            Cppi_setTag (Cppi_DescType_HOST, (Cppi_Desc *) (lvDscPtr), &lvTag);

            my_mfence ();

            Qmss_queuePushDesc (lvQueueIdx, (void *)my_aton((void*)(lvDscPtr)));

            lvBufPtr = lvBufPtr + lvBufSize;
            lvDscPtr = lvDscPtr + lvDscSize;
        }
    }

    return EM_OK;
}

void
Emti_initHwConfig (
    ti_em_hw_config2_t hwConfig2,
    Qmss_MemRegInfo * regionConfigTbl,
    Cppi_Handle cppiHdl,
    Cppi_Handle srioCppiHdl,
    uint32_t poolNum,
    ti_em_pool_config2_t * poolConfig2Tbl,
    uint32_t coreNum,
    ti_em_pl_pool_config2_t * plPoolConfig2Tbl,
    ti_em_config_t * hwConfigPtr,
    ti_em_preload_config_t * preloadPtr,
    Qmss_PdspId emPdspIdx
)
{
    int i;

    hwConfigPtr->my_device_idx = hwConfig2.my_device_idx;
    hwConfigPtr->my_process_idx = hwConfig2.my_process_idx;
    hwConfigPtr->preload_config_ptr = preloadPtr;
    hwConfigPtr->pdsp_idx_tbl[0] = (ti_em_pdsp_id_t) emPdspIdx;
    hwConfigPtr->pdsp_num = 1;
    hwConfigPtr->ap_region_queue_idx = hwConfig2.ap_private_free_queue_idx;
    hwConfigPtr->cd_region_queue_idx = hwConfig2.cd_private_free_queue_idx;
    hwConfigPtr->hw_queue_base_idx = hwConfig2.hw_queue_base_idx;
    hwConfigPtr->dma_queue_base_idx = hwConfig2.dma_queue_base_idx;
    hwConfigPtr->pool_num = poolNum;
    for (i = 0; i < poolNum; i++)
    {
        hwConfigPtr->pool_config_tbl[i].free_queue_idx = poolConfig2Tbl[i].pool_config.free_queue_idx;
        hwConfigPtr->pool_config_tbl[i].buf_mode = poolConfig2Tbl[i].pool_config.buf_mode;
        hwConfigPtr->pool_config_tbl[i].dsc_wsize = poolConfig2Tbl[i].pool_config.dsc_wsize;
    }
    for (i = 0; i < MY_EM_CORE_NUM; i++)
    {
        preloadPtr->local_free_queue_idx_tbl[i] = 0;
    }
    for (i = 0; i < coreNum; i++)
    {
        uint32_t lvCoreIdx;

        lvCoreIdx = plPoolConfig2Tbl[i].core_idx;
        preloadPtr->local_free_queue_idx_tbl[lvCoreIdx] = plPoolConfig2Tbl[i].free_queue_idx;
    }
    hwConfigPtr->hw_sem_idx = hwConfig2.hw_sem_idx;
    hwConfigPtr->dma_idx = (uint32_t) cppiHdl;
    preloadPtr->preload_size_a = hwConfig2.preload_size_a;
    preloadPtr->preload_size_b = hwConfig2.preload_size_b;
    preloadPtr->preload_size_c = hwConfig2.preload_size_c;
    hwConfigPtr->chain_config_ptr = hwConfig2.chain_config_ptr;
    hwConfigPtr->chain_config_ptr->xge_config_ptr = hwConfig2.chain_config_ptr->xge_config_ptr;
    hwConfigPtr->chain_config_ptr->rio_config_ptr = hwConfig2.chain_config_ptr->rio_config_ptr;
    hwConfigPtr->chain_config_ptr->poststore_config_ptr = NULL;
    hwConfigPtr->chain_config_ptr->dma_idx = (uint32_t) cppiHdl;

    hwConfigPtr->chain_config_ptr->rio_config_ptr->dma_idx = (uint32_t) srioCppiHdl;
    hwConfigPtr->chain_config_ptr->rio_config_ptr->dma_tx_queue_idx =
        hwConfig2.chain_config_ptr->rio_config_ptr->dma_tx_queue_idx;
    hwConfigPtr->chain_config_ptr->rio_config_ptr->service_class =
        hwConfig2.chain_config_ptr->rio_config_ptr->service_class;
    hwConfigPtr->chain_config_ptr->rio_config_ptr->my_node_idx[0] =
        hwConfig2.chain_config_ptr->rio_config_ptr->my_node_idx[0];
    hwConfigPtr->chain_config_ptr->rio_config_ptr->my_node_idx[1] =
        hwConfig2.chain_config_ptr->rio_config_ptr->my_node_idx[1];

    printf ("This test-case uses the following configuration:\n");
    printf ("  - my_device_idx        : %u\n", hwConfigPtr->my_device_idx);
    printf ("  - my_process_idx       : %u\n", hwConfigPtr->my_process_idx);
    printf ("  - nb scheduler pdsp    : %u\n", hwConfigPtr->pdsp_num);
    printf ("  - scheduler_pdsp_idx   : %u\n", hwConfigPtr->pdsp_idx_tbl[0]);
    printf ("  - AP private_free_queue: %u\n", hwConfigPtr->ap_region_queue_idx);
    printf ("  - CD private_free_queue: %u\n", hwConfigPtr->cd_region_queue_idx);
    printf ("  - hw_queue_base_idx    : %u\n", hwConfigPtr->hw_queue_base_idx);
    printf ("  - dma_queue_base_idx   : %u\n", hwConfigPtr->dma_queue_base_idx);
    printf ("  - pool_num             : %u\n", hwConfigPtr->pool_num);
    printf ("  - SRIO Chaining        : enabled\n");
    printf ("  - SRIO TX queue        : %u\n", hwConfigPtr->chain_config_ptr->rio_config_ptr->dma_tx_queue_idx);
    printf ("  - SRIO Node Idx        : %u\n\n",
            (uint32_t) (hwConfigPtr->chain_config_ptr->rio_config_ptr->my_node_idx[0] +
                        (hwConfigPtr->chain_config_ptr->rio_config_ptr->my_node_idx[1] << 8)));
    return;
}

#pragma DATA_SECTION (Emti_cppiHdl, ".tiEmSvPublicEventDscMem");
Cppi_Handle Emti_cppiHdl;
#pragma DATA_SECTION (Emti_srioCppiHdl, ".tiEmSvPublicEventDscMem");
Cppi_Handle Emti_srioCppiHdl;

em_status_t
ti_em_init_global2 (
    uint32_t region_num,
    Qmss_MemRegInfo * region_config_tbl,
    uint32_t pdsp_num,
    Qmss_PdspCfg * pdsp_config_tbl,
    uint32_t pool_num,
    ti_em_pool_config2_t * pool_config2_tbl,
    uint32_t core_num,
    ti_em_pl_pool_config2_t * pl_pool_config2_tbl,
    ti_em_hw_config2_t hw_config2
)
{
    Qmss_MemRegInfo lvRegionCfgTbl[Emti_REGION_NUM];
    int32_t lvRegionNum;
    Qmss_PdspCfg lvPdspCfgTbl[Emti_PDSP_NUM];
    int32_t lvPdspNum;
    ti_em_config_t lvHwConfig;
    ti_em_preload_config_t lvPreloadConfig;
    Qmss_PdspId emPdspIdx;

    if (pool_num > TI_EM_POOL_NUM)
        return EM_ERR;
    if (core_num > MY_EM_CORE_NUM)
        return EM_ERR;

    // Build table of PDSP configurations
    lvPdspNum = Emti_initPdspConfig (pdsp_num, pdsp_config_tbl, lvPdspCfgTbl, &emPdspIdx);
    if (lvPdspNum < 0)
        return EM_ERR;

    // Build table of region configurations
    lvRegionNum = Emti_initRegionConfig (lvPdspCfgTbl, region_num, region_config_tbl, lvRegionCfgTbl);
    if (lvRegionNum < 0)
        return EM_ERR;

    // Initialize QMSS
    if (Emti_initQmss (lvRegionNum, lvRegionCfgTbl, lvPdspNum, lvPdspCfgTbl, NULL, &Emti_cppiHdl) != EM_OK)
        return EM_ERR;

    // Initialize SRIO
    if (Emti_initSrio (&Emti_srioCppiHdl) != EM_OK)
        return EM_ERR;

    // initialize private free queue
    ti_em_init_private_events (lvRegionCfgTbl);

    // Build table of pool configurations and initialize events
    if (Emti_initEvents (lvRegionCfgTbl, pool_num, pool_config2_tbl) != EM_OK)
        return EM_ERR;

    // Build table of preload pool configurations and initialize local events
    if (Emti_initPlEvents (lvRegionCfgTbl, core_num, pl_pool_config2_tbl) != EM_OK)
        return EM_ERR;

    // Set EM configuration and initialize EM
    Emti_initHwConfig (hw_config2, lvRegionCfgTbl, Emti_cppiHdl, Emti_srioCppiHdl, 3, pool_config2_tbl, core_num,
                       pl_pool_config2_tbl, &lvHwConfig, &lvPreloadConfig, emPdspIdx);

    if (ti_em_init_global (&lvHwConfig) != EM_OK)
        return EM_ERR;

    {
        /*
         *  Adding a fake route to device #1.
         *  It's a fake in a sense that SRIO is in digital loopback mode.
         */
        ti_em_device_rio_route_t lvRoute;

        printf ("Adding a route to RIO device:\n");
        printf ("  - device_idx           : %u\n", DEST_DEVICE_IDX);
        printf ("  - node_idx             : %u\n\n", DEST_RIO_NODE_IDX);

        lvRoute.node_idx[0] = (uint8_t) (DEST_RIO_NODE_IDX & 0xFF);
        lvRoute.node_idx[1] = (uint8_t) ((DEST_RIO_NODE_IDX >> 8) & 0xFF);
        if (ti_em_device_add_rio_route (DEST_DEVICE_IDX, lvRoute) != EM_OK)
            return EM_ERR;
    }

    return EM_OK;
}

em_status_t
ti_em_init_local2 (
    void
)
{
    if (Qmss_start () != QMSS_SOK)
        return EM_ERR;

    if (ti_em_init_local () != EM_OK)
        return EM_ERR;

    return EM_OK;
}

em_status_t
ti_em_exit_global2 (
    void
)
{
    uint32_t idx;

    if (ti_em_exit_global () != EM_OK)
        return EM_ERR;

    /*
     * Releasing public queues
     */
    if (Qmss_queueClose ((Qmss_QueueHnd) MY_EM_CD_PRIVATE_FREE_QUEUE_IDX) != 0)
        return EM_ERR;
    if (Qmss_queueClose ((Qmss_QueueHnd) MY_EM_AP_PRIVATE_FREE_QUEUE_IDX) != 0)
        return EM_ERR;
    for (idx = 0; idx < MY_EM_POOL_NUM; idx++)
    {
        Qmss_QueueHnd lvQueueHdl = (Qmss_QueueHnd) (MY_EM_FREE_QUEUE_BASE_IDX + idx);
        if (Qmss_queueClose ((Qmss_QueueHnd) lvQueueHdl) != 0)
            return EM_ERR;
    }
    for (idx = 0; idx < MY_EM_CORE_NUM; idx++)
    {
        Qmss_QueueHnd lvQueueHdl = (Qmss_QueueHnd) (MY_EM_PL_FREE_QUEUE_BASE_IDX + idx);
        if (Qmss_queueClose ((Qmss_QueueHnd) lvQueueHdl) != 0)
            return EM_ERR;
    }
    if (Qmss_queueClose ((Qmss_QueueHnd) GARBAGE_LEN_QUEUE) != 0)
        return EM_ERR;
    if (Qmss_queueClose ((Qmss_QueueHnd) GARBAGE_TOUT_QUEUE) != 0)
        return EM_ERR;
    if (Qmss_queueClose ((Qmss_QueueHnd) GARBAGE_RETRY_QUEUE) != 0)
        return EM_ERR;
    if (Qmss_queueClose ((Qmss_QueueHnd) GARBAGE_TRANS_ERR_QUEUE) != 0)
        return EM_ERR;
    if (Qmss_queueClose ((Qmss_QueueHnd) GARBAGE_PROG_QUEUE) != 0)
        return EM_ERR;
    if (Qmss_queueClose ((Qmss_QueueHnd) GARBAGE_SSIZE_QUEUE) != 0)
        return EM_ERR;


    /*
     * Looking at the CPPI Pkdma side
     */
    {
        Cppi_Result lvRes = Cppi_close (Emti_cppiHdl);
        if (lvRes != 0)
            return EM_ERR;
    }

    return EM_OK;
}
