/*
 *   Copyright (c) 2012, Nokia Siemens Networks
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions are met:
 *       * Redistributions of source code must retain the above copyright
 *         notice, this list of conditions and the following disclaimer.
 *       * Redistributions in binary form must reproduce the above copyright
 *         notice, this list of conditions and the following disclaimer in the
 *         documentation and/or other materials provided with the distribution.
 *       * Neither the name of Nokia Siemens Networks nor the
 *         names of its contributors may be used to endorse or promote products
 *         derived from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *   DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
 *   DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


/**
 *
 * Contributors
 *
 * The initial developer of the original code is Nokia Siemens Networks.
 *  The following contributors have made changes to 
 *  the original code as described below.
 * 
 *----------------------------------------------------------------------------------------------
 * Copyright (c) 2012, Texas Instruments Incorporated. All Rights Reserved. - http://www.ti.com/
 * 
 * Changes were done on the original NSN header files in order to avoid 
 * circular dependencies when called in the TI implementation files.
 *
 * From the original event_machine.h, the list of changes are:
 *   - Functions definitions are now located in:
 *     - openem/event_machine_functions.h
 *   - Doxygen tags are added for API documentation generation.
 *
 * The event_machine.h is now just an aggregator of include files. It will
 * be called by the Open Event Machine user. 
 *
 * Finally, the Open Event Machine is a RTSC component.
 * See http://rtsc.eclipse.org/docs-tip/Main_Page for details.
 *
 *----------------------------------------------------------------------------------------------
 */


/**
 * @file : event_machine.h
 *
 * @brief
 * Open Event Machine API v1.0
 *
 */
  
#ifndef EVENT_MACHINE_H
#define EVENT_MACHINE_H
/**
 * @mainpage
 *
 * This document contains information about the Texas Instruments implementation of the Open Event Machine.
 *
 * The file event_machine.h represents the interface of the event machine library.
 * It includes all the specifications needed to use the library.
 * It shall be included as follows to access the event machine functionality.
 * @code #include <ti/runtime/openem/event_machine.h>
 * @endcode
 * 
 * The file event_machine_firmware.h contains the binary of the micro risc controller implementing the event machine scheduler.
 * The binary shall be loaded in the micro risc controller when initializing the QMSS.
 * This file shall be included as follows.
 * @code #include <ti/runtime/openem/firmware/event_machine_firmware.h>
 * @endcode

 * Limitations of the Texas Instruments implementation of the Open Event Machine are listed in the release notes document.
 *
 * @n This document is divided into the following sections:
 *
 * - @subpage DOC_intro
 *
 * - @subpage DOC_error
 *
 * - @subpage DOC_trace
 *
 * - @subpage DOC_api_generic
 *
 * - @subpage DOC_api_specific
 */



#define EM_API_VERSION_MAJOR 1     /**< Major API version number. Step if not backwards compatible  */
#define EM_API_VERSION_MINOR 0     /**< Minor API version number. Updates and additions             */


/** 
 *
 * @page DOC_intro Introduction
 * @section section_1 General
 * Event Machine (EM) is an architectural abstraction and framework of an event driven
 * multicore optimized processing concept originally developed for networking data plane.
 * It offers an easy programming concept for scalable and dynamically load balanced
 * multicore applications with a very low overhead run-to-completion principle.
 *
 * Main elements in the concept are events, queues, scheduler, dispatcher and the
 * execution objects (EO). Event is an application specific piece of data (like a message
 * or a network packet) describing work, something to do. Any processing in EM must be
 * triggered by an event. Events are sent to asynchronous application specific queues.
 * A single thread on all cores of an EM instance runs a dispatcher loop (a "core" is
 * used here to refer to a core or one HW thread on multi-threaded cores). Dispatcher
 * interfaces with the scheduler and asks for an event. Scheduler evaluates the state
 * of all queues and gives the highest priority event to the dispatcher, which forwards
 * it to the EO mapped to the queue the event came from by calling the registered receive
 * function. As the event is handled and receive function returns, the next event is
 * received from the scheduler and again forwarded to the mapped EO. This happens in
 * parallel on all cores included. Everything is highly efficient run to completion
 * single thread, no context switching nor pre-emption (priorities are handled by the
 * scheduler).
 * EM can run on bare metal for best performance or under an operating system with special
 * arrangements (e.g. one thread per core with thread affinity).
 *
 * The concept and this API are made to be easy to implement for multiple general purpose or
 * networking oriented multicore packet processing systems on chip, which typically also
 * contain accelerators for packet processing needs. Efficient integration with modern HW
 * accelerators has been a major driver in EM concept.
 *
 * One general principle of this API is that all calls are multicore safe, i.e.
 * no data structure gets broken, if calls are simultaneously made by multiple cores, but
 * unless explicitly documented per API call, the application also needs to take the parallel
 * world into consideration. For example if one core asks for a queue mode and another one
 * changes the mode at the same time, the returned mode may be invalid (valid data, but the old
 * or the new!). Thus modifications should be done under atomic context (if load balancing is
 * used) or otherwise synchronized by the application. One simple way of achieving this is to use
 * one EO with an atomic queue to do all the management functionality. That guarantees
 * synchronized operations (but also serializes them limiting performance)
 *
 * EM_64_BIT or EM_32_BIT (needs to be defined at makefile) defines whether (most of)
 * the types used in the API are 32 or 64 bits wide. NOTE, that this is a major decision,
 * since it may limit value passing between different systems using the defined types directly.
 * Using 64-bits may allow more efficient underlying implementation, as more data can be coded
 * in 64-bit identifiers for instance.
 *
 * @section section_2 Some principles
 * - This API attempts to guide towards a portable application architecture, but is not defined
 *  for portability by re-compilation. Many things are system specific giving more possibilities
 *  for efficient use of HW resources.
 * - EM does not define event content (one exception, see em_alloc()). This is a choice made for
 * performance, since most HW devices use proprietary descriptors. This API enables to use those directly.
 * - EM does not define detailed queue scheduling disciplines or API to set those up (or actually
 * anything to configure a system). The priority value in this API is a (mapped) system specific
 * QoS class label only
 * - In general EM does not implement full SW platform or middleware solution, it implements a sub-
 * set of such, a driver level part. For best performance it can be used directly from applications.
 *
 */



#ifdef __cplusplus
extern "C" {
#endif


/**
 * @defgroup DOC_api_generic Generic API
 * @brief Generic declarations.
 */
/**
 * @defgroup DOC_api_specific TI specific API
 * @brief Implementation specific declarations.
 * @attention Unless otherwise specified values of symbolic constants cannot be changed.
 */

/** 
 *
 * @page DOC_error Error
 * @section section_101 mechanism
 * 
 * The error handler prototype to be provided by the application to the OpenEM when activating the error mecanism shall comply with the following prototype:
 * 
 * 
 * em_status_t (*em_error_handler_t)(em_eo_t eo, em_status_t error, em_escope_t escope, va_list args)
 * 
 * 
 * In the current implementation, the OpenEM does not consider using other parameters other than the 
 * execution object (eo), the error code (error) and the error scope (escope). 
 * 
 * @see em_error_handler_t(), em_register_error_handler(), em_unregister_error_handler, em_eo_register_error_handler, em_eo_unregister_error_handler()
 * 
 */

/** 
 *
 * @page DOC_trace Trace
 * @section section_201 mechanism
 * 
 * The trace handler prototype to be provided by the application to the OpenEM when activating the trace mechanism shall comply with the following prototype:
 * 
 * 
 * em_status_t (*ti_em_trace_handler_t)(ti_em_tscope_t tscope, ...)
 * 
 * 
 * In the curent implementation, the OpenEM assumes the trace of the following APIs to be associated to the following set of input parameters:
 * 
 * - void ti_em_preschedule(void)                                                          -> TI_EM_TSCOPE_PRESCHEDULE
 * 
 * - em_status_t ti_em_dispatch_once (void)                                                -> TI_EM_TSCOPE_DISPATCH
 * 
 * - void* ti_em_claim_local (void)                                                        -> TI_EM_TSCOPE_CLAIM_LOCAL, (void*)(buffer_ptr)
 * 
 * - em_event_t em_alloc(size_t size, em_event_type_t type, em_pool_id_t pool_id)          -> TI_EM_TSCOPE_ALLOC, (em_event_t)(event), (uint32_t)(size), (em_event_type_t)(type), (em_pool_id_t)(pool_id)
 * 
 * - void em_atomic_processing_end(void)                                                   -> TI_EM_TSCOPE_ATOMIC_PROCESSING_END
 * 
 * - void em_free(em_event_t event)                                                        -> TI_EM_TSCOPE_FREE, (em_event_t)(event)
 * 
 * - em_status_t em_send(em_event_t event, em_queue_t queue)                               -> TI_EM_TSCOPE_SEND, (em_event_t)(event), (em_queue_t)(queue)
 * 
 * - em_status_t em_send_group(em_event_t event, em_queue_t queue, em_event_group_t group) -> TI_EM_TSCOPE_SEND, (em_event_t)(event), (em_queue_t)(queue), (em_event_group_t)(eventGroup)
 * 
 * 
 * @see ti_em_trace_handler_t(), ti_em_register_trace_handler(), ti_em_unregister_trace_handler() 
 * 
 */

/* Basic EM types and HW configuration */
#include <ti/runtime/openem/event_machine_macros.h>
#include <ti/runtime/openem/event_machine_types.h>
#include <ti/runtime/openem/event_machine_functions.h>
#include <ti/runtime/openem/event_machine_group.h>
#include <ti/runtime/openem/event_machine_core_mask.h>
#include <ti/runtime/openem/event_machine_helper.h>

/* EM HW */
#include <ti/runtime/openem/event_machine_hw.h>

/* EM SW */
#include <ti/runtime/openem/event_machine_sw.h>

#ifdef __cplusplus
}
#endif

#endif /* EVENT_MACHINE_H */


