/*
 *   Copyright (c) 2012, Nokia Siemens Networks
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions are met:
 *       * Redistributions of source code must retain the above copyright
 *         notice, this list of conditions and the following disclaimer.
 *       * Redistributions in binary form must reproduce the above copyright
 *         notice, this list of conditions and the following disclaimer in the
 *         documentation and/or other materials provided with the distribution.
 *       * Neither the name of Nokia Siemens Networks nor the
 *         names of its contributors may be used to endorse or promote products
 *         derived from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *   DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
 *   DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
 

/**
 *
 * Contributors
 *
 * The initial developer of the original code is Nokia Siemens Networks.
 *  The following contributors have made changes to 
 *  the original code as described below.
 * 
 *----------------------------------------------------------------------------------------------
 * Copyright (c) 2012, Texas Instruments Incorporated. All Rights Reserved. - http://www.ti.com/
 * 
 * Changes were done on the original NSN header files in order to avoid 
 * circular dependencies when called in the TI implementation files.
 *
 * From the original event_machine_helper.h, the list of changes are:
 *   - The included files are now event_machine_types.h and event_machine_macros.h
 *   - Doxygen tags are added for API documentation generation.
 *       - Doxygen comments were copied from the 64-BIT implementation to the 32-BIT section
 *         since the TI implementation only support 32-BIT section.
 *----------------------------------------------------------------------------------------------
 *
 */


/**
 * @file: event_machine_core_mask.h
 *
 * @brief
 *    Event Machine core mask functions. This implementation is generic C - non HW optimized. 
 *
 */

#ifndef EVENT_MACHINE_CORE_MASK_H_
#define EVENT_MACHINE_CORE_MASK_H_

#ifdef __cplusplus
extern "C" {
#endif

#include <ti/runtime/openem/event_machine_macros.h>
#include <ti/runtime/openem/event_machine_types.h>




#if EM_CORE_MASK_SIZE != 64
#error Core mask functions support only 64 bit mask size
#endif



#ifdef EM_64_BIT
/*
 *
 * 64 bit versions.
 * --------------------------------------------
 */


/**
 * @ingroup DOC_api_generic
 *
 * Zero the whole mask.
 *
 * @param mask      Core mask
 */
static inline void em_core_mask_zero(em_core_mask_t* mask)
{
  mask->u64[0] = 0;
}


/**
 * @ingroup DOC_api_generic
 *
 * Set a bit in the mask.
 *
 * @param core      Core id
 * @param mask      Core mask
 */
static inline void em_core_mask_set(int core, em_core_mask_t* mask)
{
  mask->u64[0] |= ((uint64_t)1 << core);
}


/**
 * @ingroup DOC_api_generic
 *
 * Clear a bit in the mask.
 *
 * @param core      Core id
 * @param mask      Core mask
 */
static inline void em_core_mask_clr(int core, em_core_mask_t* mask)
{
  mask->u64[0] &= ~((uint64_t)1 << core);
}


/**
 * @ingroup DOC_api_generic
 *
 * Test if a bit is set in the mask.
 *
 * @param core      Core id
 * @param mask      Core mask
 *
 * @return Non-zero if core id is set in the mask
 */
static inline int em_core_mask_isset(int core, const em_core_mask_t* mask)
{
  return (mask->u64[0] & ((uint64_t)1 << core));
}


/**
 * @ingroup DOC_api_generic
 *
 * Test if the mask is all zero.
 *
 * @param mask      Core mask
 *
 * @return Non-zero if the mask is all zero
 */
static inline int em_core_mask_iszero(const em_core_mask_t* mask)
{
  return (mask->u64[0] == 0);
}


/**
 * @ingroup DOC_api_generic
 *
 * Test if two masks are equal
 *
 * @param mask1     First core mask
 * @param mask2     Second core mask
 *
 * @return Non-zero if the two masks are equal
 */
static inline int em_core_mask_equal(const em_core_mask_t* mask1, const em_core_mask_t* mask2)
{
  return (mask1->u64[0] == mask2->u64[0]);
}


/**
 * @ingroup DOC_api_generic
 *
 * Set a range (0...count-1) of bits in the mask.
 * 
 * @param count     Number of bits to set
 * @param mask      Core mask
 */
static inline void em_core_mask_set_count(int count, em_core_mask_t* mask)
{
  mask->u64[0] |= (((uint64_t)1 << count) - 1);
}


/**
 * @ingroup DOC_api_generic
 *
 * Copy core mask
 *
 * @param dest      Destination core mask
 * @param src       Source core mask
 */
static inline void em_core_mask_copy(em_core_mask_t* dest, const em_core_mask_t* src)
{
  dest->u64[0] = src->u64[0];
}


/**
 * @ingroup DOC_api_generic
 *
 * Count the number of bits set in the mask.
 *
 * @param mask      Core mask
 *
 * @return Number of bits set
 */
static inline int em_core_mask_count(const em_core_mask_t* mask)
{
  uint64_t n = mask->u64[0];
  int      cnt;
  

  for(cnt = 0; n; cnt++)
  {
    n &= (n - 1); // Clear the least significant bit set
  }
  
  return cnt;
}




#elif defined(EM_32_BIT)
/*
 *
 * 32 bit versions.
 * --------------------------------------------
 */


/**
 * @ingroup DOC_api_generic
 *
 * Zero the whole mask.
 *
 * @param mask      Core mask
 */
static inline void em_core_mask_zero(em_core_mask_t* mask)
{
  mask->u32[0] = 0;
  mask->u32[1] = 0;
}


/**
 * @ingroup DOC_api_generic
 *
 * Set a bit in the mask.
 *
 * @param core      Core id
 * @param mask      Core mask
 */
static inline void em_core_mask_set(int core, em_core_mask_t* mask)
{
  if(core < 32)
  {
    mask->u32[1] |= ((uint32_t)1 << core);
  }
  else
  {
    mask->u32[0] |= ((uint32_t)1 << (core - 32) );
  }
}


/**
 * @ingroup DOC_api_generic
 *
 * Clear a bit in the mask.
 *
 * @param core      Core id
 * @param mask      Core mask
 */
static inline void em_core_mask_clr(int core, em_core_mask_t* mask)
{
  if(core < 32)
  {
    mask->u32[1] &= ~((uint32_t)1 << core);
  }
  else
  {
    mask->u32[0] &= ~((uint32_t)1 << (core - 32) );
  }
}


/**
 * @ingroup DOC_api_generic
 *
 * Test if a bit is set in the mask.
 *
 * @param core      Core id
 * @param mask      Core mask
 *
 * @return Non-zero if core id is set in the mask
 */
static inline int em_core_mask_isset(int core, const em_core_mask_t* mask)
{
  if(core < 32)
  {
    return (mask->u32[1] & ((uint32_t)1 << core));
  }
  else
  {
    return (mask->u32[0] & ((uint32_t)1 << (core - 32)));
  }
}


/**
 * @ingroup DOC_api_generic
 *
 * Test if the mask is all zero.
 *
 * @param mask      Core mask
 *
 * @return Non-zero if the mask is all zero
 */
static inline int em_core_mask_iszero(const em_core_mask_t* mask)
{
  return ((mask->u32[0] == 0) && (mask->u32[1] == 0));
}


/**
 * @ingroup DOC_api_generic
 *
 * Test if two masks are equal
 *
 * @param mask1     First core mask
 * @param mask2     Second core mask
 *
 * @return Non-zero if the two masks are equal
 */
static inline int em_core_mask_equal(const em_core_mask_t* mask1, const em_core_mask_t* mask2)
{
  return ((mask1->u32[0] == mask2->u32[0]) && (mask1->u32[1] == mask2->u32[1]));
}


/**
 * @ingroup DOC_api_generic
 *
 * Set a range (0...count-1) of bits in the mask.
 * 
 * @param count     Number of bits to set
 * @param mask      Core mask
 */
static inline void em_core_mask_set_count(int count, em_core_mask_t* mask)
{
  if(count <= 32)
  {
    mask->u32[1] |= (((uint32_t)1 << count) - 1);
  }
  else
  {
    mask->u32[1] |= 0xffffffff;
    mask->u32[0] |= (((uint32_t)1 << (count - 32)) - 1);
  }
}


/**
 * @ingroup DOC_api_generic
 *
 * Copy core mask
 *
 * @param dest      Destination core mask
 * @param src       Source core mask
 */
static inline void em_core_mask_copy(em_core_mask_t* dest, const em_core_mask_t* src)
{
  dest->u32[0] = src->u32[0];
  dest->u32[1] = src->u32[1];
}


/**
 * @ingroup DOC_api_generic
 *
 * Count the number of bits set in the mask.
 *
 * @param mask      Core mask
 *
 * @return Number of bits set
 */
static inline int em_core_mask_count(const em_core_mask_t* mask)
{
  uint32_t n;
  int      cnt, i;
  
  for(i = 0, cnt = 0; i < 2; i++)
  {
    n = mask->u32[i];
    
    for(; n; cnt++) {
      n &= (n - 1); // clear the least significant bit set
    }
  }
  
  return cnt;
}

#endif




/*
 *  These mask functions could be added also
 *
 *  void em_core_mask_and(em_core_mask_t* dest, const em_core_mask_t* src1, const em_core_mask_t* src2);
 *  void em_core_mask_or(em_core_mask_t* dest, const em_core_mask_t* src1, const em_core_mask_t* src2);
 *  void em_core_mask_xor(em_core_mask_t* dest, const em_core_mask_t* src1, const em_core_mask_t* src2);
 *
 */




#ifdef __cplusplus
}
#endif


#endif /* EVENT_MACHINE_CORE_MASK_H_ */

