/*
 * Copyright (c) 2012, Texas Instruments Incorporated - http://www.ti.com/
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions are met:
 *       * Redistributions of source code must retain the above copyright
 *         notice, this list of conditions and the following disclaimer.
 *       * Redistributions in binary form must reproduce the above copyright
 *         notice, this list of conditions and the following disclaimer in the
 *         documentation and/or other materials provided with the distribution.
 *       * Neither the name of Texas Instruments Incorporated nor the
 *         names of its contributors may be used to endorse or promote products
 *         derived from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *   DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
 *   DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


/**
 * @file: event_machine_hw_ti_functions.h
 *
 * @brief
 *    Open Event Machine HW TI specific functions declarations
 *
 */

#ifndef EVENT_MACHINE_HW_TI_FUNCTIONS_H_
#define EVENT_MACHINE_HW_TI_FUNCTIONS_H_

#include <ti/runtime/openem/event_machine_group.h>
#include <ti/runtime/openem/event_machine_hw_ti_types.h>

/**
 * @ingroup DOC_api_specific
 *
 * Enable miss redirection for Ethertype/service type.
 *
 * @param miss_type     miss type enabled.
 *
 * @return status, EM_OK on success
 *
 */
em_status_t ti_em_xge_rx_miss_enable(ti_em_xge_rx_miss_type_t miss_type);

/**
 * @ingroup DOC_api_specific
 *
 * Disable miss redirection for Ethertype/service type.
 *
 * @param miss_type     miss type disabled.
 *
 * @return status, EM_OK on success
 *
 */
em_status_t ti_em_xge_rx_miss_disable(ti_em_xge_rx_miss_type_t miss_type);

/**
 * @ingroup DOC_api_specific
 *
 * Returns the buffer size.
 *
 * Returns the size of the event attached data buffer.
 *
 * @param event         Event handle.
 *
 * @return              Size of the attached buffer.
 *
 */
size_t ti_em_buffer_size(em_event_t event);

/**
 * @ingroup DOC_api_specific
 *
 * Returns the event payload size.
 *
 * Returns the size of the event attached data buffers.
 *
 * @param event         Event handle.
 *
 * @return              Size of the attached buffers.
 *
 */
size_t ti_em_event_size(em_event_t event);

/**
 * @ingroup DOC_api_specific
 *
 * Indicates that the scheduler can pre-schedule events to this core.
 *
 */
void ti_em_preschedule(void);

/**
 * @ingroup DOC_api_specific
 *
 * Indicates that the atomic processing locality has been achieved by the scheduler.
 *
 * @return              Hint if the atomic processing locality has been achieved.
 *
 */
uint32_t ti_em_atomic_processing_locality(void);

/**
 * @ingroup DOC_api_specific
 *
 * The function returns the needed size for the dispatcher control block in bytes. It is stored in the .tiEmLobal memory section.
 *
 * @return    size for the Dispatcher Control Block structure
 */
size_t ti_em_get_tcb_size (void);

/**
 * @ingroup DOC_api_specific
 *
 * The function returns the needed size for the master control block plus the runtime master control block in bytes. It is stored in the .tiEmGlobalFast memory section.
 *
 * @return    size for the Master Control Block structure
 */
size_t ti_em_get_pcb_size (void);

/**
 * @ingroup DOC_api_specific
 *
 * The function returns the needed size for the slow part of the EO descriptor in bytes. It is stored in the .tiEmGlobalSlow memory section preferably located in non cached MSMC RAM or DDR3 RAM.
 *
 * @return    size for the slow part of a EO structure
 */
size_t ti_em_get_eo_size_slow (void);

/**
 * @ingroup DOC_api_specific
 *
 * The function returns the needed size for the fast part of the EO descriptor in bytes. It is stored in the .tiEmGlobalFast memory section preferably located in non cached MSMC RAM or DDR3 RAM.
 *
 * @return    size for the fast part of a EO structure
 */
size_t ti_em_get_eo_size_fast (void);

/**
 * @ingroup DOC_api_specific
 *
 * The function returns the needed size for the slow part of the event group descriptor in bytes. It is stored in the .tiEmGlobalSlow memory section preferably located in non cached MSMC RAM or DDR3 RAM.
 *
 * @return    size for the slow part of a event group structure
 *
 */
size_t ti_em_get_event_group_size_slow (void);

/**
 * @ingroup DOC_api_specific
 *
 * The function returns the needed size for the fast part of the event group descriptor in bytes. It is stored in the .tiEmGlobalFast memory section preferably located in non cached MSMC RAM or DDR3 RAM.
 *
 * @return    size for the fast part of a event group structure
 *
 */
size_t ti_em_get_event_group_size_fast (void);

/**
 * @ingroup DOC_api_specific
 *
 * The function returns the needed size for the slow part of the queue group descriptor in bytes. It is stored in the .tiEmGlobalSlow memory section preferably located in non cached MSMC RAM or DDR3 RAM. 
 *
 * @return    size for the slow part of a queue group structure
 *
 */
size_t ti_em_get_queue_group_size_slow (void);

/**
 * @ingroup DOC_api_specific
 *
 * The function returns the needed size for the fast part of the queue group descriptor in bytes. It is stored in the .tiEmGlobalFast memory section preferably located in non cached MSMC RAM or DDR3 RAM. 
 *
 * @return    size for the fast part of a queue group structure
 *
 */
size_t ti_em_get_queue_group_size_fast (void);

/**
 * @ingroup DOC_api_specific
 *
 * The function returns the needed size for the slow part of the queue descriptor in bytes. It is stored in the .tiEmGlobalSlow memory section preferably located in non cached MSMC RAM or DDR3 RAM.
 *
 * @return    size for the slow part of a queue structure
 *
 */
size_t ti_em_get_queue_size_slow (void);

/**
 * @ingroup DOC_api_specific
 *
 * The function returns the needed size for the fast part of the queue descriptor in bytes. It is stored in the .tiEmGlobalFast memory section preferably located in non cached MSMC RAM or DDR3 RAM.
 *
 * @return    size for the fast part of a queue structure
 *
 */
size_t ti_em_get_queue_size_fast (void);

/**
 * @ingroup DOC_api_specific
 *
 * Global initialization of EM internals.
 *
 * Only one core does this and this must be called before any other call.
 *
 * @param config_ptr   Pointer to the hardware configuration structure.
 *
 * @return status, EM_OK on success
 */
em_status_t ti_em_init_global(const ti_em_config_t* config_ptr);

/**
 * @ingroup DOC_api_specific
 *
 * Local initialization of EM internals.
 *
 * All cores call this and it must be called after em_init_global(), but before any other call.
 * Implementation may be actually empty, but this might be needed later for some
 * core specific initializations, so application startup should call this always.
 *
 * @return status, EM_OK on success
 */
em_status_t ti_em_init_local(void);

/**
 * @ingroup DOC_api_specific
 *
 * Global shutdown of EM internals.
 *
 * Only one core does this and after this call, no other call is allowed.
 *
 * @return status, EM_OK on success
 */
em_status_t ti_em_exit_global (void);

/**
 * @ingroup DOC_api_specific
 *
 * The function runs the dispatcher once.
 *
 * The dispatcher code is responsible for calling a execution object receive API.
 * It checks once if a event has been scheduled to its core and calls the receive function.
 *
 * @return returns EM_ERR_NOT_FOUND if no receive function was executed.
 *         returns EM_OK when a public event was executed.
 */
em_status_t ti_em_dispatch_once (void);

/**
 * @ingroup DOC_api_specific
 *
 * The function returns the local event buffer pointer for a local event or 
 * NULL if there is no local event.
 *
 * @return        Pointer to the event buffer.
 *
 */
void* ti_em_claim_local (void);

/**
 * @ingroup DOC_api_specific
 *
 * Allocate a local event (for post-storing feature).
 *
 * Memory address of the allocated event is system specific and
 * can depend on given pool id, event size and type. Returned 
 * event (handle) may refer to a memory buffer or a HW specific 
 * descriptor, i.e. the event structure is system specific. 
 *
 * Use em_event_pointer() to convert an event (handle) to a pointer to 
 * the event structure. 
 *
 * EM_EVENT_TYPE_SW with minor type 0 is reserved for direct portability.
 * It is always guaranteed to return a 64-bit aligned contiguous 
 * data buffer, that can directly be used by the application up to
 * the given size (no HW specific descriptors etc are visible).
 *
 * @param size          Event size in octets
 * @param type          Event type to allocate
 *
 * @return the allocated event or EM_EVENT_UNDEF on an error
 *
 * @see em_free(), em_send(), em_event_pointer(), em_receive_func_t()
 */
em_event_t ti_em_alloc_local(size_t size, em_event_type_t type);

/**
 * @ingroup DOC_api_specific
 *
 * Helper function that converts an event handle into a packet.
 *
 * @param    event    Event handle.
 *
 * @return            Pointer to the packet.
 *
 */
ti_em_packet_t* ti_em_to_packet(em_event_t event);

/**
 * @ingroup DOC_api_specific
 *
 * Helper function that converts into an event handle from a packet.
 *
 * @param    packet_ptr      Pointer to the packet.
 *
 * @return                   Event handle.
 *
 */
em_event_t ti_em_from_packet(ti_em_packet_t* packet_ptr);

/**
 * @ingroup DOC_api_specific
 *
 * Helper function which sets the 12 first words of an event to their default values.
 *
 * The CPPI type is set to Host, The EPIB flag is ON and the PS location flag is ON.
 *
 * @param packet_ptr      Pointer to the packet.
 *
 */
void ti_em_packet_set_default(ti_em_packet_t* packet_ptr);

/**
 * @ingroup DOC_api_specific
 *
 * Helper function which sets the event pool information in the packet.
 *
 * @param packet_ptr    Pointer to the packet.
 * @param pool_config   Structure containing the pool info.
 *
 */
void ti_em_packet_set_pool_info(ti_em_packet_t* packet_ptr, ti_em_pool_config_t pool_config);

/**
 * @ingroup DOC_api_specific
 *
 * Helper function which sets the buffer information in the packet.
 *
 * @param packet_ptr    Pointer to the packet.
 * @param buffer_config Structure containing the buffer info.
 *
 */
void ti_em_packet_set_buffer_info(ti_em_packet_t* packet_ptr, ti_em_buffer_config_t buffer_config);

/**
 * @ingroup DOC_api_specific
 *
 * Helper function which restores the free information in the packet.
 *
 * @param packet_ptr     Pointer to the packet.
 *
 */
void ti_em_packet_restore_free_info(ti_em_packet_t * packet_ptr);

/**
 * @ingroup DOC_api_specific
 *
 * Helper function which sets the event type in the packet.
 *
 * @param     packet_ptr    Pointer to the packet.
 * @param     event_type    Event type of the packet.
 *
 */
void ti_em_packet_set_type(ti_em_packet_t* packet_ptr, em_event_type_t event_type);

/**
 * @ingroup DOC_api_specific
 *
 * Helper function which sets the queue information in the packet.
 *
 * @param packet_ptr    Pointer to the packet.
 * @param queue_hdl     Queue handle to set.
 *
 */
void ti_em_packet_set_queue(ti_em_packet_t* packet_ptr, em_queue_t queue_hdl);

/**
 * @ingroup DOC_api_specific
 *
 * Helper function which sets the event group information in the packet.
 *
 * @param packet_ptr         Pointer to the packet.
 * @param event_group_hdl    Event group handle to set.
 *
 */
void ti_em_packet_set_event_group(ti_em_packet_t* packet_ptr, em_event_group_t event_group_hdl);

/**
 * @ingroup DOC_api_specific
 *
 * Helper function which sets the event type in the event.
 *
 * @param event         Event handle.
 * @param event_type    Event type to set.
 *
 */
void ti_em_set_type(em_event_t event, em_event_type_t event_type);

/**
 * @ingroup DOC_api_specific
 *
 * Helper function which returns the event type.
 *
 * @param     event    Event handle.
 *
 * @return             Event type of the descriptor.
 *
 */
em_event_type_t ti_em_get_type(const em_event_t event);

/**
 * @ingroup DOC_api_specific
 *
 * Helper function which extracts the preload policy from the event type.
 *
 * @param     type     Event type.
 *
 * @return             Preload policy.
 *
 */
em_event_type_t ti_em_get_type_preload(em_event_type_t type);

/**
 * @ingroup DOC_api_specific
 *
 * Helper function which returns the buffer mode.
 *
 * @param     event    Event handle.
 *
 * @return             Buffer mode of the event.
 *
 */
ti_em_buf_mode_t ti_em_get_buf_mode(const em_event_t event);

/**
 * @ingroup DOC_api_specific
 *
 * Helper function which returns the cache coherency mode.
 *
 * @param     event    Event handle.
 *
 * @return             Cache coherency mode of the event.
 *
 */
ti_em_coh_mode_t ti_em_get_coh_mode(const em_event_t event);

/**
 * @ingroup DOC_api_specific
 *
 * Helper function which sets the queue and event group information in the event.
 *
 * @param event              Event handle.
 * @param queue_hdl          Queue handle to set.
 * @param event_group_hdl    Event group handle to set.
 *
 */
void ti_em_set_queue(em_event_t event, em_queue_t queue_hdl, em_event_group_t event_group_hdl);

/**
 * @ingroup DOC_api_specific
 *
 * Helper function which sets the queue information in the event tag location.
 *
 * @param tag_ptr            Pointer to tag location.
 * @param queue_hdl          Queue handle to set.
 *
 */
void ti_em_tag_set_queue(uint32_t* tag_ptr, em_queue_t queue_hdl);

/**
 * @ingroup DOC_api_specific
 *
 * Helper function which sets the event type in the tag location.
 *
 * @param tag_ptr            Pointer to tag location.
 * @param event_type         Event type to set.
 *
 */
void ti_em_tag_set_type(uint32_t* tag_ptr, em_event_type_t event_type);

/**
 * @ingroup DOC_api_specific
 *
 * Helper function which generates a global queue handle.
 *
 * @param     queue_hdl     Local queue handle.
 * @param     device_idx    Device index.
 * @param     process_idx   Process index.
 *
 * @return                  Global queue handle.
 *
 */
em_queue_t ti_em_queue_make_global(em_queue_t queue_hdl, ti_em_device_id_t device_idx, ti_em_process_id_t process_idx);

/**
 * @ingroup DOC_api_specific
 *
 * Helper function which extracts the device index from the queue handle.
 *
 * @param     queue_hdl     Queue handle.
 *
 * @return                  Device index.
 *
 */
ti_em_device_id_t ti_em_queue_get_device_id(em_queue_t queue_hdl);

/**
 * @ingroup DOC_api_specific
 *
 * Helper function which extracts the process index from the queue handle.
 *
 * @param     queue_hdl     Queue handle.
 *
 * @return                  Process index.
 *
 */
ti_em_process_id_t ti_em_queue_get_process_id(em_queue_t queue_hdl);

/**
 * @ingroup DOC_api_specific
 *
 * Helper function which extracts the queue mode (HW or SD) from the queue handle.
 *
 * @param     queue_hdl     Queue handle.
 *
 * @return                  Queue mode.
 *
 */
ti_em_queue_mode_t ti_em_queue_get_mode(em_queue_t queue_hdl);

/**
 * @ingroup DOC_api_specific
 *
 * Helper function which returns the index of the associated Multicore Navigator queue.
 *
 * @param     queue_hdl     Queue handle.
 *
 * @return                  Device index.
 *
 */
ti_em_queue_id_t ti_em_queue_get_queue_id(em_queue_t queue_hdl);

/**
 * @ingroup DOC_api_specific
 *
 * Registers the trace handler.
 *
 * @param handler   Trace handler.
 *
 * @return Success code.
 */
em_status_t ti_em_register_trace_handler(ti_em_trace_handler_t handler);

/**
 * @ingroup DOC_api_specific
 *
 * Unregisters the trace handler.
 *
 * @return Success code is 0,
 */
em_status_t ti_em_unregister_trace_handler(void);

/**
 * @ingroup DOC_api_specific
 *
 * Allocates an event while providing buffer(s) to attach.
 *
 * It allocates a primary event from pools with zero-buffer descriptors.\n
 * It attaches one or more buffers to the event.\n
 * The event comes with optional free function and coherency mode for each buffer.
 *
 * @param event_type         Event type to allocate
 * @param pool_id            Event pool id
 * @param buffer_num         Event buffers number
 * @param buffer_config_tbl  Data buffers configuration table
 *
 * @return                   Event handle
 *
 * @see ti_em_free_with_buffers()
 *
 */
em_event_t ti_em_alloc_with_buffers(em_event_type_t event_type, em_pool_id_t pool_id, int32_t buffer_num, ti_em_buffer_config_t* buffer_config_tbl);

/**
 * @ingroup DOC_api_specific
 *
 * Frees an event with buffers.
 *
 * It returns an event to its free pool(s).\n
 * It doesn't call free function, even if provided.\n
 * It fills out the configuration for each buffer that has been attached with ti_em_alloc_with_buffers().\n
 * It returns the number of such buffers - error if too many buffer pointers to return (wrt to buffer_num).\n
 *
 * @param event              Event to be freed
 * @param buffer_num         Event buffers number
 * @param buffer_config_tbl  Data buffers configuration table
 *
 * @return                   Event handle.
 *
 * @see ti_em_alloc_with_buffers()
 *
 */
em_event_t ti_em_free_with_buffers(em_event_t event, int32_t buffer_num, ti_em_buffer_config_t* buffer_config_tbl);

/**
 * @ingroup DOC_api_specific
 *
 * Combines two (2) event handles into one scattered event handle.\n
 * When returned, the iterator buffer points to last descriptor of the old pair.head_event.\n
 *
 * @param          event_pair   pair to combine.
 *
 * @param          iterator_ptr Pointer to iterator.
 *
 * @return         event handle.
 */
em_event_t ti_em_combine(ti_em_pair_t event_pair, ti_em_iterator_t * iterator_ptr);

/**
 * @ingroup DOC_api_specific
 *
 * Splits one scattered event into two events.\n
 * when returned, the iterator buffer points to the last buffer of pair.head_event (can be recombined right away).\n
 * Second and following buffers of pair.tail_event are cache coherent.\n
 *
 * @param   event           Event to split.
 * @param   iterator_ptr    Pointer to iterator where the split starts.
 *
 * @return                  Events pair.
 */
ti_em_pair_t ti_em_split(em_event_t event, ti_em_iterator_t* iterator_ptr);

/**
 * @ingroup DOC_api_specific
 *
 * Returns the size of the iterator attached data buffer.
 *
 * @param iterator_ptr      Pointer to iterator.
 *
 * @return                  Size of the attached buffer.
 *
 */
size_t ti_em_iterator_size(ti_em_iterator_t* iterator_ptr);

/**
 * @ingroup DOC_api_specific
 *
 * Initializes the iterator to the first buffer. Two iterators may not be started at the same time for a single event.
 *
 * @param event             Event handle to start.
 * @param iterator_ptr      Pointer to iterator.
 *
 * @return                  returns EM_OK.
 *
 */
em_status_t ti_em_iterator_start(em_event_t event, ti_em_iterator_t* iterator_ptr);

/**
 * @ingroup DOC_api_specific
 *
 * It registers all the touched buffers.\n
 * It is mandatory to perform ti_em_iterator_stop() before em_free(), em_send() or em_event_group_apply().\n
 * It is forbidden to use the iterator or any iterator content from this point onwards.
 *
 * @param iterator_ptr      Pointer to iterator.
 *
 * @return                  returns EM_OK when successfull, EM_ERR_BAD_STATE when iterator is in a wrong state.
 *
 */
em_status_t ti_em_iterator_stop(ti_em_iterator_t* iterator_ptr);

/**
 * @ingroup DOC_api_specific
 *
 * Moves the iterator to the next buffer.
 *
 * @param iterator_ptr      Pointer to iterator.
 *
 * @return                  returns EM_OK when successfull, EM_ERR_BAD_STATE when iterator is in a wrong state. EM_ERR_NOT_FOUND when the iterator points to the last buffer.
 *
 */
em_status_t ti_em_iterator_next(ti_em_iterator_t* iterator_ptr);

/**
 * @ingroup DOC_api_specific
 *
 * Moves the iterator to the previous buffer.
 *
 * @param iterator_ptr      Pointer to iterator.
 *
 * @return                  returns EM_OK when successfull, EM_ERR_BAD_STATE when iterator is in a wrong state. EM_ERR_NOT_FOUND when the iterator points to the first buffer.
 *
 */
em_status_t ti_em_iterator_previous(ti_em_iterator_t* iterator_ptr);

/**
 * @ingroup DOC_api_specific
 *
 * Returns a pointer to the iterator attached data buffer.
 *
 * @param iterator_ptr      Pointer to iterator.
 *
 * @return                  Pointer to the buffer.
 *
 */
void* ti_em_iterator_pointer(ti_em_iterator_t* iterator_ptr);

/**
 * @ingroup DOC_api_specific
 *
 * Opens an hard-ware queue on the data-base.
 *
 * @pre   The hard-ware queue shall not be used by another
 *        part of the application on the device.\n
 *        If the hard-ware queue is already used, the function
 *        is not successful.
 *
 * @remark The implementation is left to the user. An
 *         example is provided in em_pdk_hal.c file
 *         using the Qmss_queueOpen PDK function.
 *
 * @param queue_idx   Index of the queue to be open.\n
 *                    It ranges from 0 to the maximum queue index
 *                    supported by the Multicore Navigator.
 *
 * @post  None.
 *
 * @return status, EM_OK on success.
 */
extern em_status_t ti_em_hw_queue_open(int queue_idx);

/**
 * @ingroup DOC_api_specific
 *
 * Closes an hard-ware queue on the data-base.
 *
 * @pre   The hard-ware queue shall already be used by the
 *        Open-Em process.\n
 *        If the hard-ware queue is not used, the function
 *        is not successful.
 *
 * @remark The implementation is left to the user. An
 *         example is provided in em_pdk_hal.c file
 *         using the Qmss_queueClose PDK function.
 *
 * @param queue_idx   Index of the queue to be open.\n
 *                    It ranges from 0 to the maximum queue index
 *                    supported by the Multicore Navigator.
 *
 * @post  None.
 *
 * @return status, EM_OK on success.
 */
extern em_status_t ti_em_hw_queue_close(int queue_idx);

/**
 * @ingroup DOC_api_specific
 *
 * Opens an hard-ware receive flow on the data-base and configure the hard-ware registers of this receive flow.
 *
 * @pre   The hard-ware receive flow shall not be used by another
 *        part of the application on the device.\n
 *        If the hard-ware receive flow is already used, the function
 *        is not successful.
 *
 * @remark The implementation is left to the user. An
 *         example is provided in em_pdk_hal.c file
 *         using the Cppi_configureRxFlow PDK function.
 *
 * @param dma_idx         Index of the dma instance.\n
 *                        This refers to the Multicore Navigator PktDMA instance.
 * @param flow_idx        Index of the receive flow to be open.\n
 *                        It ranges from 0 to the maximum Rx Flow index
 *                        supported by the Multicore Navigator instance.
 * @param dst_queue_idx   Index of the destination queue.
 * @param free_queue_idx  Index of the destination free queue.\n
 *                        Queue indexes range from 0 to the maximum queue index
 *                        supported by the Multicore Navigator instance.
 * @param error_handling  Receive flow error handling mode when starvation occurs.\n
 *                        0 = Starvation errors result in dropping packet.\n
 *                        1 = Starvation errors result in subsequent re-try.
 *
 * @post  On success, the RX_FLOW_CONFIG registers shall be correctly configured.
 *
 * @return status, EM_OK on success
 */
extern em_status_t ti_em_rx_flow_open(int dma_idx, int flow_idx, int dst_queue_idx, int free_queue_idx, int error_handling);

/**
 * @ingroup DOC_api_specific
 *
 * Closes an hard-ware receive flow on the data-base.
 *
 * @pre   The hard-ware receive flow shall be used by the Open-EM process.\n
 *        If the hard-ware receive flow is not used, the function
 *        is not successful.
 *
 * @remark The implementation is left to the user. An
 *         example is provided in em_pdk_hal.c file
 *         using the Cppi_closeRxFlow PDK function.
 *
 * @param dma_idx         Index of the dma instance.\n
 *                        This refers to the Multicore Navigator PktDMA instance.
 * @param flow_idx        Index of the receive flow to be closed.\n
 *                        It ranges from 0 to the maximum Rx Flow index
 *                        supported by the Multicore Navigator instance.
 *
 * @return status, EM_OK on success
 */
extern em_status_t ti_em_rx_flow_close(int dma_idx, int flow_idx);

/**
 * @ingroup DOC_api_specific
 *
 * Opens an hard-ware xge receive flow on the data-base and configure the hard-ware registers of this receive flow.
 *
 * @pre   The hard-ware receive flow shall not be used by another
 *        part of the application on the device.\n
 *        If the hard-ware receive flow is already used, the function
 *        is not successful.
 *
 * @remark The implementation is left to the user. An
 *         example is provided in em_pdk_hal.c file
 *         using the Cppi_configureRxFlow PDK function.
 *
 * @param dma_idx         Index of the dma instance.\n
 *                        This refers to the Multicore Navigator PktDMA instance.
 * @param flow_idx        Index of the receive flow to be open.\n
 *                        It ranges from 0 to the maximum Rx Flow index
 *                        supported by the Multicore Navigator instance.
 * @param dst_queue_idx   Index of the destination queue.
 * @param free_queue_idx0 Index of the destination free queue.\n
 *                        Queue indexes range from 0 to the maximum queue index
 *                        supported by the Multicore Navigator instance.
 * @param free_queue_idx1 Index of the destination free queue.\n
 *                        Queue indexes range from 0 to the maximum queue index
 *                        supported by the Multicore Navigator instance.
 * @param error_handling  Receive flow error handling mode when starvation occurs.\n
 *                        0 = Starvation errors result in dropping packet.\n
 *                        1 = Starvation errors result in subsequent re-try.
 *
 * @post  On success, the RX_FLOW_CONFIG registers shall be correctly configured.
 *
 * @return status, EM_OK on success
 */
extern em_status_t ti_em_xge_rx_flow_open(int dma_idx, int flow_idx, int dst_queue_idx, int free_queue_idx0, int free_queue_idx1, int error_handling);

/**
 * @ingroup DOC_api_specific
 *
 * Closes an hard-ware XGE receive flow on the data-base.
 *
 * @pre   The hard-ware receive flow shall be used by the Open-EM process.\n
 *        If the hard-ware receive flow is not used, the function
 *        is not successful.
 *
 * @remark The implementation is left to the user. An
 *         example is provided in em_pdk_hal.c file
 *         using the Cppi_closeRxFlow PDK function.
 *
 * @param dma_idx         Index of the dma instance.\n
 *                        This refers to the Multicore Navigator PktDMA instance.
 * @param flow_idx        Index of the receive flow to be closed.\n
 *                        It ranges from 0 to the maximum Rx Flow index
 *                        supported by the Multicore Navigator instance.
 *
 * @return status, EM_OK on success
 */
extern em_status_t ti_em_xge_rx_flow_close (int dma_idx, int flow_idx);

/**
 * @ingroup DOC_api_specific
 *
 * Opens an hard-ware process RIO receive flow on the data-base and configure the hard-ware registers of this receive flow.
 * It also creates the mapping between the Type9 message parameters and the opened flow.
 *
 * @pre   The hard-ware receive flow shall not be used by another
 *        part of the application on the device.\n
 *        If the hard-ware receive flow is already used, the function
 *        is not successful.
 *
 * @remark The implementation is left to the user. An
 *         example is provided in em_pdk_hal.c file
 *         using the Cppi_configureRxFlow PDK function and the CSL_SRIO_MapType9MessageToQueue CSL API.
 *
 * @param dma_idx                Index of the dma instance.\n
 *                               This refers to the Multicore Navigator PktDMA instance.
 * @param flow_idx               Index of the receive flow to be open.\n
 *                               It ranges from 0 to the maximum Rx Flow index
 *                               supported by the Multicore Navigator instance.
 * @param dst_queue_idx          Index of the destination queue.
 * @param free_queue_idxSizeA    Index of the destination free queue for packet length\n 
 *                               smaller than sizeA. Queue indexes range from 0 to the 
 *                               maximum queue index supported by the Multicore Navigator 
 *                               instance.
 * @param sizeA                  Maximal size in bytes of the packet length for free queue A
 * @param free_queue_idxSizeB    Index of the destination free queue for packet length\n 
 *                               smaller than sizeB. Queue indexes range from 0 to the 
 *                               maximum queue index supported by the Multicore Navigator 
 *                               instance.
 * @param sizeB                  Maximal size in bytes of the packet length for free queue B
 * @param free_queue_idxSizeC    Index of the destination free queue for packet length\n 
 *                               smaller than sizeC. Queue indexes range from 0 to the 
 *                               maximum queue index supported by the Multicore Navigator 
 *                               instance.
 * @param sizeC                  Maximal size in bytes of the packet length for free queue C
 * @param free_queue_idxSizeD    Index of the destination free queue for packet length\n 
 *                               smaller than sizeD. Queue indexes range from 0 to the 
 *                               maximum queue index supported by the Multicore Navigator 
 *                               instance.
 * @param sizeD                  Maximal size in bytes of the packet length for free queue D
 * @param free_queue_idxOverflow Index of the destination free queue for packet length\n 
 *                               with overflow. Queue indexes range from 0 to the 
 *                               maximum queue index supported by the Multicore Navigator 
 *                               instance.
 * @param deviceIdx              Index of the current device
 * @param processIdx             Index of the current process
 * @param error_handling         Receive flow error handling mode when starvation occurs.\n
 *                               0 = Starvation errors result in dropping packet.\n
 *                               1 = Starvation errors result in subsequent re-try.
 *
 * @post  On success, the RX_FLOW_CONFIG registers shall be correctly configured.
 *
 * @return status, EM_OK on success
 */
extern em_status_t ti_em_rio_rx_flow_open(int dma_idx,
					  int flow_idx,
					  int dst_queue_idx,
					  int free_queue_idxSizeA,
					  int sizeA,
					  int free_queue_idxSizeB,
					  int sizeB,
					  int free_queue_idxSizeC,
					  int sizeC,
					  int free_queue_idxSizeD,
					  int sizeD,
					  int free_queue_idxOverflow,
					  int deviceIdx,
					  int processIdx,
					  int error_handling);
/**
 * @ingroup DOC_api_specific
 *
 * Opens an hard-ware process chaining receive flow on the data-base and configure the hard-ware registers of this receive flow.
 *
 * @pre   The hard-ware receive flow shall not be used by another
 *        part of the application on the device.\n
 *        If the hard-ware receive flow is already used, the function
 *        is not successful.
 *
 * @remark The implementation is left to the user. An
 *         example is provided in em_pdk_hal.c file
 *         using the Cppi_configureRxFlow PDK function.
 *
 * @param dma_idx                Index of the dma instance.\n
 *                               This refers to the Multicore Navigator PktDMA instance.
 * @param flow_idx               Index of the receive flow to be open.\n
 *                               It ranges from 0 to the maximum Rx Flow index
 *                               supported by the Multicore Navigator instance.
 * @param dst_queue_idx          Index of the destination queue.
 * @param free_queue_idxSizeA    Index of the destination free queue for packet length\n 
 *                               smaller than sizeA. Queue indexes range from 0 to the 
 *                               maximum queue index supported by the Multicore Navigator 
 *                               instance.
 * @param sizeA                  Maximal size in bytes of the packet length for free queue A
 * @param free_queue_idxSizeB    Index of the destination free queue for packet length\n 
 *                               smaller than sizeB. Queue indexes range from 0 to the 
 *                               maximum queue index supported by the Multicore Navigator 
 *                               instance.
 * @param sizeB                  Maximal size in bytes of the packet length for free queue B
 * @param free_queue_idxSizeC    Index of the destination free queue for packet length\n 
 *                               smaller than sizeC. Queue indexes range from 0 to the 
 *                               maximum queue index supported by the Multicore Navigator 
 *                               instance.
 * @param sizeC                  Maximal size in bytes of the packet length for free queue C
 * @param free_queue_idxSizeD    Index of the destination free queue for packet length\n 
 *                               smaller than sizeD. Queue indexes range from 0 to the 
 *                               maximum queue index supported by the Multicore Navigator 
 *                               instance.
 * @param sizeD                  Maximal size in bytes of the packet length for free queue D
 * @param free_queue_idxOverflow Index of the destination free queue for packet length\n 
 *                               with overflow. Queue indexes range from 0 to the 
 *                               maximum queue index supported by the Multicore Navigator 
 *                               instance.
 * @param error_handling         Receive flow error handling mode when starvation occurs.\n
 *                               0 = Starvation errors result in dropping packet.\n
 *                               1 = Starvation errors result in subsequent re-try.
 *
 * @post  On success, the RX_FLOW_CONFIG registers shall be correctly configured.
 *
 * @return status, EM_OK on success
 */
extern em_status_t ti_em_chain_rx_flow_open(int dma_idx,
					    int flow_idx,
					    int dst_queue_idx,
					    int free_queue_idxSizeA,
					    int sizeA,
					    int free_queue_idxSizeB,
					    int sizeB,
					    int free_queue_idxSizeC,
					    int sizeC,
					    int free_queue_idxSizeD,
					    int sizeD,
					    int free_queue_idxOverflow,
					    int error_handling);

/**
 * @ingroup DOC_api_specific
 *
 * Opens an hard-ware transmit channel on the data-base and configure the hard-ware registers of this transmit channel.
 *
 * @pre   The hard-ware transmit channel shall not be used by another
 *        part of the application on the device.\n
 *        If the hard-ware transmit channel is already used, the function
 *        is not successful.
 *
 * @remark The implementation is left to the user. An
 *         example is provided in em_pdk_hal.c file
 *         using the Cppi_txChannelOpen PDK function.
 *
 * @param dma_idx         Index of the dma instance.\n
 *                        This refers to the Multicore Navigator PktDMA instance.
 * @param channel_idx     Index of the transmit channel to be open.\n
 *                        It ranges from 0 to the maximum Tx Channel index
 *                        supported by the Multicore Navigator instance.
 *
 * @post  On success, the TX_CHANNEL_GLOBAL_CONFIG_REG_A register shall be enabled.\n
 *        The TX_CHANNEL_SCHEDULER_CONFIG_REG_PRIORITY register shall be configured to the desired value.
 *
 * @return status, EM_OK on success
 */
extern em_status_t ti_em_tx_channel_open(int dma_idx, int channel_idx);

/**
 * @ingroup DOC_api_specific
 *
 * Closes an hard-ware transmit channel on the data-base.
 *
 * @pre   The hard-ware transmit channel shall be used by the Open-EM
 *        process.\n
 *        If the hard-ware transmit channel is not used, the function
 *        is not successful.
 *
 * @remark The implementation is left to the user. An
 *         example is provided in em_pdk_hal.c file
 *         using the Cppi_channelClose PDK function.
 *
 * @param dma_idx         Index of the dma instance.\n
 *                        This refers to the Multicore Navigator PktDMA instance.
 * @param channel_idx     Index of the transmit channel to be closed.\n
 *                        It ranges from 0 to the maximum Tx Channel index
 *                        supported by the Multicore Navigator instance.
 *
 * @return status, EM_OK on success
 */
extern em_status_t ti_em_tx_channel_close(int dma_idx, int channel_idx);

/**
 * @ingroup DOC_api_specific
 *
 * Opens an hard-ware xge transmit channel on the data-base and configure the hard-ware registers of this transmit channel.
 *
 * @pre   The hard-ware transmit channel shall not be used by another
 *        part of the application on the device.\n
 *        If the hard-ware transmit channel is already used, the function
 *        is not successful.
 *
 * @remark The implementation is left to the user. An
 *         example is provided in em_pdk_hal.c file
 *         using the Cppi_txChannelOpen PDK function.
 *
 * @param dma_idx         Index of the dma instance.\n
 *                        This refers to the Multicore Navigator PktDMA instance.
 * @param channel_idx     Index of the transmit channel to be open.\n
 *                        It ranges from 0 to the maximum Tx Channel index
 *                        supported by the Multicore Navigator instance.
 *
 * @post  On success, the TX_CHANNEL_GLOBAL_CONFIG_REG_A register shall be enabled.\n
 *        The TX_CHANNEL_SCHEDULER_CONFIG_REG_PRIORITY register shall be configured to the desired value.
 *
 * @return status, EM_OK on success
 */
extern em_status_t ti_em_xge_tx_channel_open(int dma_idx, int channel_idx);

/**
 * @ingroup DOC_api_specific
 *
 * Closes an hard-ware XGE transmit channel on the data-base.
 *
 * @pre   The hard-ware transmit channel shall be used by the Open-EM
 *        process.\n
 *        If the hard-ware transmit channel is not used, the function
 *        is not successful.
 *
 * @remark The implementation is left to the user. An
 *         example is provided in em_pdk_hal.c file
 *         using the Cppi_channelClose PDK function.
 *
 * @param dma_idx         Index of the dma instance.\n
 *                        This refers to the Multicore Navigator PktDMA instance.
 * @param channel_idx     Index of the transmit channel to be closed.\n
 *                        It ranges from 0 to the maximum Tx Channel index
 *                        supported by the Multicore Navigator instance.
 *
 * @return status, EM_OK on success
 */
extern em_status_t ti_em_xge_tx_channel_close (int dma_idx, int channel_idx);

/**
 * @ingroup DOC_api_specific
 *
 * Opens an hard-ware SRIO transmit channel on the data-base and configure the hard-ware registers of this transmit channel.
 *
 * @pre   The hard-ware transmit channel shall not be used by another
 *        part of the application on the device.\n
 *        If the hard-ware transmit channel is already used, the function
 *        is not successful.
 *
 * @remark The implementation is left to the user. An
 *         example is provided in em_pdk_hal.c file
 *         using the Cppi_txChannelOpen PDK function.
 *
 * @param dma_idx         Index of the dma instance.\n
 *                        This refers to the Multicore Navigator PktDMA instance.
 *
 * @post  On success, the TX_CHANNEL_GLOBAL_CONFIG_REG_A register shall be enabled.\n
 *        The TX_CHANNEL_SCHEDULER_CONFIG_REG_PRIORITY register shall be configured to the desired value.
 *
 * @return status, EM_OK on success
 */
extern em_status_t ti_em_rio_tx_channel_open(int dma_idx);


/**
 * @ingroup DOC_api_specific
 *
 * Opens an hard-ware receive channel on the data-base and configure the hard-ware registers of this receive channel.
 *
 * @pre   The hard-ware receive channel shall not be used by another
 *        part of the application on the device.\n
 *        If the hard-ware receive channel is already used, the function
 *        is not successful.
 *
 * @remark The implementation is left to the user. An
 *         example is provided in em_pdk_hal.c file
 *         using the Cppi_rxChannelOpen PDK function.
 *
 * @param dma_idx         Index of the dma instance.\n
 *                        This refers to the Multicore Navigator PktDMA instance.
 * @param channel_idx     Index of the receive channel to be open.\n
 *                        It ranges from 0 to the maximum Rx Channel index
 *                        supported by the Multicore Navigator instance.
 *
 * @post  On success, the RX_CHANNEL_GLOBAL_CONFIG_REG register shall be enabled.
 *
 * @return status, EM_OK on success
 */
extern em_status_t ti_em_rx_channel_open(int dma_idx, int channel_idx);

/**
 * @ingroup DOC_api_specific
 *
 * Closes an hard-ware receive channel on the data-base.
 *
 * @pre   The hard-ware receive channel shall be used by the
 *        Open-EM process.\n
 *        If the hard-ware receive channel is not used, the function
 *        is not successful.
 *
 * @remark The implementation is left to the user. An
 *         example is provided in em_pdk_hal.c file
 *         using the Cppi_channelClose PDK function.
 *
 * @param dma_idx         Index of the dma instance.\n
 *                        This refers to the Multicore Navigator PktDMA instance.
 * @param channel_idx     Index of the receive channel to be closed.\n
 *                        It ranges from 0 to the maximum Rx Channel index
 *                        supported by the Multicore Navigator instance.
 *
 * @return status, EM_OK on success
 */
extern em_status_t ti_em_rx_channel_close(int dma_idx, int channel_idx);

/**
 * @ingroup DOC_api_specific
 *
 * Opens an hard-ware xge receive channel on the data-base and configure the hard-ware registers of this receive channel.
 *
 * @pre   The hard-ware receive channel shall not be used by another
 *        part of the application on the device.\n
 *        If the hard-ware receive channel is already used, the function
 *        is not successful.
 *
 * @remark The implementation is left to the user. An
 *         example is provided in em_pdk_hal.c file
 *         using the Cppi_rxChannelOpen PDK function.
 *
 * @param dma_idx         Index of the dma instance.\n
 *                        This refers to the Multicore Navigator PktDMA instance.
 * @param channel_idx     Index of the receive channel to be open.\n
 *                        It ranges from 0 to the maximum Rx Channel index
 *                        supported by the Multicore Navigator instance.
 *
 * @post  On success, the RX_CHANNEL_GLOBAL_CONFIG_REG register shall be enabled.
 *
 * @return status, EM_OK on success
 */
extern em_status_t ti_em_xge_rx_channel_open(int dma_idx, int channel_idx);

/**
 * @ingroup DOC_api_specific
 *
 * Closes an hard-ware XGE receive channel on the data-base.
 *
 * @pre   The hard-ware receive channel shall be used by the
 *        Open-EM process.\n
 *        If the hard-ware receive channel is not used, the function
 *        is not successful.
 *
 * @remark The implementation is left to the user. An
 *         example is provided in em_pdk_hal.c file
 *         using the Cppi_channelClose PDK function.
 *
 * @param dma_idx         Index of the dma instance.\n
 *                        This refers to the Multicore Navigator PktDMA instance.
 * @param channel_idx     Index of the receive channel to be closed.\n
 *                        It ranges from 0 to the maximum Rx Channel index
 *                        supported by the Multicore Navigator instance.
 *
 * @return status, EM_OK on success
 */
extern em_status_t ti_em_xge_rx_channel_close (int dma_idx, int channel_idx);

/**
 * @ingroup DOC_api_specific
 *
 * Opens an hard-ware xge transmit queue on the data-base and configure the hard-ware registers of this transmit queue.
 *
 * @pre   The hard-ware transmit queue shall not be used by another
 *        part of the application on the device.\n
 *        If the hard-ware transmit queue is already used, the function
 *        is not successful.
 *
 * @remark The implementation is left to the user. An
 *         example is provided in em_pdk_hal.c file
 *         using the ti_em_hw_queue_open function.
 *
 * @param queue_base_idx          Index of the 1st XGE tx queue
 * @param vlan_priority           VLAN priority.
 *
 * @return index of the XGE transmit queue
 */
extern ti_em_queue_id_t ti_em_xge_tx_queue_open(ti_em_queue_id_t queue_base_idx, int    vlan_priority);

/**
 * @ingroup DOC_api_specific
 *
 * Returns the first hard-ware xge transmit queue among TI_EM_XGE_VLAN_PRIO_NUM (8).
 *
 * @pre   The hard-ware transmit queue shall not be used by another
 *        part of the application on the device.\n
 *        If the hard-ware transmit queue is already used, the function
 *        is not successful.
 *
 * @remark The implementation is left to the user. An
 *         example is provided in em_pdk_hal.c file
 *         using the ti_em_xge_tx_queue_base_idx_get function.
 *
 * @param .
 *
 * @return index of the first XGE transmit queue
 */
extern ti_em_queue_id_t ti_em_xge_tx_queue_base_idx_get (void);

/**
 * @ingroup DOC_api_specific
 *
 * Opens an hard-ware SRIO transmit queue on the data-base and configure the hard-ware registers of this transmit queue.
 *
 * @pre   The hard-ware transmit queue shall not be used by another
 *        part of the application on the device.\n
 *        If the hard-ware transmit queue is already used, the function
 *        is not successful.
 *
 * @remark The implementation is left to the user. An
 *         example is provided in em_pdk_hal.c file
 *         using the ti_em_hw_queue_open function.
 *
 * @param rio_tx_queue_idx           RIO queue index
 *
 * @return index of the RIO transmit queue or error
 */
extern ti_em_queue_id_t ti_em_rio_tx_queue_open(int rio_tx_queue_idx);


/**
 * @ingroup DOC_api_specific
 *
 * Retrieves the absolute queue index from the associated Multicore Navigator PkDMA engine index
 * and the relative TX queue index.
 *
 * @remark The implementation is left to the user. An
 *         example is provided in em_pdk_hal.c file
 *         using the CPPI LLD.
 *
 * @param dma_idx         Index of the dma instance.\n
 *                        This refers to the Multicore Navigator PktDMA instance.
 *
 * @param queue_idx       Relative index of the queue.
 *
 * @return Absolute index of the queue.
 */
extern ti_em_queue_id_t ti_em_get_absolute_queue_id(ti_em_dma_id_t dma_idx, ti_em_queue_id_t queue_idx);

/**
 * @ingroup DOC_api_specific
 *
 * Loops over all data buffers of the event and perform a cache write back invalidate if the data buffer is dirty.
 *
 * @param event           Event handle.
 *
 */
void ti_em_flush(em_event_t event);

/**
 * @ingroup DOC_api_specific
 *
 * Helper function which sets the ps words in the event.
 *
 * @param event         Event handle.
 * @param ps_word_ptr   Pointer to ps words.
 * @param ps_wsize      Number of ps words.
 *
 * @return status, EM_OK on success
 */
em_status_t ti_em_set_ps_words(em_event_t event, uint32_t* ps_word_ptr, size_t ps_wsize);

/**
 * @ingroup DOC_api_specific
 *
 * Helper function which gets the ps words from the event.
 *
 * @param event         Event handle.
 * @param ps_word_ptr   Pointer to ps words.
 * @param ps_wsize      Number of ps words.
 *
 * @return status, EM_OK on success
 */
em_status_t ti_em_get_ps_words(em_event_t event, uint32_t* ps_word_ptr, size_t ps_wsize);

/**
 * @ingroup DOC_api_specific
 *
 * Helper function which sets the number of ps words in the event.
 *
 * @param event         Event handle.
 * @param ps_wsize      Number of ps words.
 *
 * @return status, EM_OK on success
 */
em_status_t ti_em_set_ps_wsize(em_event_t event, size_t ps_wsize);

/**
 * @ingroup DOC_api_specific
 *
 * Helper function which gets the number of ps words from the event.
 *
 * @param event         Event handle.
 *
 * @return Number of ps words
 */
size_t ti_em_get_ps_wsize(em_event_t event);

/**
 * @ingroup DOC_api_specific
 *
 * Adds a process route to the device router.
 *
 * @param process_route  Parameters of the process route.
 *
 * @param process_idx  Process index.
 *
 * @return status, EM_OK on success
 */
em_status_t ti_em_process_add_route(ti_em_process_route_t process_route, ti_em_process_id_t process_idx);

/**
 * @ingroup DOC_api_specific
 *
 * Adds an XGE route to the device router.
 *
 * @param device_idx  Device index.
 *
 * @param device_route  Device route.
 *
 * @return status, EM_OK on success
 */
em_status_t ti_em_device_add_xge_route(ti_em_device_id_t device_idx, ti_em_device_xge_route_t device_route);

/**
 * @ingroup DOC_api_specific
 *
 * Adds an SRIO route to the device router.
 *
 * @param device_idx  Device index.
 *
 * @param device_route  Device route.
 *
 * @return status, EM_OK on success
 */
em_status_t ti_em_device_add_rio_route(ti_em_device_id_t device_idx, ti_em_device_rio_route_t device_route);

/**
 * @ingroup DOC_api_specific
 *
 * Creates a static HW queue.
 *
 * @param name          Queue name for debugging purposes (optional, NULL ok)
 *
 * @param hw_queue_idx  Hw queue index
 *
 * @param queue         Requested queue id from the static range
 *
 * @return  EM_OK on success or EM_QUEUE_UNDEF on an error
 */
em_status_t ti_em_queue_create_hw_static(const char* name, ti_em_queue_id_t hw_queue_idx, em_queue_t queue);

/**
 * @ingroup DOC_api_specific
 *
 * Creates a HW queue.
 *
 * @param name          Queue name for debugging purposes (optional, NULL ok)
 *
 * @param hw_queue_idx  Hw queue index
 *
 * @return new queue id or EM_QUEUE_UNDEF on an error
 */
em_queue_t ti_em_queue_create_hw(const char* name, ti_em_queue_id_t hw_queue_idx);

/**
 * @ingroup DOC_api_specific
 *
 * Receives an event from a queue.
 *
 * Event must have been allocated with em_alloc().
 *
 * @param queue         Receiving queue
 *
 * @return the received event or EM_EVENT_UNDEF if the queue is empty.
 * 
 * @see em_alloc()
 */
em_event_t ti_em_receive(em_queue_t queue);

/**
 * @ingroup DOC_api_specific
 *
 * Enables the interrupt generation for the attached dispatcher.
 * The interrupt channel index used is based on the configuration
 * hw_queue_base_idx and the physical core index.
 * If the interrupt channel index is inferior to 32, it is mapped to
 * one of the 32 HI channels in the associated INTD register.
 * Otherwise it is mapped to one of the 16 LOW channels.
 * By default, the interrupt are not enabled.
 *
 * @return status, EM_OK on success
 */
em_status_t ti_em_interrupt_enable(void);

/**
 * @ingroup DOC_api_specific
 *
 * Disables the interrupt generation for the attached dispatcher.
 *
 * @return status, EM_OK on success
 */
em_status_t ti_em_interrupt_disable(void);

/**
 * @ingroup DOC_api_specific
 *
 * Returns the value of the counter specified by the "type" parameter. 
 * If the "flag" parameter is set to 1, the counter is read and then reset to 0
 * in the API call. The returned counter value is written at the "value_ptr" 
 * address.
 *
 * @param type         type
 *
 * @param flag         reset flag; if set to 1, the counter is reset to 0.
 *
 * @param value_ptr    pointer to the variable where to write the counter value.
 *
 * @return status, EM_OK on success
 */
em_status_t ti_em_counter_get(
    ti_em_counter_type_t type, 
    ti_em_counter_value_t* value_ptr,
    int flag);

#endif /* EVENT_MACHINE_HW_TI_FUNCTIONS_H_ */

