/*
 *   Copyright (c) 2012, Nokia Siemens Networks
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions are met:
 *       * Redistributions of source code must retain the above copyright
 *         notice, this list of conditions and the following disclaimer.
 *       * Redistributions in binary form must reproduce the above copyright
 *         notice, this list of conditions and the following disclaimer in the
 *         documentation and/or other materials provided with the distribution.
 *       * Neither the name of Nokia Siemens Networks nor the
 *         names of its contributors may be used to endorse or promote products
 *         derived from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *   DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
 *   DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


/**
 *
 * Contributors
 *
 * The initial developer of the original code is Nokia Siemens Networks.
 *  The following contributors have made changes to 
 *  the original code as described below.
 * 
 *----------------------------------------------------------------------------------------------
 * Copyright (c) 2012, Texas Instruments Incorporated. All Rights Reserved. - http://www.ti.com/
 * 
 * Changes were done on the original NSN header files in order to avoid 
 * circular dependencies when called in the TI implementation files.
 *
 * This file is a merge of the type definitions from the original
 * event_machine.h and event_machine_types.h files.
 * The list of changes are:
 *   - Doxygen tags are added for API documentation generation.
 *----------------------------------------------------------------------------------------------
 *
 */


/**
 * @file: event_machine_types.h
 *
 * @brief
 *    Open Event Machine basic types
 */

#ifndef EVENT_MACHINE_TYPES_H_
#define EVENT_MACHINE_TYPES_H_

#ifdef __cplusplus
extern "C" {
#endif


#include <inttypes.h>
#include <stdarg.h>
#include <stdlib.h>
#include <ti/runtime/openem/event_machine_macros.h>
#include <ti/runtime/openem/event_machine_hw_types.h>

#ifdef EM_64_BIT

/*
 * Optimize EM for 64 bit architecture
 * =====================================
 */

 /* page  64 bit version */
/**
 *
 * @page DOC_intro Introduction
 * @section page_version 64 bit version
 * This is documentation represent the 64 bit version of Event Machine API.
 * Define EM_64_BIT or EM_32_BIT to select between 64 and 32 bit versions.
 *
 */


/*
 * Printf formats
 */


/**
 * @ingroup DOC_api_generic
 *
 *  Execution Object identifier
 *
 * @see em_eo_create() 
 */
typedef uint64_t em_eo_t;

/**
 * @ingroup DOC_api_generic
 *
 * Queue identifier
 *
 * @see em_queue_create(), em_receive_func_t(), em_send()
 */
typedef uint64_t em_queue_t;


/**
 * @ingroup DOC_api_generic
 *
 * Queue group identifier
 *
 * Each queue belongs to one queue group, that defines a core mask for scheduling
 * events, i.e. define which cores participate in the load balancing. Group can
 * also allow only a single core for no load balancing.
 *
 * Groups needs to be created as needed. One default group (EM_QUEUE_GROUP_DEFAULT)
 * always exists, and that allows scheduling to all the cores running this execution
 * binary instance.
 *
 * @see em_queue_group_create()
 */
typedef uint64_t em_queue_group_t;




#elif defined(EM_32_BIT)
/*
 * Optimize EM for 32 bit architecture
 * =====================================
 */

 /* @page page_version 32 bit version */

/**
 *
 * @page DOC_intro Introduction
 * @section page_version 32 bit version
 * This is documentation represent the 32 bit version of Event Machine API.
 * Define EM_64_BIT or EM_32_BIT to select between 64 and 32 bit versions.
 *
 */


/* ************* 32 bit ************************ */




/**
 * @ingroup DOC_api_generic
 *
 * Execution Object identifier
 *
 * @see em_eo_create() 
 */
typedef uint32_t em_eo_t;

/**
 * @ingroup DOC_api_generic
 *
 * Queue identifier
 *
 * @see em_queue_create(), em_receive_func_t(), em_send()
 */
typedef uint32_t em_queue_t;

/**
 * @ingroup DOC_api_generic
 *
 * Queue group identifier
 *
 * Each queue belongs to one queue group, that defines a core mask for scheduling
 * events, i.e. define which cores participate in the load balancing. Group can
 * also allow only a single core for no load balancing.
 *
 * Groups needs to be created as needed. One default group (EM_QUEUE_GROUP_DEFAULT)
 * always exists, and that allows scheduling to all the cores running this execution
 * binary instance.
 *
 * @see em_queue_group_create()
 */
typedef uint32_t em_queue_group_t;



#else

  #error Missing architecture definition. Define EM_64_BIT or EM_32_BIT!

  /**
   *
   * @page page_version 64/32 bit version not selected
   * This is documentation has not selected between 64/32 bit version of Event Machine API.
   * Some types might be missing. Define EM_64_BIT or EM_32_BIT to select between 64 and 32 bit versions.
   *
   */
   
#endif



/**
 * @ingroup DOC_api_generic
 *
 * Error/Status code.
 * EM_OK (0) is the general code for success, other values
 * describe failed operation.
 * 
 * @see event_machine_hw_config.h, em_error_handler_t(), em_error()
 */
typedef uint32_t em_status_t;



/**
 * @ingroup DOC_api_generic
 *
 * Error scope.
 *
 * Identifies the error scope for interpreting error codes and variable arguments.
 * 
 * @see em_error_handler_t(), em_error()
 */
typedef uint32_t em_escope_t;


/**
 * @ingroup DOC_api_generic
 *
 * Queue type.
 *
 * Affects the scheduling principle
 *
 * @see em_queue_create(), event_machine_hw_config.h
 */
typedef uint32_t em_queue_type_t;


/**
 * @ingroup DOC_api_generic
 *
 * Queue priority class
 *
 * Queue priority defines a system dependent QoS class, not just an absolute
 * priority. EM gives freedom to implement the actual scheduling disciplines
 * and the corresponding numeric values as needed, i.e. the actual values are
 * system dependent and thus not portable, but the 5 pre-defined enums
 * (em_queue_prio_e) are always valid.
 * Application platform or middleware needs to define and distribute the
 * other available values.
 *
 * @see em_queue_create(), event_machine_hw_config.h
 */
typedef uint32_t em_queue_prio_t;


/**
 * @ingroup DOC_api_generic
 *
 * Event type. This is given to application for each received
 * event and also needed for event allocation.
 * It's an integer, but split into major and minor part.
 * major-field categorizes the event and minor is
 * more detailed system specific description.
 * Major-part will not change by HW, but minor can be HW/SW platform
 * specific and thus could be split into more sub-fields as needed.
 * Application should use the access functions for reading major 
 * and minor part.
 *
 * The only event type with defined content is EM_EVENT_TYPE_SW with
 * minor type 0, which needs to be portable (direct pointer to data).
 *
 * @see em_get_type_major(), em_get_type_minor(), em_receive_func_t()
 */
typedef uint32_t em_event_type_t;


/**
 * @ingroup DOC_api_generic
 *
 * Memory pool id.
 *
 * Defines memory pool in em_alloc(). Default pool id is defined by
 * EM_POOL_DEFAULT.
 *
 * @see em_alloc(), event_machine_hw_config.h
 */
typedef uint32_t em_pool_id_t;

/**
 * @ingroup DOC_api_generic
 *
 * Type for queue group core mask.
 * Each bit represents one core, core 0 is the lsb (1 << em_core_id())
 * Note, that EM will enumerate the core identifiers to always start from 0 and
 * be contiguous meaning the core numbers are not necessarily physical.
 * This type can handle up to 64 cores.
 * 
 * @see em_queue_group_create()
 */
typedef union em_core_mask_t em_core_mask_t;

union em_core_mask_t
{
  uint8_t   u8[EM_CORE_MASK_SIZE / 8];
  uint16_t u16[EM_CORE_MASK_SIZE / 16];
  uint32_t u32[EM_CORE_MASK_SIZE / 32];
  uint64_t u64[EM_CORE_MASK_SIZE / 64];
};





/**
 * \internal
 *--------------------------------------------------------------------------
 * Copyright (c) 2012, Texas Instruments Incorporated - http://www.ti.com/
 * This part of file is a copy from the original event_machine.h 
 *--------------------------------------------------------------------------
 * \endinternal
 */



/**
 * @ingroup DOC_api_generic
 *
 * Notification structure allows user to define a notification event and
 * destination queue pair. EM will notify user by sending the event into
 * the queue.
 */
typedef struct em_notif_t
{
  em_event_t  event;  /**< User defined notification event */
  em_queue_t  queue;  /**< Destination queue */

} em_notif_t;



/*
 *
 * From Event Machine to application interface (call-back functions)
 * ----------------------------------------------------------------------------
 * 
 */


/**
 * @ingroup DOC_api_generic
 *
 * Receive event
 *
 * Application receives events through the EO receive function. It implements main part of
 * the application logic. EM calls the receive function when it has dequeued an event from one
 * of the EO's queues. Application processes the event and returns immediately (in 
 * run-to-completion fashion).
 *
 * On multicore systems several events (from the same or different queues) may be dequeued
 * in parallel and thus same receive function may be executed concurrently on several cores.
 * Parallel execution may be limited by queue group setup or using queues with atomic
 * scheduling mode.
 *
 * EO and queue context pointers are user defined (in EO and queue creation time) and 
 * may be used any way needed. For example, EO context may be used to store global EO state
 * information, which is common to all queues and events. In addition, queue context may be
 * used to store queue (or user data flow) specific state data. EM never touches the data, 
 * just passes the pointers. 
 *
 * Event (handle) must be converted to an event structure pointer with em_event_pointer().
 * Event type specifies the event structure, which is implementation or application specific.
 * Queue id specifies the queue which the event was dequeued from.
 *
 * @param eo_ctx        EO context data. The pointer is passed in em_eo_create(), EM does not touch the data.
 * @param event         Event handle
 * @param type          Event type
 * @param queue         Queue from which this event came from
 * @param q_ctx         Queue context data. The pointer is passed in em_queue_set_context(), EM does not touch the data.
 *
 * @see em_event_pointer(), em_free(), em_alloc(), em_send(), em_queue_set_context(), em_eo_create()
 */
typedef void (*em_receive_func_t)(void* eo_ctx, em_event_t event, em_event_type_t type, em_queue_t queue, void* q_ctx);

/**
 * @ingroup DOC_api_generic
 *
 * Execution object start, global.
 * 
 * If load balancing/several cores share the EO, this function is called
 * once on one core only (any). Purpose of this global start is to provide
 * a placeholder for first level initialization, like allocating memory and
 * initializing shared data. After this global start returns, the core local
 * version (if defined), is called (see em_start_local_func_t below). If there
 * is no core local start, then event dispatching is immediately enabled.
 *
 * If Execution object does not return EM_OK,
 * the system will not call the core local init and will not enable event
 * dispatching.
 *
 * This function should never be called directly from the application,
 * but using the, which maintains state information!
 *
 * @param eo_ctx        Execution object internal state/instance data
 * @param eo            Execution object id
 *
 * @return EM_OK if successful.
 *
 * @see em_eo_start(), em_eo_create()
 */
typedef em_status_t (*em_start_func_t)(void* eo_ctx, em_eo_t eo);

/**
 * @ingroup DOC_api_generic
 *
 * Execution object start, core local.
 * 
 * This is similar to the global start above, but this one is called after the
 * global start has completed and is called on all cores.
 *
 * Purpose of this optional local start is to work as a placeholder for
 * core local initialization, e.g. allocating core local memory for example.
 * The global start is only called on one core. The use of local start is
 * optional.
 * Note, that application should never directly call this function, this
 * will be called via em_eo_start().
 *
 * If this does not return EM_OK on all involved cores, the event dispatching
 * is not enabled.
 *
 * @param eo_ctx        Execution object internal state/instance data
 * @param eo            Execution object id
 *
 * @return EM_OK if successful.
 *
 * @see em_eo_start(), em_eo_create()
 */
typedef em_status_t (*em_start_local_func_t)(void* eo_ctx, em_eo_t eo);

/**
 * @ingroup DOC_api_generic
 *
 * Execution object stop, core local.
 *
 * If load balancing/several cores share the EO, this function is
 * called once on each core before the global stop (reverse order of start).
 * System disables event dispatching before calling this
 * and also makes sure this does not get called before the core
 * has been notified the stop condition for this EO (won't dispatch new events)
 * Return value is only for logging purposes, EM does not use it currently.
 *
 * Note, that application should never directly call this stop function,
 * em_eo_stop() will trigger this.
 *
 * @param eo_ctx        Execution object internal state data
 * @param eo            Execution object id
 *
 * @return EM_OK if successful.
 *
 * @see em_eo_stop(), em_eo_create()
 */
typedef em_status_t (*em_stop_local_func_t)(void* eo_ctx, em_eo_t eo);

/**
 * @ingroup DOC_api_generic
 *
 * Execution object stop, global.
 *
 * If load balancing/several cores share the EO, this function is
 * called once on one core (any) after the (optional) core local
 * stop return on all cores.
 * System disables event dispatching before calling this and also
 * makes sure this does not get called before all cores have
 * been notified the stop condition for this EO (can't dispatch new events)
 * event if there is no core local stop defined.
 * Return value is only for logging purposes, EM does not use it currently.
 *
 * Note, that application should never directly call this stop function,
 * but use the em_eo_stop() instead, which maintains state information and
 * guarantees synchronized operation.
 *
 * @param eo_ctx        Execution object internal state data
 * @param eo            Execution object id
 *
 * @return EM_OK if successful.
 *
 * @see em_eo_stop(), em_eo_create()
 */
typedef em_status_t (*em_stop_func_t)(void* eo_ctx, em_eo_t eo);

/**
 * @ingroup DOC_api_generic
 *
 * Error handler.
 *
 * Error handler maybe called after EM notices an error or user have called em_error().
 * 
 * User can register EO specific and/or EM global error handlers. When an error is noticed, 
 * EM calls EO specific error handler, if registered. If there's no EO specific handler registered 
 * (for the EO) or the error is noticed outside of an EO context, EM calls the global error 
 * handler (if registered). If no error handlers are found, EM just returns an error code depending on 
 * the API function.
 *
 * Error handler is called with the original error code from the API call or em_error(). Error 
 * scope identifies the source of the error and how the error code and variable arguments 
 * should be interpreted (number of arguments and types).
 *
 * @param eo            Execution object id
 * @param error         The error code
 * @param escope        Error scope. Identifies the scope for interpreting the error code and variable arguments.
 * @param args          Variable number and type of arguments
 *
 * @return The function may not return depending on implementation/error code/error scope. If it
 * returns, it can return the original or modified error code or even EM_OK, if it could fix the problem.
 *
 * @see em_register_error_handler(), em_eo_register_error_handler()
 */
#ifndef TI_EM_GCC
typedef em_status_t (*em_error_handler_t)(em_eo_t eo, em_status_t error, em_escope_t escope, va_list args);
#else /* TI_EM_GCC */
typedef em_status_t (*em_error_handler_t)(em_eo_t eo, em_status_t error, em_escope_t escope, ...);
#endif /* TI_EM_GCC */


/**
 * \internal
 *--------------------------------------------------------------------------
 * Copyright (c) 2012, Texas Instruments Incorporated - http://www.ti.com/
 * End of the copy from the original event_machine.h 
 *--------------------------------------------------------------------------
 * \endinternal
 */


#ifdef __cplusplus
}
#endif

#endif /* EVENT_MACHINE_TYPES_H_ */

