/*
 * Copyright (c) 2012, Texas Instruments Incorporated - http://www.ti.com/
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions are met:
 *       * Redistributions of source code must retain the above copyright
 *         notice, this list of conditions and the following disclaimer.
 *       * Redistributions in binary form must reproduce the above copyright
 *         notice, this list of conditions and the following disclaimer in the
 *         documentation and/or other materials provided with the distribution.
 *       * Neither the name of Texas Instruments Incorporated nor the
 *         names of its contributors may be used to endorse or promote products
 *         derived from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *   DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
 *   DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


/**
 *   @file  event_machine_private_functions_linux.h
 *
 *   @brief
 *      Internal private functions used by the Open EM implementation for Linux.
 *      These functions are not available for use by the application.
 *
 */

#ifndef EVENT_MACHINE_PRIVATE_FUNCTIONS_LINUX_H
#define EVENT_MACHINE_PRIVATE_FUNCTIONS_LINUX_H

#include <stddef.h>
#include <inttypes.h>
#include <ti/runtime/openem/event_machine_macros.h>
#include <ti/runtime/openem/event_machine_group.h>
#include <ti/runtime/openem/event_machine_hw_ti_macros.h>
#include <ti/runtime/openem/event_machine_hw_ti_types.h>
#include <ti/runtime/openem/event_machine_hw_ti_functions.h>
#include <ti/runtime/openem/src/event_machine_hwpform.h>
#include <ti/runtime/openem/src/event_machine_private_macros.h>
#include <ti/runtime/openem/src/event_machine_private_types.h>
#include <ti/runtime/openem/src/event_machine_private_variables.h>
#include <ti/runtime/openem/src/event_machine_private_functions.h>

#include <ti_em_hw_functions.h> /* per-architecture/OS definitions */
#include <ti_em_hw_csl.h>       /* CPPI/QMSS descriptor definitions */
#include <ti_em_hw_qmss.h>      /* CPPI/QMSS descriptor definitions */
#include <ti_em_hw_cppi.h>      /* CPPI/QMSS descriptor definitions */
#include <ti_em_osal.h>         /* OS abstraction layer for generic system services */
#include <ti_em_osal_queue.h>   /* OS abstraction layer for queue management */
#include <ti_em_osal_uaccess.h> /* OS abstraction layer for user access and VM management */
#include <ti_em_osal_core.h>    /* OS abstraction layer for core management */
#include <ti_em_osal_shm.h>     /* OS abstraction layer for shared memory management */


/**
 * @internal
 *
 * The API is used to ensure data have been written to memory
 */
static inline void Emti_mfence(void)
{
    ti_em_osal_mfence();
}

/**
 * @internal
 *
 * The API is used to ensure data have been written to memory after a cache coherency operation
 */
static inline void Emti_mfenceWithCache(void)
{
    ti_em_osal_mfence();
}

/**
 * @internal
 *
 *  The function locks the semaphore to ensure critical sections
 *  across multiple cores.
 *
 * @return
 *  Not applicable.
 */
static inline void Emti_lockSemaphore(void)
{
    ti_em_osal_core_lock();
}

/**
 * @internal
 *
 *  The function unlocks the semaphore which had been locked
 *
 * @return
 *  Not applicable.
 */
static inline void Emti_unlockSemaphore(void)
{
    ti_em_osal_core_unlock();
}

/**
 * @internal
 *
 *  The function is used for logging debug information messages.
 *
 *  The user can re-implement this function. But he would need
 *  to recompile the library and the objects that are using
 *  inlined Open EM functions.
 *
 * @return
 *  Not applicable.
 */
#define Emti_logDbg(fmt, args...) TI_EM_OSAL_TRACE(3, fmt, ## args)

/**
 * @internal
 *
 *  The function is used for logging debug error messages.
 *
 *  The user can re-implement this function. But he would need
 *  to recompile the library and the objects that are using
 *  inlined Open EM functions.
 *
 * @return
 *  Not applicable.
 */
#define Emti_logError(fmt, args...) TI_EM_OSAL_ERROR(fmt, ## args)

/**
 * @internal
 *
 * The API is used to writeback invalidate the L1D cache for the specified address and size.
 */
static inline void Emti_cacheWbInvL1d
(
    void*       blockPtr,
    uint32_t    byteCnt)
{
    ti_em_osal_cache_wbinv_L1D(blockPtr, byteCnt);
}

/**
 * @internal
 *
 * The API is used to writeback invalidate the L2 cache for the specified address and size.
 */
static inline void Emti_cacheWbInvL2
(
    void*       blockPtr,
    uint32_t    byteCnt)
{
    ti_em_osal_cache_wbinv_L2(blockPtr, byteCnt);
}

/**
 * @internal
 *
 * The API is used to invalidate the L1D cache for the specified address and size.
 */
static inline void Emti_cacheInvL1d
(
    void*       blockPtr,
    uint32_t    byteCnt)
{
    ti_em_osal_cache_inv_L1D(blockPtr, byteCnt);
}

/**
 * @internal
 *
 * The API is used to invalidate the L2 cache for the specified address and size.
 */
static inline void Emti_cacheInvL2
(
    void*       blockPtr,
    uint32_t    byteCnt)
{
    ti_em_osal_cache_inv_L2(blockPtr, byteCnt);
}

/**
 * @internal
 *
 * The API is used to invalidate the whole L2 cache.
 */
static inline void Emti_cacheInvAllL2(void)
{
    ti_em_osal_cache_inv_all_L2();
}

/**
 * @internal
 *
 * The API is used to invalidate the XMC prefetch buffer
 */
static inline void Emti_prefetchBufferInv(void)
{
    ti_em_osal_cache_inv_prefetch();
}

/**
 * @internal
 *
 * The function invalidates the data buffer of the descriptor if required.
 *
 * @return 1 when invalidate operation has been applied, 0 otherwise.
 */
static inline uint32_t
Emti_inv (
    em_event_t event
)
{
    return (0);
}

/**
 * @internal
 *
 * The function writes back the data buffer of the descriptor if required.
 *
 * @return 1 when write back operation has been applied, 0 otherwise.
 */
static inline uint32_t
Emti_wbinv (
    em_event_t event
)
{
    return (0);
}

static inline void Emti_queueEmpty (uint16_t queueIdx)
{
    (void) ti_em_osal_hw_queue_empty((int) queueIdx);
}

static inline void Emti_queueDivert(uint16_t srcQueueIdx, uint16_t dstQueueIdx, uint8_t location)
{
    ti_em_osal_hw_queue_divert(srcQueueIdx, dstQueueIdx, location);
}

static inline void Emti_pushEvent (uint16_t queueIdx, void * eventPtr, short mapping)
{
    ti_em_osal_hw_queue_push(queueIdx, eventPtr, mapping);
}

static inline void Emti_pushEventToHead (uint16_t queueIdx, void * eventPtr, short mapping)
{
    ti_em_osal_hw_queue_push_to_head(queueIdx, eventPtr, mapping);
}

static inline void * Emti_popEvent (uint16_t queueIdx, short mapping)
{
    return (void*)((uint32_t) ti_em_osal_hw_queue_pop(queueIdx, mapping) & ~0xf);
}

static inline void Emti_setQueueThreshold(uint16_t queueIdx, uint16_t hilo, uint8_t threshold)
{
    ti_em_osal_hw_queue_set_threshold(queueIdx, hilo, threshold);
}

static inline uint32_t Emti_getQueueEntryCount(uint16_t queueIdx)
{
    return (uint32_t) ti_em_osal_hw_queue_peek(queueIdx);
}

static inline uint32_t Emti_getPdspDataRamBaseAddress(uint8_t pdspIdx)
{
    return (uint32_t) (void *) ti_em_osal_mem_phys_to_virt((TI_EM_PDSP_DRAM(pdspIdx)),
							   TI_EM_MEM_PDSP_DRAM);
}

/**
 * @internal
 *
 * Provide physical to/from virtual descriptor address translations
 * 
 */
static inline void* Emti_publicDesc_ptov(uint32_t paddr)
{
    return ti_em_osal_queue_desc_ptov(paddr, TI_EM_MEM_PUBLIC_DESC);
}

static inline uint32_t Emti_publicDesc_vtop(void* vaddr)
{
    return ti_em_osal_queue_desc_vtop(vaddr, TI_EM_MEM_PUBLIC_DESC);
}

static inline void* Emti_privateDesc_ptov(uint32_t paddr)
{
    return ti_em_osal_queue_desc_ptov(paddr, TI_EM_MEM_PRIVATE_DESC);
}

static inline uint32_t Emti_privateDesc_vtop(void* vaddr)
{
    return ti_em_osal_queue_desc_vtop(vaddr, TI_EM_MEM_PRIVATE_DESC);
}

/**
 * @internal
 *
 * Provide physical to/from virtual buffer pointer address translations
 * 
 */
static inline void* Emti_bufferPtr_ptov(Cppi_HostDesc *descAddr)
{
    uint32_t poolId = Emti_getPoolIndex(descAddr);
    return (void *) ti_em_osal_mem_phys_to_virt(descAddr->buffPtr,
						TI_EM_MEM_BUFFER + poolId);
}

static inline void* Emti_origBufferPtr_ptov(Cppi_HostDesc *descAddr)
{
    uint32_t poolId = Emti_getPoolIndex(descAddr);
    return (void *) ti_em_osal_mem_phys_to_virt(descAddr->origBuffPtr,
						TI_EM_MEM_BUFFER + poolId);
}

static inline uint32_t Emti_bufferPtr_vtop(void* vaddr, uint32_t poolId)
{
    return (uint32_t) ti_em_osal_mem_virt_to_phys((uint32_t) vaddr,
						  TI_EM_MEM_BUFFER + poolId);
}

/**
 * @internal
 *
 * Provide fake native to alias address translation
 * 
 */
static inline uint32_t Emti_getUncachedAlias(uint32_t naddr)
{
    return naddr;
}

/**
 * @internal
 *
 * Provide fake alias to native address translation
 * 
 */
static inline uint32_t Emti_getUncachedNative(uint32_t aaddr)
{
    return aaddr;
}

/**
 * @internal
 *
 * Provide physical to/from virtual QMMS register address translations
 * 
 */
static inline void* Emti_QmssRegs_ptov(uint32_t paddr)
{
    return (void *) ti_em_osal_io_phys_to_virt(paddr);
}

static inline uint32_t Emti_QmssRegs_vtop(void* vaddr)
{
    return (uint32_t) ti_em_osal_io_virt_to_phys((uint32_t) vaddr);
}

/**
 * @internal
 *
 * Copy CPU memory to PDSP communication memory
 *
 * Because it is difficult to retrieve the ptr physical address
 * and be sure that PDSP has coherent access to this memory,
 * we did a copy of the ptr content to a known memory location
 * (the PDSP communication memory based on the core index).
 * This memory must be coherent between the CPU and the PDSP
 * must have a known physical address in a known mapping.
 *
 * Of course the *ptr content must not be modified by the CPU until the
 * Emti_pdspGet() operation hase been performed.
 *
 */
static inline uint32_t Emti_pdspPut(uint32_t slot, void *ptr, uint32_t size)
{
    phys_addr_t pdsp_mem_paddr = ti_em_osal_get_pdsp_comm_mem_base(slot);

    memcpy((void *) ti_em_osal_mem_phys_to_virt(pdsp_mem_paddr, TI_EM_MEM_PDSP_COMM),
	   ptr, size);
    
    ti_em_osal_mfence();

    return (uint32_t) pdsp_mem_paddr;
}

/**
 * @internal
 *
 * Get PDSP memory to CPU
 *
 */
static inline void Emti_pdspGet(uint32_t slot, void* ptr, uint32_t size)
{
    /*
     * We synchronize the PDSP communication memory with the CPU memory.
     */
    phys_addr_t pdsp_mem_paddr = ti_em_osal_get_pdsp_comm_mem_base(slot);
    memcpy(ptr, (const void *) ti_em_osal_mem_phys_to_virt(pdsp_mem_paddr, TI_EM_MEM_PDSP_COMM), size);
}

/**
 * @internal
 *
 * Map CPU memory to PDSP communication memory
 *
 */
static inline uint32_t Emti_pdspMap(uint32_t slot, void **ptr)
{
    phys_addr_t pdsp_mem_paddr = ti_em_osal_get_pdsp_comm_mem_base(slot);
    *ptr = (void *) ti_em_osal_mem_phys_to_virt(pdsp_mem_paddr, TI_EM_MEM_PDSP_COMM);
    return pdsp_mem_paddr;
}

/**
 * @internal
 *
 * Initialize internal shared memory
 *
 */
static inline int Emti_sharedMemInit(unsigned int size, int init)
{
    Emti_logDbg("Initializing shared memory size %d\n", size);
    return ti_em_osal_shm_init(0, "openem_internal", size, init);
}

/**
 * @internal
 *
 * Release internal shared memory
 *
 */
static inline void Emti_sharedMemRelease(void)
{
    Emti_logDbg("Releasing shared memory\n");
    ti_em_osal_shm_release(0);
}

/**
 * @internal
 *
 * Allocate shared memory
 *
 */
static inline void* Emti_sharedMemAlloc(unsigned int size)
{
    return ti_em_osal_shm_alloc(0, size);
}

/**
 * @internal
 *
 * Free shared memory
 *
 */
static inline void Emti_sharedMemFree(void *ptr)
{
    ti_em_osal_shm_free(ptr);
}
#endif /* EVENT_MACHINE_PRIVATE_FUNCTIONS_H */
