/*
 * Copyright (c) 2012, Texas Instruments Incorporated - http://www.ti.com/
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions are met:
 *       * Redistributions of source code must retain the above copyright
 *         notice, this list of conditions and the following disclaimer.
 *       * Redistributions in binary form must reproduce the above copyright
 *         notice, this list of conditions and the following disclaimer in the
 *         documentation and/or other materials provided with the distribution.
 *       * Neither the name of Texas Instruments Incorporated nor the
 *         names of its contributors may be used to endorse or promote products
 *         derived from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *   DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
 *   DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


#include "my_receive.h"
#include "my_em_init.h"

#include <ti_em_osal.h>
#include <ti_em_osal_core.h>
#include <ti_em_osal_shm.h>

#define MY_DATA_SIZE_MAX    (32)

#define MY_JOB_HDR_SIZE     (CACHE_L2_LINESIZE)
#define MY_JOB_PLD_SIZE     (MY_DATA_SIZE_MAX * sizeof(uint32_t))
#define MY_JOB_SIZE         (MY_JOB_HDR_SIZE + MY_JOB_PLD_SIZE)

/* Pointed memory of these pointers must be shared accross cores */
volatile my_job_desc_t *my_sv_job_desc_tbl;
volatile uint32_t      *my_sv_running_job_num;

int my_init_jobs()
{
	uint32_t i, j;

	if (MY_JOB_NUM > 2*MY_EM_EVENT_NUM)
		return -1;
	if (MY_FLOW_NUM > MY_EM_PROC_QUEUE_NUM)
		return -1;
	if (MY_JOB_SIZE > MY_EM_EVENT_BUF_SIZE)
		return -1;
	if (sizeof(my_job_desc_t) > MY_JOB_HDR_SIZE)
		return -1;

	for (i = 0; i < MY_FLOW_NUM; i++)
	{
		for (j = 0; j < MY_JOB_NUM_PER_FLOW; j++)
		{
			uint32_t job_idx;
			job_idx = i * MY_JOB_NUM_PER_FLOW + j;
			my_sv_job_desc_tbl[job_idx].job_idx      = job_idx;
			my_sv_job_desc_tbl[job_idx].flow_idx     = i;
			my_sv_job_desc_tbl[job_idx].data_size    = MY_JOB_SIZE;
			my_sv_job_desc_tbl[job_idx].success_flag = 0;
			my_sv_job_desc_tbl[job_idx].core_idx     = MY_EM_CORE_NUM;
			my_sv_job_desc_tbl[job_idx].start_time   = 0;
			my_sv_job_desc_tbl[job_idx].stop_time    = 0;
		}
	}

	*my_sv_running_job_num = MY_JOB_NUM;

	return 0;
}

static void my_fill_buffer(uint8_t *buffer_ptr, uint32_t seed, uint32_t size)
{
	uint32_t *ptr = (uint32_t *) buffer_ptr;
	uint32_t i;

	for (i = 0; i < (size >> 2); i++) {
		uint32_t word = (seed << 20) + i;
		*ptr++ = word;
		ti_em_osal_mfence();
	}
}

static uint32_t my_check_buffer(uint8_t *buffer_ptr, uint32_t seed, uint32_t size)
{
	uint32_t *ptr = (uint32_t *) buffer_ptr;
	uint32_t i;
	uint32_t nb_mismatches = 0;

	for (i = 0; i < (size >> 2); i++) {
		uint32_t word = (seed << 20) + i;
		if (*ptr++ != word)
			nb_mismatches++;
	}

	return nb_mismatches;
}

static int my_copy_scattered_buffer(uint8_t *buffer_out_ptr, em_event_t event)
{
	uint8_t         *buff_ptr  = (uint8_t*) 
		((uint32_t) em_event_pointer(event) + MY_JOB_HDR_SIZE);
	size_t           buff_size = ti_em_buffer_size(event);
	size_t           left_size = ti_em_event_size(event);

	left_size -= buff_size;
	memcpy(buffer_out_ptr, buff_ptr, buff_size - MY_JOB_HDR_SIZE);
	
	if (left_size > 0) {
		em_status_t      status;
		ti_em_iterator_t iterator;
		
		status = ti_em_iterator_start(event, &iterator);
		if (status != EM_OK)
			return 0;

		while (left_size > 0) {
			buff_size = ti_em_iterator_size(&iterator);
			buff_ptr = (uint8_t *) ti_em_iterator_pointer(&iterator);
			
			memcpy(buffer_out_ptr, buff_ptr, buff_size);
			left_size -= buff_size;
			ti_em_iterator_next(&iterator);
		}
		ti_em_iterator_stop(&iterator);
	}
	return 1;
}

int my_source_jobs(void)
{
	uint32_t i;

	/* loop over all jobs to create */
	for (i = 0; i < MY_JOB_NUM; i++)
	{
		volatile em_event_t event;
		uint8_t            *job_buf;
		my_job_desc_t      *job_desc;
		uint32_t            flow_idx;
		uint32_t            data_size;

		flow_idx     = my_sv_job_desc_tbl[i].flow_idx;
		data_size    = my_sv_job_desc_tbl[i].data_size;

		/* allocate event */
		event = EM_EVENT_UNDEF;
		while(event == EM_EVENT_UNDEF) {
			event = em_alloc(data_size,
					 MY_EM_PROC_EVENT_TYPE,
					 MY_EM_POOL_IDX);
		}

		/* acess event buffer */
		job_desc = (my_job_desc_t*) em_event_pointer(event);
		job_buf  = (uint8_t*) ((uint32_t) job_desc + MY_JOB_HDR_SIZE);

		/* set ctrl information */
		job_desc->job_idx      = i;
		job_desc->flow_idx     = flow_idx;
		job_desc->data_size    = data_size;
		job_desc->success_flag = 1;

		/* set job input data */
		my_fill_buffer(job_buf, i, data_size - MY_JOB_HDR_SIZE);
		
		/* send event to EOs */
		em_send(event, my_em_proc_queue_hdl_tbl[flow_idx]);

		TI_EM_OSAL_TRACE(2, "event for job %d sent\n", i);
	}

	return 0;
}

/*
 * The main EM job of the execution objet
 */
void my_process_job(
	void           *ctx,
	em_event_t      event,
	em_event_type_t event_type,
	em_queue_t      queue,
	void           *queue_ctx)
{
	uint8_t            *job_buf_in;
	uint8_t            *job_buf_out;
	my_job_desc_t      *job_desc;
	volatile em_event_t output_event;
	uint32_t            job_idx;
	uint32_t            data_size;
	uint32_t            success_flag;
	uint64_t            start_time;
	uint64_t            stop_time;

	/* prefetch request */
	ti_em_preschedule();

	/* access event buffer of received event */
	job_desc    = (my_job_desc_t*) em_event_pointer(event);
	job_buf_in  = (uint8_t*) ((uint32_t) job_desc  + MY_JOB_HDR_SIZE);

	/*
	 * retrieve ctrl information from received
	 * event buffer
	 */
	job_idx      = job_desc->job_idx;
	data_size    = job_desc->data_size;
	success_flag = job_desc->success_flag;

	/* allocate a new event */
	output_event = EM_EVENT_UNDEF;
	do {
		output_event = em_alloc(data_size,
					TI_EM_EVENT_TYPE_PRELOAD_OFF,
					MY_EM_POOL_IDX);
	} while(output_event == EM_EVENT_UNDEF);
	
	/* access event buffer of new event */
	job_desc     = (my_job_desc_t*) em_event_pointer(output_event);
	job_buf_out  = (uint8_t*) ((uint32_t) job_desc  + MY_JOB_HDR_SIZE);

	/* fill new event buffer with ctrl information */
	job_desc->job_idx      = job_idx;
	job_desc->data_size    = data_size;
	job_desc->success_flag = success_flag;

	/* get start time for statistics */
	start_time = ti_em_osal_timer_read(TI_EM_OSAL_TIMER_LOCAL);

	/* copy the scattered received event payload into event out buffer */
	success_flag = my_copy_scattered_buffer(job_buf_out, event);

	TI_EM_OSAL_TRACE(2, "received/assembled/copied buffer for job %d\n", job_idx);
	sched_yield();

	/* get stop time for statistics */
	stop_time = ti_em_osal_timer_read(TI_EM_OSAL_TIMER_LOCAL);

	/* store information for statistics */
	my_sv_job_desc_tbl[job_idx].core_idx   = ti_em_osal_get_core_idx();
	my_sv_job_desc_tbl[job_idx].start_time = start_time;
	my_sv_job_desc_tbl[job_idx].stop_time  = stop_time;

	/* free received event */
	em_free(event);

	/* send new event */
	em_send(output_event, *my_em_sink_queue_hdl);
}

void my_sink_job(
	void*           ctx,
	em_event_t      event,
	em_event_type_t event_type,
	em_queue_t      queue,
	void           *queue_ctx)
{
	uint8_t       *job_buf;
	my_job_desc_t *job_desc;
	uint32_t       job_idx;

	TI_EM_OSAL_TRACE(3, "calling sink job\n");

	/* prefetch request */
	ti_em_preschedule();

	/* access event buffer of received event */
	job_desc = (my_job_desc_t*) em_event_pointer(event);
	job_buf  = (uint8_t*) ((uint32_t) job_desc + MY_JOB_HDR_SIZE);

	/* 
	 * retrieve ctrl information from received
	 * event buffer
	 */
	job_idx = job_desc->job_idx;

	/* check buffer content */
	if (my_check_buffer(job_buf, job_idx, job_desc->data_size - MY_JOB_HDR_SIZE) != 0)
		my_sv_job_desc_tbl[job_idx].success_flag = 0;
	else
		my_sv_job_desc_tbl[job_idx].success_flag = 1;

	/* free received event */
	em_free(event);

	/* decrement shared job counter */
	ti_em_osal_core_lock();
	(*my_sv_running_job_num)--;
	ti_em_osal_core_unlock();
}

void my_stats_jobs(void)
{
	uint32_t job_idx;
	uint32_t flow_idx;
	uint32_t core_idx;
	uint32_t job_num        = 0;
	uint32_t min_start_time = 0xffffffff;
	uint32_t max_stop_time  = 0;
	uint32_t run_time       = 0;
	uint32_t success_flag   = 1;
	uint32_t proc_time      = 0;
	uint32_t start_time;
	uint32_t stop_time;
	uint32_t min_proc_time  = 0xffffffff;
	uint32_t max_proc_time  = 0;
	uint32_t mean_proc_time = 0;
	uint32_t min_proc_times[MY_EM_CORE_NUM];
	uint32_t max_proc_times[MY_EM_CORE_NUM];
	uint32_t mean_proc_times[MY_EM_CORE_NUM];
	uint32_t job_nums[MY_EM_CORE_NUM][MY_FLOW_NUM];
	uint32_t stop_times[MY_EM_CORE_NUM];
	uint32_t ovh_time;
	uint32_t min_ovh_time   = 0xffffffff;
	uint32_t max_ovh_time   = 0;
	uint32_t mean_ovh_time  = 0;
	uint32_t mean_ovh_count = 0;
	uint32_t min_ovh_times[MY_EM_CORE_NUM];
	uint32_t max_ovh_times[MY_EM_CORE_NUM];
	uint32_t mean_ovh_times[MY_EM_CORE_NUM];

	/* initialize local variables */
	for (core_idx = 0; core_idx < MY_EM_CORE_NUM; core_idx++)
	{
		min_proc_times[core_idx]  = 0xffffffff;
		max_proc_times[core_idx]  = 0;
		mean_proc_times[core_idx] = 0;

		for (flow_idx = 0; flow_idx < MY_FLOW_NUM; flow_idx++)
			job_nums[core_idx][flow_idx] = 0;

		stop_times[core_idx]     = 0;
		min_ovh_times[core_idx]  = 0xffffffff;
		max_ovh_times[core_idx]  = 0;
		mean_ovh_times[core_idx] = 0;
	}
	
	/* compute statistics */
	for (job_idx = 0; job_idx < MY_JOB_NUM; job_idx++)
	{
		start_time = my_sv_job_desc_tbl[job_idx].start_time;
		stop_time  = my_sv_job_desc_tbl[job_idx].stop_time;
		core_idx   = my_sv_job_desc_tbl[job_idx].core_idx;
		flow_idx   = my_sv_job_desc_tbl[job_idx].flow_idx;
		
		if(my_sv_job_desc_tbl[job_idx].success_flag == 0) {
			TI_EM_OSAL_PRINTF("job %x not success\n", job_idx);
			success_flag = 0;
		}
		
		/* detect start time for first event */
		if(min_start_time > start_time)
			min_start_time = start_time;
		
		/* detect stop time for last event */
		if(max_stop_time < stop_time)
			max_stop_time = stop_time;
		
		/* compute total duration */
		run_time = max_stop_time - min_start_time;
		
		/* increment total number of jobs */
		job_num++;
		
		/* compute job duration time */
		proc_time = stop_time - start_time;
		
		/* check with min duration time */
		if (proc_time < min_proc_time)
			min_proc_time = proc_time;
		if (proc_time < min_proc_times[core_idx])
			min_proc_times[core_idx] = proc_time;
		
		/* check with max duration time */
		if (proc_time > max_proc_time)
			max_proc_time = proc_time;
		if (proc_time > max_proc_times[core_idx])
			max_proc_times[core_idx] = proc_time;
		
		job_nums[core_idx][flow_idx]++;
		mean_proc_time            += proc_time;
		mean_proc_times[core_idx] += proc_time;
		
		if (stop_times[core_idx] != 0) {
			/* compute duration time */
			ovh_time = start_time - stop_times[core_idx];
			
			/* check with min duration time */
			if (ovh_time < min_ovh_time)
				min_ovh_time = ovh_time;
			if (ovh_time < min_ovh_times[core_idx])
				min_ovh_times[core_idx] = ovh_time;
			
			/* check with max duration time */
			if (ovh_time > max_ovh_time)
				max_ovh_time = ovh_time;
			if (ovh_time > max_ovh_times[core_idx])
				max_ovh_times[core_idx] = ovh_time;
			
			mean_ovh_count ++;
			mean_ovh_time            += ovh_time;
			mean_ovh_times[core_idx] += ovh_time;
		}
		stop_times[core_idx] = stop_time;
	}
	
	TI_EM_OSAL_PRINTF("\n==========\n");

	if (success_flag == 1)
		TI_EM_OSAL_PRINTF("Run PASSED!\n");
	else 
		TI_EM_OSAL_PRINTF("Run FAILED!\n");
	
	/* per core statistics */
	for (core_idx = 0; core_idx < MY_EM_CORE_NUM; core_idx++)
	{
		uint32_t job_num2 = 0;

		for (flow_idx = 0; flow_idx < MY_FLOW_NUM; flow_idx++)
			job_num2 += job_nums[core_idx][flow_idx];
		
		TI_EM_OSAL_PRINTF("\nStatistics for core %d:\n", core_idx);
		TI_EM_OSAL_PRINTF("Number of jobs: %d\n", job_num2);
		TI_EM_OSAL_PRINTF("Number of jobs per flow: ");

		for (flow_idx = 0; flow_idx < MY_FLOW_NUM; flow_idx++)
			TI_EM_OSAL_PRINTF("%d ",job_nums[core_idx][flow_idx]);

		TI_EM_OSAL_PRINTF("\n");
		
		if (job_num2 > 0)
			TI_EM_OSAL_PRINTF("Min/mean/max processing cycles: %d/%d/%d\n",
				     min_proc_times[core_idx],
				     mean_proc_times[core_idx] / job_num2,
				     max_proc_times[core_idx]);
		
		if (job_num2 > 1)
			TI_EM_OSAL_PRINTF("Min/mean/max overhead cycles: %d/%d/%d\n\n",
				     min_ovh_times[core_idx],
				     mean_ovh_times[core_idx] / (job_num2 - 1),
				     max_ovh_times[core_idx]);
	}
	
	/* print statistics */
	TI_EM_OSAL_PRINTF("\nStatistics for all cores\n");
	TI_EM_OSAL_PRINTF("Total number of jobs: %d\n", job_num);
	TI_EM_OSAL_PRINTF("Total number of jobs per flow: ");

	for (flow_idx = 0; flow_idx < MY_FLOW_NUM; flow_idx++)
	{
		uint32_t job_num2 = 0;
		
		for (core_idx = 0; core_idx < MY_EM_CORE_NUM; core_idx++)
			job_num2 += job_nums[core_idx][flow_idx];

		TI_EM_OSAL_PRINTF("%d ", job_num2);
	}
	TI_EM_OSAL_PRINTF("\n");
	
	if (job_num > 0)
		TI_EM_OSAL_PRINTF("Min/mean/max processing cycles: %d/%d/%d\n",
			     min_proc_time,
			     mean_proc_time / job_num,
			     max_proc_time);
	
	if (mean_ovh_count > 0)
		TI_EM_OSAL_PRINTF("Min/mean/max overhead cycles: %d/%d/%d\n\n",
			     min_ovh_time,
			     mean_ovh_time / mean_ovh_count,
			     max_ovh_time);
	
	TI_EM_OSAL_PRINTF("Start/stop/run cycles: %d/%d/%d\n",
		     min_start_time,
		     max_stop_time,
		     run_time);

	TI_EM_OSAL_PRINTF("\n==========\n");
}
