/*
 * Copyright (c) 2012, Texas Instruments Incorporated - http://www.ti.com/
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions are met:
 *       * Redistributions of source code must retain the above copyright
 *         notice, this list of conditions and the following disclaimer.
 *       * Redistributions in binary form must reproduce the above copyright
 *         notice, this list of conditions and the following disclaimer in the
 *         documentation and/or other materials provided with the distribution.
 *       * Neither the name of Texas Instruments Incorporated nor the
 *         names of its contributors may be used to endorse or promote products
 *         derived from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *   DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
 *   DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


#include "my_em_init.h"
#include "my_receive.h"

/* Importing some of the Open EM defined symbols */
#include "my_event_machine.h"

#include <ti_em_osal.h>
#include <ti_em_osal_core.h>
#include <ti_em_osal_shm.h>

#include <ti_em_rh.h>

#include <getopt.h>

static uint32_t my_em_core_num = MY_EM_CORE_NUM;

static int my_global_init(void)
{
	TI_EM_OSAL_TRACE(1, "master core initialization\n");

	/* example shared objects creation */
	my_em_shared_init(1);
	
	/* OpenEM global initialization */
	if (my_em_init_global() < 0)
		goto error;
	
	/* application timer global initialization */
	if (ti_em_osal_timer_init(TI_EM_OSAL_TIMER_GLOBAL) < 0)
		goto error;

	/* local initialization */
	if (my_em_init_local() < 0)
		goto error;

	/* OpenEM objects creation */
	if (my_em_init_queues(my_em_core_num) < 0) {
		TI_EM_OSAL_ERROR("objects creation error\n");
		goto error;
	}
	TI_EM_OSAL_TRACE(1, "objects creation done\n");

	return 0;
error:
	ti_em_osal_shm_release(MY_SHARED_MEM_ID);
	return -1;
}

static int my_dispatcher_local_init(void)
{
	TI_EM_OSAL_TRACE(1, "local init\n");
	
	/* OpenEM local initialization */
	if (my_em_init_local() < 0)
		return -1;
	
	/* application timer local initialization */
	if (ti_em_osal_timer_init(TI_EM_OSAL_TIMER_LOCAL) < 0)
		return -1;

	return 0;
}

static int my_dispatcher(void* arg)
{
	TI_EM_OSAL_TRACE(2, "entering dispatcher\n");

	/* request first event from scheduler */
	ti_em_preschedule();

	/* call event dispatcher */
	while (1) {
		ti_em_rh_dispatch(MY_EM_DISPATCH_MODE, 0, 0);
	}

	return 0; /* not reached */
}

int main(int argc, char **argv)
{
	int c;
	int lop;
	int server = 0;
	int client = 0;
	int core_id = -1;

	struct option long_options[] = {
		{ "core-num", 1, 0, 'n'},
		{ "server", 0, 0, 's'},
		{ "client", 0, 0, 'c'},
		{ "core-id", 1, 0, 'C'},
                { "job-num", 1, 0, 'j'},
                { "tx-interval", 1, 0, 'i'},
                { "data-size", 1, 0, 'z'},
		{ "tx-fragment-size", 1, 0, 'f'},
		{ "rx-fragment-size", 1, 0, 'r'},
		{ "event-buf-size", 1, 0, 'b'},
		{ "pktdma", 0, 0, 'p'},
		{ "buf-mode", 0, 0, 'B'},
		{ 0, 0, 0, 0 }
	};

	TI_EM_OSAL_PRINTF("starting Example 1\n");

	/* runtime options parse */
	while ((c = getopt_long(argc, argv, "n:j:i:z:f:r:b:B:scC:p", long_options, &lop)) != EOF) {
		switch(c) {
		case 'n':
			my_em_core_num = strtoul(optarg, (char **) NULL, 10);
			break;
		case 's':
			TI_EM_OSAL_PRINTF("launched as server...\n");
			server = 1;
			break;
		case 'c':
			TI_EM_OSAL_PRINTF("launched as client...\n");
			client = 1;
			break;
		case 'C':
			core_id = strtoul(optarg, (char **) NULL, 10);
			TI_EM_OSAL_PRINTF("using Core Id %d\n", core_id);
			break;
		case 'j':
			my_job_num = strtoul(optarg, (char **) NULL, 10);
			TI_EM_OSAL_PRINTF("set job number to %d\n", my_job_num);
			break;
		case 'i':
			my_job_interval = strtoul(optarg, (char **) NULL, 10);
			TI_EM_OSAL_PRINTF("set job transmit interval to %d\n", my_job_interval);
			break;
		case 'z':
			my_data_size = strtoul(optarg, (char **) NULL, 10);
			TI_EM_OSAL_PRINTF("set data size to %d\n", my_data_size);
			break;
		case 'f':
			my_tx_xge_fragment_size = strtoul(optarg, (char **) NULL, 10);
			TI_EM_OSAL_PRINTF("set XGE transmit fragment size to %d\n", my_tx_xge_fragment_size);
			break;
		case 'r':
			my_rx_xge_fragment_size = strtoul(optarg, (char **) NULL, 10);
			TI_EM_OSAL_PRINTF("set XGE receive fragment size to %d\n", my_rx_xge_fragment_size);
			break;
		case 'b':
			my_event_buf_size = strtoul(optarg, (char **) NULL, 10);
			TI_EM_OSAL_PRINTF("set event buf size to %d\n", my_event_buf_size);
			break;
		case 'p':
			my_xge = 0;
			TI_EM_OSAL_PRINTF("use infra-PktDMA only (no XGE)\n");
			break;
		case 'B':
			if (!strcmp(optarg, "loose"))
				my_buf_mode = TI_EM_BUF_MODE_LOOSE;
			else
				my_buf_mode = TI_EM_BUF_MODE_TIGHT;
			TI_EM_OSAL_PRINTF("use buffer-mode %s\n",
					  (my_buf_mode == TI_EM_BUF_MODE_LOOSE) ?
					  "loose": "tight");
			break;
		default:
			break;
		}
	}

	if (((client) || (server)) && (my_buf_mode == TI_EM_BUF_MODE_LOOSE)) {
		TI_EM_OSAL_ERROR("cannot use joining processes when using loose buffers\n");
		return -1;
	}

	TI_EM_OSAL_PRINTF("\n");
	TI_EM_OSAL_PRINTF("Job number is %d\n", my_job_num);
	TI_EM_OSAL_PRINTF("Transmit interval is %d\n", my_job_interval);
	TI_EM_OSAL_PRINTF("Data size (job size) is %d\n", my_data_size);
	if (my_xge) {
		TI_EM_OSAL_PRINTF("XGE tx fragment size is %d\n", my_tx_xge_fragment_size);
		TI_EM_OSAL_PRINTF("XGE rx fragment size is %d\n", my_rx_xge_fragment_size);
	} else
		TI_EM_OSAL_PRINTF("No XGE\n");
	TI_EM_OSAL_PRINTF("Event buf size is %d\n", my_event_buf_size);
	TI_EM_OSAL_PRINTF("\n");

	if (client) {
		/* retrieve shared example objects */
		my_em_shared_init(0);

		/* join existing instance with a new OpenEM core id */
		ti_em_rh_join(core_id);

		/* OpenEM local init */
		if (ti_em_init_local() != EM_OK) {
			TI_EM_OSAL_ERROR("local EM initialization failed\n");
			return -1;
		}
	
		/* allocates events, initialize event buffers, send events */
		if (my_source_jobs() < 0) {
			TI_EM_OSAL_ERROR("jobs failed\n");
			return -1;
		}

		TI_EM_OSAL_TRACE(1, "exiting...\n");

		ti_em_rh_release_local();
		ti_em_osal_core_exit(0);
	}
		
	/* 
	 * OSAL core runtime library initialization (must only called by main core) 
	 */
	ti_em_osal_core_init_global();
	
	if ((my_em_core_num > MY_EM_CORE_NUM) || (my_em_core_num < 1)) {
		TI_EM_OSAL_ERROR("invalid number of cores\n");
		exit(-1);
	}  else
		TI_EM_OSAL_TRACE(1, "running %d cores\n", my_em_core_num);
	       
	/* initialization */
	if (my_global_init() < 0) {
		return -1;
	}

	/* initialize jobs */
	if (my_init_jobs() < 0) {
		TI_EM_OSAL_ERROR("jobs initialization failed\n");
		return -1;
	}

	if (!server) {
		/* allocates events, initialize event buffers, send events */
		if (my_source_jobs() < 0) {
			TI_EM_OSAL_ERROR("jobs failed\n");
			return -1;
		}
	}
	
	/* create and starts cores */
 	if (ti_em_rh_create_cores(my_em_core_num,
				  my_dispatcher,
				  my_dispatcher_local_init,
				  MY_EM_INIT_CORE_IDX) < 0) {
		TI_EM_OSAL_ERROR("core creation failed\n");
		return -1;
	}

	/* application timer local initialization */
	if (ti_em_osal_timer_init(TI_EM_OSAL_TIMER_LOCAL) < 0)
		return -1;

	/* request first event from scheduler */
	ti_em_preschedule();

	/* call event dispatcher */
	while (*my_sv_running_job_num > 0) {
		ti_em_rh_dispatch(TI_EM_RH_POLL_MODE, 0, 0);
	}

	/* compute job statistics */
	my_stats_jobs();

	/* terminate other cores */
	if (my_em_exit_global(my_em_core_num) != 0) {
	    /* eventually kill unfinished cores and wait cores termination */
	    ti_em_rh_kill_cores(my_em_core_num);
	    ti_em_rh_wait_cores(my_em_core_num);
	}

	/* remove test shared memory */
	ti_em_osal_shm_release(MY_SHARED_MEM_ID);

	/* release open EM*/
	ti_em_rh_release_local();
	ti_em_rh_release_global();

	/* exit */
	ti_em_osal_core_exit(0);

	return -1;
}
