/*
 * Copyright (c) 2012, Texas Instruments Incorporated - http://www.ti.com/
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions are met:
 *       * Redistributions of source code must retain the above copyright
 *         notice, this list of conditions and the following disclaimer.
 *       * Redistributions in binary form must reproduce the above copyright
 *         notice, this list of conditions and the following disclaimer in the
 *         documentation and/or other materials provided with the distribution.
 *       * Neither the name of Texas Instruments Incorporated nor the
 *         names of its contributors may be used to endorse or promote products
 *         derived from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *   DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
 *   DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


#include "my_receive.h"
#include "my_em_init.h"

#include <ti_em_osal.h>
#include <ti_em_osal_core.h>
#include <ti_em_osal_shm.h> 
#include <ti_em_osal_xge.h>

#define MY_DATA_SIZE_MAX   (32)

#define MY_JOB_HDR_SIZE    sizeof(my_job_desc_t)
#define MY_JOB_PLD_SIZE    (MY_DATA_SIZE_MAX * sizeof(uint32_t))
#define MY_JOB_SIZE        (MY_JOB_HDR_SIZE + MY_JOB_PLD_SIZE)

#ifdef DEVICE_K2K
/* On K2 use XGE by default */
#define MY_XGE             (1)
#else
/* Do not use XGE on K1 */
#define MY_XGE             (0)
#endif

uint32_t my_job_num      = MY_JOB_NUM;
uint32_t my_job_interval = 0;
uint32_t my_data_size    = MY_JOB_SIZE;
uint8_t  my_xge          = MY_XGE;
uint8_t  my_buf_mode     = MY_EM_BUF_MODE;

/* Pointed memory of these pointers must be shared accross cores */
volatile my_job_desc_t    *my_sv_job_desc_tbl;
volatile uint32_t         *my_sv_running_job_num;
volatile ti_em_queue_id_t *my_xge_rx_queue_idx;

static em_event_t my_em_alloc(size_t data_size,
			      size_t buffer_size)
{
	em_event_t event = (em_event_t) 0;

	if (my_buf_mode == TI_EM_BUF_MODE_TIGHT) {
		/*
		 * Tight buffer case
		 */
		int        remaining_size = (int) data_size;
		em_event_t previous_event = (em_event_t) 0;
		ti_em_iterator_t iterator;

		while (remaining_size > 0) {
			size_t alloc_size = remaining_size < buffer_size ?
				remaining_size : buffer_size;

			event = em_alloc(alloc_size,
					 MY_EM_PROC_EVENT_TYPE,
					 MY_EM_POOL_IDX);
			
			if (event == EM_EVENT_UNDEF)
				continue;

			remaining_size -= alloc_size;

			/* combine event if needed */
			if (previous_event) {
				ti_em_pair_t my_event_pair;
				
				my_event_pair.head_event_hdl = previous_event;
				my_event_pair.tail_event_hdl = event;

				event = ti_em_combine(my_event_pair, &iterator);
			} else {
				em_status_t status;

				/* if it is the first event, then start the iterator */
				status = ti_em_iterator_start(event, &iterator);
				if (status != EM_OK)
					return 0;
			}

			previous_event = event;
		}

		/* stop the iterator */
		ti_em_iterator_stop(&iterator);
	} else {
		/*
		 * Loose buffer case
		 */
		ti_em_buffer_config_t buffer_config_tbl[16]; /* max of 16 fragments */
		int remaining_size = (int) data_size;
		int index = 0;

		memset(&buffer_config_tbl[0], 0, sizeof(buffer_config_tbl));

		while(remaining_size > 0) {
			size_t alloc_size = remaining_size < buffer_size ?
				remaining_size : buffer_size;

			buffer_config_tbl[index].buffer_size = alloc_size;
			buffer_config_tbl[index].buffer_ptr  = my_em_event_buf_mem;
			buffer_config_tbl[index].orig_buffer_size = alloc_size;
			buffer_config_tbl[index].orig_buffer_ptr  = my_em_event_buf_mem;

			my_em_event_buf_mem += alloc_size;
			remaining_size      -= alloc_size;
			index++;
		}

		event = ti_em_alloc_with_buffers(MY_EM_PROC_EVENT_TYPE,
						 MY_EM_POOL_IDX,
						 index,
						 &buffer_config_tbl[0]);
	}
	return event;
}

int my_init_jobs()
{
	uint32_t job_idx;

	if (my_job_num > MY_JOB_NUM_MAX)
		return -1;
	if (my_job_num >= MY_EM_EVENT_NUM)
		return -1;
	if (MY_FLOW_NUM > MY_EM_PROC_QUEUE_NUM)
		return -1;
	if (sizeof(my_job_desc_t) > MY_JOB_HDR_SIZE)
		return -1;

	for (job_idx = 0; job_idx < my_job_num; job_idx++)
	{
		my_sv_job_desc_tbl[job_idx].job_idx      = job_idx;
		my_sv_job_desc_tbl[job_idx].flow_idx     = job_idx % MY_FLOW_NUM;
		my_sv_job_desc_tbl[job_idx].data_size    = my_data_size;
		my_sv_job_desc_tbl[job_idx].success_flag = 0;
		my_sv_job_desc_tbl[job_idx].core_idx     = MY_EM_CORE_NUM;
		my_sv_job_desc_tbl[job_idx].start_time   = 0;
		my_sv_job_desc_tbl[job_idx].stop_time    = 0;
	}

	*my_sv_running_job_num = my_job_num;

	return 0;
}

static void my_fill_buffer(uint32_t *ptr, uint32_t seed, uint32_t start, uint32_t size)
{
	uint32_t i;

	TI_EM_OSAL_TRACE(4, "fill buffer with size %d, seed %d, start %d\n", size, seed, start);

	for (i = 0; i < (size >> 2); i++) {
		uint32_t word = (seed << 20) + i + start;
		*ptr++ = word;
	}
}

static int my_fill_scattered_buffer(em_event_t event, uint32_t seed, uint32_t size)
{
	size_t           event_size = ti_em_event_size(event);
	int              left_size = (int) size;
	int              fill_size;
	uint32_t         index = 0;
	em_status_t      status;
	ti_em_iterator_t iterator;
	uint32_t        *buff_ptr;
	size_t           buff_size;
	int              first_iteration = 1;

	TI_EM_OSAL_TRACE(4, "fill scattered buffer with size %d, event_size %d, buff_size %d\n", size, event_size, buff_size);

	if (event_size < size) {
		size = event_size;
		TI_EM_OSAL_ERROR("total buffer size of event(s) is smaller than data size!!!\n");
	}

	status = ti_em_iterator_start(event, &iterator);
	if (status != EM_OK)
		return 0;

	while (left_size > 0) {
		/* fill the scattered event buffers */
		buff_size = ti_em_iterator_size(&iterator);
		buff_ptr  = (uint32_t *) ti_em_iterator_pointer(&iterator);
		fill_size = left_size < buff_size ? left_size : buff_size; 
		
		if (first_iteration) {
			/* this is the first iteration */
			if (buff_size <= MY_JOB_HDR_SIZE) {
				TI_EM_OSAL_ERROR("buffer size is not big enough to contain job header!!!\n");
				return 0;
			}
			/* skip the job header */ 
			buff_ptr = (uint32_t *)(((uint32_t) buff_ptr) + MY_JOB_HDR_SIZE);
			fill_size -= MY_JOB_HDR_SIZE;
			left_size -= MY_JOB_HDR_SIZE;
			first_iteration = 0;
		}
		
		my_fill_buffer(buff_ptr, seed, index, fill_size);
		index += (buff_size >> 2);
		left_size -= fill_size;
		ti_em_iterator_next(&iterator);
	}

	ti_em_iterator_stop(&iterator);

	return 1;
}

static int my_check_buffer(uint32_t *ptr, uint32_t seed, uint32_t start, uint32_t size)
{
	uint32_t i;
	uint32_t nb_mismatches = 0;

	TI_EM_OSAL_TRACE(4, "check for size %d\n", size);

	for (i = 0; i < (size >> 2); i++) {
		uint32_t word = (seed << 20) + i + start;
		if (*ptr++ != word) {
			TI_EM_OSAL_TRACE(4, "mismatch at offset %d (found 0x%x != expected 0x%x)\n", (i<<2), *(ptr - 1), word);
			nb_mismatches++;
		}
	}
	return nb_mismatches;
}

static int my_check_scattered_buffer(em_event_t event, uint32_t seed, uint32_t size)
{
	size_t           event_size = ti_em_event_size(event);
	int              left_size = (int) size;
	int              check_size;
	uint32_t         index = 0;
	em_status_t      status;
	ti_em_iterator_t iterator;
	uint32_t        *buff_ptr;
	size_t           buff_size;
	int              first_iteration = 1;
	int              nb_mismatches = 0;

	TI_EM_OSAL_TRACE(4, "check scattered buffer with size %d, event_size %d, buff_size %d\n", size, event_size, buff_size);

	if (event_size < size) {
		size = event_size;
		TI_EM_OSAL_ERROR("total buffer size of event(s) is smaller than data size!!!\n");
	}

	status = ti_em_iterator_start(event, &iterator);
	if (status != EM_OK)
		return 0;

	while (left_size > 0) {
		/* fill the scattered event buffers */
		buff_size = ti_em_iterator_size(&iterator);
		buff_ptr  = (uint32_t *) ti_em_iterator_pointer(&iterator);
		check_size = left_size < buff_size ? left_size : buff_size; 
		
		if (first_iteration) {
			/* this is the first iteration */
			if (buff_size <= MY_JOB_HDR_SIZE) {
				TI_EM_OSAL_ERROR("buffer size is not big enough to contain job header!!!\n");
				return 0;
			}
			/* skip the job header */ 
			buff_ptr = (uint32_t *)(((uint32_t) buff_ptr) + MY_JOB_HDR_SIZE);
			check_size -= MY_JOB_HDR_SIZE;
			left_size  -= MY_JOB_HDR_SIZE;
			first_iteration = 0;
		}
		
		nb_mismatches = my_check_buffer(buff_ptr, seed, index, check_size);
		if (nb_mismatches) {
			TI_EM_OSAL_TRACE(1, "mistmatch %d\n", nb_mismatches);
			ti_em_iterator_stop(&iterator);
			return nb_mismatches;
		}
		index += (buff_size >> 2);
		left_size -= check_size;
		ti_em_iterator_next(&iterator);
	}

	ti_em_iterator_stop(&iterator);

	return nb_mismatches;
}

static int my_copy_scattered_buffer(em_event_t event_out, em_event_t event, size_t size)
{

	uint8_t         *buff_out_ptr;
	uint8_t         *buff_ptr;
	size_t           buff_size = ti_em_buffer_size(event);
	int              left_size = (int) size;
	em_status_t      status;
	ti_em_iterator_t iterator;
	ti_em_iterator_t iterator_out;
	int              first_iteration = 1;

	if (size > ti_em_event_size(event)) {
		TI_EM_OSAL_ERROR("the job data size (%d) is greater than the total event size (%d)!\n",
				 size, ti_em_event_size(event));
		return 0;
	}

	TI_EM_OSAL_TRACE(4, "copy scattered buffer with event size %d, buffer size %d, data size %d\n", ti_em_event_size(event), buff_size, size);
	
	status = ti_em_iterator_start(event, &iterator);
	if (status != EM_OK)
		return 0;
	
	status = ti_em_iterator_start(event_out, &iterator_out);
	if (status != EM_OK)
		return 0;
	
	while (left_size > 0) {
		buff_size = ti_em_iterator_size(&iterator);
		buff_ptr = (uint8_t *) ti_em_iterator_pointer(&iterator);
		buff_out_ptr = (uint8_t *) ti_em_iterator_pointer(&iterator_out);
		
		if (first_iteration) {
			/* this is the first iteration */
			if (buff_size <= MY_JOB_HDR_SIZE) {
				TI_EM_OSAL_ERROR("buffer size is not big enough to contain job header!!!\n");
				return 0;
			}
			/* skip the job header */ 
			buff_ptr = (uint8_t *)(((uint32_t) buff_ptr) + MY_JOB_HDR_SIZE);
			buff_out_ptr = (uint8_t *)(((uint32_t) buff_out_ptr) + MY_JOB_HDR_SIZE);
			buff_size -= MY_JOB_HDR_SIZE;
			left_size -= MY_JOB_HDR_SIZE;
			first_iteration = 0;
		}
		
		memcpy(buff_out_ptr, buff_ptr, buff_size);
		left_size -= buff_size;
		
		ti_em_iterator_next(&iterator);
		ti_em_iterator_next(&iterator_out);
	}
	ti_em_iterator_stop(&iterator);
	ti_em_iterator_stop(&iterator_out);

	return 1;
}

int my_source_jobs(void)
{
	uint32_t       i;
	em_event_t     event;
	my_job_desc_t *job_desc;
	uint32_t       flow_idx;
	uint32_t       data_size;
	em_status_t    res;
	em_queue_t     dest_queue;

	/* loop over all jobs to create */
	for (i = 0; i < my_job_num; i++) {
		flow_idx  = my_sv_job_desc_tbl[i].flow_idx;
		data_size = my_sv_job_desc_tbl[i].data_size;

		/* allocate event */
		event = my_em_alloc(data_size, my_event_buf_size);
		if (event == EM_EVENT_UNDEF) {
			TI_EM_OSAL_ERROR("cannot allocate output event\n");
			return -1;
		}

		/* acess event buffer */
		job_desc = (my_job_desc_t*) em_event_pointer(event);

		/* set ctrl information */
		job_desc->job_idx      = i;
		job_desc->flow_idx     = flow_idx;
		job_desc->data_size    = data_size;
		job_desc->success_flag = 1;

		/* set job input data */
		my_fill_scattered_buffer(event, i, data_size);

		/* send event to EQs of the remote device or process */
		if (my_xge)
			dest_queue = my_em_proc_queue_hdl_tbl[flow_idx] + 0x1000000;
		else
			dest_queue = my_em_proc_queue_hdl_tbl[flow_idx] + 0x4000;
		
		/* send event to EQs of the remote device or process */		
		while ((res = em_send(event, dest_queue)) == EM_ERR_NOT_FOUND) {
			TI_EM_OSAL_TRACE(2, "no free resource for sending event!\n");
			usleep(1);
		}			
			
		if (res != EM_OK)
			TI_EM_OSAL_ERROR("cannot send event for job %d!\n", i);

		if (my_xge) {
			/* eventually add a CRC field in the packet if using fake XGE transfers */
			ti_em_osal_xge_rx_insert_crc(MY_EM_VLAN_PRIORITY,
						     MY_EM_FREE_QUEUE_BASE_IDX + MY_EM_DTX_FRAG_POOL_IDX,
						     my_xge_rx_queue_idx[MY_EM_VLAN_PRIORITY],
						     ((my_data_size + (my_tx_xge_fragment_size - 1)) / my_tx_xge_fragment_size));
		}

		TI_EM_OSAL_TRACE(2, "event for job %d sent\n", i);

		if (my_job_interval)
			usleep(my_job_interval * 1000);
	}
	return 0;
}

/*
 * The main EM job of the execution object
 */
void my_process_job(
	void           *ctx,
	em_event_t      event,
	em_event_type_t event_type,
	em_queue_t      queue,
	void           *queue_ctx)
{
	my_job_desc_t      *job_desc;
	volatile em_event_t output_event;
	uint32_t            job_idx;
	uint32_t            data_size;
	uint32_t            success_flag;
	uint64_t            start_time;
	uint64_t            stop_time;

	/* prefetch request */
	ti_em_preschedule();

	/* access event buffer of received event */
	job_desc = (my_job_desc_t*) em_event_pointer(event);

	/*
	 * retrieve ctrl information from received
	 * event buffer
	 */
	job_idx      = job_desc->job_idx;
	data_size    = job_desc->data_size;
	success_flag = job_desc->success_flag;

	/* allocate a new event */
	output_event = my_em_alloc(data_size, my_event_buf_size);
	if (output_event == EM_EVENT_UNDEF) {
		TI_EM_OSAL_ERROR("cannot allocate output event\n");
		job_desc->success_flag = 0;
		em_free(event);
		return;
	}

	/* access event buffer of new event */
	job_desc = (my_job_desc_t*) em_event_pointer(output_event);

	/* fill new event buffer with ctrl information */
	job_desc->job_idx      = job_idx;
	job_desc->data_size    = data_size;
	job_desc->success_flag = success_flag;

	/* get start time for statistics */
	start_time = ti_em_osal_timer_read(TI_EM_OSAL_TIMER_LOCAL);

	/* copy the scattered received event payload into event out buffer */
	success_flag = my_copy_scattered_buffer(output_event, event, data_size);

	TI_EM_OSAL_TRACE(2, "received/assembled/copied buffer for job %d\n", job_idx);
	sched_yield();

	/* get stop time for statistics */
	stop_time = ti_em_osal_timer_read(TI_EM_OSAL_TIMER_LOCAL);

	/* store information for statistics */
	my_sv_job_desc_tbl[job_idx].core_idx   = ti_em_osal_get_core_idx();
	my_sv_job_desc_tbl[job_idx].start_time = start_time;
	my_sv_job_desc_tbl[job_idx].stop_time  = stop_time;

	/* free received event */
	em_free(event);

	/* send new event */
	em_send(output_event, *my_em_sink_queue_hdl);
}

void my_sink_job(
	void*           ctx,
	em_event_t      event,
	em_event_type_t event_type,
	em_queue_t      queue,
	void           *queue_ctx)
{
	my_job_desc_t *job_desc;
	uint32_t       job_idx;

	/* prefetch request */
	ti_em_preschedule();

	/* access event buffer of received event */
	job_desc = (my_job_desc_t*) em_event_pointer(event);

	/* 
	 * retrieve ctrl information from received
	 * event buffer
	 */
	job_idx = job_desc->job_idx;

	TI_EM_OSAL_TRACE(3, "calling sink job for job %d (size = %d)\n", job_idx, job_desc->data_size);

	/* check buffer content */
	my_sv_job_desc_tbl[job_idx].success_flag = 1;

	if (job_desc->data_size > MY_JOB_HDR_SIZE) {
		if (my_check_scattered_buffer(event, job_idx, job_desc->data_size))
			my_sv_job_desc_tbl[job_idx].success_flag = 0;
		else
			my_sv_job_desc_tbl[job_idx].success_flag = 1;
	} else {
		TI_EM_OSAL_TRACE(3, "skip data verification (no payload data)\n");
	}

	/* free received event */
	em_free(event);

	/* decrement shared job counter */
	ti_em_osal_core_lock();
	(*my_sv_running_job_num)--;
	ti_em_osal_core_unlock();
}

void my_stats_jobs(void)
{
	uint32_t job_idx;
	uint32_t flow_idx;
	uint32_t core_idx;
	uint32_t job_num        = 0;
	uint32_t min_start_time = 0xffffffff;
	uint32_t max_stop_time  = 0;
	uint32_t run_time       = 0;
	uint32_t success_flag   = 1;
	uint32_t proc_time      = 0;
	uint32_t start_time;
	uint32_t stop_time;
	uint32_t min_proc_time  = 0xffffffff;
	uint32_t max_proc_time  = 0;
	uint32_t mean_proc_time = 0;
	uint32_t min_proc_times[MY_EM_CORE_NUM];
	uint32_t max_proc_times[MY_EM_CORE_NUM];
	uint32_t mean_proc_times[MY_EM_CORE_NUM];
	uint32_t job_nums[MY_EM_CORE_NUM][MY_FLOW_NUM];
	uint32_t stop_times[MY_EM_CORE_NUM];
	uint32_t ovh_time;
	uint32_t min_ovh_time   = 0xffffffff;
	uint32_t max_ovh_time   = 0;
	uint32_t mean_ovh_time  = 0;
	uint32_t mean_ovh_count = 0;
	uint32_t min_ovh_times[MY_EM_CORE_NUM];
	uint32_t max_ovh_times[MY_EM_CORE_NUM];
	uint32_t mean_ovh_times[MY_EM_CORE_NUM];

	/* initialize local variables */
	for (core_idx = 0; core_idx < MY_EM_CORE_NUM; core_idx++)
	{
		min_proc_times[core_idx]  = 0xffffffff;
		max_proc_times[core_idx]  = 0;
		mean_proc_times[core_idx] = 0;

		for (flow_idx = 0; flow_idx < MY_FLOW_NUM; flow_idx++)
			job_nums[core_idx][flow_idx] = 0;

		stop_times[core_idx]     = 0;
		min_ovh_times[core_idx]  = 0xffffffff;
		max_ovh_times[core_idx]  = 0;
		mean_ovh_times[core_idx] = 0;
	}
	
	/* compute statistics */
	for (job_idx = 0; job_idx < my_job_num; job_idx++)
	{
		start_time = my_sv_job_desc_tbl[job_idx].start_time;
		stop_time  = my_sv_job_desc_tbl[job_idx].stop_time;
		core_idx   = my_sv_job_desc_tbl[job_idx].core_idx;
		flow_idx   = my_sv_job_desc_tbl[job_idx].flow_idx;
		
		if(my_sv_job_desc_tbl[job_idx].success_flag == 0) {
			TI_EM_OSAL_PRINTF("job %x not success\n", job_idx);
			success_flag = 0;
		}
		
		/* detect start time for first event */
		if(min_start_time > start_time)
			min_start_time = start_time;
		
		/* detect stop time for last event */
		if(max_stop_time < stop_time)
			max_stop_time = stop_time;
		
		/* compute total duration */
		run_time = max_stop_time - min_start_time;
		
		/* increment total number of jobs */
		job_num++;
		
		/* compute job duration time */
		proc_time = stop_time - start_time;
		
		/* check with min duration time */
		if (proc_time < min_proc_time)
			min_proc_time = proc_time;
		if (proc_time < min_proc_times[core_idx])
			min_proc_times[core_idx] = proc_time;
		
		/* check with max duration time */
		if (proc_time > max_proc_time)
			max_proc_time = proc_time;
		if (proc_time > max_proc_times[core_idx])
			max_proc_times[core_idx] = proc_time;
		
		job_nums[core_idx][flow_idx]++;
		mean_proc_time            += proc_time;
		mean_proc_times[core_idx] += proc_time;
		
		if (stop_times[core_idx] != 0) {
			/* compute duration time */
			ovh_time = start_time - stop_times[core_idx];
			
			/* check with min duration time */
			if (ovh_time < min_ovh_time)
				min_ovh_time = ovh_time;
			if (ovh_time < min_ovh_times[core_idx])
				min_ovh_times[core_idx] = ovh_time;
			
			/* check with max duration time */
			if (ovh_time > max_ovh_time)
				max_ovh_time = ovh_time;
			if (ovh_time > max_ovh_times[core_idx])
				max_ovh_times[core_idx] = ovh_time;
			
			mean_ovh_count ++;
			mean_ovh_time            += ovh_time;
			mean_ovh_times[core_idx] += ovh_time;
		}
		stop_times[core_idx] = stop_time;
	}
	
	TI_EM_OSAL_PRINTF("\n==========\n");

	if (success_flag == 1)
		TI_EM_OSAL_PRINTF("Run PASSED!\n");
	else 
		TI_EM_OSAL_PRINTF("Run FAILED!\n");
	
	/* per core statistics */
	for (core_idx = 0; core_idx < MY_EM_CORE_NUM; core_idx++)
	{
		uint32_t job_num2 = 0;

		for (flow_idx = 0; flow_idx < MY_FLOW_NUM; flow_idx++)
			job_num2 += job_nums[core_idx][flow_idx];
		
		TI_EM_OSAL_PRINTF("\nStatistics for core %d:\n", core_idx);
		TI_EM_OSAL_PRINTF("Number of jobs: %d\n", job_num2);
		TI_EM_OSAL_PRINTF("Number of jobs per flow: ");

		for (flow_idx = 0; flow_idx < MY_FLOW_NUM; flow_idx++)
			TI_EM_OSAL_PRINTF("%d ",job_nums[core_idx][flow_idx]);

		TI_EM_OSAL_PRINTF("\n");
		
		if (job_num2 > 0)
			TI_EM_OSAL_PRINTF("Min/mean/max processing cycles: %d/%d/%d\n",
				     min_proc_times[core_idx],
				     mean_proc_times[core_idx] / job_num2,
				     max_proc_times[core_idx]);
		
		if (job_num2 > 1)
			TI_EM_OSAL_PRINTF("Min/mean/max overhead cycles: %d/%d/%d\n\n",
				     min_ovh_times[core_idx],
				     mean_ovh_times[core_idx] / (job_num2 - 1),
				     max_ovh_times[core_idx]);
	}
	
	/* print statistics */
	TI_EM_OSAL_PRINTF("\nStatistics for all cores\n");
	TI_EM_OSAL_PRINTF("Total number of jobs: %d\n", job_num);
	TI_EM_OSAL_PRINTF("Total number of jobs per flow: ");

	for (flow_idx = 0; flow_idx < MY_FLOW_NUM; flow_idx++)
	{
		uint32_t job_num2 = 0;
		
		for (core_idx = 0; core_idx < MY_EM_CORE_NUM; core_idx++)
			job_num2 += job_nums[core_idx][flow_idx];

		TI_EM_OSAL_PRINTF("%d ", job_num2);
	}
	TI_EM_OSAL_PRINTF("\n");
	
	if (job_num > 0)
		TI_EM_OSAL_PRINTF("Min/mean/max processing cycles: %d/%d/%d\n",
			     min_proc_time,
			     mean_proc_time / job_num,
			     max_proc_time);
	
	if (mean_ovh_count > 0)
		TI_EM_OSAL_PRINTF("Min/mean/max overhead cycles: %d/%d/%d\n\n",
			     min_ovh_time,
			     mean_ovh_time / mean_ovh_count,
			     max_ovh_time);
	
	TI_EM_OSAL_PRINTF("Start/stop/run cycles: %d/%d/%d\n",
		     min_start_time,
		     max_stop_time,
		     run_time);

	TI_EM_OSAL_PRINTF("\n==========\n");
}
