/*
 * Copyright (c) 2012, Texas Instruments Incorporated - http://www.ti.com/
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions are met:
 *       * Redistributions of source code must retain the above copyright
 *         notice, this list of conditions and the following disclaimer.
 *       * Redistributions in binary form must reproduce the above copyright
 *         notice, this list of conditions and the following disclaimer in the
 *         documentation and/or other materials provided with the distribution.
 *       * Neither the name of Texas Instruments Incorporated nor the
 *         names of its contributors may be used to endorse or promote products
 *         derived from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *   DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
 *   DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


#include "my_em_init.h"
#include "my_receive.h"

#include <ti_em_osal.h>
#include <ti_em_osal_queue.h>
#include <ti_em_osal_core.h>
#include <ti_em_osal_uaccess.h>
#include <ti_em_osal_cma.h>
#include <ti_em_osal_shm.h>

#include <ti_em_rh.h>

#include <ti/runtime/openem/firmware/event_machine_firmware.h>

/* 
 * If defined, we use local buffers (allocated with malloc) with virtual addresses
 * In this case they cannot be used for chaining.
 */
static em_eo_t    my_em_sink_eo_hdl;
static em_eo_t    my_em_exit_eo_hdl;
static em_queue_t my_em_exit_queue_hdl;

em_queue_t *my_em_proc_queue_hdl_tbl;
em_queue_t *my_em_sink_queue_hdl;

static void my_exit_local(
	void*           ctx,
	em_event_t      event,
	em_event_type_t event_type,
	em_queue_t      queue,
	void*           queue_txt)
{
	em_free(event);
	(*my_sv_running_job_num)++;
	em_atomic_processing_end ();

	TI_EM_OSAL_TRACE(1, "exiting...\n");

	ti_em_rh_release_local();
	ti_em_osal_core_exit(0);
}

static int my_init_pool_config(ti_em_rh_pool_config_t* pool_config)
{
	/* Pool of application events initialization */
	pool_config[MY_EM_POOL_IDX].pool.free_queue_idx =
		MY_EM_FREE_QUEUE_BASE_IDX + MY_EM_POOL_IDX;
	pool_config[MY_EM_POOL_IDX].pool.buf_mode  = MY_EM_BUF_MODE;
	pool_config[MY_EM_POOL_IDX].pool.dsc_wsize = MY_EM_PUBLIC_EVENT_DSC_SIZE;

	pool_config[MY_EM_POOL_IDX].mem.paddr  = 0; /* use CMA */
	pool_config[MY_EM_POOL_IDX].mem.vaddr  = 0;
	pool_config[MY_EM_POOL_IDX].mem.size   = 0; /* computed */
	pool_config[MY_EM_POOL_IDX].mem.offset = 0; /* none */
	pool_config[MY_EM_POOL_IDX].mem.flags  = TI_EM_OSAL_MEM_CACHED;

       	pool_config[MY_EM_POOL_IDX].event_num   = MY_EM_EVENT_NUM;
	pool_config[MY_EM_POOL_IDX].free_policy = 0;
	pool_config[MY_EM_POOL_IDX].buffer_size = MY_EM_EVENT_BUF_SIZE;
	pool_config[MY_EM_POOL_IDX].offset      = 0;
	pool_config[MY_EM_POOL_IDX].ps_word_num = 0;
    
	/* Pool of exit events initialization */
	pool_config[MY_EM_EXIT_POOL_IDX].pool.free_queue_idx =
	    MY_EM_FREE_QUEUE_BASE_IDX + MY_EM_EXIT_POOL_IDX;
	pool_config[MY_EM_EXIT_POOL_IDX].pool.buf_mode  = TI_EM_BUF_MODE_TIGHT;
	pool_config[MY_EM_EXIT_POOL_IDX].pool.dsc_wsize = MY_EM_PUBLIC_EVENT_DSC_SIZE;

	pool_config[MY_EM_EXIT_POOL_IDX].mem.paddr = -1; /* no payload */

	pool_config[MY_EM_EXIT_POOL_IDX].event_num   = MY_EM_EXIT_EVENT_NUM;
	pool_config[MY_EM_EXIT_POOL_IDX].free_policy = 0;
	pool_config[MY_EM_EXIT_POOL_IDX].buffer_size = 0;
	pool_config[MY_EM_EXIT_POOL_IDX].offset      = 0;
	pool_config[MY_EM_EXIT_POOL_IDX].ps_word_num = 0;

	/* Pools for inter-process chaining */
	pool_config[MY_EM_PROCESS_CHAIN0_POOL_IDX].pool.free_queue_idx =
		MY_EM_FREE_QUEUE_BASE_IDX + MY_EM_PROCESS_CHAIN0_POOL_IDX;
	pool_config[MY_EM_PROCESS_CHAIN0_POOL_IDX].pool.buf_mode  = MY_EM_PROCESS_CHAIN0_BUF_MODE;
	pool_config[MY_EM_PROCESS_CHAIN0_POOL_IDX].pool.dsc_wsize = MY_EM_PUBLIC_EVENT_DSC_SIZE;

	pool_config[MY_EM_PROCESS_CHAIN0_POOL_IDX].mem.paddr  = 0; /* use CMA */
	pool_config[MY_EM_PROCESS_CHAIN0_POOL_IDX].mem.vaddr  = 0;
	pool_config[MY_EM_PROCESS_CHAIN0_POOL_IDX].mem.size   = 0; /* computed */
	pool_config[MY_EM_PROCESS_CHAIN0_POOL_IDX].mem.offset = 0; /* none */
	pool_config[MY_EM_PROCESS_CHAIN0_POOL_IDX].mem.flags  = TI_EM_OSAL_MEM_CACHED;

	pool_config[MY_EM_PROCESS_CHAIN0_POOL_IDX].event_num   = MY_EM_PROCESS_CHAIN0_EVENT_NUM;
	pool_config[MY_EM_PROCESS_CHAIN0_POOL_IDX].free_policy = 1;
	pool_config[MY_EM_PROCESS_CHAIN0_POOL_IDX].buffer_size = MY_EM_PROCESS_CHAIN0_EVENT_BUF_SIZE;
	pool_config[MY_EM_PROCESS_CHAIN0_POOL_IDX].offset      = 0;
	pool_config[MY_EM_PROCESS_CHAIN0_POOL_IDX].ps_word_num = 0;

	return 0;
}

static void my_em_init_hw_chain_config(ti_em_chain_config_t *chain_config)
{
	/* Index of the various pools used for chaining */
	chain_config->pool_idx_size_a   = MY_EM_PROCESS_CHAIN0_POOL_IDX;;
	chain_config->pool_idx_size_b   = MY_EM_PROCESS_CHAIN0_POOL_IDX;;
	chain_config->pool_idx_size_c   = MY_EM_PROCESS_CHAIN0_POOL_IDX;;
	chain_config->pool_idx_size_d   = MY_EM_PROCESS_CHAIN0_POOL_IDX;;
	chain_config->pool_idx_overflow = MY_EM_PROCESS_CHAIN0_POOL_IDX;;

	/* Relative queue for chaining */
	chain_config->dma_tx_queue_base_idx = MY_EM_DMA_QUEUE_BASE_IDX;
	chain_config->dma_idx               = Cppi_CpDma_QMSS_CPDMA;
}

int my_em_init_global(void)
{
	ti_em_osal_hw_region_config_t region_configs[MY_EM_REGION_NUM];
	ti_em_rh_pool_config_t        pool_configs[MY_EM_POOL_NUM];
	ti_em_rh_hw_config_t          hw_config;
	ti_em_chain_config_t          hw_chain_config;
	ti_em_process_route_t         process_route;
	int                           res;

	memset(&hw_config, 0, sizeof(ti_em_rh_hw_config_t));
	memset(&hw_chain_config, 0, sizeof(ti_em_chain_config_t));
	memset(&process_route, 0, sizeof(ti_em_process_route_t));

	/* Initialize pools and their configuration */
	res = my_init_pool_config(pool_configs);
	if (res < 0)
		return res;

	/* Initialize chaining configurations */
	my_em_init_hw_chain_config(&hw_chain_config);

	/* Set OpenEM initialization parameters */
	hw_config.private_free_queue_idx = MY_EM_PRIVATE_FREE_QUEUE_IDX;
	hw_config.hw_queue_base_idx      = MY_EM_HW_QUEUE_BASE_IDX;
	hw_config.dma_idx                = Cppi_CpDma_QMSS_CPDMA;
	hw_config.dma_queue_base_idx     = MY_EM_DMA_QUEUE_BASE_IDX + 1;
	hw_config.device_id              = MY_EM_DEVICE_ID;
	hw_config.process_id             = MY_EM_PROCESS_ID;
	hw_config.chain_config_ptr       = &hw_chain_config;
	hw_chain_config.xge_config_ptr   = NULL;

	/* The location of the PDSP communication memory (physical address) */
	hw_config.pdsp_comm_mem_config.paddr  = MY_EM_PDSP_COMM_MEM_BASE;
	hw_config.pdsp_comm_mem_config.vaddr  = MY_EM_PDSP_COMM_MEM_VBASE;
	hw_config.pdsp_comm_mem_config.size   = MY_EM_PDSP_COMM_MEM_SIZE;
	hw_config.pdsp_comm_mem_config.offset = MY_EM_PDSP_COMM_MEM_OFFSET;

	TI_EM_OSAL_TRACE(2, "physical address of the PDSP communication memory is 0x%x\n",
			 hw_config.pdsp_comm_mem_config.paddr);
	
	/* Define descriptor regions */
	region_configs[TI_EM_RH_PUBLIC].region_idx   = MY_EM_PUBLIC_REGION_IDX;
	region_configs[TI_EM_RH_PUBLIC].desc_size    = MY_EM_PUBLIC_EVENT_DSC_SIZE;
	region_configs[TI_EM_RH_PUBLIC].desc_num     = MY_EM_PUBLIC_EVENT_NUM;
	region_configs[TI_EM_RH_PUBLIC].desc_base    = MY_EM_PUBLIC_DESC_BASE;
	region_configs[TI_EM_RH_PUBLIC].desc_vbase   = MY_EM_PUBLIC_DESC_VBASE;
	region_configs[TI_EM_RH_PUBLIC].desc_offset  = MY_EM_PUBLIC_DESC_OFFSET;
	region_configs[TI_EM_RH_PUBLIC].desc_flag    = TI_EM_RH_UNMANAGED_DESCRIPTOR;
	region_configs[TI_EM_RH_PUBLIC].start_idx    = MY_EM_PUBLIC_START_DESC_IDX;

	region_configs[TI_EM_RH_PRIVATE].region_idx  = MY_EM_PRIVATE_REGION_IDX;
	region_configs[TI_EM_RH_PRIVATE].desc_size   = TI_EM_PRIVATE_EVENT_DSC_SIZE;
	region_configs[TI_EM_RH_PRIVATE].desc_num    = TI_EM_RH_PRIVATE_EVENT_NUM;
	region_configs[TI_EM_RH_PRIVATE].desc_base   = MY_EM_PRIVATE_DESC_BASE;
	region_configs[TI_EM_RH_PRIVATE].desc_vbase  = MY_EM_PRIVATE_DESC_VBASE;
	region_configs[TI_EM_RH_PRIVATE].desc_offset = MY_EM_PRIVATE_DESC_OFFSET;
	region_configs[TI_EM_RH_PRIVATE].desc_flag   = TI_EM_RH_UNMANAGED_DESCRIPTOR;
	region_configs[TI_EM_RH_PRIVATE].start_idx   = MY_EM_PRIVATE_START_DESC_IDX;

	hw_config.region_num     = MY_EM_REGION_NUM;
	hw_config.region_configs = &region_configs[0];

	/* Define PDSP configuration */
	hw_config.pdsp_num = 1;
	hw_config.pdsp_sched[0].pdsp_id = MY_EM_SCHED_PDSP_IDX;
	if (hw_config.pdsp_sched[0].pdsp_id == 1) {
	    hw_config.pdsp_sched[0].firmware = &TI_EM_FIRMWARE_SCHED_PDSP2;
	    hw_config.pdsp_sched[0].size     = sizeof(TI_EM_FIRMWARE_SCHED_PDSP2);
	} else {
	    hw_config.pdsp_sched[0].firmware = &TI_EM_FIRMWARE_SCHED_PDSP1;
	    hw_config.pdsp_sched[0].size     = sizeof(TI_EM_FIRMWARE_SCHED_PDSP1);
	}
	hw_config.pdsp_router.pdsp_id  = -1; /* do not use router (no XGE chaining) */

	TI_EM_OSAL_TRACE(1, "calling EM global initialization\n");

	/* call OpenEM global initialization */
	if (ti_em_rh_init_global(MY_EM_POOL_NUM,
				 pool_configs,
				 MY_EM_CORE_NUM,
				&hw_config) != EM_OK) {
		TI_EM_OSAL_ERROR("EM global initialization failed!\n");
		return -1;
	}

	process_route.dma_idx          = Cppi_CpDma_QMSS_CPDMA;
        process_route.dma_tx_queue_idx = MY_DSP_EM_FLOW_IDX;

	if (ti_em_process_add_route(process_route,
				    MY_DSP_EM_PROCESS_IDX) != EM_OK) {
	    TI_EM_OSAL_ERROR("Adding process route failed!\n");
	    return -1;
	}

	TI_EM_OSAL_TRACE(0, "add route to process done\n");

	return 0;
}

int my_em_init_local(void)
{
	/* call OpenEM local initialization */
	if (ti_em_rh_init_local() != EM_OK)
		return -1;

	return 0;
}

int my_em_shared_init(int init)
{
	/* initialize my shared memory */
	ti_em_osal_shm_init(MY_SHARED_MEM_ID,
			    MY_SHARED_MEM_NAME,
			    MY_SHARED_MEM_SIZE,
			    init);
	
	/* allocate queue handler table */
	my_em_proc_queue_hdl_tbl = (em_queue_t*) 
		ti_em_osal_shm_alloc(MY_SHARED_MEM_ID,
				     sizeof(em_queue_t) * MY_EM_PROC_QUEUE_NUM);
	
	my_em_sink_queue_hdl = (em_queue_t*) 
		ti_em_osal_shm_alloc(MY_SHARED_MEM_ID, sizeof(em_queue_t));
	
	/* alloc global job counter */
	my_sv_running_job_num = (typeof(my_sv_running_job_num)) 
	    ti_em_osal_shm_alloc(MY_SHARED_MEM_ID, 
				 sizeof(*my_sv_running_job_num));
	
	/* 
	 * Allocate global job descriptor table (accross all cores)
	 */
	my_sv_job_desc_tbl =
		(my_job_desc_t*) ti_em_osal_shm_alloc(MY_SHARED_MEM_ID,
						      MY_JOB_NUM * sizeof(my_job_desc_t));
	
	if (!my_sv_job_desc_tbl) {
		TI_EM_OSAL_ERROR("job descriptor table allocation failed");
		return -1;
	}

	return 0;
}

static em_status_t eo_start_default(void* eo_ctxt_ptr, em_eo_t eo_hdl)
{
	return EM_OK;
}

static em_status_t eo_stop_default(void* eo_ctxt_ptr, em_eo_t eo_hdl)
{
	return EM_OK;
}

int my_em_init_queues(uint32_t core_num)
{
	uint32_t i;
	em_core_mask_t cmask;
	em_queue_group_t exit_queue_group;

	TI_EM_OSAL_TRACE(2, "EO creation\n");

	/* "Sink EO" creation */
	my_em_sink_eo_hdl = em_eo_create("Sink EO",
					 eo_start_default, NULL,
					 eo_stop_default, NULL,
					 my_sink_job, NULL);

	if (my_em_sink_eo_hdl == EM_EO_UNDEF)
		return -1;
	
	/* "Sink Queue" create */
	*my_em_sink_queue_hdl = em_queue_create("Sink Queue",
						EM_QUEUE_TYPE_PARALLEL,
						EM_QUEUE_PRIO_LOW,
						EM_QUEUE_GROUP_DEFAULT);
	
	if (*my_em_sink_queue_hdl == EM_QUEUE_UNDEF)
		return -1;
	
	/* add "Sink Queue" to "Sink EO" */
	if (em_eo_add_queue(my_em_sink_eo_hdl,
			    *my_em_sink_queue_hdl) != EM_OK)
		return -1;
	
	/* "Sink EO" start */
	if (em_eo_start(my_em_sink_eo_hdl,
			NULL,
			0,
			NULL) != EM_OK) return -1;

	/* "Exit EO" creation */
	my_em_exit_eo_hdl = em_eo_create("Exit EO",
					 eo_start_default, NULL,
					 eo_stop_default, NULL,
					 my_exit_local, NULL);
	
	if (my_em_exit_eo_hdl == EM_EO_UNDEF)
		return -1;

	/* "Exit group" create */
	em_core_mask_zero(&cmask);
	for (i = 1; i < core_num; i++)
		em_core_mask_set(i, &cmask);

	exit_queue_group = em_queue_group_create("Exit_Queue_Group",
						 (const em_core_mask_t*) &cmask,
						 0,
						 NULL);
	
	/* "Exit Queue" create */
	my_em_exit_queue_hdl = em_queue_create("Exit Queue",
					       EM_QUEUE_TYPE_ATOMIC,
					       EM_QUEUE_PRIO_LOWEST,
					       exit_queue_group);
	
	if (my_em_exit_queue_hdl == EM_QUEUE_UNDEF)
		return -1;
	
	/* add "Exit Queue" to "Exit EO" */
	if (em_eo_add_queue(my_em_exit_eo_hdl,
			    my_em_exit_queue_hdl) != EM_OK)
		return -1;
	
	/* "Exit EO" start */
	if (em_eo_start(my_em_exit_eo_hdl,
			NULL,
			0,
			NULL) != EM_OK)
		return -1;

	/* Adding a HW static queue to communicate w/ DSP process */
	if (ti_em_queue_create_hw_static("DspToLinuxHwQueue",
					 (ti_em_queue_id_t) MY_DSP_EM_IS_CONFIGURED_QUEUE_IDX,
					 (em_queue_t) MY_LINUX_HW_QUEUE_HANDLE) != EM_OK)
		return -1;

	return 0;
}


static void dsp_wait(void)
{
	void *desc;

	desc = NULL;
	while(desc == NULL) {
		desc = ti_em_osal_hw_queue_pop(MY_DSP_EM_IS_CONFIGURED_QUEUE_IDX,
					       TI_EM_MEM_PUBLIC_DESC);
	}
}

int my_em_dsp_sync(void)
{
	void *desc;

	desc = ti_em_osal_hw_queue_pop(MY_EM_FREE_QUEUE_BASE_IDX + MY_EM_POOL_IDX,
				       TI_EM_MEM_PUBLIC_DESC);

	if (desc == NULL) {
		TI_EM_OSAL_ERROR("cannot get free desc\n");
		return -1;
	}

	ti_em_osal_hw_queue_push(MY_LINUX_EM_IS_CONFIGURED_QUEUE_IDX,
				 desc,
				 TI_EM_MEM_PUBLIC_DESC);

	dsp_wait();

	return 0;
}

int my_em_exit_global(uint32_t core_num)
{
	int i;

	*my_sv_running_job_num = 0;

	/* Wait the DSP termination */
	dsp_wait();
	
	for (i = 1; i < core_num; i++) {
		em_event_t event;
		
		event = em_alloc(0,
				 TI_EM_EVENT_TYPE_PRELOAD_OFF,
				 MY_EM_EXIT_POOL_IDX);
		
		if (event == 0)
			return -1;
		
		em_send(event, my_em_exit_queue_hdl);
	}

	while (*my_sv_running_job_num  < (core_num -1))
	    ti_em_rh_dispatch(TI_EM_RH_POLL_MODE, 0, 0);

	return 0;
}
