/*
 * Copyright (c) 2012, Texas Instruments Incorporated - http://www.ti.com/
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions are met:
 *       * Redistributions of source code must retain the above copyright
 *         notice, this list of conditions and the following disclaimer.
 *       * Redistributions in binary form must reproduce the above copyright
 *         notice, this list of conditions and the following disclaimer in the
 *         documentation and/or other materials provided with the distribution.
 *       * Neither the name of Texas Instruments Incorporated nor the
 *         names of its contributors may be used to endorse or promote products
 *         derived from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *   DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
 *   DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


#include "my_receive.h"
#include "my_em_init.h"

#include <ti_em_osal.h>
#include <ti_em_osal_core.h>
#include <ti_em_osal_queue.h>
#include <ti_em_osal_shm.h>

#define MY_FFT_SIZE_MIN     (128)
#define MY_FFT_SIZE_MAX     (4 * 128)

#define MY_JOB_HDR_SIZE     (128) /* DSP cacheline size */
#define MY_JOB_PLD_SIZE     (2 * MY_FFT_SIZE_MAX * sizeof(int16_t))
#define MY_JOB_SIZE         (MY_JOB_HDR_SIZE + MY_JOB_PLD_SIZE)

/* Pointed memory of these pointers must be shared accross cores */
volatile my_job_desc_t *my_sv_job_desc_tbl;
volatile uint32_t      *my_sv_running_job_num;

int my_init_jobs()
{
	uint32_t                 i, j;

	if (MY_JOB_NUM > 2 * MY_EM_EVENT_NUM)
		return -1;
	if (MY_FLOW_NUM > MY_EM_PROC_QUEUE_NUM)
		return -1;
	if (MY_JOB_SIZE > MY_EM_EVENT_BUF_SIZE)
		return -1;
	if (sizeof(my_job_desc_t) > MY_JOB_HDR_SIZE)
		return -1;

	for (i = 0; i < MY_FLOW_NUM; i++) {
		for (j = 0; j < MY_JOB_NUM_PER_FLOW; j++) {

			uint32_t job_idx;
			uint32_t fft_size;

			fft_size = MY_FFT_SIZE_MIN * (1 << (rand() % 2));
			if (fft_size > MY_FFT_SIZE_MAX)
				fft_size = MY_FFT_SIZE_MAX;
			
			job_idx = i * MY_JOB_NUM_PER_FLOW + j;
			my_sv_job_desc_tbl[job_idx].job_idx   = job_idx;
			my_sv_job_desc_tbl[job_idx].flow_idx  = i;
			my_sv_job_desc_tbl[job_idx].fft_size  = fft_size;
			my_sv_job_desc_tbl[job_idx].data_size = 
				MY_JOB_HDR_SIZE + 2 * fft_size * sizeof (int16_t);
			my_sv_job_desc_tbl[job_idx].success_flag = 0;
			my_sv_job_desc_tbl[job_idx].core_idx     = MY_EM_CORE_NUM;
			my_sv_job_desc_tbl[job_idx].start_time   = 0;
			my_sv_job_desc_tbl[job_idx].stop_time    = 0;
		}
	}
	
	*my_sv_running_job_num = MY_JOB_NUM;

	return 0;
}

static void my_fill_buffer(uint8_t *buffer_ptr, uint32_t seed, uint32_t size)
{
	int16_t  *data_tbl = (int16_t *)  buffer_ptr;
	uint32_t  j;

	/*
         * set FFT input data 
         */
        for (j = 0; j < size; j++) {
		switch (j & 0x3) {
                case 0:
			*data_tbl++ = 1;
			*data_tbl++ = 0;
			break;
                case 1:
			*data_tbl++ = 0;
			*data_tbl++ = 1;
			break;
                case 2:
			*data_tbl++ = (-1);
			*data_tbl++ = 0;
			break;
                case 3:
			*data_tbl++ = 0;
			*data_tbl++ = (-1);
			break;
		default:
			break;
		}
        }
}

int my_interprocess_init(void)
{
	uint32_t                 i;
	em_event_t               inter_process_event = 0;
	my_queuedesc_from_linux *buff1_ptr;
	my_queuedesc_to_linux   *buff0_ptr;

	TI_EM_OSAL_TRACE(2, "Wait inter process event from %d\n",
			 MY_DSP_EM_IS_CONFIGURED_QUEUE_IDX);

	/* wait inter process event from DSP */
	while (inter_process_event == 0)
		inter_process_event = (em_event_t)
			ti_em_osal_hw_queue_pop(MY_DSP_EM_IS_CONFIGURED_QUEUE_IDX,
						TI_EM_MEM_PUBLIC_DESC);

	buff0_ptr = (my_queuedesc_to_linux *) em_event_pointer (inter_process_event);

	/* fill the queue handler table with information sent from DSP */
	for (i = 0; i < MY_EM_PROC_QUEUE_NUM; i++) {
		my_em_proc_queue_hdl_tbl[i] = buff0_ptr->dsp_proc_queue_hdl_tbl[i];
		TI_EM_OSAL_TRACE(2, "DSP: my_em_svProcQueueHdlTbl[%u] = 0x%08X\n",
				 i, my_em_proc_queue_hdl_tbl[i]);
	}

	em_free(inter_process_event);

	inter_process_event = em_alloc (sizeof(my_queuedesc_from_linux),
					0,
					MY_EM_POOL_IDX);
	
	buff1_ptr = (my_queuedesc_from_linux *) em_event_pointer (inter_process_event);
	buff1_ptr->nb_jobs = MY_JOB_NUM;
	buff1_ptr->linux_sink_queue_hdl = *my_em_sink_queue_hdl;

	TI_EM_OSAL_TRACE(2, "nb_jobs = %u\n", buff1_ptr->nb_jobs);
	TI_EM_OSAL_TRACE(2, "linux_sink_queue_hdl = 0x%x\n", buff1_ptr->linux_sink_queue_hdl);
	TI_EM_OSAL_TRACE(2, "send event 0x%x to queue 0x%x\n",
			 inter_process_event, MY_DSP_HW_QUEUE_HANDLE);
	
	if (em_send(inter_process_event, MY_DSP_HW_QUEUE_HANDLE) != EM_OK)
		return -1;
	
	return 0;
}

int my_source_jobs(void)
{
	uint32_t i;

	/* loop over all jobs to create */
	for (i = 0; i < MY_JOB_NUM; i++)
	{
		volatile em_event_t event;
		uint8_t            *job_buf;
		my_job_desc_t      *job_desc;
		uint32_t            flow_idx;
		uint32_t            data_size;
		uint32_t            fft_size;
		uint32_t            success_flag;

		flow_idx     = my_sv_job_desc_tbl[i].flow_idx;
		data_size    = my_sv_job_desc_tbl[i].data_size;
		fft_size     = my_sv_job_desc_tbl[i].fft_size;
		success_flag = my_sv_job_desc_tbl[i].success_flag;

		/* allocate event */
		event = EM_EVENT_UNDEF;
		while(event == EM_EVENT_UNDEF) {
			event = em_alloc(data_size,
					 MY_EM_PROC_EVENT_TYPE,
					 MY_EM_POOL_IDX);
		}

		/* acess event buffer */
		job_desc = (my_job_desc_t*) em_event_pointer(event);
		job_buf  = (uint8_t*) ((uint32_t) job_desc + MY_JOB_HDR_SIZE);

		/* set ctrl information */
		job_desc->job_idx      = i;
		job_desc->flow_idx     = flow_idx;
		job_desc->data_size    = data_size;
		job_desc->fft_size     = fft_size;
		job_desc->success_flag = success_flag;

		/* set job input data */
		my_fill_buffer(job_buf, i, fft_size);
		
		/* send event to EQs of the remote device */
		em_send(event, my_em_proc_queue_hdl_tbl[flow_idx]);

		TI_EM_OSAL_TRACE(3, "event for job %d sent to 0x%x\n",
				 i, my_em_proc_queue_hdl_tbl[flow_idx]);
	}

	return 0;
}

void my_sink_job(
	void*           ctx,
	em_event_t      event,
	em_event_type_t event_type,
	em_queue_t      queue,
	void           *queue_ctx)
{
	my_job_desc_t *job_desc;
	uint32_t       job_idx;

	TI_EM_OSAL_TRACE(3, "calling sink job\n");

	/* prefetch request */
	ti_em_preschedule();

	/* access event buffer of received event */
	job_desc = (my_job_desc_t*) em_event_pointer(event);

	/* 
	 * retrieve ctrl information from received
	 * event buffer
	 */
	job_idx = job_desc->job_idx;

	my_sv_job_desc_tbl[job_idx].success_flag = job_desc->success_flag;

	/* free received event */
	em_free(event);

	/* decrement shared job counter */
	ti_em_osal_core_lock();
	(*my_sv_running_job_num)--;
	ti_em_osal_core_unlock();
}

void my_stats_jobs(void)
{
	uint32_t job_idx;
	uint32_t job_num      = 0;
	uint32_t success_flag = 1;

	/* compute statistics */
	for (job_idx = 0; job_idx < MY_JOB_NUM; job_idx++) {	
		if(my_sv_job_desc_tbl[job_idx].success_flag == 0) {
			TI_EM_OSAL_PRINTF("job %x not success\n", job_idx);
			success_flag = 0;
		}
		/* increment total number of jobs */
		job_num++;
	}
	
	TI_EM_OSAL_PRINTF("\n==========\n");

	if (success_flag == 1)
		TI_EM_OSAL_PRINTF("Run PASSED!\n");
	else 
		TI_EM_OSAL_PRINTF("Run FAILED!\n");
	
	TI_EM_OSAL_PRINTF("Total number of jobs: %d\n", job_num);

	TI_EM_OSAL_PRINTF("\n==========\n");
}
