/*
 * Copyright (c) 2012, Texas Instruments Incorporated - http://www.ti.com/
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions are met:
 *       * Redistributions of source code must retain the above copyright
 *         notice, this list of conditions and the following disclaimer.
 *       * Redistributions in binary form must reproduce the above copyright
 *         notice, this list of conditions and the following disclaimer in the
 *         documentation and/or other materials provided with the distribution.
 *       * Neither the name of Texas Instruments Incorporated nor the
 *         names of its contributors may be used to endorse or promote products
 *         derived from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *   DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
 *   DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


#include "my_em_init.h"
#include "my_receive.h"

/* Importing some of the Open EM defined symbols */
#include "my_event_machine.h"

#include <ti_em_osal.h>
#include <ti_em_osal_core.h>
#include <ti_em_osal_shm.h>

#include <ti_em_rh.h>

#include <getopt.h>

static uint32_t my_em_core_num      = MY_EM_CORE_NUM;
static uint32_t my_em_iteration     = 1;
static uint32_t my_em_delay         = 10;
static uint32_t my_em_dispatch_mode = MY_EM_DISPATCH_MODE;

static int my_global_init(void)
{
	TI_EM_OSAL_TRACE(1, "master core initialization\n");

	/* example shared objects creation */
	my_em_shared_init(1);
	
	/* OpenEM global initialization */
	if (my_em_init_global() < 0)
		goto error;
	
	/* application timer global initialization */
	if (ti_em_osal_timer_init(TI_EM_OSAL_TIMER_GLOBAL) < 0)
		goto error;

	/* local initialization */
	if (my_em_init_local() < 0)
		goto error;

	/* OpenEM objects creation */
	if (my_em_init_queues(my_em_core_num) < 0) {
		TI_EM_OSAL_ERROR("objects creation error\n");
		goto error;
	}
	TI_EM_OSAL_TRACE(1, "objects creation done\n");

	return 0;
error:
	ti_em_osal_shm_release(MY_SHARED_MEM_ID);
	return -1;
}

static int my_dispatcher_local_init(void)
{
	TI_EM_OSAL_TRACE(1, "local init\n");
	
	/* OpenEM local initialization */
	if (my_em_init_local() < 0)
		return -1;
	
	/* application timer local initialization */
	if (ti_em_osal_timer_init(TI_EM_OSAL_TIMER_LOCAL) < 0)
		return -1;

	return 0;
}

static int my_dispatcher(void* arg)
{
       	TI_EM_OSAL_TRACE(2, "entering dispatcher\n");

	/* request first event from scheduler */
	ti_em_preschedule();

	/* call event dispatcher */
	while (1) {
		ti_em_rh_dispatch(my_em_dispatch_mode, 10, 10);
	}

	return 0; /* not reached */
}

int main(int argc, char **argv)
{
	int c;
	int lop;
	int iter;

	struct option long_options[] = {
		{ "core-num", 1, 0, 'n'},
		{ "multi-pdsp", 0, 0, 'P'},
		{ "iteration", 1, 0, 'I'},
		{ "delay", 1, 0, 'D'},
		{ "poll-mode", 0, 0, 'p'},
		{ 0, 0, 0, 0 }
	};

	TI_EM_OSAL_PRINTF("starting Example 3\n");

	/* runtime options parse */
	while ((c = getopt_long(argc, argv, "n:PI:D:p", long_options, &lop)) != EOF) {
		switch(c) {
		case 'n':
			my_em_core_num = strtoul(optarg, (char **) NULL, 10);
			break;
		case 'P':
			my_em_pdsp_num = MY_EM_PDSP_NUM;
			break;
		case 'I':
			my_em_iteration = strtoul(optarg, (char **) NULL, 10);
			break;
		case 'D':
			my_em_delay = strtoul(optarg, (char **) NULL, 10);
			break;
		case 'p':
			my_em_dispatch_mode = TI_EM_RH_POLL_MODE;
			break;
		default:
			break;
		}
	}

	if (my_em_dispatch_mode == TI_EM_RH_INTERRUPT_MODE)
		TI_EM_OSAL_PRINTF("using interrupt mode\n");
	else
		TI_EM_OSAL_PRINTF("using polling mode\n");

	TI_EM_OSAL_PRINTF("using %d delay (usec)\n", my_em_delay);
	TI_EM_OSAL_PRINTF("using %d scheduler(s)\n", my_em_pdsp_num);
		
	/* 
	 * OSAL core runtime library initialization (must only called by main core) 
	 */
	ti_em_osal_core_init_global();

	if ((my_em_core_num > MY_EM_CORE_NUM) || (my_em_core_num < 1)) {
		TI_EM_OSAL_ERROR("invalid number of cores\n");
		exit(-1);
	}  else
		TI_EM_OSAL_TRACE(1, "running %d cores\n", my_em_core_num);
	       
	/* initialization */
	if (my_global_init() < 0) {
		return -1;
	}

	/* create and starts cores */
 	if (ti_em_rh_create_cores(my_em_core_num,
				  my_dispatcher,
				  my_dispatcher_local_init,
				  MY_EM_INIT_CORE_IDX) < 0) {
		TI_EM_OSAL_ERROR("core creation failed\n");
		return -1;
	}

	/* application timer local initialization */
	if (ti_em_osal_timer_init(TI_EM_OSAL_TIMER_LOCAL) < 0)
		return -1;

	TI_EM_OSAL_PRINTF("\nrunning %d iterations\n", my_em_iteration);

	/* start test */
	for (iter = 0; iter < my_em_iteration; iter++) {
		my_init_jobs();
		my_source_jobs(my_em_delay);
		
		/* wait until end of test */
		while (*my_sv_running_job_num > 0);
		
		if ((my_em_iteration > 1) && (iter == 0)) {
			TI_EM_OSAL_PRINTF("\ndummy iteration\n");
			my_stats_jobs(0);
		}
	}
	/* compute job statistics */
	my_stats_jobs(1);

	/* terminate other cores */
	if (my_em_exit_global(my_em_core_num) != 0) {
		/* eventually kill unfinished cores and wait cores termination */
		ti_em_rh_kill_cores(my_em_core_num);
		ti_em_rh_wait_cores(my_em_core_num);
	}

	/* remove test shared memory */
	ti_em_osal_shm_release(MY_SHARED_MEM_ID);

	/* release open EM*/
	ti_em_rh_release_local();
	ti_em_rh_release_global();

	/* exit */
	ti_em_osal_core_exit(0);

	return -1;
}
