/*
 * Copyright (c) 2012, Texas Instruments Incorporated - http://www.ti.com/
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions are met:
 *       * Redistributions of source code must retain the above copyright
 *         notice, this list of conditions and the following disclaimer.
 *       * Redistributions in binary form must reproduce the above copyright
 *         notice, this list of conditions and the following disclaimer in the
 *         documentation and/or other materials provided with the distribution.
 *       * Neither the name of Texas Instruments Incorporated nor the
 *         names of its contributors may be used to endorse or promote products
 *         derived from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *   DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
 *   DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


/**
 *   @file  ti_em_hw_cppi.h
 *
 *   @brief   
 *      This is the CPPI Descriptor Management include file.
 *
 */

#ifndef TI_EM_HW_CPPI_H_
#define TI_EM_HW_CPPI_H_

#ifdef __cplusplus
extern "C" {
#endif

#include <string.h> /* for memcpy() */

/** CPPI maximum number of CPDMAs */
#define CPPI_MAX_CPDMA                      13

/** Used as input parameter when queue number is 
 * not known and not specified */
#define CPPI_PARAM_NOT_SPECIFIED            -1

/** CPPI Low level Driver return and Error Codes */
/** CPPI successful return code */
#define CPPI_SOK                            0
/** CPPI Error Base */       
#define CPPI_LLD_EBASE                      (-128)
/** CPPI CPDMA not yet initialized */
#define CPPI_CPDMA_NOT_INITIALIZED          CPPI_LLD_EBASE-1
/** CPPI invalid parameter */
#define CPPI_INVALID_PARAM                  CPPI_LLD_EBASE-2
/** CPPI Rx/Tx channel not yet enabled */
#define CPPI_CHANNEL_NOT_OPEN               CPPI_LLD_EBASE-3
/** CPPI Rx flow not yet enabled */
#define CPPI_FLOW_NOT_OPEN                  CPPI_LLD_EBASE-4
/** CPPI Tx channels are still open. 
 * All Tx channels should be closed 
 * before calling CPPI_close */
#define CPPI_TX_CHANNELS_NOT_CLOSED         CPPI_LLD_EBASE-5
/** CPPI Rx channels are still open. 
 * All Rx channels should be closed 
 * before calling CPPI_close */
#define CPPI_RX_CHANNELS_NOT_CLOSED         CPPI_LLD_EBASE-6
/** CPPI Rx flows are still open. 
 * All Rx flows should be closed 
 * before calling CPPI_close */
#define CPPI_RX_FLOWS_NOT_CLOSED            CPPI_LLD_EBASE-7

/** Queue Manager subsystem memory region not enabled */
#define CPPI_QMSS_MEMREGION_NOT_INITIALIZED CPPI_LLD_EBASE-8
/** Queue open error */
#define CPPI_QUEUE_OPEN_ERROR               CPPI_LLD_EBASE-9
/** CPPI extended packet information block not present in descriptor */
#define CPPI_EPIB_NOT_PRESENT               CPPI_LLD_EBASE-10
/** CPPI protocol specific data not present in descriptor */
#define CPPI_PSDATA_NOT_PRESENT             CPPI_LLD_EBASE-11
/** CPPI CPDMA instances are still open. 
 * All CPDMA instances should be closed 
 * before calling CPPI_exit */
#define CPPI_CPDMA_NOT_CLOSED               CPPI_LLD_EBASE-12
/** CPPI resource initialization permission denied */
#define CPPI_RESOURCE_INIT_DENIED           CPPI_LLD_EBASE-13
/** CPPI resource usage permission denied */
#define CPPI_RESOURCE_USE_DENIED            CPPI_LLD_EBASE-14

/**
@}
*/

/** 
 * @brief CPPI CPDMA types
 */
typedef enum
{
    /** SRIO */
    Cppi_CpDma_SRIO_CPDMA = 0,
    /** AIF */
    Cppi_CpDma_AIF_CPDMA,
    /** FFTC A */
    Cppi_CpDma_FFTC_A_CPDMA,
    /** FFTC B */
    Cppi_CpDma_FFTC_B_CPDMA,
    /** FFTC C */
    Cppi_CpDma_FFTC_C_CPDMA,
    /** FFTC D */
    Cppi_CpDma_FFTC_D_CPDMA,
    /** FFTC E */
    Cppi_CpDma_FFTC_E_CPDMA,
    /** FFTC F */
    Cppi_CpDma_FFTC_F_CPDMA,
    /** PASS */
    Cppi_CpDma_PASS_CPDMA,
    /** QMSS in first QM */
    Cppi_CpDma_QMSS_CPDMA,
    /** QMSS in second QM */
    Cppi_CpDma_QMSS_QM2_CPDMA,
    /** BCP */
    Cppi_CpDma_BCP_CPDMA,
    /** XGE */
    Cppi_CpDma_XGE_CPDMA,
}Cppi_CpDma;

/** 
 * @brief CPPI Channel type 
 */
typedef enum
{
    /** Receive Channel */
    Cppi_ChType_RX_CHANNEL = 0,
    /** Transmit Channel */
    Cppi_ChType_TX_CHANNEL
}Cppi_ChType;

/** 
 * @brief CPPI Channel Enable
 */
typedef enum
{
    /** Disable Channel */
    Cppi_ChState_CHANNEL_DISABLE = 0,
    /** Enable Channel */
    Cppi_ChState_CHANNEL_ENABLE 
}Cppi_ChState;

/** 
 * @brief CPPI Wait after Channel Teardown
 */
typedef enum
{
    /** No wait */
    Cppi_Wait_NO_WAIT = 0,
    /** Wait */
    Cppi_Wait_WAIT
}Cppi_Wait;

/**
@}
*/

/** @addtogroup CPPI_LLD_DATASTRUCT
@{ 
*/

/** 
 * @brief CPPI CPDMA configuration structure
 */
typedef struct
{
    /** CPDMA configuring control registers */
    Cppi_CpDma                  dmaNum;

    /** This field sets the depth of the write arbitration FIFO which stores write transaction information
     * between the command arbiter and write data arbiters in the Bus Interface Unit. Setting this field to smaller 
     * values will cause prevent the CDMAHP from having an excess of write transactions outstanding whose data is 
     * still waiting to be transferred.
     * System performance can suffer if write commands are allowed to be issued long before the corresponding 
     * write data will be transferred.  This field allows the command count to be optimized based on system dynamics
     *
     * Valid range is 1 to 32. If writeFifoDepth field is set to 0, this field will not be configured. The reset/default value is 20.
     */
    uint8_t                     writeFifoDepth;
    /** This field sets the timeout duration in clock cycles.  This field controls the minimum 
     * amount of time that an Rx channel will be required to wait when it encounters a buffer starvation 
     * condition and the Rx error handling bit is set to 1 (packet is to be preserved - no discard).  
     * If the Rx error handling bit in the flow table is cleared, this field will have no effect on the Rx operation.  
     * When this field is set to 0, the Rx engine will not force an Rx channel to wait after encountering a starvation 
     * event (the feature is disabled).  When this field is set to a value other than 0, the Rx engine will force any 
     * channel whose associated flow had the Rx error handling bit asserted and which encounters starvation to wait for 
     * at least the specified # of clock cycles before coming into context again to retry the access to the QM
     */ 
    uint16_t                    timeoutCount;
    /** The QM N Queues Region Base Address Register is used to provide a programmable 
     * pointer to the base address of the queues region in Queue Manager N in the system 
     */

    /** Queue Manager 0 base address register - 0 means use default from Cppi_GlobalConfigParams */
    uint32_t            qm0BaseAddress;
    /** Queue Manager 1 base address register - 0 means use default from Cppi_GlobalConfigParams */
    uint32_t            qm1BaseAddress;
    /** Queue Manager 2 base address register - 0 means use default from Cppi_GlobalConfigParams */
    uint32_t            qm2BaseAddress;
    /** Queue Manager 3 base address register - 0 means use default from Cppi_GlobalConfigParams */
    uint32_t            qm3BaseAddress;
}Cppi_CpDmaInitCfg;

/** 
 * @brief CPPI transmit channel configuration structure
 */
typedef struct
{
    /** Channel number */
    /** If channelNum is set to CPPI_PARAM_NOT_SPECIFIED then the next 
     * available channel will be allocated */
    int32_t           channelNum;
    /** Enable Tx Channel on creation. If not set use CPPI_channelEnable() API to enable it later */
    Cppi_ChState    txEnable;
    /** Tx scheduling priority for channelNum */
    uint8_t           priority;
    /** Tx Filter Software Info.  This field controls whether or not the DMA controller will pass the 
     * extended packet information fields (if present) from the descriptor to the back end application.
     * 0 - DMA controller will pass extended packet info fields if they are present in the descriptor
     * 1 - DMA controller will filter extended packet info fields
     */
    uint16_t             filterEPIB;
    /** Filter Protocol Specific Words. This field controls whether or not the DMA controller will 
     * pass the protocol specific words (if present) from the descriptor to the back end application.
     * 0 - DMA controller will pass PS words if present in descriptor
     * 1 - DMA controller will filter PS words
     */
    uint16_t             filterPS;
    /**
     * AIF Specific Monolithic Packet Mode. This field when set indicates that all monolithic packets 
     * which will be transferred on this channel will be formatted in an optimal configuration as needed 
     * by the Antenna Interface Peripheral.  The AIF configuration uses a fixed descriptor format which 
     * includes the 3 mandatory descriptor info words, a single Protocol Specific Word and data 
     * immediately following (data offset = 16).
     */
    uint16_t             aifMonoMode;
}Cppi_TxChInitCfg;

/** 
 * @brief CPPI receive channel configuration structure
 */
typedef struct
{
    /** Channel number */
    /** If channelNum is set to CPPI_PARAM_NOT_SPECIFIED then the next 
     * available channel will be allocated */
    int32_t           channelNum;
    /** Enable Rx Channel on creation. If not set use CPPI_channelEnable() API to enable it later */
    Cppi_ChState    rxEnable;
}Cppi_RxChInitCfg;

/** 
 * @brief CPPI receive flow configuration structure
 */
typedef struct 
{
    /** Rx flow configuration register A */

    /** flow ID number */
    /** If flowIdNum is set to CPPI_PARAM_NOT_SPECIFIED then the next available flow ID will be allocated */
    int16_t           flowIdNum;
    /** This field indicates the default receive queue that this channel should use */
    uint16_t          rx_dest_qnum;
    /** This field indicates the default receive queue manager that this channel should use */
    uint16_t          rx_dest_qmgr;
    /** This field specifies the number of bytes that are to be skipped in the SOP buffer before beginning 
     * to write the payload or protocol specific bytes(if they are in the sop buffer).  This value must
     * be less than the minimum size of a buffer in the system */
    uint16_t          rx_sop_offset;
    /** This field controls where the Protocol Specific words will be placed in the Host Mode CPPI data structure 
     * 0 - protocol specific information is located in descriptor 
     * 1 - protocol specific information is located in SOP buffer */
    uint16_t             rx_ps_location;
    /** This field indicates the descriptor type to use 1 = Host, 2 = Monolithic */
    uint8_t           rx_desc_type;
    /** This field controls the error handling mode for the flow and is only used when channel errors occurs 
     * 0 = Starvation errors result in dropping packet and reclaiming any used descriptor or buffer resources 
     * back to the original queues/pools they were allocated to
     * 1 = Starvation errors result in subsequent re-try of the descriptor allocation operation.  
     */
    uint16_t             rx_error_handling;
    /** This field controls whether or not the Protocol Specific words will be present in the Rx Packet Descriptor 
     * 0 - The port DMA will set the PS word count to 0 in the PD and will drop any PS words that are presented 
     * from the back end application.
     * 1 - The port DMA will set the PS word count to the value given by the back end application and will copy 
     * the PS words from the back end application to the location 
     */
    uint16_t             rx_psinfo_present;
    /** This field controls whether or not the Extended Packet Info Block will be present in the Rx Packet Descriptor.  
     * 0 - The port DMA will clear the Extended Packet Info Block Present bit in the PD and will drop any extended 
     * packet info words that are presented from the back end application. 
     * 1 - The port DMA will set the Extended Packet Info Block Present bit in the PD and will copy any extended packet
     * info words that are presented across the Rx streaming interface into the extended packet info words in the descriptor.
     * If no extended packet info words are presented from the back end application, the port DMA will overwrite the fields with zeroes.
     */
    uint16_t             rx_einfo_present;

    /** Rx flow configuration register B */

    /** This is the value to insert into bits 7:0 of the destination tag if the rx_dest_tag_lo_sel is set to 1 */
    uint8_t           rx_dest_tag_lo;
    /** This is the value to insert into bits 15:8 of the destination tag if the rx_dest_tag_hi_sel is set to 1 */
    uint8_t           rx_dest_tag_hi;
    /** This is the value to insert into bits 7:0 of the source tag if the rx_src_tag_lo_sel is set to 1 */
    uint8_t           rx_src_tag_lo;
    /** This is the value to insert into bits 15:8 of the source tag if the rx_src_tag_hi_sel is set to 1 */
    uint8_t           rx_src_tag_hi;    

    /** Rx flow configuration register C */
    /** This bits control whether or not the flow will compare the packet size received from the back end application 
     * against the rx_size_thresh0 fields to determine which FDQ to allocate the SOP buffer from.  
     * The bits in this field is encoded as follows:
     * 0 = Do not use the threshold.
     * 1 = Use the thresholds to select SOP FDQ rx_fdq0_sz0_qnum/rx_fdq0_sz0_qmgr.
     */
    uint8_t             rx_size_thresh0_en;
    /** This bits control whether or not the flow will compare the packet size received from the back end application 
     * against the rx_size_thresh1 fields to determine which FDQ to allocate the SOP buffer from.  
     * The bits in this field is encoded as follows:
     * 0 = Do not use the threshold.
     * 1 = Use the thresholds to select SOP FDQ rx_fdq0_sz1_qnum/rx_fdq0_sz1_qmgr.
     */
    uint8_t             rx_size_thresh1_en;
        /** This bits control whether or not the flow will compare the packet size received from the back end application 
     * against the rx_size_thresh2 fields to determine which FDQ to allocate the SOP buffer from.  
     * The bits in this field is encoded as follows:
     * 0 = Do not use the threshold.
     * 1 = Use the thresholds to select SOP FDQ rx_fdq0_sz2_qnum/rx_fdq0_sz2_qmgr.
     */
    uint8_t             rx_size_thresh2_en;

    /** This field specifies the source for bits 7:0 of the source tag field in word 1 of the output PD.
     * This field is encoded as follows:
     * 0 = do not overwrite
     * 1 = overwrite with value given in rx_dest_tag_lo
     * 2 = overwrite with flow_id[7:0] from back end application
     * 3 = RESERVED
     * 4 = overwrite with dest_tag[7:0] from back end application
     * 5 = overwrite with dest_tag[15:8] from back end application
     * 6-7 = RESERVED
     */    
    uint8_t           rx_dest_tag_lo_sel;
    /** This field specifies the source for bits 15:8 of the source tag field in the word 1 of the output PD.
     * This field is encoded as follows:
     * 0 = do not overwrite
     * 1 = overwrite with value given in rx_dest_tag_hi
     * 2 = overwrite with flow_id[7:0] from back end application
     * 3 = RESERVED
     * 4 = overwrite with dest_tag[7:0] from back end application
     * 5 = overwrite with dest_tag[15:8] from back end application
     * 6-7 = RESERVED
     */
    uint8_t           rx_dest_tag_hi_sel;
    /** This field specifies the source for bits 7:0 of the source tag field in the output packet descriptor.
     * This field is encoded as follows:
     * 0 = do not overwrite
     * 1 = overwrite with value given in rx_src_tag_lo
     * 2 = overwrite with flow_id[7:0] from back end application
     * 3 = RESERVED
     * 4 = overwrite with src_tag[7:0] from back end application
     * 5 = RESERVED
     * 6-7 = RESERVED
     */
    uint8_t           rx_src_tag_lo_sel;
    /** This field specifies the source for bits 15:8 of the source tag field in the output packet descriptor.
     * This field is encoded as follows:
     * 0 = do not overwrite
     * 1 = overwrite with value given in rx_src_tag_hi
     * 2 = overwrite with flow_id[7:0] from back end application
     * 3 = RESERVED
     * 4 = overwrite with src_tag[7:0] from back end application
     * 5 = RESERVED
     * 6-7 = RESERVED
     */
    uint8_t           rx_src_tag_hi_sel;    
 
    /** Rx flow configuration register D */

    /** This field specifies which Free Descriptor Queue should be used for the 2nd Rx buffer in a host type packet */
    uint16_t          rx_fdq1_qnum;
    /** This field specifies which Queue Manager should be used for the 2nd Rx buffer in a host type packet */
    uint16_t          rx_fdq1_qmgr;
    /** This field specifies which Free Descriptor Queue should be used for the 1st Rx buffer in a packet whose 
     * size is less than or equal to the rx_size0 value */
    uint16_t          rx_fdq0_sz0_qnum;
    /** This field specifies which Queue Manager should be used for the 1st Rx buffer in a packet whose size 
     * is less than or equal to the rx_size0 value */
    uint16_t          rx_fdq0_sz0_qmgr;

    /** Rx flow configuration register E */

    /** This field specifies which Free Descriptor Queue should be used for the 4th or later Rx
     *  buffers in a host type packet */
    uint16_t          rx_fdq3_qnum;
    /** This field specifies which Queue Manager should be used for the 4th or later Rx buffers 
     * in a host type packet */
    uint16_t          rx_fdq3_qmgr;
    /** This field specifies which Free Descriptor Queue should be used for the 3rd Rx buffer in a host type packet */
    uint16_t          rx_fdq2_qnum;
    /** This field specifies which Queue Manager should be used for the 3rd Rx buffer in a host type packet */
    uint16_t          rx_fdq2_qmgr;

    /** Rx flow configuration register F */

    /** This value is left shifted by 5 bits and compared against the packet size to determine which free descriptor 
     * queue should be used for the SOP buffer in the packet.  If the  packet size is greater than the rx_size_thresh0 
     * but is less than or equal to the value given in this threshold, the DMA controller in the port will allocate the 
     * SOP buffer from the queue given by the rx_fdq0_sz1_qmgr and rx_fdq0_sz1_qnum fields. 
     * If enabled, this value must be greater than the value given in the rx_size_thresh0 field. This field is optional.
     */
    uint16_t          rx_size_thresh1;
    /** This value is left shifted by 5 bits and compared against the packet size to determine which free descriptor 
     * queue should be used for the SOP buffer in the packet.  If the packet size is less than or equal to the value 
     * given in this threshold, the DMA controller in the port will allocate the SOP buffer from the queue given by 
     * the rx_fdq0_sz0_qmgr and rx_fdq0_sz0_qnum fields. This field is optional.
     */
    uint16_t          rx_size_thresh0;
    
    /** Rx flow configuration register G */

    /** This field specifies which Queue should be used for the 1st Rx buffer in a packet whose size is 
     * less than or equal to the rx_size0 value */
    uint16_t          rx_fdq0_sz1_qnum;
    /** This field specifies which Queue Manager should be used for the 1st Rx buffer in a packet whose size 
     * is less than or equal to the rx_size0 value */
    uint16_t          rx_fdq0_sz1_qmgr;
    /** This value is left shifted by 5 bits and compared against the packet size to determine which free descriptor 
     * queue should be used for the SOP buffer in the packet.  If the  packet size is less than or equal to the value
     * given in this threshold, the DMA controller in the port will allocate the SOP buffer from the queue given by the 
     * rx_fdq0_sz2_qmgr and rx_fdq0_sz2_qnum fields.
     * If enabled, this value must be greater than the value given in the rx_size_thresh1 field. This field is optional.
     */
    uint16_t  		rx_size_thresh2;

    /** Rx flow configuration register H */

    /** This field specifies which Free Descriptor Queue should be used for the 1st Rx buffer in a
     * packet whose size is less than or equal to the rx_size3 value */
    uint16_t          rx_fdq0_sz3_qnum;
    /** This field specifies which Free Descriptor Queue Manager should be used for the 1st Rx buffer in a 
     * packet whose size is less than or equal to the rx_size3 value */
    uint16_t          rx_fdq0_sz3_qmgr;
    /** This field specifies which Free Descriptor Queue should be used for the 1st Rx buffer in a packet 
     * whose size is less than or equal to the rx_size2 value */
    uint16_t          rx_fdq0_sz2_qnum;
    /** This field specifies which Free Descriptor Queue Manager should be used for the 1st Rx buffer in a packet 
     * whose size is less than or equal to the rx_size2 value */
    uint16_t          rx_fdq0_sz2_qmgr;
}Cppi_RxFlowCfg;

/** 
 * @brief CPPI return result
 */
typedef int32_t   Cppi_Result;

/** 
 * @brief CPPI handle
 */
typedef uint32_t  *Cppi_Handle;

/** 
 * @brief CPPI channel handle
 */
typedef uint32_t  *Cppi_ChHnd;

/** 
 * @brief CPPI receive flow handle
 */
typedef uint32_t  *Cppi_FlowHnd;

/** Monolithic descriptor extended packet information block size */
#define CPPI_MONOLITHIC_DESC_EPIB_SIZE      20
/** Host descriptor extended packet information block size */
#define CPPI_HOST_DESC_EPIB_SIZE            16

/** 
 * @brief CPPI descriptor types
 */
typedef enum
{
    /** Host descriptor */
    Cppi_DescType_HOST = 0,
    /** Monolithic descriptor */
    Cppi_DescType_MONOLITHIC = 2
}Cppi_DescType;

/** 
 * @brief Packet return policy
 */
typedef enum
{
    /** Return entire packet */
    Cppi_ReturnPolicy_RETURN_ENTIRE_PACKET = 0,
    /** Return one buffer at a time */
    Cppi_ReturnPolicy_RETURN_BUFFER 
}Cppi_ReturnPolicy;

/** 
 * @brief protocol specific information location 
 */
typedef enum
{
    /** protocol specific information is located in descriptor */
    Cppi_PSLoc_PS_IN_DESC = 0,
    /** protocol specific information is located in SOP buffer */
    Cppi_PSLoc_PS_IN_SOP
}Cppi_PSLoc;

/** 
 * @brief extended packet information block
 */
typedef enum
{
    /** extended packet information block is not present in descriptor */
    Cppi_EPIB_NO_EPIB_PRESENT = 0,
    /** extended packet information block is present in descriptor  */
    Cppi_EPIB_EPIB_PRESENT
}Cppi_EPIB;

/** 
 * @brief Descriptor resource management
 */
typedef enum
{
    /** LLD initializes the descriptors with specified values */
    Cppi_InitDesc_INIT_DESCRIPTOR = 0,
    /** LLD does not initialize the descriptor with specified values */    
    Cppi_InitDesc_BYPASS_INIT 
}Cppi_InitDesc;

/** 
@} 
*/

/** @addtogroup CPPI_LLD_DATASTRUCT
@{ 
*/

/** 
 * @brief CPPI host descriptor configuration structure
 */
typedef struct 
{
    /** Indicates return policy for the packet. 
     * Valid only for host descriptor */
    Cppi_ReturnPolicy       returnPolicy;
    /** Indicates protocol specific location CPPI_PS_DESC - located in descriptor, CPPI_PS_SOP - located in SOP buffer 
     * Valid only for host descriptor */
    Cppi_PSLoc              psLocation;
}Cppi_HostDescCfg;

/** 
 * @brief CPPI monolithic descriptor configuration structure
 */
typedef struct 
{
    /** Byte offset from byte 0 of monolithic descriptor to the location where the valid data begins */
    uint32_t                  dataOffset;
}Cppi_MonolithicDescCfg;

/** 
 * @brief CPPI descriptor configuration structure
 */
typedef struct 
{
    /** Memory Region corresponding to the descriptor. */
    Qmss_MemRegion          memRegion;
    /** Group for memRegion (only applicable in when QMSS is in split mode) */
    uint32_t                queueGroup;
    /** Number of descriptors that should be configured with value below */
    uint32_t                  descNum;
    /** Queue where the descriptor is stored. If destQueueNum is set to QMSS_PARAM_NOT_SPECIFIED then the next 
     * available queue of type Qmss_QueueType will be allocated */
    int32_t                   destQueueNum;
    /** If destQueueNum is set to QMSS_PARAM_NOT_SPECIFIED then the next available queue of type 
     * Qmss_QueueType will be allocated */
    Qmss_QueueType          queueType;

    /** Descriptor configuration parameters */
    /** Indicates if the descriptor should be initialized with parameters listed below */
    Cppi_InitDesc           initDesc;

    /** Type of descriptor - Host or Monolithic */
    Cppi_DescType           descType;
    /** Indicates return Queue Manager and Queue Number. If both qMgr and qNum in returnQueue is 
     * set to QMSS_PARAM_NOT_SPECIFIED then the destQueueNum is configured in returnQueue of the descriptor */
    Qmss_Queue              returnQueue;
    /** Indicates how the CPDMA returns descriptors to free queue */
    Qmss_Location           returnPushPolicy;
    /** Indicates presence of EPIB */
    Cppi_EPIB               epibPresent;

    /** Union contains configuration that should be initialized in for host or monolithic descriptor. 
     * The configuration for host or monolithic descriptor is choosen based on "descType" field. 
     * The approriate structure fields must be specified if "initDesc" field is set to CPPI_INIT_DESCRIPTOR. */
    union{
    /** Host descriptor configuration parameters */
    Cppi_HostDescCfg        host;
    /** Monolithic  descriptor configuration parameters */
    Cppi_MonolithicDescCfg  mono;
    }cfg;

}Cppi_DescCfg;

/** 
 * @brief CPPI descriptor Word 1 Tag information
 */
typedef struct {
    uint8_t srcTagHi;
    uint8_t srcTagLo;
    uint8_t destTagHi;
    uint8_t destTagLo;
}Cppi_DescTag;

/** 
 * @brief CPPI host descriptor layout
 */
typedef struct {
    /** Descriptor type, packet type, protocol specific region location, packet length */
    uint32_t          descInfo;  
    /** Source tag, Destination tag */
    uint32_t          tagInfo;
    /** EPIB present, PS valid word count, error flags, PS flags, return policy, return push policy, 
     * packet return QM number, packet return queue number */
    uint32_t          packetInfo;
    /** Number of valid data bytes in the buffer */
    uint32_t          buffLen;
    /** Byte aligned memory address of the buffer associated with this descriptor */
    uint32_t          buffPtr;
    /** 32-bit word aligned memory address of the next buffer descriptor */
    uint32_t          nextBDPtr;       
    /** Completion tag, original buffer size */
    uint32_t          origBufferLen;
    /** Original buffer pointer */
    uint32_t          origBuffPtr;
    /** Optional EPIB word0 */
    uint32_t          timeStamp;
    /** Optional EPIB word1 */
    uint32_t          softwareInfo0;
    /** Optional EPIB word2 */
    uint32_t          softwareInfo1;
    /** Optional EPIB word3 */
    uint32_t          softwareInfo2;
    /** Optional protocol specific data */
    uint32_t          psData; 
}Cppi_HostDesc;

/** 
 * @brief CPPI monolithic descriptor layout
 */
typedef struct {
    /** Descriptor type, packet type, data offset, packet length */
    uint32_t          descInfo;  
    /** Source tag, Destination tag */
    uint32_t          tagInfo;
    /** EPIB present, PS valid word count, error flags, PS flags, return push policy, 
     * packet return QM number, packet return queue number */
    uint32_t          packetInfo;
    /** NULL word to align the extended packet words to a 128 bit boundary */
    uint32_t          Reserved;
    /** Optional EPIB word0 */
    uint32_t          timeStamp;
    /** Optional EPIB word1 */
    uint32_t          softwareInfo0;
    /** Optional EPIB word2 */
    uint32_t          softwareInfo1;
    /** Optional EPIB word3 */
    uint32_t          softwareInfo2;
    /** Optional protocol specific data */
    uint32_t          psData; 
}Cppi_MonolithicDesc;

/** 
 * @brief CPPI descriptor
 */
typedef union {
    /** Host descriptor */
    Cppi_HostDesc       *ptrHostDesc;  
    /** Monolithic descriptor */
    Cppi_MonolithicDesc *ptrMonoDesc;
}Cppi_Desc;

/**
 * @brief CPPI global configuration structure
 */
typedef struct
{
    /** CPDMA this configuration belongs to */
    Cppi_CpDma      dmaNum;
    /** Maximum supported Rx Channels */
    uint32_t          maxRxCh;
    /** Maximum supported Tx Channels */
    uint32_t          maxTxCh;
    /** Maximum supported Rx Flows */
    uint32_t          maxRxFlow;
    /** Priority for all Rx transactions of this CPDMA */
    uint8_t           rxPriority;
    /** Priority for all Tx transactions of this CPDMA */
    uint8_t           txPriority;

    /** Base address for the CPDMA overlay registers */

    /** Global Config registers */
    CSL_Cppidma_global_configRegs       *gblCfgRegs;
    /** Tx Channel Config registers */
    CSL_Cppidma_tx_channel_configRegs   *txChRegs;
    /** Rx Channel Config registers */
    CSL_Cppidma_rx_channel_configRegs   *rxChRegs;
    /** Tx Channel Scheduler registers */
    CSL_Cppidma_tx_scheduler_configRegs *txSchedRegs;
    /** Rx Flow Config registers */
    CSL_Cppidma_rx_flow_configRegs      *rxFlowRegs;
} Cppi_GlobalCPDMAConfigParams;
    
/**
 *  @b Description
 *  @n  
 *      This function is used to set the original buffer information in the host descriptor. 
 *
 *      **No validation is done on the input parameters**.
 *
 *  @param[in]  descType
 *      Type of descriptor - Cppi_DescType_HOST
 *
 *  @param[in]  descAddr
 *      Memory address of descriptor.
 *
 *  @param[in]  buffAddr
 *      Memory address of data buffer pointer.
 *      
 *  @param[in]  buffLen
 *      Size of the original data buffer.
 *
 *  @pre  
 *      Descriptor must be allocated and be valid.
 *
 *  @post  
 *      In case of host descriptor Word 6(bits 0 to 31) and Word 7 are updated
 *
 *  @retval
 *      None.
 */
static inline void ti_em_cppi_set_original_buf_info (Cppi_DescType descType, 
						     Cppi_Desc    *descAddr,
						     uint32_t      buffAddr, 
						     uint32_t      buffLen)
{
    Cppi_HostDesc *hostDescPtr = (Cppi_HostDesc *) descAddr;

    hostDescPtr->origBuffPtr = buffAddr;
    CSL_FINSR (hostDescPtr->origBufferLen, 21, 0, buffLen);

    return;
}

/**
 *  @b Description
 *  @n  
 *      This function is used to retrieve the original buffer information from host descriptor. 
 *
 *      **No validation is done on the input parameters**.
 *
 *  @param[in]  descType
 *      Type of descriptor - Cppi_DescType_HOST
 *
 *  @param[in]  descAddr
 *      Memory address of descriptor.
 *
 *  @param[out]  buffAddr
 *      Memory address of data buffer pointer.
 *      
 *  @param[out]  buffLen
 *      Size of the original data buffer.
 *
 *  @pre  
 *      Descriptor must be allocated and be valid.
 *
 *  @post  
 *      In case of host descriptor Word 6(bits 0 to 21) and Word 7 are returned
 *
 *  @retval
 *      None.
 */
static inline void ti_em_cppi_get_original_buf_info (Cppi_DescType descType, 
						     Cppi_Desc *descAddr,
						     uint8_t **buffAddr, 
						     uint32_t *buffLen)
{
    Cppi_HostDesc   *hostDescPtr = (Cppi_HostDesc *) descAddr;

    *buffAddr = (uint8_t *) hostDescPtr->origBuffPtr;
    *buffLen = CSL_FEXTR (hostDescPtr->origBufferLen, 21, 0);

    return;
}

/**
 *  @b Description
 *  @n  
 *      This function is used to set the packet type in host or monolithic descriptor.
 *
 *      **No validation is done on the input parameters**.
 *
 *  @param[in]  descType
 *      Type of descriptor - Cppi_DescType_HOST, Cppi_DescType_MONOLITHIC
 *
 *  @param[in]  descAddr
 *      Memory address of descriptor.
 *
 *  @param[in]  packetType
 *      Indicates type of packet. Valid range is 0 to 31.
 *      
 *  @pre  
 *      Descriptor must be allocated and be valid.
 *
 *  @post  
 *      Word 0 (bits 25 to 29) of host or monolithic descriptor are updated.
 *
 *  @retval
 *      None
 */
static inline void ti_em_cppi_set_packet_type (Cppi_DescType descType, 
					       Cppi_Desc *descAddr,
					       uint8_t packetType)
{
    Cppi_HostDesc   *descPtr = (Cppi_HostDesc *) descAddr;

    CSL_FINSR (descPtr->descInfo, 29, 25, packetType);
    return;
}

/**
 *  @b Description
 *  @n  
 *      This function is used to get the packet type from host or monolithic descriptor.
 *
 *      **No validation is done on the input parameters**.
 *
 *  @param[in]  descType
 *      Type of descriptor - Cppi_DescType_HOST, Cppi_DescType_MONOLITHIC
 *      
 *  @param[in]  descAddr
 *      Memory address of descriptor.
 *      
 *  @pre  
 *      Descriptor must be allocated and be valid.
 *
 *  @retval
 *      5 bit packet type field. Valid range is 0 to 31.
 */
static inline uint8_t ti_em_cppi_get_packet_type (Cppi_DescType descType, 
						  Cppi_Desc *descAddr)
{
    Cppi_HostDesc   *descPtr = (Cppi_HostDesc *) descAddr;

    return (CSL_FEXTR (descPtr->descInfo, 29, 25));
}

/**
 *  @b Description
 *  @n  
 *      This function copies the data buffer to the host or monolithic descriptor. 
 *      It is assumed that enough words are available to store the payload in case 
 *      of monolithic descriptor.
 *
 *      **No validation is done on the input parameters**.
 *      **Does not update packet length**.
 *
 *  @param[in]  descType
 *      Type of descriptor - Cppi_DescType_HOST, Cppi_DescType_MONOLITHIC
 *
 *  @param[in]  descAddr
 *      Memory address of descriptor.
 *
 *  @param[in]  buffAddr
 *      Memory address of data buffer.
 *      
 *  @param[in]  buffLen
 *      Size of the data buffer.
 *
 *  @pre  
 *      Descriptor and data buffer must be allocated and be valid.
 *
 *  @post  
 *      In case of host descriptor 
 *          Word 3 and Word 4 are updated
 *
 *      In case of monolithic descriptor 
 *          Word 0 (bits 0 to 15) and payload are updated.
 *
 *  @retval
 *      None.
 */
static inline void ti_em_cppi_set_data (Cppi_DescType descType,
					Cppi_Desc *descAddr,
					uint8_t *buffAddr,
					uint32_t buffLen)
{
    Cppi_HostDesc   *hostDescPtr;

    /* Does not update packet length */
    if (descType == Cppi_DescType_HOST)
    {
        hostDescPtr = (Cppi_HostDesc *) descAddr;
        hostDescPtr->buffPtr = (uint32_t) buffAddr;
        hostDescPtr->buffLen = (uint32_t) buffLen;
    }
    else
    {
        Cppi_MonolithicDesc *monolithicDescPtr = (Cppi_MonolithicDesc *) descAddr;
        uint16_t              dataOffset;

        dataOffset = CSL_FEXTR (monolithicDescPtr->descInfo, 24, 16);
        memcpy ((void *) (((uint8_t *) monolithicDescPtr) + dataOffset), buffAddr, buffLen);
        CSL_FINSR (monolithicDescPtr->descInfo, 15, 0, buffLen);
    }
    return;
}

/**
 *  @b Description
 *  @n  
 *      This function is update protocol specific data length in host or monolithic descriptor. 
 *
 *      **No validation is done on the input parameters**.
 *      **Does not update packet length**.
 *
 *
 *  @param[in]  descType
 *      Type of descriptor - Cppi_DescType_HOST, Cppi_DescType_MONOLITHIC
 *      
 *  @param[in]  descAddr
 *      Memory address of descriptor.
 *
 *  @param[in]  dataLen
 *      Size of PS data.
 *      
 *  @pre  
 *      Descriptor must be allocated and be valid.
 *
 *  @post  
 *      Word 2 (bits 24 to 29) are updated.
 *
 *  @retval
 *      None
 */
static inline void ti_em_cppi_set_pslen (Cppi_DescType descType, 
					 Cppi_Desc *descAddr,
					 uint32_t dataLen)
{
    Cppi_HostDesc   *descPtr = (Cppi_HostDesc *) descAddr;
 
    /* Does not update packet length */
    CSL_FINSR (descPtr->packetInfo, 29, 24, (dataLen / 4));
    return;
}

/**
 *  @b Description
 *  @n  
 *      This function is used to set source and destination tags in 
 *      host or monolithic descriptor. 
 *
 *      **No validation is done on the input parameters**.
 *
 *  @param[in]  descType
 *      Type of descriptor - Cppi_DescType_HOST, Cppi_DescType_MONOLITHIC
 *
 *  @param[in]  descAddr
 *      Memory address of descriptor.
 *
 *  @param[in]  tag
 *      Destination and source low and High tag value.
 *      
 *  @pre  
 *      Descriptor must be allocated and be valid.
 *
 *  @post  
 *      Word 1 of host or monolithic descriptor is updated.
 *
 *  @retval
 *      None
 */
static inline void ti_em_cppi_set_tag (Cppi_DescType descType, 
				       Cppi_Desc *descAddr,
				       Cppi_DescTag *tag)
{
    Cppi_HostDesc   *descPtr = (Cppi_HostDesc *) descAddr;

    CSL_FINSR (descPtr->tagInfo, 7, 0, tag->destTagLo);
    CSL_FINSR (descPtr->tagInfo, 15, 8, tag->destTagHi);
    CSL_FINSR (descPtr->tagInfo, 23, 16, tag->srcTagLo);
    CSL_FINSR (descPtr->tagInfo, 31, 24, tag->srcTagHi);
    return;
}

/**
 *  @b Description
 *  @n  
 *      This function configures a receive flow.
 */
static inline int ti_em_cppi_configure_rx_flow (
    Cppi_Handle hnd,
    Cppi_GlobalCPDMAConfigParams *cppiGblCfgParams,
    Cppi_RxFlowCfg *cfg)
{
    CSL_Cppidma_rx_flow_configRegs *rxFlowRegs;
    uint32_t index  = 0;
    int32_t  flowId = -1;
    uint32_t temp   = 0;
    uint32_t reg;

    if (cfg == NULL)
        return -1;

    flowId = cfg->flowIdNum;
    index = flowId;

    rxFlowRegs = cppiGblCfgParams[(uint32_t) hnd].rxFlowRegs;

    /* Rx flow configuration register A */
    reg = 0;
    CSL_FINS (reg, CPPIDMA_RX_FLOW_CONFIG_RX_FLOW_CONFIG_REG_A_RX_DEST_QNUM, cfg->rx_dest_qnum);
    CSL_FINS (reg, CPPIDMA_RX_FLOW_CONFIG_RX_FLOW_CONFIG_REG_A_RX_DEST_QMGR, cfg->rx_dest_qmgr);
    CSL_FINS (reg, CPPIDMA_RX_FLOW_CONFIG_RX_FLOW_CONFIG_REG_A_RX_SOP_OFFSET, cfg->rx_sop_offset);
    CSL_FINS (reg, CPPIDMA_RX_FLOW_CONFIG_RX_FLOW_CONFIG_REG_A_RX_PS_LOCATION, cfg->rx_ps_location);
    CSL_FINS (reg, CPPIDMA_RX_FLOW_CONFIG_RX_FLOW_CONFIG_REG_A_RX_DESC_TYPE, cfg->rx_desc_type);
    CSL_FINS (reg, CPPIDMA_RX_FLOW_CONFIG_RX_FLOW_CONFIG_REG_A_RX_ERROR_HANDLING, cfg->rx_error_handling);
    CSL_FINS (reg, CPPIDMA_RX_FLOW_CONFIG_RX_FLOW_CONFIG_REG_A_RX_PSINFO_PRESENT, cfg->rx_psinfo_present);
    CSL_FINS (reg, CPPIDMA_RX_FLOW_CONFIG_RX_FLOW_CONFIG_REG_A_RX_EINFO_PRESENT, cfg->rx_einfo_present);
    rxFlowRegs->RX_FLOW_CONFIG[index].RX_FLOW_CONFIG_REG_A = reg;

    /* Rx flow configuration register B */
    reg = 0;
    CSL_FINS (reg, CPPIDMA_RX_FLOW_CONFIG_RX_FLOW_CONFIG_REG_B_RX_DEST_TAG_LO, cfg->rx_dest_tag_lo);
    CSL_FINS (reg, CPPIDMA_RX_FLOW_CONFIG_RX_FLOW_CONFIG_REG_B_RX_DEST_TAG_HI, cfg->rx_dest_tag_hi);
    CSL_FINS (reg, CPPIDMA_RX_FLOW_CONFIG_RX_FLOW_CONFIG_REG_B_RX_SRC_TAG_LO, cfg->rx_src_tag_lo);
    CSL_FINS (reg, CPPIDMA_RX_FLOW_CONFIG_RX_FLOW_CONFIG_REG_B_RX_SRC_TAG_HI, cfg->rx_src_tag_hi);
    rxFlowRegs->RX_FLOW_CONFIG[index].RX_FLOW_CONFIG_REG_B = reg;

    /* Rx flow configuration register C */                
    temp = ((cfg->rx_size_thresh0_en) | (cfg->rx_size_thresh1_en << 1) | (cfg->rx_size_thresh2_en << 2));
    reg = 0;    
    CSL_FINS (reg, CPPIDMA_RX_FLOW_CONFIG_RX_FLOW_CONFIG_REG_C_RX_SIZE_THRESH_EN, temp);
    CSL_FINS (reg, CPPIDMA_RX_FLOW_CONFIG_RX_FLOW_CONFIG_REG_C_RX_DEST_TAG_LO_SEL, cfg->rx_dest_tag_lo_sel);
    CSL_FINS (reg, CPPIDMA_RX_FLOW_CONFIG_RX_FLOW_CONFIG_REG_C_RX_DEST_TAG_HI_SEL, cfg->rx_dest_tag_hi_sel);
    CSL_FINS (reg, CPPIDMA_RX_FLOW_CONFIG_RX_FLOW_CONFIG_REG_C_RX_SRC_TAG_LO_SEL, cfg->rx_src_tag_lo_sel);
    CSL_FINS (reg, CPPIDMA_RX_FLOW_CONFIG_RX_FLOW_CONFIG_REG_C_RX_SRC_TAG_HI_SEL, cfg->rx_src_tag_hi_sel);
    rxFlowRegs->RX_FLOW_CONFIG[index].RX_FLOW_CONFIG_REG_C = reg;

    /* Rx flow configuration register D */
    reg = 0;
    CSL_FINS (reg, CPPIDMA_RX_FLOW_CONFIG_RX_FLOW_CONFIG_REG_D_RX_FDQ1_QNUM, cfg->rx_fdq1_qnum);
    CSL_FINS (reg, CPPIDMA_RX_FLOW_CONFIG_RX_FLOW_CONFIG_REG_D_RX_FDQ1_QMGR, cfg->rx_fdq1_qmgr);
    CSL_FINS (reg, CPPIDMA_RX_FLOW_CONFIG_RX_FLOW_CONFIG_REG_D_RX_FDQ0_SZ0_QNUM, cfg->rx_fdq0_sz0_qnum);
    CSL_FINS (reg, CPPIDMA_RX_FLOW_CONFIG_RX_FLOW_CONFIG_REG_D_RX_FDQ0_SZ0_QMGR, cfg->rx_fdq0_sz0_qmgr);
    rxFlowRegs->RX_FLOW_CONFIG[index].RX_FLOW_CONFIG_REG_D = reg;

    /* Rx flow configuration register E */
    reg = 0;
    CSL_FINS (reg, CPPIDMA_RX_FLOW_CONFIG_RX_FLOW_CONFIG_REG_E_RX_FDQ3_QNUM, cfg->rx_fdq3_qnum);
    CSL_FINS (reg, CPPIDMA_RX_FLOW_CONFIG_RX_FLOW_CONFIG_REG_E_RX_FDQ3_QMGR, cfg->rx_fdq3_qmgr);
    CSL_FINS (reg, CPPIDMA_RX_FLOW_CONFIG_RX_FLOW_CONFIG_REG_E_RX_FDQ2_QNUM, cfg->rx_fdq2_qnum);
    CSL_FINS (reg, CPPIDMA_RX_FLOW_CONFIG_RX_FLOW_CONFIG_REG_E_RX_FDQ2_QMGR, cfg->rx_fdq2_qmgr);
    rxFlowRegs->RX_FLOW_CONFIG[index].RX_FLOW_CONFIG_REG_E = reg;

    /* Rx flow configuration register F */
    reg = 0;
    CSL_FINS (reg, CPPIDMA_RX_FLOW_CONFIG_RX_FLOW_CONFIG_REG_F_RX_SIZE_THRESH1, (cfg->rx_size_thresh1 >> 5));
    CSL_FINS (reg, CPPIDMA_RX_FLOW_CONFIG_RX_FLOW_CONFIG_REG_F_RX_SIZE_THRESH0, (cfg->rx_size_thresh0  >> 5));
    rxFlowRegs->RX_FLOW_CONFIG[index].RX_FLOW_CONFIG_REG_F = reg;

    /* Rx flow configuration register G */
    reg = 0;
    CSL_FINS (reg, CPPIDMA_RX_FLOW_CONFIG_RX_FLOW_CONFIG_REG_G_RX_FDQ0_SZ1_QNUM, cfg->rx_fdq0_sz1_qnum);
    CSL_FINS (reg, CPPIDMA_RX_FLOW_CONFIG_RX_FLOW_CONFIG_REG_G_RX_FDQ0_SZ1_QMGR, cfg->rx_fdq0_sz1_qmgr);
    CSL_FINS (reg, CPPIDMA_RX_FLOW_CONFIG_RX_FLOW_CONFIG_REG_G_RX_SIZE_THRESH2, (cfg->rx_size_thresh2) >> 5);
    rxFlowRegs->RX_FLOW_CONFIG[index].RX_FLOW_CONFIG_REG_G = reg;

    /* Rx flow configuration register H */
    reg = 0;
    CSL_FINS (reg, CPPIDMA_RX_FLOW_CONFIG_RX_FLOW_CONFIG_REG_H_RX_FDQ0_SZ3_QNUM, cfg->rx_fdq0_sz3_qnum);
    CSL_FINS (reg, CPPIDMA_RX_FLOW_CONFIG_RX_FLOW_CONFIG_REG_H_RX_FDQ0_SZ3_QMGR, cfg->rx_fdq0_sz3_qmgr);
    CSL_FINS (reg, CPPIDMA_RX_FLOW_CONFIG_RX_FLOW_CONFIG_REG_H_RX_FDQ0_SZ2_QNUM, cfg->rx_fdq0_sz2_qnum);
    CSL_FINS (reg, CPPIDMA_RX_FLOW_CONFIG_RX_FLOW_CONFIG_REG_H_RX_FDQ0_SZ2_QMGR, cfg->rx_fdq0_sz2_qmgr);
    rxFlowRegs->RX_FLOW_CONFIG[index].RX_FLOW_CONFIG_REG_H = reg;

    return 0;
}
    
/**
 *  @b Description
 *  @n  
 *      This function opens a CPPI transmit channel. 
 */
static inline int ti_em_cppi_tx_channel_open (
    Cppi_Handle hnd,
    Cppi_GlobalCPDMAConfigParams *cppiGblCfgParams,
    Cppi_TxChInitCfg *cfg)
{
    CSL_Cppidma_tx_scheduler_configRegs *txSchedRegs;
    CSL_Cppidma_tx_channel_configRegs   *txChRegs;
    int32_t chNum  = -1;

    if (cfg == NULL)
        return -1;

    chNum       = cfg->channelNum;
    txSchedRegs = cppiGblCfgParams[(uint32_t) hnd].txSchedRegs;
    txChRegs    = cppiGblCfgParams[(uint32_t) hnd].txChRegs;

    CSL_FINS (txSchedRegs->TX_CHANNEL_SCHEDULER_CONFIG_REG[chNum],
	      CPPIDMA_TX_SCHEDULER_CONFIG_TX_CHANNEL_SCHEDULER_CONFIG_REG_PRIORITY, cfg->priority);

    CSL_FINS (txChRegs->TX_CHANNEL_GLOBAL_CONFIG[chNum].TX_CHANNEL_GLOBAL_CONFIG_REG_B,
	      CPPIDMA_TX_CHANNEL_CONFIG_TX_CHANNEL_GLOBAL_CONFIG_REG_B_TX_FILT_EINFO, cfg->filterEPIB);
    
    CSL_FINS (txChRegs->TX_CHANNEL_GLOBAL_CONFIG[chNum].TX_CHANNEL_GLOBAL_CONFIG_REG_B, 
	      CPPIDMA_TX_CHANNEL_CONFIG_TX_CHANNEL_GLOBAL_CONFIG_REG_B_TX_FILT_PSWORDS, cfg->filterPS);
    
    CSL_FINS (txChRegs->TX_CHANNEL_GLOBAL_CONFIG[chNum].TX_CHANNEL_GLOBAL_CONFIG_REG_B, 
	      CPPIDMA_TX_CHANNEL_CONFIG_TX_CHANNEL_GLOBAL_CONFIG_REG_B_TX_AIF_MONO_MODE, cfg->aifMonoMode);
    
    if (cfg->txEnable)
    {
	uint32_t value = 0;
	CSL_FINS (value, CPPIDMA_TX_CHANNEL_CONFIG_TX_CHANNEL_GLOBAL_CONFIG_REG_A_TX_ENABLE, cfg->txEnable);
	txChRegs->TX_CHANNEL_GLOBAL_CONFIG[chNum].TX_CHANNEL_GLOBAL_CONFIG_REG_A = value;
    }
    return 0;
}

/**
 *  @b Description
 *  @n  
 *      This function opens a CPPI transmit channel. 
 */
static inline int ti_em_cppi_rx_channel_open (
    Cppi_Handle hnd,
    Cppi_GlobalCPDMAConfigParams *cppiGblCfgParams,
    Cppi_RxChInitCfg *cfg)
{
    CSL_Cppidma_rx_channel_configRegs *rxChRegs;
    int32_t chNum = -1;

    if (cfg == NULL)
        return -1;

    chNum    = cfg->channelNum;
    rxChRegs = cppiGblCfgParams[(uint32_t) hnd].rxChRegs;

    if (cfg->rxEnable)
    {
	uint32_t value = 0;
	CSL_FINS (value, CPPIDMA_RX_CHANNEL_CONFIG_RX_CHANNEL_GLOBAL_CONFIG_REG_RX_ENABLE, cfg->rxEnable);
	rxChRegs->RX_CHANNEL_GLOBAL_CONFIG[chNum].RX_CHANNEL_GLOBAL_CONFIG_REG = value;
    }
    return 0;
}

/**
 * @brief Map PDK HAL to Linux OSAL
 */
#define ti_em_rx_flow_open          ti_em_osal_cppi_rx_flow_open
#define ti_em_rx_channel_open       ti_em_osal_cppi_rx_channel_open
#define ti_em_tx_channel_open       ti_em_osal_cppi_tx_channel_open
#define ti_em_rx_flow_close         ti_em_osal_cppi_rx_flow_close
#define ti_em_rx_channel_close      ti_em_osal_cppi_rx_channel_close
#define ti_em_tx_channel_close      ti_em_osal_cppi_tx_channel_close

#define ti_em_chain_rx_flow_open    ti_em_osal_chain_rx_flow_open

#define ti_em_xge_rx_flow_open      ti_em_osal_xge_rx_flow_open
#define ti_em_xge_tx_channel_open   ti_em_osal_xge_tx_channel_open
#define ti_em_xge_rx_channel_open   ti_em_osal_xge_rx_channel_open
#define ti_em_xge_tx_queue_open     ti_em_osal_xge_tx_queue_open
#define ti_em_xge_rx_flow_close     ti_em_osal_xge_rx_flow_close
#define ti_em_xge_rx_channel_close  ti_em_osal_xge_rx_channel_close
#define ti_em_xge_tx_channel_close  ti_em_osal_xge_tx_channel_close
#define ti_em_xge_tx_queue_base_idx_get ti_em_osal_xge_tx_queue_base_idx_get

#define ti_em_rio_rx_flow_open      ti_em_osal_rio_rx_flow_open
#define ti_em_rio_tx_channel_open   ti_em_osal_rio_tx_channel_open
#define ti_em_rio_tx_queue_open     ti_em_osal_rio_tx_queue_open

#define ti_em_get_absolute_queue_id ti_em_osal_cppi_get_absolute_queue_id

#ifdef __cplusplus
}
#endif

#endif /*TI_EM_HW_CPPI_H_*/
