/*
 * File name: em_mod.c
 *
 * Description: OpenEM utility module to allow device user access and CMA
 *
 * Copyright (C) 2013 Texas Instruments, Incorporated
 * Author: Aurelien Jacquiot <a-jacquiot@ti.com>
 * 
 * Some code borrowed from Linux virtio_udma driver
 * Copyright (C) 2012, 2013 Texas Instruments Incorporated
 * Author: Cyril Chemparathy <cyril@ti.com>
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation version 2.
 *
 * This program is distributed "as is" WITHOUT ANY WARRANTY of any kind,
 * whether express or implied; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */
#include <linux/init.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/dma-mapping.h>
#include <linux/version.h>
#include <linux/wait.h>
#include <linux/sched.h>
#include <linux/uaccess.h>
#include <linux/spinlock.h>
#include <linux/io.h>
#include <linux/hrtimer.h>
#include <linux/miscdevice.h>
#include <linux/slab.h>
#include <linux/fs.h>
#include <linux/of.h>
#include <linux/ioctl.h>
#include <linux/module.h>

#include <asm/cacheflush.h>

/* Number of total Linux OpenEM instances to be managed */
#define TI_EM_INSTANCE_NUM    4

struct ti_em_intd_irq_config {
	int channel_base;
	int intd;
};

#define EM_IOC_MAGIC	      'O'
#define EM_IOC_ALLOC	      _IOWR(EM_IOC_MAGIC, 0, int)
#define EM_IOC_ENABLEPMU      _IO(EM_IOC_MAGIC, 1)
#define EM_IOC_WAITEVENT      _IO(EM_IOC_MAGIC, 2)
#define EM_IOC_WAITEVENTINIT  _IOWR(EM_IOC_MAGIC, 3, struct ti_em_intd_irq_config)
#define EM_IOC_WAITEVENTFREE  _IO(EM_IOC_MAGIC, 4)

#define EM_MOD_DEVNAME        "openem"

#undef EM_MOD_DEBUG
#ifdef EM_MOD_DEBUG
#define DPRINTK(fmt, args...) printk(KERN_DEBUG "OpenEM module: [%s] " fmt, __FUNCTION__, ## args)
#else
#define DPRINTK(fmt, args...)
#endif

#define EPRINTK(fmt, args...) printk(KERN_CRIT "OpenEM module: [%s] ERROR " fmt, __FUNCTION__, ## args)

#ifndef CONFIG_CMA /* for old kernels */
#define dma_mmap_coherent(dev, vma, cpu_addr, dma_addr, size)	\
	remap_pfn_range(vma, (vma)->vm_start, (vma)->vm_pgoff,	\
			size, (vma)->vm_page_prot)
#define kstrtoul(c, b, v) (*(v)) = simple_strtoul((c), NULL, (b))
#else
#include <linux/dma-contiguous.h>
#endif

#ifdef CONFIG_ARM
#ifdef DEVICE_K2K

#define TI_EM_USE_INTERRUPT

/* MPAX for cache-coherency */
struct mpax {
	u32	mpaxl;
	u32	mpaxh;
};

struct msms_regs {
	u32	pid;
	u32	_res_04;
	u32	smcerrar;
	u32	smcerrxr;
	u32	smedcc;
	u32	smcea;
	u32	smsecc;
	u32	smpfar;
	u32	smpfxr;
	u32	smpfr;
	u32	smpfcr;
	u32	_res_2c;
	u32	sbndc[8];
	u32	sbndm;
	u32	sbnde;
	u32	_res_58;
	u32	cfglck;
	u32	cfgulck;
	u32	cfglckstat;
	u32	sms_mpax_lck;
	u32	sms_mpax_ulck;
	u32	sms_mpax_lckstat;
	u32	ses_mpax_lck;
	u32	ses_mpax_ulck;
	u32	ses_mpax_lckstat;
	u32	smestat;
	u32	smirstat;
	u32	smirc;
	u32	smiestat;
	u32	smiec;
	u32	_res_94_c0[12];
	u32	smncerrar;
	u32	smncerrxr;
	u32	smncea;
	u32	_res_d0_1fc[76];
	struct mpax sms[16][8];
	struct mpax ses[16][8];
};

void msmc_share_segment(int id)
{
	struct msms_regs* msmc = (struct msms_regs*) ioremap(0x0bc00000, 0x100000);
	int j;

	DPRINTK("id %d sharing segment 0x%08x; sms lock 0x%08x; ses lock 0x%08x\n",
		id, msmc->pid, msmc->sms_mpax_lckstat, msmc->ses_mpax_lckstat);
	
	for (j = 0; j < 8; j++) {
		DPRINTK("[%d]SMS MPAXL 0x%08x; MPAXH 0x%08X\t", j,
			msmc->sms[id][j].mpaxl,
			msmc->sms[id][j].mpaxh);
		DPRINTK("[%d]SES MPAXL 0x%08x; MPAXH 0x%08X\n", j,
			msmc->ses[id][j].mpaxl,
			msmc->ses[id][j].mpaxh);
		
		msmc->sms[id][j].mpaxh &= 0xffffff7ful;
		msmc->ses[id][j].mpaxh &= 0xffffff7ful;
		msmc->sms[id][j].mpaxl |= 0x3ful;
		msmc->ses[id][j].mpaxl |= 0x3ful;
	}
	
	DPRINTK("after MSMC setup\n");

	for (j = 0; j< 8; j++) {
		DPRINTK("[%d]SMS MPAXL 0x%08x; MPAXH 0x%08X\t", j,
			msmc->sms[id][j].mpaxl,
			msmc->sms[id][j].mpaxh);
		DPRINTK("[%d]SES MPAXL 0x%08x; MPAXH 0x%08X\n", j,
			msmc->ses[id][j].mpaxl,
			msmc->ses[id][j].mpaxh);
	}
}
#endif /* DEVICE_K2K */

#define MPU0_BASE_ADDRESS     0x02360000
#define MPU1_BASE_ADDRESS     0x02368000
#define MPU2_BASE_ADDRESS     0x02370000
#define MPU3_BASE_ADDRESS     0x02378000
#define MPU4_BASE_ADDRESS     0x02380000
#define MPU5_BASE_ADDRESS     0x02388000
#define MPU6_BASE_ADDRESS     0x02390000
#define MPU7_BASE_ADDRESS     0x02398000

#ifdef DEVICE_K2K
#define MPU0_RANGE_NUM        16 /* Main CFG SCR/TeraNEt  */
#define MPU1_RANGE_NUM        16 /* QMSS data port */
#define MPU2_RANGE_NUM        16 /* QMSS cfg port 1 */
#define MPU5_RANGE_NUM        16 /* QMSS cfg port 2 */
#define MPU7_RANGE_NUM        16 /* DDR3 */
#else
#define MPU0_RANGE_NUM        16 /* Main CFG SCR/TeraNEt  */
#define MPU1_RANGE_NUM        5  /* QMSS data port */
#define MPU2_RANGE_NUM        16 /* QMSS cfg port */
#define MPU6_RANGE_NUM        16 /* DDR3 */
#endif

#define MPU_PROG_MPSAR(n)     (0x200 + ((n) << 4))
#define MPU_PROG_MPEAR(n)     (0x204 + ((n) << 4))
#define MPU_PROG_MPPA(n)      (0x208 + ((n) << 4))

static void mpu_enable_user_mode_access(void)
{
	unsigned int  reg_addr, i, count, def_reg_val;
        void __iomem *temp_reg;
	
	def_reg_val = 0x03FFFFFF;
	reg_addr    = MPU1_BASE_ADDRESS + MPU_PROG_MPPA(0);
	count       = MPU1_RANGE_NUM;
	for (i = 1; i <= count; i++) {
		temp_reg = ioremap(reg_addr, 4);
		__raw_writel(def_reg_val, temp_reg) ;
		iounmap(temp_reg);
		reg_addr = reg_addr + 0x10;
    	}
	
	reg_addr    = MPU2_BASE_ADDRESS + MPU_PROG_MPPA(0);
	count       = MPU2_RANGE_NUM;
	for (i = 1; i <= count; i++)
	{
		temp_reg = ioremap(reg_addr, 4);
		__raw_writel(def_reg_val, temp_reg) ;
		iounmap(temp_reg);
		reg_addr = reg_addr + 0x10;
	}

#ifdef DEVICE_K2K
	reg_addr    = MPU5_BASE_ADDRESS + MPU_PROG_MPPA(0);
	count       = MPU5_RANGE_NUM;
	for (i = 1; i <= count; i++)
	{
		temp_reg = ioremap(reg_addr, 4);
		__raw_writel(def_reg_val, temp_reg) ;
		iounmap(temp_reg);
		reg_addr = reg_addr + 0x10;
	}

	reg_addr    = MPU7_BASE_ADDRESS + MPU_PROG_MPPA(0);
	count       = MPU7_RANGE_NUM;
	for (i = 1; i <= count; i++)
	{
		temp_reg = ioremap(reg_addr, 4);
		__raw_writel(def_reg_val, temp_reg) ;
		iounmap(temp_reg);
		reg_addr = reg_addr + 0x10;
	}

#else /* DEVICE_K2K */

	reg_addr    = MPU6_BASE_ADDRESS + MPU_PROG_MPPA(0);
	count       = MPU6_RANGE_NUM;

	for (i = 1; i <= count; i++)
	{
		temp_reg = ioremap(reg_addr, 4);
		__raw_writel(def_reg_val, temp_reg) ;
		iounmap(temp_reg);
		reg_addr = reg_addr + 0x10;
	}
#endif /* DEVICE_K2K */
}

static void pmu_enable_user_mode_access(void)
{
	unsigned long val;

	/* single clock, reset, enable (pmcr) */
	val = 0x4 | 0x1;
	asm volatile("mcr p15, 0, %0, c9, c12, 0" : : "r"(val));
	
	/* enable user space access (userenr) */
	val = 1;
	asm volatile("mcr p15, 0, %0, c9, c14, 0" : : "r"(val));
	
	/* start it */
	val = 0x80000000;
	asm volatile("mcr p15, 0, %0, c9, c12, 1" ::  "r"(val));
}
#endif /* CONFIG_ARM */

#ifdef DEVICE_K2K
static void pktdma_qm_init(void)
{
	void __iomem *reg;
	
	reg = ioremap(0x02a08000, 0x100);
	__raw_writel(0x23a80000, (void*)((unsigned int) reg + 0x10));
	__raw_writel(0x23a90000, (void*)((unsigned int) reg + 0x14));
	iounmap(reg);

	reg = ioremap(0x02a0a000, 0x100);
	__raw_writel(0x23aa0000, (void*)((unsigned int) reg + 0x10));
	__raw_writel(0x23ab0000, (void*)((unsigned int) reg + 0x14));
	iounmap(reg);
}
#else /* DEVICE_K2K */
static void pktdma_qm_init(void)
{
	void __iomem *reg;

	reg = ioremap(0x02a6c000, 0x100);
	__raw_writel(0x34020000, (void*)((unsigned int) reg + 0x10));
	__raw_writel(0x34030000, (void*)((unsigned int) reg + 0x14));
	iounmap(reg);
}
#endif /* DEVICE_K2K */

/*
 * User-space contiguous memory allocator based on CMA
 */
struct em_device {
	struct miscdevice  misc;
	struct list_head   users;
	struct kref        refcount;
	wait_queue_head_t  wqh;
	int                in_use;
	atomic_t           event;
	int                intd;
	int                channel_base;
	unsigned long      irq_status;
	char               name[32];
};

struct em_user {
	int                instance;
	struct em_device  *em;
	struct list_head   node;
	struct list_head   chunks;
	struct file       *file;
	int                event_count;
};

struct em_chunk {
	struct em_user    *user;
	struct list_head   node;
	size_t             size;
	dma_addr_t         paddr;
	void              *kvaddr;
};

#define from_misc(misc)	container_of(misc, struct em_device, misc)
#define to_misc(em)	(&(em)->misc)

struct em_device *em_dev[TI_EM_INSTANCE_NUM];

static DEFINE_SPINLOCK(em_lock);

static void em_device_release(struct kref *kref)
{
	struct em_device *em;

	em = container_of(kref, struct em_device, refcount);

	WARN_ON(!list_empty(&em->users));

	DPRINTK("release em device instance 0x%x\n", (unsigned int) em);

	kfree(em);
}

static inline struct em_device *em_device_get(struct em_device *em)
{
	kref_get(&em->refcount);
	return em;
}

static inline void em_device_put(struct em_device *em)
{
	kref_put(&em->refcount, em_device_release);
}

static void em_add_user(struct em_device *em, struct em_user *user)
{
	spin_lock(&em_lock);
	list_add_tail(&user->node, &em->users);
	spin_unlock(&em_lock);
	em_device_get(em);
}

static void em_del_user(struct em_device *em, struct em_user *user)
{
	spin_lock(&em_lock);
	list_del(&user->node);
	spin_unlock(&em_lock);
	em_device_put(em);
}

static void em_user_add_chunk(struct em_user *user, struct em_chunk *chunk)
{
	spin_lock(&em_lock);
	list_add_tail(&chunk->node, &user->chunks);
	spin_unlock(&em_lock);
}

static void em_user_del_chunk(struct em_user *user, struct em_chunk *chunk)
{
	spin_lock(&em_lock);
	list_del(&chunk->node);
	spin_unlock(&em_lock);
}

static struct em_chunk *em_user_first_chunk(struct em_user *user)
{
	struct em_chunk *chunk = NULL;

	spin_lock(&em_lock);
	if (!list_empty(&user->chunks))
		chunk = list_first_entry(&user->chunks, struct em_chunk, node);
	spin_unlock(&em_lock);
	return chunk;
}

static int em_chunk_alloc(struct em_user *user,
			  size_t          size,
			  dma_addr_t     *paddr)
{
	struct em_chunk *chunk;

	chunk = kzalloc(sizeof(*chunk), GFP_KERNEL);
	if (!chunk) {
		EPRINTK("failed to allocate chunk info\n");
		return -ENOMEM;
	}

	chunk->user   = user;
	chunk->size   = size;
	chunk->kvaddr = dma_alloc_coherent(NULL, 
					   chunk->size, 
					   &chunk->paddr, 
					   GFP_KERNEL | GFP_DMA);
	if (!chunk->kvaddr) {
		EPRINTK("failed to allocate coherent memory\n");
		kfree(chunk);
		return -ENOMEM;
	}
	memset(chunk->kvaddr, 0, chunk->size);

	if (paddr)
		*paddr = chunk->paddr;

	em_user_add_chunk(user, chunk);

	return 0;
}

static void em_chunk_release(struct em_chunk *chunk)
{
	struct em_user *user = chunk->user;

	DPRINTK("releasing chunk kern 0x%lx, phys 0x%lx, size 0x%lx\n",
		(unsigned long) chunk->kvaddr,
		(unsigned long) chunk->paddr, 
		(unsigned long) chunk->size);

	dma_free_coherent(NULL, chunk->size, chunk->kvaddr, chunk->paddr);
	em_user_del_chunk(user, chunk);

	kfree(chunk);
}

static inline struct em_chunk *em_find_chunk(struct em_user *user, 
					     dma_addr_t      paddr)
{
	struct em_chunk  *chunk;

	list_for_each_entry(chunk, &user->chunks, node) {
		if (chunk->paddr == paddr)
			return chunk;
	}
	return NULL;
}

static int em_dev_fop_mmap(struct file *file, struct vm_area_struct *vma)
{
	dma_addr_t       paddr;
	unsigned int     pfn;
	size_t           size;
	int              ret;

	size = vma->vm_end - vma->vm_start;

	DPRINTK("vma->vm_start     = %#lx\n", vma->vm_start);
	DPRINTK("vma->vm_end       = %#lx\n", vma->vm_end);
	DPRINTK("vma->vm_pgoff     = %#lx\n", vma->vm_pgoff);
	DPRINTK("size              = 0x%x\n", size);

#ifdef CONFIG_ARM
	if (file->f_flags & O_DSYNC) {
                vma->vm_flags |= VM_IO;
		DPRINTK("D_SYNC -> (noncached/coherent memory)\n");
		vma->vm_page_prot = pgprot_noncached(vma->vm_page_prot);
	} else {
		DPRINTK("D_SYNC -> (cached/dmacoherent memory)\n");
	}
#endif

	paddr = (dma_addr_t) vma->vm_pgoff << PAGE_SHIFT;

#ifdef DEVICE_K2K
	if (paddr >= KEYSTONE_LOW_PHYS_START) {
		pfn = vma->vm_pgoff + PFN_DOWN(KEYSTONE_HIGH_PHYS_START - KEYSTONE_LOW_PHYS_START);
	} else
#endif
		pfn = vma->vm_pgoff;
	
	ret = remap_pfn_range(vma, 
			      vma->vm_start,
			      pfn,
			      size,
			      vma->vm_page_prot);

	if (ret) {
		EPRINTK("failed to user map dma memory\n");
		return -ENOMEM;
	}

	return 0;
}

#ifndef CONFIG_MMU
static unsigned long em_dev_fop_get_unmapped_area(struct file  *file,
						  unsigned long addr,
						  unsigned long len,
						  unsigned long pgoff,
						  unsigned long flags)
{
	return pgoff << PAGE_SHIFT;
}
#endif

#ifdef TI_EM_USE_INTERRUPT
/*
 * Manage incoming OpenEM interrupt-based event
 */
static int em_irq[TI_EM_INSTANCE_NUM];

static DEFINE_SPINLOCK(em_irq_lock);

#if !defined(CONFIG_OF_IRQ) || !defined(CONFIG_IRQ_DOMAIN) || !defined(CONFIG_ARM_GIC)
#error "only GIC based interrupt are supported"
#endif
#include <linux/interrupt.h>
#include <linux/irqdomain.h>
#include <linux/of_irq.h>

/* base of ARM hwirq for the various INTD */
#define QMSS_INTD1_HWIRQ_BASE            179
#define QMSS_INTD2_HWIRQ_BASE            229

#define GIC_IRQ_TYPE                     0xf01 /* pulsed interrupt */

/* QMSS INTD registers */
#define QMSS_INTD_EOI                    0x010
#define QMSS_INTD_STATUS0                0x200
#define QMSS_INTD_STATUS1                0x204
#define QMSS_INTD_CLEAR0                 0x280
#define QMSS_INTD_CLEAR1                 0x284
#define QMSS_INTD_COUNT                  0x300

/* QMSS INTD macros */
#define QMSS_INTD_COUNT_IRQ(c)           (QMSS_INTD_COUNT + ((c) << 2))
#define QMSS_INTD_CHANNEL(b, i)          ((b) + (i))
#define QMSS_INTD_CHANNEL_TO_HWIRQ(c, n) (em_hwirq_map[n] + (c))
#define QMSS_INTD_CHANNEL_TO_OFFSET(c)   (((c) & ~0x1f) >> 3)
#define QMSS_INTD_CHANNEL_TO_BITNUM(c)   ((c) & 0x1f)
#define QMSS_INTD_CHANNEL_TO_EOI(c)      ((c) + 2)

static int em_hwirq_map[2] = { QMSS_INTD1_HWIRQ_BASE, QMSS_INTD2_HWIRQ_BASE };
static void __iomem *em_intd_regs[2];

/* GIC device tree match tables */
static const struct of_device_id irq_gic_match[] = {
        { .compatible = "arm,cortex-a15-gic", },
	{}
};

static int em_event_create_gic_virqs(int hwirq)
{
	struct device_node *np;
	const u32 intspec[3] = { 0, hwirq, GIC_IRQ_TYPE };

	np = of_find_matching_node(NULL, irq_gic_match);
	if (!np) {
		EPRINTK("Unable to find gic device node\n");
		return -1;
	}

	return irq_create_of_mapping(np, (const u32*) &intspec, 3);
}

static inline void em_event_irq_unmask(int irq, int instance)
{
	unsigned long flags;

	spin_lock_irqsave(&em_irq_lock, flags);
	if (test_and_clear_bit(0, &em_dev[instance]->irq_status))
		enable_irq(irq);
	spin_unlock_irqrestore(&em_irq_lock, flags);
}

static inline void em_event_irq_mask(int irq, int instance)
{
	unsigned long flags;

	spin_lock_irqsave(&em_irq_lock, flags);
	if (!test_and_set_bit(0, &em_dev[instance]->irq_status))
		disable_irq(irq);
	spin_unlock_irqrestore(&em_irq_lock, flags);
}

static inline int em_event_intd_clear(int channel, int intd)
{
	int status_offset = QMSS_INTD_STATUS0 + QMSS_INTD_CHANNEL_TO_OFFSET(channel);
	int clear_offset  = QMSS_INTD_CLEAR0 + QMSS_INTD_CHANNEL_TO_OFFSET(channel);
	int chan_bitnum   = QMSS_INTD_CHANNEL_TO_BITNUM(channel);

	void __iomem *intd_status_reg = (void *) ((int) em_intd_regs[intd] + status_offset);
	void __iomem *intd_clear_reg  = (void *) ((int) em_intd_regs[intd] + clear_offset);

	DPRINTK("INTD status (0x%x) = 0x%x\n",
		intd_status_reg, __raw_readl(intd_status_reg));

	if(__raw_readl(intd_status_reg) & (1 << chan_bitnum)) {
    		DPRINTK("clear channel %d\n", chan_bitnum);
		__raw_writel(1 << chan_bitnum, intd_clear_reg);
		return 1;
	}

	return 0;
}

static inline void em_event_intd_ack(int channel, int intd, int in_irq)
{
	void __iomem *intd_count_reg  = (void *) ((int) em_intd_regs[intd] + QMSS_INTD_COUNT_IRQ(channel));
	void __iomem *intd_eoi_reg    = (void *) ((int) em_intd_regs[intd] + QMSS_INTD_EOI);

	int count = __raw_readl(intd_count_reg);

	/* Acknowledge the interrupt */
	if (in_irq) {
		__raw_writel(count, intd_count_reg);
		__raw_writel(QMSS_INTD_CHANNEL_TO_EOI(channel), intd_eoi_reg);
	} else 
		__raw_writel(0, intd_count_reg);
}

static irqreturn_t em_event_interrupt_handler(int irq, void *_data)
{
	int instance = (int) _data;
	int channel  = QMSS_INTD_CHANNEL(em_dev[instance]->channel_base, instance);
	int intd     = em_dev[instance]->intd;
	
	DPRINTK("Handler called for instance %d (irq %d)\n", instance, irq);

	/* clear interrupt */
	if (em_event_intd_clear(channel, intd)) {

		/* increase logical event count*/
		atomic_inc(&em_dev[instance]->event);

		/* wake-up the waiting process if any*/
		if (waitqueue_active(&em_dev[instance]->wqh)) {
			wake_up_interruptible(&em_dev[instance]->wqh);
		}
	} else
		return IRQ_NONE;

	/* acknowledge the interrupt */
	em_event_intd_ack(channel, intd, 1);
	
	return IRQ_HANDLED;
}

static int em_event_wait(int instance, struct em_user *user)
{
	int ret;
	int event_count;
	unsigned long flags;
	DECLARE_WAITQUEUE(wait, current);

	if (em_irq[instance] == -1)
		return -1;

	spin_lock_irqsave(&em_irq_lock, flags);
	if (em_dev[instance]->in_use) {
		spin_unlock_irqrestore(&em_irq_lock, flags);
		return -EBUSY;
	}
	em_dev[instance]->in_use = 1;
	spin_unlock_irqrestore(&em_irq_lock, flags);

	DPRINTK("Waiting event for OpenEM instance %d\n", instance);

	add_wait_queue(&em_dev[instance]->wqh, &wait);

	do {
		set_current_state(TASK_INTERRUPTIBLE);

		/* check if some events has been received */
		event_count = atomic_read(&em_dev[instance]->event);
		if (event_count != user->event_count) {
			DPRINTK("prepare to wake up: instance %d, event_count %d, last event_count %d\n",
				instance, event_count, user->event_count);
			user->event_count = event_count;
			break;
		}

		if (signal_pending(current)) {
			ret = -ERESTARTSYS;
			break;
		}

		schedule();

	} while (1);

	__set_current_state(TASK_RUNNING);
	remove_wait_queue(&em_dev[instance]->wqh, &wait);

	DPRINTK("Receiving event for OpenEM instance %d, event = %d\n",
		instance, atomic_read(&em_dev[instance]->event));
	
	spin_lock_irqsave(&em_irq_lock, flags);
	em_dev[instance]->in_use = 0;	
	spin_unlock_irqrestore(&em_irq_lock, flags);
	
	return 0;
}

static int em_event_interrupt_init(void)
{
	em_intd_regs[0] = ioremap(0x02A0C000, 0x1000);
	if (em_intd_regs[0] == 0) {
		EPRINTK("cannot remap QMMS INTD1\n");
		return -1;
	}
	
	em_intd_regs[1] = ioremap(0x02A0D000, 0x1000);
	if (em_intd_regs[1] == 0) {
		EPRINTK("cannot remap QMMS INTD1\n");
		return -1;
	}
	return 0;
}

static int em_event_init(int instance, int channel_base, int intd)
{
	int res;
	int hwirq;
	int channel = QMSS_INTD_CHANNEL(channel_base, instance);

	DPRINTK("creating mapping for instance %d, channel %d, intd %d\n",
		instance, channel, intd);

	if (em_irq[instance] != -1)
		free_irq(em_irq[instance], (void*) instance);

	em_event_intd_clear(channel, intd);
	em_event_intd_ack(channel, intd, 0);

	/* Retrieve and register interrupts */
	hwirq = QMSS_INTD_CHANNEL_TO_HWIRQ(channel, intd);
	em_irq[instance] = em_event_create_gic_virqs(hwirq);

	DPRINTK("hwirq %d mapped to Linux virq %d\n", hwirq, em_irq[instance]);
	
	res = request_irq(em_irq[instance],
			  em_event_interrupt_handler,
			  0,
			  em_dev[instance]->name,
			  (void*) instance);
	if (res < 0) {
		EPRINTK("cannot request irq %d\n", em_irq[instance]);
		em_irq[instance] = -1;
	} 

	em_dev[instance]->channel_base = channel_base;
	em_dev[instance]->intd = intd;
	
	init_waitqueue_head(&em_dev[instance]->wqh);

	return 0;
}

static void em_event_release(int instance) 
{
	DPRINTK("releasing instance %d, freeing irq %d\n", 
		instance, em_irq[instance]);
	
	if (em_irq[instance] != -1)
		free_irq(em_irq[instance], (void*) instance);
	
	if (em_dev[instance]->channel_base != -1) {
		int channel = 
			QMSS_INTD_CHANNEL(em_dev[instance]->channel_base, instance);
		em_event_intd_clear(channel, em_dev[instance]->intd);
		em_event_intd_ack(channel, em_dev[instance]->intd, 0);
	}
	
	em_irq[instance] = -1;
	em_dev[instance]->channel_base = -1;
}
#else  /* TI_EM_USE_INTERRUPT */

static int em_event_init(int instance, int channel_base, int intd)
{
	return ENOTSUPP;
}

static int em_event_wait(int instance, struct em_user *user)
{
	return ENOTSUPP;
}

static void em_event_release(int instance) {}

#endif /* TI_EM_USE_INTERRUPT */

/*
 * File operation
 */
static long em_dev_fop_ioctl(struct file  *file,
			     unsigned int  cmd,
			     unsigned long arg)
{
	struct em_user *user = file->private_data;
	void __user    *argp = (void __user *)arg;
	dma_addr_t      paddr;

	switch(cmd) {
	case EM_IOC_ALLOC:
	{
		int    res;
		size_t size;

		if (copy_from_user(&size, argp, sizeof(size)))
			return -EFAULT;
		
		res = em_chunk_alloc(user, size, &paddr);
		if (res)
			return res;

		if (copy_to_user(argp, &paddr, sizeof(paddr)))
			return -EFAULT;

		break;
	}
	case EM_IOC_WAITEVENTINIT:
	{	/* initialize event notification */
		struct ti_em_intd_irq_config irq_config;

		if (copy_from_user(&irq_config, argp, sizeof(irq_config)))
			return -EFAULT;

		return em_event_init(user->instance,
				     irq_config.channel_base,
				     irq_config.intd);
		break;
	}
	case EM_IOC_WAITEVENTFREE:
		/* free event notification */
		em_event_release(user->instance);
		break;
	case EM_IOC_WAITEVENT:
		/* wait an incoming event */
		return em_event_wait(user->instance, user);
		break;
#ifdef CONFIG_ARM
	case EM_IOC_ENABLEPMU:
		/* enable user access to PMU */
		pmu_enable_user_mode_access();
		break;
#endif
	default:
		return -EINVAL;
	}

	return 0;
}

static int em_dev_fop_open(struct inode *inode, struct file *file)
{
	struct em_user *user;
	unsigned long instance;
	int res;

	res = kstrtoul(file->f_dentry->d_iname + (strlen(EM_MOD_DEVNAME)),
		       10, &instance);
	if (res != 0)
		return res;
	
	user = kzalloc(sizeof(*user), GFP_KERNEL);
	if (!user) {
		EPRINTK("failed to allocate user\n");
		return -ENOMEM;
	}
	
	INIT_LIST_HEAD(&user->chunks);
	user->em          = em_dev[instance];
	user->instance    = instance;
	user->file        = file;
	user->event_count = atomic_read(&em_dev[instance]->event);

	file->private_data = user;

	em_add_user(em_dev[instance], user);

	return 0;
}

static int em_dev_fop_release(struct inode *inode, struct file *file)
{
	struct em_user   *user = file->private_data;
	struct em_device *em   = user->em;
	struct em_chunk  *chunk;

	for (;;) {
		chunk = em_user_first_chunk(user);
		if (!chunk)
			break;
		em_chunk_release(chunk);
	}

	em_del_user(em, user);
	kfree(user);
	file->private_data = NULL;
	return 0;
}

static const struct file_operations em_dev_fops = {
	.owner             = THIS_MODULE,
	.open              = em_dev_fop_open,
	.release           = em_dev_fop_release,
	.mmap		   = em_dev_fop_mmap,
#ifndef CONFIG_MMU
	.get_unmapped_area = em_dev_fop_get_unmapped_area,	
#endif	
	.unlocked_ioctl    = em_dev_fop_ioctl,
};

static int em_dev_register_device(void)
{
	struct miscdevice *misc;
	int res;
	int i;
	
#ifdef TI_EM_USE_INTERRUPT
	res = em_event_interrupt_init();
	if (res)
		return -ENOMEM;
#endif /* TI_EM_USE_INTERRUPT */
	
	for (i = 0; i < TI_EM_INSTANCE_NUM; i++) {
		em_dev[i] = kzalloc(sizeof(struct em_device), GFP_KERNEL);
		if (!em_dev[i]) {
			EPRINTK("cannot allocate device memory\n");
			return -ENOMEM;
		}

		sprintf((char *)&em_dev[i]->name, "%s%d", EM_MOD_DEVNAME, i);
	
		INIT_LIST_HEAD(&em_dev[i]->users);
		kref_init(&em_dev[i]->refcount);
		
		misc = to_misc(em_dev[i]);
		misc->minor = MISC_DYNAMIC_MINOR;
		misc->name  = em_dev[i]->name;
		misc->fops  = &em_dev_fops;

		res = misc_register(misc);
		if (res) {
			EPRINTK("could not register misc device for instance %d\n", i);
			return res;
		}

#ifdef TI_EM_USE_INTERRUPT
		em_irq[i] = -1;
#endif
	}

	return 0;
}

static void em_dev_unregister_device(void)
{
	struct miscdevice *misc;
	int i;

	for (i = 0; i < TI_EM_INSTANCE_NUM; i++) {
		int res;
		em_event_release(i);
		misc = to_misc(em_dev[i]);
		res = misc_deregister(misc);
		if (res)
			DPRINTK("cannot unregister instance %d (res = %d)\n", i, res);
		em_device_put(em_dev[i]);
	}
}

static int __init em_mod_init_module(void)
{
#ifdef CONFIG_ARM
	/* enable user access to qmss h/w */
	mpu_enable_user_mode_access();

	/* enable user access to PMU for calling core */
	pmu_enable_user_mode_access();

#ifdef DEVICE_K2K
	msmc_share_segment(8);
	msmc_share_segment(9);
	msmc_share_segment(10);
#endif /* DEVICE_K2K */
#endif /* CONFIG_ARM */

	/* Setup PktDMA logical QMs (PktDMA clouds) */
	pktdma_qm_init();

	/* register special device for CMA */
	em_dev_register_device();

	return 0;
}

static void __exit em_mod_cleanup_module(void)
{
	em_dev_unregister_device();
}

module_init(em_mod_init_module);
module_exit(em_mod_cleanup_module);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Aurelien Jacquiot <a-jacquiot@ti.com>");
MODULE_DESCRIPTION("TI KeyStone OpenEM module");

