/*
 * Copyright (c) 2013, Texas Instruments Incorporated - http://www.ti.com/
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions are met:
 *       * Redistributions of source code must retain the above copyright
 *         notice, this list of conditions and the following disclaimer.
 *       * Redistributions in binary form must reproduce the above copyright
 *         notice, this list of conditions and the following disclaimer in the
 *         documentation and/or other materials provided with the distribution.
 *       * Neither the name of Texas Instruments Incorporated nor the
 *         names of its contributors may be used to endorse or promote products
 *         derived from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *   DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
 *   DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


/**
 *   @file  ti_em_osal_cma.c
 *
 *   @brief   
 *      This is the OS abstraction layer for the CMA and kernel services management.
 *
 */
#include <ti/runtime/openem/event_machine.h>

#include <ti_em_osal.h>
#include <ti_em_osal_core.h>
#include <ti_em_osal_uaccess.h>
#include <ti_em_osal_cma.h>

static int __ti_em_osal_cma_fd;
static int __ti_em_osal_cma_fd_uncached;

static inline int __get_fd_from_flags(int flags) {
	if (flags & TI_EM_OSAL_MEM_UNCACHED)
		return  __ti_em_osal_cma_fd_uncached;
	else
		return __ti_em_osal_cma_fd;
}

int ti_em_osal_cma_init(void)
{
	/* Standard cached memory */
	__ti_em_osal_cma_fd = open(TI_EM_OSAL_CMA_DEVICE, O_RDWR);
	if (__ti_em_osal_cma_fd == -1) {
		TI_EM_OSAL_ERROR("cannot open %s device!\n", TI_EM_OSAL_CMA_DEVICE);
		return -1;
	}

	/* Specifying O_SYNC for uncached memory allocation */
	__ti_em_osal_cma_fd_uncached = open(TI_EM_OSAL_CMA_DEVICE, O_RDWR | O_SYNC);
	if (__ti_em_osal_cma_fd_uncached == -1) {
		TI_EM_OSAL_ERROR("cannot open %s device!\n", TI_EM_OSAL_CMA_DEVICE);
		return -1;
	}

	return 0;
}

void ti_em_osal_cma_exit(void)
{
	/* This will release all CMA allocations and mappings! */
	if (__ti_em_osal_cma_fd)
		close(__ti_em_osal_cma_fd);
}

void* ti_em_osal_cma_alloc(unsigned int size, int flags, unsigned int *paddr)
{
	void         *vaddr;
	unsigned int  data = size;
	int           fd   = __get_fd_from_flags(flags);
	int           res;

	TI_EM_OSAL_ASSERT(paddr);

	/* align size to page size */
	size = PAGE_ALIGN_UP(size);

	res = ioctl(fd, EM_IOC_ALLOC, &data);
	if (res != 0) {
		TI_EM_OSAL_ERROR("failed to allocate CMA for size=0x%x, errno=%d!\n",
				 size, errno);
		return (void *) -1;
	} 

	*paddr = (unsigned int) data;

	vaddr = mmap(0, size, PROT_READ | PROT_WRITE, MAP_SHARED, fd, (off_t) *paddr);
	if (vaddr == (void *) -1) {
		TI_EM_OSAL_ERROR("failed to map CMA for size=0x%x, errno=%d!\n",
				 size, errno);
		return (void*) -1;
	}
	TI_EM_OSAL_TRACE(2, "CMA returned user vaddr=0x%x\n", (unsigned int) vaddr);

	return vaddr;
}

int ti_em_osal_cma_enable_user_pmu(void)
{
	int res;
	int fd;

	fd = open(TI_EM_OSAL_CMA_DEVICE, O_RDWR | O_SYNC);
	if (fd == -1) {
		TI_EM_OSAL_ERROR("cannot open %s device!\n", TI_EM_OSAL_CMA_DEVICE);
		return -1;
	}

	res = ioctl(__ti_em_osal_cma_fd, EM_IOC_ENABLEPMU);
	if (res != 0) {
		TI_EM_OSAL_ERROR("failed to enable user-access to PMU\n");
		return -1;
	}	

	close(fd);

	return 0;
}
