/*
 * Copyright (c) 2013, Texas Instruments Incorporated - http://www.ti.com/
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions are met:
 *       * Redistributions of source code must retain the above copyright
 *         notice, this list of conditions and the following disclaimer.
 *       * Redistributions in binary form must reproduce the above copyright
 *         notice, this list of conditions and the following disclaimer in the
 *         documentation and/or other materials provided with the distribution.
 *       * Neither the name of Texas Instruments Incorporated nor the
 *         names of its contributors may be used to endorse or promote products
 *         derived from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *   DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
 *   DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


/**
 *   @file  ti_em_osal_core.c
 *
 *   @brief
 *      This is the OS abstraction layer for the core execution services.
 *
 */
#include <ti_em_osal_core.h>
#include <ti_em_osal_shm.h>

/*
 * This is the core Id as needed by OpenEM
 * It is not necessarly related to real hw core id on Linux but first to either
 * process or thread. These processes/threads may then be mapped to hw core id
 * (i.e.CPU)
 */
TI_EM_OSAL_CORE_LOCAL_VARIABLE(uint32_t, __ti_em_osal_core_id);

#define __TI_EM_OSAL_CORE_ID_MASK  ((1 << (TI_EM_OSAL_CORE_NUM)) - 1)
#define TI_EM_OSAL_CORE_IDX(c)     (1 << (c))

static int __ti_em_osal_get_new_core_idx(volatile uint32_t *core_id_bf_global)
{
	uint32_t new_core_id;

	if (!core_id_bf_global)
		return -1;

	ti_em_osal_core_lock();

	/* check if there are available core Ids */
	if ((*core_id_bf_global & __TI_EM_OSAL_CORE_ID_MASK) == __TI_EM_OSAL_CORE_ID_MASK) {
		ti_em_osal_core_unlock();
		return -1;
	}

	new_core_id = _ffz(*core_id_bf_global);
	*core_id_bf_global |= TI_EM_OSAL_CORE_IDX(new_core_id);

	ti_em_osal_core_unlock();

	return new_core_id;
}

static int __ti_em_osal_release_core_idx(volatile uint32_t *core_id_bf_global, uint32_t core_id)
{
	if (!core_id_bf_global)
		return -1;

	ti_em_osal_core_lock();
	*core_id_bf_global &= ~(TI_EM_OSAL_CORE_IDX(core_id));
	ti_em_osal_core_unlock();

	return 0;
}

#ifdef TI_EM_THREAD_MODEL
/*
 * These are the core functions when using the thread model
 */
static uint32_t __ti_em_osal_core_id_bf_global;
static uint32_t __ti_em_osal_master_core_id = 0;

struct  __ti_em_osal_create_core_info {
	int (*start_routine)(void*);
	void *arg;
}; 

static int __ti_em_osal_create_core_trampoline(void *arg)
{
	uint32_t ___ti_em_osal_core_id;
	struct __ti_em_osal_create_core_info *p_info
		= (struct __ti_em_osal_create_core_info*) arg;
	int (*start_routine)(void*) = p_info->start_routine;
	void *c_arg                 = p_info->arg;
	int ret;

	ret = __ti_em_osal_get_new_core_idx(&__ti_em_osal_core_id_bf_global);
	if (ret < 0) {
		free(p_info);
		return -1;
	}
	___ti_em_osal_core_id = (uint32_t) ret;

	TI_EM_OSAL_CORE_LOCAL_STORE(__ti_em_osal_core_id, &___ti_em_osal_core_id);

	free(p_info);

	/* call the thread function */
	ret = (int) ((start_routine)(c_arg));

	__ti_em_osal_release_core_idx(&__ti_em_osal_core_id_bf_global, ___ti_em_osal_core_id);

	return ret;
}

int ti_em_osal_core_spawn(int(*start_routine)(void*), void *arg)
{
	int       ret;
	pthread_t thread;

	ret = pthread_create(&thread,
			     0,
			     (void*(*)(void*))(start_routine),
			     (void *)arg);

	if (ret != 0) {
		return -1;
	}

	return (int) thread;
}

int ti_em_osal_core_instance_create(int(*start_routine)(void*), void *arg)
{
	int       ret;
	struct __ti_em_osal_create_core_info *p_info;

	p_info = (struct __ti_em_osal_create_core_info*)
		malloc(sizeof(struct __ti_em_osal_create_core_info));
	
	if (p_info == NULL)
		return -1;

	p_info->start_routine = start_routine;
	p_info->arg           = arg;

	ret = ti_em_osal_core_spawn(__ti_em_osal_create_core_trampoline,
				    (void *)p_info);

	if (ret == -1)
		free(p_info);

	return ret;
}

int ti_em_osal_core_exit(int retval)
{
	uint32_t core_id = ti_em_osal_get_core_idx();
	__ti_em_osal_release_core_idx(&__ti_em_osal_core_id_bf_global, core_id);

	if (core_id)
		pthread_exit((void*) retval);
	else {
		ti_em_osal_shm_release(TI_EM_OSAL_SHM_OSAL_ID);
		exit(retval);
	}
}

int ti_em_osal_core_wait_exit(int core_handler, int *retval)
{
	int res;
	res = pthread_join((pthread_t) core_handler, (void **)retval);
	if (res != 0)
		return -1;

	return 0;
}

int ti_em_osal_core_kill(int core_handler, int sig)
{
	int res;
	res = pthread_kill((pthread_t) core_handler, sig);
	if (res != 0)
		return -1;

	return 0;

}

int ti_em_osal_core_set_affinity(int core_handler, int hw_core_id)
{
	int       res;
	cpu_set_t cpu_set;

	CPU_ZERO(&cpu_set);
	CPU_SET(hw_core_id, &cpu_set);

	if (core_handler == 0)
		res = sched_setaffinity((pid_t) 0, sizeof(cpu_set), &cpu_set);
	else
		res = pthread_setaffinity_np((pthread_t) core_handler, 
					     sizeof(cpu_set), &cpu_set);
	if (res != 0) 
		return -1;

	return 0;
}

int ti_em_osal_core_set_rt_priority(int core_handler, int priority)
{
	int       res;

	struct sched_param param;

	if (priority == 0)
		param.sched_priority = sched_get_priority_max(SCHED_RR);
	else
		param.sched_priority = priority;

	if (core_handler == 0)
		res = sched_setscheduler((pid_t) 0, SCHED_RR, &param);
	else
		res = pthread_setschedparam((pthread_t) core_handler, SCHED_RR, &param);
	if (res != 0) 
		return -1;

	return 0;
}

static pthread_mutex_t __ti_em_osal_core_mutex = PTHREAD_MUTEX_INITIALIZER;

void ti_em_osal_core_lock(void)
{
	pthread_mutex_lock(&__ti_em_osal_core_mutex);
}

void ti_em_osal_core_unlock(void)
{
	pthread_mutex_unlock(&__ti_em_osal_core_mutex);
}

static pthread_barrier_t __ti_em_osal_core_barrier;

int ti_em_osal_core_barrier_init(int core_num)
{
	return pthread_barrier_init(&__ti_em_osal_core_barrier,
				    NULL,
				    core_num);
}

void ti_em_osal_core_barrier_wait(void)
{
	(void) pthread_barrier_wait(&__ti_em_osal_core_barrier);
}

void ti_em_osal_core_init_global(void)
{
	TI_EM_OSAL_CORE_LOCAL_INIT(__ti_em_osal_core_id);
	TI_EM_OSAL_CORE_LOCAL_STORE(__ti_em_osal_core_id, &__ti_em_osal_master_core_id);
	TI_EM_OSAL_CORE_LOCAL_WRITE(__ti_em_osal_core_id, 0);
	
	__ti_em_osal_core_id_bf_global = TI_EM_OSAL_CORE_IDX(0);
}

int ti_em_osal_core_join(int core_id)
{
	/* Join is not supported in thread model */
	return -1;
}
#else /* TI_EM_THREAD_MODEL */
/*
 * These are the core functions when using the process model
 */
#include <ti_em_hw_mach.h>
#include <ti_em_osal_shm.h>

/* The pointed data of these pointers must be shared accross OpenEM cores */
static sem_t *__ti_em_osal_core_mutex;
static sem_t *__ti_em_osal_core_barrier;
static volatile int *__ti_em_osal_core_barrier_count;
static uint32_t __ti_em_osal_core_id_bf_init = TI_EM_OSAL_CORE_IDX(0);
static volatile uint32_t *__ti_em_osal_core_id_bf_global = &__ti_em_osal_core_id_bf_init;

int ti_em_osal_core_spawn(int(*start_routine)(void*), void *arg)
{
	int ret;

	ret = fork();
	switch (ret) {
	case 0:
		ret = ((start_routine)(arg));
		exit(ret);
		break;
	case -1:
		/* fork failed */
		return -1; 
		break;
	default:
		break;
		/* We are the parent process */
	}
	/* return PID of child */
	return ret;
}

int ti_em_osal_core_instance_create(int(*start_routine)(void*), void *arg)
{
	int ret;

	ret = fork();
	switch (ret) {
	case 0:
		/* we are the child process */
		ret = __ti_em_osal_get_new_core_idx(__ti_em_osal_core_id_bf_global);
		if (ret < 0)
			exit(-1);

		__ti_em_osal_core_id = (uint32_t) ret;

		TI_EM_OSAL_TRACE(3, "child %d core %d\n", getpid(),
				 __ti_em_osal_core_id);

		ret = ((start_routine)(arg));

		__ti_em_osal_release_core_idx(__ti_em_osal_core_id_bf_global, __ti_em_osal_core_id);
		exit(ret);
		break;
	case -1:
		/* fork failed */
		return -1; 
		break;
	default:
		break;
		/* We are the parent process */
	}
	/* return PID of child */
	return ret;
}

int ti_em_osal_core_exit(int retval)
{
	uint32_t core_id = ti_em_osal_get_core_idx();
	__ti_em_osal_release_core_idx(__ti_em_osal_core_id_bf_global, core_id);
	if (core_id == 0) {
		ti_em_osal_shm_release(TI_EM_OSAL_SHM_OSAL_ID);
		ti_em_osal_shm_release(TI_EM_OSAL_SHM_OSAL_MAPPING_DB_ID);
	}
	exit(retval);
}

int ti_em_osal_core_wait_exit(int core_handler, int *retval)
{
	int   res;
	
	res = waitpid((pid_t) core_handler, retval, 0);
	if (res == -1)
		return -1;

	return 0;
}

int ti_em_osal_core_kill(int core_handler, int sig)
{
	int res;
	res = kill((pid_t) core_handler, sig);
	if (res != 0)
		return -1;

	return 0;

}

int ti_em_osal_core_set_affinity(int core_handler, int hw_core_id)
{
	int       res;
	cpu_set_t cpu_set;

	TI_EM_OSAL_TRACE(3, "core handler %d mapped to hw core id %d\n",
			 core_handler, hw_core_id);

	CPU_ZERO(&cpu_set);
	CPU_SET(hw_core_id, &cpu_set);

	res = sched_setaffinity((pid_t) core_handler, sizeof(cpu_set), &cpu_set);
	if (res != 0) 
		return -1;

	return 0;
}

int ti_em_osal_core_set_rt_priority(int core_handler, int priority)
{
	int       res;

	struct sched_param param;

	if (priority == 0)
		param.sched_priority = sched_get_priority_max(SCHED_FIFO);
	else
		param.sched_priority = priority;

	res = sched_setscheduler((pid_t) core_handler, SCHED_FIFO, &param);
	if (res != 0) 
		return -1;

	return 0;
}

void ti_em_osal_core_lock(void)
{
	sem_wait(__ti_em_osal_core_mutex);
}

void ti_em_osal_core_unlock(void)
{
	sem_post(__ti_em_osal_core_mutex);
}

int ti_em_osal_core_barrier_init(int core_num)
{
	ti_em_osal_core_lock();
	*__ti_em_osal_core_barrier_count = core_num;
	ti_em_osal_mfence();
	ti_em_osal_core_unlock();

	return 0;
}

void ti_em_osal_core_barrier_wait(void)
{
	ti_em_osal_core_lock();
	*__ti_em_osal_core_barrier_count -= 1;
	ti_em_osal_mfence();
	ti_em_osal_core_unlock();

	if (*__ti_em_osal_core_barrier_count == 0)
		(void) sem_post(__ti_em_osal_core_barrier);

	(void) sem_wait(__ti_em_osal_core_barrier);
	(void) sem_post(__ti_em_osal_core_barrier);
}

void ti_em_osal_core_init_global(void)
{
	TI_EM_OSAL_CORE_LOCAL_WRITE(__ti_em_osal_core_id, 0);

	/* One page is enough for our purpose */
	ti_em_osal_shm_init(TI_EM_OSAL_SHM_OSAL_ID,
			    TI_EM_OSAL_SHM_OSAL_NAME,
			    PAGE_SIZE,
			    1);

	__ti_em_osal_core_barrier_count = (int *) ti_em_osal_shm_alloc(TI_EM_OSAL_SHM_OSAL_ID, sizeof(int));
	__ti_em_osal_core_mutex         = (sem_t *) ti_em_osal_shm_alloc(TI_EM_OSAL_SHM_OSAL_ID, sizeof(sem_t));
	__ti_em_osal_core_barrier       = (sem_t *) ti_em_osal_shm_alloc(TI_EM_OSAL_SHM_OSAL_ID, sizeof(sem_t));
	__ti_em_osal_core_id_bf_global  = (uint32_t *) ti_em_osal_shm_alloc(TI_EM_OSAL_SHM_OSAL_ID, sizeof(uint32_t));
 
       /* Initialize semaphore as a shared mutex */
       sem_init(__ti_em_osal_core_mutex, 1, 1);
       sem_init(__ti_em_osal_core_barrier, 1, 0);

       *__ti_em_osal_core_id_bf_global = TI_EM_OSAL_CORE_IDX(0);
}

int ti_em_osal_core_join(int core_id)
{
	int ret;

	/* Retrieve the OSAL database */
	ti_em_osal_shm_init(TI_EM_OSAL_SHM_OSAL_ID,
			    TI_EM_OSAL_SHM_OSAL_NAME,
			    PAGE_SIZE,
			    0);
	
	__ti_em_osal_core_barrier_count = (int *) ti_em_osal_shm_alloc(TI_EM_OSAL_SHM_OSAL_ID, sizeof(int));
	__ti_em_osal_core_mutex         = (sem_t *) ti_em_osal_shm_alloc(TI_EM_OSAL_SHM_OSAL_ID, sizeof(sem_t));
	__ti_em_osal_core_barrier       = (sem_t *) ti_em_osal_shm_alloc(TI_EM_OSAL_SHM_OSAL_ID, sizeof(sem_t));
	__ti_em_osal_core_id_bf_global  = (uint32_t *) ti_em_osal_shm_alloc(TI_EM_OSAL_SHM_OSAL_ID, sizeof(uint32_t));

	/* Check if the joining process need a free OpenEM core Id */
	if (core_id < 0) {
		/* Get a new core Id */
		ret = __ti_em_osal_get_new_core_idx(__ti_em_osal_core_id_bf_global);
		if (ret < 0)
			return -1;
		__ti_em_osal_core_id = (uint32_t) ret;
	} else {
		/* Otherwise use the provided one */
		if (core_id > TI_EM_OSAL_CORE_NUM)
			return -1;
		__ti_em_osal_core_id = core_id;
	}
	return 0;
}
#endif  /* TI_EM_THREAD_MODEL */
