/*
 * Copyright (c) 2013, Texas Instruments Incorporated - http://www.ti.com/
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions are met:
 *       * Redistributions of source code must retain the above copyright
 *         notice, this list of conditions and the following disclaimer.
 *       * Redistributions in binary form must reproduce the above copyright
 *         notice, this list of conditions and the following disclaimer in the
 *         documentation and/or other materials provided with the distribution.
 *       * Neither the name of Texas Instruments Incorporated nor the
 *         names of its contributors may be used to endorse or promote products
 *         derived from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *   DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
 *   DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


/**
 *   @file  ti_em_osal_cppi.c
 *
 *   @brief
 *      This is the OS abstraction layer for the PktDMA (CPPI) management.
 *
 */

#include <ti/runtime/openem/event_machine.h>

#include <ti_em_hw_functions.h>
#include <ti_em_osal.h>
#include <ti_em_osal_core.h>
#include <ti_em_osal_queue.h>
#include <ti_em_osal_cppi.h>

#include <ti_em_hw_csl.h>
#include <ti_em_hw_cppi.h>

Cppi_GlobalCPDMAConfigParams ti_em_osal_cppi_global_cfg_params[CPPI_MAX_CPDMA];

em_status_t ti_em_osal_cppi_rx_flow_open(int dma_idx,
					 int flow_idx,
					 int dst_queue,
					 int free_queue,
					 int error_handling)
{
	Cppi_RxFlowCfg rx_flow_cfg;
	int            res;

       	memset ((void *) &rx_flow_cfg, 0, sizeof (Cppi_RxFlowCfg));

	rx_flow_cfg.flowIdNum          = flow_idx;
	rx_flow_cfg.rx_dest_qnum       = (dst_queue & 0xFFF);
	rx_flow_cfg.rx_dest_qmgr       = (dst_queue >> 12);
	rx_flow_cfg.rx_desc_type       = Cppi_DescType_HOST;

	/* Mandatory parameters to configure the RX FLOW */
	rx_flow_cfg.rx_einfo_present   = 1;
	rx_flow_cfg.rx_dest_tag_lo_sel = 4;
	rx_flow_cfg.rx_dest_tag_hi_sel = 5;
	rx_flow_cfg.rx_src_tag_hi_sel  = 4;
	rx_flow_cfg.rx_psinfo_present  = 1;
	rx_flow_cfg.rx_ps_location     = 0;
	rx_flow_cfg.rx_error_handling  = error_handling;
	rx_flow_cfg.rx_fdq0_sz0_qnum   = (free_queue & 0xFFF);
	rx_flow_cfg.rx_fdq0_sz0_qmgr   = (free_queue >> 12);
	rx_flow_cfg.rx_fdq0_sz1_qnum   = (free_queue & 0xFFF);
	rx_flow_cfg.rx_fdq0_sz1_qmgr   = (free_queue >> 12);
	rx_flow_cfg.rx_fdq0_sz2_qnum   = (free_queue & 0xFFF);
	rx_flow_cfg.rx_fdq0_sz2_qmgr   = (free_queue >> 12);
	rx_flow_cfg.rx_fdq0_sz3_qnum   = (free_queue & 0xFFF);
	rx_flow_cfg.rx_fdq0_sz3_qmgr   = (free_queue >> 12);

	/* Other fields are left to be modified by user */

	TI_EM_OSAL_TRACE(2, "opening PktDMA infra flow %d (idx %d) for dma %d, dst queue = 0x%x (%d), free_queue = 0x%x (%d)\n",
			 rx_flow_cfg.flowIdNum, flow_idx, dma_idx, dst_queue, dst_queue, free_queue, free_queue);

	/* Configure Rx flow */
	res = ti_em_cppi_configure_rx_flow((Cppi_Handle) dma_idx,
					   &ti_em_osal_cppi_global_cfg_params[0],
					   &rx_flow_cfg);
	if (res < 0)
		return (em_status_t) EM_ERR;
	
        return (em_status_t) EM_OK;
}

em_status_t ti_em_osal_cppi_tx_channel_open(int dma_idx, int channel_idx)
{
	Cppi_TxChInitCfg tx_ch_cfg;
	int              res;

	memset ((void *) &tx_ch_cfg, 0, sizeof (Cppi_TxChInitCfg));

	tx_ch_cfg.channelNum = channel_idx;
	tx_ch_cfg.priority   = 2;
	tx_ch_cfg.txEnable   = Cppi_ChState_CHANNEL_ENABLE;

	TI_EM_OSAL_TRACE(2, "opening PktDMA infra Tx channel %d (idx %d) for dma %d\n",
			 tx_ch_cfg.channelNum, channel_idx, dma_idx);

	/* Open Tx Channel */
	res = ti_em_cppi_tx_channel_open((Cppi_Handle) dma_idx, 
					 &ti_em_osal_cppi_global_cfg_params[0],
					 &tx_ch_cfg);
	if (res < 0)
		return (em_status_t) EM_ERR;
	
        return (em_status_t) EM_OK;
}

em_status_t ti_em_osal_cppi_rx_channel_open(int dma_idx, int channel_idx)
{
	Cppi_RxChInitCfg rx_ch_cfg;
	int              res;
	
	memset ((void *) &rx_ch_cfg, 0, sizeof (Cppi_RxChInitCfg));

	rx_ch_cfg.channelNum = channel_idx;
	rx_ch_cfg.rxEnable   = Cppi_ChState_CHANNEL_ENABLE;

	TI_EM_OSAL_TRACE(2, "opening PktDMA infra Rx channel %d (idx %d) for dma %d\n",
			 rx_ch_cfg.channelNum, channel_idx, dma_idx);

	/* Open Rx Channel */
	res = ti_em_cppi_rx_channel_open((Cppi_Handle) dma_idx, 
					 &ti_em_osal_cppi_global_cfg_params[0],
					 &rx_ch_cfg);
	if (res < 0)
		return (em_status_t) EM_ERR;
	
        return (em_status_t) EM_OK;
}

em_status_t ti_em_osal_cppi_rx_flow_close(int dma_idx,
					  int flow_idx)
{
        return (em_status_t) EM_OK;
}

em_status_t ti_em_osal_cppi_tx_channel_close(int dma_idx, int channel_idx)
{
        return (em_status_t) EM_OK;
}

em_status_t ti_em_osal_cppi_rx_channel_close(int dma_idx, int channel_idx)
{
        return (em_status_t) EM_OK;
}

/*
 * Perform initialization of CPPI structure that will contain all variouss 
 * CPPI registers (with virtual address)
 */
em_status_t ti_em_osal_cppi_init(void)
{
	Cppi_GlobalCPDMAConfigParams *cppi_conf = &ti_em_osal_cppi_global_cfg_params[0];

	memset(cppi_conf, 0, sizeof(ti_em_osal_cppi_global_cfg_params));

	/* SRIO CPDMA regs */
	cppi_conf[Cppi_CpDma_SRIO_CPDMA].gblCfgRegs  = (void*) ti_em_osal_io_phys_to_virt(CSL_SRIO_CFG_PKTDMA_GLOBAL_CFG_REGS);
	cppi_conf[Cppi_CpDma_SRIO_CPDMA].txChRegs    = (void*) ti_em_osal_io_phys_to_virt(CSL_SRIO_CFG_PKTDMA_TX_CFG_REGS);
	cppi_conf[Cppi_CpDma_SRIO_CPDMA].rxChRegs    = (void*) ti_em_osal_io_phys_to_virt(CSL_SRIO_CFG_PKTDMA_RX_CFG_REGS);
	cppi_conf[Cppi_CpDma_SRIO_CPDMA].txSchedRegs = (void*) ti_em_osal_io_phys_to_virt(CSL_SRIO_CFG_PKTDMA_TX_SCHEDULER_CFG_REGS);
        cppi_conf[Cppi_CpDma_SRIO_CPDMA].rxFlowRegs  = (void*) ti_em_osal_io_phys_to_virt(CSL_SRIO_CFG_PKTDMA_RX_FLOW_CFG_REGS);

	/* PASS CPDMA regs */
	cppi_conf[Cppi_CpDma_PASS_CPDMA].gblCfgRegs  = (void*) ti_em_osal_io_phys_to_virt(CSL_NETCP_CFG_PKTDMA_GLOBAL_CFG_REGS);
	cppi_conf[Cppi_CpDma_PASS_CPDMA].txChRegs    = (void*) ti_em_osal_io_phys_to_virt(CSL_NETCP_CFG_PKTDMA_TX_CFG_REGS);
	cppi_conf[Cppi_CpDma_PASS_CPDMA].rxChRegs    = (void*) ti_em_osal_io_phys_to_virt(CSL_NETCP_CFG_PKTDMA_RX_CFG_REGS);
	cppi_conf[Cppi_CpDma_PASS_CPDMA].txSchedRegs = (void*) ti_em_osal_io_phys_to_virt(CSL_NETCP_CFG_PKTDMA_TX_SCHEDULER_CFG_REGS);
	cppi_conf[Cppi_CpDma_PASS_CPDMA].rxFlowRegs  = (void*) ti_em_osal_io_phys_to_virt(CSL_NETCP_CFG_PKTDMA_RX_FLOW_CFG_REGS);

#ifdef DEVICE_K2K

	/* QMSS (Infrastructure) CPDMA 1 regs */
	cppi_conf[Cppi_CpDma_QMSS_CPDMA].gblCfgRegs  = (void*) ti_em_osal_io_phys_to_virt(CSL_QMSS_CFG_PKTDMA_1_GLOBAL_CFG_REGS);
	cppi_conf[Cppi_CpDma_QMSS_CPDMA].txChRegs    = (void*) ti_em_osal_io_phys_to_virt(CSL_QMSS_CFG_PKTDMA_1_TX_CFG_REGS);
	cppi_conf[Cppi_CpDma_QMSS_CPDMA].rxChRegs    = (void*) ti_em_osal_io_phys_to_virt(CSL_QMSS_CFG_PKTDMA_1_RX_CFG_REGS);
	cppi_conf[Cppi_CpDma_QMSS_CPDMA].txSchedRegs = (void*) ti_em_osal_io_phys_to_virt(CSL_QMSS_CFG_PKTDMA_1_TX_SCHEDULER_CFG_REGS);
	cppi_conf[Cppi_CpDma_QMSS_CPDMA].rxFlowRegs  = (void*) ti_em_osal_io_phys_to_virt(CSL_QMSS_CFG_PKTDMA_1_RX_FLOW_CFG_REGS);	

	/* QMSS (Infrastructure) CPDMA 2 regs */
	cppi_conf[Cppi_CpDma_QMSS_QM2_CPDMA].gblCfgRegs  = (void*) ti_em_osal_io_phys_to_virt(CSL_QMSS_CFG_PKTDMA_2_GLOBAL_CFG_REGS);
	cppi_conf[Cppi_CpDma_QMSS_QM2_CPDMA].txChRegs    = (void*) ti_em_osal_io_phys_to_virt(CSL_QMSS_CFG_PKTDMA_2_TX_CFG_REGS);
	cppi_conf[Cppi_CpDma_QMSS_QM2_CPDMA].rxChRegs    = (void*) ti_em_osal_io_phys_to_virt(CSL_QMSS_CFG_PKTDMA_2_RX_CFG_REGS);
	cppi_conf[Cppi_CpDma_QMSS_QM2_CPDMA].txSchedRegs = (void*) ti_em_osal_io_phys_to_virt(CSL_QMSS_CFG_PKTDMA_2_TX_SCHEDULER_CFG_REGS);
	cppi_conf[Cppi_CpDma_QMSS_QM2_CPDMA].rxFlowRegs  = (void*) ti_em_osal_io_phys_to_virt(CSL_QMSS_CFG_PKTDMA_2_RX_FLOW_CFG_REGS);

	/* XGE CPDMA regs */
	cppi_conf[Cppi_CpDma_XGE_CPDMA].gblCfgRegs  = (void*) ti_em_osal_io_phys_to_virt(CSL_XGE_CFG_PKTDMA_GLOBAL_CFG_REGS);
	cppi_conf[Cppi_CpDma_XGE_CPDMA].txChRegs    = (void*) ti_em_osal_io_phys_to_virt(CSL_XGE_CFG_PKTDMA_TX_CFG_REGS);
	cppi_conf[Cppi_CpDma_XGE_CPDMA].rxChRegs    = (void*) ti_em_osal_io_phys_to_virt(CSL_XGE_CFG_PKTDMA_RX_CFG_REGS);
	cppi_conf[Cppi_CpDma_XGE_CPDMA].txSchedRegs = (void*) ti_em_osal_io_phys_to_virt(CSL_XGE_CFG_PKTDMA_TX_SCHEDULER_CFG_REGS);
	cppi_conf[Cppi_CpDma_XGE_CPDMA].rxFlowRegs  = (void*) ti_em_osal_io_phys_to_virt(CSL_XGE_CFG_PKTDMA_RX_FLOW_CFG_REGS);

#else /* DEVICE_K2K */

	/* QMSS (Infrastructure) CPDMA regs */
	cppi_conf[Cppi_CpDma_QMSS_CPDMA].gblCfgRegs  = (void*) ti_em_osal_io_phys_to_virt(CSL_QM_SS_CFG_CPPI_DMA_GLOBAL_CFG_REGS);
	cppi_conf[Cppi_CpDma_QMSS_CPDMA].txChRegs    = (void*) ti_em_osal_io_phys_to_virt(CSL_QM_SS_CFG_CPPI_DMA_TX_CFG_REGS);
	cppi_conf[Cppi_CpDma_QMSS_CPDMA].rxChRegs    = (void*) ti_em_osal_io_phys_to_virt(CSL_QM_SS_CFG_CPPI_DMA_RX_CFG_REGS);
	cppi_conf[Cppi_CpDma_QMSS_CPDMA].txSchedRegs = (void*) ti_em_osal_io_phys_to_virt(CSL_QM_SS_CFG_CPPI_DMA_TX_SCHEDULER_CFG_REGS);
	cppi_conf[Cppi_CpDma_QMSS_CPDMA].rxFlowRegs  = (void*) ti_em_osal_io_phys_to_virt(CSL_QM_SS_CFG_CPPI_DMA_RX_FLOW_CFG_REGS);	

#endif /* DEVICE_K2K */

	return EM_OK;
}

em_status_t ti_em_osal_chain_rx_flow_open(int dma_idx,
					  int flow_idx,
					  int dst_queue,
					  int free_queue_sizea,
					  int sizea,
					  int free_queue_sizeb,
					  int sizeb,
					  int free_queue_sizec,
					  int sizec,
					  int free_queue_sized,
					  int sized,
					  int free_queue_overflow,
					  int error_handling)
{
	Cppi_RxFlowCfg rx_flow_cfg;
	int            res;

       	memset ((void *) &rx_flow_cfg, 0, sizeof (Cppi_RxFlowCfg));

	rx_flow_cfg.flowIdNum          = flow_idx;
	rx_flow_cfg.rx_dest_qnum       = (dst_queue & 0xFFF);
	rx_flow_cfg.rx_dest_qmgr       = (dst_queue >> 12);
	rx_flow_cfg.rx_desc_type       = Cppi_DescType_HOST;

	/* Mandatory parameters to configure the RX FLOW */
	rx_flow_cfg.rx_einfo_present   = 1;
	rx_flow_cfg.rx_dest_tag_lo_sel = 4;
	rx_flow_cfg.rx_dest_tag_hi_sel = 5;
	rx_flow_cfg.rx_src_tag_hi_sel  = 4;
	rx_flow_cfg.rx_psinfo_present  = 1;
	rx_flow_cfg.rx_ps_location     = 0;
	rx_flow_cfg.rx_error_handling  = error_handling;
	rx_flow_cfg.rx_fdq0_sz0_qnum   = (free_queue_sizea & 0xFFF);
	rx_flow_cfg.rx_fdq0_sz0_qmgr   = (free_queue_sizea >> 12);
	rx_flow_cfg.rx_fdq0_sz1_qnum   = (free_queue_sizeb & 0xFFF);
	rx_flow_cfg.rx_fdq0_sz1_qmgr   = (free_queue_sizeb >> 12);
	rx_flow_cfg.rx_fdq0_sz2_qnum   = (free_queue_sizec & 0xFFF);
	rx_flow_cfg.rx_fdq0_sz2_qmgr   = (free_queue_sizec >> 12);
	rx_flow_cfg.rx_fdq0_sz3_qnum   = (free_queue_sized & 0xFFF);
	rx_flow_cfg.rx_fdq0_sz3_qmgr   = (free_queue_sized >> 12);
	rx_flow_cfg.rx_fdq1_qnum       = (free_queue_overflow & 0xFFF);
	rx_flow_cfg.rx_fdq1_qmgr       = (free_queue_overflow >> 12);
	rx_flow_cfg.rx_fdq2_qnum       = (free_queue_overflow & 0xFFF);
	rx_flow_cfg.rx_fdq2_qmgr       = (free_queue_overflow >> 12);
	rx_flow_cfg.rx_fdq3_qnum       = (free_queue_overflow & 0xFFF);
	rx_flow_cfg.rx_fdq3_qmgr       = (free_queue_overflow >> 12);

	/* Check if thresholds need to be used and set them accordingly to queue sizes */
	if (free_queue_sizeb == free_queue_sizea)
		goto skip_threshold;

	rx_flow_cfg.rx_size_thresh0_en = 1;
	rx_flow_cfg.rx_size_thresh0    = sizea;
	
	TI_EM_OSAL_TRACE(2, "enabling threshold 0 for size A %d with free queue = 0x%x (%d)\n",
			 sizea, free_queue_sizea, free_queue_sizea);

	if (free_queue_sizec == free_queue_sizeb)
		goto skip_threshold;

	rx_flow_cfg.rx_size_thresh1_en = 1;
	rx_flow_cfg.rx_size_thresh1    = sizeb;

	TI_EM_OSAL_TRACE(2, "enabling threshold 1 for size B %d with free queue = 0x%x (%d)\n",
			 sizeb, free_queue_sizeb, free_queue_sizeb);

	if (free_queue_sized == free_queue_sizec)
		goto skip_threshold;

	rx_flow_cfg.rx_size_thresh2_en = 1;
	rx_flow_cfg.rx_size_thresh2    = sizec;

	TI_EM_OSAL_TRACE(2, "enabling threshold 2 for size C %d with free queue = 0x%x (%d)\n",
			 sizec, free_queue_sizec, free_queue_sizec);

skip_threshold:

	TI_EM_OSAL_TRACE(2, "opening PktDMA infra flow %d (idx %d) for dma %d, dst queue = 0x%x (%d)\n",
			 rx_flow_cfg.flowIdNum, flow_idx, dma_idx, dst_queue, dst_queue);

	/* Configure Rx flow */
	res = ti_em_cppi_configure_rx_flow((Cppi_Handle) dma_idx,
					   &ti_em_osal_cppi_global_cfg_params[0],
					   &rx_flow_cfg);
	if (res < 0)
		return (em_status_t) EM_ERR;
	
        return (em_status_t) EM_OK;
}

ti_em_queue_id_t ti_em_osal_cppi_get_absolute_queue_id(ti_em_dma_id_t dma_idx, ti_em_queue_id_t queue_idx)
{
	ti_em_queue_id_t queue;
   
	queue = ti_em_qmss_get_dma_queue_base(dma_idx) + queue_idx;

	TI_EM_OSAL_TRACE(2, "absolute queue for dma %d, queue idx %d is 0x%x (%d)\n",
			 dma_idx, queue_idx, queue, queue);

	return queue;
}
