/*
 * Copyright (c) 2013, Texas Instruments Incorporated - http://www.ti.com/
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions are met:
 *       * Redistributions of source code must retain the above copyright
 *         notice, this list of conditions and the following disclaimer.
 *       * Redistributions in binary form must reproduce the above copyright
 *         notice, this list of conditions and the following disclaimer in the
 *         documentation and/or other materials provided with the distribution.
 *       * Neither the name of Texas Instruments Incorporated nor the
 *         names of its contributors may be used to endorse or promote products
 *         derived from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *   DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
 *   DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


/**
 *   @file  ti_em_osal_event.c
 *
 *   @brief   
 *      This is the OS abstraction layer for waiting events
 *
 */
#include <ti/runtime/openem/event_machine.h>

#include <ti_em_osal.h>
#include <ti_em_osal_core.h>
#include <ti_em_osal_event.h>

static int __ti_em_osal_event_fd[TI_EM_CORE_NUM];

/* Number or IRQ configurations */
#define IRQ_CONFIG_NUM    4

/* INTD mappings for each PDSP */
static int ti_em_irq_map[IRQ_CONFIG_NUM] = {
	0,     /* PDSP 0,4 */
	0,     /* PDSP 1,5 */
	1,     /* PDSP 2,6 */
	1,     /* PDSP 3,7 */
};

int ti_em_osal_event_init(int pdsp, int channel_base)
{
	struct ti_em_intd_irq_config irq_config;
	char event_file[64];
	int res;
	int core = ti_em_osal_get_core_idx();

	if (__ti_em_osal_event_fd[core] > 0) {
		close(__ti_em_osal_event_fd[core]);
	}
		
	/* Compute the special file for our core */
	res = sprintf(event_file, "%s%d", TI_EM_OSAL_EVENT_DEVICE_BASENAME, core);
	if (res < 0) {
		TI_EM_OSAL_ERROR("cannot create event filename!\n");
		return -1;
	}

	__ti_em_osal_event_fd[core] = open(event_file, O_RDWR);
	if (__ti_em_osal_event_fd[core] == -1) {
		TI_EM_OSAL_ERROR("cannot open %s device!\n", event_file);
		return -1;
	}

	irq_config.channel_base = channel_base;
	irq_config.intd         = ti_em_irq_map[pdsp % IRQ_CONFIG_NUM];

	res = ioctl(__ti_em_osal_event_fd[core], EM_IOC_WAITEVENTINIT, &irq_config);
	if (res != 0) {
		TI_EM_OSAL_ERROR("failed to initialize wait event (errno=%d)\n", errno);
		return -1;
	} 

	TI_EM_OSAL_TRACE(2, "opening event for PDSP %d\n", pdsp);

	return 0;
}

int ti_em_osal_event_release(void)
{
	int core = ti_em_osal_get_core_idx();
	
	if (__ti_em_osal_event_fd[core] > 0) {
		int res;

		res = ioctl(__ti_em_osal_event_fd[core], EM_IOC_WAITEVENTFREE);
		if (res != 0) {
			TI_EM_OSAL_ERROR("failed to release wait event (errno=%d)\n", errno);
			return -1;
		} 
		close(__ti_em_osal_event_fd[core]);
	}
	return 0;
}

int ti_em_osal_event_wait(void)
{
	int core = ti_em_osal_get_core_idx();
	int res;

	TI_EM_OSAL_TRACE(3, "wait event\n");

	/* sleep until an incomming event is received */
	res = ioctl(__ti_em_osal_event_fd[core], EM_IOC_WAITEVENT, NULL);
	if (res != 0) {
		TI_EM_OSAL_ERROR("failed to wait event (errno=%d)\n", errno);
		return -1;
	}

	TI_EM_OSAL_TRACE(3, "get event\n");
	return 0;
}
