/*
 * Copyright (c) 2013, Texas Instruments Incorporated - http://www.ti.com/
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions are met:
 *       * Redistributions of source code must retain the above copyright
 *         notice, this list of conditions and the following disclaimer.
 *       * Redistributions in binary form must reproduce the above copyright
 *         notice, this list of conditions and the following disclaimer in the
 *         documentation and/or other materials provided with the distribution.
 *       * Neither the name of Texas Instruments Incorporated nor the
 *         names of its contributors may be used to endorse or promote products
 *         derived from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *   DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
 *   DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


/**
 *   @file  ti_em_osal_firmware.c
 *
 *   @brief   
 *      This is the OS abstraction layer for the firmware management.
 *
 */
#include <ti/runtime/openem/event_machine.h>

#include <ti_em_hw_functions.h>
#include <ti_em_osal.h>
#include <ti_em_osal_core.h>
#include <ti_em_osal_firmware.h>
#include <ti_em_osal_uaccess.h>

#define PDSP_HALT_TIMEOUT (1000000ULL)

static uint32_t __pdsp_ctrl_reg[QMSS_MAX_PDSP] = {
#ifdef DEVICE_K2K
	CSL_QMSS_CFG_PDSP1_REGS,
	CSL_QMSS_CFG_PDSP2_REGS,
	CSL_QMSS_CFG_PDSP3_REGS,
	CSL_QMSS_CFG_PDSP4_REGS,
	CSL_QMSS_CFG_PDSP5_REGS,
	CSL_QMSS_CFG_PDSP6_REGS,
	CSL_QMSS_CFG_PDSP7_REGS,
	CSL_QMSS_CFG_PDSP8_REGS
#else
	CSL_QM_SS_CFG_ADSP1_REGS,
	CSL_QM_SS_CFG_ADSP2_REGS
#endif
};

static uint32_t __pdsp_iram_reg[QMSS_MAX_PDSP] = {
#ifdef DEVICE_K2K
	CSL_QMSS_CFG_PDSP1_IRAM,
	CSL_QMSS_CFG_PDSP2_IRAM,
	CSL_QMSS_CFG_PDSP3_IRAM,
	CSL_QMSS_CFG_PDSP4_IRAM,
	CSL_QMSS_CFG_PDSP5_IRAM,
	CSL_QMSS_CFG_PDSP6_IRAM,
	CSL_QMSS_CFG_PDSP7_IRAM,
	CSL_QMSS_CFG_PDSP8_IRAM,
#else
	CSL_QM_SS_CFG_APDSP1_RAM_REGS,
	CSL_QM_SS_CFG_APDSP2_RAM_REGS
#endif
};

static uint32_t __pdsp_cmd_reg[QMSS_MAX_PDSP] = {
#ifdef DEVICE_K2K
	CSL_QMSS_CFG_PDSP1_SRAM,
	CSL_QMSS_CFG_PDSP2_SRAM,
	CSL_QMSS_CFG_PDSP3_SRAM,
	CSL_QMSS_CFG_PDSP4_SRAM,
	CSL_QMSS_CFG_PDSP5_SRAM,
	CSL_QMSS_CFG_PDSP6_SRAM,
	CSL_QMSS_CFG_PDSP7_SRAM,
	CSL_QMSS_CFG_PDSP8_SRAM,
#else
	CSL_QM_SS_CFG_SCRACH_RAM1_REGS,
	CSL_QM_SS_CFG_SCRACH_RAM2_REGS,
#endif
};

/*
 * Reset a given QMSS PDSP
 */
em_status_t ti_em_osal_pdsp_stop(short pdsp_id)
{
	CSL_PdspRegs      *pdsp_ctrl_reg = (CSL_PdspRegs *) ti_em_osal_io_phys_to_virt(__pdsp_ctrl_reg[pdsp_id]);
	volatile uint32_t  enable;
	uint64_t           start_time, current_time;

	TI_EM_OSAL_TRACE(2, "PDSP%d status is 0x%x\n", pdsp_id,
			 pdsp_ctrl_reg->PDSP_CONTROL_REG);
	
	/* Reset the PDSP */
	CSL_FINS (pdsp_ctrl_reg->PDSP_CONTROL_REG,
		  PDSP_PDSP_CONTROL_REG_PDSP_ENABLE, 0);
	
	/* Confirm PDSP has halted */
	start_time = ti_em_osal_timer_read(TI_EM_OSAL_TIMER_GLOBAL);
	do { 
		enable = CSL_FEXT (pdsp_ctrl_reg->PDSP_CONTROL_REG,
				   PDSP_PDSP_CONTROL_REG_PDSP_STATE);
		current_time = ti_em_osal_timer_read(TI_EM_OSAL_TIMER_GLOBAL);
	} while ((enable) && (current_time - start_time < PDSP_HALT_TIMEOUT));
	
	if (enable) {
		TI_EM_OSAL_ERROR("PDSP%d not halted (0x%x)!\n", pdsp_id,
				 pdsp_ctrl_reg->PDSP_CONTROL_REG);
		return EM_ERROR;
	}
		
	TI_EM_OSAL_TRACE(2, "PDSP%d halted (0x%x)\n", pdsp_id,
			 pdsp_ctrl_reg->PDSP_CONTROL_REG);
	
	return EM_OK;
}

/*
 * Reset, load firmware and enable a given QMSS PDSP
 */
em_status_t ti_em_osal_pdsp_download_firmware(
	short        pdsp_id,
	void*        firmware,
	unsigned int size)
{
	uint32_t           i, count;
	uint32_t          *data = (uint32_t *) firmware;
	CSL_PdspRegs      *pdsp_ctrl_reg = (CSL_PdspRegs *)      ti_em_osal_io_phys_to_virt(__pdsp_ctrl_reg[pdsp_id]);
	volatile uint32_t *pdsp_iram_reg = (volatile uint32_t *) ti_em_osal_io_phys_to_virt(__pdsp_iram_reg[pdsp_id]);
	volatile uint32_t *pdsp_cmd_reg  = (volatile uint32_t *) ti_em_osal_io_phys_to_virt(__pdsp_cmd_reg[pdsp_id]);

	count = size / 4;
	
	/* Download the firmware */
	TI_EM_OSAL_TRACE(2, "PDSP%d download firmware (count=%d, iram=0x%x)\n", pdsp_id,
			 count, (uint32_t) pdsp_iram_reg);
	
	for (i = 0; i < count; i++)
		pdsp_iram_reg[i] = data[i];
	
	/* Use the command register to sync the PDSP */
	*pdsp_cmd_reg = 0xFFFFFFFF;
	
	/* Wait to the memory write to land */
	for(i = 0; i < 20000 && *pdsp_cmd_reg != 0xFFFFFFFF; i++);
	
	/* Reset program counter to zero */
	CSL_FINS (pdsp_ctrl_reg->PDSP_CONTROL_REG,
		  PDSP_PDSP_CONTROL_REG_PCOUNTER_RST_VAL, 0);
	
	/* Set soft reset to zero to load the program counter */
	CSL_FINS (pdsp_ctrl_reg->PDSP_CONTROL_REG,
		  PDSP_PDSP_CONTROL_REG_SOFT_RST_N,
		  0);
	
	/* Enable the PDSP */
	CSL_FINS (pdsp_ctrl_reg->PDSP_CONTROL_REG,
		  PDSP_PDSP_CONTROL_REG_PDSP_ENABLE,
		  1);
	
	TI_EM_OSAL_TRACE(2, "PDSP%d wait command (0x%x)\n", pdsp_id, *pdsp_cmd_reg);
	
	/* Wait for the command register to clear */
	while (*pdsp_cmd_reg);
	
	return EM_OK;
}
