/*
 * Copyright (c) 2013, Texas Instruments Incorporated - http://www.ti.com/
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions are met:
 *       * Redistributions of source code must retain the above copyright
 *         notice, this list of conditions and the following disclaimer.
 *       * Redistributions in binary form must reproduce the above copyright
 *         notice, this list of conditions and the following disclaimer in the
 *         documentation and/or other materials provided with the distribution.
 *       * Neither the name of Texas Instruments Incorporated nor the
 *         names of its contributors may be used to endorse or promote products
 *         derived from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *   DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
 *   DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


/**
 *   @file  ti_em_osal_queue.c
 *
 *   @brief
 *      This is the OS abstraction layer for the QMSS management.
 *
 */
#include <ti/runtime/openem/event_machine.h>

#include <ti_em_hw_functions.h>
#include <ti_em_osal.h>
#include <ti_em_osal_core.h>
#include <ti_em_osal_queue.h>

em_status_t ti_em_osal_hw_queue_empty(int queue)
{
	TI_EM_OSAL_TRACE(4, "emptying queue %d\n", queue);
	ti_em_osal_hw_queue_push(queue, NULL, TI_EM_MEM_NONE);
        return (em_status_t) EM_OK;
}

em_status_t ti_em_osal_hw_queue_open(int queue)
{
	ti_em_osal_hw_queue_empty(queue);
	return (em_status_t) EM_OK;
}

em_status_t ti_em_osal_hw_queue_close(int queue)
{
	ti_em_osal_hw_queue_empty(queue);
	return (em_status_t) EM_OK;
}

/*
 * This function is performing insertion of QMSS region
 * Note that no check are performed on base addresss, start index for linking RAM, etc.
 */
em_status_t ti_em_osal_hw_queue_insert_region(ti_em_osal_hw_region_config_t *config, int qmss)
{
	CSL_Qm_descriptor_region_configRegs *qm_vaddr;
	uint32_t pow_reg_size;
	uint32_t reg;

	TI_EM_OSAL_TRACE(2, "inserting region %d to QMSS%d\n", config->region_idx, qmss);
	
	qm_vaddr = (void *) ti_em_osal_io_phys_to_virt(TI_EM_QMSS_CFG_DESCRIPTOR_REGS(qmss));

	TI_EM_OSAL_TRACE(3, "qm_vaddr = 0x%x\n", (uint32_t) qm_vaddr);

	qm_vaddr->MEMORY_REGION_BASE_ADDRESS_GROUP[config->region_idx].MEMORY_REGION_START_INDEX_REG
		= config->start_idx;
	
	TI_EM_OSAL_TRACE(2, "start index = 0x%x\n", config->start_idx);
	
	qm_vaddr->MEMORY_REGION_BASE_ADDRESS_GROUP[config->region_idx].MEMORY_REGION_BASE_ADDRESS_REG
		= config->desc_base;

	TI_EM_OSAL_TRACE(2, "desc base = 0x%x\n", config->desc_base);
	
	for (pow_reg_size = 0;
	     (32UL << pow_reg_size) < config->desc_num;
	     pow_reg_size++);
	
	reg = ((((config->desc_size / 16) - 1)) << 16) | pow_reg_size;
	qm_vaddr->MEMORY_REGION_BASE_ADDRESS_GROUP[config->region_idx].MEMORY_REGION_DESCRIPTOR_SETUP_REG
		= reg;
	
	TI_EM_OSAL_TRACE(2, "start desc_size = 0x%x, desc_num = 0x%x, reg = 0x%x\n",
		    config->desc_size, config->desc_num, reg);

        return (em_status_t) EM_OK;
}

unsigned int ti_em_osal_hw_queue_get_memory_region_addr(short memory_idx, int qmss)
{
	CSL_Qm_descriptor_region_configRegs * qm_vaddr;

	qm_vaddr = (void *) ti_em_osal_io_phys_to_virt(TI_EM_QMSS_CFG_DESCRIPTOR_REGS(qmss));
	return qm_vaddr->MEMORY_REGION_BASE_ADDRESS_GROUP[memory_idx].MEMORY_REGION_BASE_ADDRESS_REG;
}

void ti_em_osal_hw_queue_push(int queue, void *desc, short mapping)
{
	CSL_Qm_queue_managementQueue_mgmt_groupRegs *qm_vaddr;

	qm_vaddr = (void *) ti_em_osal_io_phys_to_virt(TI_EM_QMSS_CFG_QUEUE_MGT_REGS);

	qm_vaddr[queue].QUEUE_REG_D = ti_em_osal_queue_desc_vtop(desc, mapping);

	TI_EM_OSAL_TRACE(4, "pushing event 0x%x (paddr=0x%x, mapping %d) to queue %d (0x%x)\n",
			 (uint32_t) desc, ti_em_osal_queue_desc_vtop(desc, mapping),
			 mapping, queue,
			 (uint32_t)(&qm_vaddr[queue].QUEUE_REG_D));
}

void ti_em_osal_hw_queue_push_to_head(int queue, void *desc, short mapping)
{
	CSL_Qm_queue_managementQueue_mgmt_groupRegs *qm_vaddr;
	uint64_t dword;
	uint32_t pdesc = ti_em_osal_queue_desc_vtop(desc, mapping);

	qm_vaddr = (void *) ti_em_osal_io_phys_to_virt(TI_EM_QMSS_CFG_DEQUEUE_REGS);

	/* build double word */
#ifdef _BIG_ENDIAN
	dword = _itoll ((uint32_t) (Qmss_Location_HEAD << 31), pdesc);
#else
	dword = _itoll (pdesc, (uint32_t) (Qmss_Location_HEAD << 31));
#endif
	/* push double word to QMSS */
	*(uint64_t *)(&qm_vaddr[queue].QUEUE_REG_C) = dword;
	
	TI_EM_OSAL_TRACE(3, "pushing event 0x%x to queue head %d (mapping %d)\n",
			 (uint32_t)desc, queue, mapping)
	return;
}

void* ti_em_osal_hw_queue_pop(int queue, short mapping)
{
	CSL_Qm_queue_managementQueue_mgmt_groupRegs *qm_vaddr;
	uint32_t desc_paddr;

	qm_vaddr = (void *) ti_em_osal_io_phys_to_virt(TI_EM_QMSS_CFG_QUEUE_MGT_REGS);

	desc_paddr = qm_vaddr[queue].QUEUE_REG_D;

	TI_EM_OSAL_TRACE(4, "poping event 0x%x from queue %d mapping %d (@0x%x)\n",
			 (uint32_t) desc_paddr, queue, mapping,
			 (uint32_t) &qm_vaddr[queue].QUEUE_REG_D);
	
	return ti_em_osal_queue_desc_ptov(desc_paddr, mapping);
}

void ti_em_osal_hw_queue_set_threshold(int queue, short hilo, char threshold)
{
	CSL_Qm_queue_status_configRegs *qm_vaddr;
	
	qm_vaddr = (void *) ti_em_osal_io_phys_to_virt(TI_EM_QMSS_CFG_QUEUE_PEEK_REGS);
	qm_vaddr->QUEUE_STATUS_CONFIG_GROUP[queue].QUEUE_STATUS_CONFIG_REG_D =
		(threshold | (hilo << CSL_QM_QUEUE_STATUS_CONFIG_QUEUE_STATUS_CONFIG_REG_D_THRESHOLD_HILO_SHIFT));
	return;
}

void ti_em_osal_hw_queue_divert(int src_queue, int dst_queue, char location)
{
	uint32_t temp = 0;
	CSL_Qm_configRegs *qm_vaddr;
	
	qm_vaddr = (void *) ti_em_osal_io_phys_to_virt(TI_EM_QMSS_CFG_STARVATION_REGS(ti_em_qmss_get(src_queue)));

	CSL_FINS (temp, QM_CONFIG_QUEUE_DIVERSION_REG_SOURCE_QNUM, src_queue);
	CSL_FINS (temp, QM_CONFIG_QUEUE_DIVERSION_REG_DEST_QNUM, dst_queue);
	CSL_FINS (temp, QM_CONFIG_QUEUE_DIVERSION_REG_HEAD_TAIL, location);
	
	qm_vaddr->QUEUE_DIVERSION_REG = temp;
}

unsigned int ti_em_osal_hw_queue_peek(int queue)
{
	CSL_Qm_queue_status_configRegs *qm_vaddr;
	
	qm_vaddr = (void *) ti_em_osal_io_phys_to_virt(TI_EM_QMSS_CFG_QUEUE_PEEK_REGS);
	
	return qm_vaddr->QUEUE_STATUS_CONFIG_GROUP[queue].QUEUE_STATUS_CONFIG_REG_A;
}
