/*
 * Copyright (c) 2013, Texas Instruments Incorporated - http://www.ti.com/
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions are met:
 *       * Redistributions of source code must retain the above copyright
 *         notice, this list of conditions and the following disclaimer.
 *       * Redistributions in binary form must reproduce the above copyright
 *         notice, this list of conditions and the following disclaimer in the
 *         documentation and/or other materials provided with the distribution.
 *       * Neither the name of Texas Instruments Incorporated nor the
 *         names of its contributors may be used to endorse or promote products
 *         derived from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *   DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
 *   DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


/**
 *   @file  ti_em_osal_shm.c
 *
 *   @brief   
 *      This is the OS abstraction layer for the shared memory management
 *
 */
#include <ti_em_osal.h>
#include <ti_em_osal_core.h>
#include <ti_em_osal_shm.h>

#ifdef TI_EM_THREAD_MODEL
#include <stdlib.h>

int ti_em_osal_shm_init(int id, char *shm_name, unsigned int size, short create)
{
	return 0;
}

void ti_em_osal_shm_release(int id)
{
}

void* ti_em_osal_shm_alloc(int id, unsigned int size)
{
	void *ptr = malloc((size_t) size);

	if (ptr)
		memset(ptr, 0, size);

	return ptr;
}

void ti_em_osal_shm_free(void *ptr)
{
	free(ptr);
}

#else /* TI_EM_THREAD_MODEL */
#include <sys/mman.h>
#include <sys/file.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>

#include <ti_em_hw_mach.h>

struct __ti_em_osal_shm_mapping_desc {
	int          fd;
	char         name[64];
	void        *base;
	void        *end;
	void        *current;
};

static struct __ti_em_osal_shm_mapping_desc __ti_em_osal_shm_mapping[TI_EM_OSAL_SHM_MAPPING_NUM];

int ti_em_osal_shm_init(int id, char *shm_name, unsigned int size, short create)
{
	int fd;
	int res;
	int flags = create ? O_CREAT | O_TRUNC | O_RDWR : O_RDWR; 

	if ((id >= TI_EM_OSAL_SHM_MAPPING_NUM) || (id < 0)) {
		TI_EM_OSAL_ERROR("invalid shm id\n");
		return -1;
	}

	/* Align size to page size + 1 page guard for alignment */
	size = PAGE_ALIGN_UP(size) + PAGE_SIZE;

	TI_EM_OSAL_TRACE(3, "initializing shared memory %s (id=%d), size=%d\n", shm_name, id, size);

	if (!shm_name)
		shm_name = "openem_anonymous";

	memset(&__ti_em_osal_shm_mapping[id], 0, sizeof(struct __ti_em_osal_shm_mapping_desc));

	fd = shm_open(shm_name, flags, 0666);
	if (fd == -1) {
		TI_EM_OSAL_ERROR("cannot create shared memory (err=%d)\n", errno);
		return -1;
	}

	if (create) {
		res = ftruncate(fd, size);
		if (res != 0) {
			TI_EM_OSAL_ERROR("invalid shared memory size (err=%d)\n", errno);
			return -1;
		}
	}

	__ti_em_osal_shm_mapping[id].base = mmap(0, 
						 size, 
						 PROT_READ | PROT_WRITE,
						 MAP_SHARED, 
						 fd,
						 0);

	if (__ti_em_osal_shm_mapping[id].base == MAP_FAILED) {
		TI_EM_OSAL_ERROR("shared memory mapping failed (err=%d)\n", errno);
		return -1;
	}

	if (create)
		memset(__ti_em_osal_shm_mapping[id].base, 0, size);

	strcpy(__ti_em_osal_shm_mapping[id].name, shm_name);
	
	__ti_em_osal_shm_mapping[id].fd      = fd;
	__ti_em_osal_shm_mapping[id].end     = 
		(void *)((unsigned int) __ti_em_osal_shm_mapping[id].base + size);
	__ti_em_osal_shm_mapping[id].current = __ti_em_osal_shm_mapping[id].base;
	
	TI_EM_OSAL_TRACE(2, "shared memory %s (id=%d) start=%p, end=%p\n", 
		    shm_name, id, __ti_em_osal_shm_mapping[id].base, __ti_em_osal_shm_mapping[id].end);

	return 0;
}

void ti_em_osal_shm_release(int id)
{
	unsigned int size = ((unsigned int) __ti_em_osal_shm_mapping[id].end
			     - (unsigned int) __ti_em_osal_shm_mapping[id].base);

	munmap(__ti_em_osal_shm_mapping[id].base, size);
	shm_unlink(__ti_em_osal_shm_mapping[id].name);
}

void* ti_em_osal_shm_alloc(int id, unsigned int size)
{
	void *ptr = __ti_em_osal_shm_mapping[id].current;
	
	/* Align size to cache line */
	size += CACHE_L2_LINESIZE;
	size &= ~(CACHE_L2_LINESIZE - 1);

	if (((unsigned int) ptr + size) <
	    ((unsigned int) __ti_em_osal_shm_mapping[id].end)) {
		__ti_em_osal_shm_mapping[id].current += size;
		TI_EM_OSAL_ASSERT(((unsigned int) ptr & (CACHE_L2_LINESIZE - 1)) == 0);
		return ptr;
	}
	return NULL;
}

void ti_em_osal_shm_free(void *ptr)
{
	/* NOT IMPLEMENTED */
}
#endif /* TI_EM_THREAD_MODEL */
