/*
 * Copyright (c) 2013, Texas Instruments Incorporated - http://www.ti.com/
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions are met:
 *       * Redistributions of source code must retain the above copyright
 *         notice, this list of conditions and the following disclaimer.
 *       * Redistributions in binary form must reproduce the above copyright
 *         notice, this list of conditions and the following disclaimer in the
 *         documentation and/or other materials provided with the distribution.
 *       * Neither the name of Texas Instruments Incorporated nor the
 *         names of its contributors may be used to endorse or promote products
 *         derived from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *   DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
 *   DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


/**
 *   @file  ti_em_osal_uaccess.c
 *
 *   @brief   
 *      This is the OS abstraction layer for the I/O accesses from user mode
 *
 */
#include <ti/runtime/openem/event_machine.h>

#include <ti_em_osal.h>
#include <ti_em_osal_core.h>
#include <ti_em_osal_shm.h>
#include <ti_em_osal_uaccess.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <fcntl.h>

#ifdef TI_EM_USE_HUGETLB
#define TI_EM_OSAL_MMAP_FLAGS (MAP_POPULATE | MAP_LOCKED | MAP_HUGETLB)
#else
#define TI_EM_OSAL_MMAP_FLAGS (MAP_POPULATE | MAP_LOCKED)
#endif

static int __ti_em_osal_mapping_fd          = -1;
static int __ti_em_osal_mapping_fd_uncached = -1;

/* Base address (physical) of the PDSP <-> CPU communication memory */
phys_addr_t __ti_em_osal_pdsp_comm_mem_base = 0;

/* 
 * Local table that contains all the potential mappings for user access
 * to physical mem & I/O.
 */
struct ti_em_osal_mem_mapping __ti_em_osal_mem_mappings[TI_EM_MEM_NUM];

#ifndef TI_EM_THREAD_MODEL

#define TI_EM_OSAL_MAPPING_FREE                0
#define TI_EM_OSAL_MAPPING_IN_USE              1

struct ti_em_osal_mem_shared_db_mapping {
	phys_addr_t  paddr;
	unsigned int size;
	off_t        offset;
	unsigned int flags;
	int          status;
};

struct ti_em_osal_mem_shared_db_mapping *__ti_em_osal_mem_mapping_db = NULL;

static int __register_mapping_db(
	short                          mapping,
	struct ti_em_osal_mem_mapping *my_mapping)
{
	struct ti_em_osal_mem_shared_db_mapping *mapping_db;

	TI_EM_OSAL_ASSERT(mapping < TI_EM_MEM_NUM);

	/* Check if the global mapping database has been initialized */
	if (__ti_em_osal_mem_mapping_db == NULL) {
		unsigned int size = 
			sizeof(struct ti_em_osal_mem_shared_db_mapping) * TI_EM_MEM_NUM;

		/* Allocate the associated shared memory region */
		ti_em_osal_shm_init(TI_EM_OSAL_SHM_OSAL_MAPPING_DB_ID, 
				    TI_EM_OSAL_SHM_OSAL_MAPPING_DB_NAME,
				    size, 
				    1);

		__ti_em_osal_mem_mapping_db = (struct ti_em_osal_mem_shared_db_mapping *)
			ti_em_osal_shm_alloc(TI_EM_OSAL_SHM_OSAL_MAPPING_DB_ID,
					     sizeof(struct ti_em_osal_mem_shared_db_mapping));

		if (__ti_em_osal_mem_mapping_db == NULL) {
			TI_EM_OSAL_ERROR("cannot allocated mapping database\n");
			return -1;
		}
	}

	mapping_db = __ti_em_osal_mem_mapping_db + mapping;

	/*  If the mapping is already registered, do not update it */
	if (mapping_db->status != TI_EM_OSAL_MAPPING_FREE)
		return -1;

	/* Register the mapping */
	mapping_db->status = TI_EM_OSAL_MAPPING_IN_USE;
	mapping_db->paddr  = my_mapping->paddr;
	mapping_db->size   = my_mapping->size;
	mapping_db->offset = my_mapping->offset;
	mapping_db->flags  = my_mapping->flags;

	return 0;
}

static int __unregister_mapping_db(
	short                          mapping,
	struct ti_em_osal_mem_mapping *my_mapping)
{
	struct ti_em_osal_mem_shared_db_mapping *mapping_db;

	if (__ti_em_osal_mem_mapping_db == NULL)
		return -1;

	mapping_db = __ti_em_osal_mem_mapping_db + mapping;

	/* Unregister the mapping */
	mapping_db->paddr  = 0;
	mapping_db->size   = 0;
	mapping_db->offset = 0;
	mapping_db->flags  = 0;
	mapping_db->status = TI_EM_OSAL_MAPPING_FREE;

	return 0;
}

/* 
 * Special function for retrieving an existing mapping from the shared mapping db 
 * This is used in process mode only for the "join" operation.
 */
int ti_em_osal_retrieve_mapping(
	short                          mapping,
	struct ti_em_osal_mem_mapping *mapping_info)
{
	struct ti_em_osal_mem_shared_db_mapping *mapping_db;
	
	TI_EM_OSAL_ASSERT(mapping_info);

	if (__ti_em_osal_mem_mapping_db == NULL) {
		unsigned int size = 
			sizeof(struct ti_em_osal_mem_shared_db_mapping) * TI_EM_MEM_NUM;
		
		/* Retrieve the associated shared memory region */
		ti_em_osal_shm_init(TI_EM_OSAL_SHM_OSAL_MAPPING_DB_ID, 
				    TI_EM_OSAL_SHM_OSAL_MAPPING_DB_NAME,
				    size, 
				    0);
		
		__ti_em_osal_mem_mapping_db = (struct ti_em_osal_mem_shared_db_mapping *)
			ti_em_osal_shm_alloc(TI_EM_OSAL_SHM_OSAL_MAPPING_DB_ID, size);
		
		if (__ti_em_osal_mem_mapping_db == NULL) {
			TI_EM_OSAL_ERROR("cannot retrieve shared memory %s\n",
				TI_EM_OSAL_SHM_OSAL_MAPPING_DB_NAME);
			return -1;
		}
	}

	mapping_db = __ti_em_osal_mem_mapping_db + mapping;

	if (mapping_db->status != TI_EM_OSAL_MAPPING_IN_USE)
		return 0;

	/* Retrieve the mapping */
	mapping_info->paddr  = mapping_db->paddr;
	mapping_info->size   = mapping_db->size;
	mapping_info->offset = mapping_db->offset;
	mapping_info->flags  = mapping_db->flags;

	return 1;
}
#endif /* !TI_EM_THREAD_MODEL */

static int __get_fd_from_flags(int flags)
{
	int  oflags;
	int *fd_ptr;

	if (flags & TI_EM_OSAL_MEM_UNCACHED) {
		fd_ptr = &__ti_em_osal_mapping_fd_uncached;
		oflags = O_DSYNC;
	} else {
		fd_ptr = &__ti_em_osal_mapping_fd;
		oflags = 0;
	}
	
	if (*fd_ptr == -1)
		*fd_ptr = open(TI_EM_OSAL_MEM_UACCESS_DEVICE, O_RDWR | oflags);
	
	return *fd_ptr;
}

/*
 * This function creates an user mapping of physical memory
 *
 * If vaddr field is zero, then we create dynamically the mapping based
 * on paddr and size otherwise we consider that the mapping has already be
 * performed outside (MMU case) or is not needed (MMU-less case).
 * 
 * If fd is provided use this file descriptor for mmap, other use default
 * one.
 */
int ti_em_osal_create_mapping(
	short                          mapping,
	struct ti_em_osal_mem_mapping *mapping_info)
{
	struct ti_em_osal_mem_mapping *my_mapping;
	void                          *vmap_base;
	unsigned int                   size;
	unsigned int                   mask;
	off_t                          cpu_paddr = mapping_info->paddr;

	if ((mapping < 0) || (mapping >= TI_EM_MEM_NUM) || (!mapping_info))
		return -1;

	if (mapping_info->size == 0)
		return -1;

	/* align size to page size */
	size = PAGE_ALIGN_UP(mapping_info->size);
	
	my_mapping = &__ti_em_osal_mem_mappings[mapping];
	my_mapping->size = size;
	mask = size -1;

	/* if virtual address is set, only register the mapping */
	if (mapping_info->vaddr) {
		*my_mapping = *mapping_info;
		goto end;
	}

	my_mapping->paddr  = mapping_info->paddr;
	my_mapping->offset = mapping_info->offset;
	my_mapping->flags  = mapping_info->flags;

	/* retrieve file descriptor if either opened with O_DYNC or without */
	if (!mapping_info->fd)
		my_mapping->fd = __get_fd_from_flags(my_mapping->flags);
	else
		my_mapping->fd = mapping_info->fd;
			
	if (my_mapping->fd == -1) {
		TI_EM_OSAL_ERROR("cannot open %s device!\n",
				 TI_EM_OSAL_MEM_UACCESS_DEVICE);
		return -1;
	}

	cpu_paddr = (off_t) my_mapping->paddr + my_mapping->offset;

	/* perform mapping */
	vmap_base = mmap(0, 
			 size,
			 PROT_READ | PROT_WRITE,
			 MAP_SHARED | TI_EM_OSAL_MMAP_FLAGS, 
			 my_mapping->fd,
			 cpu_paddr & ~mask);

	if(vmap_base == (void *) -1) {
		TI_EM_OSAL_ERROR("failed to mmap %s for 0x%x (size=0x%x)!\n",
				 TI_EM_OSAL_MEM_UACCESS_DEVICE, my_mapping->paddr,
				 size);
		my_mapping->vaddr = 0;
		return -1;
	}
	
	my_mapping->vaddr = (unsigned int)
		vmap_base + ((off_t)my_mapping->paddr & mask);

	mapping_info->vaddr = my_mapping->vaddr;
end:

#ifndef TI_EM_THREAD_MODEL
	(void) __register_mapping_db(mapping, my_mapping);
#endif

	TI_EM_OSAL_TRACE(2, "mapping %d: vaddr=0x%08x paddr=0x%08x cpu_paddr=0x%08x size=0x%08x %s\n",
			 mapping, my_mapping->vaddr, my_mapping->paddr, (unsigned int) cpu_paddr, my_mapping->size,
			 !my_mapping->fd ? "(fixed)": "(dynamic)");

	return 0;
}

/*
 * This function releases an user mapping of physical memory
 */
void ti_em_osal_release_mapping(short mapping)
{
	struct ti_em_osal_mem_mapping *my_mapping;

	if ((mapping < 0) || (mapping >= TI_EM_MEM_NUM))
		return;

	my_mapping = &__ti_em_osal_mem_mappings[mapping];

	if (my_mapping->size && my_mapping->vaddr)
		(void) munmap((void*) my_mapping->vaddr, my_mapping->size);

	if (my_mapping->fd)
		close(my_mapping->fd);
	
#ifndef TI_EM_THREAD_MODEL
	(void) __unregister_mapping_db(mapping, my_mapping);
#endif

	memset(my_mapping, 0, sizeof(*my_mapping));
}
