/*
 * Copyright (c) 2013, Texas Instruments Incorporated - http://www.ti.com/
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions are met:
 *       * Redistributions of source code must retain the above copyright
 *         notice, this list of conditions and the following disclaimer.
 *       * Redistributions in binary form must reproduce the above copyright
 *         notice, this list of conditions and the following disclaimer in the
 *         documentation and/or other materials provided with the distribution.
 *       * Neither the name of Texas Instruments Incorporated nor the
 *         names of its contributors may be used to endorse or promote products
 *         derived from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *   DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
 *   DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


/**
 *   @file  ti_em_osal_xge.c
 *
 *   @brief
 *      This is the OS abstraction layer for the 10Gig Ethernet management.
 *
 */
#include <ti/runtime/openem/event_machine.h>

#include <ti_em_hw_functions.h>
#include <ti_em_osal.h>
#include <ti_em_osal_core.h>
#include <ti_em_osal_queue.h>
#include <ti_em_osal_cppi.h>

static ti_em_queue_id_t __rx_flow_dest_queue[TI_EM_XGE_FLOW_NUM];

em_status_t ti_em_osal_xge_rx_flow_open(int dma_idx,
					int flow_idx,
					int dst_queue,
					int free_queue0,
					int free_queue1,
					int error_handling)
{
	Cppi_RxFlowCfg rx_flow_cfg;
	int            res;

       	memset ((void *) &rx_flow_cfg, 0, sizeof (Cppi_RxFlowCfg));

	/* Save the real destination queue for this flow */
	__rx_flow_dest_queue[flow_idx] = (ti_em_queue_id_t) dst_queue;

	rx_flow_cfg.flowIdNum          = TI_EM_XGE_FLOW_IDX_START + flow_idx;

#ifdef TI_EM_XGE_LOOPBACK
	/* We need to use a temporary queue in order to append the missing CRC field */
	rx_flow_cfg.rx_dest_qnum       = ((TI_EM_XGE_TMP_RX_QUEUE + flow_idx) & 0xFFF);
	rx_flow_cfg.rx_dest_qmgr       = ((TI_EM_XGE_TMP_RX_QUEUE + flow_idx) >> 12);
#else /* TI_EM_XGE_LOOPBACK */
	rx_flow_cfg.rx_dest_qnum       = (dst_queue & 0xFFF);
	rx_flow_cfg.rx_dest_qmgr       = (dst_queue >> 12);
#endif /* TI_EM_XGE_LOOPBACK */
	rx_flow_cfg.rx_desc_type       = Cppi_DescType_HOST;

	/* Mandatory parameters to configure the RX FLOW */
	rx_flow_cfg.rx_einfo_present   = 1;
	rx_flow_cfg.rx_dest_tag_lo_sel = 4;
	rx_flow_cfg.rx_dest_tag_hi_sel = 5;
	rx_flow_cfg.rx_src_tag_hi_sel  = 4;
	rx_flow_cfg.rx_psinfo_present  = 1;
	rx_flow_cfg.rx_ps_location     = 0;
	rx_flow_cfg.rx_error_handling  = error_handling;
	rx_flow_cfg.rx_fdq0_sz0_qnum   = (free_queue0 & 0xFFF);
	rx_flow_cfg.rx_fdq0_sz0_qmgr   = (free_queue0 >> 12);
	rx_flow_cfg.rx_fdq0_sz1_qnum   = (free_queue0 & 0xFFF);
	rx_flow_cfg.rx_fdq0_sz1_qmgr   = (free_queue0 >> 12);
	rx_flow_cfg.rx_fdq0_sz2_qnum   = (free_queue0 & 0xFFF);
	rx_flow_cfg.rx_fdq0_sz2_qmgr   = (free_queue0 >> 12);
	rx_flow_cfg.rx_fdq0_sz3_qnum   = (free_queue0 & 0xFFF);
	rx_flow_cfg.rx_fdq0_sz3_qmgr   = (free_queue0 >> 12);
	rx_flow_cfg.rx_fdq1_qnum       = (free_queue1 & 0xFFF);
	rx_flow_cfg.rx_fdq1_qmgr       = (free_queue1 >> 12);
	rx_flow_cfg.rx_fdq2_qnum       = (free_queue1 & 0xFFF);
	rx_flow_cfg.rx_fdq2_qmgr       = (free_queue1 >> 12);
	rx_flow_cfg.rx_fdq3_qnum       = (free_queue1 & 0xFFF);
	rx_flow_cfg.rx_fdq3_qmgr       = (free_queue1 >> 12);

	/* Other fields are left to be modified by user */

	TI_EM_OSAL_TRACE(2, "opening XGE flow %d for dma %d dst_queue %d free_queue0 %d free_queue1 %d\n",
			 TI_EM_XGE_FLOW_IDX_START + flow_idx, dma_idx, 
			 dst_queue, free_queue0, free_queue1);

	/* Configure Rx flow */
	res = ti_em_cppi_configure_rx_flow((Cppi_Handle) dma_idx,
					   &ti_em_osal_cppi_global_cfg_params[0],
					   &rx_flow_cfg);
	if (res < 0)
		return (em_status_t) EM_ERR;
	
        return (em_status_t) EM_OK;
}

em_status_t ti_em_osal_xge_tx_channel_open(int dma_idx, int channel_idx)
{
	TI_EM_OSAL_TRACE(2, "opening XGE Tx channel %d for dma %d\n", 
			 TI_EM_XGE_FLOW_IDX_START + channel_idx, dma_idx);

	return ti_em_osal_cppi_tx_channel_open(dma_idx, 
					       TI_EM_XGE_FLOW_IDX_START + channel_idx);
}

em_status_t ti_em_osal_xge_rx_channel_open(int dma_idx, int channel_idx)
{
	TI_EM_OSAL_TRACE(2, "opening XGE Rx channel %d for dma %d\n", 
			 TI_EM_XGE_FLOW_IDX_START + channel_idx, dma_idx);

	return ti_em_osal_cppi_rx_channel_open(dma_idx,
					       TI_EM_XGE_FLOW_IDX_START + channel_idx);
}

ti_em_queue_id_t ti_em_osal_xge_tx_queue_open(ti_em_queue_id_t queue_base_idx, int vlan_priority)
{
	em_status_t status;

	/* open xge Tx DMA Tx queue */
	TI_EM_OSAL_TRACE(2, "opening XGE queue %d for vlan prio %d\n", 
			 queue_base_idx + TI_EM_XGE_FLOW_IDX_START + vlan_priority,
			 vlan_priority);

	status = ti_em_osal_hw_queue_open(queue_base_idx + TI_EM_XGE_FLOW_IDX_START + vlan_priority);
	if (status != EM_OK)
		return EM_ERR_ALLOC_FAILED;

	return (ti_em_queue_id_t) (queue_base_idx + TI_EM_XGE_FLOW_IDX_START + vlan_priority);
}

em_status_t ti_em_osal_xge_rx_flow_close(int dma_idx, int flow_idx)
{
	return EM_OK;
}

em_status_t ti_em_osal_xge_rx_channel_close(int dma_idx, int channel_idx)
{
	return EM_OK;
}

em_status_t ti_em_osal_xge_tx_channel_close(int dma_idx, int channel_idx)
{
	return EM_OK;
}

ti_em_queue_id_t ti_em_osal_xge_tx_queue_base_idx_get(void)
{
	TI_EM_OSAL_TRACE(3, "queue base index is %d\n", TI_EM_XGE_TX_QUEUE_BASE + TI_EM_XGE_FLOW_IDX_START);
	return (ti_em_queue_id_t) (TI_EM_XGE_TX_QUEUE_BASE + TI_EM_XGE_FLOW_IDX_START);
}

ti_em_queue_id_t ti_em_osal_xge_rx_queue_idx_get(int flow_idx)
{
	TI_EM_OSAL_TRACE(3, "dst queue base index is %d\n", __rx_flow_dest_queue[flow_idx]);
	return __rx_flow_dest_queue[flow_idx];
}

#ifdef TI_EM_XGE_LOOPBACK
/* 
 * This function is used to add a dummy CRC field in order to emulate what the real 
 * XGE hw does.
 */
em_status_t ti_em_osal_xge_rx_insert_crc(int              flow_idx,
					 ti_em_queue_id_t rx_fragment_free_queue,
					 ti_em_queue_id_t xge_dst_queue,
					 int              count)
{
	int            event_count = 0;
	uint32_t       packet_size;
	Cppi_HostDesc *desc;
        Cppi_HostDesc *next_desc;
        Cppi_HostDesc *last_desc;
        Cppi_HostDesc *crc_desc;

	TI_EM_OSAL_TRACE(3, "using flow_idx %d tmp rx queue %d, real rx queue %d, fragment queue %d, count %d\n",
			 TI_EM_XGE_FLOW_IDX_START + flow_idx,
			 TI_EM_XGE_TMP_RX_QUEUE + flow_idx,
			 xge_dst_queue,
			 rx_fragment_free_queue,
			 count);

	if (xge_dst_queue == 0) {
		TI_EM_OSAL_ERROR("XGE Rx queue not initialized\n");
		return EM_ERR;
	}

        /* Waiting for incoming events in the temporary XGE rx queue */
        while(event_count != count) {
		desc = (Cppi_HostDesc *)
			((uint32_t) ti_em_osal_hw_queue_pop(TI_EM_XGE_TMP_RX_QUEUE + flow_idx,
							    TI_EM_MEM_PUBLIC_DESC) & ~0xf);

		if (desc != NULL) {
			TI_EM_OSAL_TRACE(4, "getting desc 0x%x, event_count %d from queue %d\n",
					 (uint32_t) desc, event_count, TI_EM_XGE_TMP_RX_QUEUE + flow_idx);
			event_count++;
			
			/* Get the latest descriptor in the current packet */
			last_desc = desc;
			next_desc = (Cppi_HostDesc *) 
				ti_em_osal_queue_desc_ptov(last_desc->nextBDPtr, TI_EM_MEM_PUBLIC_DESC);
			
			while(next_desc != NULL) {
				last_desc = next_desc;
				next_desc = (Cppi_HostDesc *)
					ti_em_osal_queue_desc_ptov(last_desc->nextBDPtr, TI_EM_MEM_PUBLIC_DESC);
			}
			
			/* Pop a new descriptor from the rx fragment free queue */
			crc_desc = (Cppi_HostDesc *)
				((uint32_t) ti_em_osal_hw_queue_pop(rx_fragment_free_queue,
								    TI_EM_MEM_PUBLIC_DESC) & ~0xf);
			if(crc_desc == NULL){
				TI_EM_OSAL_ERROR("no more free receive fragment descriptors for CRC\n");
				return EM_ERR;
			}
			
			/* Set buffer length to CRC size (4 Bytes) */
			crc_desc->buffLen = TI_EM_XGE_CRC_SIZE;
			
			/* Link CRC descriptor to the packet */
			last_desc->nextBDPtr = ti_em_osal_queue_desc_vtop(crc_desc,
									  TI_EM_MEM_PUBLIC_DESC);
			/* Adjust packet size */
			packet_size    = (desc->descInfo & 0x003fffff) + TI_EM_XGE_CRC_SIZE;
			desc->descInfo = (desc->descInfo & 0xffc00000) + packet_size;
			
			/* Push the descriptor to the original XGE rx queue */
			ti_em_osal_hw_queue_push(xge_dst_queue, desc, TI_EM_MEM_PUBLIC_DESC);
		}
        }
	return (em_status_t) EM_OK;
}
#endif /* TI_EM_XGE_LOOPBACK */
