/*
 * Copyright (c) 2013, Texas Instruments Incorporated - http://www.ti.com/
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions are met:
 *       * Redistributions of source code must retain the above copyright
 *         notice, this list of conditions and the following disclaimer.
 *       * Redistributions in binary form must reproduce the above copyright
 *         notice, this list of conditions and the following disclaimer in the
 *         documentation and/or other materials provided with the distribution.
 *       * Neither the name of Texas Instruments Incorporated nor the
 *         names of its contributors may be used to endorse or promote products
 *         derived from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *   DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
 *   DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/**
 *   @file  ti_em_osal_cma.h
 *
 *   @brief
 *      This is the OS abstraction layer for basic services (log, timers, ...).
 *
 */

#ifndef TI_EM_OSAL_H_
#define TI_EM_OSAL_H_

#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <sched.h>
#include <signal.h>

#include <ti_em_hw_functions.h>

/*
 * Trace, error and assert logging macros
 */
#define __TI_EM_OSAL_PRINTF(tl, type, fmt, args...)			\
    do{ if ((tl) <= TI_EM_TRACE_LEVEL) printf("EM[%s][C%d] %s(): " fmt, type, ti_em_osal_get_core_idx(), __FUNCTION__ , ## args); } while(0);
#define TI_EM_OSAL_ERROR(fmt, args...)      __TI_EM_OSAL_PRINTF(1,  "error", fmt, ## args)
#define TI_EM_OSAL_TRACE(tl, fmt, args...)  __TI_EM_OSAL_PRINTF(tl, "trace", fmt, ## args)
#define TI_EM_OSAL_PRINTF(fmt, args...)     printf(fmt, ## args)

#if (EM_CHECK_LEVEL >= 1)
#define TI_EM_OSAL_ASSERT(cond)             if (!(cond)) TI_EM_OSAL_TRACE(0, "ASSERT %s FAILURE\n", # cond)
#else
#define TI_EM_OSAL_ASSERT(cond)
#endif

#define PAGE_ALIGN_UP(x)     (((x) + (PAGE_SIZE - 1)) & ~(PAGE_SIZE - 1))
#define PAGE_ALIGN_LOW(x)    (((x) & ~(PAGE_SIZE - 1)))

/*
 * Functions to map OS basic services
 */
#define TI_EM_OSAL_TIMER_LOCAL    (0)
#define TI_EM_OSAL_TIMER_GLOBAL   (1)

static inline int ti_em_osal_timer_init(int type)
{
	if (type ==  TI_EM_OSAL_TIMER_LOCAL) {
#ifdef TI_EM_HW_HAS_FAST_TSC
		return TI_EM_HW_INIT_TSC();
#endif
	}
	return 0;
}

/*
 * Retrieve current time:
 * - local time is in cycles (direct core timestamp counter)
 * - global time is in nanoseconds (Linux syscall) but is consistent accross cores
 */
static inline uint64_t ti_em_osal_timer_read(int type)
{
	struct timespec ts;
	
#ifdef TI_EM_HW_HAS_FAST_TSC
	if (type == TI_EM_OSAL_TIMER_LOCAL)
		return TI_EM_HW_GET_TSC();
#endif	
	(void) clock_gettime(type == TI_EM_OSAL_TIMER_LOCAL ?
			     CLOCK_PROCESS_CPUTIME_ID : CLOCK_MONOTONIC, &ts);

	return (((uint64_t) ts.tv_sec * 1000000000ULL) + (uint64_t) ts.tv_nsec); 
}

/*
 * Wait until data written to memory
 */
static inline void ti_em_osal_mfence(void)
{
	TI_EM_ARCH_MEM_BARRIER();
}

/*
 * Cache management functions (not used on Linux)
 */
static inline void ti_em_osal_cache_wbinv_L1D(void* block_ptr, unsigned int count) { }
static inline void ti_em_osal_cache_wbinv_L2(void* block_ptr, unsigned int count) { }
static inline void ti_em_osal_cache_wb_L1D(void* block_ptr, unsigned int count) { }
static inline void ti_em_osal_cache_wb_L2(void* block_ptr, unsigned int count) { }
static inline void ti_em_osal_cache_inv_L1D(void* block_ptr, unsigned int count) { }
static inline void ti_em_osal_cache_inv_L2(void* block_ptr, unsigned int count) { }
static inline void ti_em_osal_cache_inv_all_L2(void) { }
static inline void ti_em_osal_cache_inv_prefetch(void) { }

/*
 * Main exit wrapper
 */
static inline void ti_em_osal_exit(int retval) 
{
	exit(retval);
}

#endif /*TI_EM_OSAL_H_*/
