/*
 * Copyright (c) 2013, Texas Instruments Incorporated - http://www.ti.com/
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions are met:
 *       * Redistributions of source code must retain the above copyright
 *         notice, this list of conditions and the following disclaimer.
 *       * Redistributions in binary form must reproduce the above copyright
 *         notice, this list of conditions and the following disclaimer in the
 *         documentation and/or other materials provided with the distribution.
 *       * Neither the name of Texas Instruments Incorporated nor the
 *         names of its contributors may be used to endorse or promote products
 *         derived from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *   DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
 *   DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


/**
 *   @file  ti_em_osal_core.h
 *
 *   @brief
 *      This is the OS abstraction layer for the core execution services.
 *
 */

#ifndef TI_EM_OSAL_CORE_H_
#define TI_EM_OSAL_CORE_H_

#include <ti_em_osal.h>

#include <sys/mman.h>

#define TI_EM_OSAL_CORE_NUM (8) /* max number of cores manageable by the OSAL library for Core Id */

#ifdef TI_EM_THREAD_MODEL
 #include <pthread.h>
 #ifdef TI_EM_HAS___THREAD
 #define TI_EM_OSAL_CORE_LOCAL_VARIABLE(type, var)  __thread type var
 #define TI_EM_OSAL_CORE_LOCAL_INIT(var)
 #define TI_EM_OSAL_CORE_LOCAL_STORE(var, value)    var = *value
 #define TI_EM_OSAL_CORE_LOCAL_READ(var, dest)      dest = var
 #define TI_EM_OSAL_CORE_LOCAL_WRITE(var, value)    var = value
 #else /* TI_EM_HAS___THREAD */
 #define TI_EM_OSAL_CORE_LOCAL_VARIABLE(type, var)  pthread_key_t __key_##var
 #define TI_EM_OSAL_CORE_LOCAL_INIT(var)            (void) pthread_key_create(&__key_##var, NULL)
 #define TI_EM_OSAL_CORE_LOCAL_STORE(var, value)    (void) pthread_setspecific(__key_##var, value)

 #define TI_EM_OSAL_CORE_LOCAL_READ(var, dest)				\
	{								\
		typeof(dest) *ptr =					\
			(typeof(dest)*) pthread_getspecific( __key_##var); \
		dest = *ptr;						\
	}

 #define TI_EM_OSAL_CORE_LOCAL_WRITE(var, value)				\
	{								\
		typeof(value) *ptr =					\
			(typeof(value)*) pthread_getspecific( __key_##var); \
		*ptr = value;						\
	}

 #endif  /* TI_EM_HAS___THREAD */
#else   /* TI_EM_THREAD_MODEL */
 /* Here are the process model definitions and macros */
 #include <unistd.h>
 #include <sys/types.h>
 #include <sys/wait.h>
 #include <semaphore.h>
 #include <sched.h>
 #define TI_EM_OSAL_CORE_LOCAL_VARIABLE(type, var)   type var
 #define TI_EM_OSAL_CORE_LOCAL_INIT(var, value)
 #define TI_EM_OSAL_CORE_LOCAL_STORE(var, value)
 #define TI_EM_OSAL_CORE_LOCAL_READ(var, dest)       dest =  var
 #define TI_EM_OSAL_CORE_LOCAL_WRITE(var, value)     var = value
#endif  /* TI_EM_THREAD_MODEL */

extern TI_EM_OSAL_CORE_LOCAL_VARIABLE(uint32_t, __ti_em_osal_core_id);

/*
 * Return the EM core index
 */
static inline uint32_t ti_em_osal_get_core_idx(void)
{
	uint32_t id;
	TI_EM_OSAL_CORE_LOCAL_READ(__ti_em_osal_core_id, id);
	return id;
}

/*
 * Return the hardware core number on which we run (aka cpu id)
 */
static inline uint32_t ti_em_osal_get_hw_core(void)
{
#ifdef TI_EM_C66X
	return 0;
#else
	return sched_getcpu();
#endif
}

static inline int ti_em_osal_core_mlock(void)
{
	return mlockall(MCL_FUTURE);
}

extern int   ti_em_osal_core_spawn(int(*start_routine)(void*), void *arg);
extern int   ti_em_osal_core_instance_create(int(*start_routine)(void*), void *arg);
extern int   ti_em_osal_core_exit(int retval);
extern int   ti_em_osal_core_wait_exit(int core_handler, int *retval);
extern int   ti_em_osal_core_kill(int core_handler, int sig);
extern int   ti_em_osal_core_set_affinity(int core_handler, int hw_core_id);
extern int   ti_em_osal_core_set_rt_priority(int core_handler, int priority);
extern void  ti_em_osal_core_lock(void);
extern void  ti_em_osal_core_unlock(void);
extern int   ti_em_osal_core_barrier_init(int core_num);
extern void  ti_em_osal_core_barrier_wait(void);
extern void* ti_em_osal_alloc_per_core(void *ptr, uint32_t size);
extern void* ti_em_osal_alloc_shared(void *ptr, uint32_t size);
extern void  ti_em_osal_core_init_global(void);
extern int   ti_em_osal_core_join(int core_id);

#endif /* TI_EM_OSAL_CORE_H_ */
