/*
 * Copyright (c) 2013, Texas Instruments Incorporated - http://www.ti.com/
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions are met:
 *       * Redistributions of source code must retain the above copyright
 *         notice, this list of conditions and the following disclaimer.
 *       * Redistributions in binary form must reproduce the above copyright
 *         notice, this list of conditions and the following disclaimer in the
 *         documentation and/or other materials provided with the distribution.
 *       * Neither the name of Texas Instruments Incorporated nor the
 *         names of its contributors may be used to endorse or promote products
 *         derived from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *   DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
 *   DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


/**
 *   @file  ti_em_rh_core.c
 *
 *   @brief
 *      This is the OpenEM runtime helper library for Linux
 *
 */

#include <ti_em_osal.h>
#include <ti_em_osal_core.h>

#include <ti_em_rh.h>

static int __hw_core_start;
static int __hw_core_num;
static int __core_num;
static int __core_hdl[TI_EM_CORE_NUM];
static int (*__dispatcher_fn)(void*);
static int (*__init_fn)(void);

static int set_rt_config(void)
{
	int res;

	/* lock our memory */
	if (ti_em_osal_core_mlock() != 0)
		TI_EM_OSAL_TRACE(0, "WARNING: memory lock failed\n");

	/* bind OpenEM core to hw core */
	res = ti_em_osal_core_set_affinity(0, (ti_em_osal_get_core_idx()
					       + __hw_core_start) % __hw_core_num);
	if (res != 0)
		TI_EM_OSAL_TRACE(0, "WARNING: set_affinity failed for core %d\n",
			    (ti_em_osal_get_core_idx() + __hw_core_start) % __hw_core_num);

#ifdef TI_EM_USE_RTPRIO /* this can stuck Linux so only set RT prio if you are not in debug phase */
#ifndef TI_EM_RTPRIO
#define TI_EM_RTPRIO 0 /* by default use the highest priority */
#endif
	/* 
	 * set to RT scheduling policy (SCHED_RR) with max priority
	 * only do this if we are one OpenEM core per processor
	 */
	if (__hw_core_num >= __core_num) {
		res = ti_em_osal_core_set_rt_priority(0, TI_EM_RTPRIO);
		if (res != 0)
			TI_EM_OSAL_TRACE(0, "WARNING: set RT scheduling failed\n");
	}
#endif
	return 0;
}

static int dispatcher_trampoline(void* arg)
{
	/* on dispatcher core */
	TI_EM_OSAL_TRACE(2, "launching dispatcher\n");

	set_rt_config();

	/* local initialization */
	if (__init_fn) {
		if ((__init_fn)() < 0)
			return -1;
	}

	/* synchronisation barrier */
	ti_em_osal_core_barrier_wait();

	TI_EM_OSAL_TRACE(2, "running dispatcher on hw core/cpu %d\n",
			 ti_em_osal_get_hw_core());

	/* run the dispatcher */
	ti_em_osal_core_exit((__dispatcher_fn)(arg));

	/* should not reached */
	return -1;
}

/*
 * This user helper function simplify creation of OpenEM cores
 *
 * int core_num is the number of OpenEM cores to launch including ourself!
 * int (*dispatcher_fn)(void*) is the dispatcher fonction pointer
 * int (*init_fn)(void) is the local init function that will be called on each core
 *     before synchronization and calling dispatcher_fn. If NULL, it will not be
 *     called.
 * int hw_core_start is the hardware core (i.e. processor) on which starting to
 *     bind OpenEM cores including ourself (the calling core)
 */
em_status_t ti_em_rh_create_cores(
	int core_num,
	int (*dispatcher_fn)(void*),
	int (*init_fn)(void),
	int hw_core_start)
{
    	int core;

	__hw_core_start = hw_core_start;
	__hw_core_num   = sysconf(_SC_NPROCESSORS_ONLN);
	__core_num      = core_num;

	/* initialization of core barrier */
	ti_em_osal_core_barrier_init(core_num);

	__core_hdl[0]    = 0; /* myself*/
	__dispatcher_fn  = dispatcher_fn;
	__init_fn        = init_fn;

	TI_EM_OSAL_TRACE(2, "launching %d cores on %d processor(s)\n", core_num, __hw_core_num);
	
	/* create other instances */
	if (core_num > 1) {
		for (core = 1; core < core_num; core++) {
			__core_hdl[core] =
				ti_em_osal_core_instance_create(dispatcher_trampoline, NULL);
			
			if (__core_hdl[core] == -1)
				return EM_ERR;
		}
	}

	/* synchronisation barrier */
	ti_em_osal_core_barrier_wait();

	set_rt_config();

	TI_EM_OSAL_TRACE(2, "running main EM core on hw core/cpu %d\n",
			 ti_em_osal_get_hw_core());

	return EM_OK;
}

em_status_t ti_em_rh_wait_cores(int core_num)
{
    	int core;

	/* only core 0 can wait other cores */
	if (ti_em_osal_get_core_idx())
		return EM_OK;

	/* wait core terminations */
	for (core = 1; core < core_num; core++) {
		int retval = -1;
		(void) ti_em_osal_core_wait_exit(__core_hdl[core], &retval);
		TI_EM_OSAL_TRACE(3, "core %d exited with retval = 0x%x\n", core, retval);
	}
	return EM_OK;
}

em_status_t ti_em_rh_kill_cores(int core_num)
{
    	int core;

	/* only core 0 can kill other cores */
	if (ti_em_osal_get_core_idx())
		return EM_OK;

	/* send termination signal to cores */
	for (core = 1; core < core_num; core++) {
		(void) ti_em_osal_core_kill(__core_hdl[core], SIGKILL);
		TI_EM_OSAL_TRACE(3, "core %d killed\n", core);
	}
	return EM_OK;
}
