/*
 * Copyright (c) 2013, Texas Instruments Incorporated - http://www.ti.com/
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions are met:
 *       * Redistributions of source code must retain the above copyright
 *         notice, this list of conditions and the following disclaimer.
 *       * Redistributions in binary form must reproduce the above copyright
 *         notice, this list of conditions and the following disclaimer in the
 *         documentation and/or other materials provided with the distribution.
 *       * Neither the name of Texas Instruments Incorporated nor the
 *         names of its contributors may be used to endorse or promote products
 *         derived from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *   DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
 *   DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


/**
 *   @file  ti_em_rh_dispatcher.c
 *
 *   @brief
 *      This is the OpenEM runtime helper library for Linux
 *
 */
#include <ti/runtime/openem/event_machine.h>

#include <ti_em_osal.h>
#include <ti_em_osal_core.h>
#include <ti_em_osal_firmware.h>
#include <ti_em_osal_event.h>

#include <ti_em_rh.h>

/*
 * Run the OpenEM dispatcher and eventually wait interrupt if using the non-polling
 * mode.
 *
 * The function returns when there is no more events based on conditions defined by 
 * mode, sleep_quota and cancel_quota parameters.
 *
 * mode can be either TI_EM_RH_POLL_MODE for polling only or TI_EM_RH_INTERRUPT_MODE 
 * for waiting on interrupt when there is no more events to dispatch.
 * 
 * sleep_quota defines the number of retries when calling the dispatcher without any 
 * event to dispatch. In polling mode it defines the quota before returning to the
 * caller.
 * In interrupt mode it defines the quota before going to wait on interrupts.
 *
 * cancel_quota defines the minimal number of received events before stopping the
 * wait on interrupt process: while going to sleep and activating interrupt 
 * generation, a bunch of events can be received. This quota is used in this case to 
 * cancel the sleep process, dispatch the incoming events and go back to the caller.
 * If cancel_quota is null, the sleep process is not canceled in this case.
 *
 */
em_status_t ti_em_rh_dispatch(int mode, uint32_t sleep_quota, uint32_t cancel_quota)
{
	em_status_t status = EM_ERROR;
	int count;
	
	count = sleep_quota;

	/* spin on dispatcher */
	while (count >= 0) {
		/* run the OpenEM dispatcher */
		status = ti_em_dispatch_once();
		if (status != EM_OK)
			/* no more event */
			count--;
		else
			/* event, so reset delay */
			count = sleep_quota;
	}
	
	/* if polling only, return */
	if (mode == TI_EM_RH_POLL_MODE)
		return status;

	/* otherwise prepare to sleep */
	TI_EM_OSAL_TRACE(3, "no more event, going to sleep\n");
	
	/* enable interrupt generation */
	ti_em_interrupt_enable();
	
	/* consume events which may occur while enabling interrupts */
	count = 0;
	while ((count < cancel_quota) || (cancel_quota == 0)) {
		/* run the OpenEM dispatcher */
		status = ti_em_dispatch_once();
		if (status != EM_OK) {
			int res;
			/* no more event, we will really go to sleep until interrupts */
			TI_EM_OSAL_TRACE(3, "calling wait\n");
			res = ti_em_osal_event_wait();
			if (res != 0)
				status = EM_ERROR;
			break;
		} else
			/* event received */
			count++;
	}

	/* disable interrupt generation */
	ti_em_interrupt_disable();

	if (status != EM_ERROR) {
		/* spin on dispatcher */
		while ((status = ti_em_dispatch_once()) == EM_OK);
	}

	return status;
}
