/*
 * Copyright (c) 2013, Texas Instruments Incorporated - http://www.ti.com/
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions are met:
 *       * Redistributions of source code must retain the above copyright
 *         notice, this list of conditions and the following disclaimer.
 *       * Redistributions in binary form must reproduce the above copyright
 *         notice, this list of conditions and the following disclaimer in the
 *         documentation and/or other materials provided with the distribution.
 *       * Neither the name of Texas Instruments Incorporated nor the
 *         names of its contributors may be used to endorse or promote products
 *         derived from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *   DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
 *   DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


/**
 *   @file  ti_em_rh_init.c
 *
 *   @brief
 *      This is the OpenEM runtime helper library for Linux
 *
 */
#include <ti/runtime/openem/event_machine.h>

#include <ti_em_osal.h>
#include <ti_em_osal_core.h>
#include <ti_em_osal_queue.h>
#include <ti_em_osal_cppi.h>
#include <ti_em_osal_cma.h>
#include <ti_em_osal_uaccess.h>
#include <ti_em_osal_firmware.h>
#include <ti_em_osal_event.h>

#include <ti_em_rh.h>

#ifdef TI_EM_COHERENT_IO
#define BUFFER_COH_MODE    TI_EM_COH_MODE_OFF
#define DESC_MAPPING_FLAGS TI_EM_OSAL_MEM_CACHED
#else
#define BUFFER_COH_MODE    TI_EM_COH_MODE_ON
#define DESC_MAPPING_FLAGS TI_EM_OSAL_MEM_UNCACHED
#endif

#define QUEUE_BASE_ALIGN_UP(q) (((q) + (TI_EM_HW_QUEUE_STEP - 1)) & ~(TI_EM_HW_QUEUE_STEP - 1))

static ti_em_rh_hw_config_t   __hw_config;
static ti_em_rh_pool_config_t __pool_configs[TI_EM_POOL_NUM];
static uint32_t               __pool_num;

static em_status_t init_pdsp(ti_em_rh_pdsp_config_t *pdsp_configs)
{
	short       pdsp = pdsp_configs->pdsp_id;
	void       *pdsp_dram;
	em_status_t res;

	if (pdsp > QMSS_MAX_PDSP)
		return EM_ERROR;

	/* if -1, do not initialize this PDSP */
	if (pdsp == -1)
		return EM_OK;

	/* stop the PDSP */
	if (ti_em_osal_pdsp_stop(pdsp_configs->pdsp_id) != EM_OK)
		return EM_ERROR;

	/* clear PDSP local memory */
	pdsp_dram = (void *) ti_em_osal_mem_phys_to_virt((TI_EM_PDSP_DRAM(pdsp_configs->pdsp_id)),
							 TI_EM_MEM_PDSP_DRAM);
	memset(pdsp_dram, 0, TI_EM_PDSP_DRAM_SIZE);

	/* download firmware and run PDSP */
	res = ti_em_osal_pdsp_download_firmware(pdsp_configs->pdsp_id,
						pdsp_configs->firmware,
						pdsp_configs->size);
	if (res != EM_OK) {
		TI_EM_OSAL_ERROR("failed to load PDSP%d firmware\n", pdsp);
		return res;
	}
	return EM_OK;;
}

static int init_mappings(ti_em_rh_hw_config_t   *hw_config,
			 uint32_t                pool_num,
			 ti_em_rh_pool_config_t *pool_configs)
{
	struct ti_em_osal_mem_mapping mapping_info;
	int mapping_id = TI_EM_MEM_BUFFER;
	int res;
	int i;
	
	/* Initialize CMA */
	ti_em_osal_cma_init();

	memset(&mapping_info, 0, sizeof(mapping_info));

	/* Map the I/O configuration registers */
	mapping_info.paddr  = TI_EM_IO_CFG_BASE;
	mapping_info.vaddr  = 0;
	mapping_info.offset = 0;
	mapping_info.size   = TI_EM_IO_CFG_SIZE;
	mapping_info.flags  = TI_EM_OSAL_MEM_UNCACHED;
	res = ti_em_osal_create_mapping(TI_EM_MEM_QMSS_REGS, &mapping_info);
	if (res == -1) {
		TI_EM_OSAL_ERROR("cannot map QMSS registers\n");
		return EM_ERR;
	}

	/* Map the PDSP local memory */
	mapping_info.paddr  = TI_EM_PDSP_DRAM(0);
	mapping_info.vaddr  = 0;
	mapping_info.offset = TI_EM_PDSP_DRAM_OFFSET;
	mapping_info.size   = TI_EM_PDSP_DRAM_SIZE * QMSS_MAX_PDSP;
	mapping_info.flags  = TI_EM_OSAL_MEM_UNCACHED;
	res = ti_em_osal_create_mapping(TI_EM_MEM_PDSP_DRAM, &mapping_info);
	if (res == -1) {
		TI_EM_OSAL_ERROR("cannot map QMSS registers\n");
		return EM_ERR;
	}

	/* Public descriptors (cached if coherent I/O) */
	mapping_info.paddr  = hw_config->region_configs[TI_EM_RH_PUBLIC].desc_base;
	mapping_info.vaddr  = hw_config->region_configs[TI_EM_RH_PUBLIC].desc_vbase;
	mapping_info.offset = hw_config->region_configs[TI_EM_RH_PUBLIC].desc_offset;
	mapping_info.size   = hw_config->region_configs[TI_EM_RH_PUBLIC].desc_size *
		hw_config->region_configs[TI_EM_RH_PUBLIC].desc_num;
	mapping_info.flags  = DESC_MAPPING_FLAGS;

	/* Test if the QMSS public descriptors must be allocated with CMA */
	if (mapping_info.paddr == 0) {
		mapping_info.vaddr = (unsigned int)
			ti_em_osal_cma_alloc(mapping_info.size,
					     mapping_info.flags,
					     &mapping_info.paddr);
		if (mapping_info.vaddr == (unsigned int) -1) {
			TI_EM_OSAL_ERROR("CMA failed for public descriptors\n");
			return EM_ERR;
		}
	}

	/* Map the QMSS public descriptors */
	res = ti_em_osal_create_mapping(TI_EM_MEM_PUBLIC_DESC, &mapping_info);
	if (res == -1) {
		TI_EM_OSAL_ERROR("cannot map public descriptors\n");
		return EM_ERR;
	}

	/* Initialize the public region */
	memset((void*) mapping_info.vaddr, 0, mapping_info.size);
	
	/* update the descriptor base with the CMA physical address */
	hw_config->region_configs[TI_EM_RH_PUBLIC].desc_base = mapping_info.paddr;

	/* Private descriptors */
	mapping_info.paddr  = hw_config->region_configs[TI_EM_RH_PRIVATE].desc_base;
	mapping_info.vaddr  = hw_config->region_configs[TI_EM_RH_PRIVATE].desc_vbase;
	mapping_info.offset = hw_config->region_configs[TI_EM_RH_PRIVATE].desc_offset;
	mapping_info.size   = hw_config->region_configs[TI_EM_RH_PRIVATE].desc_size *
		hw_config->region_configs[TI_EM_RH_PRIVATE].desc_num;
	mapping_info.flags  = TI_EM_OSAL_MEM_UNCACHED;

	if (mapping_info.paddr == 0) {
		TI_EM_OSAL_ERROR("private descriptors cannot be allocated by CMA\n");
		return EM_ERR;
	}

	/* Map the QMSS private descritors */
	res = ti_em_osal_create_mapping(TI_EM_MEM_PRIVATE_DESC, &mapping_info);
	if (res == -1) {
		TI_EM_OSAL_ERROR("cannot map private descriptors\n");
		return EM_ERR;
	}

	/* Initialize the private region */
	memset((void*) mapping_info.vaddr, 0, mapping_info.size);

	/* Map the PSDP communication memory */
	mapping_info.paddr  = hw_config->pdsp_comm_mem_config.paddr;
	mapping_info.vaddr  = hw_config->pdsp_comm_mem_config.vaddr;
	mapping_info.offset = hw_config->pdsp_comm_mem_config.offset;
	mapping_info.size   = hw_config->pdsp_comm_mem_config.size;
	mapping_info.flags  = TI_EM_OSAL_MEM_UNCACHED;

	/* Test if the PDSP communication memory must be allocated with CMA */
	if (mapping_info.paddr == 0) {
		mapping_info.vaddr = (unsigned int)
			ti_em_osal_cma_alloc(mapping_info.size,
					     mapping_info.flags,
					     &mapping_info.paddr);
		if (mapping_info.vaddr == (unsigned int) -1) {
			TI_EM_OSAL_ERROR("CMA failed for PDSP communication memory\n");
			return EM_ERR;
		}
	}

	res = ti_em_osal_create_mapping(TI_EM_MEM_PDSP_COMM, &mapping_info);
	if (res == -1) {
		TI_EM_OSAL_ERROR("PDSP communication memory cannot be allocated by CMA\n");
		return EM_ERR;
	}

	/* Update the PDSP communication memory base with the CMA physical address */
	hw_config->pdsp_comm_mem_config.paddr = mapping_info.paddr;

	/* Initialize the PDSP communication memory */
	memset((void*) mapping_info.vaddr, 0, mapping_info.size);

	/* Store the PDSP communication memory base */
	ti_em_osal_set_pdsp_comm_mem_base(hw_config->pdsp_comm_mem_config.paddr);

	/* Buffer (descriptor payload) memory */
	if (pool_num < 1) {
		TI_EM_OSAL_TRACE(2, "warning: no descriptor payload memory defined!\n");
		return EM_OK;
	}

	for (i = 0; i < pool_num; i++) {
		mapping_info.paddr  = pool_configs[i].mem.paddr;
		mapping_info.vaddr  = pool_configs[i].mem.vaddr;
		mapping_info.offset = pool_configs[i].mem.offset;
		mapping_info.flags  = pool_configs[i].mem.flags;

		if (pool_configs[i].mem.size)
			mapping_info.size = pool_configs[i].mem.size;
		else
			mapping_info.size = pool_configs[i].event_num *
				((pool_configs[i].buffer_size + 7) & (~7));
		
		if ((mapping_info.paddr == -1) || (mapping_info.size == 0)) {
			mapping_id++;
			continue; /* no payload */
		}

		if ((mapping_id >= TI_EM_MEM_NUM) || (i > TI_EM_BUFFER_POOL_ID_NUM)) {
			TI_EM_OSAL_ERROR("too many buffer memory pools (%d): cannot allocate resource\n", i);
			return EM_ERR;
		}

		/*  Test if the QMSS public descriptors must be allocated with CMA */
		if (mapping_info.vaddr) {
			/*
			 * If not physical address is provided and virtual one is defined
			 * we consider that this payload memory is local only and thus
			 * not allocated through CMA and not mapped.
			 * In this case we set physical address equal to virtual one.
			 * Of course PktDMA should not be used on these buffers!!!
			 */
			mapping_info.paddr = mapping_info.vaddr;
		} else if (mapping_info.paddr == 0) {
			mapping_info.vaddr = (unsigned int)
				ti_em_osal_cma_alloc(mapping_info.size,
						     mapping_info.flags,
						     &mapping_info.paddr);;
			if (mapping_info.vaddr == (unsigned int) -1) {
				TI_EM_OSAL_ERROR("CMA failed for descriptor payload memory\n");
				return EM_ERR;
			}
		}

		/* Map the QMSS public descriptors */
		res = ti_em_osal_create_mapping(mapping_id, &mapping_info);
		if (res == -1) {
			TI_EM_OSAL_ERROR("descriptor payload memory cannot be allocated by CMA\n");
			return EM_ERR;
		}
		
		/* Register the mapping Id */
		pool_configs[i].mem.mapping_id = mapping_id++;

		/* Initialize the public region */
		memset((void*) mapping_info.vaddr, 0, mapping_info.size);
	
		/* Update the descriptor base with the CMA physical and virtual addresses */
		pool_configs[i].mem.paddr = mapping_info.paddr;
		pool_configs[i].mem.vaddr = mapping_info.vaddr;
	}

	return EM_OK;
}

static em_status_t init_private_events(
	ti_em_osal_hw_region_config_t *region_configs,
	uint32_t                       queue_idx)
{
	uint8_t *desc;
	uint32_t desc_size;
	int      i;

	/* Open CD private event queue */
	if (ti_em_osal_hw_queue_open(queue_idx + 1) != EM_OK)
		return EM_ERR;
	
	TI_EM_OSAL_TRACE(2, "initializing private CD events (queue 0x%x)\n", queue_idx + 1);

	/* 
	 * Assume private region is at base address of the pdsp data ram,
	 * first descriptor is reserved for cd queue identification.
	 */
	desc = (uint8_t*) ti_em_osal_mem_phys_to_virt(region_configs[TI_EM_RH_PRIVATE].desc_base,
						      TI_EM_MEM_PRIVATE_DESC);
	desc_size = region_configs[TI_EM_RH_PRIVATE].desc_size;

	/* 
	 * Skip one descriptor because beginning of the private memory
	 * is used for PDSP init/management.
	 */
	desc += desc_size;

	/* Push private descriptors in private free queue */
	for (i = 0; i < (TI_EM_CD_PRIVATE_EVENT_NUM - 1); i++) {
		ti_em_osal_hw_queue_push(queue_idx + 1,
					 (void*)desc,
					 TI_EM_MEM_PRIVATE_DESC);
		desc += desc_size;
	}

	/* Open AP private event queue */
	if (ti_em_osal_hw_queue_open(queue_idx) != EM_OK)
		return EM_ERR;
	
	TI_EM_OSAL_TRACE(2, "initializing private AP events (queue 0x%x)\n", queue_idx);

	/* push private descriptors in private free queue */
	for (i = 0; i < TI_EM_AP_PRIVATE_EVENT_NUM; i++) {
		ti_em_osal_hw_queue_push(queue_idx,
					 (void*)desc,
					 TI_EM_MEM_PRIVATE_DESC);
		desc += desc_size;
	}
	
	return EM_OK;
}

static em_status_t init_events(
	ti_em_osal_hw_region_config_t *region_configs,
	uint32_t                       pool_num,
	ti_em_rh_pool_config_t        *pool_configs)
{	
	uint8_t *desc;
	uint32_t desc_size;
	uint32_t event_num_total = 1; /* because we skip the first descriptor */
	uint32_t event_num_max   = region_configs[TI_EM_RH_PUBLIC].desc_num;
	int      i;

	desc = (uint8_t*) ti_em_osal_mem_phys_to_virt(region_configs[TI_EM_RH_PUBLIC].desc_base,
						      TI_EM_MEM_PUBLIC_DESC);

	desc_size = region_configs[TI_EM_RH_PUBLIC].desc_size;

	/* Skip the first descriptor that cannot be assigned to any pool */
	desc += desc_size;

	for (i = 0; i < pool_num; i++) {
		phys_addr_t buf_paddr;
		uint32_t    buf_vaddr;
		uint32_t    buf_size;
		uint32_t    offset;
		uint32_t    event_num;
		uint32_t    queue_idx;
		int         j;

		buf_size         = pool_configs[i].buffer_size;
		event_num        = pool_configs[i].event_num;
		queue_idx        = pool_configs[i].pool.free_queue_idx;
		event_num_total += event_num;

		if (event_num_total > event_num_max) {
			TI_EM_OSAL_ERROR("not enough space in the public descriptor region to allocate %d events\n", event_num);
			return EM_ERR;
		}

		/* Adjust data buffer for payload by (optional) offset */
		offset           = pool_configs[i].offset;
		buf_paddr        = pool_configs[i].mem.paddr + offset;
		buf_vaddr        = pool_configs[i].mem.vaddr + offset;

		if (ti_em_osal_hw_queue_open(queue_idx) != EM_OK)
			return EM_ERR;

		TI_EM_OSAL_TRACE(2, "initializing pool %d starting at 0x%x, desc size 0x%x, event num 0x%x, buffer 0x%x, buffer size %d, free queue %d, offset %d\n",
				 i, (uint32_t) desc, desc_size, event_num, buf_paddr, buf_size, queue_idx, offset);

		for (j = 0; j < event_num; j++)	{
			Cppi_DescTag          tag;
			ti_em_buffer_config_t buffer_config;

			ti_em_packet_set_default((ti_em_packet_t*)desc);

			ti_em_packet_set_pool_info((ti_em_packet_t*)(desc),
						   (pool_configs[i].pool));

			if (buf_paddr != 0) {
				buffer_config.buffer_ptr       = (void*) buf_paddr;
				buffer_config.buffer_size      = (uint32_t) buf_size - offset;
				buffer_config.orig_buffer_ptr  = (void*) buf_paddr;
				buffer_config.orig_buffer_size = (uint32_t) buf_size - offset;
				buffer_config.orig_buffer_pool = (uint32_t) i;
				buffer_config.coh_mode         = BUFFER_COH_MODE;
				buffer_config.free_func        = (ti_em_free_func_t) NULL;
				buffer_config.free_push_policy = TI_EM_PUSH_POLICY_TAIL;
				
				ti_em_packet_set_buffer_info((ti_em_packet_t*) (desc),
							     buffer_config);
			}
			
			ti_em_cppi_set_pslen(Cppi_DescType_HOST,
					     (Cppi_Desc *)(desc),
					     pool_configs[i].ps_word_num << 2);

			tag.srcTagHi  = 0x00;
			tag.srcTagLo  = 0xFF;
			tag.destTagHi = 0x00;
			tag.destTagLo = 0x00;
			
			ti_em_cppi_set_tag(Cppi_DescType_HOST,
					   (Cppi_Desc*)(desc),
					   &tag);

			ti_em_osal_mfence();

			ti_em_osal_hw_queue_push(queue_idx,
						 (void*)desc, 
						 TI_EM_MEM_PUBLIC_DESC);

			buf_paddr += (buf_size + 7) & ~(7);
			buf_vaddr += (buf_size + 7) & ~(7);
			desc      += desc_size;
		}
	}

	TI_EM_OSAL_TRACE(2, "total number of allocated event is %d\n", event_num_total);
	
	return EM_OK;
}

static void release_events(
	uint32_t                pool_num,
	ti_em_rh_pool_config_t *pool_configs)
{
	int i;

	/* Release free queues associated to public event pools  */
	for (i = 0; i < pool_num; i++) {
		uint32_t queue_idx;
		queue_idx = pool_configs[i].pool.free_queue_idx;
		ti_em_osal_hw_queue_close(queue_idx);
	}
}

static void release_private_events(uint32_t queue_idx)
{
	/* Close CD private event queue */
	ti_em_osal_hw_queue_close(queue_idx + 1);

	/* Close AP private event queue */
	ti_em_osal_hw_queue_open(queue_idx);
}

static uint32_t round_desc_num(uint32_t event_num)
{
	if (event_num <  0x00000020)
		event_num = 0x00000020;
	else {
		uint32_t event_num_tmp = 0x00004000 >> 1;
		while(event_num <= event_num_tmp)
			event_num_tmp >>= 1;
		event_num = event_num_tmp << 1;
	}
	return event_num;
}

/*
 * Initialize QMSS
 */
static em_status_t init_qmss(
	uint32_t                       region_num,
	ti_em_osal_hw_region_config_t *region_configs,
	Cppi_Handle                   *cppi_hdl_ptr)
{
	uint32_t desc_num = 0;
	int      i, j;
	int      res;
	
	for (i = 0; i < region_num; i++) {
		TI_EM_OSAL_TRACE(2, "region %d QMSS hw descriptor number is %d\n", i, region_configs[i].desc_num);
		desc_num += region_configs[i].desc_num;
	}

	TI_EM_OSAL_TRACE(2, "total QMSS hw descriptor number allocated is %d\n", desc_num);
	
	/* Insert descriptor regions */
	for (i = 0; i < region_num; i++) {
		for (j = 0; j < QMSS_MAX_QMGR_GROUPS; j++) {
			res = ti_em_osal_hw_queue_insert_region(&region_configs[i], j);
			if (res != EM_OK) {
				TI_EM_OSAL_ERROR("region %d insertion failed", i);
				return res;
			}
		}
	}
	return EM_OK;
}

/*
 * Setup QMSS descriptor memory regions configuration
 *
 * At least 1 region for public global events.
 * Exactly 1 region for private events.
 */
static int init_region_config(
	uint32_t                       region_num,
	ti_em_osal_hw_region_config_t *region_configs)
{
	uint32_t desc_num = 0;
	uint32_t link_idx = 0;
	int      i;

	if ((region_num > QMSS_MAX_MEM_REGIONS) && (region_num < 1))
		return -1;
	
	/* Fixup descriptor number */
	region_configs[TI_EM_RH_PUBLIC].desc_num  = round_desc_num(region_configs[TI_EM_RH_PUBLIC].desc_num);
	region_configs[TI_EM_RH_PRIVATE].desc_num = round_desc_num(region_configs[TI_EM_RH_PRIVATE].desc_num);

	TI_EM_OSAL_ASSERT(region_configs[0].desc_base != 0);

	/* 
	 * Automatically calculate desc base and linking RAM index
	 * if fields are set to 0.
	 */
	for (i = 0; i < region_num; i++) {
		if ((region_configs[i].desc_base == 0) && (i != 0)) {
			region_configs[i].desc_base = 
				region_configs[i - 1].desc_base
				+ (region_configs[i - 1].desc_num
				   * region_configs[i - 1].desc_size);
		}
		if (region_configs[i].start_idx != -1)
			link_idx = region_configs[i].start_idx; 
	}

	/* Assign linking RAM start index */
	for (i = 0; i < region_num; i++)
	{
		if (region_configs[i].start_idx == -1)
			region_configs[i].start_idx  = link_idx + desc_num;
		desc_num += region_configs[i].desc_num;
		link_idx  = region_configs[i].start_idx;
		
		TI_EM_OSAL_TRACE(2, "region %d: region_idx is %d, link_idx is %d, desc_base is 0x%x\n",
			    i,
			    region_configs[i].region_idx, 
			    region_configs[i].start_idx,
			    region_configs[i].desc_base);
	}
	return region_num;
}

static void init_hw_config(
	uint32_t                core_num,
	ti_em_rh_hw_config_t   *hw_config,
	uint32_t                pool_num,
	ti_em_rh_pool_config_t *pool_configs,
	ti_em_config_t         *em_hw_config_ptr)
{
	int i;

	memset(em_hw_config_ptr, 0, sizeof(*em_hw_config_ptr));

	em_hw_config_ptr->my_device_idx          = hw_config->device_id;
	em_hw_config_ptr->my_process_idx         = hw_config->process_id;
	em_hw_config_ptr->chain_config_ptr       = hw_config->chain_config_ptr;
	em_hw_config_ptr->hw_queue_base_idx      = QUEUE_BASE_ALIGN_UP(hw_config->hw_queue_base_idx);
	em_hw_config_ptr->dma_idx                = hw_config->dma_idx;
	em_hw_config_ptr->dma_queue_base_idx     = hw_config->dma_queue_base_idx;
	em_hw_config_ptr->pool_num               = pool_num;
	
	for (i = 0; i < hw_config->pdsp_num; i++)
		em_hw_config_ptr->pdsp_idx_tbl[i] = hw_config->pdsp_sched[i].pdsp_id;

	em_hw_config_ptr->pdsp_num               = hw_config->pdsp_num;
	em_hw_config_ptr->ap_region_queue_idx    = hw_config->private_free_queue_idx;
	em_hw_config_ptr->cd_region_queue_idx    = hw_config->private_free_queue_idx + 1;
	em_hw_config_ptr->preload_config_ptr     = NULL;
	em_hw_config_ptr->hw_interrupt_base_idx  = hw_config->hw_interrupt_base_idx;

	for (i = 0; i < pool_num; i++)
	{
		em_hw_config_ptr->pool_config_tbl[i].free_queue_idx   = pool_configs[i].pool.free_queue_idx;
		em_hw_config_ptr->pool_config_tbl[i].buf_mode         = pool_configs[i].pool.buf_mode;
		em_hw_config_ptr->pool_config_tbl[i].dsc_wsize        = pool_configs[i].pool.dsc_wsize;
	}
}

em_status_t ti_em_rh_init_global(
	uint32_t                pool_num,
	ti_em_rh_pool_config_t *pool_configs,
	uint32_t                core_num,
	ti_em_rh_hw_config_t   *hw_config)
{
	ti_em_config_t          em_hw_config;
	int                     pdsp;
	int                     res;

	if (pool_num > TI_EM_POOL_NUM)
		return EM_ERR;

	if (core_num > TI_EM_CORE_NUM)
		return EM_ERR;


	if (!hw_config)
		return EM_ERR;

	/* Initialize and register the mappings */
	res = init_mappings(hw_config, pool_num, pool_configs);
	if (res != EM_OK) {
		TI_EM_OSAL_ERROR("Mappings init failed\n");
		return res;
	}

	/* Set and configure QMSS regions */
	TI_EM_OSAL_TRACE(1, "set QMSS region init\n");
	res = init_region_config(hw_config->region_num,
				 hw_config->region_configs);
	if (res != 2) {
		TI_EM_OSAL_ERROR("QMSS region init failed (res=%d)\n", res);
		return -1;
	}

	/* Initialize QMSS */
	TI_EM_OSAL_TRACE(1, "initialize QMSS\n");
	res = init_qmss(hw_config->region_num,
			hw_config->region_configs,
			NULL);
	if (res < 0) {
		TI_EM_OSAL_ERROR("QMSS init failed (res=%d)\n", res);
		return -1;
	}

	/* Save our hardware configuration for release function */
	memcpy(&__hw_config, hw_config, sizeof(__hw_config));
	memcpy(&__pool_configs[0], &pool_configs[0],
	       sizeof(ti_em_rh_pool_config_t) * pool_num);
	__pool_num = pool_num;

	/* Load PDSP firmware(s) for scheduler */
	if (hw_config->pdsp_num >= TI_EM_SCHEDULER_THREAD_NUM) {
		TI_EM_OSAL_ERROR("Invalid number of scheduler (%d)\n",
			hw_config->pdsp_num);
			return EM_ERR;
	}
		
	TI_EM_OSAL_TRACE(1, "load PDSP scheduler firmware\n");
	for (pdsp = 0; pdsp < hw_config->pdsp_num; pdsp++) {
		res = init_pdsp(&hw_config->pdsp_sched[pdsp]);
		if (res != EM_OK) {
			TI_EM_OSAL_ERROR("PDSP scheduler firmware %d loading failed\n", pdsp);
			return EM_ERR;
		}
	}

	/* Load PDSP firmware for router */
	TI_EM_OSAL_TRACE(1, "load PDSP router firmware\n");
	res = init_pdsp(&hw_config->pdsp_router);
	if (res != EM_OK) {
		TI_EM_OSAL_ERROR("PDSP scheduler firmware loading failed\n");
		return EM_ERR;
	}

	/* Initialize CPPI library */
	ti_em_osal_cppi_init();

	/* Initialize private (PDSP) events */
	if (init_private_events(hw_config->region_configs,
				hw_config->private_free_queue_idx) != EM_OK) {
		TI_EM_OSAL_ERROR("private event initialization failed!\n");
		return EM_ERR;
	}

	/* Build table of pool configurations and initialize events */
	if (init_events(hw_config->region_configs,
			pool_num,
			pool_configs) != EM_OK) {
		TI_EM_OSAL_ERROR("event initialization failed!\n");
		return EM_ERR;
	}

	/* Set EM configuration and initialize EM */
	TI_EM_OSAL_TRACE(1, "set EM configuration and initialize EM\n");
	init_hw_config(core_num,
		       hw_config,
		       pool_num,
		       pool_configs,
		       &em_hw_config);

	if (ti_em_init_global(&em_hw_config) != EM_OK) {
		TI_EM_OSAL_ERROR("EM configuration and initialization failed!\n");
		return EM_ERR;
	}

	return EM_OK;
}

em_status_t ti_em_rh_release_global(void)
{
	int pdsp;

	/* EM exit procedure */
	ti_em_exit_global();

	/* Release events (public and private) and associated queues */
	release_events(__pool_num, &__pool_configs[0]);
	release_private_events(__hw_config.private_free_queue_idx);

	/* Stop PDSP */
	for (pdsp = 0; pdsp < __hw_config.pdsp_num; pdsp++)
		if (__hw_config.pdsp_sched[pdsp].pdsp_id >= 0)
			ti_em_osal_pdsp_stop(__hw_config.pdsp_sched[pdsp].pdsp_id);

	if (__hw_config.pdsp_router.pdsp_id >= 0)
		ti_em_osal_pdsp_stop(__hw_config.pdsp_router.pdsp_id);

	return EM_OK;
}

em_status_t ti_em_rh_init_local(void)
{
	TI_EM_OSAL_TRACE(1, "EM local configuration\n");

	if ((__hw_config.dispatch_mode == TI_EM_RH_INTERRUPT_MODE) 
	    && (__hw_config.pdsp_sched[0].pdsp_id >= 0)) {
		int res;

		/* Initialize per instance event notification */
		TI_EM_OSAL_TRACE(2, "initialize interrupt event notification\n");
		res = ti_em_osal_event_init(__hw_config.pdsp_sched[0].pdsp_id,
					    __hw_config.hw_interrupt_base_idx);
		if (res < 0) {
			TI_EM_OSAL_ERROR("interrupt event notification init failed\n");
			return EM_ERR;
		}
	}

 	if (ti_em_init_local() != EM_OK) {
		TI_EM_OSAL_ERROR("EM local configuration failed!\n");
		return EM_ERR;
	}

	return EM_OK;
}

em_status_t ti_em_rh_release_local(void)
{
	if ((__hw_config.dispatch_mode == TI_EM_RH_INTERRUPT_MODE) 
	    && (__hw_config.pdsp_sched[0].pdsp_id >= 0)) {
		int res;
		
		/* Release per instance event notification */
		TI_EM_OSAL_TRACE(2, "release interrupt event notification\n");
		res = ti_em_osal_event_release();
		if (res < 0) {
			TI_EM_OSAL_ERROR("interrupt event notification release failed\n");
			return EM_ERR;
		}
	}

	return EM_OK;
}
