/*
 * Copyright (c) 2013, Texas Instruments Incorporated - http://www.ti.com/
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions are met:
 *       * Redistributions of source code must retain the above copyright
 *         notice, this list of conditions and the following disclaimer.
 *       * Redistributions in binary form must reproduce the above copyright
 *         notice, this list of conditions and the following disclaimer in the
 *         documentation and/or other materials provided with the distribution.
 *       * Neither the name of Texas Instruments Incorporated nor the
 *         names of its contributors may be used to endorse or promote products
 *         derived from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *   DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
 *   DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


/**
 *   @file  ti_em_rh_join.c
 *
 *   @brief
 *      This is the OpenEM runtime helper library for Linux
 *
 */
#include <ti/runtime/openem/event_machine.h>

#include <ti_em_osal.h>
#include <ti_em_osal_core.h>
#include <ti_em_osal_uaccess.h>

#include <ti_em_rh.h>

#ifdef TI_EM_THREAD_MODEL
em_status_t ti_em_rh_join(int core_id)
{
	TI_EM_OSAL_ERROR("cannot join in thread model\n");
	return EM_ERR_NOT_IMPLEMENTED;
}
#else /* TI_EM_THREAD_MODEL */
em_status_t ti_em_rh_join(int core_id)
{
	struct ti_em_osal_mem_mapping mapping_info;
	short                         mapping;
	int                           res;

	/*
	 * Join the OSAL library
	 */
	res = ti_em_osal_core_join(core_id);
	if (res < 0) {
	    TI_EM_OSAL_ERROR("cannot join the OSAL database\n");
	    return EM_ERR;
	}

	/*
	 * Retrieve all mappings for the mapping shared db
	 */
	for (mapping = 0; mapping < TI_EM_MEM_NUM; mapping++) {
		int res;

		memset(&mapping_info, 0, sizeof(mapping_info));

		/* retrieve first the (if existing) mapping */
		res = ti_em_osal_retrieve_mapping(mapping, &mapping_info);
		
		if (res == 0)
			continue;
		
		if (res < 0) {
			TI_EM_OSAL_ERROR("error when retrieving mapping %d\n", 
					 mapping);
			return EM_ERR;
		}

		/* create a new one in this process and register it locally */
		res = ti_em_osal_create_mapping(mapping, &mapping_info);
		if (res == -1) {
			TI_EM_OSAL_ERROR("cannot map QMSS registers\n");
			return EM_ERR;
		}

		if (mapping == TI_EM_MEM_PDSP_COMM) {
			/* Store the PDSP communication memory base */
			ti_em_osal_set_pdsp_comm_mem_base(mapping_info.paddr);
		}

	}

	/*
	 * Join OpenEM 
	 */
	if (ti_em_init_global(NULL) != EM_OK) {
		TI_EM_OSAL_ERROR("EM join failed!\n");
		return EM_ERR;
	}

	return EM_OK;
}
#endif /* !TI_EM_THREAD_MODEL */
