/*
 * Copyright (c) 2012, Texas Instruments Incorporated - http://www.ti.com/
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions are met:
 *       * Redistributions of source code must retain the above copyright
 *         notice, this list of conditions and the following disclaimer.
 *       * Redistributions in binary form must reproduce the above copyright
 *         notice, this list of conditions and the following disclaimer in the
 *         documentation and/or other materials provided with the distribution.
 *       * Neither the name of Texas Instruments Incorporated nor the
 *         names of its contributors may be used to endorse or promote products
 *         derived from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *   DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
 *   DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


/**
 *   @file  event_machine_hw_ti_chain_pktdma.c
 *
 *   @brief   
 *
 */
#include <ti/runtime/openem/src/event_machine_private_types.h>
#include <ti/runtime/openem/src/event_machine_private_variables.h>
#include <ti/runtime/openem/src/event_machine_private_functions.h>
#include <ti/runtime/openem/modules/event_machine_modules.h>

static em_status_t
Emti_pktdmaAddRoute (
    ti_em_destination_id_t dest_id,
    void *route
)
{
    ti_em_process_route_t *process_route_ptr = (ti_em_process_route_t *) route;
    ti_em_process_id_t process_id = (ti_em_process_id_t) dest_id;
    ti_em_queue_id_t lvProcessTxQueueIdx;

    Emti_logDbg ("[PKTDMA-CHAIN] add route to %d\n", process_id);

    lvProcessTxQueueIdx = ti_em_get_absolute_queue_id (process_route_ptr->dma_idx, process_route_ptr->dma_tx_queue_idx);

    Emti_setField (Emti_svPsbPtr->processRouteTbl[process_id].routeInfo, PROCESS_ROUTE_INFO_ROUTE_ENABLED, 1);
    Emti_setField (Emti_svPsbPtr->processRouteTbl[process_id].routeInfo, PROCESS_ROUTE_INFO_TX_QUEUE_IDX,
                   lvProcessTxQueueIdx);

    return EM_OK;
}

static em_status_t
Emti_pktdmaConnect (
    void
)
{
    em_status_t lvStatus;
    uint32_t lvSchedIdx = 0;

    Emti_logDbg ("[PKTDMA-CHAIN] connect\n");

    /*
     * For the time being, let say only one is responsible to do the routing. 
     */
    /*
     * for (lvSchedIdx = 0; lvSchedIdx < Emti_svPcb.pdspNum; lvSchedIdx++) 
     */
    {
        lvStatus =
            Emti_connectFirmware (Emti_svPcb.cdQueueIdx[lvSchedIdx], Emti_svPcb.chainDsc.ctCommandFreeQueueIdx,
                                  Emti_svPcb.chainDsc.rpQueueIdx, (uint32_t) & Emti_svPcb.shadowRDsc,
                                  (uint32_t) sizeof (Emti_svPcb.shadowRDsc), (uint32_t) Emti_CtType_ROUTER_CONNECT,
                                  (uint32_t) TI_EM_SLOT_RPCB, (ti_em_process_id_t) 0);
    }
    return lvStatus;
}

static em_status_t
Emti_pktdmaGlobalInit (
    ti_em_chain_config_t * chain_config_ptr,
    uint32_t * queue_id_ptr,
    ti_em_queue_id_t private_event_free_queue,
    uint32_t * private_event_number_ptr
)
{
    ti_em_queue_id_t lvChainingQueueBaseIdx;
    ti_em_dma_id_t lvChainingDmaIdx;
    em_status_t lvRetStatus;
    uint32_t lvIndex;
    volatile Emti_CdEventDsc_s *cdEventDscPtr;

    /*
     * skip initialization if no chaining information
     */
    if (!chain_config_ptr)
        return EM_ERR_NOT_FOUND;

    /*
     * open the chaining process Response (RP) Queue
     */
    lvRetStatus = ti_em_hw_queue_open (*queue_id_ptr);
    if (lvRetStatus != EM_OK)
        return EM_ERR_ALLOC_FAILED;
    Emti_svPcb.chainDsc.rpQueueIdx = (uint16_t) (*queue_id_ptr)++;

    Emti_logDbg ("[PKTDMA-CHAIN] process response queue 0x%x \n", Emti_svPcb.chainDsc.rpQueueIdx);

    /*
     * chaining process command (CD) Queue
     */
    lvRetStatus = ti_em_hw_queue_open (*queue_id_ptr);
    if (lvRetStatus != EM_OK)
        return EM_ERR_ALLOC_FAILED;

    /*
     * store cd queue idx
     */
    Emti_svPcb.chainDsc.cdQueueIdx = (uint16_t) * queue_id_ptr;
    Emti_svPcb.shadowRDsc.cdQueueIdx = (uint16_t) (*queue_id_ptr)++;
    Emti_queueEmpty (Emti_svPcb.chainDsc.cdQueueIdx);

    Emti_logDbg ("[PKTDMA-CHAIN] chaining command Queue 0x%x \n", Emti_svPcb.chainDsc.cdQueueIdx);

    /*
     * chaining process RX Queue
     */
    lvRetStatus = ti_em_hw_queue_open (*queue_id_ptr);
    if (lvRetStatus != EM_OK)
        return EM_ERR_ALLOC_FAILED;

    /*
     * store cd queue idx
     */
    Emti_svPcb.chainDsc.rxQueueIdx = (uint16_t) * queue_id_ptr;
    Emti_svPcb.shadowRDsc.rxQueueIdx = (uint16_t) (*queue_id_ptr)++;
    Emti_queueEmpty (Emti_svPcb.chainDsc.rxQueueIdx);

    lvChainingQueueBaseIdx = chain_config_ptr->dma_tx_queue_base_idx;
    lvChainingDmaIdx = chain_config_ptr->dma_idx;

    /*
     * chaining process command free Queue
     */
    lvRetStatus = ti_em_hw_queue_open (*queue_id_ptr);
    if (lvRetStatus != EM_OK)
        return EM_ERR_ALLOC_FAILED;

    /*
     * store cd queue idx
     */
    Emti_svPcb.chainDsc.ctCommandFreeQueueIdx = (uint16_t) (*queue_id_ptr)++;
    Emti_queueEmpty (Emti_svPcb.chainDsc.ctCommandFreeQueueIdx);

    /*
     * DMA Chaining process TX Queue
     */
    for (lvIndex = 0; lvIndex < TI_EM_CHAIN_TX_QUEUE_NUM; lvIndex++)
    {
        ti_em_queue_id_t lvDmaTxQueueId =
            ti_em_get_absolute_queue_id (lvChainingDmaIdx, lvChainingQueueBaseIdx + lvIndex);
        lvRetStatus = ti_em_hw_queue_open (lvDmaTxQueueId);
        if (lvRetStatus != EM_OK)
            return EM_ERR_ALLOC_FAILED;

        /*
         * store cd queue idx
         */
        if (lvIndex == 0)
        {
            Emti_svPcb.chainDsc.txQueueBaseIdx = (uint16_t) lvDmaTxQueueId;
            Emti_svPcb.shadowRDsc.txQueueBaseIdx = (uint16_t) lvDmaTxQueueId;
            Emti_svPcb.shadowRChainTxQueueBaseIdx = (uint16_t) lvDmaTxQueueId;
        }
        Emti_queueEmpty (lvDmaTxQueueId);
    }

    Emti_logDbg ("[PKTDMA-CHAIN]  Tx Queue base 0x%x \n", Emti_svPcb.chainDsc.txQueueBaseIdx);

    /*
     * process chain RX Flows
     */
    for (lvIndex = 0; lvIndex < TI_EM_CHAIN_TX_QUEUE_NUM; lvIndex++)
    {
        lvRetStatus =
            ti_em_chain_rx_flow_open (lvChainingDmaIdx,
				      lvChainingQueueBaseIdx + lvIndex,
                                      Emti_svPcb.chainDsc.rxQueueIdx,
                                      Emti_svPcb.poolConfigTbl[Emti_svPcb.chainDsc.poolIdxSizeA].free_queue_idx,
                                      Emti_svPcb.poolConfigTbl[Emti_svPcb.chainDsc.poolIdxSizeA].buf_size,
                                      Emti_svPcb.poolConfigTbl[Emti_svPcb.chainDsc.poolIdxSizeB].free_queue_idx,
                                      Emti_svPcb.poolConfigTbl[Emti_svPcb.chainDsc.poolIdxSizeB].buf_size,
                                      Emti_svPcb.poolConfigTbl[Emti_svPcb.chainDsc.poolIdxSizeC].free_queue_idx,
                                      Emti_svPcb.poolConfigTbl[Emti_svPcb.chainDsc.poolIdxSizeC].buf_size,
                                      Emti_svPcb.poolConfigTbl[Emti_svPcb.chainDsc.poolIdxSizeD].free_queue_idx,
                                      Emti_svPcb.poolConfigTbl[Emti_svPcb.chainDsc.poolIdxSizeD].buf_size,
                                      Emti_svPcb.poolConfigTbl[Emti_svPcb.chainDsc.poolIdxOverflow].free_queue_idx,
				      1);
        if (lvRetStatus != EM_OK)
        {
            Emti_logError ("[PKTDMA-CHAIN] Error process chain Opening Rx flow\n");
            return EM_ERR;
        }
    }
    Emti_logDbg ("[PKTDMA-CHAIN] Opened process chain Rx flow: %d\n", lvChainingQueueBaseIdx);

    /*
     * process chain TX Channel
     */
    for (lvIndex = 0; lvIndex < TI_EM_CHAIN_TX_QUEUE_NUM; lvIndex++)
    {
        lvRetStatus = ti_em_tx_channel_open (lvChainingDmaIdx, lvChainingQueueBaseIdx + lvIndex);
        if (lvRetStatus != EM_OK)
        {
            Emti_logError ("[PKTDMA-CHAIN] Error Opening process chain Tx channel\n");
            return EM_ERR;
        }
    }
    Emti_logDbg ("[PKTDMA-CHAIN] Opened process chain Tx channel: %d\n", lvChainingQueueBaseIdx);

    /*
     * process chain RX Channels
     */
    for (lvIndex = 0; lvIndex < TI_EM_CHAIN_TX_QUEUE_NUM; lvIndex++)
    {
        lvRetStatus = ti_em_rx_channel_open (lvChainingDmaIdx, lvChainingQueueBaseIdx + lvIndex);
        if (lvRetStatus != EM_OK)
        {
            Emti_logError ("[PKTDMA-CHAIN] Error Opening process chain Rx channel\n");
            return EM_ERR;
        }
    }
    Emti_logDbg ("[PKTDMA-CHAIN] Opened process chain Rx channel: %d\n", lvChainingQueueBaseIdx);

    Emti_svPcb.chainDsc.dma_idx = chain_config_ptr->dma_idx;
    Emti_svPcb.chainDsc.dma_queue_base_idx = chain_config_ptr->dma_tx_queue_base_idx;

    /*
     * Pop off a private event dsc from the queue
     */
    cdEventDscPtr = (volatile Emti_CdEventDsc_s *) Emti_popEvent (private_event_free_queue, TI_EM_MEM_PRIVATE_DESC);

    if (cdEventDscPtr == NULL)
        return EM_ERR_ALLOC_FAILED;

    /*
     * Increase number of used private events
     */
    (*private_event_number_ptr)++;

    /*
     * Now push this descriptor into the CT command free queue
     */
    Emti_pushEvent (Emti_svPcb.chainDsc.ctCommandFreeQueueIdx, (uint32_t *) cdEventDscPtr, TI_EM_MEM_PRIVATE_DESC);

    return lvRetStatus;
}

static em_status_t
Emti_pktdmaLocalInit (
    void
)
{
    Emti_logDbg ("[PKTDMA-CHAIN] local init\n");
    return EM_OK;
}

static em_status_t
Emti_pktdmaGlobalExit (
    void
)
{
    ti_em_queue_id_t lvChainingQueueBaseIdx;
    ti_em_dma_id_t lvChainingDmaIdx;
    em_status_t lvRetStatus;
    uint32_t lvIndex;


    Emti_logDbg ("[PKTDMA-CHAIN] global exit\n");

    /*
     * open the chaining process Response (RP) Queue
     */
    lvRetStatus = ti_em_hw_queue_close (Emti_svPcb.chainDsc.rpQueueIdx);
    if (lvRetStatus != EM_OK)
        return EM_ERR_ALLOC_FAILED;

    /*
     * chaining process command (CD) Queue
     */
    lvRetStatus = ti_em_hw_queue_close (Emti_svPcb.chainDsc.cdQueueIdx);
    if (lvRetStatus != EM_OK)
        return EM_ERR_ALLOC_FAILED;

    /*
     * chaining process RX Queue
     */
    lvRetStatus = ti_em_hw_queue_close (Emti_svPcb.chainDsc.rxQueueIdx);
    if (lvRetStatus != EM_OK)
        return EM_ERR_ALLOC_FAILED;

    /*
     * chaining process RX Queue
     */
    lvRetStatus = ti_em_hw_queue_close (Emti_svPcb.chainDsc.ctCommandFreeQueueIdx);
    if (lvRetStatus != EM_OK)
        return EM_ERR_ALLOC_FAILED;

    lvChainingQueueBaseIdx = Emti_svPcb.chainDsc.dma_queue_base_idx;
    lvChainingDmaIdx = Emti_svPcb.chainDsc.dma_idx;

    /*
     * process chain RX Flows
     */
    for (lvIndex = 0; lvIndex < TI_EM_CHAIN_TX_QUEUE_NUM; lvIndex++)
    {
        lvRetStatus = ti_em_rx_flow_close (lvChainingDmaIdx, lvChainingQueueBaseIdx + lvIndex);
        if (lvRetStatus != EM_OK)
            return EM_ERR_ALLOC_FAILED;
    }

    /*
     * DMA Chaining process TX Queue
     */
    for (lvIndex = 0; lvIndex < TI_EM_CHAIN_TX_QUEUE_NUM; lvIndex++)
    {
        lvRetStatus = ti_em_hw_queue_close (Emti_svPcb.chainDsc.txQueueBaseIdx + lvIndex);
        if (lvRetStatus != EM_OK)
            return EM_ERR_ALLOC_FAILED;
    }

    /*
     * process chain TX Channel
     */
    for (lvIndex = 0; lvIndex < TI_EM_CHAIN_TX_QUEUE_NUM; lvIndex++)
    {
        lvRetStatus = ti_em_tx_channel_close (lvChainingDmaIdx, lvChainingQueueBaseIdx + lvIndex);
        if (lvRetStatus != EM_OK)
        {
            return EM_ERR;
        }
    }

    /*
     * process chain RX Channels
     */
    for (lvIndex = 0; lvIndex < TI_EM_CHAIN_TX_QUEUE_NUM; lvIndex++)
    {
        lvRetStatus = ti_em_rx_channel_close (lvChainingDmaIdx, lvChainingQueueBaseIdx + lvIndex);
        if (lvRetStatus != EM_OK)
        {
            return EM_ERR;
        }
    }

    return lvRetStatus;
}

Emti_moduleChainRegister (TI_EM_CHAINING_PKTDMA, "PKTDMA-CHAIN", Emti_pktdmaGlobalInit, Emti_pktdmaLocalInit,
                          Emti_pktdmaConnect, Emti_pktdmaAddRoute, NULL, Emti_pktdmaGlobalExit, NULL, NULL, NULL)
