/*
 * Copyright (c) 2012, Texas Instruments Incorporated - http://www.ti.com/
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions are met:
 *       * Redistributions of source code must retain the above copyright
 *         notice, this list of conditions and the following disclaimer.
 *       * Redistributions in binary form must reproduce the above copyright
 *         notice, this list of conditions and the following disclaimer in the
 *         documentation and/or other materials provided with the distribution.
 *       * Neither the name of Texas Instruments Incorporated nor the
 *         names of its contributors may be used to endorse or promote products
 *         derived from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *   DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
 *   DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


/**
 *   @file  event_machine_hw_ti_chain_poststore.c
 *
 *   @brief   
 *
 */
#include <ti/runtime/openem/src/event_machine_private_types.h>
#include <ti/runtime/openem/src/event_machine_private_variables.h>
#include <ti/runtime/openem/src/event_machine_private_functions.h>
#include <ti/runtime/openem/modules/event_machine_modules.h>

/**
 * @internal
 *
 * Poststore disabled on Open-EM process.
 */
#define TI_EM_POSTSTORE_DISABLED          (0u)

/**
 * @internal
 *
 * Poststore enabled on Open-EM process.
 */
#define TI_EM_POSTSTORE_ENABLED           (1u)


static em_status_t
Emti_poststoreGlobalInit (
    ti_em_chain_config_t * chain_config_ptr,
    uint32_t * queue_id_ptr,
    ti_em_queue_id_t private_event_free_queue,
    uint32_t * private_event_number_ptr
)
{
    uint32_t lvFreeQueueIdx;
    uint32_t localEventNum = 0xffffffff;
    uint32_t lvQueueEntryCount;

    /*
     * skip initialization if no poststore chaining information
     */
    if ((!chain_config_ptr) || (!chain_config_ptr->poststore_config_ptr))
        return (em_status_t) (EM_ERR_NOT_FOUND);

    Emti_logDbg ("[POSTSTORE-CHAIN] global init\n");

    /*
     * copy the poststore configuration. 
     */
    Emti_svPcb.chainPoststoreDsc.poststoreSizeMax = chain_config_ptr->poststore_config_ptr->poststore_size_max;
    for (lvFreeQueueIdx = 0; lvFreeQueueIdx < 2 * TI_EM_CORE_NUM; lvFreeQueueIdx++)
    {
        Emti_svPcb.chainPoststoreDsc.localFreeQueueIdxTbl[lvFreeQueueIdx] =
            chain_config_ptr->poststore_config_ptr->local_free_queue_idx_tbl[lvFreeQueueIdx];
        Emti_svPcb.chainPoststoreDsc.localHeapPtrTbl[lvFreeQueueIdx] =
            chain_config_ptr->poststore_config_ptr->local_heap_ptr_tbl[lvFreeQueueIdx];

        /*
         * store minimal number of events in free queues 
         */
        lvQueueEntryCount = Emti_getQueueEntryCount (Emti_svPcb.chainPoststoreDsc.localFreeQueueIdxTbl[lvFreeQueueIdx]);
        if (lvQueueEntryCount < localEventNum)
            localEventNum = lvQueueEntryCount;
    }

    Emti_svPcb.chainPoststoreDsc.localEventNum = localEventNum;

    return (em_status_t) (EM_OK);
}

static em_status_t
Emti_poststoreLocalInit (
    void
)
{
    uint32_t lvFreeQueueIdx;
    uint32_t lvEventIdx;
    em_event_t lvEventHdl;
    Cppi_HostDesc *lvHostDescPtr;

    Emti_logDbg ("[POSTSTORE-CHAIN] local init\n");

    /*
     * copy Emti_ChainPoststorePcb_s to Emti_ChainPoststoreTcb_s 
     */
    Emti_gvTcb.chainPoststoreDsc.localEventNum = Emti_svPcb.chainPoststoreDsc.localEventNum;
    Emti_gvTcb.chainPoststoreDsc.localFreeQueueIdxTbl[0] =
        Emti_svPcb.chainPoststoreDsc.localFreeQueueIdxTbl[(DNUM * 2) + 0];
    Emti_gvTcb.chainPoststoreDsc.localFreeQueueIdxTbl[1] =
        Emti_svPcb.chainPoststoreDsc.localFreeQueueIdxTbl[(DNUM * 2) + 1];
    Emti_gvTcb.chainPoststoreDsc.localHeapPtrTbl[0] = Emti_svPcb.chainPoststoreDsc.localHeapPtrTbl[(DNUM * 2) + 0];
    Emti_gvTcb.chainPoststoreDsc.localHeapPtrTbl[1] = Emti_svPcb.chainPoststoreDsc.localHeapPtrTbl[(DNUM * 2) + 1];
    Emti_gvTcb.chainPoststoreDsc.poststoreSizeMax = Emti_svPcb.chainPoststoreDsc.poststoreSizeMax;

    /*
     * loop over all localEventNum events of all free queues 
     */
    for (lvFreeQueueIdx = 0; lvFreeQueueIdx < 2; lvFreeQueueIdx++)
    {
        for (lvEventIdx = 0; lvEventIdx < Emti_gvTcb.chainPoststoreDsc.localEventNum; lvEventIdx++)
        {
            lvEventHdl =
                (em_event_t) Emti_popEvent (Emti_gvTcb.chainPoststoreDsc.localFreeQueueIdxTbl[lvFreeQueueIdx],
                                            TI_EM_MEM_PUBLIC_DESC);
            lvHostDescPtr = Emti_getCppiHostDescPtr (lvEventHdl);
            if (lvHostDescPtr == NULL)
                return (em_event_t) EM_ERR_NOT_FOUND;

            /*
             * set buffer mode to TI_EM_BUF_MODE_TIGHT 
             */
            Emti_setBufMode (lvHostDescPtr, TI_EM_BUF_MODE_TIGHT);

            /*
             * configure buffer info 
             */
            {
                ti_em_buffer_config_t lvBufferConfig;

                lvBufferConfig.buffer_ptr = (void *) (Emti_gvTcb.chainPoststoreDsc.localHeapPtrTbl[lvFreeQueueIdx]);
                lvBufferConfig.buffer_size = (uint32_t) (Emti_gvTcb.chainPoststoreDsc.poststoreSizeMax);
                lvBufferConfig.orig_buffer_ptr =
                    (void *) (Emti_gvTcb.chainPoststoreDsc.localHeapPtrTbl[lvFreeQueueIdx]);
                lvBufferConfig.orig_buffer_size = (uint32_t) (Emti_gvTcb.chainPoststoreDsc.poststoreSizeMax);
                lvBufferConfig.coh_mode = (ti_em_coh_mode_t) (TI_EM_COH_MODE_OFF);
                lvBufferConfig.free_func = (ti_em_free_func_t) NULL;
                lvBufferConfig.free_push_policy = (ti_em_push_policy_t) TI_EM_PUSH_POLICY_TAIL;

                ti_em_packet_set_buffer_info ((ti_em_packet_t *) lvHostDescPtr, lvBufferConfig);
            }

            Emti_pushEvent (Emti_gvTcb.chainPoststoreDsc.localFreeQueueIdxTbl[lvFreeQueueIdx],
                            (uint32_t *) (lvEventHdl), TI_EM_MEM_PUBLIC_DESC);
        }
    }


    return (em_status_t) (EM_OK);
}

static em_status_t
Emti_poststoreGlobalExit (
    void
)
{

    Emti_logDbg ("[POSTSTORE-CHAIN] global exit\n");

    return (em_status_t) (EM_OK);
}

Emti_moduleChainRegister (TI_EM_CHAINING_POSTSTORE, "POSTSTORE-CHAIN", Emti_poststoreGlobalInit,
                          Emti_poststoreLocalInit, NULL, NULL, NULL, Emti_poststoreGlobalExit, NULL, NULL, NULL)
