/*
 * Copyright (c) 2012, Texas Instruments Incorporated - http://www.ti.com/
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions are met:
 *       * Redistributions of source code must retain the above copyright
 *         notice, this list of conditions and the following disclaimer.
 *       * Redistributions in binary form must reproduce the above copyright
 *         notice, this list of conditions and the following disclaimer in the
 *         documentation and/or other materials provided with the distribution.
 *       * Neither the name of Texas Instruments Incorporated nor the
 *         names of its contributors may be used to endorse or promote products
 *         derived from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *   DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
 *   DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


/**
 *   @file  event_machine_hw_ti_chain_rio.c
 *
 *   @brief   
 *
 */
#include <ti/runtime/openem/src/event_machine_private_types.h>
#include <ti/runtime/openem/src/event_machine_private_variables.h>
#include <ti/runtime/openem/src/event_machine_private_functions.h>
#include <ti/runtime/openem/modules/event_machine_modules.h>


/**
 * @ingroup DOC_api_specific
 *
 * Chaining over RIO disabled on Open-EM process.
 */
#define TI_EM_RIO_DISABLED (0u)

/**
 * @ingroup DOC_api_specific
 *
 * Chaining over RIO enabled on Open-EM process.
 */
#define TI_EM_RIO_ENABLED  (1u)

/**
 * @internal
 *
 * RIO Chain Header structure
 *
 */
struct Emti_RioChainHdr_s_
{
    uint16_t dstNodeIdx;
    uint16_t srcNodeIdx;
    uint8_t serviceClass;
    uint8_t transportInfo;
    uint16_t dstInfo;
};

/**
 * @internal
 *
 * RIO Chain Route Info structure
 */
struct Emti_RioChainRouteInfo_s_
{
    /*
     * this must always be the first field
     */
    uint8_t routeInfo;

    /*
     * these are the RIO specific fields
     */
    uint8_t rioNodeIdx[2];
};

typedef struct Emti_RioChainHdr_s_ Emti_RioChainHdr_s;
typedef struct Emti_RioChainRouteInfo_s_ Emti_RioChainRouteInfo_s;

/**
 * @internal
 *
 * This is the RIO chaining send method
 *
 */
static em_status_t
Emti_rioSend (
    em_event_t event,
    em_queue_t queue
)
{
    ti_em_device_id_t lvDeviceIdx;
    ti_em_queue_id_t lvQueueIdx;
    ti_em_process_id_t lvProcessIdx;

    Cppi_HostDesc *lvEventDscPtr;
    Emti_RioChainHdr_s *lvRioChainHdrPtr;
    Emti_RioChainRouteInfo_s *lvRioChainRouteInfoPtr;

    uint16_t lvHwQueueIdx;

    lvDeviceIdx = (ti_em_device_id_t) Emti_getField (queue, QUEUE_HANDLE_DEVICE_IDX);
    lvProcessIdx = (ti_em_process_id_t) Emti_getField (queue, QUEUE_HANDLE_PROCESS_IDX);
    lvQueueIdx = (ti_em_queue_id_t) Emti_getField (queue, QUEUE_HANDLE_QUEUE_IDX);

    lvEventDscPtr = (Cppi_HostDesc *) ((uint32_t) event & 0xFFFFFFF0);
    lvRioChainHdrPtr = (Emti_RioChainHdr_s *) (&lvEventDscPtr->psData);

    lvRioChainRouteInfoPtr = (Emti_RioChainRouteInfo_s *) & Emti_svPsbPtr->deviceRouteTbl[lvDeviceIdx].u.privateData[0];
    lvRioChainHdrPtr->srcNodeIdx = Emti_gvTcb.chainRioDsc.myNodeIdx[0] + (Emti_gvTcb.chainRioDsc.myNodeIdx[1] << 8);
    lvRioChainHdrPtr->dstNodeIdx = lvRioChainRouteInfoPtr->rioNodeIdx[0] + (lvRioChainRouteInfoPtr->rioNodeIdx[1] << 8);
    lvRioChainHdrPtr->dstInfo = 0;

    Emti_setField (lvRioChainHdrPtr->dstInfo, RIO_CHAIN_HDR_DST_INFO_PROCESS_IDX, lvProcessIdx);
    Emti_setField (lvRioChainHdrPtr->dstInfo, RIO_CHAIN_HDR_DST_INFO_QUEUE_IDX, lvQueueIdx);
    lvRioChainHdrPtr->serviceClass = Emti_gvTcb.chainRioDsc.serviceClass;
    lvRioChainHdrPtr->transportInfo = 0;
    Emti_setField (lvRioChainHdrPtr->transportInfo, RIO_CHAIN_HDR_TPT_INFO_TPT_TYPE, 1);
    Emti_setField (lvRioChainHdrPtr->transportInfo, RIO_CHAIN_HDR_TPT_INFO_COS_MASK, 1);
    lvHwQueueIdx = Emti_gvTcb.chainRioDsc.dmaTxQueueIdx;

    /*
     * set PS data Length (8 Bytes) 
     */
    lvEventDscPtr->packetInfo &= 0xF0FFFFFF;
    lvEventDscPtr->packetInfo |= 0x02000000;

    /*
     * set Packet Type to 30 (Type9)    
     */
    lvEventDscPtr->descInfo &= ~Emti_EVENT_SIZE_EVENT_TAG_MASK;
    lvEventDscPtr->descInfo |= (Emti_EVENT_TAG_RIO << Emti_EVENT_SIZE_EVENT_TAG_SHIFT);

    /*
     * do the cache coherency 
     */
    if (Emti_wbinv (event) == 0)
    {
        Emti_mfence ();
    }

    /*
     * set descriptor size  
     */
    lvEventDscPtr = (Cppi_HostDesc *) (((uint32_t) lvEventDscPtr & 0xFFFFFFF0) + Emti_DSC_HINT_RIO);

    /*
     * push Event to hw queue   
     */
    Emti_pushEvent (lvHwQueueIdx, (uint32_t *) lvEventDscPtr, TI_EM_MEM_PUBLIC_DESC);

    return EM_OK;
}

static em_status_t
Emti_rioAddRoute (
    ti_em_destination_id_t dest_id,
    void *route
)
{
    ti_em_device_rio_route_t *lvDeviceRoutePtr = (ti_em_device_rio_route_t *) route;
    ti_em_device_id_t lvDeviceIdx = (ti_em_device_id_t) dest_id;
    Emti_RioChainRouteInfo_s *lvRioChainRouteInfoPtr;

    lvRioChainRouteInfoPtr = (Emti_RioChainRouteInfo_s *) & Emti_svPsbPtr->deviceRouteTbl[lvDeviceIdx].u.privateData[0];

    Emti_logDbg ("[RIO-CHAIN] add route to %d\n", lvDeviceIdx);

    lvRioChainRouteInfoPtr->routeInfo = 0;
    Emti_setField (lvRioChainRouteInfoPtr->routeInfo, DEVICE_ROUTE_INFO_ROUTE_TYPE, TI_EM_CHAINING_RIO);
    Emti_setField (lvRioChainRouteInfoPtr->routeInfo, DEVICE_ROUTE_INFO_ENABLED, 1);


    lvRioChainRouteInfoPtr->rioNodeIdx[0] = lvDeviceRoutePtr->node_idx[0];
    lvRioChainRouteInfoPtr->rioNodeIdx[1] = lvDeviceRoutePtr->node_idx[1];

    return EM_OK;
}

static em_status_t
Emti_rioConnect (
    void
)
{
    em_status_t lvStatus;
    uint32_t lvSchedIdx = 0;

    Emti_logDbg ("[RIO-CHAIN] connect\n");

    /*
     * for (lvSchedIdx = 0; lvSchedIdx < Emti_svPcb.pdspNum; lvSchedIdx++) 
     */
    {
        lvStatus =
            Emti_connectFirmware (Emti_svPcb.cdQueueIdx[lvSchedIdx], Emti_svPcb.chainDsc.ctCommandFreeQueueIdx,
                                  Emti_svPcb.chainDsc.rpQueueIdx, (uint32_t) & Emti_svPcb.shadowRDsc,
                                  (uint32_t) sizeof (Emti_svPcb.shadowRDsc), (uint32_t) Emti_CtType_ROUTER_CONNECT_RIO,
                                  (uint32_t) TI_EM_SLOT_RPCB, (ti_em_process_id_t) 0);
    }
    return lvStatus;
}

static em_status_t
Emti_rioGlobalInit (
    ti_em_chain_config_t * chain_config_ptr,
    uint32_t * queue_id_ptr,
    ti_em_queue_id_t private_event_free_queue,
    uint32_t * private_event_number_ptr
)
{
    em_status_t lvRetStatus;

    /*
     * skip initialization if no RIO chaining information
     */
    if ((!chain_config_ptr) || (!chain_config_ptr->rio_config_ptr))
        return EM_ERR_NOT_FOUND;

    /*
     * copy the RIO configuration. 
     */
    Emti_svPcb.chainRioDsc.dmaIdx = chain_config_ptr->rio_config_ptr->dma_idx;
    Emti_svPcb.chainRioDsc.dmaTxQueueIdx = chain_config_ptr->rio_config_ptr->dma_tx_queue_idx;
    Emti_svPcb.chainRioDsc.serviceClass = chain_config_ptr->rio_config_ptr->service_class;
    Emti_svPcb.chainRioDsc.myNodeIdx[0] = chain_config_ptr->rio_config_ptr->my_node_idx[0];
    Emti_svPcb.chainRioDsc.myNodeIdx[1] = chain_config_ptr->rio_config_ptr->my_node_idx[1];

    /*
     * open RIO DMA Tx queue 
     */
    Emti_svPcb.chainRioDsc.dmaTxQueueIdx = ti_em_rio_tx_queue_open ((uint16_t) Emti_svPcb.chainRioDsc.dmaTxQueueIdx);

    Emti_logDbg ("RIO-CHAIN] RIO Tx DMA queue 0x%x \n", Emti_svPcb.chainRioDsc.dmaTxQueueIdx);

    /*
     * open RIO Rx DMA queue 
     */
    lvRetStatus = ti_em_hw_queue_open (*queue_id_ptr);
    if (lvRetStatus != EM_OK)
        return EM_ERR_ALLOC_FAILED;

    /*
     * store RIO queue idx in DSP SW context 
     */
    Emti_svPcb.chainRioDsc.rxQueueIdx = (uint16_t) * queue_id_ptr;
    Emti_svPcb.shadowRDsc.chainRioDsc.rxQueueIdx = (uint16_t) (*queue_id_ptr)++;
    Emti_queueEmpty (Emti_svPcb.chainRioDsc.rxQueueIdx);

    Emti_logDbg ("[RIO-CHAIN] RIO Rx DMA queue 0x%x \n", Emti_svPcb.chainDsc.rxQueueIdx);

    /*
     * open RIO Rx flow
     */
    {
        uint32_t lvRxFlowNb = Emti_svPcb.chainRioDsc.dmaTxQueueIdx - QMSS_SRIO_QUEUE_BASE;
        lvRetStatus =
            ti_em_rio_rx_flow_open (Emti_svPcb.chainRioDsc.dmaIdx,
				    lvRxFlowNb,
				    Emti_svPcb.chainRioDsc.rxQueueIdx,
                                    Emti_svPcb.poolConfigTbl[Emti_svPcb.chainDsc.poolIdxSizeA].free_queue_idx,
				    Emti_svPcb.poolConfigTbl[Emti_svPcb.chainDsc.poolIdxSizeA].buf_size,
                                    Emti_svPcb.poolConfigTbl[Emti_svPcb.chainDsc.poolIdxSizeB].free_queue_idx,
				    Emti_svPcb.poolConfigTbl[Emti_svPcb.chainDsc.poolIdxSizeB].buf_size,
                                    Emti_svPcb.poolConfigTbl[Emti_svPcb.chainDsc.poolIdxSizeC].free_queue_idx,
				    Emti_svPcb.poolConfigTbl[Emti_svPcb.chainDsc.poolIdxSizeC].buf_size,
                                    Emti_svPcb.poolConfigTbl[Emti_svPcb.chainDsc.poolIdxSizeD].free_queue_idx,
				    Emti_svPcb.poolConfigTbl[Emti_svPcb.chainDsc.poolIdxSizeD].buf_size,
                                    Emti_svPcb.poolConfigTbl[Emti_svPcb.chainDsc.poolIdxOverflow].free_queue_idx,
                                    Emti_svPcb.myDeviceIdx,
				    Emti_svPcb.myProcessIdx,
				    1);

        if (lvRetStatus != EM_OK)
        {
            Emti_logError ("[RIO-CHAIN] Error opening RIO Rx flow\n");
            return EM_ERR;
        }
        Emti_logDbg ("[RIO-CHAIN] Opened RIO Rx flow: %d\n", Emti_svPcb.myDeviceIdx);
    }

    /*
     * open RIO Tx channel
     */
    lvRetStatus = ti_em_rio_tx_channel_open (Emti_svPcb.chainRioDsc.dmaIdx);
    if (lvRetStatus != EM_OK)
    {
        Emti_logError ("[RIO-CHAIN] Error Opening RIO Tx channel\n");
        return EM_ERR;
    }
    Emti_svPcb.shadowRDsc.rioEnabled = TI_EM_RIO_ENABLED;

    Emti_logDbg ("[RIO-CHAIN] Opened RIO Tx channel: %d \n", Emti_svPcb.chainRioDsc.dmaIdx);

    return lvRetStatus;
}

static em_status_t
Emti_rioLocalInit (
    void
)
{
    Emti_logDbg ("[RIO-CHAIN] local init\n");

    /*
     * copy Emti_ChainRioPcb_s to Emti_ChainRioTcb_s 
     */
    Emti_gvTcb.chainRioDsc.dmaTxQueueIdx = Emti_svPcb.chainRioDsc.dmaTxQueueIdx;
    Emti_gvTcb.chainRioDsc.serviceClass = Emti_svPcb.chainRioDsc.serviceClass;
    Emti_gvTcb.chainRioDsc.myNodeIdx[0] = Emti_svPcb.chainRioDsc.myNodeIdx[0];
    Emti_gvTcb.chainRioDsc.myNodeIdx[1] = Emti_svPcb.chainRioDsc.myNodeIdx[1];

    return EM_OK;
}

static em_status_t
Emti_rioGlobalExit (
    void
)
{
    em_status_t lvRetStatus;

    Emti_logDbg ("[RIO-CHAIN] global exit\n");

    /*
     * close RIO DMA Tx queue 
     */
    lvRetStatus = ti_em_hw_queue_close (Emti_svPcb.chainRioDsc.dmaTxQueueIdx);
    if (lvRetStatus != EM_OK)
        return EM_ERR_ALLOC_FAILED;

    /*
     * close RIO Rx DMA queue 
     */
    lvRetStatus = ti_em_hw_queue_close (Emti_svPcb.chainRioDsc.rxQueueIdx);
    if (lvRetStatus != EM_OK)
        return EM_ERR_ALLOC_FAILED;

    /*
     * close RIO Rx flow
     */
    {
        uint32_t lvRxFlowNb = Emti_svPcb.chainRioDsc.dmaTxQueueIdx - QMSS_SRIO_QUEUE_BASE;
        lvRetStatus = ti_em_rx_flow_close (Emti_svPcb.chainRioDsc.dmaIdx, lvRxFlowNb);

        if (lvRetStatus != EM_OK)
        {
            return EM_ERR;
        }
    }

    /*
     * open RIO Tx channel
     */
    lvRetStatus = ti_em_tx_channel_close (Emti_svPcb.chainRioDsc.dmaIdx, 0);
    if (lvRetStatus != EM_OK)
    {
        return EM_ERR;
    }

    return lvRetStatus;
}

Emti_moduleChainRegister (TI_EM_CHAINING_RIO, "RIO-CHAIN", Emti_rioGlobalInit, Emti_rioLocalInit, Emti_rioConnect,
                          Emti_rioAddRoute, Emti_rioSend, Emti_rioGlobalExit, NULL, NULL, NULL)
