/*
 * Copyright (c) 2012, Texas Instruments Incorporated - http://www.ti.com/
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions are met:
 *       * Redistributions of source code must retain the above copyright
 *         notice, this list of conditions and the following disclaimer.
 *       * Redistributions in binary form must reproduce the above copyright
 *         notice, this list of conditions and the following disclaimer in the
 *         documentation and/or other materials provided with the distribution.
 *       * Neither the name of Texas Instruments Incorporated nor the
 *         names of its contributors may be used to endorse or promote products
 *         derived from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *   DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
 *   DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


/**
 *   @file  event_machine_hw_ti_chain_xge.c
 *
 *   @brief   
 *
 */
#include <ti/runtime/openem/src/event_machine_private_types.h>
#include <ti/runtime/openem/src/event_machine_private_variables.h>
#include <ti/runtime/openem/src/event_machine_private_functions.h>
#include <ti/runtime/openem/modules/event_machine_modules.h>

/**
 * @internal
 *
 * Chaining over XGE disabled on Open-EM process
 */
#define TI_EM_XGE_DISABLED          (0u)

/**
 * @internal
 *
 * Chaining over XGE enabled on Open-EM process
 */
#define TI_EM_XGE_ENABLED           (1u)

/**
 * @internal
 *
 * Push policy to head
 */
#define TI_EM_XGE_EVENT_SEQ_IDX_MAX (64u)

/**
 * @internal
 *
 * XGE Chain Header structure
 *
 */
struct Emti_XgeChainHdr_s_
{
    uint16_t streamIdx;
    uint16_t eventSize;
    uint8_t eventSeqIdx;
    uint8_t fragmentIdx;
    uint16_t dstInfo;
};

/**
 * @internal
 *
 * Ethernet Header structure
 *
 */
struct Emti_EtherHdr_s_
{
    uint8_t dstMacAddrB0;
    uint8_t dstMacAddrB1;
    uint8_t dstMacAddrB2;
    uint8_t dstMacAddrB3;
    uint8_t dstMacAddrB4;
    uint8_t dstMacAddrB5;
    uint8_t srcMacAddrB0;
    uint8_t srcMacAddrB1;
    uint8_t srcMacAddrB2;
    uint8_t srcMacAddrB3;
    uint8_t srcMacAddrB4;
    uint8_t srcMacAddrB5;
    uint8_t vlanTagB0;
    uint8_t vlanTagB1;
    uint8_t vlanTagB2;
    uint8_t vlanTagB3;
    uint8_t etherTypeB0;
    uint8_t etherTypeB1;
};

/**
 * @internal
 *
 * XGE Chain Route Info structure
 */
struct Emti_XgeChainRouteInfo_s_
{
    /*
     * this must always be the first field
     */
    uint8_t routeInfo;

    /*
     * these are the XGE specific fields
     */
    uint8_t xgeEventSeqIdx;
    uint16_t xgeFragmentSize;
    uint8_t xgeMacAddress[6];
    //acTicket302uint8_t xgeServiceType;
    uint8_t xgeVlanTag[4];
};

typedef struct Emti_XgeChainHdr_s_ Emti_XgeChainHdr_s;
typedef struct Emti_EtherHdr_s_ Emti_EtherHdr_s;
typedef struct Emti_XgeChainRouteInfo_s_ Emti_XgeChainRouteInfo_s;

/**
 * @internal
 *
 * Free header and fragment(s)
 *
 */
static inline void
Emti_fragmentFree (
    Cppi_HostDesc * lvHeaderDscPtr
)
{
    lvHeaderDscPtr->descInfo = 0;
    lvHeaderDscPtr = (Cppi_HostDesc *) (((uint32_t) lvHeaderDscPtr & 0xfffffff0) | Emti_DSC_HINT_RS);

    Emti_mfence ();

    Emti_pushEvent (Emti_svPcb.rsTxQueueIdx, (uint32_t *) lvHeaderDscPtr, TI_EM_MEM_PUBLIC_DESC);
}

/**
 * @internal
 *
 * Send header and fragment(s)
 *
 */
static inline void
Emti_fragmentSend (
    Cppi_HostDesc * lvHeaderDscPtr,
    uint16_t lvTxQueueIdx
)
{
    lvHeaderDscPtr = (Cppi_HostDesc *) (((uint32_t) lvHeaderDscPtr & 0xfffffff0) | Emti_DSC_HINT_XGE);
    Emti_mfence ();
    Emti_pushEvent (lvTxQueueIdx, (uint32_t *) lvHeaderDscPtr, TI_EM_MEM_PUBLIC_DESC);
}

/**
 * @internal
 *
 * Allocate a divert queue
 *
 */
static inline int
Emti_divertQueueGet (
    Emti_DvEventDsc_s ** lvDvEventDscPtr
)
{
    uint16_t lvDivertQueueIdx;

    *lvDvEventDscPtr =
        (Emti_DvEventDsc_s *) Emti_popEvent (Emti_gvTcb.chainXgeDsc.dvFreeQueueIdx, TI_EM_MEM_PRIVATE_DESC);

    if (*lvDvEventDscPtr == NULL)
	return -1;

    lvDivertQueueIdx = (uint16_t) (*lvDvEventDscPtr)->txDivertQueueIdx;

    return lvDivertQueueIdx;
}

/**
 * @internal
 *
 * Push header and fragment(s) into the divert queue
 *
 */
static inline void
Emti_divertQueuePush (
    Cppi_HostDesc * lvHeaderDscPtr,
    uint16_t lvDivertQueueIdx
)
{
    lvHeaderDscPtr = (Cppi_HostDesc *) (((uint32_t) lvHeaderDscPtr & 0xfffffff0) | Emti_DSC_HINT_XGE);
    Emti_pushEvent (lvDivertQueueIdx, lvHeaderDscPtr, TI_EM_MEM_PUBLIC_DESC);
}

/**
 * @internal
 *
 * Flush a given divert queue
 *
 */
static inline void
Emti_divertQueueFlush (
    Cppi_HostDesc * lvHeaderDscPtr,
    uint16_t lvDivertQueueIdx
)
{
    while (lvHeaderDscPtr != NULL)
    {
        /*
         * free header and fragment(s)
         */
        Emti_fragmentFree (lvHeaderDscPtr);

        /*
         * dequeue descriptors from the divertQueue
         */
        lvHeaderDscPtr =
            (Cppi_HostDesc *) ((uint32_t) Emti_popEvent (lvDivertQueueIdx, TI_EM_MEM_PUBLIC_DESC) & 0xfffffff0);
    }
}

/**
 * @internal
 *
 * Send divert queue
 *
 */
static inline void
Emti_divertQueueSend (
    uint16_t lvDivertQueueIdx,
    uint16_t lvTxQueueIdx
)
{
    /*
     * memory fence before pushing to divert: all modifications made to any descriptor(s) will be completed
     */
    Emti_mfence ();

    Emti_queueDivert (lvDivertQueueIdx, lvTxQueueIdx, 0);
}

/**
 * @internal
 *
 * Release a divert queue
 *
 */
static inline void
Emti_divertQueueRelease (
    Emti_DvEventDsc_s * lvDvEventDscPtr
)
{
    /*
     * push the DV event into the XGE divert free queue
     */
    Emti_mfence ();

    Emti_pushEvent (Emti_gvTcb.chainXgeDsc.dvFreeQueueIdx, (uint32_t *) lvDvEventDscPtr, TI_EM_MEM_PRIVATE_DESC);
}

/**
 * @internal
 *
 * This is the XGE chaining send method
 *
 */
static em_status_t
Emti_xgeSend (
    em_event_t event,
    em_queue_t queue
)
{
    ti_em_device_id_t lvDeviceIdx;
    ti_em_queue_id_t lvQueueIdx;
    ti_em_process_id_t lvProcessIdx;

    Cppi_HostDesc *lvEventDscPtr;
    Cppi_HostDesc *lvCompletionDscPtr;
    Cppi_HostDesc *lvHeaderDscPtr = NULL;
    Cppi_HostDesc *lvPreviousDscPtr;
    Cppi_HostDesc *lvCurrentDscPtr = NULL;
    Emti_EtherHdr_s *lvEtherHdrPtr;
    Emti_XgeChainHdr_s *lvXgeChainHdrPtr;
    Emti_XgeChainRouteInfo_s *lvXgeChainRouteInfoPtr;
    Emti_DvEventDsc_s *lvDvEventDscPtr = NULL;
    uint32_t *lvW32Ptr;
    uint32_t lvPacketSize;
    uint32_t lvBufferPtr;
    uint32_t lvEventSeqIdx;
    uint32_t lvVlanTag;
    uint16_t lvEventSize;
    uint16_t lvEventSizeLeft;
    uint16_t lvFragmentSizeLeft;
    uint16_t lvFragmentSizeUsed;
    uint16_t lvBufferSizeLeft;
    uint16_t lvBufferSize;
    uint16_t lvDivertQueueIdx = 0xffff;
    uint16_t lvHeaderFreeQueueIdx;
    uint16_t lvFragmentFreeQueueIdx;
    uint16_t lvTxQueueIdx;
    uint16_t lvFragmentSize;
    uint8_t lvVlanPrio;
    uint8_t lvFragmentIndex;

    lvEventDscPtr = Emti_getCppiHostDescPtr (event);
    lvHeaderFreeQueueIdx = Emti_gvTcb.chainXgeDsc.txHeaderFreeQueueIdx;
    lvFragmentFreeQueueIdx = Emti_gvTcb.chainXgeDsc.txFragmentFreeQueueIdx;

    lvDeviceIdx = (ti_em_device_id_t) Emti_getField (queue, QUEUE_HANDLE_DEVICE_IDX);
    lvProcessIdx = (ti_em_process_id_t) Emti_getField (queue, QUEUE_HANDLE_PROCESS_IDX);
    lvQueueIdx = (ti_em_queue_id_t) Emti_getField (queue, QUEUE_HANDLE_QUEUE_IDX);

    lvXgeChainRouteInfoPtr = (Emti_XgeChainRouteInfo_s *) & Emti_svPsbPtr->deviceRouteTbl[lvDeviceIdx].u.privateData[0];

    lvVlanTag =
        (uint32_t) lvXgeChainRouteInfoPtr->xgeVlanTag[0] << 24 | (uint32_t) lvXgeChainRouteInfoPtr->
        xgeVlanTag[1] << 16 | (uint32_t) lvXgeChainRouteInfoPtr->
        xgeVlanTag[2] << 8 | (uint32_t) lvXgeChainRouteInfoPtr->xgeVlanTag[3];
    lvVlanPrio = (uint8_t) ((lvVlanTag >> 13) & 0x07);
    lvTxQueueIdx = Emti_gvTcb.chainXgeDsc.txQueueBaseIdx + lvVlanPrio;

    lvEventSize = (uint16_t) lvEventDscPtr->descInfo & 0x003FFFFF;  /* read 21 LSB of descInfo (CPPI W0) */
    lvFragmentSize = lvXgeChainRouteInfoPtr->xgeFragmentSize;

    /*
     * check if we must use a divert queue
     */
    if (lvEventSize > lvFragmentSize)
    {
	int lvDivertQueueRes;
        lvDivertQueueRes = Emti_divertQueueGet (&lvDvEventDscPtr);
        if (lvDivertQueueRes < 0)
        {
            /*
             * there is no more divert queue available 
             */
            Emti_logDbg ("[XGE-CHAIN] no free divert queue available\n");
            return EM_ERR_NOT_FOUND;
        }
	lvDivertQueueIdx = (uint16_t) lvDivertQueueRes;
    }

    /*
     * do the cache coherency before touching the original event
     */
    Emti_wbinv (event);

    lvEventSeqIdx = Emti_gvTcb.myProcessIdx * TI_EM_XGE_EVENT_SEQ_IDX_MAX + lvXgeChainRouteInfoPtr->xgeEventSeqIdx;
    lvFragmentIndex = 0;
    lvEventSizeLeft = lvEventSize;

    lvBufferPtr = lvEventDscPtr->buffPtr;
    lvBufferSizeLeft = (uint16_t) lvEventDscPtr->buffLen;

    while (lvEventSizeLeft > 0)
    {
        lvHeaderDscPtr =
            (Cppi_HostDesc *) ((uint32_t) Emti_popEvent (lvHeaderFreeQueueIdx, TI_EM_MEM_PUBLIC_DESC) & 0xfffffff0);

        if (lvHeaderDscPtr == 0)
        {
            if (lvEventSize > lvFragmentSize)
            {
                /*
                 * dequeue all events in the divert queue
                 */
                while ((lvHeaderDscPtr =
                        (Cppi_HostDesc *) ((uint32_t) Emti_popEvent (lvDivertQueueIdx, TI_EM_MEM_PUBLIC_DESC) &
                                           0xfffffff0)) != NULL)

                    Emti_fragmentFree (lvHeaderDscPtr);
            }
            return (em_status_t) EM_ERR_NOT_FOUND;
        }

        lvPreviousDscPtr = lvHeaderDscPtr;
        lvFragmentSizeLeft = lvFragmentSize;

        if (lvEventSizeLeft < lvFragmentSizeLeft)
        {
            lvFragmentSizeLeft = lvEventSizeLeft;
        }

        lvFragmentSizeUsed = 0;

        while (lvFragmentSizeLeft > 0)
        {
            lvBufferSize = (lvFragmentSizeLeft < lvBufferSizeLeft) ? lvFragmentSizeLeft : lvBufferSizeLeft;
            lvBufferSizeLeft -= lvBufferSize;
            lvFragmentSizeLeft -= lvBufferSize;
            lvFragmentSizeUsed += lvBufferSize;
            lvEventSizeLeft -= lvBufferSize;

            if (lvBufferSizeLeft > 0)
            {
                lvCurrentDscPtr =
                    (Cppi_HostDesc *) ((uint32_t) Emti_popEvent (lvFragmentFreeQueueIdx, TI_EM_MEM_PUBLIC_DESC) &
                                       0xfffffff0);

                if (lvCurrentDscPtr == 0)
                {
                    if (lvEventSize > lvFragmentSize)
                    {
                        Emti_divertQueueFlush (lvHeaderDscPtr, lvDivertQueueIdx);
                    }
                    return (em_status_t) EM_ERR_NOT_FOUND;
                }

                lvCurrentDscPtr->buffPtr = lvBufferPtr;
                lvCurrentDscPtr->buffLen = lvBufferSize;
                lvBufferPtr += lvBufferSize;
                lvEventDscPtr->buffPtr = lvBufferPtr;
                lvEventDscPtr->buffLen = lvBufferSizeLeft;

            }
            else
            {
                lvCurrentDscPtr = lvEventDscPtr;
                lvEventDscPtr = (Cppi_HostDesc *) Emti_publicDesc_ptov (lvEventDscPtr->nextBDPtr);
                if (lvEventDscPtr)
                {
                    lvBufferPtr = lvEventDscPtr->buffPtr;
                    lvBufferSizeLeft = lvEventDscPtr->buffLen;
                }
            }
            lvPreviousDscPtr->nextBDPtr = Emti_publicDesc_vtop (lvCurrentDscPtr);
            lvPreviousDscPtr = lvCurrentDscPtr;
        }

        /*
         * SDS pseudo code 20130911
         * if eventSizeLeft > 0
         *  set completionDscPtr = 0
         * else // exception
         *  set completionDscPtr = eventDscPtr
         */
        if (lvEventSizeLeft > 0)
            lvCompletionDscPtr = NULL;
        else
            lvCompletionDscPtr = lvEventDscPtr;

        if (lvFragmentSizeUsed < TI_EM_XGE_PAYLOAD_SIZE_MIN)
        {
            lvCurrentDscPtr =
                (Cppi_HostDesc *) ((uint32_t) Emti_popEvent (lvFragmentFreeQueueIdx, TI_EM_MEM_PUBLIC_DESC) &
                                   0xfffffff0);

            if (lvCurrentDscPtr == 0)
            {
                if (lvEventSize > lvFragmentSize)
                {
                    Emti_divertQueueFlush (lvHeaderDscPtr, lvDivertQueueIdx);
                }
                return (em_status_t) EM_ERR_NOT_FOUND;
            }

            /*
             * the frag buffer is in fact the unused part of the header buffer 
             */
            lvBufferPtr = lvHeaderDscPtr->buffPtr;
            lvBufferPtr += TI_EM_XGE_HEADER_SIZE;
            lvBufferSize = TI_EM_XGE_PAYLOAD_SIZE_MIN - lvFragmentSizeUsed;
            lvFragmentSizeUsed = TI_EM_XGE_PAYLOAD_SIZE_MIN;
            lvPreviousDscPtr->nextBDPtr = Emti_publicDesc_vtop (lvCurrentDscPtr);
            lvCurrentDscPtr->buffPtr = lvBufferPtr;
            lvCurrentDscPtr->buffLen = lvBufferSize;
        }

        /*
         * SDS pseudo code 20130911
         * write completionDscPtr to currentDscPtr->nextDscPtr
         */
        lvCurrentDscPtr->nextBDPtr = Emti_publicDesc_vtop (lvCompletionDscPtr);

        lvEtherHdrPtr = (Emti_EtherHdr_s *) Emti_getUncachedAlias((uint32_t)(Emti_bufferPtr_ptov(lvHeaderDscPtr)));

        lvEtherHdrPtr->dstMacAddrB0 = (uint8_t) (lvXgeChainRouteInfoPtr->xgeMacAddress[0]);
        lvEtherHdrPtr->dstMacAddrB1 = (uint8_t) (lvXgeChainRouteInfoPtr->xgeMacAddress[1]);
        lvEtherHdrPtr->dstMacAddrB2 = (uint8_t) (lvXgeChainRouteInfoPtr->xgeMacAddress[2]);
        lvEtherHdrPtr->dstMacAddrB3 = (uint8_t) (lvXgeChainRouteInfoPtr->xgeMacAddress[3]);
        lvEtherHdrPtr->dstMacAddrB4 = (uint8_t) (lvXgeChainRouteInfoPtr->xgeMacAddress[4]);
        lvEtherHdrPtr->dstMacAddrB5 = (uint8_t) (lvXgeChainRouteInfoPtr->xgeMacAddress[5]);
        lvEtherHdrPtr->srcMacAddrB0 = (uint8_t) (Emti_gvTcb.chainXgeDsc.myMacAddress[0]);
        lvEtherHdrPtr->srcMacAddrB1 = (uint8_t) (Emti_gvTcb.chainXgeDsc.myMacAddress[1]);
        lvEtherHdrPtr->srcMacAddrB2 = (uint8_t) (Emti_gvTcb.chainXgeDsc.myMacAddress[2]);
        lvEtherHdrPtr->srcMacAddrB3 = (uint8_t) (Emti_gvTcb.chainXgeDsc.myMacAddress[3]);
        lvEtherHdrPtr->srcMacAddrB4 = (uint8_t) (Emti_gvTcb.chainXgeDsc.myMacAddress[4]);
        lvEtherHdrPtr->srcMacAddrB5 = (uint8_t) (Emti_gvTcb.chainXgeDsc.myMacAddress[5]);
        lvEtherHdrPtr->vlanTagB0 = (uint8_t) (lvXgeChainRouteInfoPtr->xgeVlanTag[0]);
        lvEtherHdrPtr->vlanTagB1 = (uint8_t) (lvXgeChainRouteInfoPtr->xgeVlanTag[1]);
        lvEtherHdrPtr->vlanTagB2 = (uint8_t) (lvXgeChainRouteInfoPtr->xgeVlanTag[2]);
        lvEtherHdrPtr->vlanTagB3 = (uint8_t) (lvXgeChainRouteInfoPtr->xgeVlanTag[3]);
        lvEtherHdrPtr->etherTypeB0 = (uint8_t) (Emti_gvTcb.chainXgeDsc.etherType[0]);
        lvEtherHdrPtr->etherTypeB1 = (uint8_t) (Emti_gvTcb.chainXgeDsc.etherType[1]);
        lvEtherHdrPtr++;

        lvXgeChainHdrPtr = (Emti_XgeChainHdr_s *) lvEtherHdrPtr;

#if defined _BIG_ENDIAN

        //acTicket302lvXgeChainHdrPtr->streamIdx = ((uint16_t) (lvXgeChainRouteInfoPtr->xgeServiceType << 8) +
        //acTicket302                               (uint16_t) (Emti_gvTcb.myDeviceIdx));
        lvXgeChainHdrPtr->streamIdx = ((uint16_t) (Emti_gvTcb.chainXgeDsc.serviceType << 8) +
                                       (uint16_t) (Emti_gvTcb.myDeviceIdx));
        lvXgeChainHdrPtr->eventSize = (uint16_t) lvEventSize;
        lvXgeChainHdrPtr->eventSeqIdx = lvEventSeqIdx;
        lvXgeChainHdrPtr->fragmentIdx = lvFragmentIndex;
        lvXgeChainHdrPtr->dstInfo = 0;
        Emti_setField (lvXgeChainHdrPtr->dstInfo, XGE_CHAIN_HDR_DST_INFO_PROCESS_IDX, lvProcessIdx);
        Emti_setField (lvXgeChainHdrPtr->dstInfo, XGE_CHAIN_HDR_DST_INFO_QUEUE_IDX, lvQueueIdx);

#else /* _BIG_ENDIAN */

        /*
         * chain header must be converterted into big endian format on a little endian device 
         */
        //acTicket302lvXgeChainHdrPtr->streamIdx = ((uint16_t) (lvXgeChainRouteInfoPtr->xgeServiceType) +
        //acTicket302                               (uint16_t) (Emti_gvTcb.myDeviceIdx << 8));
        lvXgeChainHdrPtr->streamIdx = ((uint16_t) (Emti_gvTcb.chainXgeDsc.serviceType) +
                                       (uint16_t) (Emti_gvTcb.myDeviceIdx << 8));
        lvXgeChainHdrPtr->eventSize = (uint16_t) ((lvEventSize & 0xFF) << 8) + ((lvEventSize >> 8) & 0xFF);
        lvXgeChainHdrPtr->eventSeqIdx = lvEventSeqIdx;
        lvXgeChainHdrPtr->fragmentIdx = lvFragmentIndex;
        lvXgeChainHdrPtr->dstInfo = 0;
        Emti_setField (lvXgeChainHdrPtr->dstInfo, XGE_CHAIN_HDR_DST_INFO_PROCESS_IDX, lvProcessIdx);
        Emti_setField (lvXgeChainHdrPtr->dstInfo, XGE_CHAIN_HDR_DST_INFO_QUEUE_IDX, lvQueueIdx);
        lvXgeChainHdrPtr->dstInfo =
            ((uint16_t) (lvXgeChainHdrPtr->dstInfo & 0xFF) << 8) + (((uint16_t) lvXgeChainHdrPtr->dstInfo >> 8) & 0xFF);

#endif /* _BIG_ENDIAN */

        /*
         * set packet (headerPayload + eventPayload) size
         */
        lvPacketSize = TI_EM_XGE_HEADER_SIZE + lvFragmentSizeUsed;
        lvHeaderDscPtr->descInfo = (lvHeaderDscPtr->descInfo & 0xFFC00000) + lvPacketSize;

        /*
         * set the dst_lo tag to zero (cf XGE user's guide)
         */
        lvW32Ptr = (uint32_t *) (&(lvHeaderDscPtr->tagInfo));
        Emti_setField (*lvW32Ptr, EVENT_DSC_W1_DST_TAG_LO, 0);

        /*
         * set the PS flag to zero
         */
        lvW32Ptr = (uint32_t *) (&(lvHeaderDscPtr->packetInfo));
        Emti_setField (*lvW32Ptr, EVENT_DSC_W2_PS_FLAG, 0);

        if (lvEventSize > lvFragmentSize)
        {
            /*
             * divert the fragment descriptors
             */
            Emti_divertQueuePush (lvHeaderDscPtr, lvDivertQueueIdx);
        }

        lvFragmentIndex = (lvFragmentIndex + 1) & 0xFF;

    }                           /* end of while(lvEventSizeLeft>0) */

    if (lvEventSize > lvFragmentSize)
    {
        /*
         * divert the queue
         */
        Emti_divertQueueSend (lvDivertQueueIdx, lvTxQueueIdx);

        /*
         * release the divert queue
         */
        Emti_divertQueueRelease (lvDvEventDscPtr);
    }
    else
    {
        /*
         * send the unfragmented event
         */
        Emti_fragmentSend (lvHeaderDscPtr, lvTxQueueIdx);
    }

    lvXgeChainRouteInfoPtr->xgeEventSeqIdx = (lvXgeChainRouteInfoPtr->xgeEventSeqIdx + 1) % TI_EM_XGE_EVENT_SEQ_IDX_MAX;

    return EM_OK;
}

static em_status_t
Emti_xgeAddRoute (
    ti_em_destination_id_t dest_id,
    void *route
)
{
    ti_em_device_xge_route_t *lvDeviceRoutePtr = (ti_em_device_xge_route_t *) route;
    ti_em_device_id_t lvDeviceIdx = (ti_em_device_id_t) dest_id;
    Emti_XgeChainRouteInfo_s *lvXgeChainRouteInfoPtr;

    lvXgeChainRouteInfoPtr = (Emti_XgeChainRouteInfo_s *) & Emti_svPsbPtr->deviceRouteTbl[lvDeviceIdx].u.privateData[0];

    if (lvDeviceIdx >= TI_EM_DEVICE_NUM)
        return EM_ERR_BAD_ID;

    Emti_logDbg ("[XGE-CHAIN] add route to %d\n", lvDeviceIdx);

    lvXgeChainRouteInfoPtr->xgeFragmentSize = lvDeviceRoutePtr->fragmentSize;

    lvXgeChainRouteInfoPtr->xgeMacAddress[0] = lvDeviceRoutePtr->mac_address[0];
    lvXgeChainRouteInfoPtr->xgeMacAddress[1] = lvDeviceRoutePtr->mac_address[1];
    lvXgeChainRouteInfoPtr->xgeMacAddress[2] = lvDeviceRoutePtr->mac_address[2];
    lvXgeChainRouteInfoPtr->xgeMacAddress[3] = lvDeviceRoutePtr->mac_address[3];
    lvXgeChainRouteInfoPtr->xgeMacAddress[4] = lvDeviceRoutePtr->mac_address[4];
    lvXgeChainRouteInfoPtr->xgeMacAddress[5] = lvDeviceRoutePtr->mac_address[5];

    //acTicket302lvXgeChainRouteInfoPtr->xgeServiceType = lvDeviceRoutePtr->service_type;

    lvXgeChainRouteInfoPtr->xgeVlanTag[0] = lvDeviceRoutePtr->vlan_tag[0];
    lvXgeChainRouteInfoPtr->xgeVlanTag[1] = lvDeviceRoutePtr->vlan_tag[1];
    lvXgeChainRouteInfoPtr->xgeVlanTag[2] = lvDeviceRoutePtr->vlan_tag[2];
    lvXgeChainRouteInfoPtr->xgeVlanTag[3] = lvDeviceRoutePtr->vlan_tag[3];

    lvXgeChainRouteInfoPtr->xgeEventSeqIdx = 0;

    lvXgeChainRouteInfoPtr->routeInfo = 0;
    Emti_setField (lvXgeChainRouteInfoPtr->routeInfo, DEVICE_ROUTE_INFO_ROUTE_TYPE, TI_EM_CHAINING_XGE);
    Emti_setField (lvXgeChainRouteInfoPtr->routeInfo, DEVICE_ROUTE_INFO_ENABLED, 1);

    return EM_OK;
}

static em_status_t
Emti_xgeConnect (
    void
)
{
    em_status_t lvStatus;
    em_status_t lvInitStatus;
    uint32_t vlanPriority;

    Emti_logDbg ("[XGE-CHAIN] connect\n");

    /*
     * The status can be :
     *  - EM_OK if it was the first ROUTER_INIT call
     *  - EM_ERR_NOT_FREE if the router as previously been initialized
     *  - EM_ERR_LIB_FAILED if an error occured during the router initialization
     */
    lvStatus =
        Emti_connectFirmware (Emti_svPcb.shadowRdcbDsc.cdQueueIdx, Emti_svPcb.chainXgeDsc.ctCommandFreeQueueIdx,
                              Emti_svPcb.chainXgeDsc.rpQueueIdx, (uint32_t) & Emti_svPcb.shadowRdcbDsc,
                              (uint32_t) sizeof (Emti_svPcb.shadowRdcbDsc), (uint32_t) Emti_CtType_ROUTER_INIT,
                              (uint32_t) TI_EM_SLOT_RDCB, Emti_svPcb.myProcessIdx);

    if (lvStatus != EM_ERR_LIB_FAILED)
    {
        lvInitStatus = lvStatus;
        lvStatus =
            Emti_connectFirmware (Emti_svPcb.shadowRdcbDsc.cdQueueIdx, Emti_svPcb.chainXgeDsc.ctCommandFreeQueueIdx,
                                  Emti_svPcb.chainXgeDsc.rpQueueIdx, (uint32_t) & Emti_svPcb.shadowRChainTxQueueBaseIdx,
                                  4, (uint32_t) Emti_CtType_ROUTER_CONNECT_XGE, (uint32_t) TI_EM_SLOT_RDCB,
                                  Emti_svPcb.myProcessIdx);

        if ((lvStatus != EM_ERR_LIB_FAILED) && (lvInitStatus != EM_ERR_NOT_FREE))
        {
            for (vlanPriority = 0; vlanPriority < TI_EM_XGE_VLAN_PRIO_NUM; vlanPriority++)
            {
                if ((Emti_svPcb.chainXgeDsc.vlanPrioMask >> vlanPriority) & 0x01)
                {
                    /*
                     * the flows, queues and channels will be opened according to vlanPriorityMask bits 
                     */

                    /*
                     * open XGE Rx flow
                     */
                    lvStatus =
                        ti_em_xge_rx_flow_open (Emti_svPcb.chainXgeDsc.dmaIdx, vlanPriority,
                                                Emti_svPcb.chainXgeDsc.rxQueueBaseIdx + vlanPriority,
                                                Emti_svPcb.chainXgeDsc.rxHeaderFreeQueueIdx,
                                                Emti_svPcb.chainXgeDsc.rxFragmentFreeQueueIdx, 1);
                    if (lvStatus != EM_OK)
                    {
                        Emti_logError ("[XGE-CHAIN] Error opening XGE Rx flow %d \n", vlanPriority);
                        return EM_ERR;
                    }
                    Emti_logDbg ("[XGE-CHAIN] Opened XGE Rx flow: %d\n", vlanPriority);

                    /*
                     * open XGE Tx channel
                     */
                    lvStatus = ti_em_xge_tx_channel_open (Emti_svPcb.chainXgeDsc.dmaIdx, vlanPriority);
                    if (lvStatus != EM_OK)
                    {
                        Emti_logError ("[XGE-CHAIN] Error Opening XGE Tx channel %d\n", vlanPriority);
                        return EM_ERR;
                    }
                    Emti_logDbg ("[XGE-CHAIN] Opened XGE Tx channel: %d\n", vlanPriority);

                    /*
                     * open XGE Rx channel
                     */
                    lvStatus = ti_em_xge_rx_channel_open (Emti_svPcb.chainXgeDsc.dmaIdx, vlanPriority);
                    if (lvStatus != EM_OK)
                    {
                        Emti_logError ("[XGE-CHAIN] Error Opening XGE Rx channel %d\n", vlanPriority);
                        return EM_ERR;
                    }
                    Emti_logDbg ("[XGE-CHAIN] Opened XGE Rx channel: %d\n", vlanPriority);
                }
            }
        }
    }

    return lvStatus;
}

static em_status_t
Emti_xgeGlobalInit (
    ti_em_chain_config_t * chain_config_ptr,
    uint32_t * queue_id_ptr,
    ti_em_queue_id_t private_event_free_queue,
    uint32_t * private_event_number_ptr
)
{
    em_status_t lvRetStatus;
    uint32_t index;
    uint32_t vlanPriority;

    uint32_t lvPrivateEventPtr;
    ti_em_queue_id_t lvCdDeviceQueueIdx;
    Emti_DvEventDsc_s *dvEventDscPtr;
    Emti_CdEventDsc_s *cdEventDscPtr;

    /*
     * skip initialization if no XGE chaining information
     */
    if ((!chain_config_ptr) || (!chain_config_ptr->xge_config_ptr))
        return EM_ERR_NOT_FOUND;

    /*
     * copy the XGE configuration
     */
    Emti_svPcb.chainXgeDsc.dmaIdx = chain_config_ptr->xge_config_ptr->dma_idx;
    Emti_svPcb.chainXgeDsc.etherType[0] = chain_config_ptr->xge_config_ptr->ether_type[0];
    Emti_svPcb.chainXgeDsc.etherType[1] = chain_config_ptr->xge_config_ptr->ether_type[1];

    for (index = 0; index < 6; index++)
    {
        Emti_svPcb.chainXgeDsc.myMacAddress[index] = chain_config_ptr->xge_config_ptr->my_mac_address[index];
    }

    Emti_svPcb.chainXgeDsc.rxHeaderFreeQueueIdx = chain_config_ptr->xge_config_ptr->rx_header_free_queue_idx;
    Emti_svPcb.chainXgeDsc.rxFragmentFreeQueueIdx = chain_config_ptr->xge_config_ptr->rx_fragment_free_queue_idx;
    Emti_svPcb.chainXgeDsc.rxMissQueueBaseIdx = chain_config_ptr->xge_config_ptr->rx_miss_queue_base_idx;
    Emti_svPcb.chainXgeDsc.txDivertQueueBaseIdx = chain_config_ptr->xge_config_ptr->tx_divert_queue_base_idx;
    Emti_svPcb.chainXgeDsc.txHeaderFreeQueueIdx = chain_config_ptr->xge_config_ptr->tx_header_free_queue_idx;
    Emti_svPcb.chainXgeDsc.txFragmentFreeQueueIdx = chain_config_ptr->xge_config_ptr->tx_fragment_free_queue_idx;
    Emti_svPcb.chainXgeDsc.vlanPrioMask = chain_config_ptr->xge_config_ptr->vlan_prio_mask;
    Emti_svPcb.chainXgeDsc.pdspIdx = chain_config_ptr->xge_config_ptr->pdsp_idx;
    Emti_svPcb.chainXgeDsc.serviceType = chain_config_ptr->xge_config_ptr->service_type;

    /*
     * fill shadow Rdsb structure
     */
    Emti_svPcb.shadowRdcbDsc.etherType[0] = chain_config_ptr->xge_config_ptr->ether_type[0];
    Emti_svPcb.shadowRdcbDsc.etherType[1] = chain_config_ptr->xge_config_ptr->ether_type[1];

    for (index = 0; index < 6; index++)
    {
        Emti_svPcb.shadowRdcbDsc.myMacAddress[index] = chain_config_ptr->xge_config_ptr->my_mac_address[index];
    }

    Emti_svPcb.shadowRdcbDsc.rxFragmentFreeQueueIdx = chain_config_ptr->xge_config_ptr->rx_fragment_free_queue_idx;
    Emti_svPcb.shadowRdcbDsc.rxHeaderFreeQueueIdx = chain_config_ptr->xge_config_ptr->rx_header_free_queue_idx;
    Emti_svPcb.shadowRdcbDsc.rxMissQueueBaseIdx = chain_config_ptr->xge_config_ptr->rx_miss_queue_base_idx;
    Emti_svPcb.shadowRdcbDsc.vlanPrioMask = chain_config_ptr->xge_config_ptr->vlan_prio_mask;
    Emti_svPcb.shadowRdcbDsc.serviceType = chain_config_ptr->xge_config_ptr->service_type;

    /*
     * buffLen of Tx header descriptors is set to TI_EM_XGE_HEADER_SIZE 
     */
    {
        Cppi_HostDesc *lvHeaderDscPtr;
        uint16_t lvQueueIdx = chain_config_ptr->xge_config_ptr->tx_header_free_queue_idx;
        uint32_t lvElemNum = Emti_getQueueEntryCount (lvQueueIdx);

        for (index = 0; index < lvElemNum; index++)
        {
            lvHeaderDscPtr =
                (Cppi_HostDesc *) ((uint32_t) Emti_popEvent (lvQueueIdx, TI_EM_MEM_PUBLIC_DESC) & 0xfffffff0);
            lvHeaderDscPtr->buffLen = TI_EM_XGE_HEADER_SIZE;
            Emti_pushEvent (lvQueueIdx, (uint32_t *) lvHeaderDscPtr, TI_EM_MEM_PUBLIC_DESC);
        }
    }
    
    /*
     * store XGE tx queue base idx
     */
    Emti_svPcb.chainXgeDsc.txQueueBaseIdx = ti_em_xge_tx_queue_base_idx_get ();
    Emti_logDbg ("[XGE-CHAIN] XGE Tx DMA queue base idx [0-%d] 0x%x \n", TI_EM_XGE_VLAN_PRIO_NUM,
                 Emti_svPcb.chainXgeDsc.txQueueBaseIdx);
                 
    /*
     * get lvCdDeviceQueueIdx in pdsp ram
     * null will indicate that no xgeGlobalInit as been done so far
     */
    lvPrivateEventPtr = (uint32_t) Emti_getPdspDataRamBaseAddress (Emti_svPcb.chainXgeDsc.pdspIdx);
    lvCdDeviceQueueIdx = (ti_em_queue_id_t) * (uint16_t *) (lvPrivateEventPtr);
    
    if (lvCdDeviceQueueIdx == 0)
    {
        /*
         * Set txQueueBaseIdx in shadowRdcb descriptor
         */
        Emti_svPcb.shadowRdcbDsc.txQueueBaseIdx = Emti_svPcb.chainXgeDsc.txQueueBaseIdx;
        
        /*
         * store XGE rx queue base idx
         */
        Emti_svPcb.chainXgeDsc.rxQueueBaseIdx = (uint16_t) * queue_id_ptr;
        Emti_svPcb.shadowRdcbDsc.rxQueueBaseIdx = (uint16_t) * queue_id_ptr;
        Emti_queueEmpty (Emti_svPcb.chainXgeDsc.rxQueueBaseIdx);
        Emti_logDbg ("[XGE-CHAIN] XGE Rx DMA queue base idx [0-%d] 0x%x \n", TI_EM_XGE_VLAN_PRIO_NUM,
                     Emti_svPcb.chainXgeDsc.rxQueueBaseIdx);

        /*
         * reserve queue number from rxQueueBaseIdx to rxQueueBaseIdx + 8 (VLAN priority num) 
         */
        *(queue_id_ptr) += TI_EM_XGE_VLAN_PRIO_NUM;

        for (vlanPriority = 0; vlanPriority < TI_EM_XGE_VLAN_PRIO_NUM; vlanPriority++)
        {
            if ((Emti_svPcb.chainXgeDsc.vlanPrioMask >> vlanPriority) & 0x01)
            {
                /*
                 * open XGE DMA Tx queue
                 */
                lvRetStatus = ti_em_hw_queue_open (Emti_svPcb.chainXgeDsc.txQueueBaseIdx + vlanPriority);
                if (lvRetStatus != EM_OK)
                {
                    return EM_ERR_ALLOC_FAILED;
                }

                /*
                 * open XGE Rx DMA queue
                 */
                lvRetStatus = ti_em_hw_queue_open (Emti_svPcb.chainXgeDsc.rxQueueBaseIdx + vlanPriority);
                if (lvRetStatus != EM_OK)
                {
                    return EM_ERR_ALLOC_FAILED;
                }

                /*
                 * open Rx Miss queue
                 */
                lvRetStatus = ti_em_hw_queue_open (Emti_svPcb.chainXgeDsc.rxMissQueueBaseIdx + vlanPriority);
                if (lvRetStatus != EM_OK)
                {
                    return EM_ERR_ALLOC_FAILED;
                }
            }
        }
        Emti_logDbg ("[XGE-CHAIN] XGE Rx miss queue base idx [0-%d] 0x%x \n", TI_EM_XGE_VLAN_PRIO_NUM,
                     Emti_svPcb.chainXgeDsc.rxMissQueueBaseIdx);
    }

    /*
     * open XGE command return queue
     */
    lvRetStatus = ti_em_hw_queue_open (*queue_id_ptr);
    if (lvRetStatus != EM_OK)
        return EM_ERR_ALLOC_FAILED;

    /*
     * store XGE queue idx
     */
    Emti_svPcb.chainXgeDsc.rpQueueIdx = (uint16_t) (*queue_id_ptr)++;

    /*
     * open XGE command free queue
     */
    lvRetStatus = ti_em_hw_queue_open (*queue_id_ptr);
    if (lvRetStatus != EM_OK)
        return EM_ERR_ALLOC_FAILED;

    /*
     * store XGE command free queue idx
     */
    Emti_svPcb.chainXgeDsc.ctCommandFreeQueueIdx = (uint16_t) (*queue_id_ptr)++;

    /*
     * pop off a private event dsc from the free queue
     */
    cdEventDscPtr = (Emti_CdEventDsc_s *) Emti_popEvent (private_event_free_queue, TI_EM_MEM_PRIVATE_DESC);

    if (cdEventDscPtr == NULL)
        return EM_ERR_ALLOC_FAILED;

    /*
     * increase number of used private events
     */
    (*private_event_number_ptr)++;

    /*
     * now push this descriptor into the CT command free queue
     */
    Emti_pushEvent (Emti_svPcb.chainXgeDsc.ctCommandFreeQueueIdx, (uint32_t *) cdEventDscPtr, TI_EM_MEM_PRIVATE_DESC);

    /*
     * open XGE divert free queue
     */
    lvRetStatus = ti_em_hw_queue_open (*queue_id_ptr);
    if (lvRetStatus != EM_OK)
        return EM_ERR_ALLOC_FAILED;

    /*
     * store XGE divert free queue idx
     */
    Emti_svPcb.chainXgeDsc.dvFreeQueueIdx = (uint16_t) (*queue_id_ptr)++;

    /*
     * setup divert queues
     */
    if (chain_config_ptr->xge_config_ptr->tx_divert_queue_base_idx)
    {
        ti_em_queue_id_t lvDivertQueueIdx =
            (ti_em_queue_id_t) chain_config_ptr->xge_config_ptr->tx_divert_queue_base_idx;

        for (index = 0; index < TI_EM_XGE_TX_DIVERT_QUEUE_NUM; index++)
        {
            /*
             * open all the DMA queues
             */
            lvRetStatus = ti_em_hw_queue_open (lvDivertQueueIdx);
            if (lvRetStatus != EM_OK)
                return EM_ERR_ALLOC_FAILED;

            /*
             * store base index queue
             */
            if (index == 0)
                Emti_svPcb.chainXgeDsc.txDivertQueueBaseIdx = lvDivertQueueIdx;

            /*
             * get a DV event for free queue
             */
            dvEventDscPtr = (Emti_DvEventDsc_s *) Emti_popEvent (private_event_free_queue, TI_EM_MEM_PRIVATE_DESC);

            if (dvEventDscPtr == NULL)
                return EM_ERR_ALLOC_FAILED;

            /*
             * Increase number of used private events
             */
            (*private_event_number_ptr)++;

            /*
             * store the physical base address of the DV event desc
             */
            if (index == 0)
                Emti_svPcb.chainXgeDsc.dvEventDscBasePtr = Emti_privateDesc_vtop ((void *) dvEventDscPtr);

            /*
             * store the divert queue index into the DV event
             */
            dvEventDscPtr->txDivertQueueIdx = (uint32_t) lvDivertQueueIdx;

            /*
             * Push the DV event into the XGE divert free queue
             */
            Emti_pushEvent (Emti_svPcb.chainXgeDsc.dvFreeQueueIdx, (uint32_t *) dvEventDscPtr, TI_EM_MEM_PRIVATE_DESC);

            lvDivertQueueIdx++;
            dvEventDscPtr++;
        }
        Emti_logDbg ("[XGE-CHAIN] XGE divert queue base idx [%d-%d] 0x%x\n", 0, TI_EM_XGE_TX_DIVERT_QUEUE_NUM,
                     Emti_svPcb.chainXgeDsc.txDivertQueueBaseIdx);
        Emti_logDbg ("[XGE-CHAIN] XGE divert free queue 0x%x\n", Emti_svPcb.chainXgeDsc.dvFreeQueueIdx);
    }

    if (lvCdDeviceQueueIdx == 0)
    {
        lvRetStatus = ti_em_hw_queue_open (*queue_id_ptr);
        if (lvRetStatus != EM_OK)
            return EM_ERR_ALLOC_FAILED;

        Emti_svPcb.chainXgeDsc.cdQueueIdx = (uint16_t) (*queue_id_ptr);
        Emti_svPcb.shadowRdcbDsc.cdQueueIdx = (uint16_t) (*queue_id_ptr)++;

        Emti_logDbg ("[XGE-CHAIN] XGE CD Queue 0x%x \n", Emti_svPcb.shadowRdcbDsc.cdQueueIdx);

        *(uint16_t *) (lvPrivateEventPtr) = (uint16_t) (Emti_svPcb.shadowRdcbDsc.cdQueueIdx);
    }
    else
    {
        Emti_svPcb.shadowRdcbDsc.cdQueueIdx = lvCdDeviceQueueIdx;
    }

    Emti_svPcb.shadowRDsc.xgeEnabled = TI_EM_XGE_ENABLED;

    return lvRetStatus;
}

static em_status_t
Emti_xgeLocalInit (
    void
)
{
    uint32_t index;

    Emti_logDbg ("[XGE-CHAIN] local init\n");

    /*
     * copy Emti_ChainXgePcb_s to Emti_ChainXgeTcb_s
     */
    Emti_gvTcb.chainXgeDsc.etherType[0] = Emti_svPcb.chainXgeDsc.etherType[0];
    Emti_gvTcb.chainXgeDsc.etherType[1] = Emti_svPcb.chainXgeDsc.etherType[1];

    for (index = 0; index < 6; index++)
    {
        Emti_gvTcb.chainXgeDsc.myMacAddress[index] = Emti_svPcb.chainXgeDsc.myMacAddress[index];
    }

    Emti_gvTcb.chainXgeDsc.txHeaderFreeQueueIdx = Emti_svPcb.chainXgeDsc.txHeaderFreeQueueIdx;
    Emti_gvTcb.chainXgeDsc.txFragmentFreeQueueIdx = Emti_svPcb.chainXgeDsc.txFragmentFreeQueueIdx;
    Emti_gvTcb.chainXgeDsc.txQueueBaseIdx = Emti_svPcb.chainXgeDsc.txQueueBaseIdx;
    Emti_gvTcb.chainXgeDsc.dvFreeQueueIdx = Emti_svPcb.chainXgeDsc.dvFreeQueueIdx;
    Emti_gvTcb.chainXgeDsc.serviceType = Emti_svPcb.chainXgeDsc.serviceType;

    return EM_OK;
}

static em_status_t
Emti_xgeGlobalExit (
    void
)
{

    em_status_t lvRetStatus;
    uint32_t index;
    uint32_t vlanPriority;

    Emti_logDbg ("[XGE-CHAIN] global exit\n");

    lvRetStatus =
        Emti_connectFirmware (Emti_svPcb.shadowRdcbDsc.cdQueueIdx, Emti_svPcb.chainXgeDsc.ctCommandFreeQueueIdx,
                              Emti_svPcb.chainXgeDsc.rpQueueIdx, (uint32_t) NULL, (uint32_t) 0,
                              (uint32_t) Emti_CtType_KILL, (uint32_t) TI_EM_SLOT_RPCB, Emti_svPcb.myProcessIdx);

    if (lvRetStatus != EM_OK)
        return EM_ERR_ALLOC_FAILED;

    /*
     *  retrieve the number of connected processes left
     */
    {
        Emti_CdEventDsc_s *cdEventDscPtr;
        uint8_t nb_process = 0;
        cdEventDscPtr =
            (Emti_CdEventDsc_s *) Emti_popEvent (Emti_svPcb.chainXgeDsc.ctCommandFreeQueueIdx, TI_EM_MEM_PRIVATE_DESC);

        if (cdEventDscPtr == NULL)
            return EM_ERR_LIB_FAILED;

        nb_process = (uint8_t) (cdEventDscPtr->returnQueue >> 24);
        if (nb_process == 0)
        {
            /*
             * close XGE Command queue
             */
            lvRetStatus = ti_em_hw_queue_close (Emti_svPcb.shadowRdcbDsc.cdQueueIdx);
            if (lvRetStatus != EM_OK)
                return EM_ERR_ALLOC_FAILED;
        }
    }

    for (vlanPriority = 0; vlanPriority < TI_EM_XGE_VLAN_PRIO_NUM; vlanPriority++)
    {
        if ((Emti_svPcb.chainXgeDsc.vlanPrioMask >> vlanPriority) & 0x01)
        {
            /*
             * close XGE DMA Tx queue
             */
            lvRetStatus = ti_em_hw_queue_close (Emti_svPcb.chainXgeDsc.txQueueBaseIdx + vlanPriority);
            if (lvRetStatus != EM_OK)
                return EM_ERR_ALLOC_FAILED;

            /*
             * close XGE Rx DMA queue
             */
            lvRetStatus = ti_em_hw_queue_close (Emti_svPcb.chainXgeDsc.rxQueueBaseIdx + vlanPriority);
            if (lvRetStatus != EM_OK)
                return EM_ERR_ALLOC_FAILED;

            /*
             * close Rx Miss queue
             */
            lvRetStatus = ti_em_hw_queue_close (Emti_svPcb.chainXgeDsc.rxMissQueueBaseIdx + vlanPriority);
            if (lvRetStatus != EM_OK)
                return EM_ERR_ALLOC_FAILED;

            /*
             * close XGE Rx flow
             */
            lvRetStatus = ti_em_xge_rx_flow_close (Emti_svPcb.chainXgeDsc.dmaIdx, vlanPriority);
            if (lvRetStatus != EM_OK)
            {
                return EM_ERR;
            }

            /*
             * close XGE Tx channel
             */
            lvRetStatus = ti_em_xge_tx_channel_close (Emti_svPcb.chainXgeDsc.dmaIdx, vlanPriority);
            if (lvRetStatus != EM_OK)
            {
                return EM_ERR;
            }

            /*
             * close XGE Rx channel
             */
            lvRetStatus = ti_em_xge_rx_channel_close (Emti_svPcb.chainXgeDsc.dmaIdx, vlanPriority);
            if (lvRetStatus != EM_OK)
            {
                return EM_ERR;
            }
        }
    }

    /*
     * close ct command free queue
     */
    lvRetStatus = ti_em_hw_queue_close (Emti_svPcb.chainXgeDsc.ctCommandFreeQueueIdx);
    if (lvRetStatus != EM_OK)
        return EM_ERR_ALLOC_FAILED;

    /*
     * close XGE divert free queue
     */
    lvRetStatus = ti_em_hw_queue_close (Emti_svPcb.chainXgeDsc.dvFreeQueueIdx);
    if (lvRetStatus != EM_OK)
        return EM_ERR_ALLOC_FAILED;

    /*
     * close divert queues
     */
    {
        uint16_t lvDivertQueueIdx = (uint16_t) Emti_svPcb.chainXgeDsc.txDivertQueueBaseIdx;
        for (index = 0; index < TI_EM_XGE_TX_DIVERT_QUEUE_NUM; index++)
        {
            lvRetStatus = ti_em_hw_queue_close (lvDivertQueueIdx);
            if (lvRetStatus != EM_OK)
                return EM_ERR_ALLOC_FAILED;

            lvDivertQueueIdx++;
        }
    }

    return lvRetStatus;
}

static em_status_t
Emti_xgeCounterGet (
    ti_em_counter_type_t type, 
    ti_em_counter_value_t * valuePtr,    
    int flag
)
{
    volatile Emti_CdEventDsc_s *cdEventDscPtr = NULL;

    /*
     * pop off the event from the CD free queue 
     */
    cdEventDscPtr = (volatile Emti_CdEventDsc_s *) Emti_popEvent (Emti_svPcb.chainXgeDsc.ctCommandFreeQueueIdx, TI_EM_MEM_PRIVATE_DESC);

    if (cdEventDscPtr == NULL)
        return (em_status_t)(EM_ERR_LIB_FAILED);
        
    /*
     * fill command
     */
    Emti_svPcb.shadowRCounterCommand = (flag <<16) | type;

    /*
     * fill event descriptor 
     */
    cdEventDscPtr->commandType = (Emti_CdQueueCommandType_e) ((uint32_t) Emti_CtType_ROUTER_COUNTER_GET | (Emti_svPcb.myProcessIdx << 16));
    cdEventDscPtr->commandPtr = Emti_pdspPut (TI_EM_SLOT_RPCB, (void *) (uint32_t) &(Emti_svPcb.shadowRCounterCommand), (uint32_t) sizeof (Emti_svPcb.shadowRCounterCommand));
    cdEventDscPtr->errorCode = Emti_CdQueueErrorCode_WAITING_CMD;
    cdEventDscPtr->returnQueue = (uint32_t) Emti_svPcb.chainXgeDsc.rpQueueIdx;

    /*
     * push event to the CD queue 
     */
    Emti_mfence ();
    Emti_pushEvent (Emti_svPcb.shadowRdcbDsc.cdQueueIdx, (uint32_t *) (((uint32_t) (cdEventDscPtr) & 0xfffffff0) | Emti_CD_COMMAND_TYPE_CT),
                    TI_EM_MEM_PRIVATE_DESC);

    /*
     * waits for the RP event 
     */
    cdEventDscPtr = NULL;
    while (cdEventDscPtr == NULL)
        cdEventDscPtr = (volatile Emti_CdEventDsc_s *) Emti_popEvent (Emti_svPcb.chainXgeDsc.rpQueueIdx, TI_EM_MEM_PRIVATE_DESC);

    cdEventDscPtr = (volatile Emti_CdEventDsc_s *) (((uint32_t) cdEventDscPtr & 0xFFFFFFF0));
    
    /*
     * read the event 
     */
    if (cdEventDscPtr->errorCode != Emti_CdQueueErrorCode_NO_ERROR)
        return (em_status_t)(EM_ERR_LIB_FAILED);

    /* 
     * return counter value
     */
    *valuePtr = (ti_em_counter_value_t)(cdEventDscPtr->commandPtr);

    /*
     * push to the CD free queue 
     */
    cdEventDscPtr->commandType = Emti_CtType_NULL;
    cdEventDscPtr->commandPtr = 0;
    Emti_mfence ();
    Emti_pushEvent (Emti_svPcb.chainXgeDsc.ctCommandFreeQueueIdx, (uint32_t *) cdEventDscPtr, TI_EM_MEM_PRIVATE_DESC);

    return (em_status_t)(EM_OK);
}

static em_status_t
Emti_xgeRxMissEnable (
    ti_em_xge_rx_miss_type_t miss_type 
)
{
    em_status_t lvRetStatus;  
    
    if (miss_type == ti_em_xge_rx_miss_type_ETHER_TYPE)
        Emti_svPcb.shadowRCounterCommand = Emti_XGE_RX_MISS_MASK_ETHER_TYPE_ON;
    
    if (miss_type == ti_em_xge_rx_miss_type_SERVICE_TYPE)
        Emti_svPcb.shadowRCounterCommand = Emti_XGE_RX_MISS_MASK_SERVICE_TYPE_ON;
      
    lvRetStatus =
    Emti_connectFirmware (Emti_svPcb.shadowRdcbDsc.cdQueueIdx, Emti_svPcb.chainXgeDsc.ctCommandFreeQueueIdx,
                          Emti_svPcb.chainXgeDsc.rpQueueIdx, (uint32_t) & Emti_svPcb.shadowRCounterCommand, (uint32_t) 4,
                          (uint32_t) Emti_CtType_ROUTER_ENABLE_RX_MISS, (uint32_t) TI_EM_SLOT_RPCB, Emti_svPcb.myProcessIdx);

    if (lvRetStatus != EM_OK)
        return EM_ERR_ALLOC_FAILED;

    return (em_status_t)(EM_OK);
}

static em_status_t
Emti_xgeRxMissDisable (
    ti_em_xge_rx_miss_type_t miss_type 
)
{
    em_status_t lvRetStatus;  
    
    if (miss_type == ti_em_xge_rx_miss_type_ETHER_TYPE)
        Emti_svPcb.shadowRCounterCommand = Emti_XGE_RX_MISS_MASK_ETHER_TYPE_ON;
    
    if (miss_type == ti_em_xge_rx_miss_type_SERVICE_TYPE)
        Emti_svPcb.shadowRCounterCommand = Emti_XGE_RX_MISS_MASK_SERVICE_TYPE_ON;
      
    lvRetStatus =
    Emti_connectFirmware (Emti_svPcb.shadowRdcbDsc.cdQueueIdx, Emti_svPcb.chainXgeDsc.ctCommandFreeQueueIdx,
                          Emti_svPcb.chainXgeDsc.rpQueueIdx, (uint32_t) & Emti_svPcb.shadowRCounterCommand, (uint32_t) 4,
                          (uint32_t) Emti_CtType_ROUTER_DISABLE_RX_MISS, (uint32_t) TI_EM_SLOT_RPCB, Emti_svPcb.myProcessIdx);

    if (lvRetStatus != EM_OK)
        return EM_ERR_ALLOC_FAILED;

    return (em_status_t)(EM_OK);
}


Emti_moduleChainRegister (TI_EM_CHAINING_XGE, "XGE-CHAIN", Emti_xgeGlobalInit, Emti_xgeLocalInit, Emti_xgeConnect,
                          Emti_xgeAddRoute, Emti_xgeSend, Emti_xgeGlobalExit, Emti_xgeCounterGet, Emti_xgeRxMissEnable,
                          Emti_xgeRxMissDisable)
