/*
 * Copyright (c) 2012, Texas Instruments Incorporated - http://www.ti.com/
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions are met:
 *       * Redistributions of source code must retain the above copyright
 *         notice, this list of conditions and the following disclaimer.
 *       * Redistributions in binary form must reproduce the above copyright
 *         notice, this list of conditions and the following disclaimer in the
 *         documentation and/or other materials provided with the distribution.
 *       * Neither the name of Texas Instruments Incorporated nor the
 *         names of its contributors may be used to endorse or promote products
 *         derived from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *   DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
 *   DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


/**
 *   @file  event_machine_modules.h
 *
 *   @brief
 *      Internal definition and functions for managing OpenEM modules
 *
 */

#ifndef EVENT_MACHINE_MODULES_H_
#define EVENT_MACHINE_MODULES_H_

#include <ti/runtime/openem/src/event_machine_private_functions.h>
#include <ti/runtime/openem/src/event_machine_private_macros.h>

typedef struct Emti_ModuleChainInfo_s_ Emti_ModuleChainInfo_s;

/**
 * @internal
 *
 * @brief
 *  The structure contains the chaining module information
 *
 * @details
 *  The structure is used to hold the information which describes an
 *  OpenEM chaining module.
 */
struct Emti_ModuleChainInfo_s_
{
    /**
     * @brief   module name
     */
    const char *name;

    /**
     * @brief   global_init method
     */
    em_status_t (*global_init) (
	ti_em_chain_config_t * chain_config_ptr,
	uint32_t * queue_id_ptr,
	ti_em_queue_id_t private_event_free_queue,
	uint32_t * private_event_number_ptr);

    /**
     * @brief   local_init method
     */
    em_status_t (*local_init) (void);

    /**
     * @brief   connect method
     */
    em_status_t (*connect) (void);

    /**
     * @brief   add_route method
     */
    em_status_t (*add_route) (
	ti_em_destination_id_t dest_id,
	void *route);

    /**
     * @brief   send method
     */
    em_status_t (*send) (
	em_event_t event,
	em_queue_t queue);

    /**
     * @brief   global_exit method
     */
    em_status_t (*global_exit) (void);
    
    /**
     * @brief   counter_get method
     */
    em_status_t (*counter_get)(
    ti_em_counter_type_t type, 
    ti_em_counter_value_t *valuePtr,
    int flag
    );    
    
    /**
     * @brief   rx_miss_enable method
     */
    em_status_t (*rx_miss_enable)(
    ti_em_xge_rx_miss_type_t miss_type
    );    
    
    /**
     * @brief   rx_miss_disable method
     */
    em_status_t (*rx_miss_disable)(
    ti_em_xge_rx_miss_type_t miss_type
    );    
};

/** 
 * @internal
 *
 * Chaining module information table
 */
extern Emti_ModuleChainInfo_s Emti_moduleChainInfo[Emti_CHAIN_MODULE_NUM];

/** 
 * @internal
 *
 * Chaining module register 
 */
#ifdef TI_EM_GCC
#define Emti_moduleChainRegister(__id, __name, __global_init, __local_init, __connect, __add_route, __send, __global_exit, __counter_get, __rx_miss_enable, __rx_miss_disable) \
    static void __attribute__((__constructor__)) __Emti_moduleChainRegisterFunc(void) \
    {									\
	Emti_moduleChainInfo[__id].name        = __name;		\
	Emti_moduleChainInfo[__id].global_init = __global_init;		\
	Emti_moduleChainInfo[__id].local_init  = __local_init;		\
	Emti_moduleChainInfo[__id].connect     = __connect;		\
	Emti_moduleChainInfo[__id].add_route   = __add_route;		\
	Emti_moduleChainInfo[__id].send        = __send;		\
	Emti_moduleChainInfo[__id].global_exit = __global_exit;		\
	Emti_moduleChainInfo[__id].counter_get = __counter_get;		\
	Emti_moduleChainInfo[__id].rx_miss_enable = __rx_miss_enable;		\
	Emti_moduleChainInfo[__id].rx_miss_disable = __rx_miss_disable;		\
    }

#define Emti_moduleAllInit()
#else /* TI_EM_GCC */
#define Emti_moduleChainRegister(__id, __name, __global_init, __local_init, __connect, __add_route, __send, __global_exit, __counter_get, __rx_miss_enable, __rx_miss_disable) \
    void __attribute__((__constructor__)) __Emti_moduleChainRegisterFunc_##__id(void) \
    {									\
	Emti_moduleChainInfo[__id].name        = __name;		\
	Emti_moduleChainInfo[__id].global_init = __global_init;		\
	Emti_moduleChainInfo[__id].local_init  = __local_init;		\
	Emti_moduleChainInfo[__id].connect     = __connect;		\
	Emti_moduleChainInfo[__id].add_route   = __add_route;		\
	Emti_moduleChainInfo[__id].send        = __send;		\
	Emti_moduleChainInfo[__id].global_exit = __global_exit;		\
	Emti_moduleChainInfo[__id].counter_get = __counter_get;		\
	Emti_moduleChainInfo[__id].rx_miss_enable = __rx_miss_enable;		\
	Emti_moduleChainInfo[__id].rx_miss_disable = __rx_miss_disable;		\
    }

#define MODULE_REGISTER_FN(__id) __Emti_moduleChainRegisterFunc_##__id

extern void MODULE_REGISTER_FN(TI_EM_CHAINING_PKTDMA) (void);
extern void MODULE_REGISTER_FN(TI_EM_CHAINING_XGE) (void);
extern void MODULE_REGISTER_FN(TI_EM_CHAINING_RIO) (void);
extern void MODULE_REGISTER_FN(TI_EM_CHAINING_POSTSTORE) (void);

#define Emti_moduleAllInit()					\
    do {							\
	MODULE_REGISTER_FN(TI_EM_CHAINING_PKTDMA) ();		\
	MODULE_REGISTER_FN(TI_EM_CHAINING_XGE) ();		\
	MODULE_REGISTER_FN(TI_EM_CHAINING_RIO) ();		\
	MODULE_REGISTER_FN(TI_EM_CHAINING_POSTSTORE) ();		\
    } while (0)
#endif /* TI_EM_GCC */

/** 
 * @internal
 *
 * Initialize all chaining modules (global)
 */
static inline em_status_t
Emti_moduleChainGlobalInit (
    ti_em_chain_config_t * chain_config_ptr,
    uint32_t * queue_id_ptr,
    ti_em_queue_id_t private_event_free_queue,
    uint32_t * private_event_number_ptr
)
{
    em_status_t res;
    uint32_t module;

    /*
     * Check if chaining information 
     */
    if (!chain_config_ptr)
    {
        Emti_svPcb.chainEnabled = TI_EM_CHAIN_DISABLED;
        return EM_ERR_NOT_FOUND;
    }

    Emti_moduleAllInit ();

    /*
     * Set generic chaining information 
     */
    Emti_svPcb.chainDsc.poolIdxSizeA = chain_config_ptr->pool_idx_size_a;
    Emti_svPcb.chainDsc.poolIdxSizeB = chain_config_ptr->pool_idx_size_b;
    Emti_svPcb.chainDsc.poolIdxSizeC = chain_config_ptr->pool_idx_size_c;
    Emti_svPcb.chainDsc.poolIdxSizeD = chain_config_ptr->pool_idx_size_d;
    Emti_svPcb.chainDsc.poolIdxOverflow = chain_config_ptr->pool_idx_overflow;

    for (module = 0; module < Emti_CHAIN_MODULE_NUM; module++)
    {
        if (Emti_moduleChainInfo[module].global_init)
        {
            res = Emti_moduleChainInfo[module].global_init (chain_config_ptr,
							    queue_id_ptr,
							    private_event_free_queue,
							    private_event_number_ptr);
            switch (res)
            {
                case EM_OK:
                    Emti_svPcb.chainDsc.moduleChainEnabled[module] = TI_EM_CHAIN_ENABLED;
                    break;

                case EM_ERR_NOT_FOUND:
                    Emti_svPcb.chainDsc.moduleChainEnabled[module] = TI_EM_CHAIN_DISABLED;
                    break;

                default:
                    Emti_logError ("Error in module %s global initialization\n", Emti_moduleChainInfo[module].name);

                    Emti_svPcb.chainDsc.moduleChainEnabled[module] = TI_EM_CHAIN_DISABLED;
                    return res;
            }
        }
    }

    /*
     * Set generic chaining enabled 
     */
    Emti_svPcb.chainEnabled = TI_EM_CHAIN_ENABLED;
    return EM_OK;
}

/** 
 * @internal
 *
 * Initialize all chaining modules (local)
 */
static inline em_status_t
Emti_moduleChainLocalInit (
    void
)
{
    em_status_t res;
    uint32_t module;

    if (Emti_svPcb.chainEnabled == TI_EM_CHAIN_DISABLED)
        return EM_ERR_NOT_FOUND;

    for (module = 0; module < Emti_CHAIN_MODULE_NUM; module++)
    {
        if ((Emti_moduleChainInfo[module].local_init) &&
            (Emti_svPcb.chainDsc.moduleChainEnabled[module] == TI_EM_CHAIN_ENABLED))
        {
            res = Emti_moduleChainInfo[module].local_init ();
            if ((res != EM_OK) && (res != EM_ERR_NOT_FOUND))
            {
                Emti_logError ("Error in module %s local initialization\n", Emti_moduleChainInfo[module].name);

                return res;
            }
        }
    }

    return EM_OK;
}

/** 
 * @internal
 *
 * Perform connect for all chaining modules
 */
static inline em_status_t
Emti_moduleChainConnect (
    void
)
{
    em_status_t res;
    uint32_t module;

    if (Emti_svPcb.chainEnabled == TI_EM_CHAIN_DISABLED)
        return EM_ERR_NOT_FOUND;

    for (module = 0; module < Emti_CHAIN_MODULE_NUM; module++)
    {
        if ((Emti_moduleChainInfo[module].connect) &&
            (Emti_svPcb.chainDsc.moduleChainEnabled[module] == TI_EM_CHAIN_ENABLED))
        {
            res = Emti_moduleChainInfo[module].connect ();
            if (res != EM_OK)
            {
                Emti_logError ("Error in module %s connect\n", Emti_moduleChainInfo[module].name);
                return res;
            }
        }
    }
    return EM_OK;
}

/** 
 * @internal
 *
 * Add route for a given chaining module
 */
static inline em_status_t
Emti_moduleChainAddRoute (
    int chain_id,
    ti_em_destination_id_t dest_id,
    void *route
)
{
    em_status_t res;

    if (Emti_svPcb.chainEnabled == TI_EM_CHAIN_DISABLED)
    {
        Emti_logError ("Error in module %s add route\n", Emti_moduleChainInfo[chain_id].name);
        return EM_ERR_LIB_FAILED;
    }

    if ((Emti_moduleChainInfo[chain_id].add_route) &&
        (Emti_svPcb.chainDsc.moduleChainEnabled[chain_id] == TI_EM_CHAIN_ENABLED))
    {
        res = Emti_moduleChainInfo[chain_id].add_route (dest_id, route);
        if ((res != EM_OK) && (res != EM_ERR_NOT_FOUND))
        {
            Emti_logError ("Error in module %s add route\n", Emti_moduleChainInfo[chain_id].name);
            return res;
        }
    }
    else
        return EM_ERR_NOT_FOUND;

    return EM_OK;
}

/** 
 * @internal
 *
 * Send method for a given chaining module
 */
static inline em_status_t
Emti_moduleChainSend (
    int chain_id,
    em_event_t event,
    em_queue_t queue
)
{
    em_status_t res = EM_ERR_NOT_FOUND;

    if ((Emti_moduleChainInfo[chain_id].send) &&
        (Emti_svPcb.chainDsc.moduleChainEnabled[chain_id] == TI_EM_CHAIN_ENABLED))
        res = Emti_moduleChainInfo[chain_id].send (event, queue);

    return res;
}

/** 
 * @internal
 *
 * Exit all chaining modules
 */
static inline em_status_t
Emti_moduleChainGlobalExit (
    void
)
{
    em_status_t res;
    uint32_t module;

    if (Emti_svPcb.chainEnabled == TI_EM_CHAIN_DISABLED)
        return EM_ERR_NOT_FOUND;

    for (module = 0; module < Emti_CHAIN_MODULE_NUM; module++)
    {
        if ((Emti_moduleChainInfo[module].global_exit) &&
            (Emti_svPcb.chainDsc.moduleChainEnabled[module] == TI_EM_CHAIN_ENABLED))
        {
            res = Emti_moduleChainInfo[module].global_exit ();
            if ((res != EM_OK) && (res != EM_ERR_NOT_FOUND))
            {
                Emti_logError ("Error in module %s exit procedure\n", Emti_moduleChainInfo[module].name);

                return res;
            }
        }
    }

    return EM_OK;
}

/** 
 * @internal
 *
 * Counter get for a given chaining module
 */
static inline em_status_t
Emti_moduleChainCounterGet (
    int chain_id,
    ti_em_counter_type_t type, 
    ti_em_counter_value_t * valuePtr,    
    int flag
)
{
    em_status_t res;

    if (Emti_svPcb.chainEnabled == TI_EM_CHAIN_DISABLED)
    {
        Emti_logError ("Error in module %s counter get\n", Emti_moduleChainInfo[chain_id].name);
        return (em_status_t)(EM_ERR_LIB_FAILED);
    }

    if ((Emti_moduleChainInfo[chain_id].counter_get) &&
        (Emti_svPcb.chainDsc.moduleChainEnabled[chain_id] == TI_EM_CHAIN_ENABLED))
    {
        res = Emti_moduleChainInfo[chain_id].counter_get(type, valuePtr, flag);
        if (res != EM_OK)
            return (em_status_t)(res);
    }
    else
        return (em_status_t)(EM_ERR_NOT_FOUND);

    return (em_status_t)(EM_OK);
}

/** 
 * @internal
 *
 * Rx miss enable for a given module
 */
static inline em_status_t
Emti_moduleChainRxMissEnable (
    int chain_id,
    ti_em_xge_rx_miss_type_t miss_type 
)
{
    em_status_t res;

    if (Emti_svPcb.chainEnabled == TI_EM_CHAIN_DISABLED)
    {
        Emti_logError ("Error in module %s rx miss enable\n", Emti_moduleChainInfo[chain_id].name);
        return (em_status_t)(EM_ERR_LIB_FAILED);
    }

    if ((Emti_moduleChainInfo[chain_id].rx_miss_enable) &&
        (Emti_svPcb.chainDsc.moduleChainEnabled[chain_id] == TI_EM_CHAIN_ENABLED))
    {
        res = Emti_moduleChainInfo[chain_id].rx_miss_enable(miss_type);
        if (res != EM_OK)
            return (em_status_t)(res);
    }
    else
        return (em_status_t)(EM_ERR_NOT_FOUND);

    return (em_status_t)(EM_OK);
}

/** 
 * @internal
 *
 * Rx miss disable for a given module
 */
static inline em_status_t
Emti_moduleChainRxMissDisable (
    int chain_id,
    ti_em_xge_rx_miss_type_t miss_type 
)
{
    em_status_t res;

    if (Emti_svPcb.chainEnabled == TI_EM_CHAIN_DISABLED)
    {
        Emti_logError ("Error in module %s rx miss disable\n", Emti_moduleChainInfo[chain_id].name);
        return (em_status_t)(EM_ERR_LIB_FAILED);
    }

    if ((Emti_moduleChainInfo[chain_id].rx_miss_disable) &&
        (Emti_svPcb.chainDsc.moduleChainEnabled[chain_id] == TI_EM_CHAIN_ENABLED))
    {
        res = Emti_moduleChainInfo[chain_id].rx_miss_disable(miss_type);
        if (res != EM_OK)
            return (em_status_t)(res);
    }
    else
        return (em_status_t)(EM_ERR_NOT_FOUND);

    return (em_status_t)(EM_OK);
}



#endif /* EVENT_MACHINE_MODULES_H_ */
