/*
 * Copyright (c) 2012, Texas Instruments Incorporated - http://www.ti.com/
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions are met:
 *       * Redistributions of source code must retain the above copyright
 *         notice, this list of conditions and the following disclaimer.
 *       * Redistributions in binary form must reproduce the above copyright
 *         notice, this list of conditions and the following disclaimer in the
 *         documentation and/or other materials provided with the distribution.
 *       * Neither the name of Texas Instruments Incorporated nor the
 *         names of its contributors may be used to endorse or promote products
 *         derived from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *   DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
 *   DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


/**
 *   @file  event_machine.c
 *
 *   @brief
 *      Open Event Machine Implementation for Keystone1 devices.
 *
 */

#include <stdio.h>
#include <inttypes.h>
#include <string.h>
#include <ti/runtime/openem/event_machine_macros.h>
#include <ti/runtime/openem/event_machine_hw_macros.h>
#include <ti/runtime/openem/event_machine_hw_ti_macros.h>
#include <ti/runtime/openem/event_machine_types.h>
#include <ti/runtime/openem/event_machine_hw_types.h>
#include <ti/runtime/openem/event_machine_hw_ti_types.h>
#include <ti/runtime/openem/event_machine_sw_ti_config.h>
#include <ti/runtime/openem/src/event_machine_hwpform.h>
#include <ti/runtime/openem/src/event_machine_private_macros.h>
#include <ti/runtime/openem/src/event_machine_private_types.h>
#include <ti/runtime/openem/src/event_machine_private_variables.h>
#include <ti/runtime/openem/src/event_machine_private_functions.h>
#include <ti/runtime/openem/event_machine_group.h>


em_event_t
em_alloc (
    size_t size,
    em_event_type_t type,
    em_pool_id_t pool_id
)
{
    Cppi_HostDesc *lvHostDescPtr;
    em_event_t lvEventHdl;
    uint8_t *lvWord2Ptr;
    uint32_t lvfreeQueueRegAdd = Emti_gvTcb.freeQueueRegDPtrTbl[pool_id];

#if (EM_CHECK_LEVEL >= 1)
    /*
     * Basic Validations: Ensure that the pool Id is in range. 
     */
    if (pool_id >= Emti_svPcb.poolNum)
    {
        Emti_error ((em_status_t) EM_ERR_BAD_ID, (em_escope_t) EM_ESCOPE_ALLOC);
        return EM_EVENT_UNDEF;
    }
#endif

    /*
     * Initialize the allocated descriptor to NULL. 
     */
    lvHostDescPtr = (Cppi_HostDesc *) Emti_publicDesc_ptov ((*((volatile uint32_t *) (lvfreeQueueRegAdd))));

    if (lvHostDescPtr == NULL)
        return (em_event_t) EM_EVENT_UNDEF;

    if (Emti_gvTcb.poolConfigTbl[pool_id].buf_mode == TI_EM_BUF_MODE_TIGHT)
    {
        /*
         * descriptor word 0 - ps location default is in descriptor. 
         */
        lvHostDescPtr->descInfo = (Cppi_PSLoc_PS_IN_DESC << 22) | (size & 0x3fffff);

        /*
         * descriptor word 3 - buffLen 
         */
        lvHostDescPtr->buffLen = size;

        /*
         * re-align the buffer pointer with the original buffer pointer 
         */
        lvHostDescPtr->buffPtr = lvHostDescPtr->origBuffPtr;
    }
    else if (Emti_gvTcb.poolConfigTbl[pool_id].buf_mode == TI_EM_BUF_MODE_LOOSE)
    {
#if (EM_CHECK_LEVEL >= 1)
        /*
         * Basic Validations: Ensure that the size is null. 
         */
        if (size != 0)
        {
            Emti_error ((em_status_t) EM_ERR_TOO_LARGE, (em_escope_t) EM_ESCOPE_ALLOC);
            return EM_EVENT_UNDEF;
        }
#endif

        Emti_setCohMode (lvHostDescPtr, (ti_em_coh_mode_t) (TI_EM_COH_MODE_OFF));

        lvHostDescPtr->buffPtr = 0;
        lvHostDescPtr->buffLen = 0;
        lvHostDescPtr->origBuffPtr = 0;
        lvHostDescPtr->origBufferLen &= 0xfffc00000;

        Emti_packet_set_free_function ((ti_em_packet_t *) (lvHostDescPtr), NULL);
        /*
         * descriptor word 0 - ps location default is in descriptor. 
         */
        lvHostDescPtr->descInfo = (Cppi_PSLoc_PS_IN_DESC << 22) | (0 & 0x3fffff);
    }

    /*
     * descriptor word 2 
     */
    lvWord2Ptr = (uint8_t *) & lvHostDescPtr->packetInfo;
#ifdef _LITTLE_ENDIAN
    lvWord2Ptr += 3;
#endif
    *lvWord2Ptr = 0x80;

    /*
     * descriptor word 5 
     */
    lvHostDescPtr->nextBDPtr = 0;

    /*
     * descriptor word 8 
     */
    lvHostDescPtr->timeStamp = (((uint8_t) type) << 24);

    /*
     * moving from host descriptor pointer to event handle 
     */
    lvEventHdl = (em_event_t) (((uint32_t) (lvHostDescPtr) & 0xfffffff0) | (Emti_EVENT_MODE_GLOBAL_HW));

    /*
     * Debug Message: 
     */
    Emti_logDbg ("Allocating descriptor %p\n", lvHostDescPtr);

#if (TI_EM_TRACE_LEVEL >=1)
    if (Emti_svTraceFunctionPtr != NULL)
    {
        Emti_svTraceFunctionPtr (TI_EM_TSCOPE_ALLOC, (em_event_t) (lvEventHdl), (uint32_t) (size),
                                 (em_event_type_t) (type), (em_pool_id_t) (pool_id));
    }
#endif /* (TI_EM_TRACE_LEVEL >=1) */

    /*
     * return event hdl 
     */
    return lvEventHdl;
}

void *
em_event_pointer (
    const em_event_t event
)
{
    Cppi_HostDesc *lvHostDescPtr = (Cppi_HostDesc *) Emti_getCppiHostDescPtr (event);

    /*
     * set buffer dirty flag 
     */
    Emti_setDirtyFlag (lvHostDescPtr, 0x1);

    /*
     * Return the pointer to the event 
     */
    return Emti_bufferPtr_ptov (lvHostDescPtr);
}

void
em_free (
    em_event_t event
)
{
    Cppi_HostDesc *lvHostDescPtr = (Cppi_HostDesc *) (Emti_getCppiHostDescPtr (event));
    Emti_event_mode_t lvEventMode = Emti_getEventModeFromEventHdl (event);

#if (EM_CHECK_LEVEL >= 1)
    /*
     * Basic Validation: Ensure that a valid event is being cleaned up. 
     */
    if (lvHostDescPtr == NULL)
        return;
#endif

#if (TI_EM_TRACE_LEVEL >=1)
    if (Emti_svTraceFunctionPtr != NULL)
        Emti_svTraceFunctionPtr (TI_EM_TSCOPE_FREE, (em_event_t) (event));
#endif /* (TI_EM_TRACE_LEVEL >=1) */

    /*
     * write Emti_EVENT_MODE_GLOBAL_HW to <current DSC>eventMode 
     */
    Emti_setEventMode (lvHostDescPtr, Emti_EVENT_MODE_GLOBAL_HW);

    /*
     * write 0xFF to <event DSC>srcTagLow 
     */
    lvHostDescPtr->tagInfo = (lvHostDescPtr->tagInfo & 0xff00ffff) | (0xff << 16);

    /*
     * Set event size to 0  
     */
    lvHostDescPtr->descInfo = 0 | (Cppi_PSLoc_PS_IN_DESC << 22);

    /*
     * manage cache coherency on first buffer 
     */
    if (Emti_inv (event) == 0)
        Emti_mfence ();

    /*
     * Emti_EVENT_MODE_GLOBAL_SW
     */
    if (lvEventMode == Emti_EVENT_MODE_GLOBAL_SW)
    {
        while (lvHostDescPtr != NULL)
        {
            uint32_t lvBufferFreeQueueIndex = Emti_getFreeQueueIndex (lvHostDescPtr);
            uint32_t lvBufferFreePushPolicy = Emti_getFreePushPolicy (lvHostDescPtr);
            uint32_t lvBufferMode = Emti_getBufMode (lvHostDescPtr);

            /*
             * manage free mode for first buffer 
             */
            if (lvBufferMode == TI_EM_BUF_MODE_LOOSE)
            {
                ti_em_free_func_t lvFreeFunctionPtr =
                    Emti_packet_get_free_function ((ti_em_packet_t *) (lvHostDescPtr));

                if (lvFreeFunctionPtr != NULL)
                {
                    /*
                     * write Emti_EVENT_MODE_GLOBAL_HW to <current DSC>eventMode 
                     */
                    Emti_setEventMode (lvHostDescPtr, Emti_EVENT_MODE_GLOBAL_HW);

                    /*
                     * call event buffer free function 
                     */
                    lvFreeFunctionPtr (Emti_origBufferPtr_ptov (lvHostDescPtr),
                                       lvHostDescPtr->origBufferLen & 0x3fffff);
                }
            }

            if (lvBufferFreePushPolicy == 1)
                Emti_pushEventToHead ((uint16_t) (lvBufferFreeQueueIndex), (uint32_t *) lvHostDescPtr,
                                      TI_EM_MEM_PUBLIC_DESC);
            else
                Emti_pushEvent ((uint16_t) (lvBufferFreeQueueIndex), (uint32_t *) lvHostDescPtr, TI_EM_MEM_PUBLIC_DESC);

            /*
             * go to next buffers 
             */
            lvHostDescPtr = (Cppi_HostDesc *) Emti_publicDesc_ptov (lvHostDescPtr->nextBDPtr);
        }
        return;
    }
    else
    {
        if (Emti_svPcb.dma_idx == (ti_em_dma_id_t) - 1)
        {
            /*
             * Manually free descriptors
             */
            while (lvHostDescPtr != NULL)
            {
                uint32_t lvBufferFreeQueueIndex = Emti_getFreeQueueIndex (lvHostDescPtr);
                uint32_t lvBufferFreePushPolicy = Emti_getFreePushPolicy (lvHostDescPtr);
                Cppi_HostDesc *lvNextBufferPtr = (Cppi_HostDesc *) Emti_publicDesc_ptov (lvHostDescPtr->nextBDPtr);

                if (lvBufferFreePushPolicy == 1)
                    Emti_pushEventToHead ((uint16_t) (lvBufferFreeQueueIndex), (uint32_t *) lvHostDescPtr,
                                          TI_EM_MEM_PUBLIC_DESC);
                else
                    Emti_pushEvent ((uint16_t) (lvBufferFreeQueueIndex), (uint32_t *) lvHostDescPtr,
                                    TI_EM_MEM_PUBLIC_DESC);

                /*
                 * go to next buffers 
                 */
                lvHostDescPtr = lvNextBufferPtr;
            }
        }
        else
        {
            /*
             * Use PktDMA for freeing descriptors
             */
            *((uint32_t *) (Emti_gvTcb.rsTxQueueRegDPtr)) =
                (uint32_t) Emti_publicDesc_vtop (lvHostDescPtr) | Emti_DSC_HINT_MAX;
        }
        return;
    }
}

em_event_type_t
em_get_type_major (
    em_event_type_t type
)
{
/**
 * \internal
 * Copyright (C) 2012 Texas Instruments Incorporated - http://www.ti.com/
 * 
 * Implementation change. Event type is only a byte in TI implementation.
 * Old implementation was: 
    return (type & 0xFF000000);
 * \endinternal
 */
    return (EM_EVENT_TYPE_UNDEF);
}

em_event_type_t
em_get_type_minor (
    em_event_type_t type
)
{
/**
 * \internal
 * Copyright (C) 2012 Texas Instruments Incorporated - http://www.ti.com/
 * 
 * Implementation change. Event type is only a byte in TI implementation.
 * Old implementation was: 
    return (type & 0x00FFFFFF);
 * \endinternal
 */
    return (EM_EVENT_TYPE_SW_DEFAULT);
}

int
em_core_id (
    void
)
{
    return DNUM;
}

int
em_core_count (
    void
)
{
    return ((int) Emti_svPsbPtr->coreNum);
}

int
em_core_id_get_physical (
    int core
)
{
    return (int) (core);
}

void
em_core_mask_get_physical (
    em_core_mask_t * phys,
    const em_core_mask_t * logic
)
{
    return;
}
