/*
 * Copyright (c) 2012, Texas Instruments Incorporated - http://www.ti.com/
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions are met:
 *       * Redistributions of source code must retain the above copyright
 *         notice, this list of conditions and the following disclaimer.
 *       * Redistributions in binary form must reproduce the above copyright
 *         notice, this list of conditions and the following disclaimer in the
 *         documentation and/or other materials provided with the distribution.
 *       * Neither the name of Texas Instruments Incorporated nor the
 *         names of its contributors may be used to endorse or promote products
 *         derived from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *   DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
 *   DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


/**
 *   @file  event_machine_dispatcher.c
 *
 *   @brief
 *      Open Event Machine Dispatcher Implementation for Keystone1 devices.
 *      Functions related to dispatching.
 *
 */

#include <stdio.h>
#include <inttypes.h>
#include <ti/runtime/openem/event_machine_macros.h>
#include <ti/runtime/openem/event_machine_types.h>
#include <ti/runtime/openem/event_machine_functions.h>
#include <ti/runtime/openem/event_machine_group.h>
#include <ti/runtime/openem/src/event_machine_private_types.h>
#include <ti/runtime/openem/src/event_machine_private_variables.h>
#include <ti/runtime/openem/src/event_machine_private_functions.h>
#include <ti/runtime/openem/modules/event_machine_modules.h>

/*
 **********************************************************************************************************************
 **********************************************************************************************************************
 */

void
Emti_dispatch_event_group (
    em_event_group_t eventGroup
)
{
    Emti_EventGroup_s *lvEventGroupPtr;
    uint8_t sendNotification;
    uint32_t index;

    /*
     * Get the event group. 
     */
    lvEventGroupPtr = &Emti_svEventGroupMdb[(uint32_t) eventGroup];

    /*
     * Critical Section Enter 
     */
    Emti_lockSemaphore ();

    /*
     * Decrement the count 
     */
    lvEventGroupPtr->eventNum--;

    /*
     * Have all the events in the group been handled? 
     */
    if (lvEventGroupPtr->eventNum == 0)
    {
        sendNotification = 1;
        lvEventGroupPtr->state = Emti_EventGroupState_ALLOCATED;
    }
    else
        sendNotification = 0;

    /*
     * Do we need to send out the notifications. 
     */
    if (sendNotification == 1)
    {
        /*
         * Yes. Send out all the notifications. 
         */
        for (index = 0; index < lvEventGroupPtr->notifNum; index++)
        {
            /*
             * Send out the internal event. 
             */
            em_send (lvEventGroupPtr->notifTbl[index].event, (em_queue_t) lvEventGroupPtr->notifTbl[index].queue);
        }
    }

    /*
     * Critical Section Exit 
     */
    Emti_unlockSemaphore ();

    return;
}

void
ti_em_preschedule (
    void
)
{
#if (TI_EM_TRACE_LEVEL >=1)
    if (Emti_svTraceFunctionPtr != NULL)
        Emti_svTraceFunctionPtr (TI_EM_TSCOPE_PRESCHEDULE, Emti_privateDesc_ptov ((Emti_gvTcb.pfCommandHdl)));
#endif /* (TI_EM_TRACE_LEVEL >=1) */

    if (!(Emti_gvTsb.actionFlags & Emti_PF_COMMAND_TO_SEND))
        return;

    *(uint32_t *) (Emti_gvTcb.cdQueueRegDPtr) = (uint32_t) (Emti_gvTcb.pfCommandHdl);

    /*
     * The queue has been processed. 
     */
    Emti_gvTsb.actionFlags &= Emti_PF_COMMAND_TO_CLEAR;

    return;
}

uint32_t
ti_em_atomic_processing_locality (
    void
)
{

    volatile uint32_t *lvHintPtr = &Emti_gvTsb.esbDscPtr->atomicLocalityHint;
    return *lvHintPtr;
}

em_status_t
ti_em_dispatch_once (
    void
)
{
    volatile Cppi_HostDesc *lvEventDscPtr;
    uint32_t *lvEpibPtr;
    uint64_t lvEpib_w0_w1;
    uint32_t lvActionFlags;
    Cppi_HostDesc *lvGlobalEventDscPtr;
    uint16_t lvEventHandleType;
    uint32_t lvEqHandle;
    em_event_t lvEventHandle;
    em_event_type_t lvEventType;
    em_event_group_t lvEventGroup;
#if (EM_CHECK_LEVEL >= 1)
    Emti_QueueSlow_s *lvQueueSlowPtr;
#endif
    uint64_t lvQueueFast64b;
    volatile uint32_t *lvAuxPtr;
    em_receive_func_t lvReceiveFuncPtr;

    {
        /*
         * We pop event from head of the PF Queue to get a PF Event or an AP Event 
         */
        lvEventDscPtr = (Cppi_HostDesc *) Emti_publicDesc_ptov ((*(volatile uint32_t *) (Emti_gvTcb.pfQueueRegDPtr)));
        if (lvEventDscPtr == NULL)
            return (em_status_t) EM_ERR_NOT_FOUND;
    }

    /*
     * ensure global event descriptor address written by the scheduler is complete 
     */
    {
        lvAuxPtr = &Emti_gvTsb.esbDscPtr->globalEventHdl;
        lvGlobalEventDscPtr = (Cppi_HostDesc *) Emti_getCppiHostDescPtr ((em_event_t) Emti_publicDesc_ptov (*lvAuxPtr));
        if (lvGlobalEventDscPtr == NULL)
        {
            do
            {
                lvGlobalEventDscPtr =
                    (Cppi_HostDesc *) Emti_getCppiHostDescPtr ((em_event_t) Emti_publicDesc_ptov (*lvAuxPtr));
            }
            while (lvGlobalEventDscPtr == NULL);
        }
    }

    /*
     * store global event free mode and coherent mode in a local variable for reuse a multiple places latter 
     */
    {
        lvEventHandleType = (uint16_t) (Emti_gvTsb.esbDscPtr->globalFreeInfo >> 24) & 0xf;
    }

    /*
     * Clear the atomic locality hint
     */
    {
        Emti_gvTsb.esbDscPtr->atomicLocalityHint = 0;
    }

    /*
     * retreive EQ handle, event group handle and event type 
     */
    {
        /*
         * Mask out the descriptor size for pointer processing 
         */
        lvEventDscPtr = (volatile Cppi_HostDesc *) Emti_getCppiHostDescPtr ((em_event_t) lvEventDscPtr);

        /*
         * Decode the host descriptor in order to retrieve the EM infos
         * * TIMESTAMP WORD AFTER SCHEDULER
         * * 31:24 : EVENT TYPE (with bits #31-30 = PRELOAD FLAG)
         * * 23:16 : SD QUEUE INDEX
         * *    15 : ATOMIC FLAG
         * * 14:00 : QUEUE HANDLE
         * * SW INFO [0] WORD AFTER SCHEDULER
         * * 15:00 : EVENT GROUP 
         */

        /*
         * Since host descriptors necessarily need to be 16-byte aligned
         * timestamp is 8-byte aligned
         * swInfo[1] is 8-byte aligned
         */

        lvEpibPtr = Emti_getTimeStampAddress ((Cppi_HostDesc *) lvEventDscPtr);
        lvEpib_w0_w1 = (uint64_t) (_mem8 (lvEpibPtr));
        lvEqHandle = (em_queue_t) Emti_LOW_LL (lvEpib_w0_w1);
        lvEventGroup = (em_event_group_t) Emti_HIGH_LL (lvEpib_w0_w1);
        lvEventType = (em_event_type_t) (lvEqHandle >> 24);
        Emti_gvTsb.eventGroupHdl = (lvEventGroup & 0xFFFF);
    }

    /*
     * reset global event to free flag 
     */
    {
        lvActionFlags = Emti_PF_COMMAND_TO_SEND;
    }

    /*
     * set <TSB>pingPong = (<TSB>pingPong + 1) % 2
     */
    Emti_gvTsb.pingPong = (Emti_gvTsb.pingPong + 1) % 2;

    /*
     * set <TSB>poststoreSize[<TSB>pingPong] = 0 
     */
    Emti_gvTsb.chainPoststoreDsc.poststoreSizeTbl[Emti_gvTsb.pingPong] = 0;


    /*
     * In case data preloading occurred:
     * *   - word #0 of original pointer has to be reset with original size
     * *   - word #2 2 LS Bytes with the free info.
     */
    if ((lvEventType & TI_EM_EVENT_TYPE_PRELOAD_MSK) != TI_EM_EVENT_TYPE_PRELOAD_OFF)
    {
        /*
         * set global event to free flag 
         */
        lvActionFlags |= Emti_LOCAL_EVENT_TO_CLAIM;

        /*
         * preload ON - build local event handle 
         */
        lvEventHandle = (em_event_t) (((uint32_t) (lvEventDscPtr) & 0xfffffff0) | lvEventHandleType);

        /*
         * restore <event DSC>eventSize from <running ECB>globalEventWord0 
         */
        lvEventDscPtr->descInfo = Emti_gvTsb.esbDsc.globalEventSize | (Cppi_PSLoc_PS_IN_DESC << 22);

        /*
         * restore <event DSC>returnPushPolicy and <event DSC>returnQueueIdx from <running ECB>globalEventWord12 
         */
        {
            volatile uint16_t *lvFreeInfoPtr = (uint16_t *) & lvEventDscPtr->packetInfo;
#ifdef _BIG_ENDIAN
            {
                lvFreeInfoPtr++;
            }
#endif
            *lvFreeInfoPtr = (uint16_t) (Emti_gvTsb.esbDsc.globalFreeInfo & 0xFFFF);
        }

        /*
         * reset local event to null
         */
        Emti_gvTcb.preloadDsc.localEventHdl = EM_EVENT_UNDEF;
    }
    else
    {
        /*
         * preload OFF - build global event 
         */
        lvEventHandle = (em_event_t) (((uint32_t) (lvEventDscPtr) & 0xfffffff0) | lvEventHandleType);
        /*
         * not strictly required ... 
         */
        Emti_gvTsb.esbDsc.globalEventHdl = (Emti_gvTsb.esbDsc.globalEventHdl & 0xfffffff0) | lvEventHandleType;
    }

    /*
     * Find out if queue is atomic 
     */
    if (Emti_isQueueAtomic (lvEqHandle) == 1)
    {
        lvActionFlags |= Emti_AP_COMMAND_TO_SEND;
    }

    Emti_gvTsb.actionFlags = lvActionFlags;

    /*
     * reset pending global event descriptor - this ensures that when a non null value is present, it is because the scheduler write has been completed 
     */
    {
        *lvAuxPtr = 0;
        Emti_gvTsb.mFence = *lvAuxPtr;
        Emti_mfence ();
    }

    /*
     * Remember the event queue which is being handled. 
     */
#if (EM_CHECK_LEVEL >= 1)
    lvQueueSlowPtr = &Emti_svQueueSlowMdb[Emti_getQueueIndex (lvEqHandle)];
#endif
    /*
     * load both the eoPtr and the queue context (struct aligned on 64 bits) 
     */
    lvQueueFast64b = (uint64_t) (_mem8 (&Emti_svQueueFastMdb[Emti_getQueueIndex (lvEqHandle)]));

    /*
     * store EO handle in global variable 
     */
#if (EM_CHECK_LEVEL >= 1)
    Emti_gvEoHandle = (em_eo_t) (lvQueueSlowPtr->eoPtr);
#endif /* (EM_CHECK_LEVEL >= 1) */

    /*
     * rebuild queue handle 
     */
    lvEqHandle = lvEqHandle & Emti_QUEUE_HANDLE_QUEUE_IDX_MASK;
    Emti_setField (lvEqHandle, QUEUE_HANDLE_DEVICE_IDX, Emti_gvTcb.myDeviceIdx);
    Emti_setField (lvEqHandle, QUEUE_HANDLE_PROCESS_IDX, Emti_gvTcb.myProcessIdx);

    /*
     * Original code left to remind us of the call parameters.
     * lvQueuePtr->eoPtr->receive((void*)(lvQueuePtr->eoPtr->contextPtr),
     *         (em_event_t)(lvEventHandle),
     *         (em_event_type_t)lvSwInfoPtr[0],
     *         (em_queue_t)queueHandle,
     *         (void*) lvQueuePtr->contextPtr);
     */

#if (TI_EM_TRACE_LEVEL >=1)
    if (Emti_svTraceFunctionPtr != NULL)
    {
        Emti_svTraceFunctionPtr (TI_EM_TSCOPE_DISPATCH, (void *) NULL, (em_event_t) (lvEventHandle),
                                 (em_event_type_t) (lvEventType), (em_queue_t) lvEqHandle,
                                 (void *) (Emti_HIGH_LL (lvQueueFast64b)));

    }
#endif /* (TI_EM_TRACE_LEVEL >=1) */

    /*
     * Flush the prefetch buffer 
     */
    Emti_prefetchBufferInv ();

    /****************************************************************************************
     * Optimization: The eo context is not anymore provided to the receive function.
     * - The idea was to remove the 2nd MSMC read to lvEo64b
     * - Only the queue context is provided. It's application's responsability to provide
     *   eo context in the queue context.
     * - The receive function is now part of the queue parameters.
     ****************************************************************************************/

    lvReceiveFuncPtr = ((em_receive_func_t) (Emti_LOW_LL (lvQueueFast64b)));

#if (EM_CHECK_LEVEL >= 1)
#define __RECEIVE_FUNC_COND lvReceiveFuncPtr
#else
#define __RECEIVE_FUNC_COND 1   /* always true */
#endif

    if (__RECEIVE_FUNC_COND)
    {
        ((em_receive_func_t) (lvReceiveFuncPtr)) ((void *) NULL, (em_event_t) lvEventHandle,
                                                  (em_event_type_t) lvEventType, (em_queue_t) lvEqHandle,
                                                  (void *) (Emti_HIGH_LL (lvQueueFast64b)));
    }
    else
        Emti_logDbg ("No receive function associated!! idx = 0x%x, Epib = %p, EventDsc = %p\n",
                     Emti_getQueueIndex (lvEqHandle), lvEpibPtr, lvEventDscPtr);

    /*
     * reset EO handle global variable 
     */
#if (EM_CHECK_LEVEL >= 1)
    Emti_gvEoHandle = (em_eo_t) (EM_EO_UNDEF);
#endif /* (EM_CHECK_LEVEL >= 1) */

    lvActionFlags = Emti_gvTsb.actionFlags;

    /*
     * We need to indicate the scheduler it can prefetch a job for that core 
     * Now push this descriptor into the prefetch SD queue. 
     */
    if (lvActionFlags & Emti_PF_COMMAND_TO_SEND)
    {
        *(uint32_t *) (Emti_gvTcb.cdQueueRegDPtr) = (uint32_t) (Emti_gvTcb.pfCommandHdl);
    }

    /*
     * if the application did not claim the global ptr 
     * we release it 
     */
    if (lvActionFlags & Emti_LOCAL_EVENT_TO_CLAIM)
    {
        if (Emti_gvTcb.preloadDsc.localEventHdl == EM_EVENT_UNDEF)
        {
            em_event_t localEventHdl = 0;
            while (localEventHdl == 0)
            {
                localEventHdl =
                    (em_event_t) Emti_popEvent (Emti_gvTcb.preloadDsc.returnQueueIdx, TI_EM_MEM_PUBLIC_DESC);
            }
            Emti_gvTcb.preloadDsc.localEventHdl = localEventHdl;
        }
        Emti_pushEvent (Emti_gvTcb.preloadDsc.localFreeQueueIdx, (void *) Emti_gvTcb.preloadDsc.localEventHdl,
                        TI_EM_MEM_PUBLIC_DESC);
    }

    /*
     * Warning: event may have been release and event references are not valid anymore 
     */
    /*
     * Indicate that the queue has done with the processing. This is required only for atomic queues 
     */
    if (lvActionFlags & Emti_AP_COMMAND_TO_SEND)
    {
        *(uint32_t *) (Emti_gvTcb.cdQueueRegDPtr) = (uint32_t) (Emti_gvTcb.apCommandHdlTbl[Emti_gvTsb.pingPong]);
    }

    /*
     * Check if the event belonged to a group or not? 
     */
    if (lvEventGroup != Emti_NO_EVENT_GROUP_TAG)
    {
        if ((lvEventGroup >> Emti_EVENT_GROUP_TAG_SHIFT) == Emti_EVENT_GROUP_TAG)
        {
            lvEventGroup = (lvEventGroup & Emti_EVENT_GROUP_ID_MASK);
            Emti_dispatch_event_group (lvEventGroup);
        }
    }

    return (em_status_t) EM_OK;
}

void *
ti_em_claim_local (
    void
)
{
    void *localBufferPtr = NULL;
    if (Emti_gvTsb.actionFlags & Emti_LOCAL_EVENT_TO_CLAIM)
    {
        if (Emti_gvTcb.preloadDsc.localEventHdl == EM_EVENT_UNDEF)
        {
            em_event_t localEventHdl = 0;
            Cppi_HostDesc *lvHostDescPtr;

            while (localEventHdl == 0)
            {
                localEventHdl =
                    (em_event_t) Emti_popEvent (Emti_gvTcb.preloadDsc.returnQueueIdx, TI_EM_MEM_PUBLIC_DESC);
            }
            Emti_gvTcb.preloadDsc.localEventHdl = localEventHdl;
            lvHostDescPtr = (Cppi_HostDesc *) Emti_getCppiHostDescPtr (localEventHdl);
            localBufferPtr = (void *) Emti_bufferPtr_ptov (lvHostDescPtr);
        }
    }
#if (TI_EM_TRACE_LEVEL >=1)
    if (Emti_svTraceFunctionPtr != NULL)
    {
        Emti_svTraceFunctionPtr (TI_EM_TSCOPE_CLAIM_LOCAL, (void *) (localBufferPtr));
    }
#endif /* (TI_EM_TRACE_LEVEL >=1) */
    return localBufferPtr;
}

void
em_atomic_processing_end (
    void
)
{

#if (TI_EM_TRACE_LEVEL >=1)
    if (Emti_svTraceFunctionPtr != NULL)
    {
        Emti_svTraceFunctionPtr (TI_EM_TSCOPE_ATOMIC_PROCESSING_END,
                                 (void *) (Emti_gvTcb.apCommandHdlTbl[Emti_gvTsb.pingPong]));
    }
#endif /* (TI_EM_TRACE_LEVEL >=1) */

    /*
     * This function is applicable only for atomic queues. 
     */
    if (!(Emti_gvTsb.actionFlags & Emti_AP_COMMAND_TO_SEND))
        return;

    *(uint32_t *) (Emti_gvTcb.cdQueueRegDPtr) = (uint32_t) (Emti_gvTcb.apCommandHdlTbl[Emti_gvTsb.pingPong]);

    /*
     * The queue has been processed. 
     */
    Emti_gvTsb.actionFlags &= Emti_AP_COMMAND_TO_CLEAR;

    return;
}
