/*
 * Copyright (c) 2012, Texas Instruments Incorporated - http://www.ti.com/
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions are met:
 *       * Redistributions of source code must retain the above copyright
 *         notice, this list of conditions and the following disclaimer.
 *       * Redistributions in binary form must reproduce the above copyright
 *         notice, this list of conditions and the following disclaimer in the
 *         documentation and/or other materials provided with the distribution.
 *       * Neither the name of Texas Instruments Incorporated nor the
 *         names of its contributors may be used to endorse or promote products
 *         derived from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *   DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
 *   DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


/**
 * @file: event_machine_event.c
 *
 * @brief
 *    Open Event Machine HW TI specific functions implementation
 *    Functions related to events.
 *
 */

#include <stdio.h>
#include <string.h>
#include <ti/runtime/openem/event_machine_macros.h>
#include <ti/runtime/openem/event_machine_types.h>
#include <ti/runtime/openem/event_machine_group.h>
#include <ti/runtime/openem/event_machine_hw_macros.h>
#include <ti/runtime/openem/event_machine_hw_types.h>
#include <ti/runtime/openem/event_machine_hw_ti_macros.h>
#include <ti/runtime/openem/event_machine_hw_ti_types.h>
#include <ti/runtime/openem/src/event_machine_hwpform.h>
#include <ti/runtime/openem/src/event_machine_private_types.h>
#include <ti/runtime/openem/src/event_machine_private_macros.h>
#include <ti/runtime/openem/src/event_machine_private_variables.h>
#include <ti/runtime/openem/src/event_machine_private_functions.h>
#include <ti/runtime/openem/modules/event_machine_modules.h>

size_t
ti_em_buffer_size (
    em_event_t event
)
{
    Cppi_HostDesc *lvHostDescPtr = (Cppi_HostDesc *) Emti_getCppiHostDescPtr (event);
    return (size_t) (lvHostDescPtr->buffLen & 0x3FFFFF);
}

size_t
ti_em_event_size (
    em_event_t event
)
{
    Cppi_HostDesc *lvHostDescPtr = (Cppi_HostDesc *) Emti_getCppiHostDescPtr (event);
    return (size_t) (lvHostDescPtr->descInfo & 0x3FFFFF);
}

void
ti_em_set_type (
    em_event_t event,
    em_event_type_t event_type
)
{
    Cppi_HostDesc *lvHostDescPtr = (Cppi_HostDesc *) Emti_getCppiHostDescPtr (event);
    uint8_t *lvTimeStampPtr = (uint8_t *) & (lvHostDescPtr->timeStamp);
#ifdef _LITTLE_ENDIAN
    lvTimeStampPtr += 3;
#endif
    *lvTimeStampPtr = (((uint8_t) event_type) & 0xff);

    return;
}

em_event_type_t
ti_em_get_type (
    const em_event_t event
)
{
    Cppi_HostDesc *lvHostDescPtr = (Cppi_HostDesc *) Emti_getCppiHostDescPtr (event);

    return (em_event_type_t) ((lvHostDescPtr->timeStamp >> 24) & (0xff));
}

em_event_type_t
ti_em_get_type_preload (
    em_event_type_t type
)
{
    return (em_event_type_t) (type & TI_EM_EVENT_TYPE_PRELOAD_MSK);
}

void
ti_em_set_queue (
    em_event_t event,
    em_queue_t queue_hdl,
    em_event_group_t event_group_hdl
)
{
    Cppi_HostDesc *lvHostDescPtr = (Cppi_HostDesc *) Emti_getCppiHostDescPtr (event);
    uint32_t *lvTimeStampPtr = &(lvHostDescPtr->timeStamp);
    uint32_t *lvSoftwareInfo0Ptr = &(lvHostDescPtr->softwareInfo0);

    /*
     * Writes the full queue handle to word 8 (FORMAT_A) 
     */
    *lvTimeStampPtr = ((*lvTimeStampPtr & Emti_EVENT_TYPE_MASK) | (queue_hdl & Emti_EQ_HANDLE_MASK));

    /*
     * Writes event group index to word 9 (FORMAT_A) 
     */
    *lvSoftwareInfo0Ptr = ((event_group_hdl & Emti_EVENT_GROUP_ID_MASK) | Emti_NO_EVENT_GROUP_TAG);

    return;
}

void
ti_em_flush (
    em_event_t event
)
{
    /*
     * check cache coherency on first buffer 
     */
    if (Emti_wbinv (event) == 0)
        Emti_mfence ();

    return;
}

size_t
ti_em_set_ps_words (
    em_event_t event,
    uint32_t * ps_word_ptr,
    size_t ps_wsize
)
{
    uint32_t lvIndex;
    ti_em_packet_t *lvPacketPtr;
    Cppi_HostDesc *lvHostDescPtr;
    uint32_t *lvPsDataPtr;
    size_t lvSize = ps_wsize;
    size_t lvPsDataMaxSize;

    lvPacketPtr = ti_em_to_packet (event);
    lvHostDescPtr = (Cppi_HostDesc *) (lvPacketPtr);
    lvPsDataPtr = (uint32_t *) (&(lvHostDescPtr->psData));

    lvPsDataMaxSize = Emti_getPacketWsize (lvHostDescPtr) - Emti_EVENT_DSC_WORD_SIZE_MIN;

    /*
     * if required, truncate ps_wsize with dscWsize - Emti_EVENT_DSC_WORD_SIZE_MIN (max number of psData allowed) 
     */
    if (lvPsDataMaxSize < ps_wsize)
        lvSize = lvPsDataMaxSize;

    /*
     * loop over ps words 
     */
    for (lvIndex = 0; lvIndex < lvSize; lvIndex++)
    {
        lvPsDataPtr[lvIndex] = ps_word_ptr[lvIndex];
    }

    /*
     * set ps word location in packet 
     */
    lvHostDescPtr->descInfo = (lvHostDescPtr->descInfo | (Cppi_PSLoc_PS_IN_DESC << 22));
    /*
     * set ps byte count in packet 
     */
    lvHostDescPtr->packetInfo = (lvHostDescPtr->packetInfo & 0xC0FFFFFF) | (lvSize << 24);

    return (size_t) (lvSize);
}

size_t
ti_em_get_ps_words (
    em_event_t event,
    uint32_t * ps_word_ptr,
    size_t ps_wsize
)
{
    uint32_t lvIndex;
    ti_em_packet_t *lvPacketPtr;
    Cppi_HostDesc *lvHostDescPtr;
    uint32_t *lvPsDataPtr;
    size_t lvSize = ps_wsize;
    size_t lvPsDataMaxSize;
    size_t lvPsSize;

    lvPacketPtr = ti_em_to_packet (event);
    lvHostDescPtr = (Cppi_HostDesc *) (lvPacketPtr);
    lvPsDataPtr = (uint32_t *) (&(lvHostDescPtr->psData));

    lvPsDataMaxSize = Emti_getPacketWsize (lvHostDescPtr) - Emti_EVENT_DSC_WORD_SIZE_MIN;

    /*
     * if required, truncate ps_wsize with dscWsize - Emti_EVENT_DSC_WORD_SIZE_MIN (max number of psData allowed) 
     */
    if (lvPsDataMaxSize < ps_wsize)
        lvSize = lvPsDataMaxSize;
    /*
     * Now check the real number of ps words) 
     */
    lvPsSize = (size_t) ((lvHostDescPtr->packetInfo >> 24) & 0x3f);
    if (lvPsSize < ps_wsize)
        lvSize = lvPsSize;

    /*
     * loop over ps words 
     */
    for (lvIndex = 0; lvIndex < lvSize; lvIndex++)
    {
        ps_word_ptr[lvIndex] = lvPsDataPtr[lvIndex];
    }
    return (size_t) (lvSize);
}

size_t
ti_em_set_ps_wsize (
    em_event_t event,
    size_t ps_wsize
)
{
    ti_em_packet_t *lvPacketPtr;
    Cppi_HostDesc *lvHostDescPtr;
    size_t lvSize = ps_wsize;
    size_t lvPsDataMaxSize;

    lvPacketPtr = ti_em_to_packet (event);
    lvHostDescPtr = (Cppi_HostDesc *) (lvPacketPtr);

    lvPsDataMaxSize = Emti_getPacketWsize (lvHostDescPtr) - Emti_EVENT_DSC_WORD_SIZE_MIN;

    /*
     * if required, truncate ps_wsize with dscWsize - Emti_EVENT_DSC_WORD_SIZE_MIN (max number of psData allowed) 
     */
    if (lvPsDataMaxSize < ps_wsize)
        lvSize = lvPsDataMaxSize;

    lvHostDescPtr->packetInfo = (lvHostDescPtr->packetInfo & 0xc0ffffff) | (lvSize << 24);
    return (size_t) (lvSize);
}

size_t
ti_em_get_ps_wsize (
    em_event_t event
)
{
    ti_em_packet_t *lvPacketPtr;
    Cppi_HostDesc *lvHostDescPtr;

    lvPacketPtr = ti_em_to_packet (event);
    lvHostDescPtr = (Cppi_HostDesc *) (lvPacketPtr);

    return (size_t) ((lvHostDescPtr->packetInfo >> 24) & 0x3f);
}

ti_em_buf_mode_t
ti_em_get_buf_mode (
    const em_event_t event
)
{
    return Emti_getBufMode ((Cppi_HostDesc *) event);
}

ti_em_coh_mode_t
ti_em_get_coh_mode (
    const em_event_t event
)
{
    return Emti_getCohMode ((Cppi_HostDesc *) event);
}
