/*
 * Copyright (c) 2012, Texas Instruments Incorporated - http://www.ti.com/
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions are met:
 *       * Redistributions of source code must retain the above copyright
 *         notice, this list of conditions and the following disclaimer.
 *       * Redistributions in binary form must reproduce the above copyright
 *         notice, this list of conditions and the following disclaimer in the
 *         documentation and/or other materials provided with the distribution.
 *       * Neither the name of Texas Instruments Incorporated nor the
 *         names of its contributors may be used to endorse or promote products
 *         derived from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *   DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
 *   DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


/**
 *   @file  event_machine_event_group.c
 *
 *   @brief
 *      Open Event Machine Implementation for Keystone1 devices.
 *      Functions related to event group.
 *
 */

#include <stdio.h>
#include <inttypes.h>
#include <string.h>
#include <ti/runtime/openem/event_machine_macros.h>
#include <ti/runtime/openem/event_machine_hw_macros.h>
#include <ti/runtime/openem/event_machine_hw_ti_macros.h>
#include <ti/runtime/openem/event_machine_types.h>
#include <ti/runtime/openem/event_machine_hw_types.h>
#include <ti/runtime/openem/event_machine_hw_ti_types.h>
#include <ti/runtime/openem/event_machine_sw_ti_config.h>
#include <ti/runtime/openem/src/event_machine_hwpform.h>
#include <ti/runtime/openem/src/event_machine_private_macros.h>
#include <ti/runtime/openem/src/event_machine_private_types.h>
#include <ti/runtime/openem/src/event_machine_private_variables.h>
#include <ti/runtime/openem/src/event_machine_private_functions.h>
#include <ti/runtime/openem/event_machine_group.h>


em_event_group_t
em_event_group_create (
    void
)
{
    Emti_EventGroup_s *ptrEventGroup;
    uint32_t startIndex;

    /*
     * Critical Section Enter 
     */
    Emti_lockSemaphore ();

    startIndex = Emti_svPsbPtr->nextFreeEventGroupIdx;

    if (startIndex == EM_MAX_EVENT_GROUPS)
    {
        /*
         * Critical Section Exit 
         */
        Emti_unlockSemaphore ();

        return (em_event_group_t) EM_EVENT_GROUP_UNDEF;
    }

    /*
     * Get the pointer to the event group. 
     */
    ptrEventGroup = &Emti_svEventGroupMdb[startIndex];

    /*
     * Mark the event group as allocated 
     */
    ptrEventGroup->state = Emti_EventGroupState_ALLOCATED;

    Emti_svPsbPtr->nextFreeEventGroupIdx = startIndex + 1;

    /*
     * Critical Section Exit 
     */
    Emti_unlockSemaphore ();

    /*
     * Return the allocated event group. 
     */
    return (em_event_group_t) startIndex;
}

em_status_t
em_event_group_delete (
    em_event_group_t event_group
)
{
    em_status_t ret = EM_ERR_NOT_IMPLEMENTED;
    return ret;
}

em_status_t
em_event_group_apply (
    em_event_group_t group,
    int count,
    int num_notif,
    const em_notif_t * notif_tbl
)
{
    Emti_EventGroup_s *ptrEventGroup;
    uint32_t lvNumNotifIndex;

#if (EM_CHECK_LEVEL >= 1)
    /*
     * Basic Validations: Ensure a valid event group was passed. 
     */
    if (group >= EM_MAX_EVENT_GROUPS)
        return EM_ERR_TOO_LARGE;
#endif

    /*
     * Get the event group. 
     */
    ptrEventGroup = &Emti_svEventGroupMdb[(uint32_t) group];

#if (EM_CHECK_LEVEL >= 1)
    /*
     * Basic Validation: Make sure that the event group was ALLOCATED 
     */
    if (ptrEventGroup->state != Emti_EventGroupState_ALLOCATED)
        return EM_ERR_BAD_STATE;
#endif

    /*
     * loop over num_notif, check cache coherency on the first buffer 
     */
    for (lvNumNotifIndex = 0; lvNumNotifIndex < num_notif; lvNumNotifIndex++)
        Emti_wbinv (notif_tbl[lvNumNotifIndex].event);

    /*
     * Critical Section Enter 
     */
    Emti_lockSemaphore ();

    /*
     * Store the event group information. 
     */
    ptrEventGroup->eventNum = count;
    ptrEventGroup->notifNum = num_notif;
    memcpy ((void *) &ptrEventGroup->notifTbl[0], (void *) notif_tbl, (sizeof (em_notif_t) * num_notif));

    /*
     * Change the state of the event group. 
     */
    ptrEventGroup->state = Emti_EventGroupState_APPLIED;

    /*
     * Critical Section Exit 
     */
    Emti_unlockSemaphore ();

    return EM_OK;
}

em_status_t
em_event_group_increment (
    int event_num
)
{
    Emti_EventGroup_s *ptrEventGroup;

#if (EM_CHECK_LEVEL >= 1)
    /*
     * Basic Validations: Ensure a valid event group was passed. 
     */
    if (Emti_gvTsb.eventGroupHdl >= EM_MAX_EVENT_GROUPS)
        return EM_ERR_BAD_CONTEXT;
#endif

    /*
     * Get the event group. 
     */
    ptrEventGroup = &Emti_svEventGroupMdb[(uint32_t) (Emti_gvTsb.eventGroupHdl)];

#if (EM_CHECK_LEVEL >= 1)
    /*
     * Basic Validation: Make sure that the event group was APPLIED 
     */
    if (ptrEventGroup->state != Emti_EventGroupState_APPLIED)
        return EM_ERR_BAD_STATE;
#endif

    /*
     * Critical Section Enter 
     */
    Emti_lockSemaphore ();

    ptrEventGroup->eventNum += event_num;

    /*
     * Critical Section Exit 
     */
    Emti_unlockSemaphore ();

    return EM_OK;
}

em_event_group_t
em_event_group_current (
    void
)
{
    return (em_event_group_t) (Emti_gvTsb.eventGroupHdl);
}
