/*
 * Copyright (c) 2012, Texas Instruments Incorporated - http://www.ti.com/
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions are met:
 *       * Redistributions of source code must retain the above copyright
 *         notice, this list of conditions and the following disclaimer.
 *       * Redistributions in binary form must reproduce the above copyright
 *         notice, this list of conditions and the following disclaimer in the
 *         documentation and/or other materials provided with the distribution.
 *       * Neither the name of Texas Instruments Incorporated nor the
 *         names of its contributors may be used to endorse or promote products
 *         derived from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *   DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
 *   DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


/**
 *   @file  event_machine_execution_object.c
 *
 *   @brief
 *      Open Event Machine Implementation for Keystone1 devices.
 *      Functions related to execution objects.
 *
 */

#include <stdio.h>
#include <inttypes.h>
#include <string.h>
#include <ti/runtime/openem/event_machine_macros.h>
#include <ti/runtime/openem/event_machine_hw_macros.h>
#include <ti/runtime/openem/event_machine_hw_ti_macros.h>
#include <ti/runtime/openem/event_machine_types.h>
#include <ti/runtime/openem/event_machine_hw_types.h>
#include <ti/runtime/openem/event_machine_hw_ti_types.h>
#include <ti/runtime/openem/event_machine_sw_ti_config.h>
#include <ti/runtime/openem/src/event_machine_hwpform.h>
#include <ti/runtime/openem/src/event_machine_private_macros.h>
#include <ti/runtime/openem/src/event_machine_private_types.h>
#include <ti/runtime/openem/src/event_machine_private_variables.h>
#include <ti/runtime/openem/src/event_machine_private_functions.h>
#include <ti/runtime/openem/event_machine_group.h>


em_eo_t
em_eo_create (
    const char *name,
    em_start_func_t start,
    em_start_local_func_t local_start,
    em_stop_func_t stop,
    em_stop_local_func_t local_stop,
    em_receive_func_t receive,
    const void *eo_ctx
)
{
    uint32_t startIndex;
    Emti_EO_s *ptrEO;

#if (EM_CHECK_LEVEL >= 1)
    /*
     * Basic Validations: Ensure that the start/stop and receive API have been specified. 
     */
    if ((start == NULL) || (stop == NULL) || (receive == NULL))
        return EM_EO_UNDEF;
#endif

    /*
     * Critical Section Enter 
     */
    Emti_lockSemaphore ();

    startIndex = Emti_svPsbPtr->nextFreeEoIdx;

    if (startIndex == EM_MAX_EOS)
    {
        /*
         * Critical Section Exit 
         */
        Emti_unlockSemaphore ();
        return (em_eo_t) (EM_EO_UNDEF);
    }

    /*
     * Get the pointer to the execution object 
     */
    ptrEO = &Emti_svEoMdb[startIndex];

    /*
     * Execution object is available. 
     */
    if (strlen (name) <= EM_EO_NAME_LEN)
        strcpy (ptrEO->name, name);
    else
        strncpy (ptrEO->name, name, EM_EO_NAME_LEN);
    ptrEO->start = start;
    ptrEO->localStart = local_start;
    ptrEO->stop = stop;
    ptrEO->localStop = local_stop;
    ptrEO->eo64b.receive = receive;
    ptrEO->eo64b.contextPtr = (void *) eo_ctx;
    ptrEO->errorHandler = NULL;

    /*
     * Mark the execution object as allocated 
     */
    ptrEO->state = Emti_EOState_ALLOCATED;

    Emti_svPsbPtr->nextFreeEoIdx = startIndex + 1;

    /*
     * Critical Section Exit 
     */
    Emti_unlockSemaphore ();

    /*
     * Return the handle to the execution object that has been allocated. 
     */
    return (em_eo_t) ptrEO;
}

em_status_t
em_eo_start (
    em_eo_t eo,
    em_status_t * result,
    int num_notif,
    const em_notif_t * notif_tbl
)
{
    Emti_EO_s *ptrEO;
    em_status_t status;

    /*
     * Get the execution object 
     */
    ptrEO = (Emti_EO_s *) eo;

#if (EM_CHECK_LEVEL >= 1)
    /*
     * Basic Parameter Validation: Ensure a valid EO was passed. 
     */
    if (ptrEO == NULL)
        return EM_ERR;

    /*
     * Basic Validation: Ensure that the EO has not already been started. 
     */
    if (ptrEO->state == Emti_EOState_STARTED)
        return EM_ERR_BAD_STATE;
#endif

    /*
     * Call the global start function. 
     */
    status = ptrEO->start (ptrEO->eo64b.contextPtr, eo);
    if (status == EM_OK)
    {
        /*
         * Critical Section Enter 
         */
        Emti_lockSemaphore ();

        /*
         * The Execution object is now started. 
         */
        ptrEO->state = Emti_EOState_STARTED;

        /*
         * Critical Section Exit 
         */
        Emti_unlockSemaphore ();
    }
    return status;
}

em_status_t
em_eo_delete (
    em_eo_t eo
)
{
    em_status_t ret = EM_ERR_NOT_IMPLEMENTED;
    return ret;
}

size_t
em_eo_get_name (
    em_eo_t eo,
    char *name,
    size_t maxlen
)
{
    Emti_EO_s *ptrEMEO;
    size_t ret;

    /*
     * If there was no buffer provided; we cannot proceed. 
     */
    if ((maxlen == 0) || (name == NULL))
        return (size_t) 0;

    /*
     * Get the pointer to the execution object. 
     */
    ptrEMEO = (Emti_EO_s *) eo;

#if (EM_CHECK_LEVEL >= 1)
    /*
     * Basic Validation: Ensure that the queue is valid. 
     */
    if (ptrEMEO == NULL)
        return EM_ERR_BAD_CONTEXT;

    /*
     * Basic Validation: Ensure that the queue was at least allocated. 
     */
    if (ptrEMEO->state == Emti_EOState_FREE)
        return EM_ERR_BAD_STATE;

    if (maxlen > EM_EO_NAME_LEN)
        return EM_ERR_TOO_LARGE;
#endif

    /*
     * Get the length of the EO name. 
     */
    ret = (size_t) strlen (ptrEMEO->name);
    if (maxlen < ret)
        ret = maxlen;
    strncpy (name, ptrEMEO->name, ret);

    /*
     * Return the length. 
     */
    return ret;
}

em_status_t
em_eo_add_queue (
    em_eo_t eo,
    em_queue_t queue
)
{
    Emti_QueueSlow_s *ptrQueueSlow;
    Emti_QueueFast_s *ptrQueueFast;
    Emti_EO_s *ptrEMEO;

    /*
     * Get the pointer to the queue 
     */
    ptrQueueSlow = &Emti_svQueueSlowMdb[Emti_getQueueIndex (queue)];
    ptrQueueFast = &Emti_svQueueFastMdb[Emti_getQueueIndex (queue)];

#if (EM_CHECK_LEVEL >= 1)
    /*
     * Basic Validation: Ensure that the queue & EO is valid. 
     */
    if ((ptrQueueSlow == NULL) || (eo == 0))
        return EM_ERR_BAD_CONTEXT;

    /*
     * Basic Validation: Ensure that the queue was not already mapped. 
     */
    if (ptrQueueSlow->state == Emti_QueueState_MAPPED)
        return EM_ERR_BAD_STATE;
#endif

    /*
     * Map the execution object and queue together. 
     */
    ptrEMEO = (Emti_EO_s *) eo;

    /*
     * Critical Section Enter 
     */
    Emti_lockSemaphore ();

    ptrQueueSlow->eoPtr = ptrEMEO;
    ptrQueueFast->receive = (em_receive_func_t) ptrEMEO->eo64b.receive;

    /*
     * The queue has been mapped. 
     */
    ptrQueueSlow->state = Emti_QueueState_MAPPED;

    /*
     * Debug Message: 
     */
    Emti_logDbg ("Mapping  Queue (%p,%p) to Execution Object %p\n", ptrQueueSlow, ptrQueueFast, ptrEMEO);

    /*
     * Critical Section Exit 
     */
    Emti_unlockSemaphore ();

    /*
     * The mapping has been done. 
     */
    return EM_OK;
}

em_status_t
em_eo_remove_queue (
    em_eo_t eo,
    em_queue_t queue,
    int num_notif,
    const em_notif_t * notif_tbl
)
{
    em_status_t ret = EM_ERR_NOT_IMPLEMENTED;
    return ret;
}

em_status_t
em_eo_stop (
    em_eo_t eo,
    int num_notif,
    const em_notif_t * notif_tbl
)
{
    em_status_t ret = EM_ERR_NOT_IMPLEMENTED;
    return ret;
}

em_status_t
em_eo_register_error_handler (
    em_eo_t eo,
    em_error_handler_t handler
)
{
    Emti_EO_s *lvEoPtr = (Emti_EO_s *) (eo);

    lvEoPtr->errorHandler = handler;

    /*
     * ensure data are written in memory 
     */
    Emti_mfence ();

    return EM_OK;
}

em_status_t
em_eo_unregister_error_handler (
    em_eo_t eo
)
{
    Emti_EO_s *lvEoPtr = (Emti_EO_s *) (eo);

    lvEoPtr->errorHandler = NULL;

    /*
     * ensure data are written in memory 
     */
    Emti_mfence ();

    return EM_OK;
}
