/*
 * Copyright (c) 2012, Texas Instruments Incorporated - http://www.ti.com/
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions are met:
 *       * Redistributions of source code must retain the above copyright
 *         notice, this list of conditions and the following disclaimer.
 *       * Redistributions in binary form must reproduce the above copyright
 *         notice, this list of conditions and the following disclaimer in the
 *         documentation and/or other materials provided with the distribution.
 *       * Neither the name of Texas Instruments Incorporated nor the
 *         names of its contributors may be used to endorse or promote products
 *         derived from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *   DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
 *   DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


/**
 *   @file  event_machine_exit.c
 *
 *   @brief
 *      Open Event Machine Implementation for Keystone1 devices.
 *      Exit procedure.
 */

#include <inttypes.h>
#include <stddef.h>
#include <string.h>
#include <ti/runtime/openem/event_machine_macros.h>
#include <ti/runtime/openem/event_machine_types.h>
#include <ti/runtime/openem/event_machine_core_mask.h>
#include <ti/runtime/openem/event_machine_hw_macros.h>
#include <ti/runtime/openem/event_machine_hw_types.h>
#include <ti/runtime/openem/event_machine_hw_ti_macros.h>
#include <ti/runtime/openem/event_machine_hw_ti_types.h>
#include <ti/runtime/openem/event_machine_hw_ti_functions.h>
#include <ti/runtime/openem/src/event_machine_hwpform.h>
#include <ti/runtime/openem/src/event_machine_private_macros.h>
#include <ti/runtime/openem/src/event_machine_private_types.h>
#include <ti/runtime/openem/src/event_machine_private_variables.h>
#include <ti/runtime/openem/src/event_machine_private_functions.h>
#include <ti/runtime/openem/modules/event_machine_modules.h>

em_status_t
ti_em_exit_global (
    void
)
{
    uint32_t queueNb;
    uint32_t dmaFlowNb = 0;
    uint32_t index;
    em_status_t lvRetStatus = EM_OK;

    /*
     * send Kill command to scheduler CD queue 
     */
    if (Emti_svPcb.pdspNum > 0)
    {
        uint32_t lvSchedIdx;
        for (lvSchedIdx = 0; lvSchedIdx < Emti_svPcb.pdspNum; lvSchedIdx++)
        {
            lvRetStatus =
                Emti_connectFirmware (Emti_svPcb.cdQueueIdx[lvSchedIdx], Emti_svPcb.ctCommandFreeQueueIdx,
                                      Emti_svPcb.rpQueueIdx, (uint32_t) NULL, (uint32_t) 0, (uint32_t) Emti_CtType_KILL,
                                      (uint32_t) TI_EM_SLOT_SPCB, (ti_em_process_id_t) 0);
            if (lvRetStatus != EM_OK)
                return lvRetStatus;
        }
    }

    /**********************************************************************
     *********************** Scheduling (SD) Queues ***********************
     **********************************************************************/
    {
        /*
         * Close all the scheduling queues. 
         */
        queueNb = (uint32_t) Emti_svPcb.sdQueueBaseIdx;
        for (index = 0; index < Emti_SD_QUEUE_NUM_UP; index++)
        {
            lvRetStatus = ti_em_hw_queue_close (queueNb);
            if (lvRetStatus != EM_OK)
                return EM_ERR_ALLOC_FAILED;

            queueNb++;
        }
    }

    if (Emti_svPcb.pdspNum == 0)
        goto skip_atomic_processing_deallocation0;

    /**********************************************************************
     ************** Atomic Processing Scheduling (APSD) Queues ************
     **********************************************************************/
    {
        /*
         * Close all the atomic-processing queues. 
         */
        queueNb = (uint32_t) Emti_svPcb.apSdQueueBaseIdx;
        for (index = 0; index < Emti_SD_QUEUE_NUM_UP; index++)
        {
            lvRetStatus = ti_em_hw_queue_close (queueNb);
            if (lvRetStatus != EM_OK)
                return EM_ERR_ALLOC_FAILED;

            queueNb++;
        }
    }

    /**********************************************************************
     ********************** Atomic Processing (AP) Queues *****************
     **********************************************************************/
    {
        /*
         * Close all the atomic hw queues. 
         */
        queueNb = (uint32_t) Emti_svPcb.apQueueBaseIdx;
        for (index = 0; index < Emti_AP_QUEUE_NUM; index++)
        {
            lvRetStatus = ti_em_hw_queue_close (queueNb);
            if (lvRetStatus != EM_OK)
                return EM_ERR_ALLOC_FAILED;

            queueNb++;
        }
    }

    /****************************************************************
     ******************** Atomic Processing Free (APfree) Queues ****
     ****************************************************************/
    {
        queueNb = (uint32_t) Emti_svPcb.apFreeQueueBaseIdx;
        for (index = 0; index < TI_EM_PRIO_NUM; index++)
        {
            lvRetStatus = ti_em_hw_queue_close (queueNb);
            if (lvRetStatus != EM_OK)
                return EM_ERR_ALLOC_FAILED;

            queueNb++;
        }
    }

  skip_atomic_processing_deallocation0:

    /****************************************************************
     ********************** Prefetch completion Queues **************
     ****************************************************************/
    {
        /*
         * These queues are the return queues when preloading is enable 
         */
        queueNb = (uint32_t) Emti_svPcb.preloadDsc.returnQueueBaseIdx;
        for (index = 0; index < TI_EM_CORE_NUM; index++)
        {
            lvRetStatus = ti_em_hw_queue_close (queueNb);
            if (lvRetStatus != EM_OK)
                return EM_ERR_ALLOC_FAILED;

            queueNb++;
        }
    }

    /******************************************* ********************
     ********************** Prefetch Queue (PF) Queues **************
     ****************************************************************/
    {
        /*
         * These queues will contain local or global prefetch events 
         */
        queueNb = (uint32_t) Emti_svPcb.pfQueueBaseIdx;
        for (index = 0; index < TI_EM_CORE_NUM; index++)
        {
            /*
             * Open the queues 
             */
            lvRetStatus = ti_em_hw_queue_close (queueNb);
            if (lvRetStatus != EM_OK)
                return EM_ERR_ALLOC_FAILED;

            queueNb++;
        }
    }

    /****************************************************************
     ******************** CT command free queue *********************
     ****************************************************************/
    {
        lvRetStatus = ti_em_hw_queue_close (Emti_svPcb.ctCommandFreeQueueIdx);
        if (lvRetStatus != EM_OK)
            return EM_ERR_ALLOC_FAILED;
    }

    /****************************************************************
     ******************** AP command free queue *********************
     ****************************************************************/
    {
        lvRetStatus = ti_em_hw_queue_close (Emti_svPcb.apCommandFreeQueueIdx);
        if (lvRetStatus != EM_OK)
            return EM_ERR_ALLOC_FAILED;
    }

    /****************************************************************
     ******************** PF command free queue *********************
     ****************************************************************/
    {
        lvRetStatus = ti_em_hw_queue_close (Emti_svPcb.pfCommandFreeQueueIdx);
        if (lvRetStatus != EM_OK)
            return EM_ERR_ALLOC_FAILED;
    }

    /****************************************************************
     ********************  Response (RP) Queue **********************
     ****************************************************************/
    {
        lvRetStatus = ti_em_hw_queue_close (Emti_svPcb.rpQueueIdx);
        if (lvRetStatus != EM_OK)
            return EM_ERR_ALLOC_FAILED;
    }

    /****************************************************************
     ********************  Command (CD) Queue ***********************
     ****************************************************************/
    if (Emti_svPcb.pdspNum > 0)
    {
        uint32_t lvSchedIdx;

        for (lvSchedIdx = 0; lvSchedIdx < Emti_svPcb.pdspNum; lvSchedIdx++)
        {
            lvRetStatus = ti_em_hw_queue_close (Emti_svPcb.cdQueueIdx[lvSchedIdx]);
            if (lvRetStatus != EM_OK)
                return EM_ERR_ALLOC_FAILED;
        }

        lvRetStatus = ti_em_hw_queue_close (Emti_svPcb.apMutexQueueIdx);
        if (lvRetStatus != EM_OK)
            return EM_ERR_ALLOC_FAILED;
    }

    if (Emti_svPcb.preloadEnabled != TI_EM_PRELOAD_ENABLED)
        goto skip_preload_exit;

    /****************************************************************
     ********************  DMA preload Tx Queue *********************
     ****************************************************************/
    {
        ti_em_queue_id_t lvDmaQueueNb = Emti_svPcb.preloadDsc.txQueueBaseIdx;
        for (index = 0; index < TI_EM_CORE_NUM; index++)
        {
            lvRetStatus = ti_em_hw_queue_close (lvDmaQueueNb);
            if (lvRetStatus != EM_OK)
                return EM_ERR_ALLOC_FAILED;

            lvDmaQueueNb++;
        }
    }

    /****************************************************************
     ********************  RX Flows *********************************
     ****************************************************************/
    {
        int lvFlowId = Emti_svPcb.dma_queue_base_idx;
        for (index = 0; index < TI_EM_CORE_NUM; index++)
        {
            em_status_t lvRxFlowStatus;

            lvRxFlowStatus = ti_em_rx_flow_close (Emti_svPcb.dma_idx, lvFlowId + index);
            if (lvRxFlowStatus != EM_OK)
            {
                return EM_ERR;
            }
            dmaFlowNb++;
        }
    }

    /****************************************************************
     ********************  TX Channels ******************************
     ****************************************************************/
    {
        uint32_t lvTxchannelNb = Emti_svPcb.dma_queue_base_idx;

        for (index = 0; index < TI_EM_CORE_NUM; index++)
        {
            em_status_t lvTxChannelStatus;

            lvTxChannelStatus = ti_em_tx_channel_close (Emti_svPcb.dma_idx, lvTxchannelNb + index);
            if (lvTxChannelStatus != EM_OK)
            {
                return EM_ERR;
            }
        }
    }

    /****************************************************************
     ********************  RX Channels ******************************
     ****************************************************************/
    {
        uint32_t lvRxchannelNb = Emti_svPcb.dma_queue_base_idx;

        for (index = 0; index < TI_EM_CORE_NUM; index++)
        {
            em_status_t lvRxChannelStatus;

            lvRxChannelStatus = ti_em_rx_channel_close (Emti_svPcb.dma_idx, lvRxchannelNb + index);
            if (lvRxChannelStatus != EM_OK)
            {
                return EM_ERR;
            }
        }
    }

  skip_preload_exit:
    /****************************************************************
     ************** Call chaining global exit           *************
     ****************************************************************/
    lvRetStatus = Emti_moduleChainGlobalExit ();
    if ((lvRetStatus != EM_OK) && (lvRetStatus != EM_ERR_NOT_FOUND))
        return lvRetStatus;

    /****************************************************************
     ******************** release DMA queues ************************
     ****************************************************************/
    lvRetStatus = ti_em_hw_queue_close (Emti_svPcb.rsTxQueueIdx);
    if (lvRetStatus != EM_OK)
        return EM_ERR_ALLOC_FAILED;

    if (Emti_svPcb.dma_idx != (ti_em_dma_id_t) - 1)
    {

        lvRetStatus = ti_em_hw_queue_close (Emti_svPcb.rsFreeQueueIdx);
        if (lvRetStatus != EM_OK)
            return EM_ERR_ALLOC_FAILED;

    /****************************************************************
     ******************** release DMA RX Flows **********************
     ****************************************************************/
        {
            em_status_t lvRxFlowStatus;
            int lvFlowId = Emti_svPcb.dma_queue_base_idx + dmaFlowNb;

            lvRxFlowStatus = ti_em_rx_flow_close (Emti_svPcb.dma_idx, lvFlowId);

            if (lvRxFlowStatus != EM_OK)
            {
                return EM_ERR;
            }
        }

    /****************************************************************
     ******************** release DMA TX Channels *******************
     ****************************************************************/
        {
            em_status_t lvTxChannelStatus;
            uint32_t lvTxchannelNb = Emti_svPcb.dma_queue_base_idx + dmaFlowNb;

            lvTxChannelStatus = ti_em_tx_channel_close (Emti_svPcb.dma_idx, lvTxchannelNb);
            if (lvTxChannelStatus != EM_OK)
            {
                return EM_ERR;
            }
        }

    /****************************************************************
     ******************** release DMA RX Channels *******************
     ****************************************************************/
        {
            em_status_t lvRxChannelStatus;
            uint32_t lvRxchannelNb = Emti_svPcb.dma_queue_base_idx + dmaFlowNb;

            lvRxChannelStatus = ti_em_rx_channel_close (Emti_svPcb.dma_idx, lvRxchannelNb);
            if (lvRxChannelStatus != EM_OK)
            {
                return EM_ERR;
            }
        }
    }

    /*
     * Release shared memory 
     */
    Emti_sharedMemRelease ();

    return lvRetStatus;
}
