/*
 * Copyright (c) 2012, Texas Instruments Incorporated - http://www.ti.com/
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions are met:
 *       * Redistributions of source code must retain the above copyright
 *         notice, this list of conditions and the following disclaimer.
 *       * Redistributions in binary form must reproduce the above copyright
 *         notice, this list of conditions and the following disclaimer in the
 *         documentation and/or other materials provided with the distribution.
 *       * Neither the name of Texas Instruments Incorporated nor the
 *         names of its contributors may be used to endorse or promote products
 *         derived from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *   DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
 *   DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


/**
 * @file: event_machine_hw_ti_functions.c
 *
 * @brief
 *    Open Event Machine HW TI specific functions implementation
 *
 */

#include <stdio.h>
#include <string.h>
#include <ti/runtime/openem/event_machine_macros.h>
#include <ti/runtime/openem/event_machine_types.h>
#include <ti/runtime/openem/event_machine_group.h>
#include <ti/runtime/openem/event_machine_hw_macros.h>
#include <ti/runtime/openem/event_machine_hw_types.h>
#include <ti/runtime/openem/event_machine_hw_ti_macros.h>
#include <ti/runtime/openem/event_machine_hw_ti_types.h>
#include <ti/runtime/openem/src/event_machine_hwpform.h>
#include <ti/runtime/openem/src/event_machine_private_types.h>
#include <ti/runtime/openem/src/event_machine_private_macros.h>
#include <ti/runtime/openem/src/event_machine_private_variables.h>
#include <ti/runtime/openem/src/event_machine_private_functions.h>
#include <ti/runtime/openem/modules/event_machine_modules.h>

size_t
ti_em_get_pcb_size (
    void
)
{
    return (size_t) (sizeof (Emti_Pcb_s));
}

size_t
ti_em_get_tcb_size (
    void
)
{
    return (size_t) sizeof (Emti_Tcb_s);
}

size_t
ti_em_get_eo_size_slow (
    void
)
{
    return (size_t) sizeof (Emti_EO_s);
}

size_t
ti_em_get_eo_size_fast (
    void
)
{
    return (size_t) (0);
}

size_t
ti_em_get_event_group_size_fast (
    void
)
{
    return (size_t) sizeof (Emti_EventGroup_s);
}

size_t
ti_em_get_event_group_size_slow (
    void
)
{
    return (size_t) (0);
}

size_t
ti_em_get_queue_group_size_fast (
    void
)
{
    return (size_t) (0);
}

size_t
ti_em_get_queue_group_size_slow (
    void
)
{
    return (size_t) sizeof (Emti_QueueGroup_s);
}

size_t
ti_em_get_queue_size_slow (
    void
)
{
    return (size_t) sizeof (Emti_QueueSlow_s);
}

size_t
ti_em_get_queue_size_fast (
    void
)
{
    return (size_t) sizeof (Emti_QueueFast_s);
}

void
ti_em_tag_set_queue (
    uint32_t * tag_ptr,
    em_queue_t queue_hdl
)
{
    /*
     * Writes the full queue handle to word 1 (FORMAT_B) 
     */
    *tag_ptr = ((*tag_ptr & Emti_EVENT_TYPE_MASK) | (queue_hdl & Emti_EQ_HANDLE_MASK));

    return;
}

void
ti_em_tag_set_type (
    uint32_t * tag_ptr,
    em_event_type_t event_type
)
{
    *tag_ptr = ((*tag_ptr & Emti_EQ_HANDLE_MASK) | ((event_type & 0xff) << 24));

    return;
}

em_event_t
ti_em_alloc_local (
    size_t size,
    em_event_type_t type
)
{
    void *lvBufferPtr;
    Cppi_HostDesc *lvHostDescPtr;

    /*
     * set freeQueue = <TCB.poststore>localFreeQueueIdxTbl[<TSB>pingPong]
     * retreive heap free queue Idx for current side 
     */
    ti_em_queue_id_t lvFreeQueueIdx = Emti_gvTcb.chainPoststoreDsc.localFreeQueueIdxTbl[Emti_gvTsb.pingPong];

    /*
     * set poststoreSize = <TSB>poststoreSize[<TSB>pingPong]
     * retreive heap size already used for current side 
     */
    size_t lvPoststoreSize = Emti_gvTsb.chainPoststoreDsc.poststoreSizeTbl[Emti_gvTsb.pingPong];

    /*
     * if poststoreSize == 0 
     *   do set localEventNum to the number of events in freeQueue
     *   while localEventNum < <TCB.poststore>localEventNum
     * check if first em_alloc_local for the event 
     */
    if (lvPoststoreSize == 0)
    {
        while (Emti_getQueueEntryCount (lvFreeQueueIdx) < Emti_gvTcb.chainPoststoreDsc.localEventNum);
    }

    /*
     * set bufferPtr = <TCB.poststore>localHeapPtrTbl[<TSB>pingPong] + poststoreSize
     * retreive heap buffer start pointer for current side 
     */
    lvBufferPtr =
        (void *) (((uint32_t) Emti_gvTcb.chainPoststoreDsc.localHeapPtrTbl[Emti_gvTsb.pingPong]) +
                  ((uint32_t) lvPoststoreSize));

    /*
     * align buffer on 8 bytes boundary - LITTLE ENDIAN only
     */
    lvBufferPtr = (void *) (((uint32_t) (lvBufferPtr) + (Emti_8_BYTES_ALIGN - 1)) & (~(Emti_8_BYTES_ALIGN - 1)));

    /*
     * set poststoreSize = poststoreSize + event_size
     * increment heap size already used with buffer size for current side 
     */
    lvPoststoreSize += size;

    /*
     * if poststoreSize > <TCB.poststore>poststoreSizeMax
     *   return EM_EVENT_UNDEF
     * check if ok with heap total size for current side 
     */
    if (lvPoststoreSize > Emti_gvTcb.chainPoststoreDsc.poststoreSizeMax)
    {
        return (em_event_t) (EM_EVENT_UNDEF);
    }

    /*
     * pop eventHdl from freeQueue 
     */
    lvHostDescPtr = (Cppi_HostDesc *) ((uint32_t) Emti_popEvent (lvFreeQueueIdx, TI_EM_MEM_LOCAL_DESC) & 0xfffffff0);
    if (lvHostDescPtr == NULL)
        return (em_event_t) EM_EVENT_UNDEF;

    /*
     * write event_size to <event DSC>eventSize 
     */
    lvHostDescPtr->descInfo = (Cppi_PSLoc_PS_IN_DESC << 22) | (size & 0x3fffff);

    /*
     * write event_size to <event DSC>bufferSize 
     */
    lvHostDescPtr->buffLen = size;

    /*
     * write bufferPtr to <event DSC>bufferPtr 
     */
    lvHostDescPtr->buffPtr = (uint32_t) (lvBufferPtr);

    /*
     * write 0 to <event DSC>psWsize 
     */
    lvHostDescPtr->packetInfo = (lvHostDescPtr->packetInfo & 0xc0ffffff) | (0 << 24);

    /*
     * write NULL to <event DSC>nextDscPtr 
     */
    lvHostDescPtr->nextBDPtr = 0;

    /*
     * write event type to <event DSC>eventType 
     */
    lvHostDescPtr->timeStamp = (((uint8_t) type) << 24);

    /*
     * set <TSB>poststoreSize[<TSB>pingPong] = poststoreSize 
     */
    Emti_gvTsb.chainPoststoreDsc.poststoreSizeTbl[Emti_gvTsb.pingPong] = lvPoststoreSize;

    /*
     * write Emti_EVENT_MODE_LOCAL to <event DSC>eventMode  
     */
    Emti_setEventMode (lvHostDescPtr, Emti_EVENT_MODE_LOCAL);

    /*
     * write Emti_EVENT_MODE_LOCAL to <event HDL>eventMode  
     */
    return (em_event_t) (((uint32_t) (lvHostDescPtr) & 0xfffffff0) | Emti_EVENT_MODE_LOCAL);
}

em_status_t ti_em_counter_get(
    ti_em_counter_type_t type, 
    ti_em_counter_value_t* value_ptr,
    int flag)
{
    em_status_t status;
        
    switch( type ) 
    {
        case ti_em_counter_type_XGE_RECEIVE :
        case ti_em_counter_type_XGE_WRONG_ETHER_TYPE :
        case ti_em_counter_type_XGE_WRONG_MAC :
        case ti_em_counter_type_XGE_WRONG_PROCESS :
        case ti_em_counter_type_XGE_WRONG_SEQUENCE :
        case ti_em_counter_type_XGE_WRONG_SERVICE_TYPE :
            status = Emti_moduleChainCounterGet(TI_EM_CHAINING_XGE, type, value_ptr, flag);
            break;
        default :
            status = EM_ERR;
            break;
    }    
    
    return status;
}

em_status_t ti_em_xge_rx_miss_enable(ti_em_xge_rx_miss_type_t miss_type)
{
    em_status_t status;
    
    switch( miss_type ) 
    {
        case ti_em_xge_rx_miss_type_ETHER_TYPE :
        case ti_em_xge_rx_miss_type_SERVICE_TYPE :
            status = Emti_moduleChainRxMissEnable(TI_EM_CHAINING_XGE, miss_type);
            break;
        default :
            status = EM_ERR;
            break;
    }    

    return status;
}

em_status_t ti_em_xge_rx_miss_disable(ti_em_xge_rx_miss_type_t miss_type)
{
    em_status_t status;
    switch( miss_type ) 
    {
        case ti_em_xge_rx_miss_type_ETHER_TYPE :
        case ti_em_xge_rx_miss_type_SERVICE_TYPE :
            status = Emti_moduleChainRxMissDisable(TI_EM_CHAINING_XGE, miss_type);
            break;
        default :
            status = EM_ERR;
            break;
    }    

    return status;
}

