/*
 * Copyright (c) 2012, Texas Instruments Incorporated - http://www.ti.com/
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions are met:
 *       * Redistributions of source code must retain the above copyright
 *         notice, this list of conditions and the following disclaimer.
 *       * Redistributions in binary form must reproduce the above copyright
 *         notice, this list of conditions and the following disclaimer in the
 *         documentation and/or other materials provided with the distribution.
 *       * Neither the name of Texas Instruments Incorporated nor the
 *         names of its contributors may be used to endorse or promote products
 *         derived from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *   DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
 *   DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


/**
 *   @file  event_machine_init.c
 *
 *   @brief
 *      Open Event Machine Implementation for Keystone1 devices.
 *
 */

#include <inttypes.h>
#include <stddef.h>
#include <string.h>
#include <ti/runtime/openem/event_machine_macros.h>
#include <ti/runtime/openem/event_machine_types.h>
#include <ti/runtime/openem/event_machine_core_mask.h>
#include <ti/runtime/openem/event_machine_hw_macros.h>
#include <ti/runtime/openem/event_machine_hw_types.h>
#include <ti/runtime/openem/event_machine_hw_ti_macros.h>
#include <ti/runtime/openem/event_machine_hw_ti_types.h>
#include <ti/runtime/openem/event_machine_hw_ti_functions.h>
#include <ti/runtime/openem/src/event_machine_hwpform.h>
#include <ti/runtime/openem/src/event_machine_private_macros.h>
#include <ti/runtime/openem/src/event_machine_private_types.h>
#include <ti/runtime/openem/src/event_machine_private_variables.h>
#include <ti/runtime/openem/src/event_machine_private_functions.h>
#include <ti/runtime/openem/modules/event_machine_modules.h>

em_status_t
ti_em_init_global (
    const ti_em_config_t * config_ptr
)
{
    int32_t priority;
    uint32_t index;
    uint32_t queueNb;
    uint32_t lvDmaQueueNb;
    em_status_t lvRetStatus;
    int32_t init = 1;
    int32_t noPreload = 0;
    Emti_Pcb_s *lvPcbPtr;
    uint32_t lvApPrivateEvtNb = 0;
    uint32_t lvCdPrivateEvtNb = 0;
    ti_em_queue_id_t lvApFreeQueueBaseIdx;
    ti_em_queue_id_t lvCdFreeQueueBaseIdx;

    /*
     * Check if we want to join a existing (and initialized instance) 
     */
    if (config_ptr == NULL)
        init = 0;

    /************************/
    /*
     * initialize variables 
     */
    /************************/

    /*
     * Reset global error function pointer 
     */
    Emti_svErrorFunctionPtr = NULL;

    /*
     * Reset global trace function pointer 
     */
    Emti_svTraceFunctionPtr = NULL;

    /*
     * If needed, initialize shared memory 
     */
    Emti_sharedMemInit (Emti_SV_RW_SECTION_LENGTH, init);

    Emti_svEoMdb = Emti_GLOBAL_INIT (__Emti_svEoMdb[0], sizeof (__Emti_svEoMdb));
    Emti_svQueueFastMdb = Emti_GLOBAL_INIT (__Emti_svQueueFastMdb[0], sizeof (__Emti_svQueueFastMdb));
    Emti_svQueueSlowMdb = Emti_GLOBAL_INIT (__Emti_svQueueSlowMdb[0], sizeof (__Emti_svQueueSlowMdb));
    Emti_svEventGroupMdb = Emti_GLOBAL_INIT (__Emti_svEventGroupMdb[0], sizeof (__Emti_svEventGroupMdb));
    Emti_svQueueGroupMdb = Emti_GLOBAL_INIT (__Emti_svQueueGroupMdb[0], sizeof (__Emti_svQueueGroupMdb));

    if (!init)
    {
        /*
         * Retrieve Psb from database 
         */
        Emti_svPsbPtr = Emti_GLOBAL_INIT (Emti_svPsb, sizeof (Emti_svPsb));

        /*
         * Retrieve Psb from database 
         */
        Emti_svPsbSlowPtr = Emti_GLOBAL_INIT (Emti_svPsbSlow, sizeof (Emti_svPsbSlow));

        /*
         * Retrieve Pcb from database 
         */
        lvPcbPtr = Emti_GLOBAL_INIT (Emti_svPcb, sizeof (Emti_Pcb_s));
        memcpy ((void *) &Emti_svPcb, (void *) lvPcbPtr, sizeof (Emti_Pcb_s));

        /*
         * Indicate that this core is not part of the instance 
         */
        Emti_svPcb.externalCore = 1;
        return EM_OK;
    }
    else
        Emti_svPcb.externalCore = 0;

    /*
     * Initialize the Global Event Machine shared variables 
     */
    memset ((void *) &Emti_svPcb, 0, sizeof (Emti_Pcb_s));
    memset ((void *) &Emti_svEoMdb[0], 0, sizeof (Emti_EO_s) * EM_MAX_EOS);
    memset ((void *) &Emti_svQueueSlowMdb[0], 0, sizeof (Emti_QueueSlow_s) * EM_MAX_QUEUES);
    memset ((void *) &Emti_svQueueFastMdb[0], 0, sizeof (Emti_QueueFast_s) * EM_MAX_QUEUES);
    memset ((void *) &Emti_svEventGroupMdb[0], 0, sizeof (Emti_EventGroup_s) * EM_MAX_EVENT_GROUPS);
    memset ((void *) &Emti_svQueueGroupMdb[0], 0, sizeof (Emti_QueueGroup_s) * EM_MAX_QUEUE_GROUPS);

    /*
     * Check that config_ptr->hw_queue_base_idx is a multiple of TI_EM_HW_QUEUE_STEP
     */
    if ((config_ptr->hw_queue_base_idx & (TI_EM_HW_QUEUE_STEP - 1)) != 0)
    {
        Emti_logError ("HW queue base index (%d) is not a multiple of TI_EM_HW_QUEUE_STEP (%d).\n",
                       config_ptr->hw_queue_base_idx, TI_EM_HW_QUEUE_STEP);
        return EM_ERR;
    }
    
    /*
     * configure the PA Tx queue Idx
     */
    Emti_svPcb.shadowRDsc.paTxQueueIdx = Emti_PA_TX_QUEUE_IDX;

    /*
     * configure number of static queues in router 
     */
    Emti_svPcb.shadowRDsc.staticQueueNum = EM_QUEUE_STATIC_NUM;
    Emti_svPcb.shadowRDsc.staticQueueNumInSet = TI_EM_STATIC_QUEUE_NUM_IN_SET;

    /*************************** 
     *
     * Copy the configuration. 
     * 
     ***************************/
    Emti_svPcb.poolNum = config_ptr->pool_num;
    if (Emti_svPcb.poolNum > TI_EM_POOL_NUM)
    {
        Emti_logError ("Too many pools defined (%u)\n", Emti_svPcb.poolNum);
        return EM_ERR;
    }
    Emti_svPcb.myDeviceIdx = config_ptr->my_device_idx;
    Emti_svPcb.myProcessIdx = config_ptr->my_process_idx;
    Emti_svPcb.hwQueueBaseIdx = config_ptr->hw_queue_base_idx;
    Emti_svPcb.hwInterruptBaseIdx = config_ptr->hw_interrupt_base_idx;
    Emti_svPcb.hwSemIdx = config_ptr->hw_sem_idx;
    Emti_svPcb.pdspNum = config_ptr->pdsp_num;

    if (Emti_svPcb.pdspNum > 5)
    {
        Emti_logError ("Keystone II OpenEm does not support a scheduler with more than 5 threads (%d).\n",
                       Emti_svPcb.pdspNum);
        return EM_ERR;
    }
    for (index = 0; index < config_ptr->pdsp_num; index++)
        Emti_svPcb.pdspIdxTbl[index] = config_ptr->pdsp_idx_tbl[index];
    lvApFreeQueueBaseIdx = config_ptr->ap_region_queue_idx;
    lvCdFreeQueueBaseIdx = config_ptr->cd_region_queue_idx;

    for (index = 0; index < Emti_svPcb.poolNum; index++)
    {
        Emti_svPcb.poolConfigTbl[index].buf_size = config_ptr->pool_config_tbl[index].buf_size;
        Emti_svPcb.poolConfigTbl[index].dsc_wsize = config_ptr->pool_config_tbl[index].dsc_wsize;
        Emti_svPcb.poolConfigTbl[index].buf_mode = config_ptr->pool_config_tbl[index].buf_mode;
        Emti_svPcb.poolConfigTbl[index].free_queue_idx = config_ptr->pool_config_tbl[index].free_queue_idx;
    }

    /*
     * store queue reg D addresses 
     */
    {
        CSL_Qm_queue_managementQueue_mgmt_groupRegs *lvQmAddr;

        lvQmAddr = (void *) CSL_QMSS_CFG_QM_1_QUEUE_MANAGEMENT_REGS;

        /*
         * public free queues 
         */
        for (index = 0; index < Emti_svPcb.poolNum; index++)
            Emti_svPcb.freeQueueRegDPtrTbl[index] =
                (uint32_t) (&(lvQmAddr[config_ptr->pool_config_tbl[index].free_queue_idx].QUEUE_REG_D));
    }

    /**********************************************************************
     ************* check and configure preload sizes **********************
     **********************************************************************/

    /*
     * Check if preloading is used 
     */
    if (config_ptr->preload_config_ptr == NULL)
    {
        noPreload = 1;
        goto skip_preload_check;
    }

    {
        ti_em_preload_config_t *preload_config_ptr = config_ptr->preload_config_ptr;

        /*
         * Check that size_a is a multiple of 16 bytes 
         */
        if ((preload_config_ptr->preload_size_a & 0xF) != 0)
        {
            Emti_logError ("Preloading SIZE_A %d is not a multiple of 16.\n", preload_config_ptr->preload_size_a);
            return EM_ERR;
        }
        /*
         * Check that size_a is in correct range 
         */
        if (((preload_config_ptr->preload_size_a >> 4) > 256) || (preload_config_ptr->preload_size_a == 0))
        {
            Emti_logError ("Preloading SIZE_A %d is out of requested range [1..256]x16 bytes .\n",
                           preload_config_ptr->preload_size_a);
            return EM_ERR;
        }
        /*
         * Check that size_b is a multiple of 16 bytes 
         */
        if ((preload_config_ptr->preload_size_b & 0xF) != 0)
        {
            Emti_logError ("Preloading SIZE_B %d is not a multiple of 16.\n", preload_config_ptr->preload_size_b);
            return EM_ERR;
        }
        /*
         * Check that size_b is in correct range 
         */
        if (((preload_config_ptr->preload_size_b >> 4) > 256) || (preload_config_ptr->preload_size_b == 0))
        {
            Emti_logError ("Preloading SIZE_B %d is out of requested range [1..256]x16 bytes .\n",
                           preload_config_ptr->preload_size_b);
            return EM_ERR;
        }
        /*
         * Check that size_c is a multiple of 16 bytes 
         */
        if ((preload_config_ptr->preload_size_c & 0xF) != 0)
        {
            Emti_logError ("Preloading SIZE_C %d is not a multiple of 16.\n", preload_config_ptr->preload_size_c);
            return EM_ERR;
        }
        /*
         * Check that size_c is in correct range 
         */
        if (((preload_config_ptr->preload_size_c >> 4) > 0x10000) || (preload_config_ptr->preload_size_c == 0))
        {
            Emti_logError ("Preloading SIZE_C %d is out of requested range [1..64K]x16 bytes .\n",
                           preload_config_ptr->preload_size_c);
            return EM_ERR;
        }

        for (index = 0; index < TI_EM_CORE_NUM; index++)
            Emti_svPcb.preloadDsc.localFreeQueueIdxTbl[index] = preload_config_ptr->local_free_queue_idx_tbl[index];

        /*
         * Convert sizes for EM PDSP scheduler copy of PCB 
         */
        Emti_svPcb.preloadDsc.preloadSizeA = (uint8_t) ((preload_config_ptr->preload_size_a >> 4) - 1);
        Emti_svPcb.shadowSDsc.preloadSizeA = (uint8_t) ((preload_config_ptr->preload_size_a >> 4) - 1);
        Emti_svPcb.preloadDsc.preloadSizeB = (uint8_t) ((preload_config_ptr->preload_size_b >> 4) - 1);
        Emti_svPcb.shadowSDsc.preloadSizeB = (uint8_t) ((preload_config_ptr->preload_size_b >> 4) - 1);
        Emti_svPcb.preloadDsc.preloadSizeC = (uint16_t) ((preload_config_ptr->preload_size_c >> 4) - 1);
        Emti_svPcb.shadowSDsc.preloadSizeC = (uint16_t) ((preload_config_ptr->preload_size_c >> 4) - 1);

    }
    Emti_svPcb.preloadEnabled = TI_EM_PRELOAD_ENABLED;

  skip_preload_check:

    queueNb = config_ptr->hw_queue_base_idx;
    lvDmaQueueNb = config_ptr->dma_queue_base_idx;
    Emti_svPcb.dma_queue_base_idx = config_ptr->dma_queue_base_idx;
    Emti_svPcb.dma_idx = config_ptr->dma_idx;

    /**********************************************************************
     *********************** Scheduling (SD) Queues ***********************
     **********************************************************************/
    {
        /*
         * Open all the scheduling queues. 
         */
        for (index = 0; index < Emti_SD_QUEUE_NUM_UP; index++)
        {
            lvRetStatus = ti_em_hw_queue_open (queueNb);
            if (lvRetStatus != EM_OK)
                return EM_ERR_ALLOC_FAILED;

            /*
             * Remember the priority queue. 
             */
            if (index == 0)
            {
                Emti_svPcb.sdQueueBaseIdx = (uint16_t) queueNb;
                Emti_svPcb.shadowSDsc.sdQueueBaseIdx = (uint16_t) queueNb;
                Emti_svPcb.shadowRDsc.sdQueueBaseIdx = (uint16_t) queueNb;
            }

            /*
             * Enable the threshold detection 
             */
            Emti_setQueueThreshold (queueNb, 1, 1);

            queueNb++;
        }
        /*
         * Debug Message: 
         */
        Emti_logDbg ("Scheduling [%d] Queue = 0x%x\n", 0, queueNb - Emti_SD_QUEUE_NUM_UP);
        Emti_logDbg ("Scheduling [%d] Queue = 0x%x\n", index - 1, queueNb - 1);
    }

    if (Emti_svPcb.pdspNum == 0)
        goto skip_atomic_processing_allocation0;

    /**********************************************************************
     ************** Atomic Processing Scheduling (APSD) Queues ************
     **********************************************************************/
    {
        /*
         * Open all the atomic-processing queues. 
         */
        for (index = 0; index < Emti_SD_QUEUE_NUM_UP; index++)
        {
            lvRetStatus = ti_em_hw_queue_open (queueNb);
            if (lvRetStatus != EM_OK)
                return EM_ERR_ALLOC_FAILED;

            /*
             * Remember the priority queue. 
             */
            if (index == 0)
            {
                Emti_svPcb.apSdQueueBaseIdx = (uint16_t) queueNb;
                Emti_svPcb.shadowSDsc.apSdQueueBaseIdx = (uint16_t) queueNb;
            }

            /*
             * Enable the threshold detection 
             */
            Emti_setQueueThreshold (queueNb, 1, 1);

            queueNb++;
        }
        /*
         * Debug Message: 
         */
        Emti_logDbg ("AP Scheduling[%d] Queue 0x%x\n", 0, queueNb - Emti_SD_QUEUE_NUM_UP);
        Emti_logDbg ("AP Scheduling[%d] Queue 0x%x\n", index - 1, queueNb - 1);
    }

    /**********************************************************************
     ********************** Atomic Processing (AP) Queues *****************
     **********************************************************************/
    {
        /*
         * Open all the atomic hw queues. 
         */
        for (index = 0; index < Emti_AP_QUEUE_NUM; index++)
        {
            lvRetStatus = ti_em_hw_queue_open (queueNb);
            if (lvRetStatus != EM_OK)
                return EM_ERR_ALLOC_FAILED;

            /*
             * Remember the priority queue. 
             */
            if (index == 0)
            {
                Emti_svPcb.apQueueBaseIdx = (uint16_t) queueNb;
                Emti_svPcb.shadowSDsc.apQueueBaseIdx = (uint16_t) queueNb;
            }
            queueNb++;
        }
        /*
         * Debug Message: 
         */
        Emti_logDbg ("Atomic Hw[%d] Queue 0x%x\n", 0, queueNb - Emti_AP_QUEUE_NUM);
        Emti_logDbg ("Atomic Hw[%d] Queue 0x%x\n", index - 1, queueNb - 1);
    }

    /****************************************************************
     ******************** Atomic Processing Free (APfree) Queues ****
     ****************************************************************/
    {
        /*
         * Open all the atomic processing free queues. 
         */
        for (index = 0; index < TI_EM_PRIO_NUM; index++)
        {
            lvRetStatus = ti_em_hw_queue_open (queueNb);
            if (lvRetStatus != EM_OK)
                return EM_ERR_ALLOC_FAILED;

            /*
             * Remember the 1st queue. 
             */
            if (index == 0)
            {
                Emti_svPcb.apFreeQueueBaseIdx = (uint16_t) queueNb;
                Emti_svPcb.shadowSDsc.apFreeQueueBaseIdx = (uint16_t) queueNb;
            }

            queueNb++;
        }
        /*
         * Debug Message: 
         */
        Emti_logDbg ("AP EventDsc Free Queue 0x%x \n", (uint16_t) queueNb - TI_EM_PRIO_NUM);
        Emti_logDbg ("AP EventDsc Free Queue 0x%x \n", (uint16_t) queueNb - 1);
    }

  skip_atomic_processing_allocation0:

    /****************************************************************
     ********************** Prefetch completion Queues **************
     ****************************************************************/
    {
        /*
         * These queues are the return queues when preloading is enable 
         */
        for (index = 0; index < TI_EM_CORE_NUM; index++)
        {
            lvRetStatus = ti_em_hw_queue_open (queueNb);
            if (lvRetStatus != EM_OK)
                return EM_ERR_ALLOC_FAILED;

            /*
             * Remember the base index 
             */
            if (index == 0)
            {
                Emti_svPcb.preloadDsc.returnQueueBaseIdx = (uint16_t) queueNb;
                Emti_svPcb.shadowSDsc.returnQueueBaseIdx = (uint16_t) queueNb;
            }

            queueNb++;
        }
        /*
         * Debug Message: 
         */
        Emti_logDbg ("PreFetch Completion Queue Hw[%d] Queue 0x%x\n", 0, queueNb - TI_EM_CORE_NUM);
        Emti_logDbg ("PreFetch Completion Queue Hw[%d] Queue 0x%x\n", index - 1, queueNb - 1);
    }

    /******************************************* ********************
     ********************** Prefetch Queue (PF) Queues **************
     ****************************************************************/
    {
        /*
         * These queues will contain local or global prefetch events 
         */
        for (index = 0; index < TI_EM_CORE_NUM; index++)
        {
            lvRetStatus = ti_em_hw_queue_open (queueNb);
            if (lvRetStatus != EM_OK)
                return EM_ERR_ALLOC_FAILED;

            /*
             * Remember the base index 
             */
            if (index == 0)
            {
                Emti_svPcb.pfQueueBaseIdx = (uint16_t) queueNb;
                Emti_svPcb.shadowSDsc.pfQueueBaseIdx = (uint16_t) queueNb;
            }

            queueNb++;
        }
        /*
         * Debug Message: 
         */
        Emti_logDbg ("PreFetch Queue Hw[%d] Queue 0x%x\n", 0, queueNb - TI_EM_CORE_NUM);
        Emti_logDbg ("PreFetch Queue Hw[%d] Queue 0x%x\n", index - 1, queueNb - 1);
    }

    /****************************************************************
     ******************** CT command free queue *********************
     ****************************************************************/
    {
        /*
         * to be filled with events at init time 
         */
        lvRetStatus = ti_em_hw_queue_open (queueNb);
        if (lvRetStatus != EM_OK)
            return EM_ERR_ALLOC_FAILED;
        Emti_svPcb.ctCommandFreeQueueIdx = (uint16_t) queueNb++;

        /*
         * Debug Message: 
         */
        Emti_logDbg ("CT command free queue 0x%x \n", Emti_svPcb.ctCommandFreeQueueIdx);
    }

    /****************************************************************
     ******************** AP command free queue *********************
     ****************************************************************/
    {
        /*
         * to be filled with events at init time 
         */
        lvRetStatus = ti_em_hw_queue_open (queueNb);
        if (lvRetStatus != EM_OK)
            return EM_ERR_ALLOC_FAILED;
        Emti_svPcb.apCommandFreeQueueIdx = (uint16_t) queueNb++;

        /*
         * Debug Message: 
         */
        Emti_logDbg ("AP command free queue 0x%x \n", Emti_svPcb.apCommandFreeQueueIdx);
    }

    /****************************************************************
     ******************** PF command free queue *********************
     ****************************************************************/
    {
        /*
         * to be filled with events at init time 
         */
        lvRetStatus = ti_em_hw_queue_open (queueNb);
        if (lvRetStatus != EM_OK)
            return EM_ERR_ALLOC_FAILED;
        Emti_svPcb.pfCommandFreeQueueIdx = (uint16_t) queueNb++;

        /*
         * Debug Message: 
         */
        Emti_logDbg ("PF command free queue 0x%x \n", Emti_svPcb.pfCommandFreeQueueIdx);
    }

    /****************************************************************
     ********************  Response (RP) Queue **********************
     ****************************************************************/
    {
        lvRetStatus = ti_em_hw_queue_open (queueNb);
        if (lvRetStatus != EM_OK)
            return EM_ERR_ALLOC_FAILED;
        Emti_svPcb.rpQueueIdx = (uint16_t) queueNb++;

        /*
         * Debug Message: 
         */
        Emti_logDbg ("Response Queue 0x%x \n", Emti_svPcb.rpQueueIdx);
    }

    /****************************************************************
     ********************  Command (CD) Queue ***********************
     ****************************************************************/
    if (Emti_svPcb.pdspNum > 0)
    {
        uint32_t lvPrivateEventPtr;
        uint32_t lvSchedIdx;

        for (lvSchedIdx = 0; lvSchedIdx < Emti_svPcb.pdspNum; lvSchedIdx++)
        {
            lvRetStatus = ti_em_hw_queue_open (queueNb);
            if (lvRetStatus != EM_OK)
                return EM_ERR_ALLOC_FAILED;

            /*
             * store cd queue idx in DSP SW context 
             */
            Emti_svPcb.cdQueueIdx[lvSchedIdx] = (uint16_t) queueNb++;
            Emti_queueEmpty (Emti_svPcb.cdQueueIdx[lvSchedIdx]);

            /*
             * retrieve private memory region base address (virtual) 
             */
            lvPrivateEventPtr = (uint32_t) Emti_getPdspDataRamBaseAddress (Emti_svPcb.pdspIdxTbl[lvSchedIdx]);

            /*
             * store cd queue idx in PDSP SW context - be carefull, keep (uint16_t*) cast as PDSP will read 2 bytes 
             */
            *(uint16_t *) (lvPrivateEventPtr) = (uint16_t) Emti_svPcb.cdQueueIdx[lvSchedIdx];
        }

        /*
         * Debug Message: 
         */
        Emti_logDbg ("Command Queue 0x%x \n", Emti_svPcb.cdQueueIdx[0]);

        /*
         * In case of multi-threaded scheduler, it is needed to use
         * an HW queue for Atomic Processing.
         */
        lvRetStatus = ti_em_hw_queue_open (queueNb);
        if (lvRetStatus != EM_OK)
            return EM_ERR_ALLOC_FAILED;

        /*
         * store AP mutex queue idx in DSP SW context 
         */
        Emti_svPcb.apMutexQueueIdx = (uint16_t) queueNb;
        Emti_svPcb.shadowSDsc.apMutexQueueIdx = (uint16_t) queueNb++;
        Emti_svPcb.shadowSDsc.processIdx = (uint16_t) Emti_svPcb.myProcessIdx;
        Emti_queueEmpty (Emti_svPcb.apMutexQueueIdx);

        /*
         * Debug Message: 
         */
        Emti_logDbg ("AP Mutex Queue 0x%x \n", Emti_svPcb.apMutexQueueIdx);

    }

    if (noPreload)
        goto skip_preload_init;

    /****************************************************************
     ********************  DMA preload Tx Queue *********************
     ****************************************************************/
    {
        ti_em_queue_id_t lvDmaTxQueueId;
        for (index = 0; index < TI_EM_CORE_NUM; index++)
        {
            lvDmaTxQueueId = ti_em_get_absolute_queue_id (config_ptr->dma_idx, lvDmaQueueNb);
            /*
             * Open all the DMA queues. 
             */
            lvRetStatus = ti_em_hw_queue_open (lvDmaTxQueueId);
            if (lvRetStatus != EM_OK)
                return EM_ERR_ALLOC_FAILED;

            /*
             * Remember the priority queue. 
             */
            if (index == 0)
            {
                Emti_svPcb.preloadDsc.txQueueBaseIdx = (uint16_t) lvDmaTxQueueId;
                Emti_svPcb.shadowSDsc.txQueueBaseIdx = (uint16_t) lvDmaTxQueueId;
            }

            lvDmaQueueNb++;
        }
        /*
         * Debug Message: 
         */
        Emti_logDbg ("DMA [%d] Queue = 0x%x\n", 0, lvDmaTxQueueId - TI_EM_CORE_NUM - 1);
        Emti_logDbg ("DMA [%d] Queue = 0x%x\n", index - 1, lvDmaTxQueueId);
    }

    /****************************************************************
     ********************  RX Flows *********************************
     ****************************************************************/
    {
        for (index = 0; index < TI_EM_CORE_NUM; index++)
        {
            em_status_t lvRxFlowStatus;

            lvRxFlowStatus =
                ti_em_rx_flow_open (config_ptr->dma_idx, config_ptr->dma_queue_base_idx + index,
                                    Emti_svPcb.preloadDsc.returnQueueBaseIdx + index,
                                    Emti_svPcb.preloadDsc.localFreeQueueIdxTbl[index], 1);
            if (lvRxFlowStatus != EM_OK)
            {
                Emti_logError ("Error Opening Rx flow\n");
                return EM_ERR;
            }
        }
        Emti_logDbg ("Opened Rx flow : %d\n", config_ptr->dma_queue_base_idx);
        Emti_logDbg ("Opened Rx flow : %d\n", config_ptr->dma_queue_base_idx - index - 1);
    }

    /****************************************************************
     ********************  TX Channels ******************************
     ****************************************************************/
    {
        uint32_t lvTxchannelNb = config_ptr->dma_queue_base_idx;

        for (index = 0; index < TI_EM_CORE_NUM; index++)
        {
            em_status_t lvTxChannelStatus;

            lvTxChannelStatus = ti_em_tx_channel_open (config_ptr->dma_idx, lvTxchannelNb + index);
            if (lvTxChannelStatus != EM_OK)
            {
                Emti_logError ("Error Opening Tx channel\n");
                return EM_ERR;
            }
        }
        Emti_logDbg ("Opened Tx channel : %d\n", lvTxchannelNb);
        Emti_logDbg ("Opened Tx channel : %d\n", lvTxchannelNb + index - 1);
    }

    /****************************************************************
     ********************  RX Channels ******************************
     ****************************************************************/
    {
        uint32_t lvRxchannelNb = config_ptr->dma_queue_base_idx;

        for (index = 0; index < TI_EM_CORE_NUM; index++)
        {
            em_status_t lvRxChannelStatus;

            lvRxChannelStatus = ti_em_rx_channel_open (config_ptr->dma_idx, lvRxchannelNb + index);
            if (lvRxChannelStatus != EM_OK)
            {
                Emti_logError ("Error Opening Rx channel\n");
                return EM_ERR;
            }
        }
        Emti_logDbg ("Opened Rx channel : %d\n", lvRxchannelNb);
        Emti_logDbg ("Opened Rx channel : %d\n", lvRxchannelNb + index - 1);
    }

  skip_preload_init:

    /****************************************************************
     ************** Call chaining global initialization *************
     ****************************************************************/
    lvRetStatus =
        Emti_moduleChainGlobalInit (config_ptr->chain_config_ptr, &queueNb, lvCdFreeQueueBaseIdx, &lvCdPrivateEvtNb);
    if ((lvRetStatus != EM_OK) && (lvRetStatus != EM_ERR_NOT_FOUND))
        return lvRetStatus;


    /****************************************************************
     ******************** release DMA queues ************************
     ****************************************************************/
#ifdef TI_EM_NO_PKTDMA_RELEASE
    Emti_svPcb.dma_idx = (ti_em_dma_id_t) - 1;
#endif

    if (Emti_svPcb.dma_idx == (ti_em_dma_id_t) - 1)
    {
        /*
         * We need at least one standard queue because the release DMA queue is used
         * by firmware in case of error to free scattered events.
         */

        ti_em_queue_id_t lvRsTxQueueId = queueNb;
        lvRetStatus = ti_em_hw_queue_open (lvRsTxQueueId);
        if (lvRetStatus != EM_OK)
            return EM_ERR_ALLOC_FAILED;
        Emti_svPcb.rsTxQueueIdx = (uint16_t) lvRsTxQueueId;
        Emti_svPcb.shadowRDsc.rsTxQueueIdx = (uint16_t) lvRsTxQueueId;
        queueNb++;

        Emti_logDbg ("release SW queue 0x%x \n", Emti_svPcb.rsTxQueueIdx);

        goto skip_release_pktdma_allocation;
    }
    else
    {
        ti_em_queue_id_t lvDmaTxQueueId;
        lvDmaTxQueueId = ti_em_get_absolute_queue_id (config_ptr->dma_idx, lvDmaQueueNb);

        /*
         * open release DMA Tx queue 
         */
        lvRetStatus = ti_em_hw_queue_open (lvDmaTxQueueId);
        if (lvRetStatus != EM_OK)
            return EM_ERR_ALLOC_FAILED;
        Emti_svPcb.rsTxQueueIdx = (uint16_t) lvDmaTxQueueId;
        Emti_svPcb.shadowRDsc.rsTxQueueIdx = (uint16_t) lvDmaTxQueueId;
        lvDmaQueueNb++;

        /*
         * Debug Message: 
         */
        Emti_logDbg ("release DMA Tx queue 0x%x \n", Emti_svPcb.rsTxQueueIdx);

        /*
         * open release DMA free Rx queue 
         */
        lvRetStatus = ti_em_hw_queue_open (queueNb);
        if (lvRetStatus != EM_OK)
            return EM_ERR_ALLOC_FAILED;
        Emti_svPcb.rsFreeQueueIdx = (uint16_t) queueNb++;
        /*
         * Debug Message: 
         */
        Emti_logDbg ("release DMA free Rx queue 0x%x \n", Emti_svPcb.rsFreeQueueIdx);
    }

    /****************************************************************
     ******************** release DMA RX Flows **********************
     ****************************************************************/
    {
        em_status_t lvRxFlowStatus;

        lvRxFlowStatus =
            ti_em_rx_flow_open (config_ptr->dma_idx, (lvDmaQueueNb - 1), Emti_svPcb.rsFreeQueueIdx,
                                Emti_svPcb.rsFreeQueueIdx, 0);

        if (lvRxFlowStatus != EM_OK)
        {
            Emti_logError ("Error Opening Rx flow\n");
            return EM_ERR;
        }
        Emti_logDbg ("Opened Rx flow : %d\n", lvRxFlowStatus);
    }

    /****************************************************************
     ******************** release DMA TX Channels *******************
     ****************************************************************/
    {
        uint32_t lvTxchannelNb = (lvDmaQueueNb - 1);
        em_status_t lvTxChannelStatus;

        lvTxChannelStatus = ti_em_tx_channel_open (config_ptr->dma_idx, lvTxchannelNb);
        if (lvTxChannelStatus != EM_OK)
        {
            Emti_logError ("Error Opening Tx channel\n");
            return EM_ERR;
        }
        Emti_logDbg ("Opened Tx channel : %d\n", lvTxchannelNb);
    }

    /****************************************************************
     ******************** release DMA RX Channels *******************
     ****************************************************************/
    {
        uint32_t lvRxchannelNb = (lvDmaQueueNb - 1);
        em_status_t lvRxChannelStatus;

        lvRxChannelStatus = ti_em_rx_channel_open (config_ptr->dma_idx, lvRxchannelNb);
        if (lvRxChannelStatus != EM_OK)
        {
            Emti_logError ("Error Opening Rx channel\n");
            return EM_ERR;
        }
        Emti_logDbg ("Opened Rx channel : %d\n", lvRxchannelNb);
    }

  skip_release_pktdma_allocation:

    /*******************************************************************************
     ***   Verification that we did not allocate more than publicized   ************
     *******************************************************************************/
    if ((queueNb - config_ptr->hw_queue_base_idx) > TI_EM_HW_QUEUE_NUM)
    {
        Emti_logError ("Error: allocated too many HW queues\n");
        return EM_ERR;
    }
    if ((lvDmaQueueNb - config_ptr->dma_queue_base_idx) > TI_EM_DMA_QUEUE_NUM)
    {
        Emti_logError ("Error: allocated too many DMA queues\n");
        return EM_ERR;
    }

    /*******************************************************************************
     ******************* push CT command event ctCommandFreeQueueIdx ***************
     *******************************************************************************/
    {
        Emti_CdEventDsc_s *cdEventDscPtr;

        /*
         * Pop off the event dsc from the queue. 
         */
        cdEventDscPtr = (Emti_CdEventDsc_s *) (Emti_popEvent (lvCdFreeQueueBaseIdx, TI_EM_MEM_PRIVATE_DESC));
        if (cdEventDscPtr == NULL)
            return EM_ERR_LIB_FAILED;
        lvCdPrivateEvtNb++;

        /*
         * Now push this descriptor into the CT command free queue. 
         */
        Emti_pushEvent (Emti_svPcb.ctCommandFreeQueueIdx, (uint32_t *) cdEventDscPtr, TI_EM_MEM_PRIVATE_DESC);
    }

    /*******************************************************************************
     ******************* push PF command events pfCommandFreeQueueIdx **************
     *******************************************************************************/
    {
        Emti_CdEventDsc_s *cdEventDscPtr;

        for (index = 0; index < TI_EM_CORE_NUM; index++)
        {
            /*
             * Pop off the event dsc from the queue. 
             */
            cdEventDscPtr = (Emti_CdEventDsc_s *) (Emti_popEvent (lvCdFreeQueueBaseIdx, TI_EM_MEM_PRIVATE_DESC));
            if (cdEventDscPtr == NULL)
                return EM_ERR_LIB_FAILED;
            lvCdPrivateEvtNb++;

            /*
             * Now push this descriptor into the PF command free queue. 
             */
            Emti_pushEvent (Emti_svPcb.pfCommandFreeQueueIdx, (uint32_t *) cdEventDscPtr, TI_EM_MEM_PRIVATE_DESC);
        }
    }

    if (Emti_svPcb.pdspNum == 0)
        goto skip_atomic_processing_allocation1;

    /*******************************************************************************
     ******************* push one token into apMutexQueueIdx    ********************
     *******************************************************************************/
    {
        Emti_CdEventDsc_s *cdEventDscPtr;

        /*
         * Pop off the event dsc from the queue. 
         */
        cdEventDscPtr = (Emti_CdEventDsc_s *) (Emti_popEvent (lvCdFreeQueueBaseIdx, TI_EM_MEM_PRIVATE_DESC));
        if (cdEventDscPtr == NULL)
            return EM_ERR_LIB_FAILED;
        lvCdPrivateEvtNb++;

        /*
         * Now push this descriptor into the CT command free queue. 
         */
        Emti_pushEvent (Emti_svPcb.apMutexQueueIdx, (uint32_t *) cdEventDscPtr, TI_EM_MEM_PRIVATE_DESC);
    }

    /*******************************************************************************
     ******************* push AP command events apCommandFreeQueueIdx **************
     *******************************************************************************/
    {
        Emti_CdEventDsc_s *cdEventDscPtr;

        for (index = 0; index < (TI_EM_CORE_NUM * (1 + TI_EM_PF_LEN)); index++)
        {
            /*
             * Pop off the event dsc from the queue. 
             */
            cdEventDscPtr = (Emti_CdEventDsc_s *) (Emti_popEvent (lvCdFreeQueueBaseIdx, TI_EM_MEM_PRIVATE_DESC));
            if (cdEventDscPtr == NULL)
                return EM_ERR_LIB_FAILED;
            lvCdPrivateEvtNb++;

            /*
             * Now push this descriptor into the AP command free queue. 
             */
            Emti_pushEvent (Emti_svPcb.apCommandFreeQueueIdx, (uint32_t *) cdEventDscPtr, TI_EM_MEM_PRIVATE_DESC);
        }
    }

    /****************************************************************
     ******************* push events to apFreeQueueBaseIdx **********
     ****************************************************************/
    {
        /*
         * Cycle through all the events and ensure that they are associated with one HW atomic queue. 
         */
        for (priority = EM_QUEUE_PRIO_LOWEST; priority <= EM_QUEUE_PRIO_HIGHEST; priority++)
        {
            for (index = 0; index < TI_EM_CORE_NUM * (1 + TI_EM_PF_LEN); index++)
            {
                Emti_EventDsc_s *ptrEventDsc;
                /*
                 * Pop off the event dsc from the queue. 
                 */
                ptrEventDsc = (Emti_EventDsc_s *) (Emti_popEvent (lvApFreeQueueBaseIdx, TI_EM_MEM_PRIVATE_DESC));
                if (ptrEventDsc == NULL)
                    return EM_ERR_LIB_FAILED;
                lvApPrivateEvtNb++;

                /*
                 * Populate the AP EventDsc with the information. 
                 */
                ptrEventDsc->atomicQueueIdx = index;
                ptrEventDsc->pad[0] = 0xffffffff;
                if ((priority == EM_QUEUE_PRIO_LOWEST) && (index == 0))
                    Emti_svPcb.apEventDscBasePtr = Emti_privateDesc_vtop ((void *) ptrEventDsc);

                /*
                 * ensure data are written in memory 
                 */
                Emti_mfence ();

                /*
                 * Now push this descriptor into the EM free queue. 
                 */
                Emti_pushEvent ((uint16_t) (Emti_svPcb.apFreeQueueBaseIdx + priority), (uint32_t *) ptrEventDsc,
                                TI_EM_MEM_PRIVATE_DESC);
            }
        }
    }

  skip_atomic_processing_allocation1:

    /*******************************************************************************
     ***   Verification that we did not allocate more than publicized   ************
     *******************************************************************************/
    if (lvApPrivateEvtNb > TI_EM_AP_PRIVATE_EVENT_NUM)
    {
        Emti_logError ("Error: allocated too many private events\n");
        return EM_ERR;
    }
    if (lvCdPrivateEvtNb > TI_EM_CD_PRIVATE_EVENT_NUM)
    {
        Emti_logError ("Error: allocated too many private events\n");
        return EM_ERR;
    }

    /**********************************************************************
     ********************* Queue Group Configuration **********************
     **********************************************************************/
    {
        em_core_mask_t mask;
        /*
         * Queue Group Configuration: Initialize the core mask. 
         */
        em_core_mask_zero (&mask);

        /*
         * Initialize all the queue groups. 
         */
        for (index = 0; index < EM_MAX_QUEUE_GROUPS; index++)
        {
            em_core_mask_copy (&(Emti_svQueueGroupMdb[index].coreMskTbl), &mask);
            Emti_svQueueGroupMdb[index].state = Emti_QueueGroupState_FREE;
            for (priority = EM_QUEUE_PRIO_LOWEST; priority <= EM_QUEUE_PRIO_HIGHEST; priority++)
            {
                Emti_svQueueGroupMdb[index].sdQueueTbl[priority] = -1;
            }
        }

        /*
         * Default: The default queue group is always created initially. Create the
         * * default core mask for this queue group which selects all the cores. 
         */
        for (index = 0; index < TI_EM_CORE_NUM; index++)
            em_core_mask_set (index, &mask);

        /*
         * Configure the queue group with the all-core mask. 
         */
        em_core_mask_copy (&(Emti_svQueueGroupMdb[EM_QUEUE_GROUP_DEFAULT].coreMskTbl), &mask);
        Emti_svQueueGroupMdb[EM_QUEUE_GROUP_DEFAULT].state = Emti_QueueGroupState_ALLOCATED;
        for (index = TI_EM_CORE_NUM; index < 32; index++)
            em_core_mask_set (index, &mask);
        em_core_mask_copy (&(Emti_svQueueGroupMdb[EM_QUEUE_GROUP_UNDEF].coreMskTbl), &mask);
        Emti_svQueueGroupMdb[EM_QUEUE_GROUP_UNDEF].state = Emti_QueueGroupState_ALLOCATED;
    }

    /*********************************************************************
     ********************* initialize Psb structure **********************
     *********************************************************************/

    {

        /*
         * Initialize the Global Event Machine PSB - fast part
         */
        Emti_svPsbPtr = Emti_GLOBAL_INIT (Emti_svPsb, sizeof (Emti_svPsb));
        memset ((void *) Emti_svPsbPtr, 0, sizeof (Emti_svPsb));

        /*
         * Initialize the Global Event Machine PSB - slow part
         */
        Emti_svPsbSlowPtr = Emti_GLOBAL_INIT (Emti_svPsbSlow, sizeof (Emti_svPsbSlow));
        memset ((void *) Emti_svPsbSlowPtr, 0, sizeof (Emti_svPsbSlow));

        /*
         * The default queue group is 2 now 
         */
        Emti_svPsbPtr->nextFreeQueueGroupIdx = EM_QUEUE_GROUP_DEFAULT + 1;

        /*
         * The default event group is 1 now 
         */
        Emti_svPsbPtr->nextFreeEventGroupIdx = 1;

        /*
         * The default eo is 1 now 
         */
        Emti_svPsbPtr->nextFreeEoIdx = 1;

        /*
         * The default queue set is 0 now as static queues are spread over all sets 
         */
        Emti_svPsbPtr->nextFreeQueueSetIdx = 0;

        for (index = 0; index < Emti_SD_QUEUE_NUM; index++)
        {
            Emti_svPsbSlowPtr->sdQueueStateTbl[index].atomicSetIdx = -1;
            Emti_svPsbSlowPtr->sdQueueStateTbl[index].nextFreeIdxInAtomicSet = -1;
            Emti_svPsbSlowPtr->sdQueueStateTbl[index].nextFreeIdxInParallelSet = -1;
            Emti_svPsbSlowPtr->sdQueueStateTbl[index].parallelSetIdx = -1;
        }

        for (index = 0; index < Emti_QM_QUEUE_NUM; index++)
        {
            Emti_svPsbSlowPtr->hwQueueStateTbl[index].setIdx = -1;
        }

        /*
         * Initialization of processRouteTbl (routeInfo)
         * and deviceRouteTbl (routeInfo) to 0 is done
         * with the initial memset(0) of Emti_svPsbPtr.
         */
    }

    /****************************************************************
     ******************* populate nextFreeSdQueueIdxTbl **********
     ****************************************************************/
    {
        for (priority = EM_QUEUE_PRIO_LOWEST; priority <= EM_QUEUE_PRIO_HIGHEST; priority++)
        {
            Emti_svPsbPtr->nextFreeSdQueueIdxTbl[priority] = priority * TI_EM_CORE_NUM;
        }
    }

    /*
     * send global init command to scheduler CD queue 
     */
    if (Emti_svPcb.pdspNum > 0)
    {
        em_status_t lvStatus;
        uint32_t lvSchedIdx;
        for (lvSchedIdx = 0; lvSchedIdx < Emti_svPcb.pdspNum; lvSchedIdx++)
        {
            Emti_svPcb.shadowSDsc.cdQueueIdx = Emti_svPcb.cdQueueIdx[lvSchedIdx];
            Emti_mfence ();
            lvStatus =
                Emti_connectFirmware (Emti_svPcb.cdQueueIdx[lvSchedIdx], Emti_svPcb.ctCommandFreeQueueIdx,
                                      Emti_svPcb.rpQueueIdx, (uint32_t) & Emti_svPcb.shadowSDsc,
                                      (uint32_t) sizeof (Emti_svPcb.shadowSDsc), (uint32_t) Emti_CtType_SCHEDULER_INIT,
                                      (uint32_t) TI_EM_SLOT_SPCB, (ti_em_process_id_t) 0);
            if (lvStatus != EM_OK)
                return lvStatus;
        }
    }

    /*
     * Perform connect for chaining
     */
    lvRetStatus = Emti_moduleChainConnect ();
    if ((lvRetStatus != EM_OK) && (lvRetStatus != EM_ERR_NOT_FOUND))
        return lvRetStatus;

    /*
     * Copy the PCB in the global shared database for join operation 
     */
    lvPcbPtr = Emti_GLOBAL_INIT (Emti_svPcb, sizeof (Emti_svPcb));
    memcpy ((void *) lvPcbPtr, (void *) &Emti_svPcb, sizeof (Emti_Pcb_s));

    /*
     * Open Event Machine has been initialized globally. 
     */
    return EM_OK;
}

em_status_t
ti_em_init_local (
    void
)
{
    uint32_t index;

    /*
     * reset global EO handle 
     */
    Emti_gvEoHandle = (em_eo_t) (EM_EO_UNDEF);

    /*
     * Initialize per-core local variables 
     */
    memset ((void *) &Emti_gvTcb, 0, sizeof (Emti_Tcb_s));
    memset ((void *) &Emti_gvTsb, 0, sizeof (Emti_Tsb_s));

    /********************************* 
     *
     * copy Emti_Pcb_s to Emti_Tcb_s 
     *
     *********************************/
    Emti_gvTcb.poolNum = Emti_svPcb.poolNum;
    Emti_gvTcb.myDeviceIdx = Emti_svPcb.myDeviceIdx;
    Emti_gvTcb.myProcessIdx = Emti_svPcb.myProcessIdx;
    Emti_gvTcb.hwSemIdx = Emti_svPcb.hwSemIdx;
    Emti_gvTcb.pfQueueIdx = Emti_svPcb.pfQueueBaseIdx + DNUM;
    Emti_gvTcb.rsTxQueueIdx = Emti_svPcb.rsTxQueueIdx;

    for (index = 0; index < TI_EM_POOL_NUM; index++)
    {
        Emti_gvTcb.poolConfigTbl[index].buf_size = Emti_svPcb.poolConfigTbl[index].buf_size;
        Emti_gvTcb.poolConfigTbl[index].dsc_wsize = Emti_svPcb.poolConfigTbl[index].dsc_wsize;
        Emti_gvTcb.poolConfigTbl[index].buf_mode = Emti_svPcb.poolConfigTbl[index].buf_mode;
        Emti_gvTcb.poolConfigTbl[index].free_queue_idx = Emti_svPcb.poolConfigTbl[index].free_queue_idx;

        Emti_gvTcb.freeQueueRegDPtrTbl[index] = Emti_svPcb.freeQueueRegDPtrTbl[index];
    }

    Emti_gvTcb.rpQueueIdx = Emti_svPcb.rpQueueIdx;
    Emti_gvTcb.sdQueueBaseIdx = Emti_svPcb.sdQueueBaseIdx;
    Emti_gvTcb.apEventDscBasePtr = Emti_svPcb.apEventDscBasePtr;
    Emti_gvTcb.pfEventDscBasePtr = Emti_svPcb.pfEventDscBasePtr;

    if (Emti_svPcb.preloadEnabled == TI_EM_PRELOAD_ENABLED)
    {
        Emti_gvTcb.preloadDsc.localFreeQueueIdx = Emti_svPcb.preloadDsc.localFreeQueueIdxTbl[DNUM];
        Emti_gvTcb.preloadDsc.localEventHdl = EM_EVENT_UNDEF;
    }

    /*******************************************
     *
     * chaining local initialization
     *
     *******************************************/
    Emti_moduleChainLocalInit ();

#ifndef TI_EM_COHERENT_IO
    /*
     * global L1D, L1P and L2 invalidate 
     */
    Emti_cacheInvAllL2 ();

    /*
     * Invalidate the XMC Prefetch buffer also. 
     */
    Emti_prefetchBufferInv ();
#endif

    /*
     * store queue reg D addresses 
     */
    {
        CSL_Qm_queue_managementQueue_mgmt_groupRegs *lvQmAddr;

        lvQmAddr = Emti_QmssRegs_ptov (CSL_QMSS_CFG_QM_1_QUEUE_MANAGEMENT_REGS);

        /*
         * PF queue (per-core) 
         */
        Emti_gvTcb.pfQueueRegDPtr = (uint32_t) (&(lvQmAddr[Emti_svPcb.pfQueueBaseIdx + DNUM].QUEUE_REG_D));

        /*
         * release Tx queue 
         */
        Emti_gvTcb.rsTxQueueRegDPtr = (uint32_t) (&(lvQmAddr[Emti_svPcb.rsTxQueueIdx].QUEUE_REG_D));

        /*
         * PL return queue 
         */
        Emti_gvTcb.preloadDsc.returnQueueIdx = Emti_svPcb.preloadDsc.returnQueueBaseIdx + DNUM;
        Emti_gvTcb.preloadDsc.returnQueueRegDPtr =
            (uint32_t) (&(lvQmAddr[Emti_svPcb.preloadDsc.returnQueueBaseIdx + DNUM].QUEUE_REG_D));

        /*
         * AP SD queue 
         */
        Emti_gvTcb.apSdQueueRegDBasePtr = (uint32_t) (&(lvQmAddr[Emti_svPcb.apSdQueueBaseIdx].QUEUE_REG_D));

    }

    /*
     * convert queue reg D addresses of public free queues into virtual (local) 
     */
    {
        uint32_t lvIndex;
        for (lvIndex = 0; lvIndex < TI_EM_POOL_NUM; lvIndex++)
            Emti_gvTcb.freeQueueRegDPtrTbl[lvIndex] =
                (uint32_t) Emti_QmssRegs_ptov (Emti_svPcb.freeQueueRegDPtrTbl[lvIndex]);
    }

    if (Emti_svPcb.externalCore)
        goto skip_local_instance_init;

    /*
     * initialize dispatcher ping-pong flag
     */
    Emti_gvTsb.pingPong = 0;

    /*
     * generate (1 + TI_EM_PF_LEN) command events of type AP per core
     */
    {
        uint32_t lvIndex;
        for (lvIndex = 0; lvIndex < (1 + TI_EM_PF_LEN); lvIndex++)
        {
            Emti_PfEventDsc_s *lvPrivDescPtr;

            /*
             * Pop off the event dsc from the command free queue. 
             */
            lvPrivDescPtr =
                (Emti_PfEventDsc_s *) Emti_popEvent (Emti_svPcb.apCommandFreeQueueIdx, TI_EM_MEM_PRIVATE_DESC);

            lvPrivDescPtr->coreIdx = DNUM;
            Emti_mfence ();

            /*
             * Now push this descriptor into the dispatcher control block 
             */
            Emti_gvTcb.apCommandHdlTbl[lvIndex] =
                ((Emti_privateDesc_vtop (lvPrivDescPtr) & (0xfffffff0)) | Emti_CD_COMMAND_TYPE_AP);
        }
    }

    /*
     * generate 1 command event of type PF per core
     */
    {
        Emti_PfEventDsc_s *lvPrivDescPtr;

        Emti_logDbg ("Pop command from %d\n", Emti_svPcb.pfCommandFreeQueueIdx);

        /*
         * Pop off the event dsc from the command free queue. 
         */
        lvPrivDescPtr = (Emti_PfEventDsc_s *) Emti_popEvent (Emti_svPcb.pfCommandFreeQueueIdx, TI_EM_MEM_PRIVATE_DESC);

        /*
         * The initialization of the ESB pointer is now done in the init_local 
         * function.
         * It makes placement of ecb structure core-independant.
         * On DSP, shadowEsbDscPtr direcly points to physical shadowEsbDsc within TCB structure,
         * on Linux shadowEsbDscPtr is overriden with the PDSP communication memory 
         * virtual address (and thus not in TCB anymore).
         */
        Emti_gvTsb.esbDscPtr = &Emti_gvTsb.esbDsc;  /* for DSP, overriden in Linux case */

        lvPrivDescPtr->ecbPtr = Emti_pdspMap (TI_EM_SLOT_ESB, (void **) &Emti_gvTsb.esbDscPtr);

        if (Emti_gvTsb.esbDscPtr != &Emti_gvTsb.esbDsc)
        {
            memcpy (Emti_gvTsb.esbDscPtr, &Emti_gvTsb.esbDsc, sizeof (Emti_gvTsb.esbDsc));
        }

        lvPrivDescPtr->coreIdx = DNUM;
        lvPrivDescPtr->pad[0] = 0x00000000;
        Emti_mfence ();

        /*
         * Now copy this descriptor into the dispatcher control block 
         */
        Emti_gvTcb.pfCommandHdl =
            Emti_privateDesc_vtop ((void *) ((((uint32_t) (lvPrivDescPtr) & (0xfffffff0)) | Emti_CD_COMMAND_TYPE_PF)));
    }

    /*
     * Increment the coreNum field in Psb 
     */
    Emti_lockSemaphore ();
    {
        uint32_t schedIdx = Emti_svPsbPtr->coreNum % Emti_svPcb.pdspNum;
        CSL_Qm_queue_managementQueue_mgmt_groupRegs *lvQmAddr;
        lvQmAddr = Emti_QmssRegs_ptov (CSL_QMSS_CFG_QM_1_QUEUE_MANAGEMENT_REGS);

        /*
         * CD queue 
         */
        Emti_gvTcb.cdQueueRegDPtr = (uint32_t) (&(lvQmAddr[Emti_svPcb.cdQueueIdx[schedIdx]].QUEUE_REG_D));
        Emti_gvTcb.cdQueueIdx = Emti_svPcb.cdQueueIdx[schedIdx];
    }
    Emti_svPsbPtr->coreNum++;
    Emti_unlockSemaphore ();

  skip_local_instance_init:

    /*
     * set flag to preschedule to be sent 
     */
    Emti_gvTsb.actionFlags = Emti_PF_COMMAND_TO_SEND;

    /*
     * Local Initialization was successful. 
     */
    return EM_OK;
}

em_status_t
ti_em_register_trace_handler (
    ti_em_trace_handler_t handler
)
{
    Emti_svTraceFunctionPtr = handler;

    /*
     * ensure data are written in memory 
     */
    Emti_mfence ();

    return EM_OK;
}

em_status_t
ti_em_unregister_trace_handler (
    void
)
{
    Emti_svTraceFunctionPtr = NULL;

    /*
     * ensure data are written in memory 
     */
    Emti_mfence ();

    return EM_OK;
}
