/*
 * Copyright (c) 2012, Texas Instruments Incorporated - http://www.ti.com/
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions are met:
 *       * Redistributions of source code must retain the above copyright
 *         notice, this list of conditions and the following disclaimer.
 *       * Redistributions in binary form must reproduce the above copyright
 *         notice, this list of conditions and the following disclaimer in the
 *         documentation and/or other materials provided with the distribution.
 *       * Neither the name of Texas Instruments Incorporated nor the
 *         names of its contributors may be used to endorse or promote products
 *         derived from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *   DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
 *   DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


/**
 *   @file  event_machine_interrupt.c
 *
 *   @brief
 *      Open Event Machine Implementation for Keystone1 and KeystoneII devices.
 *      Management of the interrupt generation.
 *
 */

#include <inttypes.h>
#include <stddef.h>
#include <string.h>
#include <ti/runtime/openem/event_machine_macros.h>
#include <ti/runtime/openem/event_machine_types.h>
#include <ti/runtime/openem/event_machine_core_mask.h>
#include <ti/runtime/openem/event_machine_hw_macros.h>
#include <ti/runtime/openem/event_machine_hw_types.h>
#include <ti/runtime/openem/event_machine_hw_ti_macros.h>
#include <ti/runtime/openem/event_machine_hw_ti_types.h>
#include <ti/runtime/openem/event_machine_hw_ti_functions.h>
#include <ti/runtime/openem/src/event_machine_hwpform.h>
#include <ti/runtime/openem/src/event_machine_private_macros.h>
#include <ti/runtime/openem/src/event_machine_private_types.h>
#include <ti/runtime/openem/src/event_machine_private_variables.h>
#include <ti/runtime/openem/src/event_machine_private_functions.h>

static em_status_t
Emti_interrupt_connect_with_firmware (
    ti_em_queue_id_t cmd_free_queue_idx,
    ti_em_queue_id_t return_queue_idx,
    uint32_t cmd_type,
    uint32_t word_to_write
)
{
    volatile Emti_CdEventDsc_s *cdEventDscPtr = NULL;
    int async_cmd = 0;

    /*
     * check if command must be synchronous (waiting RP evnent) or asynchronous (faster but no guarantee) 
     */
    if (return_queue_idx == (ti_em_queue_id_t) - 1)
        async_cmd = 1;

    /*
     * pop off the event from the CD free queue 
     */
    cdEventDscPtr = (volatile Emti_CdEventDsc_s *) Emti_popEvent (cmd_free_queue_idx, TI_EM_MEM_PRIVATE_DESC);

    while (cdEventDscPtr == NULL)
        cdEventDscPtr = (volatile Emti_CdEventDsc_s *) Emti_popEvent (cmd_free_queue_idx, TI_EM_MEM_PRIVATE_DESC);

    cdEventDscPtr = (volatile Emti_CdEventDsc_s *) (((uint32_t) cdEventDscPtr & 0xFFFFFFF0));

    /*
     * fill event descriptor 
     */
    cdEventDscPtr->commandType = (Emti_CdQueueCommandType_e) cmd_type;
    cdEventDscPtr->commandPtr = word_to_write;
    cdEventDscPtr->errorCode = Emti_CdQueueErrorCode_WAITING_CMD;
    if (async_cmd)
        cdEventDscPtr->returnQueue = (uint32_t) cmd_free_queue_idx;
    else
        cdEventDscPtr->returnQueue = (uint32_t) return_queue_idx;

    /*
     * push event to the thread CD queue 
     */
    Emti_mfence ();
    Emti_pushEvent (Emti_gvTcb.cdQueueIdx, (uint32_t *) (((uint32_t) cdEventDscPtr) | Emti_CD_COMMAND_TYPE_CT),
                    TI_EM_MEM_PRIVATE_DESC);

    /*
     * if it is an asynchronous command, do not wait for reply
     */
    if (async_cmd)
        return EM_OK;

    /*
     * waits for the RP event 
     */
    cdEventDscPtr = NULL;
    while (cdEventDscPtr == NULL)
        cdEventDscPtr = (volatile Emti_CdEventDsc_s *) Emti_popEvent (return_queue_idx, TI_EM_MEM_PRIVATE_DESC);

    cdEventDscPtr = (volatile Emti_CdEventDsc_s *) (((uint32_t) cdEventDscPtr & 0xFFFFFFF0));

    /*
     * read the event 
     */
    if (cdEventDscPtr->errorCode != Emti_CdQueueErrorCode_NO_ERROR)
        return EM_ERR_LIB_FAILED;

    /*
     * push to the CD free queue 
     */
    cdEventDscPtr->commandType = Emti_CtType_NULL;
    cdEventDscPtr->commandPtr = 0;
    Emti_mfence ();
    Emti_pushEvent (cmd_free_queue_idx, (uint32_t *) cdEventDscPtr, TI_EM_MEM_PRIVATE_DESC);

    return EM_OK;
}

em_status_t
ti_em_interrupt_enable (
    void
)
{
    em_status_t lvStatus;
    uint32_t enable_word;

    /*
     * The interrupt channel is equal to the HW Interrupt Base Index + 
     * Core Physical Index.
     */
    enable_word = ((DNUM + Emti_svPcb.hwInterruptBaseIdx) & 0xFF) + (1 << 8) + (DNUM << 16);
    lvStatus = Emti_interrupt_connect_with_firmware (Emti_svPcb.ctCommandFreeQueueIdx, Emti_svPcb.rpQueueIdx,   /* Synchronous to avoid race condition */
                                                     (uint32_t) Emti_CtType_SCHEDULER_UPDATE_INTERRUPT,
                                                     (uint32_t) enable_word);
    return lvStatus;
}

em_status_t
ti_em_interrupt_disable (
    void
)
{
    em_status_t lvStatus;
    uint32_t disable_word;

    disable_word = 0 + (DNUM << 16);
    lvStatus = Emti_interrupt_connect_with_firmware (Emti_svPcb.ctCommandFreeQueueIdx, (ti_em_queue_id_t) - 1,  /* Asynchronous */
                                                     (uint32_t) Emti_CtType_SCHEDULER_UPDATE_INTERRUPT,
                                                     (uint32_t) disable_word);
    return lvStatus;
}
