/*
 * Copyright (c) 2012, Texas Instruments Incorporated - http://www.ti.com/
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions are met:
 *       * Redistributions of source code must retain the above copyright
 *         notice, this list of conditions and the following disclaimer.
 *       * Redistributions in binary form must reproduce the above copyright
 *         notice, this list of conditions and the following disclaimer in the
 *         documentation and/or other materials provided with the distribution.
 *       * Neither the name of Texas Instruments Incorporated nor the
 *         names of its contributors may be used to endorse or promote products
 *         derived from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *   DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
 *   DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


/**
 * @file: event_machine_packet.c
 *
 * @brief
 *    Open Event Machine HW TI specific functions implementation
 *    Functions related to packet.
 *
 */

#include <stdio.h>
#include <string.h>
#include <ti/runtime/openem/event_machine_macros.h>
#include <ti/runtime/openem/event_machine_types.h>
#include <ti/runtime/openem/event_machine_group.h>
#include <ti/runtime/openem/event_machine_hw_macros.h>
#include <ti/runtime/openem/event_machine_hw_types.h>
#include <ti/runtime/openem/event_machine_hw_ti_macros.h>
#include <ti/runtime/openem/event_machine_hw_ti_types.h>
#include <ti/runtime/openem/src/event_machine_hwpform.h>
#include <ti/runtime/openem/src/event_machine_private_types.h>
#include <ti/runtime/openem/src/event_machine_private_macros.h>
#include <ti/runtime/openem/src/event_machine_private_variables.h>
#include <ti/runtime/openem/src/event_machine_private_functions.h>
#include <ti/runtime/openem/modules/event_machine_modules.h>

ti_em_packet_t *
ti_em_to_packet (
    em_event_t event
)
{
    Cppi_HostDesc *lvHostDescPtr = (Cppi_HostDesc *) Emti_getCppiHostDescPtr (event);
    return (ti_em_packet_t *) lvHostDescPtr;
}

em_event_t
ti_em_from_packet (
    ti_em_packet_t * packet_ptr
)
{
    Cppi_HostDesc *lvHostDescPtr = (Cppi_HostDesc *) packet_ptr;
    uint32_t lvEventHandle;
    lvEventHandle = ((uint32_t) packet_ptr) | (Emti_getEventMode (lvHostDescPtr) & 0xF);
    return (em_event_t) lvEventHandle;
}

void
ti_em_packet_set_default (
    ti_em_packet_t * packet_ptr
)
{
    uint32_t lvIdx;
    uint32_t *lvPtr = (uint32_t *) packet_ptr;
    Cppi_HostDesc *lvHostDescPtr = (Cppi_HostDesc *) packet_ptr;

    Emti_setCohMode (lvHostDescPtr, TI_EM_COH_MODE_OFF);
    Emti_setDirtyFlag (lvHostDescPtr, 0x0);

    /*
     * dscWsize reset to 16 (4 psData), dscWsize will be configured in ti_em_packet_set_pool_info()
     */
    Emti_setPacketWsize (lvHostDescPtr, Emti_EVENT_DSC_WORD_SIZE_4PSDATA);
    Emti_setPoolIndex (lvHostDescPtr, 0x0);
    Emti_setFreeInfo (lvHostDescPtr, 0);

    /*
     * set free policy to 1 (word -1) 
     */
    Emti_setFreePolicy (lvHostDescPtr, 1);
    Emti_setBufMode (lvHostDescPtr, (ti_em_buf_mode_t) 0x0);
    Emti_setEventMode (lvHostDescPtr, (ti_em_buf_mode_t) 0x0);

    /*
     * descriptor is set to host descriptor, by default ps location is set to descriptor. 
     */
    *lvPtr++ = (Cppi_DescType_HOST << 30) | (Cppi_PSLoc_PS_IN_DESC << 22);

    /*
     * tag info is set to 0 
     */
    *lvPtr++ = 0;

    /*
     * EPIB set to present, return policy set to 1 
     */
    *lvPtr++ = (((uint32_t) Cppi_EPIB_EPIB_PRESENT) << 31) | (1 << 15);

    /*
     * reset words from 3 to Emti_EVENT_DSC_WORD_SIZE_MIN to 0 
     */
    for (lvIdx = 3; lvIdx < Emti_EVENT_DSC_WORD_SIZE_MIN; lvIdx++)
        *lvPtr++ = 0;

    return;
}

void
ti_em_packet_set_buffer_info (
    ti_em_packet_t * packet_ptr,
    ti_em_buffer_config_t buffer_config
)
{
    Cppi_HostDesc *lvHostDescPtr;
    uint16_t lvWord2 = 0;

    lvHostDescPtr = (Cppi_HostDesc *) (packet_ptr);

    /*
     * original buffer pointer 
     */
    lvHostDescPtr->origBuffPtr = (uint32_t) (buffer_config.orig_buffer_ptr);

    /*
     * original buffer length
     */
    Emti_setOrigBufferLength (lvHostDescPtr, buffer_config.orig_buffer_size);

    /*
     * set pool id
     */
    Emti_setPoolIndex (lvHostDescPtr, buffer_config.orig_buffer_pool);

    /*
     * buffer pointer 
     */
    lvHostDescPtr->buffPtr = (uint32_t) (buffer_config.buffer_ptr);

    /*
     * buffer length 
     */
    lvHostDescPtr->buffLen = buffer_config.buffer_size & 0x0003fffff;

    /*
     * free function pointer 
     */
    Emti_packet_set_free_function ((ti_em_packet_t *) (lvHostDescPtr), buffer_config.free_func);

    /*
     * buffer coherency mode 
     */
    Emti_setCohMode (lvHostDescPtr, (ti_em_coh_mode_t) (buffer_config.coh_mode));

    /*
     * push policy 
     */
    Emti_setFreePushPolicy (lvHostDescPtr, buffer_config.free_push_policy);

    /*
     * Updates word 2 
     */
    lvWord2 = (uint16_t) (buffer_config.free_push_policy << 14);
    lvHostDescPtr->packetInfo = (lvHostDescPtr->packetInfo & 0xffffbfff) | lvWord2;

    return;
}

void
ti_em_packet_set_pool_info (
    ti_em_packet_t * packet_ptr,
    ti_em_pool_config_t pool_config
)
{
    Cppi_HostDesc *lvHostDescPtr = (Cppi_HostDesc *) packet_ptr;
    uint16_t lvWord2 = 0;

    Emti_setFreeQueueIndex (lvHostDescPtr, pool_config.free_queue_idx);
    Emti_setFreePolicy (lvHostDescPtr, 0x1);
    Emti_setBufMode (lvHostDescPtr, pool_config.buf_mode);
    Emti_setPacketWsize (lvHostDescPtr, pool_config.dsc_wsize);

    lvWord2 = (uint16_t) ((pool_config.free_queue_idx & 0x3FFF) | (1 << 15));

    /*
     * Updates word 2 
     */
    lvHostDescPtr->packetInfo = (lvHostDescPtr->packetInfo & 0xffff0000) | lvWord2;

    return;
}

void
ti_em_packet_restore_free_info (
    ti_em_packet_t * packet_ptr
)
{
    Cppi_HostDesc *lvHostDescPtr = (Cppi_HostDesc *) packet_ptr;
    uint32_t *lvPacketInfoPtr = (uint32_t *) & (lvHostDescPtr->packetInfo);
    uint32_t *lvEventDescStartPtr = (uint32_t *) (Emti_getEventDescStartAdd ((uint32_t) (lvHostDescPtr)));

    /*
     * Copy Word #12 2 LS bytes into Word #2 2 LS Bytes 
     */
    *lvPacketInfoPtr = (*lvPacketInfoPtr & 0xffff0000) | ((lvEventDescStartPtr[1] & 0x0000ffff) | (1 << 15));

    return;
}

void
ti_em_packet_set_type (
    ti_em_packet_t * packet_ptr,
    em_event_type_t event_type
)
{
    Cppi_HostDesc *lvHostDescPtr = (Cppi_HostDesc *) packet_ptr;
    uint8_t *lvTimestampPtr = (uint8_t *) & lvHostDescPtr->timeStamp;
#ifdef _LITTLE_ENDIAN
    lvTimestampPtr += 3;
#endif
    *lvTimestampPtr = (uint8_t) (event_type & 0xFF);

    return;
}

void
ti_em_packet_set_queue (
    ti_em_packet_t * packet_ptr,
    em_queue_t queue_hdl
)
{
    Cppi_HostDesc *lvHostDescPtr = (Cppi_HostDesc *) packet_ptr;
    uint32_t *lvTimeStampPtr = &(lvHostDescPtr->timeStamp);

    /*
     * Writes the full queue handle to word 8 (FORMAT_A) 
     */
    *lvTimeStampPtr = ((*lvTimeStampPtr & Emti_EVENT_TYPE_MASK) | (queue_hdl & Emti_EQ_HANDLE_MASK));

    return;
}

void
ti_em_packet_set_event_group (
    ti_em_packet_t * packet_ptr,
    em_event_group_t event_group_hdl
)
{
    Cppi_HostDesc *lvHostDescPtr = (Cppi_HostDesc *) packet_ptr;
    uint32_t *lvSoftwareInfo0Ptr = &(lvHostDescPtr->softwareInfo0);

    /*
     * Writes event group index to word 9 (FORMAT_A) 
     */
    *lvSoftwareInfo0Ptr = ((event_group_hdl & Emti_EVENT_GROUP_ID_MASK) | Emti_NO_EVENT_GROUP_TAG);

    return;
}
