/*
 * Copyright (c) 2012, Texas Instruments Incorporated - http://www.ti.com/
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions are met:
 *       * Redistributions of source code must retain the above copyright
 *         notice, this list of conditions and the following disclaimer.
 *       * Redistributions in binary form must reproduce the above copyright
 *         notice, this list of conditions and the following disclaimer in the
 *         documentation and/or other materials provided with the distribution.
 *       * Neither the name of Texas Instruments Incorporated nor the
 *         names of its contributors may be used to endorse or promote products
 *         derived from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *   DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
 *   DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


/**
 *   @file  event_machine_private_functions.h
 *
 *   @brief
 *      Internal private functions used by the Open EM implementation.
 *      These functions are not available for use by the application.
 *
 */

#ifndef EVENT_MACHINE_PRIVATE_FUNCTIONS_H
#define EVENT_MACHINE_PRIVATE_FUNCTIONS_H

/**********************************************************************
 ************************* Inlined functions  **************************
 **********************************************************************/

 /**
 * @internal
 *
 * The API is used to retreive the first word address of the event descriptor
 */
static inline uint32_t
Emti_getEventDescStartAdd (
    uint32_t add
)
{
    uint32_t lvWordsToGoToPsData = 0;
    uint32_t lvWordsToGoBack = 2;

    /*
     * return current ccpi host descriptor start address minus 2 32-bit words 
     * to get openem descriptor start address 
     */
    return (add + (lvWordsToGoToPsData << 2) - (lvWordsToGoBack << 2));
}

/**
 * @internal
 *
 * return cppi host descriptor pointer from event handle
 *
 * @return event descriptor pointer
 *
 */
static inline Cppi_HostDesc *
Emti_getCppiHostDescPtr (
    em_event_t eventHandle
)
{
    return ((Cppi_HostDesc *) (eventHandle & 0xfffffff0));
}

/**
 * @internal
 *
 * Generate all functions for accessing event descriptor fields
 *
 */
#define EMTI_DECLARE_SET_FN(__name, __define, __type) static inline void Emti_set##__name(Cppi_HostDesc *descPtr, __type var) \
    {                                    \
    uint32_t *lvPsDataPtr = (uint32_t *) (Emti_getEventDescStartAdd ((uint32_t) (descPtr))); \
    lvPsDataPtr[(Emti_##__define##_WORD)] = (lvPsDataPtr[(Emti_##__define##_WORD)] \
                         & ~((Emti_##__define##_MASK) << (Emti_##__define##_SHIFT))) \
        | ((var & (Emti_##__define##_MASK)) << (Emti_##__define##_SHIFT)); \
    }

#define EMTI_DECLARE_GET_FN(__name, __define, __type) static inline __type Emti_get##__name(Cppi_HostDesc *descPtr) \
    {                                    \
    uint32_t *lvPsDataPtr = (uint32_t *) (Emti_getEventDescStartAdd ((uint32_t) (descPtr))); \
    return (__type) ((lvPsDataPtr[(Emti_##__define##_WORD)] >> Emti_##__define##_SHIFT) & Emti_##__define##_MASK); \
    }

EMTI_DECLARE_GET_FN (CohMode, COH_MODE, ti_em_coh_mode_t) EMTI_DECLARE_SET_FN (CohMode, COH_MODE,
                                                                               ti_em_coh_mode_t)
EMTI_DECLARE_GET_FN (DirtyFlag, DIRTY_FLAG, uint32_t) EMTI_DECLARE_SET_FN (DirtyFlag, DIRTY_FLAG,
                                                                           uint32_t) EMTI_DECLARE_GET_FN (PacketWsize,
                                                                                                          PACKET_WSIZE,
                                                                                                          size_t)
EMTI_DECLARE_SET_FN (PacketWsize, PACKET_WSIZE, size_t) EMTI_DECLARE_GET_FN (PoolIndex, POOL_INDEX,
                                                                             uint32_t) EMTI_DECLARE_SET_FN (PoolIndex,
                                                                                                            POOL_INDEX,
                                                                                                            uint32_t)
EMTI_DECLARE_GET_FN (FreeInfo, FREE_INFO, uint32_t) EMTI_DECLARE_SET_FN (FreeInfo, FREE_INFO,
                                                                         uint32_t) EMTI_DECLARE_GET_FN (FreeQueueIndex,
                                                                                                        FREE_QUEUE_INDEX,
                                                                                                        uint32_t)
EMTI_DECLARE_SET_FN (FreeQueueIndex, FREE_QUEUE_INDEX, uint32_t) EMTI_DECLARE_GET_FN (FreePushPolicy, FREE_PUSH_POLICY,
                                                                                      uint32_t)
EMTI_DECLARE_SET_FN (FreePushPolicy, FREE_PUSH_POLICY, uint32_t) EMTI_DECLARE_GET_FN (FreePolicy, FREE_POLICY,
                                                                                      uint32_t)
EMTI_DECLARE_SET_FN (FreePolicy, FREE_POLICY, uint32_t) EMTI_DECLARE_GET_FN (BufMode, BUF_MODE,
                                                                             ti_em_buf_mode_t)
EMTI_DECLARE_SET_FN (BufMode, BUF_MODE, ti_em_buf_mode_t) EMTI_DECLARE_GET_FN (EventMode, EVENT_MODE,
                                                                               ti_em_buf_mode_t)
EMTI_DECLARE_SET_FN (EventMode, EVENT_MODE, ti_em_buf_mode_t)
#ifdef TI_EM_LINUX
#include <ti/runtime/openem/linux/event_machine_private_functions_linux.h>
#else /* !TI_EM_LINUX */
#include <ti/runtime/openem/dsp/event_machine_private_functions_dsp.h>
#endif /* !TI_EM_LINUX */
 /**
 * @internal
 *
 * The API is used to set the free function in the event descriptor (word 10)
 *
 * @param packet_ptr event descriptor pointer
 *
 * @param free free function
 *
 */
     static inline void Emti_packet_set_free_function (
    ti_em_packet_t * packet_ptr,
    ti_em_free_func_t free
)
{
    Cppi_HostDesc *lvHostDescPtr = (Cppi_HostDesc *) packet_ptr;
    if (free != NULL)
    {
        Emti_setEventMode (lvHostDescPtr, Emti_EVENT_MODE_GLOBAL_SW);
    }
    /*
     * Write in the word #14 
     */
    lvHostDescPtr->softwareInfo1 = (uint32_t) free;

    return;
}

 /**
 * @internal
 *
 * The API is used to get the free function from the event descriptor (word 10)
 *
 * @param packet_ptr event descriptor pointer
 * 
 * @return ti_em_free_func_t
 *
 */
static inline ti_em_free_func_t
Emti_packet_get_free_function (
    ti_em_packet_t * packet_ptr
)
{
    Cppi_HostDesc *lvHostDescPtr = (Cppi_HostDesc *) packet_ptr;
    return (ti_em_free_func_t) (lvHostDescPtr->softwareInfo1);
}


/**
 * @internal
 *
 * initialize buffer config structure
 *
 * @param eventDescAdd event descriptor
 *
 * @return ti_em_coh_mode_t
 *
 */

static inline void
Emti_setBufferConfigFromEventDsc (
    Cppi_HostDesc * descPtr,
    ti_em_buffer_config_t * bufferconfigPtr
)
{
    bufferconfigPtr->buffer_ptr = Emti_bufferPtr_ptov (descPtr);
    bufferconfigPtr->buffer_size = descPtr->buffLen;
    bufferconfigPtr->orig_buffer_ptr = Emti_origBufferPtr_ptov (descPtr);
    bufferconfigPtr->orig_buffer_size = descPtr->origBufferLen & 0x3fffff;
    bufferconfigPtr->orig_buffer_pool = (descPtr->origBufferLen >> 28) & 0xf;
    bufferconfigPtr->coh_mode = Emti_getCohMode (descPtr);
    bufferconfigPtr->free_func = Emti_packet_get_free_function ((ti_em_packet_t *) (descPtr));
    return;
}

/**
 * @internal
 *
 * return event event mode from event handle
 *
 * @param eventHandle event handle
 *
 * @return Emti_event_mode_t
 *
 */

static inline Emti_event_mode_t
Emti_getEventModeFromEventHdl (
    em_event_t eventHandle
)
{
    return ((Emti_event_mode_t) (eventHandle & Emti_EVENT_HDL_EVENT_MODE_MASK));
}

/**
 * @internal
 *
 * generic function reporting errors.
 *
 * Reported errors are handled by the appropriate (EO specific or global) error handler or returned back to
 * the callee function.
 *
 * @param error    Error code
 * @param escope   Error scope. Identifies the scope for interpreting the error code and variable arguments.
 * @param ...      Variable number and type of arguments
 *
 */
static inline em_status_t
Emti_error (
    em_status_t error,
    em_escope_t escope,
    ...
)
{
    Emti_EO_s *lvEoPtr = (Emti_EO_s *) (Emti_gvEoHandle);

    /*
     * call eo error function pointer 
     */
    if ((lvEoPtr != NULL) && (lvEoPtr->errorHandler != NULL))
    {
        return (lvEoPtr->errorHandler ((em_eo_t) (lvEoPtr), (em_status_t) (error), (em_escope_t) (escope), NULL));
    }
    /*
     * call global error function pointer 
     */
    else if (Emti_svErrorFunctionPtr != NULL)
    {
        return (Emti_svErrorFunctionPtr (EM_EO_UNDEF, (em_status_t) (error), (em_escope_t) (escope), NULL));
    }
    /*
     * return error code 
     */
    else
    {
        return error;
    }
}


/**
 * @internal
 *
 *  The 3 following functions are used to manipulate the 4 lsb of a descriptor
 *
 * @return
 *  Not applicable.
 */

static inline Cppi_HostDesc *
Emti_setTypeToDesc (
    Cppi_HostDesc * descPtr,
    em_queue_type_t type
)
{
    uint32_t lPtr = (((uint32_t) descPtr) & 0xFFFFFFF0) | (((uint32_t) type) & 0x0F);
    return (Cppi_HostDesc *) lPtr;
}

static inline em_queue_type_t
Emti_getTypeFromDesc (
    Cppi_HostDesc * descPtr
)
{
    em_queue_type_t type = (em_queue_type_t) (((uint32_t) descPtr) & 0x0F);
    return type;
}

/**
 * @internal
 *
 *  The following functions are used to manipulate the 2nd word of event and EQ handle
 *
 * @return
 *  Not applicable.
 */

static inline uint16_t
Emti_getQueueIndex (
    em_queue_t eqHandle
)
{
    return ((uint16_t) (eqHandle & Emti_QUEUE_HANDLE_QUEUE_IDX_MAX));
}

static inline uint8_t
Emti_getSdQueueIndex (
    em_queue_t eqHandle
)
{
    uint8_t lvQueueSetIdx = (eqHandle >> Emti_QUEUE_HANDLE_QUEUE_SET_IDX_SHIFT) & (0xff);
    return (uint8_t) (Emti_getField
                      (Emti_svPsbPtr->queueRouteTbl[lvQueueSetIdx].routeInfo,
                       QUEUE_ROUTE_INFO_SD_QUEUE_IDX) + (Emti_svPcb.sdQueueBaseIdx));
}

static inline uint8_t
Emti_isQueueAtomic (
    em_queue_t eqHandle
)
{
    return ((uint8_t) Emti_getField (eqHandle, EVENT_DSC_W8_QUEUE_TYPE));
}

static inline em_queue_t
Emti_setEqHandle (
    uint16_t qIdx,
    uint16_t sdIdx,
    uint8_t isAtomic
)
{
    uint32_t ret;
    /*
     * Only the 3 LS bytes are relevant. 
     */
    ret =
        (qIdx & Emti_QUEUE_HANDLE_QUEUE_IDX_MAX) | ((sdIdx & Emti_QUEUE_HANDLE_SD_QUEUE_IDX_MAX) << 16) |
        ((isAtomic & 0x1) << 15);
    return ((em_queue_t) ret);
}

static inline void
Emti_getData (
    Cppi_HostDesc * descAddr,
    uint8_t ** buffAddr
)
{
    *buffAddr = (uint8_t *) Emti_bufferPtr_ptov (descAddr);
    return;
}

static inline uint32_t *
Emti_getTimeStampAddress (
    Cppi_HostDesc * descAddr
)
{
    return (uint32_t *) & descAddr->timeStamp;
}

static inline uint32_t
Emti_getTagInfo (
    Cppi_HostDesc * descAddr
)
{
    return descAddr->tagInfo;
}

static inline uint32_t
Emti_getTimestamp (
    Cppi_HostDesc * descAddr
)
{
    return descAddr->timeStamp;
}

static inline void
Emti_setTimestamp (
    Cppi_HostDesc * descAddr,
    uint32_t value
)
{
    descAddr->timeStamp = value;
    return;
}

static inline void
Emti_setSoftwareInfo0 (
    Cppi_HostDesc * descAddr,
    uint32_t value
)
{
    /*
     * Does not update packet length 
     */
    descAddr->softwareInfo0 = value;
    return;
}

static inline void
Emti_setSoftwareInfo1 (
    Cppi_HostDesc * descAddr,
    uint32_t value
)
{
    descAddr->softwareInfo1 = value;
    return;
}

static inline void
Emti_setSoftwareInfo2 (
    Cppi_HostDesc * descAddr,
    uint32_t value
)
{
    descAddr->softwareInfo2 = value;
    return;
}

/**
 * @internal
 *
 * The API is used to set and retrieve the original buffer size in descriptor
 */

static inline em_status_t
Emti_setOrigBufferLength (
    Cppi_HostDesc * descPtr,
    size_t orig_buffer_len
)
{
    descPtr->origBufferLen =
        (uint32_t) (((descPtr->origBufferLen) & ~Emti_EVENT_DSC_ORIG_BUFF_LEN_MASK) |
                    ((orig_buffer_len) & Emti_EVENT_DSC_ORIG_BUFF_LEN_MASK));
    return (em_status_t) (EM_OK);
}

static inline size_t
Emti_getOrigBufferLength (
    Cppi_HostDesc * descPtr
)
{
    return (size_t) (descPtr->origBufferLen & Emti_EVENT_DSC_ORIG_BUFF_LEN_MASK);
}

/**
 * @internal
 *
 * Connect with firmware
 *
 */
static inline em_status_t
Emti_connectFirmware (
    ti_em_queue_id_t cmd_queue_idx,
    ti_em_queue_id_t cmd_free_queue_idx,
    ti_em_queue_id_t return_queue_idx,
    uint32_t base_address,
    uint32_t size_to_transfer,
    uint32_t cmd_type,
    uint32_t mapping,
    ti_em_process_id_t process_idx
)
{
    volatile Emti_CdEventDsc_s *cdEventDscPtr = NULL;
    uint32_t svSpcbPhysAddr = 0;
    uint32_t lvErrorCode;

    /*
     * pop off the event from the CD free queue 
     */
    cdEventDscPtr = (volatile Emti_CdEventDsc_s *) Emti_popEvent (cmd_free_queue_idx, TI_EM_MEM_PRIVATE_DESC);

    if (cdEventDscPtr == NULL)
        return EM_ERR_LIB_FAILED;

    /*
     * copy SPCB to memory accessible by PDSP and return its physical address 
     */
    if (base_address != 0)
    {
        Emti_logDbg ("Put PDSP memory (global) for core %d\n", DNUM);
        svSpcbPhysAddr = Emti_pdspPut (mapping, (void *) base_address, size_to_transfer);
    }

    /*
     * fill event descriptor 
     */
    cdEventDscPtr->commandType = (Emti_CdQueueCommandType_e) (cmd_type | (process_idx << 16));
    cdEventDscPtr->commandPtr = svSpcbPhysAddr;
    cdEventDscPtr->errorCode = Emti_CdQueueErrorCode_WAITING_CMD;
    cdEventDscPtr->returnQueue = (uint32_t) return_queue_idx;

    /*
     * push event to the CD queue 
     */
    Emti_mfence ();
    Emti_pushEvent (cmd_queue_idx, (uint32_t *) (((uint32_t) (cdEventDscPtr) & 0xfffffff0) | Emti_CD_COMMAND_TYPE_CT),
                    TI_EM_MEM_PRIVATE_DESC);

    /*
     * waits for the RP event 
     */
    cdEventDscPtr = NULL;
    while (cdEventDscPtr == NULL)
        cdEventDscPtr = (volatile Emti_CdEventDsc_s *) Emti_popEvent (return_queue_idx, TI_EM_MEM_PRIVATE_DESC);

    cdEventDscPtr = (volatile Emti_CdEventDsc_s *) (((uint32_t) cdEventDscPtr & 0xFFFFFFF0));
    /*
     * read the event 
     */
    if ((cdEventDscPtr->errorCode != Emti_CdQueueErrorCode_NO_ERROR) &&
        (cdEventDscPtr->errorCode != Emti_CdQueueErrorCode_ALREADY_EXECUTED))
        return EM_ERR_LIB_FAILED;

    lvErrorCode = (uint32_t) cdEventDscPtr->errorCode;

    /*
     * copy SPCB from PDSP communication memory 
     */
    if (base_address != 0)
    {
        Emti_pdspGet (mapping, (void *) base_address, size_to_transfer);
        Emti_logDbg ("Get PDSP memory (global) for core %d\n", DNUM);
    }

    /*
     * push to the CD free queue 
     */
    cdEventDscPtr->commandType = Emti_CtType_NULL;
    cdEventDscPtr->commandPtr = 0;
    Emti_mfence ();
    Emti_pushEvent (cmd_free_queue_idx, (uint32_t *) cdEventDscPtr, TI_EM_MEM_PRIVATE_DESC);

    if (lvErrorCode == Emti_CdQueueErrorCode_ALREADY_EXECUTED)
        return EM_ERR_NOT_FREE;

    return EM_OK;
}

/**
 * @internal
 *
 * The dispatcher calls this routine when event group is defined.
 *
 */
void Emti_dispatch_event_group (
    em_event_group_t eventGroup
);

/**
 * @internal
 *
 * The function returns the queue offset to add to the TxQueueBaseIdx to get 
 * the Tx queue to use.
 *
 * @return Tx queue offset.
 */
static inline ti_em_queue_id_t
Emti_hash (
    ti_em_queue_id_t queueIdx
)
{
    uint8_t lvQueueSetIdx;
    uint8_t lvQueueIdxInSet;

    /*
     * set queueSetIdx = queueIdx.queueSetIdx
     */
    lvQueueSetIdx = (uint8_t) Emti_getField (queueIdx, QUEUE_HANDLE_QUEUE_SET_IDX);

    /*
     * set queueIdxInSet = queueIdx.queueIdxInSet
     */
    lvQueueIdxInSet = (uint8_t) Emti_getField (queueIdx, QUEUE_HANDLE_QUEUE_IDX_IN_SET);

    /*
     * set key1 = queueSetIdx & Emti_HASH_QUEUE_SET_IDX_MASK
     * set key2 = queueIdxInSet & Emti_HASH_QUEUE_IDX_IN_SET_MASK
     * set key = key1 | key2
     * return key
     */
    return ((lvQueueSetIdx & Emti_HASH_QUEUE_SET_IDX_MASK) | (lvQueueIdxInSet & Emti_HASH_QUEUE_IDX_IN_SET_MASK));
}

#endif /* EVENT_MACHINE_PRIVATE_FUNCTIONS_H */
