/*
 * Copyright (c) 2012, Texas Instruments Incorporated - http://www.ti.com/
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions are met:
 *       * Redistributions of source code must retain the above copyright
 *         notice, this list of conditions and the following disclaimer.
 *       * Redistributions in binary form must reproduce the above copyright
 *         notice, this list of conditions and the following disclaimer in the
 *         documentation and/or other materials provided with the distribution.
 *       * Neither the name of Texas Instruments Incorporated nor the
 *         names of its contributors may be used to endorse or promote products
 *         derived from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *   DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
 *   DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


/**
 *   @file  event_machine_private_variables.c
 *
 *   @brief
 *      Internal private variables used by the Open EM implementation.
 *      These variables are not available for use by the application.
 *      The file defines global variables & their data sections.
 *      These variables are used internally by the Open EM implementation.
 *      The data sections however are required to be configured by the
 *      applications into the appropriate memory sections. 
 *
 */

#include <ti/runtime/openem/event_machine_types.h>
#include <ti/runtime/openem/event_machine_hw_macros.h>
#include <ti/runtime/openem/event_machine_hw_ti_types.h>
#include <ti/runtime/openem/event_machine_sw_config.h>
#include <ti/runtime/openem/src/event_machine_private_types.h>
#include <ti/runtime/openem/modules/event_machine_modules.h>

/**********************************************************************
 ************************* Global Definitions *************************
 **********************************************************************/


/**
 * @brief openem process configuration information:
 * .
 */
#ifndef TI_EM_GCC
#pragma DATA_SECTION (Emti_svPcb, ".tiEmGlobalFast");
#endif
Emti_Pcb_s Emti_svPcb;

 /**
 * @brief openem thread configuration information
 * .
 */
#ifndef TI_EM_GCC
#pragma DATA_SECTION (Emti_gvTcb, ".tiEmLocal");
#endif
Emti_LOCAL_VARIABLE (Emti_Tcb_s, Emti_gvTcb);

/**
 * @brief openem process state information - fast part:
 * .
 */
#ifndef TI_EM_GCC
#pragma DATA_SECTION (Emti_svPsb, ".tiEmGlobalFast");
#endif
Emti_GLOBAL_VARIABLE (Emti_Psb_s, Emti_svPsb);
Emti_Psb_s Emti_svPsb;

/**
 * @brief openem process state information - slow part:
 * .
 */
#ifndef TI_EM_GCC
#pragma DATA_SECTION (Emti_svPsbSlow, ".tiEmGlobalSlow");
#endif
Emti_GLOBAL_VARIABLE (Emti_Psb_slow_s, Emti_svPsbSlow);
Emti_Psb_slow_s Emti_svPsbSlow;

/**
 * @brief openem process state information Pointer - fast part:
 * .
 */
#ifndef TI_EM_GCC
#pragma DATA_SECTION (Emti_svPsbPtr, ".tiEmGlobalFast");
#endif
Emti_Psb_s *Emti_svPsbPtr;

/**
 * @brief openem process state information Pointer - slow part:
 * .
 */
#ifndef TI_EM_GCC
#pragma DATA_SECTION (Emti_svPsbSlowPtr, ".tiEmGlobalSlow");
#endif
Emti_Psb_slow_s *Emti_svPsbSlowPtr;

/**
 * @brief openem thread state information
 * .
 */
#ifndef TI_EM_GCC
#pragma DATA_SECTION (Emti_gvTsb, ".tiEmLocal");
#endif
Emti_LOCAL_VARIABLE (Emti_Tsb_s, Emti_gvTsb);

/**
 * @brief   Execution Object Database:
 * The structures stores all the execution objects which can be present in 
 * the system.
 */
#ifndef TI_EM_GCC
#pragma DATA_SECTION (__Emti_svEoMdb, ".tiEmGlobalSlow");
#pragma DATA_SECTION (Emti_svEoMdb, ".tiEmGlobalSlow");
#endif
Emti_GLOBAL_VARIABLE (Emti_EO_s, __Emti_svEoMdb[EM_MAX_EOS]);
Emti_EO_s *Emti_svEoMdb;

/**
 * @brief    Slow Queue Database:
 *  The structure stores all the queue attributes in slow memory.
 */
#ifndef TI_EM_GCC
#pragma DATA_SECTION (__Emti_svQueueSlowMdb, ".tiEmGlobalSlow");
#pragma DATA_SECTION (Emti_svQueueSlowMdb, ".tiEmGlobalSlow");
#endif
Emti_GLOBAL_VARIABLE (Emti_QueueSlow_s, __Emti_svQueueSlowMdb[EM_MAX_QUEUES]);
Emti_QueueSlow_s *Emti_svQueueSlowMdb;

/**
 * @brief    Fast Queue Database:
 *  The structure stores all the queue attributes in fast memory.
 */
#ifndef TI_EM_GCC
#pragma DATA_SECTION (__Emti_svQueueFastMdb, ".tiEmGlobalFast");
#pragma DATA_SECTION (Emti_svQueueFastMdb, ".tiEmGlobalFast");
#endif
Emti_GLOBAL_VARIABLE (Emti_QueueFast_s, __Emti_svQueueFastMdb[EM_MAX_QUEUES]);
Emti_QueueFast_s *Emti_svQueueFastMdb;

/**
 * @brief   Event Group Database:
 *  The structure stores all the Event Groups which can be present in the
 *  system.
 */
#ifndef TI_EM_GCC
#pragma DATA_SECTION (__Emti_svEventGroupMdb, ".tiEmGlobalFast");
#pragma DATA_SECTION (Emti_svEventGroupMdb, ".tiEmGlobalFast");
#endif
Emti_GLOBAL_VARIABLE (Emti_EventGroup_s, __Emti_svEventGroupMdb[EM_MAX_EVENT_GROUPS]);
Emti_EventGroup_s *Emti_svEventGroupMdb;

/**
 * @brief   Queue Group Database:
 *  The structure stores all the Queue Groups which can be present in the
 *  system.
 */
#ifndef TI_EM_GCC
#pragma DATA_SECTION (__Emti_svQueueGroupMdb, ".tiEmGlobalSlow");
#pragma DATA_SECTION (Emti_svQueueGroupMdb, ".tiEmGlobalSlow");
#endif
Emti_GLOBAL_VARIABLE (Emti_QueueGroup_s, __Emti_svQueueGroupMdb[EM_MAX_QUEUE_GROUPS]);
Emti_QueueGroup_s *Emti_svQueueGroupMdb;

/** 
 * @brief Global variable which indicates the EO handle (EO identifier) when processing
 * the receive function. When not in a receive function, it is set to EO_UNDEF. This is 
 * maintained on a per core basis. This does not need to be exported to an application. 
 */
#ifndef TI_EM_GCC
#pragma DATA_SECTION (Emti_gvEoHandle, ".tiEmLocal");
#endif
Emti_LOCAL_VARIABLE (em_eo_t, Emti_gvEoHandle);

/** 
 * @brief Global shared variable which contains the global error function pointer.
 * This does not need to be exported to an application. */
#ifndef TI_EM_GCC
#pragma DATA_SECTION (Emti_svErrorFunctionPtr, ".tiEmGlobalFast");
#endif
em_error_handler_t Emti_svErrorFunctionPtr;

/** 
 * @brief Global shared variable which contains the global trace function pointer.
 * This does not need to be exported to an application. */
#ifndef TI_EM_GCC
#pragma DATA_SECTION (Emti_svTraceFunctionPtr, ".tiEmGlobalFast");
#endif
ti_em_trace_handler_t Emti_svTraceFunctionPtr;

/** 
 * @brief Global shared array which contains information of each chaining modules
 * This does not need to be exported to an application. */
#ifndef TI_EM_GCC
#pragma DATA_SECTION (Emti_moduleChainInfo, ".tiEmGlobalFast");
#endif
Emti_ModuleChainInfo_s Emti_moduleChainInfo[Emti_CHAIN_MODULE_NUM];
