/*
 * Copyright (c) 2012, Texas Instruments Incorporated - http://www.ti.com/
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions are met:
 *       * Redistributions of source code must retain the above copyright
 *         notice, this list of conditions and the following disclaimer.
 *       * Redistributions in binary form must reproduce the above copyright
 *         notice, this list of conditions and the following disclaimer in the
 *         documentation and/or other materials provided with the distribution.
 *       * Neither the name of Texas Instruments Incorporated nor the
 *         names of its contributors may be used to endorse or promote products
 *         derived from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *   DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
 *   DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


/**
 *   @file  event_machine_private_variables.h
 *
 *   @brief
 *      Internal private variables used by the Open EM implementation.
 *      These variables are not available for use by the application.
 *
 */

#ifndef EVENT_MACHINE_PRIVATE_VARIABLES_H_
#define EVENT_MACHINE_PRIVATE_VARIABLES_H_

#include <ti/runtime/openem/event_machine_sw_config.h>
#include <ti/runtime/openem/event_machine_hw_ti_types.h>


/**
 * @brief openem process configuration information:
 * .
 */
extern Emti_Pcb_s Emti_svPcb;

 /**
 * @brief openem thread configuration information
 * .
 */
extern Emti_LOCAL_VARIABLE (
    Emti_Tcb_s,
    Emti_gvTcb
);

/**
 * @brief openem process state information - fast part
 * .
 */
extern Emti_Psb_s Emti_svPsb;

/**
 * @brief openem process state information - slow part
 * .
 */
extern Emti_Psb_slow_s Emti_svPsbSlow;

/**
 * @brief openem process state information pointer 
 * .
 */
extern Emti_Psb_s *Emti_svPsbPtr;

/**
 * @brief openem process state information pointer 
 * .
 */
extern Emti_Psb_slow_s *Emti_svPsbSlowPtr;

/**
 * @brief openem thread state information
 * .
 */
extern Emti_LOCAL_VARIABLE (
    Emti_Tsb_s,
    Emti_gvTsb
);

/**
 * @internal
 *
 * @brief   Execution Object Database:
 *  This stores information for all the Execution Objects present in the system
 *  and it should be accessible by all the cores.
 */
extern Emti_EO_s __Emti_svEoMdb[EM_MAX_EOS];
extern Emti_EO_s *Emti_svEoMdb;

/**
 * @internal
 *
 * @brief    Queue Database:
 *  This stores information for all the Event Queues present in slow memory in the system
 *  and it should be accessible by all the cores.
 */
extern Emti_QueueSlow_s __Emti_svQueueSlowMdb[EM_MAX_QUEUES];
extern Emti_QueueSlow_s *Emti_svQueueSlowMdb;

/**
 * @internal
 *
 * @brief    Queue Database:
 *  This stores information for all the Event Queues present in fast memory in the system
 *  and it should be accessible by all the cores.
 */
extern Emti_QueueFast_s __Emti_svQueueFastMdb[EM_MAX_QUEUES];
extern Emti_QueueFast_s *Emti_svQueueFastMdb;

/**
 * @internal
 *
 * @brief   Event Group Database:
 *  This stores information for all the Event Groups present in the system
 *  and it should be accessible by all the cores. 
 */
extern Emti_EventGroup_s __Emti_svEventGroupMdb[EM_MAX_EVENT_GROUPS];
extern Emti_EventGroup_s *Emti_svEventGroupMdb;

/**
 * @internal
 *
 * @brief   Queue Group Database:
 *  This stores information for all the Queue Groups present in the system
 *  and it should be accessible by all the cores.
 */
extern Emti_QueueGroup_s __Emti_svQueueGroupMdb[EM_MAX_QUEUE_GROUPS];
extern Emti_QueueGroup_s *Emti_svQueueGroupMdb;

/** 
 * @internal
 *
 * @brief Global variable which indicates the EO handle (EO identifier) when processing
 * the receive function. When not in a receive function, it is set to EO_UNDEF. This is 
 * maintained on a per core basis. This does not need to be exported to an application. 
 */
extern Emti_LOCAL_VARIABLE (
    em_eo_t,
    Emti_gvEoHandle
);

/** 
 * @internal
 *
 * @brief Global shared variable which contains the global error function pointer.
 * This does not need to be exported to an application. */
extern em_error_handler_t Emti_svErrorFunctionPtr;

/** 
 * @internal
 *
 * @brief Global shared variable which contains the global trace function pointer.
 * This does not need to be exported to an application. */
extern ti_em_trace_handler_t Emti_svTraceFunctionPtr;

/* This define the size of all R/W shared structures */
#define Emti_SV_RW_SECTION_LENGTH  (sizeof(Emti_svPsb) +                \
                                    sizeof(Emti_svPsbSlow) +            \
                                    sizeof(Emti_svPcb) +                \
                                    sizeof(__Emti_svEoMdb) +            \
                                    sizeof(__Emti_svQueueFastMdb) +     \
                                    sizeof(__Emti_svQueueSlowMdb) +     \
                                    sizeof(__Emti_svEventGroupMdb) +    \
                                    sizeof(__Emti_svQueueGroupMdb))

/** 
 * @internal
 *
 * @brief Used to access per-core structure when __thread is not available
 * in thread model
 */
#if defined(TI_EM_THREAD_MODEL) && !defined(TI_EM_HAS___THREAD)
#define Emti_gvTcb       __Emti_gvTcbArray[DNUM]
#define Emti_gvEoHandle  __Emti_gvEoHandleArray[DNUM]
#define Emti_gvDspMcb    __Emti_gvDspMcbArray[DNUM]
#define Emti_gvTsb       __Emti_gvTsbArray[DNUM]
#endif /* defined(TI_EM_THREAD_MODEL) && !defined(TI_EM_HAS___THREAD) */

#endif /*EVENT_MACHINE_PRIVATE_VARIABLES_H_ */
