/*
 * Copyright (c) 2012, Texas Instruments Incorporated - http://www.ti.com/
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions are met:
 *       * Redistributions of source code must retain the above copyright
 *         notice, this list of conditions and the following disclaimer.
 *       * Redistributions in binary form must reproduce the above copyright
 *         notice, this list of conditions and the following disclaimer in the
 *         documentation and/or other materials provided with the distribution.
 *       * Neither the name of Texas Instruments Incorporated nor the
 *         names of its contributors may be used to endorse or promote products
 *         derived from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *   DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
 *   DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


/**
 *   @file  event_machine_queue.c
 *
 *   @brief
 *      Open Event Machine Implementation for Keystone1 devices.
 *      This file contains the queue related functions
 *
 */

#include <inttypes.h>
#include <string.h>
#include <ti/runtime/openem/event_machine_macros.h>
#include <ti/runtime/openem/event_machine_types.h>
#include <ti/runtime/openem/event_machine_core_mask.h>
#include <ti/runtime/openem/event_machine_hw_types.h>
#include <ti/runtime/openem/src/event_machine_private_types.h>
#include <ti/runtime/openem/src/event_machine_private_variables.h>
#include <ti/runtime/openem/src/event_machine_private_functions.h>

/*
 **********************************************************************************************************************
 **********************************************************************************************************************
 */

static inline uint32_t
Emti_issetCoreMaskBit (
    uint32_t * arrayPtr,
    uint32_t prio,
    uint32_t coreId
)
{
    uint32_t *ptr = arrayPtr;
    uint32_t mask = (1 << prio);
    ptr += coreId;
    return (*ptr & mask);
}

static inline void
Emti_setCoreMaskBit (
    uint32_t * arrayPtr,
    uint32_t prio,
    uint32_t coreId
)
{
    uint32_t *ptr = arrayPtr;
    uint32_t mask = (1 << prio);
    ptr += coreId;
    *ptr |= mask;
    return;
}

static em_status_t
Emti_queueGroupCheck (
    Emti_QueueGroup_s * queueGroupPtr,
    int32_t * sdQueueOffsetPtr,
    uint8_t * sdQueuePendingPtr,
    em_queue_prio_t prio
)
{
    em_core_mask_t lvLocalMask;
    uint32_t lvIndex;
    int32_t lvSdQueueOffset;
    uint8_t sdQueuePending = *sdQueuePendingPtr;

    if (queueGroupPtr->state == Emti_QueueGroupState_FREE)
    {
        return (em_status_t) EM_ERR_BAD_STATE;
    }

    /*
     * if <group>sdQueueOfsTbl[priority] == -1 
     */
    if (queueGroupPtr->sdQueueTbl[prio] == -1)
    {

        /*
         * if <group>coreMask overlaps with <PSB>coreMaskTbl[priority]
         */
        {
            /*
             * copy the mask from the queue group 
             */
            em_core_mask_copy (&lvLocalMask, &queueGroupPtr->coreMskTbl);

            /*
             * Check for no overlapping core masks for a given priority level 
             */
            for (lvIndex = 0; lvIndex < TI_EM_CORE_NUM; lvIndex++)
            {
                if (em_core_mask_isset (lvIndex, &lvLocalMask))
                {
                    if (Emti_issetCoreMaskBit (Emti_svPsbPtr->coreMskTbl, prio, lvIndex))
                    {
                        return (em_status_t) EM_ERR_BAD_CONTEXT;
                    }
                    else
                        Emti_setCoreMaskBit (Emti_svPsbPtr->coreMskTbl, prio, lvIndex);
                }
            }
        }

        /*
         * set sdQueueOfs = <PSB>nextFreeSdQueueOfsTbl[priority]        
         * if sdQueueOfs == -1
         */
        lvSdQueueOffset = (int32_t) Emti_svPsbPtr->nextFreeSdQueueIdxTbl[prio];
        /*
         * No more available SD queue for that priority level 
         */
        if (lvSdQueueOffset == -1)
        {
            return (em_status_t) EM_ERR_ALLOC_FAILED;
        }
        sdQueuePending = 1;

        /*
         * for each core c selected in <group>coreMask
         * set <PSB>sdQueueMask[c][priority] = tmpSdQueueMaskTbl[c] 
         */
        {
            uint32_t lvLineSdQueueOffset = (uint32_t) lvSdQueueOffset;
            uint32_t *lvPtr;

            lvPtr = (uint32_t *) (&(Emti_svPsbPtr->sdQueueMaskTbl[0][lvLineSdQueueOffset >> 5]));
            lvLineSdQueueOffset = (1 << (lvLineSdQueueOffset & 0x1F));

            for (lvIndex = 0; lvIndex < TI_EM_CORE_NUM; lvIndex++)
            {
                if (em_core_mask_isset (lvIndex, &lvLocalMask))
                {
                    *lvPtr |= lvLineSdQueueOffset;
                }
                lvPtr += 4;
            }
        }

        /*
         * update shadow structure 
         */
        for (lvIndex = 0; lvIndex < TI_EM_CORE_NUM; lvIndex++)
        {
            Emti_svPsbPtr->shadowSDsc.sdQueueMaskTbl[lvIndex][0] = Emti_svPsbPtr->sdQueueMaskTbl[lvIndex][0];
            Emti_svPsbPtr->shadowSDsc.sdQueueMaskTbl[lvIndex][1] = Emti_svPsbPtr->sdQueueMaskTbl[lvIndex][1];
            Emti_svPsbPtr->shadowSDsc.sdQueueMaskTbl[lvIndex][2] = Emti_svPsbPtr->sdQueueMaskTbl[lvIndex][2];
            Emti_svPsbPtr->shadowSDsc.sdQueueMaskTbl[lvIndex][3] = Emti_svPsbPtr->sdQueueMaskTbl[lvIndex][3];
        }

        /*
         * Communicate to Scheduler PDSP update to COREMASK
         * if needed
         */
        if (Emti_svPcb.pdspNum > 0)
        {
            /*
             * send global init command to CD queue 
             */
            volatile Emti_CdEventDsc_s *lvCdEventDscPtr = NULL;
            uint32_t svSpsbPhysAddr;
            uint32_t schedIdx = 0;

            /*
             * pop off the event from the CD free queue 
             */
            lvCdEventDscPtr =
                (volatile Emti_CdEventDsc_s *) Emti_popEvent (Emti_svPcb.ctCommandFreeQueueIdx, TI_EM_MEM_PRIVATE_DESC);

            if (lvCdEventDscPtr == NULL)
            {
                return (em_status_t) EM_ERR_BAD_POINTER;
            }
            lvCdEventDscPtr = (volatile Emti_CdEventDsc_s *) (((uint32_t) lvCdEventDscPtr) & 0xFFFFFFF0);

            /*
             * copy SPSB to memory accessible by PDSP and return its physical address 
             */
            svSpsbPhysAddr =
                Emti_pdspPut (TI_EM_SLOT_SPSB, (void *) &Emti_svPsbPtr->shadowSDsc, sizeof (Emti_svPsbPtr->shadowSDsc));

            for (schedIdx = 0; schedIdx < Emti_svPcb.pdspNum; schedIdx++)
            {
                /*
                 * fill event descriptor 
                 */
                lvCdEventDscPtr->commandType = Emti_CtType_SCHEDULER_UPDATE_MASK;
                lvCdEventDscPtr->commandPtr = svSpsbPhysAddr;
                lvCdEventDscPtr->errorCode = Emti_CdQueueErrorCode_WAITING_CMD;
                lvCdEventDscPtr->returnQueue = (uint32_t) Emti_svPcb.rpQueueIdx;

                /*
                 * push event to the CD queue 
                 */
                Emti_mfence ();
                Emti_pushEvent (Emti_svPcb.cdQueueIdx[schedIdx],
                                (uint32_t *) (((uint32_t) (lvCdEventDscPtr) & 0xfffffff0) | Emti_CD_COMMAND_TYPE_CT),
                                TI_EM_MEM_PRIVATE_DESC);

                /*
                 * waits for the RP event 
                 */
                lvCdEventDscPtr = NULL;
                while (lvCdEventDscPtr == NULL)
                    lvCdEventDscPtr =
                        (volatile Emti_CdEventDsc_s *) Emti_popEvent (Emti_svPcb.rpQueueIdx, TI_EM_MEM_PRIVATE_DESC);

                lvCdEventDscPtr = (volatile Emti_CdEventDsc_s *) (((uint32_t) lvCdEventDscPtr) & 0xFFFFFFF0);
                /*
                 * read the event 
                 */
                if (lvCdEventDscPtr->errorCode != Emti_CdQueueErrorCode_NO_ERROR)
                {
                    return (em_status_t) EM_ERR_LIB_FAILED;
                }
            }

            /*
             * Copy SPSB to CPU memory 
             */
            Emti_pdspGet (TI_EM_SLOT_SPSB, (void *) &Emti_svPsbPtr->shadowSDsc, sizeof (Emti_svPsbPtr->shadowSDsc));

            /*
             * push to the CD free queue 
             */
            lvCdEventDscPtr->commandType = Emti_CtType_NULL;
            lvCdEventDscPtr->commandPtr = 0;
            Emti_mfence ();

            Emti_pushEvent (Emti_svPcb.ctCommandFreeQueueIdx, (uint32_t *) lvCdEventDscPtr, TI_EM_MEM_PRIVATE_DESC);
        }

    }
    else
    {
        /*
         * set sdQueueOfs = <group>sdQueueOfsTbl[priority] 
         */
        lvSdQueueOffset = (int32_t) queueGroupPtr->sdQueueTbl[prio];
    }
    *sdQueueOffsetPtr = lvSdQueueOffset;
    *sdQueuePendingPtr = sdQueuePending;
    return (em_status_t) EM_OK;
}

static void
Emti_queuePsbUpdate (
    Emti_QueueGroup_s * queueGroupPtr,
    int32_t sdQueueOffset,
    em_queue_prio_t prio
)
{
    /*
     * Increment nextFreeSdQueueIdxTbl[prio]
     */
    if (((sdQueueOffset + 1) & 0xF) == TI_EM_CORE_NUM)
        Emti_svPsbPtr->nextFreeSdQueueIdxTbl[prio] = -1;
    else
        Emti_svPsbPtr->nextFreeSdQueueIdxTbl[prio] = (int8_t) (sdQueueOffset + 1);

    /*
     * set <group>sdQueueOfsTbl[priority] = sdQueueOfs 
     */
    queueGroupPtr->sdQueueTbl[prio] = (Qmss_QueueHnd) sdQueueOffset;
    return;
}

static void
Emti_queueSlowQueueUpdate (
    Emti_QueueSlow_s * queueSlowPtr,
    const char *name,
    em_queue_type_t type,
    em_queue_prio_t prio,
    em_queue_group_t group
)
{
    /*
     * set <queue>state = Emti_QueueState_ALLOCATED 
     */
    queueSlowPtr->state = Emti_QueueState_ALLOCATED;

    /*
     * set <queue>name = name 
     */
    {
        if (strlen (name) <= EM_QUEUE_NAME_LEN)
            strcpy (queueSlowPtr->name, name);
        else
            strncpy (queueSlowPtr->name, name, EM_QUEUE_NAME_LEN);
    }

    /*
     * set <queue>mode = TI_EM_QUEUE_MODE_SD 
     */
    queueSlowPtr->queueMode = TI_EM_QUEUE_MODE_SD;

    /*
     * set <queue>type = type 
     */
    queueSlowPtr->type = type;

    /*
     * set <queue>priority = priority 
     */
    queueSlowPtr->priority = prio;

    /*
     * set <queue>queueGroupHdl = groupHdl 
     */
    queueSlowPtr->queueGroupPtr = group;

    return;
}

static em_status_t
Emti_updateRoute (
    uint16_t lvObjectIdx,
    Emti_CdQueueCommandType_e commandType
)
{
    /*
     * send global init command to CD queue 
     */
    volatile Emti_CdEventDsc_s *lvCdEventDscPtr = NULL;
    uint32_t svRpsbPhysAddr;

    /*
     * pop off the event from the CD free queue 
     */
    lvCdEventDscPtr =
        (volatile Emti_CdEventDsc_s *) Emti_popEvent (Emti_svPcb.chainDsc.ctCommandFreeQueueIdx,
                                                      TI_EM_MEM_PRIVATE_DESC);

    if (lvCdEventDscPtr == NULL)
        return (em_status_t) EM_QUEUE_UNDEF;

    /*
     * copy RPSB to memory accessible by PDSP and return its physical address 
     */
    svRpsbPhysAddr =
        Emti_pdspPut (TI_EM_SLOT_RPSB, (void *) &Emti_svPsbPtr->shadowRDsc, sizeof (Emti_svPsbPtr->shadowRDsc));

    /*
     * fill event descriptor 
     */
    lvCdEventDscPtr->commandType = commandType;
    lvCdEventDscPtr->commandPtr = svRpsbPhysAddr;
    lvCdEventDscPtr->errorCode = Emti_CdQueueErrorCode_WAITING_CMD;
    lvCdEventDscPtr->returnQueue = (uint32_t) Emti_svPcb.chainDsc.rpQueueIdx;

    /*
     * the 16 MSB of the commandType word are used to store objectIdx
     */
    lvCdEventDscPtr->commandType =
        (Emti_CdQueueCommandType_e) ((((uint32_t) lvObjectIdx) << 16) + lvCdEventDscPtr->commandType);

    /*
     * push event to the CD queue 
     */
    Emti_mfence ();
    Emti_pushEvent (Emti_svPcb.chainDsc.cdQueueIdx,
                    (uint32_t *) (((uint32_t) (lvCdEventDscPtr) & 0xfffffff0) | Emti_CD_COMMAND_TYPE_CT),
                    TI_EM_MEM_PRIVATE_DESC);

    /*
     * waits for the RP event 
     */
    lvCdEventDscPtr = NULL;
    while (lvCdEventDscPtr == NULL)
        lvCdEventDscPtr =
            (volatile Emti_CdEventDsc_s *) Emti_popEvent (Emti_svPcb.chainDsc.rpQueueIdx, TI_EM_MEM_PRIVATE_DESC);

    /*
     * read the event 
     */
    if (lvCdEventDscPtr->errorCode != Emti_CdQueueErrorCode_NO_ERROR)
        return EM_ERR_LIB_FAILED;

    /*
     * copy RPSB to CPU memory 
     */
    Emti_pdspGet (TI_EM_SLOT_RPSB, (void *) &Emti_svPsbPtr->shadowRDsc, sizeof (Emti_svPsbPtr->shadowRDsc));

    /*
     * push to the CD free queue 
     */
    lvCdEventDscPtr->commandType = Emti_CtType_NULL;
    lvCdEventDscPtr->commandPtr = 0;
    Emti_mfence ();

    Emti_pushEvent (Emti_svPcb.chainDsc.ctCommandFreeQueueIdx,
                    (uint32_t *) (((uint32_t) (lvCdEventDscPtr) & 0xfffffff0)), TI_EM_MEM_PRIVATE_DESC);

    return EM_OK;
}

static void
Emti_queueUpdateRouteInfo (
    uint16_t * routeInfoPtr,
    int32_t sdQueueOffset,
    int32_t queueMode,
    em_queue_type_t type
)
{
    uint16_t routeInfo = 0;

    /*
     * set tmpQueueRoute.enabled = 1 
     */
    Emti_setField (routeInfo, QUEUE_ROUTE_INFO_ROUTE_ENABLED, 1);

    /*
     * set tmpQueueRoute.queueMode = queueMode 
     */
    Emti_setField (routeInfo, QUEUE_ROUTE_INFO_QUEUE_MODE, queueMode);

    if (queueMode == TI_EM_QUEUE_MODE_SD)
    {
        /*
         * set tmpQueueRoute.queueType = type 
         */
        Emti_setField (routeInfo, QUEUE_ROUTE_INFO_QUEUE_TYPE, type);

        /*
         * set tmpQueueRoute.queueTag = sdQueueOfs 
         */
        Emti_setField (routeInfo, QUEUE_ROUTE_INFO_SD_QUEUE_IDX, sdQueueOffset);
    }
    else
    {
        /*
         * set tmpQueueRoute.hwQueueIdx = hw_queue_idx
         */
        Emti_setField (routeInfo, QUEUE_ROUTE_INFO_HW_QUEUE_IDX, sdQueueOffset);
    }

    *routeInfoPtr = routeInfo;

    return;
}

static inline em_queue_t
Emti_queueMakeGlobal (
    em_queue_t queue_hdl,
    ti_em_device_id_t device_idx,
    ti_em_process_id_t process_idx
)
{
    em_queue_t lvEqHandle = (em_queue_t) EM_QUEUE_UNDEF;

    /*
     * set handle.deviceIdx = <TCB>myDeviceIdx 
     */
    Emti_setField (lvEqHandle, QUEUE_HANDLE_DEVICE_IDX, device_idx);

    /*
     * set handle.processIdx = <TCB>myProcessIdx 
     */
    Emti_setField (lvEqHandle, QUEUE_HANDLE_PROCESS_IDX, process_idx);

    /*
     * set handle.queueIdx = queueIdx 
     */
    Emti_setField (lvEqHandle, QUEUE_HANDLE_QUEUE_IDX, queue_hdl);

    return (em_queue_t) lvEqHandle;
}

static inline em_status_t
Emti_queueStaticInputCheck (
    em_queue_t queue,
    ti_em_device_id_t * deviceIdxPtr,
    ti_em_process_id_t * processIdxPtr,
    ti_em_queue_id_t * queueIdxPtr
)
{
    ti_em_device_id_t lvDeviceIdx;
    ti_em_process_id_t lvProcessIdx;
    ti_em_queue_id_t lvQueueIdx;

    /*
     * deviceIdx = handle.deviceIdx
     * if deviceIdx != <PCB>myDeviceIdx
     * return error 
     */
    lvDeviceIdx = (ti_em_device_id_t) Emti_getField (queue, QUEUE_HANDLE_DEVICE_IDX);
    if (lvDeviceIdx != Emti_svPcb.myDeviceIdx)
    {
        return (em_status_t) EM_ERR_BAD_ID;
    }

    /*
     * processIdx = handle.processIdx
     * if processIdx != <PCB>myProcessIdx
     * return error 
     */
    lvProcessIdx = (ti_em_process_id_t) Emti_getField (queue, QUEUE_HANDLE_PROCESS_IDX);
    if (lvProcessIdx != Emti_svPcb.myProcessIdx)
    {
        return (em_status_t) EM_ERR_BAD_ID;
    }

    /*
     * queueIdx = handle.queueIdx
     * if queueIdx === EM_QUEUE_UNDEF
     * return error
     * if queueIdx < EM_QUEUE_STATIC_MIN
     * return error 
     * if queueIdx > EM_QUEUE_STATIC_MAX
     * return error 
     */
    lvQueueIdx = (ti_em_queue_id_t) Emti_getField (queue, QUEUE_HANDLE_QUEUE_IDX);
    if (lvQueueIdx == EM_QUEUE_UNDEF)
    {
        return (em_status_t) EM_ERR_BAD_ID;
    }

    if (lvQueueIdx < EM_QUEUE_STATIC_MIN)
    {
        return (em_status_t) EM_ERR_BAD_ID;
    }

    if (lvQueueIdx > EM_QUEUE_STATIC_MAX)
    {
        return (em_status_t) EM_ERR_BAD_ID;
    }
    *deviceIdxPtr = lvDeviceIdx;
    *processIdxPtr = lvProcessIdx;
    *queueIdxPtr = lvQueueIdx;
    return (em_status_t) EM_OK;
}

#if (EM_CHECK_LEVEL >= 1)
static inline em_status_t
Emti_queueParamsCheck (
    em_queue_type_t type,
    em_queue_prio_t prio,
    em_queue_group_t group
)
{
    /*
     * Basic Validations: We currently dont support PARALLEL ORDERED queues. 
     */
    if (type == EM_QUEUE_TYPE_PARALLEL_ORDERED)
    {
        return (em_status_t) EM_ERR;
    }
    /*
     * Basic Validations: Check that the priority level is in the correct range. 
     */
    if ((prio < EM_QUEUE_PRIO_LOWEST) || (prio > EM_QUEUE_PRIO_HIGHEST))
    {
        return (em_status_t) EM_ERR_BAD_ID;
    }
    /*
     * Basic Validations: Check that the group is in the correct range. 
     */
    if (group >= EM_MAX_QUEUE_GROUPS)
    {
        return (em_status_t) EM_ERR_TOO_LARGE;
    }
    return EM_OK;
}
#endif

/*
 **********************************************************************************************************************
 **********************************************************************************************************************
 */

em_queue_t
em_queue_create (
    const char *name,
    em_queue_type_t type,
    em_queue_prio_t prio,
    em_queue_group_t group
)
{
    Emti_QueueSlow_s *lvQueueSlowPtr;
    Emti_QueueGroup_s *lvQueueGroupPtr;
    int32_t lvSdQueueOffset;
    Emti_SdQueueState_s *lvSdQueueStatePtr;
    int8_t lvQueueIdxInSet;
    int8_t lvNextQueueIdxInSet;
    uint8_t sdQueuePending = 0;
    int16_t lvQueueSetIdx;
    int16_t lvNextQueueSetIdx;
    uint16_t lvQueueIdx;
    em_queue_t lvEqHandle;
    em_status_t lvStatus;

#if (EM_CHECK_LEVEL >= 1)
    lvStatus = Emti_queueParamsCheck (type, prio, group);
    if (lvStatus != EM_OK)
    {
        Emti_error ((em_status_t) lvStatus, (em_escope_t) EM_ESCOPE_QUEUE_CREATE);
        return (em_queue_t) EM_QUEUE_UNDEF;
    }
#endif

    /*
     * if TI_EM_STATIC_QUEUE_NUM_IN_SET == TI_EM_QUEUE_NUM_IN_SET
     * return EM_QUEUE_UNDEF 
     */
    if (TI_EM_STATIC_QUEUE_NUM_IN_SET == TI_EM_QUEUE_NUM_IN_SET)
    {
        return (em_queue_t) EM_QUEUE_UNDEF;
    }

    /*
     * Critical Section Enter 
     */
    Emti_lockSemaphore ();

    /*
     * Populate the queue configuration. 
     */
    lvQueueGroupPtr = &Emti_svQueueGroupMdb[group];

    lvStatus = Emti_queueGroupCheck (lvQueueGroupPtr, &lvSdQueueOffset, &sdQueuePending, prio);
    if (lvStatus != EM_OK)
    {
        /*
         * Critical Section Exit 
         */
        Emti_unlockSemaphore ();
        Emti_error (lvStatus, (em_escope_t) EM_ESCOPE_QUEUE_CREATE);
        return (em_queue_t) EM_QUEUE_UNDEF;
    }

    /*
     * set queueTagState = <PSB>queueTagState[sdQueueOfs] 
     */
    lvSdQueueStatePtr = &(Emti_svPsbSlowPtr->sdQueueStateTbl[lvSdQueueOffset]);

    /*
     * if type == EM_QUEUE_TYPE_PARALLEL
     * set queueIdxInSet = queueTagState.nextFreeIdxInParallelSdSet
     * else
     * set queueIdxInSet = queueTagState.nextFreeIdxInAtomicSdSet 
     */
    {
        if (type == EM_QUEUE_TYPE_PARALLEL)
            lvQueueIdxInSet = lvSdQueueStatePtr->nextFreeIdxInParallelSet;
        else
            lvQueueIdxInSet = lvSdQueueStatePtr->nextFreeIdxInAtomicSet;
    }

    /*
     * if queueIdxInSet == -1 
     */
    if (lvQueueIdxInSet == -1)
    {
        /*
         * set queueSetIdx = <PSB>nextFreeQueueSetIdx
         * if queueSetIdx == -1
         * return EM_QUEUE_UNDEF 
         */
        {
            lvQueueSetIdx = Emti_svPsbPtr->nextFreeQueueSetIdx;
            if (lvQueueSetIdx == -1)
            {
                /*
                 * Critical Section Exit 
                 */
                Emti_unlockSemaphore ();
                return (em_queue_t) EM_QUEUE_UNDEF;
            }
        }

        /*
         * set queueIdxInSet = 0 
         */
        lvQueueIdxInSet = TI_EM_STATIC_QUEUE_NUM_IN_SET;

        Emti_queueUpdateRouteInfo (&(Emti_svPsbPtr->shadowRDsc.queueRouteTbl[lvQueueSetIdx].routeInfo), lvSdQueueOffset,
                                   TI_EM_QUEUE_MODE_SD, type);

        /*
         * send ROUTE_UPDATE command to router
         */
        if (Emti_svPcb.chainEnabled == TI_EM_CHAIN_ENABLED)
        {
            em_status_t lvRetStatus;

            lvRetStatus = Emti_updateRoute (lvQueueSetIdx, Emti_CtType_ROUTER_UPDATE_ROUTE);
            if (lvRetStatus != EM_OK)
            {
                Emti_unlockSemaphore ();
                return lvRetStatus;
            }
        }

        /*
         * set <PSB>queueRouteTbl[queueSetIdx] = tmpQueueRoute 
         */
        {
            Emti_svPsbPtr->queueRouteTbl[lvQueueSetIdx].routeInfo =
                Emti_svPsbPtr->shadowRDsc.queueRouteTbl[lvQueueSetIdx].routeInfo;
        }

        /*
         * set nextQueueSetIdx = queueSetIdx + 1
         * if nextQueueSetIdx == TI_EM_QUEUE_SET_NUM
         * set nextQueueSetIdx = -1 
         */
        {
            lvNextQueueSetIdx = lvQueueSetIdx + 1;
            if (lvNextQueueSetIdx == TI_EM_QUEUE_SET_NUM)
                lvNextQueueSetIdx = -1;
        }

        /*
         * if type == EM_QUEUE_TYPE_PARALLEL
         * set queueTagState.parallelSdSetIdx = queueSetIdx
         * else
         * set queueTagState.atomicSdSetIdx = queueSetIdx 
         */
        {
            if (type == EM_QUEUE_TYPE_PARALLEL)
                lvSdQueueStatePtr->parallelSetIdx = lvQueueSetIdx;
            else
                lvSdQueueStatePtr->atomicSetIdx = lvQueueSetIdx;
        }

        /*
         * set <PSB>nextFreeQueueSetIdx = nextQueueSetIdx 
         */
        Emti_svPsbPtr->nextFreeQueueSetIdx = lvNextQueueSetIdx;
    }
    else
    {
        if (type == EM_QUEUE_TYPE_PARALLEL)
            lvQueueSetIdx = lvSdQueueStatePtr->parallelSetIdx;
        else
            lvQueueSetIdx = lvSdQueueStatePtr->atomicSetIdx;
    }

    /*
     * set nextQueueIdxInSet = queueIdxInSet + 1
     * if nextQueueIdxInSet == TI_EM_QUEUE_NUM_IN_SET
     * set nextQueueIdxInSet = -1 
     */
    {
        lvNextQueueIdxInSet = lvQueueIdxInSet + 1;
        if (lvNextQueueIdxInSet == TI_EM_QUEUE_NUM_IN_SET)
            lvNextQueueIdxInSet = -1;
    }

    if (sdQueuePending == 1)
    {
        Emti_queuePsbUpdate (lvQueueGroupPtr, lvSdQueueOffset, prio);
    }

    /*
     * if type == EM_QUEUE_TYPE_PARALLEL
     * set queueTagState.nextFreeIdxInParallelSdSet = nextQueueIdxInSet
     * else
     * set queueTagState.nextFreeIdxInAtomicSdSet = nextQueueIdxInSet 
     */
    {
        if (type == EM_QUEUE_TYPE_PARALLEL)
            lvSdQueueStatePtr->nextFreeIdxInParallelSet = lvNextQueueIdxInSet;
        else
            lvSdQueueStatePtr->nextFreeIdxInAtomicSet = lvNextQueueIdxInSet;
    }

    /*
     * set queueIdx = queueSetIdx*TI_EM_QUEUE_NUM_IN_SET + queueIdxInSet 
     */
    lvQueueIdx = (lvQueueIdxInSet * TI_EM_QUEUE_SET_NUM) + lvQueueSetIdx;
    if (lvQueueIdx >= EM_MAX_QUEUES)
    {
        /*
         * Critical Section Exit 
         */
        Emti_unlockSemaphore ();

        return (em_queue_t) EM_QUEUE_UNDEF;
    }
    lvQueueSlowPtr = &Emti_svQueueSlowMdb[lvQueueIdx];

    Emti_queueSlowQueueUpdate (lvQueueSlowPtr, name, type, prio, group);

    lvEqHandle =
        (em_queue_t) Emti_queueMakeGlobal ((em_queue_t) lvQueueIdx, Emti_gvTcb.myDeviceIdx, Emti_gvTcb.myProcessIdx);

    /*
     * Critical Section Exit 
     */
    Emti_unlockSemaphore ();

    /*
     * Debug Message: 
     */
    Emti_logDbg (" Queue maps to HW Queue 0x%x\n", lvSdQueueOffset);

    /*
     * return handle 
     */
    return (em_queue_t) lvEqHandle;
}

em_status_t
em_queue_create_static (
    const char *name,
    em_queue_type_t type,
    em_queue_prio_t prio,
    em_queue_group_t group,
    em_queue_t queue
)
{
    Emti_QueueSlow_s *lvQueueSlowPtr;
    Emti_QueueGroup_s *lvQueueGroupPtr;
    ti_em_device_id_t lvDeviceIdx;
    ti_em_process_id_t lvProcessIdx;
    ti_em_queue_id_t lvQueueIdx;
    int32_t lvSdQueueOffset;
    uint8_t sdQueuePending = 0;
    em_status_t lvStatus;

#if (EM_CHECK_LEVEL >= 1)
    lvStatus = Emti_queueParamsCheck (type, prio, group);
    if (lvStatus != EM_OK)
    {
        Emti_error ((em_status_t) lvStatus, (em_escope_t) EM_ESCOPE_QUEUE_CREATE_STATIC);
        return (em_status_t) lvStatus;
    }
#endif

    lvStatus = Emti_queueStaticInputCheck (queue, &lvDeviceIdx, &lvProcessIdx, &lvQueueIdx);
    if (lvStatus != EM_OK)
    {
        Emti_error (lvStatus, (em_escope_t) EM_ESCOPE_QUEUE_CREATE_STATIC);
        return lvStatus;
    }

    /*
     * if <queue>state == Emti_QueueState_ALLOCATED
     * return error 
     */
    lvQueueSlowPtr = &Emti_svQueueSlowMdb[lvQueueIdx];
    if (lvQueueSlowPtr->state == Emti_QueueState_ALLOCATED)
    {
        Emti_error ((em_status_t) EM_ERR_BAD_STATE, (em_escope_t) EM_ESCOPE_QUEUE_CREATE_STATIC);
        return (em_status_t) EM_ERR_BAD_STATE;
    }

    /*
     * Critical Section Enter 
     */
    Emti_lockSemaphore ();

    /*
     * Populate the queue configuration. 
     */
    lvQueueGroupPtr = &Emti_svQueueGroupMdb[group];

    lvStatus = Emti_queueGroupCheck (lvQueueGroupPtr, &lvSdQueueOffset, &sdQueuePending, prio);
    if (lvStatus != EM_OK)
    {
        /*
         * Critical Section Exit 
         */
        Emti_unlockSemaphore ();
        Emti_error (lvStatus, (em_escope_t) EM_ESCOPE_QUEUE_CREATE_STATIC);
        return lvStatus;
    }

    Emti_queueUpdateRouteInfo (&(Emti_svPsbPtr->shadowRDsc.staticQueueRouteTbl[lvQueueIdx].routeInfo), lvSdQueueOffset,
                               TI_EM_QUEUE_MODE_SD, type);

    /*
     * send ROUTE_UPDATE_STATIC command to router
     */
    if (Emti_svPcb.chainEnabled == TI_EM_CHAIN_ENABLED)
    {
        em_status_t lvRetStatus;

        lvRetStatus = Emti_updateRoute (lvQueueIdx, Emti_CtType_ROUTER_UPDATE_ROUTE_STATIC);
        if (lvRetStatus != EM_OK)
        {
            Emti_unlockSemaphore ();
            return lvRetStatus;
        }
    }

    if (sdQueuePending == 1)
    {
        Emti_queuePsbUpdate (lvQueueGroupPtr, lvSdQueueOffset, prio);
    }

    /*
     * set <PSB>staticQueueRouteTbl[queueIdx] = tmpStaticQueueRoute 
     */
    {
        Emti_svPsbPtr->staticQueueRouteTbl[lvQueueIdx].routeInfo =
            Emti_svPsbPtr->shadowRDsc.staticQueueRouteTbl[lvQueueIdx].routeInfo;
    }

    Emti_queueSlowQueueUpdate (lvQueueSlowPtr, name, type, prio, group);

    /*
     * Critical Section Exit 
     */
    Emti_unlockSemaphore ();

    /*
     * Debug Message: 
     */
    Emti_logDbg (" Queue maps to HW Queue 0x%x\n", lvSdQueueOffset);

    return (em_status_t) (EM_OK);
}

em_queue_t
ti_em_queue_create_hw (
    const char *name,
    ti_em_queue_id_t hw_queue_idx
)
{
    Emti_QueueSlow_s *lvQueueSlowPtr;
    int32_t lvHwQueueBaseIdx;
    Emti_HwQueueState_s lvHwQueueState;
    int8_t lvQueueSetIdx;
    int8_t lvNextQueueSetIdx;
    uint32_t lvQueueSetPending;
    uint8_t lvQueueIdxInSet;
    ti_em_queue_id_t lvQueueIdx;
    em_queue_t lvEqHandle;
    em_status_t lvRetStatus;

    /*
     * if TI_EM_STATIC_QUEUE_NUM_IN_SET == TI_EM_QUEUE_NUM_IN_SET
     * return EM_QUEUE_UNDEF 
     */
    if (TI_EM_STATIC_QUEUE_NUM_IN_SET == TI_EM_QUEUE_NUM_IN_SET)
    {
        return (em_queue_t) EM_QUEUE_UNDEF;
    }

    /*
     * Critical Section Enter 
     */
    Emti_lockSemaphore ();

    /*
     * Open an HW queue throu the application resource manager.
     */
    lvRetStatus = ti_em_hw_queue_open (hw_queue_idx);
    if (lvRetStatus != EM_OK)
    {
        /*
         * Critical Section Exit 
         */
        Emti_unlockSemaphore ();
        return (em_queue_t) EM_QUEUE_UNDEF;
    }

    /*
     * set hwQueueBaseIdx = hw_queue_idx/(TI_EM_QUEUE_NUM_IN_SET - TI_EM_STATIC_QUEUE_NUM_IN_SET) 
     */
    lvHwQueueBaseIdx = hw_queue_idx / (TI_EM_QUEUE_NUM_IN_SET - TI_EM_STATIC_QUEUE_NUM_IN_SET);

    /*
     * set hwQueueBaseIdx = hwQueueBaseIdx*(TI_EM_QUEUE_NUM_IN_SET - TI_EM_STATIC_QUEUE_NUM_IN_SET) 
     */
    lvHwQueueBaseIdx = lvHwQueueBaseIdx * (TI_EM_QUEUE_NUM_IN_SET - TI_EM_STATIC_QUEUE_NUM_IN_SET);

    /*
     * set hwQueueStateTbl = <PSBSLOW>hwQueueStateTbl[hwQueueBaseIdx] 
     */
    lvHwQueueState = Emti_svPsbSlowPtr->hwQueueStateTbl[lvHwQueueBaseIdx];

    /*
     * set queueSetIdx = hwQueueStateTbl.setIdx 
     */
    lvQueueSetIdx = (int8_t) lvHwQueueState.setIdx;

    /*
     * set queueSetPending = 0 
     */
    lvQueueSetPending = 0;

    /*
     * if queueSetIdx == -1
     * set queueSetPending = 1
     * set queueSetIdx = <PSB>nextFreeQueueSetIdx
     * if queueSetIdx == -1
     * return EM_QUEUE_UNDEF 
     */
    if (lvQueueSetIdx == -1)
    {
        lvQueueSetPending = 1;
        lvQueueSetIdx = Emti_svPsbPtr->nextFreeQueueSetIdx;
        if (lvQueueSetIdx == -1)
        {
            /*
             * Critical Section Exit 
             */
            Emti_unlockSemaphore ();
            return (em_queue_t) EM_QUEUE_UNDEF;
        }
    }

    Emti_queueUpdateRouteInfo (&(Emti_svPsbPtr->shadowRDsc.queueRouteTbl[lvQueueSetIdx].routeInfo),
                               (int32_t) lvHwQueueBaseIdx, TI_EM_QUEUE_MODE_HW, 0);

    /*
     * send ROUTE_UPDATE command to router
     */
    if (Emti_svPcb.chainEnabled == TI_EM_CHAIN_ENABLED)
    {
        em_status_t lvRetStatus;

        lvRetStatus = Emti_updateRoute (lvQueueSetIdx, Emti_CtType_ROUTER_UPDATE_ROUTE);
        if (lvRetStatus != EM_OK)
        {
            Emti_unlockSemaphore ();
            return lvRetStatus;
        }
    }

    /*
     * set queueIdxInSet = hw_queue_idx - hwQueueBaseIdx 
     */
    lvQueueIdxInSet = TI_EM_STATIC_QUEUE_NUM_IN_SET + hw_queue_idx - lvHwQueueBaseIdx;

    /*
     * set queueIdx = queueIdxInSet*TI_EM_QUEUE_SET*NUM + queueSetIdx 
     */
    lvQueueIdx = lvQueueIdxInSet * TI_EM_QUEUE_SET_NUM + lvQueueSetIdx;

    /*
     * if <queue>state == Emti_QueueState_ALLOCATED
     * return EM_QUEUE_UNDEF 
     */
    lvQueueSlowPtr = &Emti_svQueueSlowMdb[lvQueueIdx];
    if (lvQueueSlowPtr->state == Emti_QueueState_ALLOCATED)
    {
        /*
         * Critical Section Exit 
         */
        Emti_unlockSemaphore ();
        return (em_queue_t) EM_QUEUE_UNDEF;
    }

    /*
     * if queueSetPending == 1 
     */
    if (lvQueueSetPending == 1)
    {
        /*
         * set <PSB>queueRouteTbl[queueSetIdx] = tmpQueueRoute 
         */
        Emti_svPsbPtr->queueRouteTbl[lvQueueSetIdx].routeInfo =
            Emti_svPsbPtr->shadowRDsc.queueRouteTbl[lvQueueSetIdx].routeInfo;

        /*
         * set hwQueueStateTbl.setIdx = queueSetIdx 
         */
        Emti_svPsbSlowPtr->hwQueueStateTbl[lvHwQueueBaseIdx].setIdx = lvQueueSetIdx;

        /*
         * set nextQueueSetIdx = queueSetIdx + 1
         * if nextQueueSetIdx == TI_EM_QUEUE_SET_NUM
         * set nextQueueSetIdx = -1 
         */
        lvNextQueueSetIdx = lvQueueSetIdx + 1;
        if (lvNextQueueSetIdx == TI_EM_QUEUE_SET_NUM)
            lvNextQueueSetIdx = -1;

        /*
         * set <PSB>nextFreeQueueSetIdx = nextQueueSetIdx 
         */
        Emti_svPsbPtr->nextFreeQueueSetIdx = lvNextQueueSetIdx;
    }

    /*
     * set <queue>state = Emti_QueueState_MAPPED 
     */
    lvQueueSlowPtr->state = Emti_QueueState_MAPPED;

    /*
     * set <queue>eoPtr = NULL 
     */
    lvQueueSlowPtr->eoPtr = NULL;

    /*
     * set <queue>name = name 
     */
    {
        if (strlen (name) <= EM_QUEUE_NAME_LEN)
            strcpy (lvQueueSlowPtr->name, name);
        else
            strncpy (lvQueueSlowPtr->name, name, EM_QUEUE_NAME_LEN);
    }

    /*
     * set <queue>mode = TI_EM_QUEUE_MODE_HW 
     */
    lvQueueSlowPtr->queueMode = TI_EM_QUEUE_MODE_HW;

    lvEqHandle =
        (em_queue_t) Emti_queueMakeGlobal ((em_queue_t) lvQueueIdx, Emti_gvTcb.myDeviceIdx, Emti_gvTcb.myProcessIdx);

    /*
     * Critical Section Exit 
     */
    Emti_unlockSemaphore ();

    /*
     * return handle 
     */
    return (em_queue_t) lvEqHandle;
}

em_status_t
ti_em_queue_create_hw_static (
    const char *name,
    ti_em_queue_id_t hw_queue_idx,
    em_queue_t queue
)
{
    Emti_QueueSlow_s *lvQueueSlowPtr;
    ti_em_device_id_t lvDeviceIdx;
    ti_em_process_id_t lvProcessIdx;
    ti_em_queue_id_t lvQueueIdx;
    em_status_t lvRetStatus;

    lvRetStatus = Emti_queueStaticInputCheck (queue, &lvDeviceIdx, &lvProcessIdx, &lvQueueIdx);
    if (lvRetStatus != EM_OK)
    {
        return lvRetStatus;
    }

    /*
     * if <queue>state == Emti_QueueState_ALLOCATED
     * return error 
     */
    lvQueueSlowPtr = &Emti_svQueueSlowMdb[lvQueueIdx];
    if (lvQueueSlowPtr->state == Emti_QueueState_ALLOCATED)
        return EM_ERR_BAD_STATE;

    /*
     * Critical Section Enter 
     */
    Emti_lockSemaphore ();

    /*
     * Open an HW queue throu the application resource manager.
     */
    lvRetStatus = ti_em_hw_queue_open (hw_queue_idx);
    if (lvRetStatus != EM_OK)
    {
        /*
         * Critical Section Exit 
         */
        Emti_unlockSemaphore ();
        return EM_ERR_ALLOC_FAILED;
    }

    Emti_queueUpdateRouteInfo (&(Emti_svPsbPtr->shadowRDsc.staticQueueRouteTbl[lvQueueIdx].routeInfo),
                               (int32_t) hw_queue_idx, TI_EM_QUEUE_MODE_HW, 0);

    /*
     * send ROUTE_UPDATE_STATIC command to router
     */
    if (Emti_svPcb.chainEnabled == TI_EM_CHAIN_ENABLED)
    {
        em_status_t lvRetStatus;

        lvRetStatus = Emti_updateRoute (lvQueueIdx, Emti_CtType_ROUTER_UPDATE_ROUTE_STATIC);
        if (lvRetStatus != EM_OK)
        {
            Emti_unlockSemaphore ();
            return lvRetStatus;
        }
    }

    /*
     * set <PSB>staticQueueRouteTbl[queueIdx] = tmpStaticQueueRoute 
     */
    {
        Emti_svPsbPtr->staticQueueRouteTbl[lvQueueIdx].routeInfo =
            Emti_svPsbPtr->shadowRDsc.staticQueueRouteTbl[lvQueueIdx].routeInfo;
    }

    /*
     * set <queue>state = Emti_QueueState_MAPPED 
     */
    lvQueueSlowPtr->state = Emti_QueueState_MAPPED;

    /*
     * set <queue>eoPtr = NULL 
     */
    lvQueueSlowPtr->eoPtr = NULL;

    /*
     * set <queue>name = name 
     */
    {
        if (strlen (name) <= EM_QUEUE_NAME_LEN)
            strcpy (lvQueueSlowPtr->name, name);
        else
            strncpy (lvQueueSlowPtr->name, name, EM_QUEUE_NAME_LEN);
    }

    /*
     * set <queue>mode = TI_EM_QUEUE_MODE_HW 
     */
    lvQueueSlowPtr->queueMode = TI_EM_QUEUE_MODE_HW;

    /*
     * Critical Section Exit 
     */
    Emti_unlockSemaphore ();

    return (em_status_t) (EM_OK);
}

em_status_t
em_queue_delete (
    em_queue_t queue
)
{
    em_status_t ret = EM_ERR_NOT_IMPLEMENTED;
    return ret;
}

em_status_t
em_queue_enable (
    em_queue_t queue
)
{
    em_status_t ret = EM_ERR_NOT_IMPLEMENTED;
    return ret;
}

em_status_t
em_queue_enable_all (
    em_eo_t eo
)
{
    em_status_t ret = EM_ERR_NOT_IMPLEMENTED;
    return ret;
}

em_status_t
em_queue_disable (
    em_queue_t queue,
    int num_notif,
    const em_notif_t * notif_tbl
)
{
    em_status_t ret = EM_ERR_NOT_IMPLEMENTED;
    return ret;
}

em_status_t
em_queue_disable_all (
    em_eo_t eo,
    int num_notif,
    const em_notif_t * notif_tbl
)
{
    em_status_t ret = EM_ERR_NOT_IMPLEMENTED;
    return ret;
}

em_status_t
em_queue_set_context (
    em_queue_t queue,
    const void *context
)
{
#if (EM_CHECK_LEVEL >= 1)
    Emti_QueueSlow_s *lvQueueSlowPtr;
#endif
    Emti_QueueFast_s *lvQueueFastPtr;
    uint16_t lvQueueIdx = Emti_getQueueIndex (queue);

#if (EM_CHECK_LEVEL >= 1)
    /*
     * Basic Validations: The queue index has to be in range. 
     */
    if (lvQueueIdx >= EM_MAX_QUEUES)
    {
        return Emti_error ((em_status_t) EM_ERR_BAD_ID, (em_escope_t) EM_ESCOPE_QUEUE_SET_CONTEXT);
    }
#endif

    /*
     * Get the pointer to the queue 
     */
#if (EM_CHECK_LEVEL >= 1)
    lvQueueSlowPtr = &Emti_svQueueSlowMdb[lvQueueIdx];
#endif
    lvQueueFastPtr = &Emti_svQueueFastMdb[lvQueueIdx];

#if (EM_CHECK_LEVEL >= 1)
    /*
     * Basic Validation: Ensure that the queue was at least allocated. 
     */
    if (lvQueueSlowPtr->state == Emti_QueueState_FREE)
    {
        return Emti_error ((em_status_t) EM_ERR_BAD_STATE, (em_escope_t) EM_ESCOPE_QUEUE_SET_CONTEXT);
    }
#endif

    /*
     * Critical Section Enter 
     */
    Emti_lockSemaphore ();

    /*
     * Set the queue context. 
     */
    lvQueueFastPtr->contextPtr = (void *) context;

    /*
     * Critical Section Exit 
     */
    Emti_unlockSemaphore ();

    /*
     * Context has been configured. 
     */
    return EM_OK;
}

void *
em_queue_get_context (
    em_queue_t queue
)
{
#if (EM_CHECK_LEVEL >= 1)
    Emti_QueueSlow_s *lvQueueSlowPtr;
#endif
    Emti_QueueFast_s *lvQueueFastPtr;
    uint16_t lvQueueIdx = Emti_getQueueIndex (queue);
    void *lvRetPtr;

#if (EM_CHECK_LEVEL >= 1)
    {
        /*
         * Basic Validations: The queue index has to be in range. 
         */
        if (lvQueueIdx >= EM_MAX_QUEUES)
        {
            Emti_error ((em_status_t) EM_ERR_BAD_ID, (em_escope_t) EM_ESCOPE_QUEUE_GET_CONTEXT);
            return NULL;
        }
    }
#endif

    /*
     * Get the pointer to the queue 
     */
#if (EM_CHECK_LEVEL >= 1)
    lvQueueSlowPtr = &Emti_svQueueSlowMdb[lvQueueIdx];
#endif
    lvQueueFastPtr = &Emti_svQueueFastMdb[lvQueueIdx];

    /*
     * Critical Section Enter 
     */
    Emti_lockSemaphore ();

    lvRetPtr = (void *) lvQueueFastPtr->contextPtr;

    /*
     * Critical Section Exit 
     */
    Emti_unlockSemaphore ();

#if (EM_CHECK_LEVEL >= 1)
    {
        /*
         * Basic Validation: Ensure that the queue was at least allocated. 
         */
        if (lvQueueSlowPtr->state == Emti_QueueState_FREE)
            return NULL;
    }
#endif

    /*
     * Return the queue context. 
     */
    return lvRetPtr;
}

size_t
em_queue_get_name (
    em_queue_t queue,
    char *name,
    size_t maxlen
)
{
    Emti_QueueSlow_s *lvQueueSlowPtr;
    size_t lvRet;
    uint16_t lvQueueIdx = Emti_getQueueIndex (queue);

    /*
     * If there was no buffer provided we cannot proceed. 
     */
    if ((maxlen == 0) || (name == NULL))
        return (size_t) 0;

#if (EM_CHECK_LEVEL >= 1)
    /*
     * Basic Validations: The queue index has to be in range. 
     */
    if (lvQueueIdx >= EM_MAX_QUEUES)
    {
        Emti_error ((em_status_t) EM_ERR_BAD_ID, (em_escope_t) EM_ESCOPE_QUEUE_GET_NAME);
        return (size_t) 0;
    }
#endif

    /*
     * Critical Section Enter 
     */
    Emti_lockSemaphore ();

    /*
     * Get the pointer to the queue 
     */
    lvQueueSlowPtr = &Emti_svQueueSlowMdb[lvQueueIdx];

#if (EM_CHECK_LEVEL >= 1)
    /*
     * Basic Validation: Ensure that the queue was at least allocated. 
     */
    if (lvQueueSlowPtr->state == Emti_QueueState_FREE)
    {
        /*
         * Critical Section Exit 
         */
        Emti_unlockSemaphore ();

        Emti_error ((em_status_t) EM_ERR_BAD_STATE, (em_escope_t) EM_ESCOPE_QUEUE_GET_NAME);

        return (size_t) 0;
    }

    if (maxlen > EM_QUEUE_NAME_LEN)
    {
        /*
         * Critical Section Exit 
         */
        Emti_unlockSemaphore ();

        Emti_error ((em_status_t) EM_ERR_TOO_LARGE, (em_escope_t) EM_ESCOPE_QUEUE_GET_NAME);

        return (size_t) 0;
    }
#endif

    /*
     * Get the length of the queue name. 
     */
    lvRet = (size_t) strlen (lvQueueSlowPtr->name);
    if (maxlen < lvRet)
        lvRet = maxlen;
    strncpy (name, lvQueueSlowPtr->name, lvRet);

    /*
     * Critical Section Exit 
     */
    Emti_unlockSemaphore ();

    /*
     * Return the length of the queue name. 
     */
    return lvRet;
}

em_queue_prio_t
em_queue_get_priority (
    em_queue_t queue
)
{
    Emti_QueueSlow_s *lvQueueSlowPtr;
    uint16_t lvQueueIdx = Emti_getQueueIndex (queue);
    em_queue_prio_t lvRet;

#if (EM_CHECK_LEVEL >= 1)
    /*
     * Basic Validations: The queue index has to be in range. 
     */
    if (lvQueueIdx >= EM_MAX_QUEUES)
    {
        Emti_error ((em_status_t) EM_ERR_BAD_ID, (em_escope_t) EM_ESCOPE_QUEUE_GET_PRIORITY);
        return (em_queue_prio_t) EM_QUEUE_PRIO_UNDEF;
    }
#endif

    /*
     * Get the pointer to the queue 
     */
    lvQueueSlowPtr = &Emti_svQueueSlowMdb[lvQueueIdx];

    /*
     * Critical Section Enter 
     */
    Emti_lockSemaphore ();

    lvRet = lvQueueSlowPtr->priority;

    /*
     * Critical Section Exit 
     */
    Emti_unlockSemaphore ();

#if (EM_CHECK_LEVEL >= 1)
    /*
     * Basic Validation: Ensure that the queue was at least allocated. 
     */
    if (lvQueueSlowPtr->state == Emti_QueueState_FREE)
    {
        Emti_error ((em_status_t) EM_ERR_BAD_STATE, (em_escope_t) EM_ESCOPE_QUEUE_GET_PRIORITY);
        return (em_queue_prio_t) EM_QUEUE_PRIO_UNDEF;
    }
#endif

    /*
     * Return the queue priority. 
     */
    return lvRet;
}

em_queue_type_t
em_queue_get_type (
    em_queue_t queue
)
{
    Emti_QueueSlow_s *lvQueueSlowPtr;
    uint16_t lvQueueIdx = Emti_getQueueIndex (queue);
    em_queue_type_t lvRet;

#if (EM_CHECK_LEVEL >= 1)
    /*
     * Basic Validations: The queue index has to be in range. 
     */
    if (lvQueueIdx >= EM_MAX_QUEUES)
    {
        Emti_error ((em_status_t) EM_ERR_BAD_ID, (em_escope_t) EM_ESCOPE_QUEUE_GET_TYPE);
        return (em_queue_type_t) (EM_QUEUE_TYPE_UNDEF);
    }
#endif

    /*
     * Get the pointer to the queue 
     */
    lvQueueSlowPtr = &Emti_svQueueSlowMdb[lvQueueIdx];

#if (EM_CHECK_LEVEL >= 1)
    /*
     * Basic Validation: Ensure that the queue was at least allocated. 
     */
    if (lvQueueSlowPtr->state == Emti_QueueState_FREE)
    {
        Emti_error ((em_status_t) EM_ERR_BAD_STATE, (em_escope_t) EM_ESCOPE_QUEUE_GET_TYPE);
        return (em_queue_type_t) (EM_QUEUE_TYPE_UNDEF);
    }
#endif

    /*
     * Critical Section Enter 
     */
    Emti_lockSemaphore ();

    lvRet = (em_queue_type_t) lvQueueSlowPtr->type;

    /*
     * Critical Section Exit 
     */
    Emti_unlockSemaphore ();

    /*
     * Return the queue type. 
     */
    return (em_queue_type_t) lvRet;
}

em_queue_t
ti_em_queue_make_global (
    em_queue_t queue_hdl,
    ti_em_device_id_t device_idx,
    ti_em_process_id_t process_idx
)
{
    return Emti_queueMakeGlobal (queue_hdl, device_idx, process_idx);
}

ti_em_device_id_t
ti_em_queue_get_device_id (
    em_queue_t queue_hdl
)
{
    return (ti_em_device_id_t) Emti_getField (queue_hdl, QUEUE_HANDLE_DEVICE_IDX);
}

ti_em_process_id_t
ti_em_queue_get_process_id (
    em_queue_t queue_hdl
)
{
    return (ti_em_process_id_t) Emti_getField (queue_hdl, QUEUE_HANDLE_PROCESS_IDX);
}

ti_em_queue_mode_t
ti_em_queue_get_mode (
    em_queue_t queue_hdl
)
{
    ti_em_queue_id_t lvQueueIdx = (uint16_t) Emti_getField (queue_hdl, QUEUE_HANDLE_QUEUE_IDX);
    uint16_t lvRouteInfo;
    uint8_t lvQueueMode;

    /*
     * If static queue  
     */
    if (lvQueueIdx < (EM_QUEUE_STATIC_MAX + 1))
    {
        lvRouteInfo = Emti_svPsbPtr->staticQueueRouteTbl[lvQueueIdx].routeInfo;
        lvQueueMode = (uint8_t) Emti_getField (lvRouteInfo, QUEUE_ROUTE_INFO_QUEUE_MODE);
    }
    /*
     * If not static queue  
     */
    else
    {
        uint8_t lvQueueSetIdx;
        lvQueueSetIdx = (uint8_t) Emti_getField (queue_hdl, QUEUE_HANDLE_QUEUE_SET_IDX);
        lvRouteInfo = Emti_svPsbPtr->queueRouteTbl[lvQueueSetIdx].routeInfo;
        lvQueueMode = (uint8_t) Emti_getField (lvRouteInfo, QUEUE_ROUTE_INFO_QUEUE_MODE);
    }
    return (ti_em_queue_mode_t) lvQueueMode;
}

ti_em_queue_id_t
ti_em_queue_get_queue_id (
    em_queue_t queue_hdl
)
{
    ti_em_queue_id_t lvQueueIdx = (uint16_t) Emti_getField (queue_hdl, QUEUE_HANDLE_QUEUE_IDX);
    uint16_t lvRouteInfo;
    ti_em_queue_id_t lvHwQueueIdx;
    uint8_t lvQueueMode;
    uint8_t lvQueueIdxInSet;
    uint8_t lvQueueSetIdx;
    uint8_t lvQueueEnabled;

    /*
     * If static queue  
     */
    if (lvQueueIdx < (EM_QUEUE_STATIC_MAX + 1))
    {
        lvRouteInfo = Emti_svPsbPtr->staticQueueRouteTbl[lvQueueIdx].routeInfo;
        lvQueueIdxInSet = 0;
        lvQueueMode = (uint8_t) Emti_getField (lvRouteInfo, QUEUE_ROUTE_INFO_QUEUE_MODE);
    }
    /*
     * If not static queue  
     */
    else
    {
        lvQueueSetIdx = (uint8_t) Emti_getField (queue_hdl, QUEUE_HANDLE_QUEUE_SET_IDX);
        lvRouteInfo = Emti_svPsbPtr->queueRouteTbl[lvQueueSetIdx].routeInfo;
        lvQueueIdxInSet = (uint8_t) Emti_getField (queue_hdl, QUEUE_HANDLE_QUEUE_IDX_IN_SET);
        lvQueueMode = (uint8_t) Emti_getField (lvRouteInfo, QUEUE_ROUTE_INFO_QUEUE_MODE);
        if (lvQueueMode == TI_EM_QUEUE_MODE_HW)
            lvQueueIdxInSet -= TI_EM_STATIC_QUEUE_NUM_IN_SET;

    }
    lvQueueEnabled = (uint8_t) Emti_getField (lvRouteInfo, QUEUE_ROUTE_INFO_ROUTE_ENABLED);
    if (lvQueueEnabled == 0)
    {
        return (ti_em_queue_id_t) 0xFFFF;
    }
    if (lvQueueMode == TI_EM_QUEUE_MODE_SD)
    {
        lvHwQueueIdx = Emti_gvTcb.sdQueueBaseIdx + Emti_getField (lvRouteInfo, QUEUE_ROUTE_INFO_SD_QUEUE_IDX);
    }
    if (lvQueueMode == TI_EM_QUEUE_MODE_HW)
    {
        lvHwQueueIdx = Emti_getField (lvRouteInfo, QUEUE_ROUTE_INFO_HW_QUEUE_IDX) + lvQueueIdxInSet;
    }

    return (ti_em_queue_id_t) lvHwQueueIdx;
}
