/*
 * Copyright (c) 2012, Texas Instruments Incorporated - http://www.ti.com/
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions are met:
 *       * Redistributions of source code must retain the above copyright
 *         notice, this list of conditions and the following disclaimer.
 *       * Redistributions in binary form must reproduce the above copyright
 *         notice, this list of conditions and the following disclaimer in the
 *         documentation and/or other materials provided with the distribution.
 *       * Neither the name of Texas Instruments Incorporated nor the
 *         names of its contributors may be used to endorse or promote products
 *         derived from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *   DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
 *   DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


/**
 *   @file  event_machine_queue_group.c
 *
 *   @brief
 *      Open Event Machine Implementation for Keystone1 devices.
 *      This file contains the queue group related functions
 *
 */

#include <inttypes.h>
#include <string.h>
#include <ti/runtime/openem/event_machine_macros.h>
#include <ti/runtime/openem/event_machine_types.h>
#include <ti/runtime/openem/event_machine_core_mask.h>
#include <ti/runtime/openem/event_machine_hw_types.h>
#include <ti/runtime/openem/src/event_machine_private_types.h>
#include <ti/runtime/openem/src/event_machine_private_variables.h>
#include <ti/runtime/openem/src/event_machine_private_functions.h>

/**
 * @internal 
 *
 * Internal Queue Group Find API
 *
 *  The function is used to find a queue group matching the specified core mask.
 *  This is an internal API.
 *
 * @param mask     Core Mask for which a matching queue group is being searched.
 *
 * @return
 *  Success -   Non Zero (Queue Group Index)
 * @return
 *  Error   -   EM_QUEUE_GROUP_UNDEF  (No Match found)
 */
static uint32_t
Emti_find_queue_group (
    const char *name
)
{
    uint32_t lvIndex;
    Emti_QueueGroup_s *lvQueueGroupPtr;

    /*
     * Run through the entire queue group database searching for a match. 
     */
    for (lvIndex = 0; lvIndex < EM_MAX_QUEUE_GROUPS; lvIndex++)
    {
        /*
         * Get the queue group. 
         */
        lvQueueGroupPtr = &Emti_svQueueGroupMdb[lvIndex];

        /*
         * Is the queue group allocated? Search only non-free groups. 
         */
        if (lvQueueGroupPtr->state != Emti_QueueGroupState_FREE)
        {
            /*
             * Check if there is a match on the core mask. If there is a match we are done 
             */
            if (!strcmp (lvQueueGroupPtr->name, name))
                return lvIndex;
        }
    }

    /*
     * Control comes here only if there is no match which is found. 
     */
    return EM_QUEUE_GROUP_UNDEF;
}

/*
 **********************************************************************************************************************
 **********************************************************************************************************************
 */

em_queue_group_t
em_queue_group_find (
    const char *name
)
{
    em_queue_group_t queueGroup;

#if (EM_CHECK_LEVEL >= 1)
    /*
     * Basic Validation: Ensure that the core name is valid. 
     */
    if (!strcmp (name, ""))
        return EM_QUEUE_GROUP_UNDEF;
#endif

    /*
     * Critical Section Enter 
     */
    Emti_lockSemaphore ();

    /*
     * Find the Event Queue Group. 
     */
    queueGroup = Emti_find_queue_group (name);

    /*
     * Critical Section Exit 
     */
    Emti_unlockSemaphore ();

    /*
     * Return the queue group index. 
     */
    return (em_queue_group_t) queueGroup;
}

em_queue_group_t
em_queue_group_create (
    const char *name,
    const em_core_mask_t * mask,
    int num_notif,
    const em_notif_t * notif_tbl
)
{
    em_core_mask_t lvMaskZero;
    int32_t lvQueueGroupIndex;
    Emti_QueueGroup_s *lvQueueGroupPtr;
    uint32_t lvStartIndex;

#if (EM_CHECK_LEVEL >= 1)
    /*
     * Basic Validation: Ensure that the core mask is valid. 
     */
    if (mask == NULL)
        return EM_QUEUE_GROUP_UNDEF;
#endif

#if (EM_CHECK_LEVEL >= 1)
    /*
     * Basic Validation: Ensure that the core name is valid. 
     */
    if (!strcmp (name, ""))
        return EM_QUEUE_GROUP_UNDEF;
#endif

    /*
     * Set the 0 core mask. 
     */
    em_core_mask_zero (&lvMaskZero);

    /*
     * Critical Section Enter 
     */
    Emti_lockSemaphore ();

    /*
     * Check if we need to create a new queue group or not? 
     */
    if (em_core_mask_equal (mask, &lvMaskZero))
    {
        /*
         * Core Mask was not specified. We will always create a new entry. 
         */
        lvQueueGroupIndex = EM_QUEUE_GROUP_UNDEF;
    }
    else
    {
        /*
         * Core Mask was specified; so search for an existing entry. If none is found
         * * we will create a new entry. 
         */
        lvQueueGroupIndex = Emti_find_queue_group (name);
    }

    /*
     * Do we need to create a new queue group entry? 
     */
    if (lvQueueGroupIndex == EM_QUEUE_GROUP_UNDEF)
    {
        lvStartIndex = Emti_svPsbPtr->nextFreeQueueGroupIdx;

        if (lvStartIndex == EM_MAX_QUEUE_GROUPS)
        {
            /*
             * Critical Section Exit 
             */
            Emti_unlockSemaphore ();

            return (em_queue_group_t) (EM_QUEUE_GROUP_UNDEF);
        }

        /*
         * Get the queue group. 
         */
        lvQueueGroupPtr = &Emti_svQueueGroupMdb[lvStartIndex];
        while (lvQueueGroupPtr->state != Emti_QueueGroupState_FREE)
        {
            lvStartIndex++;
            if (lvStartIndex == EM_MAX_QUEUE_GROUPS)
            {
                /*
                 * Critical Section Exit 
                 */
                Emti_unlockSemaphore ();

                return (em_queue_group_t) (EM_QUEUE_GROUP_UNDEF);
            }
            lvQueueGroupPtr = &Emti_svQueueGroupMdb[lvStartIndex];
        }

        lvQueueGroupIndex = lvStartIndex;

        /*
         * YES. Populate the queue group mask with all zeros. 
         */
        em_core_mask_copy (&(lvQueueGroupPtr->coreMskTbl), mask);

        if (strlen (name) <= EM_QUEUE_GROUP_NAME_LEN)
            strcpy (lvQueueGroupPtr->name, name);
        else
            strncpy (lvQueueGroupPtr->name, name, EM_QUEUE_GROUP_NAME_LEN);

        /*
         * The Queue Group has now been allocated. 
         */
        lvQueueGroupPtr->state = Emti_QueueGroupState_ALLOCATED;

        Emti_svPsbPtr->nextFreeQueueGroupIdx = lvQueueGroupIndex + 1;
    }
    else
    {
        lvQueueGroupIndex = EM_QUEUE_GROUP_UNDEF;
    }

    /*
     * Critical Section Exit 
     */
    Emti_unlockSemaphore ();

    /*
     * Return the queue group index. 
     */
    return (em_queue_group_t) lvQueueGroupIndex;
}

em_queue_group_t
em_queue_get_group (
    em_queue_t queue
)
{
    Emti_QueueSlow_s *lvQueueSlowPtr;
    uint16_t lvQueueIdx = Emti_getQueueIndex (queue);
    em_queue_group_t lvRet;

#if (EM_CHECK_LEVEL >= 1)
    /*
     * Basic Validations: The queue index has to be in range. 
     */
    if (lvQueueIdx >= EM_MAX_QUEUES)
    {
        Emti_error ((em_status_t) EM_ERR_BAD_ID, (em_escope_t) EM_ESCOPE_QUEUE_GET_GROUP);
        return (em_queue_group_t) EM_QUEUE_GROUP_UNDEF;
    }
#endif

    /*
     * Get the pointer to the queue 
     */
    lvQueueSlowPtr = &Emti_svQueueSlowMdb[lvQueueIdx];

    /*
     * Critical Section Enter 
     */
    Emti_lockSemaphore ();

    lvRet = lvQueueSlowPtr->queueGroupPtr;

    /*
     * Critical Section Exit 
     */
    Emti_unlockSemaphore ();

#if (EM_CHECK_LEVEL >= 1)
    /*
     * Basic Validation: Ensure that the queue was at least allocated. 
     */
    if (lvQueueSlowPtr->state == Emti_QueueState_FREE)
    {
        Emti_error ((em_status_t) EM_ERR_BAD_STATE, (em_escope_t) EM_ESCOPE_QUEUE_GET_GROUP);
        return (em_queue_group_t) EM_QUEUE_GROUP_UNDEF;
    }
#endif

    /*
     * Return the queue group. 
     */
    return lvRet;
}

em_status_t
em_queue_group_delete (
    em_queue_group_t group,
    int num_notif,
    const em_notif_t * notif_tbl
)
{
    em_status_t ret = EM_ERR_NOT_IMPLEMENTED;
    return ret;
}

em_status_t
em_queue_group_modify (
    em_queue_group_t group,
    const em_core_mask_t * new_coremask,
    int num_notif,
    const em_notif_t * notif_tbl
)
{
    em_status_t ret = EM_ERR_NOT_IMPLEMENTED;
    return ret;
}

em_status_t
em_queue_group_mask (
    em_queue_group_t group,
    em_core_mask_t * mask
)
{
    Emti_QueueGroup_s *lvQueueGroupPtr;

#if (EM_CHECK_LEVEL >= 1)
    /*
     * Basic Validation: Ensure that the queue group is valid and within range 
     */
    if ((group == EM_QUEUE_GROUP_UNDEF) || (group >= EM_MAX_QUEUE_GROUPS))
        return EM_ERR_BAD_CONTEXT;

    /*
     * Basic Validation: Ensure that a valid core mask is specified. 
     */
    if (mask == NULL)
        return EM_ERR;
#endif

    /*
     * Get the pointer to the queue group 
     */
    lvQueueGroupPtr = &Emti_svQueueGroupMdb[group];

#if (EM_CHECK_LEVEL >= 1)
    /*
     * Basic Validation: Ensure that the queue group is allocated. 
     */
    if (lvQueueGroupPtr->state == Emti_QueueGroupState_FREE)
        return EM_ERR_BAD_STATE;
#endif

    /*
     * Copy the Mask Information 
     */
    em_core_mask_copy (mask, &(lvQueueGroupPtr->coreMskTbl));
    return EM_OK;
}
