/*
 * Copyright (c) 2012, Texas Instruments Incorporated - http://www.ti.com/
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions are met:
 *       * Redistributions of source code must retain the above copyright
 *         notice, this list of conditions and the following disclaimer.
 *       * Redistributions in binary form must reproduce the above copyright
 *         notice, this list of conditions and the following disclaimer in the
 *         documentation and/or other materials provided with the distribution.
 *       * Neither the name of Texas Instruments Incorporated nor the
 *         names of its contributors may be used to endorse or promote products
 *         derived from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *   DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
 *   DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


/**
 *   @file  event_machine_scattered.c
 *
 *   @brief
 *      Open Event Machine Implementation for Keystone devices.
 *      Functions that are managing scattered events.
 *
 */

#include <inttypes.h>
#include <stddef.h>
#include <string.h>
#include <ti/runtime/openem/event_machine_macros.h>
#include <ti/runtime/openem/event_machine_types.h>
#include <ti/runtime/openem/event_machine_core_mask.h>
#include <ti/runtime/openem/event_machine_hw_macros.h>
#include <ti/runtime/openem/event_machine_hw_types.h>
#include <ti/runtime/openem/event_machine_hw_ti_macros.h>
#include <ti/runtime/openem/event_machine_hw_ti_types.h>
#include <ti/runtime/openem/event_machine_hw_ti_functions.h>
#include <ti/runtime/openem/src/event_machine_hwpform.h>
#include <ti/runtime/openem/src/event_machine_private_macros.h>
#include <ti/runtime/openem/src/event_machine_private_types.h>
#include <ti/runtime/openem/src/event_machine_private_variables.h>
#include <ti/runtime/openem/src/event_machine_private_functions.h>
#include <ti/runtime/openem/modules/event_machine_modules.h>

em_event_t
ti_em_alloc_with_buffers (
    em_event_type_t event_type,
    em_pool_id_t pool_id,
    int32_t buffer_num,
    ti_em_buffer_config_t * buffer_config_tbl
)
{
    Cppi_HostDesc *lvHostDescPtr = NULL;
    Cppi_HostDesc *lvPreviousHostDescPtr = NULL;
    Cppi_HostDesc *lvPrimaryHostDescPtr = NULL;
    uint32_t lvBufferIndex;
    uint32_t lvPacketsize = 0;
    Emti_event_mode_t lvEventMode = Emti_EVENT_MODE_GLOBAL_HW;
    uint32_t lvfreeQueueRegAdd = Emti_gvTcb.freeQueueRegDPtrTbl[pool_id];

    /*
     * check if the pool buf mode is TI_EM_BUF_MODE_TIGHT 
     */
    if (Emti_gvTcb.poolConfigTbl[pool_id].buf_mode == TI_EM_BUF_MODE_TIGHT)
    {
        return EM_EVENT_UNDEF;
    }

    /*
     * loop over buffers 
     */
    for (lvBufferIndex = 0; lvBufferIndex < buffer_num; lvBufferIndex++)
    {
        ti_em_buffer_config_t *lvBufferCfgPtr;

        /*
         * initialize the allocated descriptor to NULL. 
         */
        lvHostDescPtr = (Cppi_HostDesc *) Emti_publicDesc_ptov ((*((volatile uint32_t *) (lvfreeQueueRegAdd))));

        /*
         * initialize pointer to buffer parameters structure 
         */
        lvBufferCfgPtr = &(buffer_config_tbl[lvBufferIndex]);

        /*
         * got null pointer, return previous allocated descriptors to free queue 
         */
        if (lvHostDescPtr == NULL)
        {
            uint32_t lvReturnIndex;
            for (lvReturnIndex = 0; lvReturnIndex < lvBufferIndex; lvReturnIndex++)
            {
                if (lvReturnIndex == 0)
                    lvHostDescPtr = lvPrimaryHostDescPtr;
                else
                    lvHostDescPtr = (Cppi_HostDesc *) Emti_publicDesc_ptov (lvHostDescPtr->nextBDPtr);

		Emti_mfence();

                *(uint32_t *) (lvfreeQueueRegAdd) = Emti_publicDesc_vtop ((void *) (lvHostDescPtr));
            }
            return (em_event_t) EM_EVENT_UNDEF;
        }

        if (lvBufferIndex == 0)
            lvPrimaryHostDescPtr = lvHostDescPtr;

        /*
         * descriptor word 0 - default ps location is in descriptor. 
         */
        lvHostDescPtr->descInfo = (Cppi_PSLoc_PS_IN_DESC << 22) | (lvBufferCfgPtr->buffer_size & 0x3fffff);

        /*
         * compute packet size 
         */
        lvPacketsize += lvBufferCfgPtr->buffer_size;

        lvHostDescPtr->packetInfo = (lvHostDescPtr->packetInfo & 0x00ffffff) | 0x80000000;

        /*
         * set buffer size 
         */
        lvHostDescPtr->buffLen = lvBufferCfgPtr->buffer_size;

        /*
         * set buffer pointer 
         */
        lvHostDescPtr->buffPtr = Emti_bufferPtr_vtop (lvBufferCfgPtr->buffer_ptr, pool_id);

        /*
         * set next descriptor pointer 
         */
        lvHostDescPtr->nextBDPtr = 0;

        /*
         * link previous descriptor pointer to our descriptor 
         */
        if (lvBufferIndex != 0)
            lvPreviousHostDescPtr->nextBDPtr = Emti_publicDesc_vtop (lvHostDescPtr);

        /*
         * set original buffer length 
         */
        Emti_setOrigBufferLength (lvHostDescPtr, lvBufferCfgPtr->orig_buffer_size);

        /*
         * set original buffer pointer 
         */
        lvHostDescPtr->origBuffPtr = Emti_bufferPtr_vtop (lvBufferCfgPtr->orig_buffer_ptr, pool_id);

        /*
         * set event mode
         */
        if (lvBufferCfgPtr->free_func != NULL)
        {
            lvEventMode = Emti_EVENT_MODE_GLOBAL_SW;
        }

        /*
         * set coherency mode
         */
        Emti_setCohMode (lvHostDescPtr,
                         lvBufferCfgPtr->coh_mode == TI_EM_COH_MODE_ON ? TI_EM_COH_MODE_ON : TI_EM_COH_MODE_OFF);

        /*
         * set pool index
         */
        Emti_setPoolIndex (lvHostDescPtr, pool_id);

        /*
         * set free function pointer 
         */
        Emti_packet_set_free_function ((ti_em_packet_t *) (lvHostDescPtr), lvBufferCfgPtr->free_func);

        /*
         * store descriptor pointer for next iteration 
         */
        lvPreviousHostDescPtr = lvHostDescPtr;
    }

    /*
     * set event type 
     */
    lvPrimaryHostDescPtr->timeStamp = (((uint8_t) event_type) << 24);

    /*
     * reset ps word size to 0 
     */
    ti_em_set_ps_wsize ((em_event_t) (lvPrimaryHostDescPtr), 0);

    /*
     * update packet length for primary event 
     */
    lvPrimaryHostDescPtr->descInfo = (lvPrimaryHostDescPtr->descInfo & 0xffc00000) | (lvPacketsize & 0x3fffff);

    /*
     * update event mode
     */
    Emti_setEventMode (lvPrimaryHostDescPtr, lvEventMode);

    /*
     * moving from host descriptor pointer to event handle and return
     */
    return (em_event_t) (((uint32_t) (lvPrimaryHostDescPtr) & 0xfffffff0) | lvEventMode);
}

em_event_t
ti_em_free_with_buffers (
    em_event_t event,
    int32_t buffer_num,
    ti_em_buffer_config_t * buffer_config_tbl
)
{
    Cppi_HostDesc *lvHostDescPtr;
    uint32_t lvBufferIndex = 0;
    uint32_t lvPacketSize = 0;
    uint32_t lvCurrentSize = 0;
    Emti_event_mode_t lvEventMode = Emti_getEventModeFromEventHdl (event);

    /*
     * manage cache coherency on first buffer 
     */
    if (Emti_inv (event) == 0)
        Emti_mfence ();

    /*
     * Get the pointer to the host descriptor from the event handle. 
     */
    lvHostDescPtr = (Cppi_HostDesc *) (Emti_getCppiHostDescPtr (event));

    /*
     * get complete packet size 
     */
    lvPacketSize = lvHostDescPtr->descInfo & 0x3fffff;

    /*
     * loop while descriptor ptr is not NULL 
     */
    while (lvHostDescPtr != NULL)
    {
        uint32_t lvBufferFreeQueueIndex = Emti_getFreeQueueIndex (lvHostDescPtr);
        uint32_t lvBufferFreePushPolicy = Emti_getFreePushPolicy (lvHostDescPtr);
        uint32_t lvBufferMode = Emti_getBufMode (lvHostDescPtr);

        Emti_setEventMode (lvHostDescPtr, Emti_EVENT_MODE_GLOBAL_HW);

        /*
         * if loose buffer 
         */
        if (lvBufferMode == TI_EM_BUF_MODE_LOOSE)
        {
            /*
             * check if split needed and return tail event from split 
             */
            if (lvBufferIndex >= buffer_num)
            {
                lvPacketSize -= lvCurrentSize;
                lvHostDescPtr->descInfo = (lvHostDescPtr->descInfo & 0xffc00000) | (lvPacketSize & 0x3fffff);
                Emti_setEventMode (lvHostDescPtr, lvEventMode);

                return (em_event_t) (((uint32_t) (lvHostDescPtr) & 0xfffffff0) | lvEventMode);
            }

            /*
             * build ti_em_buffer_config_t 
             */
            {
                ti_em_buffer_config_t *lvBufferConfigPtr;
                lvBufferConfigPtr = (ti_em_buffer_config_t *) (&(buffer_config_tbl[lvBufferIndex]));
                Emti_setBufferConfigFromEventDsc (lvHostDescPtr, lvBufferConfigPtr);
            }

            lvBufferIndex++;
        }

        if (lvBufferFreePushPolicy == 1)
            Emti_pushEventToHead ((uint16_t) (lvBufferFreeQueueIndex), (uint32_t *) lvHostDescPtr,
                                  TI_EM_MEM_PUBLIC_DESC);
        else
            Emti_pushEvent ((uint16_t) (lvBufferFreeQueueIndex), (uint32_t *) lvHostDescPtr, TI_EM_MEM_PUBLIC_DESC);

        /*
         * get packet size from descriptor 
         */
        lvCurrentSize += (lvHostDescPtr->buffLen & 0x3fffff);

        /*
         * get next pointer 
         */
        lvHostDescPtr = (Cppi_HostDesc *) Emti_publicDesc_ptov (lvHostDescPtr->nextBDPtr);
    }

    /*
     * check if buffer_config_tbl is not full, then tag first unused entry with NULL. Buffer
     * * index already points to next unused index 
     */
    if (lvBufferIndex < buffer_num)
    {
        buffer_config_tbl[lvBufferIndex].buffer_ptr = NULL;
        buffer_config_tbl[lvBufferIndex].buffer_size = 0;
        buffer_config_tbl[lvBufferIndex].orig_buffer_ptr = NULL;
        buffer_config_tbl[lvBufferIndex].orig_buffer_size = 0;
        buffer_config_tbl[lvBufferIndex].coh_mode = (ti_em_coh_mode_t) (NULL);
        buffer_config_tbl[lvBufferIndex].free_func = (ti_em_free_func_t) (NULL);
    }

    return (em_event_t) (EM_EVENT_UNDEF);
}

em_event_t
ti_em_combine (
    ti_em_pair_t event_pair,
    ti_em_iterator_t * iterator_ptr
)
{
    Emti_event_mode_t lvHeadEventMode;
    Emti_event_mode_t lvTailEventMode;
    ti_em_packet_t *lvHeadPacketPtr;
    ti_em_packet_t *lvTailPacketPtr;
    Cppi_HostDesc *lvHeadDescPtr;
    Cppi_HostDesc *lvCurrentDescPtr;
    Cppi_HostDesc *lvTailDescPtr;
    uint32_t lvTailSize;
    uint32_t lvEventSize;
    uint32_t lvHeadSize;
    em_status_t lvNextStatus;

    /* set headEventHdl = event_pair.head_event */
    em_event_t lvHeadEvent = event_pair.head_event_hdl;
    
    /* set tailEventHdl = event_pair.tail_event */
    em_event_t lvTailEvent = event_pair.tail_event_hdl;

    if (ti_em_to_packet((em_event_t)(((Emti_Iterator_s *) (iterator_ptr))->event_hdl)) != ti_em_to_packet(lvHeadEvent))
        return EM_EVENT_UNDEF;

    /* set headEventMode = headEventHdl.eventMode */
    lvHeadEventMode = Emti_getEventModeFromEventHdl (lvHeadEvent);

    /* set tailEventMode = tailEventHdl.eventMode */
    lvTailEventMode = Emti_getEventModeFromEventHdl (lvTailEvent);

    if (lvHeadEventMode == Emti_EVENT_MODE_LOCAL)
    {
        if (lvTailEventMode == Emti_EVENT_MODE_GLOBAL_SW)
            return EM_EVENT_UNDEF;
        else
            lvTailEventMode = Emti_EVENT_MODE_GLOBAL_HW;
    }

    if (lvHeadEventMode == Emti_EVENT_MODE_GLOBAL_SW)
    {
        if (lvTailEventMode == Emti_EVENT_MODE_LOCAL)
            return EM_EVENT_UNDEF;
        else
            lvTailEventMode = Emti_EVENT_MODE_GLOBAL_HW;
    }

    if (lvHeadEventMode == Emti_EVENT_MODE_GLOBAL_HW)
    {
        if (lvTailEventMode == Emti_EVENT_MODE_LOCAL)
            lvHeadEventMode = Emti_EVENT_MODE_LOCAL;
        if (lvTailEventMode == Emti_EVENT_MODE_GLOBAL_SW)
            lvHeadEventMode = Emti_EVENT_MODE_GLOBAL_SW;
        lvTailEventMode = Emti_EVENT_MODE_GLOBAL_HW;
    }

    /* set headEventDscPtr = headEventHdl.dscPtr */
    lvHeadPacketPtr = ti_em_to_packet (lvHeadEvent);
    lvHeadDescPtr = (Cppi_HostDesc *) lvHeadPacketPtr;
    
    /* read headEventSize from headEventDscPtr->eventSize */
    lvHeadSize = (uint32_t) (lvHeadDescPtr->descInfo & 0x3FFFFF);

    /* set tailEventDscPtr = tailEventHdl.dscPtr */
    lvTailPacketPtr = ti_em_to_packet (lvTailEvent);
    lvTailDescPtr = (Cppi_HostDesc *) lvTailPacketPtr;
    
    /* read tailEventSize from tailEventDscPtr->eventSize */
    lvTailSize = (uint32_t) (lvTailDescPtr->descInfo & 0x3FFFFF);

    /* Emti_wbinv(tailEventHdl) */
    Emti_wbinv (event_pair.tail_event_hdl);

    /* set eventSize = headEventSize + tailEventSize */
    lvEventSize = lvHeadSize + lvTailSize;
    lvNextStatus = EM_OK;

    while (lvNextStatus != EM_ERR_NOT_FOUND)
    	lvNextStatus = ti_em_iterator_next(iterator_ptr);
    
    /* write eventSize to headEventDscPtr->eventSize */
    lvHeadDescPtr->descInfo = lvEventSize | (Cppi_PSLoc_PS_IN_DESC << 22);

    /* set currentDscPtr = iteratorPtr->currentDscPtr */
    lvCurrentDescPtr = (Cppi_HostDesc *)(((Emti_Iterator_s *) (iterator_ptr))->current_desc_ptr);

    /* write tailEventDscPtr to currentDscPtr->nextDscPtr */
    lvCurrentDescPtr->nextBDPtr = Emti_publicDesc_vtop (lvTailPacketPtr);

    /* write tailEventMode to tailEventDscPtr->eventMode */
    Emti_setEventMode (lvTailDescPtr, lvTailEventMode);
    
    /* write headEventMode to headEventDscPtr->eventMode */
    Emti_setEventMode (lvHeadDescPtr, lvHeadEventMode);
    
    /* set headEventHdl.eventMode = headEventMode */
    lvHeadEvent = (em_event_t) (((uint32_t) lvHeadPacketPtr | (lvHeadEventMode << 0)));

    /* return head event handle */
    return lvHeadEvent;
}

ti_em_pair_t
ti_em_split (
    em_event_t event,
    ti_em_iterator_t * iterator_ptr
)
{
    ti_em_pair_t lvEventPair;
    Emti_event_mode_t lvHeadEventMode;
    uint32_t lvCohLevel;
    uint32_t lvCurrentLevel;
    uint32_t lvEventSize;
    uint32_t lvHeadEventSize;
    uint32_t lvTailEventSize;
    uint32_t lvIndex = 0;
    Cppi_HostDesc *lvHeadDescPtr;
    Cppi_HostDesc *lvTailDescPtr;
    Cppi_HostDesc *lvCurrentDescPtr;

    if (ti_em_to_packet((em_event_t)(((Emti_Iterator_s *) (iterator_ptr))->event_hdl)) != ti_em_to_packet(event))
    {
        lvEventPair.head_event_hdl = (em_event_t) (EM_EVENT_UNDEF);
        lvEventPair.tail_event_hdl = (em_event_t) (EM_EVENT_UNDEF);
        return lvEventPair;
    }

    /* set headEventMode = headEventHdl.eventMode */
    lvHeadEventMode = Emti_getEventModeFromEventHdl (event);

    /* set headEventDscPtr = headEventHdl.dscPtr */
    lvHeadDescPtr = (Cppi_HostDesc *) (Emti_getCppiHostDescPtr (event));

    /* set currentDscPtr = iterator_ptr->currentDscPtr */
    lvCurrentDescPtr = (Cppi_HostDesc *)(((Emti_Iterator_s *) (iterator_ptr))->current_desc_ptr);
    
    /* read tailEventDscPtr from currentDscPtr->nextDscPtr */
    lvTailDescPtr = (Cppi_HostDesc *) Emti_publicDesc_ptov ((lvCurrentDescPtr->nextBDPtr));

    if (lvTailDescPtr == NULL)
    {
        lvEventPair.head_event_hdl = (em_event_t) (EM_EVENT_UNDEF);
        lvEventPair.tail_event_hdl = (em_event_t) (EM_EVENT_UNDEF);
        return lvEventPair;
    }
        
    /* set tailEventHdl.eventMode = headEventMode */
    Emti_setEventMode (lvTailDescPtr, lvHeadEventMode);
    
    /* read currentDscPtr from tailEventDscPtr->nextDscPtr */
    lvCurrentDescPtr = (Cppi_HostDesc *) Emti_publicDesc_ptov ((lvTailDescPtr->nextBDPtr));
    
    /* set currentLevel = iterator_ptr->currentLevel */
    lvCurrentLevel = ((Emti_Iterator_s *) (iterator_ptr))->current_level;
    
    /* set cohLevel = iterator_ptr->cohLevel */
    lvCohLevel = ((Emti_Iterator_s *) (iterator_ptr))->coh_level;
    
    for (lvIndex = (lvCurrentLevel + 2); lvIndex <= lvCohLevel; lvIndex++)
    {
        Emti_wbinv((em_event_t)(lvCurrentDescPtr));
        lvCurrentDescPtr = (Cppi_HostDesc *) Emti_publicDesc_ptov ((lvCurrentDescPtr->nextBDPtr));
    }
    
    /* set currentDscPtr = iterator_ptr->currentDscPtr */
    lvCurrentDescPtr = (Cppi_HostDesc *) (((Emti_Iterator_s *) (iterator_ptr))->current_desc_ptr);

    /* write 0 to currentDscPtr->nextDscPtr */
    lvCurrentDescPtr->nextBDPtr = 0;
    
    /* set iterator_ptr->cohLevel = currentLevel */
    ((Emti_Iterator_s *) (iterator_ptr))->coh_level = lvCurrentLevel;
    
    /* read eventSize from headEventDscPtr->eventSize */
    lvEventSize = (lvHeadDescPtr->descInfo & 0x0003fffff);
    
    /* set headEventSize = iterator_ptr->currentSize + iterator_ptr->currentBufSize */
    lvHeadEventSize = (((Emti_Iterator_s *) (iterator_ptr))->current_size & 0x0003fffff) +
                      (((Emti_Iterator_s *) (iterator_ptr))->current_buf_size);
    
    /* write headEventSize to headEventDscPtr->eventSize */
    lvHeadDescPtr->descInfo = (lvHeadDescPtr->descInfo & 0xffc00000) | (lvHeadEventSize);
    
    /* set tailEventSize = eventSize - headEventSize */
    lvTailEventSize = lvEventSize - lvHeadEventSize;
    
    /* write tailEventSize to tailEventDscPtr->eventSize */
    lvTailDescPtr->descInfo = (lvTailDescPtr->descInfo & 0xffc00000) | (lvTailEventSize);
    
    /* write headEventMode to tailEventDscPtr->eventMode */
    Emti_setEventMode (lvTailDescPtr, lvHeadEventMode);
    
    /* set event_pair.head_event = headEventHdl */
    lvEventPair.head_event_hdl = (em_event_t) (event);

    /* set event_pair.tail_event = tailEventHdl */
    lvEventPair.tail_event_hdl = (em_event_t) (((uint32_t) (lvTailDescPtr) & 0xfffffff0) | (lvHeadEventMode));

    /* return event_pair */
    return lvEventPair;
}

size_t
ti_em_iterator_size (
    ti_em_iterator_t * iterator_ptr
)
{
    if (((Emti_Iterator_s *) (iterator_ptr))->event_hdl == EM_EVENT_UNDEF)
        return 0;

    return (size_t) (((Emti_Iterator_s *) (iterator_ptr))->current_buf_size);
}

em_status_t
ti_em_iterator_start (
    em_event_t event,
    ti_em_iterator_t * iterator_ptr
)
{
    Cppi_HostDesc *lvDescPtr;
    uint32_t lvBufferSize;
    uint32_t lvBufferPtr;

    lvDescPtr = (Cppi_HostDesc *) (Emti_getCppiHostDescPtr (event));

    lvBufferSize = (lvDescPtr->buffLen & 0x3fffff);
    lvBufferPtr = (lvDescPtr->buffPtr);

    ((Emti_Iterator_s *) (iterator_ptr))->event_hdl = (uint32_t) (event);
    ((Emti_Iterator_s *) (iterator_ptr))->current_desc_ptr = (uint32_t) (lvDescPtr);
    ((Emti_Iterator_s *) (iterator_ptr))->current_buf_ptr = lvBufferPtr;
    ((Emti_Iterator_s *) (iterator_ptr))->current_buf_size = lvBufferSize;
    ((Emti_Iterator_s *) (iterator_ptr))->current_size = 0;
    ((Emti_Iterator_s *) (iterator_ptr))->current_level = 0;
    ((Emti_Iterator_s *) (iterator_ptr))->coh_level = 0;
    ((Emti_Iterator_s *) (iterator_ptr))->coh_desc_ptr = (uint32_t) (NULL);

    return EM_OK;
}

em_status_t
ti_em_iterator_stop (
    ti_em_iterator_t * iterator_ptr
)
{
    Cppi_HostDesc *lvCurrentDescPtr = (Cppi_HostDesc *) ((((Emti_Iterator_s *) (iterator_ptr))->coh_desc_ptr));
    uint32_t lvEventCohLevel = ((Emti_Iterator_s *) (iterator_ptr))->coh_level;
    uint32_t lvBufferIndex;

    if (((Emti_Iterator_s *) (iterator_ptr))->event_hdl == EM_EVENT_UNDEF)
        return EM_ERR_BAD_STATE;
        
    if (lvCurrentDescPtr == NULL)
        return EM_OK;
            
    for (lvBufferIndex = 1; lvBufferIndex <= lvEventCohLevel; lvBufferIndex++)
    {
        Emti_wbinv ((em_event_t) (lvCurrentDescPtr));
        lvCurrentDescPtr = (Cppi_HostDesc *) Emti_publicDesc_ptov (lvCurrentDescPtr->nextBDPtr);
    }

    ((Emti_Iterator_s *) (iterator_ptr))->event_hdl = EM_EVENT_UNDEF;

    return EM_OK;
}

em_status_t
ti_em_iterator_next (
    ti_em_iterator_t * iterator_ptr
)
{
    Cppi_HostDesc *lvCurrentDescPtr = (Cppi_HostDesc *) (((Emti_Iterator_s *) (iterator_ptr))->current_desc_ptr);
    Cppi_HostDesc *lvNextDescPtr;
    uint32_t lvNextBufferSize;
    uint32_t lvNextBufferPtr;

    if (((Emti_Iterator_s *) (iterator_ptr))->event_hdl == EM_EVENT_UNDEF)
        return EM_ERR_BAD_STATE;

    lvNextDescPtr = (Cppi_HostDesc *) Emti_publicDesc_ptov (lvCurrentDescPtr->nextBDPtr);
    if (lvNextDescPtr == NULL)
        return EM_ERR_NOT_FOUND;

    lvNextBufferSize = (lvNextDescPtr->buffLen & 0x3fffff);
    lvNextBufferPtr = lvNextDescPtr->buffPtr;

    /*
     * write current descriptor pointer to <next DSC>previousDscPtr 
     */
    lvNextDescPtr->softwareInfo2 = (uint32_t) (lvCurrentDescPtr);

    ((Emti_Iterator_s *) (iterator_ptr))->current_desc_ptr = (uint32_t) (lvNextDescPtr);
    ((Emti_Iterator_s *) (iterator_ptr))->current_buf_ptr = lvNextBufferPtr;
    ((Emti_Iterator_s *) (iterator_ptr))->current_buf_size = lvNextBufferSize;
    ((Emti_Iterator_s *) (iterator_ptr))->current_size +=
        ((Emti_Iterator_s *) (iterator_ptr))->current_buf_size;
    ((Emti_Iterator_s *) (iterator_ptr))->current_level++;
    
    if (((Emti_Iterator_s *) (iterator_ptr))->coh_desc_ptr == 0)
        ((Emti_Iterator_s *) (iterator_ptr))->coh_desc_ptr = (uint32_t)(lvNextDescPtr);
    
    return EM_OK;
}

em_status_t
ti_em_iterator_previous (
    ti_em_iterator_t * iterator_ptr
)
{
    Cppi_HostDesc *lvCurrentDescPtr = (Cppi_HostDesc *) (((Emti_Iterator_s *) (iterator_ptr))->current_desc_ptr);
    Cppi_HostDesc *lvPreviousDescPtr;
    uint32_t lvPreviousBufferSize;
    uint32_t lvPreviousBufferPtr;

    if (((Emti_Iterator_s *) (iterator_ptr))->event_hdl == EM_EVENT_UNDEF)
        return EM_ERR_BAD_STATE;

    if (((Emti_Iterator_s *) (iterator_ptr))->current_level == 0)
        return EM_ERR_NOT_FOUND;

    lvPreviousDescPtr = (Cppi_HostDesc *) (lvCurrentDescPtr->softwareInfo2);
    lvPreviousBufferSize = (lvPreviousDescPtr->buffLen & 0x3fffff);
    lvPreviousBufferPtr = (lvPreviousDescPtr->buffPtr);

    ((Emti_Iterator_s *) (iterator_ptr))->current_level--;
    ((Emti_Iterator_s *) (iterator_ptr))->current_size -= lvPreviousBufferSize;
    ((Emti_Iterator_s *) (iterator_ptr))->current_buf_size = lvPreviousBufferSize;
    ((Emti_Iterator_s *) (iterator_ptr))->current_buf_ptr = lvPreviousBufferPtr;
    ((Emti_Iterator_s *) (iterator_ptr))->current_desc_ptr = (uint32_t) (lvPreviousDescPtr);

    return EM_OK;
}

void *
ti_em_iterator_pointer (
    ti_em_iterator_t * iterator_ptr
)
{
    Cppi_HostDesc *lvDescPtr = (Cppi_HostDesc *) (((Emti_Iterator_s *) (iterator_ptr))->current_desc_ptr);

    if (((Emti_Iterator_s *) (iterator_ptr))->event_hdl == EM_EVENT_UNDEF)
        return NULL;

    Emti_setDirtyFlag (lvDescPtr, 0x1);

    if (((Emti_Iterator_s *) (iterator_ptr))->current_level > ((Emti_Iterator_s *) (iterator_ptr))->coh_level)
        if (Emti_getCohMode (lvDescPtr) == TI_EM_COH_MODE_ON)
            ((Emti_Iterator_s *) (iterator_ptr))->coh_level = ((Emti_Iterator_s *) (iterator_ptr))->current_level;

    return Emti_bufferPtr_ptov (lvDescPtr);
}
