/*
 * Copyright (c) 2012, Texas Instruments Incorporated - http://www.ti.com/
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions are met:
 *       * Redistributions of source code must retain the above copyright
 *         notice, this list of conditions and the following disclaimer.
 *       * Redistributions in binary form must reproduce the above copyright
 *         notice, this list of conditions and the following disclaimer in the
 *         documentation and/or other materials provided with the distribution.
 *       * Neither the name of Texas Instruments Incorporated nor the
 *         names of its contributors may be used to endorse or promote products
 *         derived from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *   DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
 *   DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


/**
 *   @file  event_machine_send.c
 *
 *   @brief   
 *      Open Event Machine Implementation for Nyquist.
 *      This file contains the send related functions.
 *
 */

#include <inttypes.h>
#include <stdio.h>
#include <ti/runtime/openem/event_machine_macros.h>
#include <ti/runtime/openem/event_machine_types.h>
#include <ti/runtime/openem/event_machine_group.h>
#include <ti/runtime/openem/event_machine_hw_types.h>
#include <ti/runtime/openem/src/event_machine_hwpform.h>
#include <ti/runtime/openem/src/event_machine_private_types.h>
#include <ti/runtime/openem/src/event_machine_private_variables.h>
#include <ti/runtime/openem/src/event_machine_private_functions.h>
#include <ti/runtime/openem/event_machine_sw_config.h>
#include <ti/runtime/openem/modules/event_machine_modules.h>

/*
 **********************************************************************************************************************
 **********************************************************************************************************************
 */
em_status_t
em_send (
    em_event_t event,
    em_queue_t queue
)
{

    Cppi_HostDesc *lvHostDescPtr;
    ti_em_device_id_t lvDeviceIdx;
    ti_em_queue_id_t lvQueueIdx;
    uint8_t lvEventMode;
    ti_em_process_id_t lvProcessIdx;
    em_event_group_t lvGroup;

    lvGroup = EM_EVENT_GROUP_UNDEF;

#if (TI_EM_TRACE_LEVEL >=1)
    if (Emti_svTraceFunctionPtr != NULL)
    {
        Emti_svTraceFunctionPtr (TI_EM_TSCOPE_SEND, (em_event_t) (event), (em_queue_t) (queue),
                                 (em_event_group_t) (lvGroup));
    }
#endif /* (TI_EM_TRACE_LEVEL >=1) */

    lvDeviceIdx = (ti_em_device_id_t) Emti_getField (queue, QUEUE_HANDLE_DEVICE_IDX);
    lvProcessIdx = (ti_em_process_id_t) Emti_getField (queue, QUEUE_HANDLE_PROCESS_IDX);
    lvQueueIdx = (ti_em_queue_id_t) Emti_getField (queue, QUEUE_HANDLE_QUEUE_IDX);

#if (EM_CHECK_LEVEL >= 1)
    if ((lvDeviceIdx == Emti_gvTcb.myDeviceIdx) && (lvProcessIdx == Emti_gvTcb.myProcessIdx))
    {
        Emti_QueueSlow_s *ptrQueueSlow;
        Emti_EO_s *ptrEO;

        /*
         * Basic Validations: The queue index has to be in range. 
         */
        if ((lvQueueIdx >= EM_MAX_QUEUES) || (lvQueueIdx == EM_QUEUE_UNDEF))
        {
            return Emti_error ((em_status_t) EM_ERR_BAD_ID, (em_escope_t) EM_ESCOPE_SEND);
        }

        if (!Emti_svPcb.externalCore)
        {
            /*
             * Get the pointer to the queue.
             * Since the queue index is in range, the following pointer is not NULL. 
             */
            ptrQueueSlow = &Emti_svQueueSlowMdb[lvQueueIdx];

            /*
             * Basic Validation: We can send an event to a queue only if it has been mapped 
             */
            if (ptrQueueSlow->state != Emti_QueueState_MAPPED)
            {
                return Emti_error ((em_status_t) EM_ERR_BAD_STATE, (em_escope_t) EM_ESCOPE_SEND);
            }

            /*
             * Get the pointer to the execution object.
             * Since the queue is mapped, the following pointer is not NULL. 
             */
            ptrEO = ptrQueueSlow->eoPtr;

            /*
             * Basic Validation: We can send an event to an execution object provided it has
             * already been started. 
             */
            if (ptrEO != NULL)
            {
                if (ptrEO->state != Emti_EOState_STARTED)
                {
                    return Emti_error ((em_status_t) EM_ERR_BAD_STATE, (em_escope_t) EM_ESCOPE_SEND);
                }
            }
        }

    }
#endif

    /*
     * Debug Message: 
     */
    Emti_logDbg ("Sending event 0x%x descriptor 0x%x\n", event, Emti_publicDesc_vtop (Emti_getCppiHostDescPtr (event)));

    /*
     * If the destination device is the current device    
     */
    if (lvDeviceIdx == Emti_gvTcb.myDeviceIdx)
    {
        uint32_t *lvW32Ptr;
        uint16_t lvRouteInfo;
        uint16_t lvHwQueueIdx;
        uint8_t lvQueueSetIdx;
        uint8_t lvQueueIdxInSet;
        uint8_t lvQueueMode;
        uint8_t lvQueueType;
        uint8_t lvQueueEnabled;

        /*
         * If the destination process is the current process    
         *
         * =>Compute the physical queue num and push event        
         */
        if (lvProcessIdx == Emti_gvTcb.myProcessIdx)
        {
            /*
             * check event is a local event 
             */
            lvEventMode = (uint8_t) Emti_getField (event, EVENT_HDL_EVENT_MODE);
            if (lvEventMode == Emti_EVENT_MODE_LOCAL)
            {
                uint32_t *lvW32Ptr;
                uint16_t lvHwQueueIdx;

                /*
                 * set eventDscPtr = eventHdl.dscPtr 
                 */
                lvHostDescPtr = (Cppi_HostDesc *) Emti_getCppiHostDescPtr (event);

                /*
                 * write Emti_EVENT_MODE_GLOBAL_HW to eventDscPtr->eventMode 
                 */
                Emti_setEventMode (lvHostDescPtr, Emti_EVENT_MODE_GLOBAL_HW);

                /*
                 * write 0xFF to eventDscPtr->srcTagLo 
                 */
                lvW32Ptr = (uint32_t *) (&(lvHostDescPtr->tagInfo));
                Emti_setField (*lvW32Ptr, EVENT_DSC_W1_SRC_TAG_LO, 0xff);

                /*
                 * write queueIdx to eventDscPtr->queueIdx 
                 */
                lvW32Ptr = (uint32_t *) (&(lvHostDescPtr->timeStamp));
                Emti_setField (*lvW32Ptr, EVENT_DSC_W8_QUEUE_IDX, lvQueueIdx);

                /*
                 * write groupHdl to eventDscPtr->eventGroupIdx 
                 */
                lvW32Ptr = (uint32_t *) (&(lvHostDescPtr->softwareInfo0));
                *lvW32Ptr = (uint32_t) ((lvGroup & Emti_EVENT_GROUP_ID_MASK) | Emti_NO_EVENT_GROUP_TAG);

                /*
                 * set eventHdl |= Emti_DSC_HINT_CH 
                 */
                lvHostDescPtr = (Cppi_HostDesc *) (((uint32_t) lvHostDescPtr & 0xFFFFFFF0) + Emti_DSC_HINT_CH);

                /*
                 * set hwQueueIdx = <PCB.chain>txQueueBaseIdx + Emti_hash(queueIdx) 
                 */
                lvHwQueueIdx = Emti_svPcb.chainDsc.txQueueBaseIdx + Emti_hash (lvQueueIdx);

                /*
                 * if Emti_wbinv(event_hdl) == 0
                 * raise memory fence 
                 */
                if (Emti_wbinv (event) == 0)
                {
                    Emti_mfence ();
                }

                /*
                 * push event to HW queue 
                 */
                Emti_pushEvent (lvHwQueueIdx, (uint32_t *) lvHostDescPtr, TI_EM_MEM_PUBLIC_DESC);

                return (em_status_t) EM_OK;
            }

            if (lvQueueIdx < (EM_QUEUE_STATIC_MAX + 1))
            {
                /*
                 * If static queue  
                 */
                lvRouteInfo = Emti_svPsbPtr->staticQueueRouteTbl[lvQueueIdx].routeInfo;
                lvQueueIdxInSet = 0;
                lvQueueMode = (uint8_t) Emti_getField (lvRouteInfo, QUEUE_ROUTE_INFO_QUEUE_MODE);
            }
            else
            {
                /*
                 * If not static queue  
                 */
                lvQueueSetIdx = (uint8_t) Emti_getField (queue, QUEUE_HANDLE_QUEUE_SET_IDX);
                lvQueueIdxInSet = (uint8_t) Emti_getField (queue, QUEUE_HANDLE_QUEUE_IDX_IN_SET);
                lvRouteInfo = Emti_svPsbPtr->queueRouteTbl[lvQueueSetIdx].routeInfo;
                lvQueueMode = (uint8_t) Emti_getField (lvRouteInfo, QUEUE_ROUTE_INFO_QUEUE_MODE);
                if (lvQueueMode == TI_EM_QUEUE_MODE_HW)
                    lvQueueIdxInSet -= TI_EM_STATIC_QUEUE_NUM_IN_SET;
            }

            lvQueueEnabled = (uint8_t) Emti_getField (lvRouteInfo, QUEUE_ROUTE_INFO_ROUTE_ENABLED);
            if (lvQueueEnabled == 0)
            {
                return EM_ERROR;
            }

            lvHostDescPtr = (Cppi_HostDesc *) Emti_getCppiHostDescPtr (event);

            if (lvQueueMode == TI_EM_QUEUE_MODE_SD)
            {
                lvHwQueueIdx = Emti_gvTcb.sdQueueBaseIdx + Emti_getField (lvRouteInfo, QUEUE_ROUTE_INFO_SD_QUEUE_IDX);
                lvQueueType = (uint8_t) Emti_getField (lvRouteInfo, QUEUE_ROUTE_INFO_QUEUE_TYPE);
                /*
                 * Filling the word 8 of the event descriptor       
                 */
                lvW32Ptr = (uint32_t *) (&(lvHostDescPtr->timeStamp));
                Emti_setField (*lvW32Ptr, EVENT_DSC_W8_QUEUE_IDX, lvQueueIdx);
                Emti_setField (*lvW32Ptr, EVENT_DSC_W8_QUEUE_TYPE, lvQueueType);
                /*
                 * Filling word 9 of the event descriptor   
                 */
                lvW32Ptr++;
                *lvW32Ptr = (uint32_t) ((lvGroup & Emti_EVENT_GROUP_ID_MASK) | Emti_NO_EVENT_GROUP_TAG);
            }

            if (lvQueueMode == TI_EM_QUEUE_MODE_HW)
            {
                uint32_t lvFreeInfo;
                uint16_t lvEventHandleType;
                Emti_event_mode_t lvGlobalEventMode;
                /*
                 * Event with an event mode cannot be sent to an HW queue. 
                 */
                lvFreeInfo = Emti_getFreeInfo (Emti_getCppiHostDescPtr (event));
                lvEventHandleType = (uint16_t) (lvFreeInfo >> 24) & 0xf;
                lvGlobalEventMode = (Emti_event_mode_t) (lvEventHandleType & Emti_EVENT_HDL_EVENT_MODE_MASK);
                if (lvGlobalEventMode == Emti_EVENT_MODE_GLOBAL_SW)
                {
                    return EM_ERROR;
                }

                lvHwQueueIdx = Emti_getField (lvRouteInfo, QUEUE_ROUTE_INFO_HW_QUEUE_IDX) + lvQueueIdxInSet;
            }
            
            /* check HwQueueIdx == PA Tx queue Idx then fill ps words and send to PA*/
            if (lvHwQueueIdx == Emti_PA_TX_QUEUE_IDX)
            {
                /* get RIO chain header pointer */
                Emti_PaHdrRioFormat_s *lvPaHdrPtr;
                uint32_t * lvUint32PaHdrPtr;
                
                lvPaHdrPtr = (Emti_PaHdrRioFormat_s *) (&(lvHostDescPtr->psData));
                lvUint32PaHdrPtr = (uint32_t*)(&(lvHostDescPtr->psData));
                
                /* clear RIO chain header */
                *lvUint32PaHdrPtr++ = 0;
                *lvUint32PaHdrPtr = 0;
            
                /* fill RIO chain header */
                Emti_setField (lvPaHdrPtr->dstInfo, RIO_CHAIN_HDR_DST_INFO_PROCESS_IDX, lvProcessIdx);
                Emti_setField (lvPaHdrPtr->dstInfo, RIO_CHAIN_HDR_DST_INFO_QUEUE_IDX, lvQueueIdx);

                /*
                 * set PS data Length (8 Bytes) 
                 */
                lvHostDescPtr->packetInfo &= 0xF0FFFFFF;
                lvHostDescPtr->packetInfo |= 0x02000000;
                
                /*
                 * set Packet Type to 30 (Type9)    
                 */
                lvHostDescPtr->descInfo &= ~Emti_EVENT_SIZE_EVENT_TAG_MASK;
                lvHostDescPtr->descInfo |= (Emti_EVENT_TAG_RIO << Emti_EVENT_SIZE_EVENT_TAG_SHIFT);
            }


            /*
             * Do the cache coherency 
             */
            if (Emti_wbinv (event) == 0)
            {
                Emti_mfence ();
            }

            /*
             * Set descriptor size to Emti_DSC_HINT_MAX  
             */
            lvHostDescPtr = (Cppi_HostDesc *) (((uint32_t) lvHostDescPtr & 0xFFFFFFF0) + Emti_DSC_HINT_MAX);

            /*
             * Push Event to hw queue   
             */
            Emti_pushEvent (lvHwQueueIdx, (uint32_t *) lvHostDescPtr, TI_EM_MEM_PUBLIC_DESC);
        }                       /* EndOf if(lvProcessIdx == Emti_gvTcb.myProcessIdx)    */
        else
        {                       /* The destination device is the current device but the process is different */
            uint32_t lvFreeInfo;
            uint16_t lvEventHandleType;
            Emti_event_mode_t lvGlobalEventMode;

            /*
             * Check if the route was enabled before.
             */
            if (Emti_getField (Emti_svPsbPtr->processRouteTbl[lvProcessIdx].routeInfo, PROCESS_ROUTE_INFO_ROUTE_ENABLED)
                == 0)
            {
                return EM_ERROR;
            }

            /*
             * Event with an event mode cannot be sent to another process.
             */
            lvFreeInfo = Emti_getFreeInfo (Emti_getCppiHostDescPtr (event));
            lvEventHandleType = (uint16_t) (lvFreeInfo >> 24) & 0xf;
            lvGlobalEventMode = (Emti_event_mode_t) (lvEventHandleType & Emti_EVENT_HDL_EVENT_MODE_MASK);
            if (lvGlobalEventMode == Emti_EVENT_MODE_GLOBAL_SW)
            {
                return EM_ERROR;
            }

            lvHostDescPtr = (Cppi_HostDesc *) Emti_getCppiHostDescPtr (event);

            /*
             * write Emti_EVENT_MODE_GLOBAL_HW to eventDscPtr->eventMode 
             */
            Emti_setEventMode (lvHostDescPtr, Emti_EVENT_MODE_GLOBAL_HW);

            /*
             * write 0xFF to eventDscPtr->srcTagLo 
             */
            lvW32Ptr = (uint32_t *) (&(lvHostDescPtr->tagInfo));
            Emti_setField (*lvW32Ptr, EVENT_DSC_W1_SRC_TAG_LO, 0xff);

            /*
             * Update Word8 of the event descriptor, (only the queue idx field) 
             */
            lvW32Ptr = (uint32_t *) (&(lvHostDescPtr->timeStamp));
            Emti_setField (*lvW32Ptr, EVENT_DSC_W8_QUEUE_IDX, lvQueueIdx);

            /*
             * set hwQueueIdx = <PCB.chain>txQueueBaseIdx
             */
            lvHwQueueIdx =
                Emti_getField (Emti_svPsbPtr->processRouteTbl[lvProcessIdx].routeInfo, PROCESS_ROUTE_INFO_TX_QUEUE_IDX);

            /*
             * Do the cache coherency 
             */
            if (Emti_wbinv (event) == 0)
            {
                Emti_mfence ();
            }

            /*
             * Set descriptor size to Emti_DSC_HINT_CH  
             */
            lvHostDescPtr = (Cppi_HostDesc *) (((uint32_t) lvHostDescPtr & 0xFFFFFFF0) + Emti_DSC_HINT_CH);

            /*
             * Push Event to hw queue   
             */
            Emti_pushEvent (lvHwQueueIdx, (uint32_t *) lvHostDescPtr, TI_EM_MEM_PUBLIC_DESC);
        }                       /* EndOf else    */
    }                           /* EndOf if(lvDeviceIdx == Emti_gvTcb.myDeviceIdx)    */
    else
    {                           /* The destination device is a distant device   */
        uint32_t lvFreeInfo;
        uint16_t lvRouteInfo;
        uint16_t lvEventHandleType;
        uint16_t lvRouteType;
        uint8_t lvEnabled;
        em_status_t status;
        Emti_event_mode_t lvGlobalEventMode;

        /*
         * Get the routing info from the device routing table
         */
        lvRouteInfo = Emti_svPsbPtr->deviceRouteTbl[lvDeviceIdx].u.routeInfo;
        lvEnabled = (uint8_t) Emti_getField (lvRouteInfo, DEVICE_ROUTE_INFO_ENABLED);
        if (lvEnabled == 0)
        {
            return EM_ERROR;
        }
        /*
         * Event with an event mode cannot be sent to another device. 
         */
        lvFreeInfo = Emti_getFreeInfo (Emti_getCppiHostDescPtr (event));
        lvEventHandleType = (uint16_t) (lvFreeInfo >> 24) & 0xf;
        lvGlobalEventMode = (Emti_event_mode_t) (lvEventHandleType & Emti_EVENT_HDL_EVENT_MODE_MASK);
        if (lvGlobalEventMode == Emti_EVENT_MODE_GLOBAL_SW)
        {
            return EM_ERROR;
        }

        /*
         * Call the chaining send method
         */
        lvRouteType = (uint16_t) Emti_getField (lvRouteInfo, DEVICE_ROUTE_INFO_ROUTE_TYPE);
        status = Emti_moduleChainSend (lvRouteType, event, queue);
        return status;
    }

    return (em_status_t) EM_OK;
}

em_status_t
em_send_group (
    em_event_t event,
    em_queue_t queue,
    em_event_group_t group
)
{
    Cppi_HostDesc *lvHostDescPtr;
    ti_em_device_id_t lvDeviceIdx;
    ti_em_queue_id_t lvQueueIdx;
    uint8_t lvEventMode;
    ti_em_process_id_t lvProcessIdx;

    lvDeviceIdx = (ti_em_device_id_t) Emti_getField (queue, QUEUE_HANDLE_DEVICE_IDX);
    lvProcessIdx = (ti_em_process_id_t) Emti_getField (queue, QUEUE_HANDLE_PROCESS_IDX);
    lvQueueIdx = (ti_em_queue_id_t) Emti_getField (queue, QUEUE_HANDLE_QUEUE_IDX);

#if (EM_CHECK_LEVEL >= 1)
    if ((lvDeviceIdx == Emti_gvTcb.myDeviceIdx) && (lvProcessIdx == Emti_gvTcb.myProcessIdx))
    {
        Emti_QueueSlow_s *ptrQueueSlow;
        Emti_EO_s *ptrEO;
        uint16_t queueIdx = Emti_getQueueIndex (queue);

        /*
         * Basic Validations: The queue index has to be in range. 
         */
        if (queueIdx >= EM_MAX_QUEUES)
        {
            return Emti_error ((em_status_t) EM_ERR_BAD_ID, (em_escope_t) EM_ESCOPE_SEND);
        }

        /*
         * Get the pointer to the queue. 
         * * Since the queue index is in range, the following pointer is not NULL. 
         */
        ptrQueueSlow = &Emti_svQueueSlowMdb[queueIdx];

        /*
         * Basic Validation: We can send an event to a queue only if it has been mapped 
         */
        if (ptrQueueSlow->state != Emti_QueueState_MAPPED)
        {
            return Emti_error ((em_status_t) EM_ERR_BAD_STATE, (em_escope_t) EM_ESCOPE_SEND);
        }

        /*
         * Get the pointer to the execution object. 
         * * Since the queue is mapped, the following pointer is not NULL. 
         */
        ptrEO = ptrQueueSlow->eoPtr;

        /*
         * Basic Validation: We can send an event to an execution object provided it has
         * * already been started. 
         */
        if (ptrEO->state != Emti_EOState_STARTED)
        {
            return Emti_error ((em_status_t) EM_ERR_BAD_STATE, (em_escope_t) EM_ESCOPE_SEND);
        }

    }
#endif

    /*
     * Debug Message: 
     */
    Emti_logDbg ("Sending event 0x%p Descriptor 0x%p\n", (void *) event, Emti_getCppiHostDescPtr (event));

    /*
     * If the destination device is the current device    
     */
    if (lvDeviceIdx == Emti_gvTcb.myDeviceIdx)
    {
        uint32_t *lvW32Ptr;
        uint16_t lvHwQueueIdx;
        uint16_t lvRouteInfo;
        uint8_t lvQueueSetIdx;
        uint8_t lvQueueIdxInSet;
        uint8_t lvQueueMode;
        uint8_t lvQueueType;
        uint8_t lvQueueEnabled;
        /*
         * If the destination process is the current process    
         */
        if (lvProcessIdx == Emti_gvTcb.myProcessIdx)
        {
            /*
             * check event is a local event 
             */
            lvEventMode = (uint8_t) Emti_getField (event, EVENT_HDL_EVENT_MODE);
            if (lvEventMode == Emti_EVENT_MODE_LOCAL)
            {
                uint16_t lvRouteInfo;
                uint8_t lvEnabled;
                uint32_t *lvW32Ptr;
                uint16_t lvHwQueueIdx;

                /*
                 * set eventDscPtr = eventHdl.dscPtr 
                 */
                lvHostDescPtr = (Cppi_HostDesc *) Emti_getCppiHostDescPtr (event);

                /*
                 * select route = <PSB.chain>processRouteTbl[<TCB>myProcessIdx] 
                 */
                lvRouteInfo = Emti_svPsbPtr->processRouteTbl[Emti_gvTcb.myProcessIdx].routeInfo;

                /*
                 * if route.routeEnabled == 0
                 * return error 
                 */
                lvEnabled = (uint8_t) Emti_getField (lvRouteInfo, PROCESS_ROUTE_INFO_ROUTE_ENABLED);
                if (lvEnabled == 0)
                {
                    return EM_ERROR;
                }

                /*
                 * write Emti_EVENT_MODE_GLOBAL_HW to eventDscPtr->eventMode 
                 */
                Emti_setEventMode (lvHostDescPtr, Emti_EVENT_MODE_GLOBAL_HW);

                /*
                 * write 0xFF to eventDscPtr->srcTagLo 
                 */
                lvW32Ptr = (uint32_t *) (&(lvHostDescPtr->tagInfo));
                Emti_setField (*lvW32Ptr, EVENT_DSC_W1_SRC_TAG_LO, 0xff);

                /*
                 * write queueIdx to eventDscPtr->queueIdx 
                 */
                lvW32Ptr = (uint32_t *) (&(lvHostDescPtr->timeStamp));
                Emti_setField (*lvW32Ptr, EVENT_DSC_W8_QUEUE_IDX, lvQueueIdx);

                /*
                 * write groupHdl to eventDscPtr->eventGroupIdx 
                 */
                lvW32Ptr = (uint32_t *) (&(lvHostDescPtr->softwareInfo0));
                *lvW32Ptr = (uint32_t) ((group & Emti_EVENT_GROUP_ID_MASK) | Emti_NO_EVENT_GROUP_TAG);


                /*
                 * set eventHdl |= Emti_DSC_HINT_CH 
                 */
                lvHostDescPtr = (Cppi_HostDesc *) (((uint32_t) lvHostDescPtr & 0xFFFFFFF0) + Emti_DSC_HINT_CH);

                /*
                 * set hwQueueIdx = <PCB.chain>txQueueBaseIdx + Emti_hash(queueIdx) 
                 */
                lvHwQueueIdx = Emti_svPcb.chainDsc.txQueueBaseIdx + Emti_hash (lvQueueIdx);

                /*
                 * if Emti_wbinv(event_hdl) == 0
                 * raise memory fence 
                 */
                if (Emti_wbinv (event) == 0)
                {
                    Emti_mfence ();
                }

                /*
                 * push event to HW queue 
                 */
                Emti_pushEvent (lvHwQueueIdx, (uint32_t *) lvHostDescPtr, TI_EM_MEM_PUBLIC_DESC);

                return (em_status_t) EM_OK;
            }

            lvQueueSetIdx = (uint8_t) Emti_getField (queue, QUEUE_HANDLE_QUEUE_SET_IDX);
            lvQueueIdxInSet = (uint8_t) Emti_getField (queue, QUEUE_HANDLE_QUEUE_IDX_IN_SET);
            /*
             * If static queue  
             */
            if (lvQueueIdx < TI_EM_STATIC_QUEUE_NUM_IN_SET)
            {
                lvRouteInfo = Emti_svPsbPtr->staticQueueRouteTbl[lvQueueIdx].routeInfo;
                lvQueueIdxInSet = 0;
            }
            /*
             * If not static queue  
             */
            else
            {
                lvRouteInfo = Emti_svPsbPtr->queueRouteTbl[lvQueueSetIdx].routeInfo;
            }
            lvQueueEnabled = (uint8_t) Emti_getField (lvRouteInfo, QUEUE_ROUTE_INFO_ROUTE_ENABLED);
            if (lvQueueEnabled == 0)
            {
                return EM_ERROR;
            }
            lvQueueMode = (uint8_t) Emti_getField (lvRouteInfo, QUEUE_ROUTE_INFO_QUEUE_MODE);
            lvHostDescPtr = (Cppi_HostDesc *) Emti_getCppiHostDescPtr (event);
            if (lvQueueMode == TI_EM_QUEUE_MODE_SD)
            {
                lvHwQueueIdx = Emti_gvTcb.sdQueueBaseIdx + Emti_getField (lvRouteInfo, QUEUE_ROUTE_INFO_SD_QUEUE_IDX);
                lvQueueType = (uint8_t) Emti_getField (lvRouteInfo, QUEUE_ROUTE_INFO_QUEUE_TYPE);
                /*
                 * Filling the word 8 of the event descriptor       
                 */
                lvW32Ptr = (uint32_t *) (&(lvHostDescPtr->timeStamp));
                Emti_setField (*lvW32Ptr, EVENT_DSC_W8_QUEUE_IDX, lvQueueIdx);
                Emti_setField (*lvW32Ptr, EVENT_DSC_W8_QUEUE_TYPE, lvQueueType);
                /*
                 * Filling word 9 of the event descriptor   
                 */
                lvW32Ptr++;
                *lvW32Ptr = (uint32_t) ((group & Emti_EVENT_GROUP_ID_MASK) | Emti_NO_EVENT_GROUP_TAG);
            }
            if (lvQueueMode == TI_EM_QUEUE_MODE_HW)
            {
                lvHwQueueIdx = Emti_getField (lvRouteInfo, QUEUE_ROUTE_INFO_HW_QUEUE_IDX) + lvQueueIdxInSet;
            }

            /*
             * Do the cache coherency 
             */
            if (Emti_wbinv (event) == 0)
            {
                Emti_mfence ();
            }
            /*
             * Push Event to hw queue   
             */
            /*
             * Set descriptor size to Emti_DSC_HINT_MAX  
             */
            lvHostDescPtr = (Cppi_HostDesc *) (((uint32_t) lvHostDescPtr & 0xFFFFFFF0) + Emti_DSC_HINT_MAX);
            Emti_pushEvent (lvHwQueueIdx, (uint32_t *) lvHostDescPtr, TI_EM_MEM_PUBLIC_DESC);

        }                       /* EndOf if(lvProcessIdx == Emti_gvTcb.myProcessIdx)    */
        else
        {                       /* The destination device is the current device but the process is different */
            return EM_ERROR;
        }                       /* EndOf else    */
    }                           /* EndOf if(lvDeviceIdx == Emti_gvTcb.myDeviceIdx)    */
    else
    {                           /* The destination device is a distant device   */
        return EM_ERROR;
    }                           /* EndOf if(lvDeviceIdx != Emti_gvTcb.myDeviceIdx)    (else)    */
    return (em_status_t) EM_OK;
}

em_event_t
ti_em_receive (
    em_queue_t queue
)
{
    ti_em_device_id_t lvDeviceIdx;
    ti_em_queue_id_t lvQueueIdx;
    ti_em_process_id_t lvProcessIdx;
    uint16_t lvRouteInfo;
    uint16_t lvHwQueueIdx;
    uint8_t lvQueueSetIdx;
    uint8_t lvQueueIdxInSet;
    uint8_t lvQueueMode;
    uint8_t lvQueueEnabled;

    lvDeviceIdx = (ti_em_device_id_t) Emti_getField (queue, QUEUE_HANDLE_DEVICE_IDX);
    lvProcessIdx = (ti_em_process_id_t) Emti_getField (queue, QUEUE_HANDLE_PROCESS_IDX);
    lvQueueIdx = (ti_em_queue_id_t) Emti_getField (queue, QUEUE_HANDLE_QUEUE_IDX);

    if ((lvDeviceIdx != Emti_gvTcb.myDeviceIdx) || (lvProcessIdx != Emti_gvTcb.myProcessIdx))
    {
        return (em_event_t) EM_EVENT_UNDEF;
    }
    /*
     * If static queue  
     */
    if (lvQueueIdx < (EM_QUEUE_STATIC_MAX + 1))
    {
        lvRouteInfo = Emti_svPsbPtr->staticQueueRouteTbl[lvQueueIdx].routeInfo;
        lvQueueIdxInSet = 0;
        lvQueueMode = (uint8_t) Emti_getField (lvRouteInfo, QUEUE_ROUTE_INFO_QUEUE_MODE);
    }
    /*
     * If not static queue  
     */
    else
    {
        lvQueueSetIdx = (uint8_t) Emti_getField (queue, QUEUE_HANDLE_QUEUE_SET_IDX);
        lvQueueIdxInSet = (uint8_t) Emti_getField (queue, QUEUE_HANDLE_QUEUE_IDX_IN_SET);
        lvRouteInfo = Emti_svPsbPtr->queueRouteTbl[lvQueueSetIdx].routeInfo;
        lvQueueMode = (uint8_t) Emti_getField (lvRouteInfo, QUEUE_ROUTE_INFO_QUEUE_MODE);
        if (lvQueueMode == TI_EM_QUEUE_MODE_HW)
            lvQueueIdxInSet -= TI_EM_STATIC_QUEUE_NUM_IN_SET;
    }
    lvQueueEnabled = (uint8_t) Emti_getField (lvRouteInfo, QUEUE_ROUTE_INFO_ROUTE_ENABLED);
    if (lvQueueEnabled == 0)
    {
        return (em_event_t) EM_EVENT_UNDEF;
    }
    if (lvQueueMode == TI_EM_QUEUE_MODE_SD)
    {
        return (em_event_t) EM_EVENT_UNDEF;
    }
    if (lvQueueMode == TI_EM_QUEUE_MODE_HW)
    {
        lvHwQueueIdx = Emti_getField (lvRouteInfo, QUEUE_ROUTE_INFO_HW_QUEUE_IDX) + lvQueueIdxInSet;
    }
    /*
     * Flush the prefetch buffer 
     */
    Emti_prefetchBufferInv ();

    return (em_event_t) Emti_popEvent (lvHwQueueIdx, TI_EM_MEM_PUBLIC_DESC);
}
